SystemOrganization addCategory: #'Qwaq-Client'!
SystemOrganization addCategory: #'Qwaq-Client-Exceptions'!
SystemOrganization addCategory: #'Qwaq-Client-Meeting'!
SystemOrganization addCategory: #'Qwaq-Client-Platform'!
SystemOrganization addCategory: #'Qwaq-Client-Undo'!
SystemOrganization addCategory: #'Qwaq-Client-Plugin'!

TTaskMain subclass: #QMenuTask
	instanceVariableNames: 'menu'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QMenuTask commentStamp: '<historical>' prior: 0!
A task to handle events while an avatar is carrying a menu.
Prevents events from going into the space until the menu is used or dismissed.

Note that this conspires in some detail with QwaqParticipantUI.
See QPUI
	avatarMenuPosted:itemData:
	fireTMenu:item:

Triggers are set up in QPUI #avatarMenuPosted:itemData:,
and here in menu:,  to ensure that the CMenu and the TMenu have the same lifetime.
That's a bit tricky.  Usually  we want the tmenu to induce dismantling the cmenu
(after it fires, or is destroyed inspace for whatever reason.)
The participant sets up #destroyed handling on the TMenu to honor that.

But one case has to go the other way... if the user dismisses the cmenu
by clicking outside  it, we need to honor that on-island by disposing the t-menu.
So there are #playerClosed triggers attached to the cmenu, 
and if we get those before the participant has told us to ignore them
(which it does by setting my menu: back to nil, in fireTMenu:item:)
then  we tell the avatar-replica to take away the underlying tmenu.


!

!QMenuTask class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/15/2009 00:13'!
cancel
	"Abandon menu mode."
	menu ifNotNil: [ 
		menu close.		"Letting the closed-handlers do their stuff."
	] ifNil: [
		self popSelf.
	].! !

!QMenuTask methodsFor: 'controls' stamp: 'bgf 10/4/2007 10:06'!
cursor

	^ nil

	"This irritates some folk...
		^ CCursorLibrary leftArrowCursor
	"! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 10/31/2006 14:48'!
doKeyDown: event

	^ false! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/15/2009 00:13'!
doKeyStroke: event
	" If we get a background stroke while a menu is up..."
	event keyCharacter = (Character escape)  ifTrue: [
		self cancel
	].! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 10/31/2006 14:49'!
doKeyUp: event

	^ false! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/14/2009 22:10'!
doPointerDown: event
	"Any pointer down event outside the menu (into the space) is ignored.
	 We use a CMenu to run the tmenu, and its lifecycle governs."
	^ false! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/14/2009 22:11'!
doPointerMove: event
	"Any pointer down event outside the menu (into the space) is ignored.
	 We use a CMenu to run the tmenu, and its lifecycle governs."
	^ false! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/14/2009 22:11'!
doPointerUp: event
	"Any pointer down event outside the menu (into the space) is ignored.
	 We use a CMenu to run the tmenu, and its lifecycle governs."
	^ false! !

!QMenuTask methodsFor: 'accessing' stamp: 'bgf 11/14/2009 23:16'!
menu: aCMenu
	"The menu task will pop itself when this menu is closed.
	 (Unless superseded by a subsequent menu: )"
	menu := aCMenu.
	self stopScript: #signalClosed.
	menu ifNotNil: [
		self startScript: #signalClosed when: { menu . #playerClosed }.
	].! !

!QMenuTask methodsFor: 'controls' stamp: 'bgf 10/31/2006 19:24'!
menuMode
	"Indicate to the harness that menuing is in action."
	^ true! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/15/2009 00:10'!
popSelf
	 (view task = self ) ifTrue: [
		"We only tell the replica to discard via cancelMenu  if we have a cmenu.
	 	  The participant clears this menu before future-invoking a tmenu operation.
		  This is how we distinguish
			cmenu playerClosed because the in-space menu destroyed itself
			cmenu playerClosed because the user clicked outside it to dismiss it.
		 For the first, we want to cancelMenu, for the second, we do not. "

		( menu notNil and: [ view readyForEvents ] )  ifTrue: [
			view avatar replica future cancelMenu.
			menu := nil.
		].
		view popTask.
	].! !

!QMenuTask methodsFor: 'events' stamp: 'bgf 11/15/2009 00:26'!
signalClosed
	"Indirect signal response to the tweak-menu playerClosed.
	 Indirect, to avoid weird tweak asynchronicities...
	 Tweak closes menus before firing the item.
	 (But all in the same step-cycle, the problem is with ordering, not delays.)
	 If we do an avatar cancelMenu before the menu gets the future-fire,
	 then nothing happens.   But we have to be able to cancelMenu from here;
	 otherwise the inspace side can (if we just click outside the tweak menu)
	 be stuck with a lingering TMenu that leaves the avatar poised to get actions for it."

	self waitTick.
	self startScript: #popSelf.
! !

CFormCache subclass: #QTextureFormCache
	instanceVariableNames: 'material'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QTextureFormCache commentStamp: '<historical>' prior: 0!
A Tweak form cache which also holds onto the material for easier processing by the client.!

!QTextureFormCache class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextureFormCache methodsFor: 'invalidating' stamp: 'bgf 7/25/2008 13:51'!
invalidate: aRect

	"Some scrolled-text + font emphasis recompositing produces invalid rects
	 that, if passed to some OGL drivers, cause un-amusing results."

	aRect top > aRect bottom ifTrue: [ ^self ].
	aRect right < aRect left ifTrue: [ ^self ].

	super invalidate: aRect.! !

!QTextureFormCache methodsFor: 'accessing' stamp: 'ar 4/7/2007 16:37'!
material
	"Answer the material associated with this cache"
	^material! !

!QTextureFormCache methodsFor: 'accessing' stamp: 'ar 4/7/2007 16:37'!
material: aTMaterial
	"Set the material associated with this cache"
	material := aTMaterial! !

TTask subclass: #QLookTask
	instanceVariableNames: 'avatar player keySymbols'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QLookTask commentStamp: '<historical>' prior: 0!
The look-task, when atop the task stack of a the mainView,
drives the avatar camera according to the pointer position
(a look-only variation on mouse-joystick control, where you steer your camera only.) 

We do look-mode as a task controlling the focal transform of the camera,
rather than using the TAvatarDriver avatar-pivot lookModeMouseDriving:,
so that look-mode does not involve engaging the driving machinery
(so it doesn't affect posture, or  your current Place, or focal input modes.

Look mode resets to forward orientation when the camera is zoomed,
as otherwise camera-steering leads to arbitary cam placements anywhere in space,
with no obvious path back to natural... so we keep your camera behind your head,
and swing it back to forward when you zoom in or out.
!

!QLookTask class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QLookTask methodsFor: 'private' stamp: 'bgf 7/14/2009 00:46'!
addPYR: v3
	"Worker method for camera turning. "
	| xform |

	xform := self curTransform clone.
	xform pitchYawRoll: (xform pitchYawRoll + v3).

	avatar onFocalTransformChanged: xform.
! !

!QLookTask methodsFor: 'events' stamp: 'bgf 7/26/2009 16:46'!
cancelLook
	"End look-override, returning to the avatar's desired camera view"

	avatar ifNotNil: [ avatar driver fullClutch ].

	"Experiment... just leave it until we start driving again."
	false ifTrue: [
		avatar ifNotNil: [
			avatar readyForEvents ifTrue: [
				avatar onFocalTransformChanged: (
						avatar replica get: #focalCameraTransform).
			] ifFalse: [
				avatar onFocalTransformChanged: (avatar localTransform clone)
			]]] 

! !

!QLookTask methodsFor: 'private' stamp: 'bgf 5/8/2009 02:33'!
curTransform
	"The current camera transform, basis for the next rotation step.
	 Starts from where-ever the camera is."
	^ avatar focalCameraTransform ifNil: [ 
				avatar localTransform].
					! !

!QLookTask methodsFor: 'controls' stamp: 'bgf 5/7/2009 15:16'!
cursor
	^ Cursor crossHair		"Maybe should be an eye... but something to distinguish from drive"! !

!QLookTask methodsFor: 'events' stamp: 'bgf 5/8/2009 01:07'!
doPointerDown: event
	"No interactors, just make sure the indicative cursor stays visible."
	self doPointerMove: event.
	! !

!QLookTask methodsFor: 'events' stamp: 'bgf 5/8/2009 01:07'!
doPointerEnter: event
	"No interactors, just make sure the indicative cursor stays visible."
	self doPointerOver: event! !

!QLookTask methodsFor: 'events' stamp: 'bgf 5/8/2009 01:07'!
doPointerMove: event
	"No interactors, just make sure the indicative cursor stays visible."
	player mouseCursor: (self cursor).! !

!QLookTask methodsFor: 'events' stamp: 'bgf 5/8/2009 01:07'!
doPointerUp: event
	"No interactors, just make sure the indicative cursor stays visible."
	player mouseCursor: (self cursor).! !

!QLookTask methodsFor: 'events' stamp: 'bgf 8/7/2009 00:49'!
keyboardLookEvent: evt
	| sym |
	"The basic behavior here is to reinterpret the keyboard-driving symbols as camera moves,
	 so we start by finding the applicable keyboard-driving symbol."

	evt isKeystroke ifTrue: [
		sym := view keyStrokeSymbolFor: evt.
	] ifFalse: [
		sym := view keySymbolFor: evt fromKeyUp: (evt type = #keyUp).
	].
	
	"Toggling out of sitting should be available while sitting."
	(evt isKeystroke and: [ evt keyCharacter = $t ])  ifTrue: [
		view owner ifAllowedToMove: [ view avatar replica future togglePosture ].
		^ true 
	].
	sym ifNil: [ ^ false ].

	"These aren't steering keys, they're broader commands that should be handled outside..."
	(#(escape resetPosition) includes: sym) ifTrue: [ ^ false ].

	"We do not mix turning with any other steering, because zooming resets direction to forward.
	 The first check notices incoming zooms when we're already turning...s top turning."
	( #(scrollUp scrollDown) includes: sym) ifTrue: [ 
		keySymbols := keySymbols intersection: #(moveForward moveBack)
	].

	sym caseOf: {
		"These have to be stroke-based rather than step-repeated."
		[#scrollUp ] -> [ avatar zoomIn ].			"Mousewheel steps synthetic."
		[#scrollDown ] -> [ avatar zoomOut ].		
		[#resetLook] -> [ self resetLook ].			"Reset is an action."
	} otherwise: [ 
		"Others become part of the keySymbols we implement in step."
		(evt isKeyDown or:[ evt isKeystroke ]) ifTrue: [
			keySymbols add: sym
		] ifFalse: [ 
			keySymbols remove: sym ifAbsent: [ ].
		]].

	"See above... this discards incoming turns while already zooming."
	(keySymbols includesAnyOf: #(moveForward moveBack))  ifTrue: [
		 keySymbols := keySymbols intersection: #(moveForward moveBack)
	].

	"If we removed the last key we're not in look mode any more..."
	keySymbols ifEmpty: [ view lookMode: false ].
	^ true


! !

!QLookTask methodsFor: 'key-driving' stamp: 'bgf 7/26/2009 16:42'!
lookDown
	self addPYR: (Vector3 x: (-2 * avatar speed) y: 0 z: 0)! !

!QLookTask methodsFor: 'key-driving' stamp: 'bgf 7/26/2009 16:42'!
lookUp
	self addPYR: (Vector3 x: (2 * avatar speed) y: 0 z: 0)
! !

!QLookTask methodsFor: 'key-driving' stamp: 'bgf 7/26/2009 16:42'!
resetLook

	view readyForEvents ifTrue: [
		avatar replica future updateFocalTransform.
		avatar driver fullClutch.
	]! !

!QLookTask methodsFor: 'key-driving' stamp: 'bgf 7/26/2009 16:41'!
rotateLeft
	"Turn a little..."
	self addPYR: (Vector3 x: 0 y: (3 * avatar speed) z: 0)
! !

!QLookTask methodsFor: 'key-driving' stamp: 'bgf 7/26/2009 16:41'!
rotateRight
	"Turn a little..."
	self addPYR: (Vector3 x: 0 y: (-3 * avatar speed) z: 0)
! !

!QLookTask methodsFor: 'events' stamp: 'bgf 7/26/2009 17:02'!
runSymbols
	"Run the look-mode driving symbols, a subset of normal driving symbols."
	keySymbols do: [ : sym |
		sym caseOf: {
			[#rotateLeft ] -> [ self rotateLeft ] .
			[#rotateRight] -> [ self rotateRight ] .
			[#lookDown] -> [ self lookDown ].
			[#lookUp ] -> [ self lookUp ].
			[#moveLeft ] -> [ self rotateLeft ].
			[#moveRight ] -> [ self rotateRight ].
			[#moveForward ] -> [ avatar zoomIn: (1.0 - (0.1 *(avatar  speed))). ] .
			[#moveBack ] -> [ avatar zoomOut: (1.0 + (0.1 * (avatar speed))) ].
		} otherwise: [ 
			avatar driver fullClutch
			 "no look-key binding " ].
	].! !

!QLookTask methodsFor: 'events' stamp: 'bgf 7/28/2009 22:30'!
step
	"Called in the display cycle when I'm the active task;
	 this is what actually drives the camera-angle changes."

	avatar ifNil: [ ^ self ].

	avatar cameraInLockStep: false.
	view event yellowButtonPressed ifTrue: [
		self updateCamera	
	] ifFalse: [
		self runSymbols
	].! !

!QLookTask methodsFor: 'controls' stamp: 'bgf 5/7/2009 15:14'!
taskEnded
	"End look-mode, returning to the avatars desired camera view"
	player ifNotNil: [ player mouseCursor: nil ].
	self cancelLook.! !

!QLookTask methodsFor: 'events' stamp: 'bgf 5/24/2009 12:10'!
updateCamera
	"Drive the camera in the direction indicated by the current mouse position.
	 (This is done from step, so long as this task is the active task of its view.)
	 The task installs and modifes a focal camera transform,
	 so the camera steering is done without other affect on the avatar stance. "
	| xform sc pyr |

	xform := self curTransform clone.	"... the camera orientation now"
 	sc := avatar speedControl.	"... the current mouse position, as a 'joystick' point." 

		"Change more slowly than driving, for visual smoothness ... 
		 camera turns relative to the avatar are easily very jerky; 
		 keeping the  speed down helps them be more natural."
		pyr := xform pitchYawRoll + ((sc y * 0.8) @ (sc x * 1.4) @ 0).
		pyr x: ((pyr x asSmallAngleDegrees) atLeast: -89 atMost: 89).	"Avoid singularities."
		pyr y: (pyr y asSmallAngleDegrees).
		pyr z: 0.
		xform := xform clone.
		xform pitchYawRoll: pyr.

		"N.B. it quite possible to translate the camera,
		 simply by giving xform a translation change governed by the speedControl
		 and the current camera orientation. 
		 (such as TAvatarDriver drive does for the avatar in normal driving.)
	       You could fly your eyes around in the space rather than just turning.
		 I'm not doing that here because it is a fairly big change in basic mechanics,
	 	 and we have no in-space indication for others that you're 'out of body'."
		
	avatar onFocalTransformChanged: (xform).! !

!QLookTask methodsFor: 'initialize' stamp: 'bgf 7/13/2009 23:29'!
view: aView
	"Sent when I become the active task for the view."
	super view: aView.
	player := view player.
	avatar := view avatar.
	keySymbols := Set new.! !

!TContentManager methodsFor: '*Qwaq-Client' stamp: 'bgf 12/3/2009 21:45'!
loadFromCache: thumb forOwner: owner
	"If the texture image file data for the given thumbnail is already
	 in the local machine's cache directory, load it and return true.
	 Otherwise return false.	

	 The caller should follow up with a document fetch request if this returns false."

	| sha entry  |

	sha := thumb get: #sha.
	sha ifNil: [ ^ false ].
	entry := (self assuredEntryForSHA: sha).
	(entry hasReferringThumb: thumb) ifTrue: [ ^ true ].	"Already requested it."

	entry addReferringThumb: thumb.
	entry privateContent ifNotNil: [  ^ true ].				"Already resolved it."

	(cacheMgr has: sha) ifFalse: [
		"If it is bundled, try to pull it straight out of the package, 
		 in case we're local/unconnected."
		| fName file data |
		fName := (thumb get: #fileName).
		(TContentManager isBundledFileName: fName) ifTrue:[
		[
			fName := thumb get: #fileName.
			file := FileStream readOnlyFileNamed: (FileDirectory default pathFromURI: fName).
			file binary.
			[data := file contents] ensure:[file close].
			cacheMgr cache: data to: sha.
		] on: Error do: [
			"NB - this  is expected once bundled content moves off-package."
			false ifTrue: [ TLogger logInfo: 'No local for bundled resource: ', fName ].
		]].
		"If we've failed to find it, return false to trigger resource fetching."
		data ifNil:[  ^ false ]].
	
	"Fallthrough: the thumb data is in the cache.
	 We prioritize anything bundled for immediate load, 
	 and any model as loadable right away, to smooth entry appearance."
	entry cacheName: sha.

	^ true 
	! !

TViewPane subclass: #QViewPane
	instanceVariableNames: 'player driveSymbols keyMap keyboardDrivingStopping keyboardDrivingStarting keyboardType annotationBlock'
	classVariableNames: 'KeyCodesMacToWin KeyCodesWinToMac'
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QViewPane commentStamp: '<historical>' prior: 0!
Camera, rendering and event delivery into a single point of view into a space,
and driving the avatar that represents the user's presence in that space.!

!QViewPane class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QViewPane class methodsFor: 'keyboard' stamp: 'jjl 7/22/2008 16:34'!
initializeKeyboardTranslations

"All of the Linux translations are incomplete and otherwise suspect.  The Mac and Windows translations are complete and still a llittle suspect for special keys"

"Keyboard maps based on http://classicteck.com/rbarticles/mackeyboard.php

						Mac   Windows	Linux
rbKeyUp 				126 	30 		103 	
rbKeyDown 				125 	31	 	108 	
rbKeyLeft 				123 	28 		105 	
rbKeyRight 				124 	29 		106 	
rbKeyBackspace 		51	 	8 		14 	
rbKeyEnter 				76 		* 		28 	
rbKeyHome 			115 	36 		102 	
rbKeyEnd 				119 	35 		107 	
rbKeyPageDown 		121 	34	 	109 	
rbKeyPageUp 			116 	33 		104 	
rbKeyReturn 			36 		13 		* 	
rbKeyDelete 			117		46 		111 	
rbKeyTab 				48	 	9 		15 	
rbKeySpacebar	 		49 		32	 	57 	
rbKeyShift 				56 		16 		* 	
rbKeyControl 			59 		17 		* 	
rbKeyMenu 				58 		18 		139 	The Alt key
rbKeyPrintScreen	 	* 		42 		210 	
rbKeyEscape 			53 		27 		1 	
rbKeyCapsLock 			57 		20 		58 	
rbKeyHelp 				114 	47 		138 	
rbKeyF1 				122 	112 	59 	
rbKeyF2 				120 	113 	60 	
rbKeyF3 				99 		114 	61 	
rbKeyF4 				118 	115 	62 	
rbKeyF5 				96 		116 	63 	
rbKeyF6 				97 		117 	64 	
rbKeyF7 				98 		118 	65 	
rbKeyF8 				100 	119 	66 	
rbKeyF9 				101 	120 	67 	
rbKeyF10 				109 	121 	68 	
rbKeyF11 				103 	122 	87 	
rbKeyF12 				111 	123 	88 	
rbKeyMacFn 			63 		* 		* 	
rbKeyMacOption 		58	 	*	 	* 	
rbKeyMacCommand 		55 		*	 	* 	
rbKeyWinLeftWindow 	* 		91	 	* 	On 'Natural' keyboards
rbKeyWinRightWindow 	* 		92	 	* 	On 'Natural' keyboards
rbKeyWinApplication 	110 	93	 	* 	On 'Natural' keyboards
rbKeyQ 					12 		81 		16 	
rbKeyW 				13 		87 		17 	
rbKeyE				 	14 		69 		18 	
rbKeyR 					15 		82 		19 	
rbKeyT 					17 		84 		20 	
rbKeyY 					16 		89 		21 	
rbKeyU 					32 		85 		22 	
rbKeyI 					34 		73 		23 	
rbKeyO 					31 		79 		24 	
rbKeyP 					35 		80 		25 	
rbKeyA 					0 		65 		30 	
rbKeyS 					1 		83 		31 	
rbKeyD 					2 		68 		32 	
rbKeyF 					3 		70 		33 	
rbKeyG 					5 		71 		34 	
rbKeyH 					4 		72 		35 	
rbKeyJ 					38 		74 		36 	
rbKeyK 					40 		75 		37 	
rbKeyL 					37 		76 		38 	
rbKeyZ 					6 		90 		44 	
rbKeyX 					7 		88 		45 	
rbKeyC 					8 		67 		46 	
rbKeyV 					9 		86 		47 	
rbKeyB 					11 		66 		48 	
rbKeyN 					45 		78 		49 	
rbKeyM 				46 		77 		50 	
rbKey0 					29 		48 		11 	
rbKey1 					18 		49 		2 	
rbKey2 					19 		50 		3 	
rbKey3 					20 		51 		4 	
rbKey4 					21 		52 		5 	
rbKey5 					23 		53 		6 	
rbKey6 					22 		54 		7 	
rbKey7 					26 		55 		8 	
rbKey8 					28 		56 		9 	
rbKey9 					25 		57 		10 	
rbKeyPeriod 			47	 	190 	52 	
rbKeyComma 			43 		188 	51 	
rbKeySlash 				44 		191 	53 	The key with /? generally next to right shift key.
rbKeyComma			41		186
rbKeyQuote				39		222
rbKeyBackSlash			42		220
rbKeyRightBracket		30		221
rbKeyLeftBracket		33		219
rbKeyEqual				24		187
rbKeyDash				27		189
rbKeyBackQuote		50		192
rbKeyNum0 			82 		96 		82 	On numeric keypad or with NumLock
rbKeyNum1 			83 		97 		79 	On numeric keypad or with NumLock
rbKeyNum2 			84 		98 		80 	On numeric keypad or with NumLock
rbKeyNum3 			85 		99 		81 	On numeric keypad or with NumLock
rbKeyNum4 			86 		100 	75 	On numeric keypad or with NumLock
rbKeyNum5 			87 		101 	76 	On numeric keypad or with NumLock
rbKeyNum6 			88 		102 	77 	On numeric keypad or with NumLock
rbKeyNum7 			89 		103 	71 	On numeric keypad or with NumLock
rbKeyNum8 			91 		104 	72 	On numeric keypad or with NumLock
rbKeyNum9 			92 		105 	73 	On numeric keypad or with NumLock
rbKeyMultiply 			67	 	106 	55 	On numeric keypad or with NumLock
rbKeyAdd 				69 		107 	78 	On numeric keypad or with NumLock
rbKeySubtract	 		78 		109 	74 	On numeric keypad or with NumLock
rbKeyDivide 			75 		111 	98 	On numeric keypad or with NumLock
rbKeyDecimal 			65 		110 	83 	On numeric keypad or with NumLock
rbKeyNumEqual 		81 		* 		117  On numeric keypad or with NumLock"

KeyCodesWinToMac := #(
255 255 255 255 255 255 255 255 117 48 
255 255 255 36 255 255 56 59 58 255 
57 255 255 255 255 255 255 53 123 124 
126 125 49 116 121 119 115 255 255 255 
255 255 255 255 255 255 51 114 29 18 
19 20 21 23 22 26 28 25 255 255 
255 255 255 255 255 0 11 8 2 14 
3 5 4 34 38 40 37 46 45 31 
35 12 15 1 17 32 9 13 7 16 
6 255 255 110 255 255 82 83 84 85 
86 87 88 89 91 92 67 69 255 78 
65 75 122 120 99 118 96 97 98 100 
101 109 103 111 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 41 24 43 27 
47 44 50 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255 
255 255 255 255 255 255 255 255 255 33 
42 30 39 ).
 
KeyCodesMacToWin := #(
65 83 68 70 72 71 90 88 67 86
255 66 81 87 69 82 89 84 49 50
51 52 54 53 187 57 55 189 56 48
221 79 85 219 73 80 13 76 74 222
75 186 220 188 191 78 77 190 9 32
192 8 255 27 255 255 16 20 18 17
255 255 255 255 255 110 255 106 255 107 
255 255 255 255 255 111 255 255 109 255 
255 255 96 97 98 99 100 101 102 103 
255 104 105 255 255 255 116 117 118 114 
119 120 255 122 255 255 255 255 255 121 
93 123 255 255 47 36 33 46 115 35 
113 34 112 28 29 31 30 ).

"
keyCodesWinToLinux := #(
255 255 255 255 255 255 255 255 14 15
255 255 255 255 255 255 255 255 139 255
58 255 255 255 255 105 103 1 108 106
255 255 57 104 109 107 102 255 255 255
255 255 210 255 255 255 111 138 11 2
3 4 5 6 7 8 9 10 255 255
255 255 255 255 255 30 48 46 32 18
33 34 35 23 36 37 38 50 49 24
25 16 19 31 20 22 47 17 45 21
44 255 255 255 255 255 82 79 80 81
75 76 77 71 72 73 55 78 255 74
83 98 59 60 61 62 63 64 65 66
67 68 87 88 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 51 52
53 ).

keyCodesMacToLinux := #(
30 31 32 33 35 34 44 45 46 47
255 48 16 17 18 19 21 20 2 3
4 5 7 6 255 10 8 255 9 11
255 24 22 255 23 25 255 38 36 255
37 255 255 51 53 49 50 52 15 57
255 111 255 1 255 255 255 58 255 139
255 255 255 255 255 255 83 255 55 255
78 255 255 255 255 255 98 28 255 74
255 255 117 82 79 80 81 75 76 77
71 255 72 73 255 255 255 63 64 65
61 66 67 255 87 255 255 255 255 255
68 255 88 255 255 138 102 104 14 62
107 60 109 59 105 106 108 103).

keyCodesLinuxToMac := #(
255 53 18 19 20 21 23 22 26 28
25 29 255 255 117 48 12 13 14 15
17 16 32 34 31 35 255 255 76 255
0 1 2 3 5 4 38 40 37 255
255 255 255 255 6 7 8 9 11 45
46 43 47 44 255 67 255 49 57 122
120 99 118 96 97 98 100 101 109 255
255 89 91 92 78 86 87 88 69 83
84 85 82 65 255 255 255 103 111 255
255 255 255 255 255 255 255 255 75 255
255 255 115 126 116 123 124 119 125 121
255 51 255 255 255 255 255 81 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 114 58
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 ).

keyCodesLinuxToWin := #(
255 27 49 50 51 52 53 54 55 56
57 48 255 255 8 9 81 87 69 82
84 89 85 73 79 80 255 255 255 255
65 83 68 70 71 72 74 75 76 255
255 255 255 255 90 88 67 86 66 78
77 188 190 191 255 106 255 20 20 112
113 114 115 116 117 118 119 120 121 255
255 103 104 105 109 100 101 102 107 97
98 99 96 110 255 255 255 122 123 255
255 255 255 255 255 255 255 255 111 255
255 255 36 26 33 25 27 35 28 34
255 46 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 47 18
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
255 255 255 255 255 255 255 255 255 255
42 ).
"! !

!QViewPane class methodsFor: 'keyboard' stamp: 'jjl 7/17/2008 11:04'!
keyboardType

	| type |
	(Smalltalk platformName = 'Win32') ifTrue:[ type := self winKeyboard.].
	(Smalltalk platformName = 'Mac OS') ifTrue:[ type := self macKeyboard.].
	type ifNil:[ type := 0 ].
	^type.
! !

!QViewPane class methodsFor: 'keyboard' stamp: 'jjl 7/17/2008 11:12'!
macKeyboard

	"This is used as both a type and a bit mask. (See buttonFix:)"
	^1024! !

!QViewPane class methodsFor: 'keyboard' stamp: 'jjl 7/17/2008 16:43'!
translateKey: key from: inputKeyboard to: hostingKeyboard

	| table size out macKB winKB |
	KeyCodesWinToMac ifNil:[ self initializeKeyboardTranslations ].

	macKB := self macKeyboard.
	winKB := self winKeyboard.

	inputKeyboard caseOf: {
		[ macKB ] -> [ (hostingKeyboard = winKB) ifTrue:[ table := KeyCodesMacToWin. ]. ].
		[ winKB ] -> [ (hostingKeyboard = macKB) ifTrue:[ table := KeyCodesWinToMac. ]. ].
	}.

	table ifNil:[ ^key ].

	size := table size.
	(key > size) ifTrue:[ ^key ].
	
	out := table at: key + 1.
	(out = 255) ifTrue:[ ^key ].

	^out

	
! !

!QViewPane class methodsFor: 'keyboard' stamp: 'jjl 7/17/2008 11:12'!
winKeyboard

	"This is used as both a type and a bit mask. (See buttonFix:)"
	^2048! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 8/18/2009 23:30'!
addDriveSymbol: aSymbol

	"Key-down of a keyboard-based driving symbol."

	self checkOpposites: aSymbol.
	driveSymbols ifNil: [
		driveSymbols := {aSymbol}.
		self leaveCurrentSelection.
	]  ifNotNil: [
		(driveSymbols includes: aSymbol) ifTrue: [
			^ self 
		].
		driveSymbols := driveSymbols copyWith: aSymbol.
		QActivityCapture countAction: #DriveKeyDown details: aSymbol asString.
	].
	avatar driveSymbols: driveSymbols.

	"Make sure there's a live driveStep: to update things immediately
	 whenever the drive-key set changes.
	 (The base system supplies this for startDriving; we don't want surplus.)
	 This ensures that key-taps ... even turn key taps while already moving...
	 are always reflected in the avatar motion/orientation, for a positive trustworthy feel."
 
	avatar driver isDriving ifFalse: [
		avatar startDriving.
	] ifTrue: [
		avatar driveStep: true.	"I.e. forced."
	].! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 7/16/2009 11:58'!
annotationBlock: block2args
	"A block taking two arguments - an OpenGL , and the view pane.
	 If not nil, the block is evaluated at the tail end of each frame render in this pane,
	 just prior to save-bits capture.     OpenGL is postured for 2D overlay rendering
	 (no lighting, etc) using beginOverlay2D just prior to each evaluation."

	annotationBlock := block2args.! !

!QViewPane methodsFor: 'session' stamp: 'bgf 6/7/2008 22:32'!
audioSession
	^ window audioSession! !

!QViewPane methodsFor: 'ui-events' stamp: 'das 8/23/2008 17:05'!
buttonFix: buttons

	| clicks extras |
	"It is possible for nrClicks to be nil, so need a guard"
	event nrClicks ifNotNil:[
		clicks := (event nrClicks bitAnd: 3) bitShift: 8.
	] ifNil:[clicks := 1 bitShift:8.].
	extras := clicks bitOr: keyboardType.
	^buttons bitOr: extras
! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 6/17/2008 23:49'!
checkDirectKeyDispatch: a2DEvent toTarget: anEventTarget

	"If the target is an embedded application, send it the 2D event.
	 return true if the event has been handled."

	window keyboardDriving ifTrue: [  ^false ].
	self isReadonlyUser ifTrue: [ ^ false ].

	^ (super checkDirectKeyDispatch: a2DEvent toTarget: anEventTarget) 
	and:[ 
		"Always deliver escape so the harness sees it... it takes out of input mode."
		((self keySymbolFor: a2DEvent) ~= #escape) ]

	! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 3/20/2009 10:52'!
checkDirectMouseDispatch: a2DEvent toTarget: anEventTarget
	"If the target is an embedded application, send it the 2D event.
	 return true if the event has been handled."
	| result |
	self isReadonlyUser ifTrue: [ ^ false ].

	window keyboardDriving ifTrue: [
		result := false.
		"Maybe it's a click on an app... if so, we change from driving to input mode."
		(a2DEvent type = #mouseDown 
		 and: [a2DEvent yellowButtonPressed not 
			   or: [ window avatar isAtApplicationPlace ]]) 
		ifTrue: [
			result := (super checkDirectMouseDispatch: a2DEvent toTarget: anEventTarget).
			result ifTrue: [ self keyboardDriving: false ].
		].
		^ result.
	].

	"Already in input mode... check the other path... we may be clicking out.
	 This allows background / off-appclicking to release the keyboard.
	 checkDirect returns true if the target is a typable app.
       (The annoyance risk here is that clicking a title bar, etc, will relinquish the keyboard.)"
	(super checkDirectMouseDispatch: a2DEvent toTarget: anEventTarget)
	 ifFalse: [
		(a2DEvent type == #mouseDown) ifTrue: [
			self keyboardDriving: true.
		].
		^ false
	].
	^true


	! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/18/2009 10:14'!
checkLookMode: ev2d
	"Make sure look-mode matches the ev2d's button states;
	 we are in look mode when using the right-mouse and shift key together,
	 or when using the right-mouse while seated.
	 Returns the new value of look-mode."
	| forced |

	forced := self owner allowedToMove not 
			 or: [ self avatar notNil  and: [ self avatar driver posture = #Sitting ] ].

	"In app-input we're never in look mode "
	self keyboardDriving ifFalse: [ ^  self lookMode: false ].

	ev2d isKeyboard ifTrue: [
		^ self lookMode: (forced).
	] ifFalse: [
		^ self lookMode: (ev2d yellowButtonPressed and: [ ev2d shiftPressed or: [ forced ] ] )
	].! !

!QViewPane methodsFor: 'driving' stamp: 'jjl 7/26/2008 18:28'!
checkOpposites: aSymbol
! !

!QViewPane methodsFor: 'initialize' stamp: 'jcg 8/30/2008 00:12'!
destroy

	"Take my avatar out of the space now, so it doesn't go zombie in replicants."
	avatar ifNotNil: [
		| tmp | 
		"We only want to try this once (aids interrupt-recovery when debugging)."
		tmp := avatar.
		avatar := nil.
		tmp replica isValidFarRef ifTrue: [
			[ 
				tmp replica future destroyAtClientTime: (DateAndTime now asUTCSeconds).
				tmp replica: nil.
				(Delay forMilliseconds: 50) wait.  "Help the destroy message on its way."
				QLogger logInfo: 'Avatar retired.'.
			] ifError: [:msg| 
				QLogger logWarning: 'Problem retiring avatar: ', msg.
			].
		].
	].
	
	super destroy.! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 12/11/2008 20:00'!
doDirectMouseDispatch: a2DEvent toTarget: anEventTarget position: pos

	"Qwaq extension: if we're showing the warp-to-place indicator,
	 don't also dispatch app events, so that warp-to is separated from launching/clicking."

	avatar ifNil: [ ^ false ].
	avatar replica isValidFarRef ifFalse: [ ^ false ].
	(avatar replica get: #warpPlaceUnderMouse: with: event) ifNotNil: [ ^ false ].

	^ super doDirectMouseDispatch: a2DEvent 
		      toTarget: anEventTarget 
			position: pos! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 6/7/2008 18:36'!
driveSymbols
	"The list of driving keys being pressed."
	^ driveSymbols! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 5/24/2009 12:15'!
driveSymbols: aList

	"The list of depressed driving symbols for keyboard driving."

	driveSymbols := aList ifEmpty: [nil].
	avatar driveSymbols: driveSymbols.
	driveSymbols ifNotNil: [
		avatar startDriving.
	] ifNil: [
		avatar stopDriving.
	].! !

!QViewPane methodsFor: 'drag-drop' stamp: 'bgf 8/25/2010 12:08'!
dropCard: cardRef position: position maxDropDistance: maxDistance

	"This is the bottleneck method by which session-side document/filecards 
	 are placed in the active space.  
	 The file card represents an imported file, 
	 or a template card from the Objects menu,
	 or a software-generated item such as a personal Web Cam.

	 The cardRef supplied must be a far ref to a QFileCard in the active space.
 	 Position is a 2D panel coordinate on my surface, or nil if the drop
	 did not originate with a placement gesture. "

	|  div offset distance trans pos evt |

	self readyForEvents ifFalse: [ ^ self ].
	avatar replica island = cardRef island ifFalse: [ ^ self ].

	"Most of this method is using the harness-side equipment to
	 prepare all the event/selection arguments for the dropCard: future message.
	 First, synthesize a pick event based on the drop position."
	pos := (position ifNil: [ (bounds center x @ ((bounds center y * 0.4 )truncated)) ]) - bounds topLeft.
	self pointerXY: (pos).
	(self fastPickSelection) ifNotNilDo: [ :sel |
		evt := event copy.
		evt selection: sel.
	] ifNil: [
		"Too busy or no OpenGL... Just take whatever's going, the last cycle-pick."
		evt := event copy.
	].
	"Some of that leaves us open to delays..."
	self readyForEvents ifFalse: [ ^ self ].

	"The following is effectively a 'pointer pickPlane' on a plane a couple of units
	into the space and perpendicular to the camera ray.  
	This is where to place the card in front of the user."
	offset := (camera pointerXY: pos) row3.
	offset := (avatar localTransform)
				transformVector: (offset x)@(offset y)@(offset z negated).
	div := (avatar localTransform transformVector: (0 @ 0 @ -0.9)) normalized 
				dot: offset.
	div abs < 0.5 ifTrue: [	"Guard"
		div := 1.8 	 "too near is better than too far"
	].
	distance := 1.8/(div abs).			"1.8 - basic distance from the avatar"
	trans := (avatar localTransform translation + (offset * distance)).

	evt utcSeconds: (Time utcSeconds).
	
	"If we've not teleported amidst all that, then finish the drop."
	(avatar replica island = cardRef island) ifTrue: [
		avatar replica future 
				dropCard: cardRef 
				event: evt
				rawCardPosition: trans 
				graspDistance: distance
				maxDropDistance: maxDistance
	].! !

!QViewPane methodsFor: 'private-render' stamp: 'bgf 11/28/2009 09:22'!
fastPick: ogl

	'BGF-Selection' ifNil: [
		"If in full-scene editing, allow picking non-pickable objects.
		 This will be enabled later, when in-scene click-selection is allowed. "
		self owner ifNotNilDo: [ : ui | 
			eventPointer pickableOnly: (ui isEditingTemplate not).
		]].
	super fastPick: ogl.
	! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 8/7/2009 00:49'!
filterKeyUp: sym for: evt

	"We filter key events to support keyboard driving."
	sym caseOf: {
		[#scrollUp] -> [].
		[#scrollDown] -> [].	"Handled in keyStroke."
		[#shift] -> [] . "Windows shift.  The Mac doesn't send these as key events."
		[#escape] -> [].
		[nil] -> [].
	} otherwise: [
		"Release the key that was driving the avatar."
		self removeDriveSymbol: sym.
		^false.
	].

	^true
! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 11/11/2009 18:13'!
finishKeyUp: sym

	window keyboardDriving ifFalse: [
		"Deliver the keyUp regardless of value, since we delivered the down."
		event target ifNotNil:[
			self isReadonlyUser ifFalse: [ task doKeyUp: event ].
		] ifNil: [
			"Fallthrough: unhandled keys in input mode just go away."
		].
	] ifTrue: [
		sym ifNil: [  avatar keyUp: event ].	"Unfiltered keys go on to the avatar."
	].

	"Finally, if the key down was to start or stop driving, we can complete the transition,
	 since the up and down have both been consumed.
	 TODO - generalize for the case where the sequence is down-up-stroke;
	 the current code assumes down-stroke-up.
	 NB - we really need only one flag (keyboardDrivingChanging) for this,
	 but for vague reasons I'm more comfortable keeping them distinct."
	keyboardDrivingStopping ifTrue: [
		self keyboardDriving:  false.
		keyboardDrivingStopping := false.
	].
	keyboardDrivingStarting ifTrue: [
		self keyboardDriving: true.
		keyboardDrivingStarting := false.
	].
	event action: #unknown.
	! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 5/7/2009 17:09'!
handleRightMouseForDown: event2D

	"Is this 2D-event right (yellow) mouse down?
	 Extended to set an indicator-cursor in the view."

	self isLookMode ifTrue: [
		"If already in look mode, just stay that way."
		self task doPointerDown: event2D.
		^ true
	].

	(super handleRightMouseForDown: event2D) ifTrue: [
		self owner viewPanel mouseCursor: (QStyles drivingCursor).
		^ true
	].
	^ false! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/10/2009 09:30'!
handleRightMouseForMove: event2D

	self checkLookMode: event2D.

	"If the yellow mouse is released during move,
	 clear the drive mouse-cursor. "
	(event2D yellowButtonChanged and: [event2D yellowButtonPressed not])  ifTrue: [
		window owner viewPanel mouseCursor: (self task cursor).
	].
	(self isLookMode not and: [ event2D yellowButtonPressed ]) ifTrue: [
		"We may have just released the look (shift) key, so turn driving back on if needed...
		 but not if we're in input mode, only if we're driving already."
		self keyboardDriving ifTrue: [
			avatar startDriving
		]].
	"Don't drive in look mode."
	self isLookMode ifTrue: [ 
		self task doPointerMove: event2D.
		^ true 
	 ].

	"Is this 2D-event right (yellow) mouse down?
	 Returns true if the event is handled here."

	(event2D yellowButtonPressed and: [ self keyboardDriving ]) ifTrue: [
		self mouseDrivingStart: event2D.
		avatar rightMouseMove: event2D.
		^ true
	] ifFalse: [
		self mouseDrivingStop: event2D.
		event2D yellowButtonChanged ifTrue: [
			avatar rightMouseUp: event2D.
		].
		^ false
	].! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/10/2009 09:36'!
handleRightMouseForUp: event2D
	"Is this 2D-event right (yellow) mouse release?
	 Extended to clear the driving mouseCursor when driving stops,
	 and to switch into or out of look mode if metakeys indicate that.."

	| wasLook |
	wasLook := self isLookMode.
	(self checkLookMode: event2D) ifTrue: [ 
		self task doPointerUp: event2D.
		^ true 
	].
	wasLook ifTrue: [ ^ true ]. "Clearing look is all we need to do to handle finishing look."

	(self keyboardDriving 
		and: [event2D yellowButtonChanged ]
		and: [event2D yellowButtonPressed not])
	 ifTrue: [
		self mouseDrivingStop: event2D.
		eventPointer pointerDownFlag: false.
		avatar rightMouseUp: event2D.
		self owner viewPanel mouseCursor: nil.
		^ true
	].
	^ false
! !

!QViewPane methodsFor: 'initialize' stamp: 'jjl 7/17/2008 11:11'!
initialize
	super initialize.

	keyMap := QDriveKeyMap new.
	keyboardDrivingStopping := false.
	keyboardDrivingStarting := false.

	keyboardType := MViewPane keyboardType.

! !

!QViewPane methodsFor: 'drag-drop' stamp: 'bgf 1/7/2010 14:55'!
insertModel: modelRef forPosition: aPoint editBox: inEditBox doScaling: bDoScale

	"Add the given model (tframe) reference to the space in an edit box, in front of the avatar;
	 the model is assumed to already have been inserted in the Island"

	| pos camT |
	pos := aPoint ifNil: [ bounds center ].
	camT := camera  pointerXY: pos.
	inEditBox ifTrue:[
		^ space future 
			addModelInEditor: modelRef
			inFrontOf: (camera localTransform * camT)
			for: avatar replica
			doScale: bDoScale
	].
	^ nil! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 5/8/2009 02:47'!
isLookMode
	"Look mode is implemented as a kind of task,
	 with some conspiracy support here to integrate it as a right-button behavior."
	^ self task class == QLookTask! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 6/19/2008 12:35'!
isReadonlyUser

	"Indicates a user who cannot do -anything- to interact with stuff in the space.
	 This is (currently) the only implementation of action-access-control
	 for inspace objects and delivery of events into the space.
	 When the event-handling regime and all the in-space objects are enhanced
	 to honor user roles in there, then this will unfold into more specific capabilities."

	^ window ifNil: [ true ]
			   ifNotNil: [ window owner isReadonlyUser ].! !

!QViewPane methodsFor: 'session' stamp: 'bgf 6/8/2008 15:31'!
jobHandler
	"The QJobHandler for harness-side activities."
	^ window jobHandler! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 12/10/2009 13:55'!
keyDown: evt
	| sym |
	window ifNil: [ ^ self ].
	window updateLastEventTime.
	sym := self keySymbolFor: evt fromKeyUp:false.
	self fillEvent: TFrame eventKeyboard withKeyboardEvent: evt.
	event action: #keyDown.

	"Check direct dispatch to underlying embedded application."
	self readyForEvents ifTrue: [
		(self checkDirectKeyDispatch: evt toTarget: event target) ifTrue: [ ^ self ].
	].

	avatar ifNil: [ ^ self ].

	self keyboardDriving ifTrue: [
		sym caseOf: {
			[#escape] -> [ ].
			[#shift] -> [] . "Windows shift. When driving, we read the shift state in setDriveSpeed:"
			[#scrollUp] -> [].	"Handled as a keyStroke."
			[#scrollDown]->[].
			[#resetPosition]-> [ avatar resetPosition ].		"Home key"
			[nil]->[  avatar keyDown: event  ].	"Unfiltered keys go on to the avatar."
		} otherwise: [
			self addDriveSymbol: sym
		].
	] ifFalse: [
		"In typing mode, <escape> takes us back to driving mode.
	       <Return> doesn't , so we can type return keys in typing mode."
		(sym = #escape) ifTrue: [
			keyboardDrivingStarting := true.
		].
		"If we are pointing to something then send the event to that, otherwise send to avatar"
		event target ifNotNil:[
			self isReadonlyUser ifFalse: [ task doKeyDown: event ].
		] ifNil: [
			"Fallthrough - keys in typing mode that go nowhere, go nowhere."
		].
	].
	"Esc always releases from zoomToWindow."
	sym = #escape ifTrue: [  window owner showPanels ].
	event action: #unknown.
! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 12/10/2009 13:56'!
keyStroke: anEvent
	"Don't start handling things until we're ready."
	| esc |
	window ifNil: [ ^ self ].
	window updateLastEventTime.
	self fillEvent: TFrame eventKeyboard withKeyboardEvent: anEvent.
	event action: #keyStroke.

	"Check direct dispatch to underlying embedded application."
	self readyForEvents ifTrue: [
		(self checkDirectKeyDispatch: anEvent toTarget: event target) ifTrue: [ ^ self ].
	].
	avatar ifNil: [ ^ self ].
	esc := (event keyValue = 27).

	"If keyboardDrivng is disabled, deliver keyStroke to event target."
	(window keyboardDriving not | esc ) ifTrue: [
		"The escape key signals return to keyboard driving, 
		 and its stroke is also delivered into the frames to serve as a 'cancel' indicator."
		self isReadonlyUser ifFalse: [
			event target ifNotNil:[
				task doKeyStroke: event.
				esc ifFalse:[ ^ self ].  "Let escape propagate below."
			].
		] ifTrue: [
			self owner statusAlert: 'As a visitor, you cannot type within the space.'.	
			esc ifFalse:[ ^ self ].  "Let escape propagate below."	
		].
	].

	"Allow alternative treatments of keyboard event."
	(self letOthersTryKey:event) ifTrue:[ ^self ].

	"Fallthrough: deliver unhandled events to the avatar, 
	 unless we're in strict typing mode; 
	 undirected keyStrokes in typing mode simply go nowhere."
	window keyboardDriving ifTrue: [
		avatar keyStroke: event.	
	].
	event action: #unknown.
! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 6/7/2008 19:04'!
keyStrokeSymbolFor: anEvent

	"The logical mapping of a key stroke event for driving control.
	 We only return the scrollUp and scrollDown event codes here;
	 all other keyStrokes are ignored, since we handle them in keyDown/keyUp."

	| result |

	result := keyMap map: anEvent keyValue.
	"Detect the synthetic keystrokes (no corresponding keyDown/keyUp)
	 that Squeak generates to indicate scroll wheel motion.
	"
	result := result caseOf: {
			[keyMap map: #scrollUp] -> [#scrollUp].
			[keyMap map: #scrollDown] -> [#scrollDown].
		} otherwise: [nil].

	^ result.
! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 1/30/2010 13:33'!
keySymbolFor: anEvent
	"For use with keyDown and keyUp events - these use different keyValues
	 than the corresponding keyStroke events."
	| result |
	result := keyMap map: anEvent keyValue.

	"On the Mac, filter out the precursors of synthetic scroll wheel events;
	 the mac sends 126, 125 (moveForward and moveBack keyPress) events
	 (that are marked as control-key down)
	 in addition to the 30,31 (Windows-style) forward and back stroke events.
	 We filter these out, so that we don't go into driving mode for these,
	 since we want to deal with them as scroll wheel  events in keyStroke:"
	(result notNil 
		and: [(Smalltalk platformName = 'Mac OS')
			and: [anEvent controlKeyPressed]]) ifTrue: [

		result caseOf: {
			[#moveForward] -> [result := nil].
			[#moveBack] -> [result := nil].
		} otherwise: [].
	].

	^ result! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 1/30/2010 13:45'!
keySymbolFor: anEvent fromKeyUp: yesno
	"Return the driving key symbol mapping for the given down or up key event.
	 For use with keyDown and keyUp events - these use different keyValues
	 than the corresponding keyStroke events."

	| sym |

	"Command-modified events need to reach the avatar-user
	  to allow normal edit-key shortcuts (#X/#C/#V, etc).
	 But, may interfere with key-driving during push-talk down (control key effect)"

	( ((Smalltalk platformName = 'Mac OS') and: [anEvent commandKeyPressed])
	  or: [ (Smalltalk platformName = 'Win32') and: [anEvent controlKeyPressed]] )
	ifTrue: [ 
		"We're going to ignore this as a drive key.  But if it's an up, we need to prevent run-on"
		yesno ifTrue: [
			sym := keyMap map: anEvent keyValue.
			sym ifNotNil: [ self removeDriveSymbol: sym ].
		].
		^ nil ].

	^ self keySymbolFor: anEvent! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 12/10/2009 13:56'!
keyUp: evt
	| sym |
	window ifNil: [ ^ self ].
	window updateLastEventTime.
	self fillEvent: TFrame eventKeyboard withKeyboardEvent: evt.
	event action: #keyUp.

	"Check direct dispatch to underlying embedded application."
	self readyForEvents ifTrue: [
		(self checkDirectKeyDispatch: evt toTarget: event target) ifTrue: [ ^ self ].
	].
	avatar ifNil: [ ^ self ].

	"We filter key events to support keyboard driving."
	sym := self keySymbolFor: evt fromKeyUp: true.
	self filterKeyUp: sym for: evt.

	self finishKeyUp: sym.! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 6/8/2008 02:09'!
keyboardDriving
	^ window keyboardDriving! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 6/8/2008 02:10'!
keyboardDriving: aBool
	window keyboardDriving: aBool! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 6/7/2008 19:02'!
letOthersTryKey:event

	"A subclass might allow event targets or 'selected' objects to handle the event."
	^false.! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 5/24/2009 12:30'!
lookMode: lookModeBool
	"Turn look mode on or off.  Look mode is implemented by pushing a look-task"
	(self isLookMode = lookModeBool) ifFalse: [
		lookModeBool ifTrue: [ 
			avatar stopDriving.		"So (for instance) we don't walk-animate."
			self pushTask: (QLookTask new).
		] ifFalse: [
			self task class = QLookTask ifTrue: [
				self popTask.
			].
		]].

	^ lookModeBool! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/18/2009 10:09'!
mouseDown: evt

	event action: #pointerDown.
	self isReadonlyUser ifFalse: [
		super mouseDown: evt.
	] ifTrue: [
		self processPointerFor: evt.
		(self handleRightMouseForDown: evt) ifFalse: [
			self owner statusAlert: 'As a visitor, you cannot click within the space.'.
		].
	].
	event action: #unknown.
! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 6/18/2008 00:08'!
mouseEnter: evt
	"A little protection against undelivered keys when the mouse-focus changes abruptly...
	 don't keep driving based on older state in this case."
	event action: #pointerEnter.
	self driveSymbols: #().
	super mouseEnter: evt.
	event action: #unknown.! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/18/2009 10:09'!
mouseLeave: evt
	"A little protection against undelivered keys when the mouse-focus is lost abruptly...
	 don't keep driving in this case."
	event action: #pointerLeave.
	self driveSymbols: #().

	self isReadonlyUser ifFalse: [
		super mouseLeave: evt.
	] ifTrue: [
		self processPointerFor: evt.
		self handleRightMouseForMove: evt.
	].

	event action: #unknown.! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/18/2009 10:09'!
mouseMove: evt

	| type |
	event action: #pointerMove.
	self isReadonlyUser ifFalse: [
		super mouseMove: evt.
	] ifTrue: [
		self processPointerFor: evt.
		type := (evt anyButtonPressed
						ifTrue: [TFrame eventPointerDown] 
						ifFalse: [TFrame eventPointerOver]).
		self fillEvent: type withUserInputEvent: evt.

		"Check/handle joystick-style mouse driving."
		self handleRightMouseForMove: evt.
	].
	event action: #unknown.! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/18/2009 10:10'!
mouseUp: evt

	event action: #pointerUp.
	self isReadonlyUser ifFalse: [
		super mouseUp: evt.
	] ifTrue: [
		self processPointerFor: evt position.
		self handleRightMouseForUp: evt
	].
	event action: #unknown.
! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 6/7/2008 19:07'!
owner
	"The participant UI"
	^ window owner! !

!QViewPane methodsFor: 'service linking' stamp: 'jjl 8/27/2008 09:24'!
performReturnPrep: target atTransform: aTransformOrNil.

	^true! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 12/4/2008 17:56'!
player
	"The QViewPlayer that manages me"
	^ player! !

!QViewPane methodsFor: 'accessing' stamp: 'bgf 12/4/2008 18:03'!
player: aCPlayer
	"The QViewPlayer that manages me"
	player := aCPlayer! !

!QViewPane methodsFor: 'ui-events' stamp: 'bgf 7/9/2009 02:28'!
pointerXY: pxy

	"Override to pass the driveSymbols as the avatar speedControl
	 whenever it is non-nil; this is the keyboard driving mechanism."

	| speed sx sy xAccel yAccel mid |
	window updateLastEventTime.
	self readyForEvents ifFalse: [ ^ self ].

	eventPointer localTransform: (camera pointerXY: pxy).
	
	mid := bounds extent / 2.0.
	"The center region (4 pixels each side of center) is no-motion."
	sx := (mid x - (pxy x)).
	sx abs < 5 ifTrue: [ sx := 0 ] ifFalse: [ sx := sx - (5 * (sx sign)) ].

	sy := (mid y - (pxy y)).
	sy abs < 5 ifTrue: [ sy := 0 ] ifFalse: [ sy := sy - (5 * (sy sign)) ].

	xAccel := sx / (mid x).
	yAccel := sy / (mid x).

	"The stick has increasing effect as you move out from center,
 	 so fine motion is possible at low accel.
	 (Accels are (0, 1); log is log10; the 3 is a floor guard.)"
	xAccel := xAccel * (((xAccel * 24) abs max:3) log).   
	yAccel := yAccel * (((yAccel * 32) abs max:3) log).

	"And reverse is 1/3 the speed of forward; with partial dampening of back-turn;
	 we suspect this may be less discomfiting to new users.  Mike Kaye's request."
	self isLookMode ifFalse: [
		yAccel < 0 ifTrue: [ yAccel := (yAccel *0.33).  xAccel := (xAccel *0.5) ].
	].

	speed := (xAccel @ yAccel).
	
	self mouseDriving ifFalse: [ window fastPick: self ].
	avatar speedControl: speed
			pointer:  eventPointer 
			redButton: mouseIsDriving.

	window eventPosition: (pxy + self origin).! !

!QViewPane methodsFor: 'ui-events' stamp: 'jjl 9/4/2008 12:12'!
processPointerFor: evt

	self pointerXY: evt position.! !

!QViewPane methodsFor: 'task' stamp: 'bgf 6/7/2008 18:55'!
pushTask: aTask
	super pushTask: aTask.
	window owner viewPanel mouseCursor: (aTask cursor).! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 5/24/2009 12:04'!
removeDriveSymbol: aSymbol

	"Key-up of a keyboard-based driving symbol."
	driveSymbols ifNil: [
		^ self
	]  ifNotNil: [
		(driveSymbols includes: aSymbol) ifTrue: [
			driveSymbols := (driveSymbols copyWithout: aSymbol) ifEmpty: [nil].
			QActivityCapture countAction: #DriveKeyUp details: aSymbol asString.
		].
	].

	avatar driveSymbols: driveSymbols.
	driveSymbols ifNil: [ avatar stopDriving ].
	! !

!QViewPane methodsFor: 'initialize' stamp: 'bgf 7/16/2009 11:52'!
renderAnnotations: ogl
	"The annotation if provided is rendered in 2D overlay posture."
	annotationBlock ifNotNil: [ 
		self beginOverlay2D: ogl.
		annotationBlock value: ogl value: self.
		self endOverlay2D: ogl.
		]! !

!QViewPane methodsFor: 'driving' stamp: 'bgf 6/7/2008 18:52'!
resetToNormalDriving

	"Clear avatar sticky state, and reset to keyboard driving."

	(avatar isSticky)  ifTrue: [
		avatar sticky: false.
	].
	self keyboardDriving: true.! !

!QViewPane methodsFor: 'session' stamp: 'bgf 6/7/2008 22:36'!
show3DPointerFor: anAvatarUserID
	"Shorthand query - given current preferences,
	 should the 3d-cursor be drawn for the avatar with the given ID?"
	
	^ window show3DPointerFor: anAvatarUserID! !

!QViewPane methodsFor: 'session' stamp: 'bgf 7/23/2009 19:29'!
showNameFor: anAvatarUserID whenSpeaking: bSpeaking
	"Shorthand query - given current preferences,
	 should the 3d-cursor be drawn for the avatar with the given ID?"
	
	^ window showNameFor: anAvatarUserID whenSpeaking: bSpeaking! !

!QViewPane methodsFor: 'session' stamp: 'bgf 6/7/2008 22:33'!
videoSession
	^ window videoSession! !

TWindowedSession subclass: #QWindowedSession
	instanceVariableNames: 'owner jobHandler sharedDesktopManager documentManager appLaunchMutex reconnecting audioSession videoSession dropModelThumb endBenchTest keyboardDriving focalDocumentName animationSession'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QWindowedSession commentStamp: '<historical>' prior: 0!
This is the Qwaq windowed-session class, that manages a client croquet session.
We extend resource support, and provide resource-content import tools here.!

!QWindowedSession class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QWindowedSession class methodsFor: 'private' stamp: 'bgf 1/17/2009 11:49'!
preloadCachePaths

	"An array of URI paths into the product installation (relative to default directory).
	 Each a directory containing SHA-named cache entries for a forum template
	 that was built in Forums and made generally available;
 	 bundling these resources allows new visitors to enter such forums 
	 without downloading every piece.  (Use reserved for 'core' product resources.)"

	^ {
			'PreCache/Core' .		"Common small resources for widgets"
			'PreCache/Avatars'.	      "Resources for standard Animated Characters."
			'PreCache/Project-Forums' .	 "Caribbean, Pacific, Baltic theme Project Fora"
			'PreCache/Library'.		"The Library template's graphics and meshes."
			'PreCache/Reception'	"The Jan-2009 Welcome Forum decorations."
	}! !

!QWindowedSession methodsFor: 'render' stamp: 'jcg 4/9/2010 17:44'!
aboutToRender
	super aboutToRender.
	audioSession ifNotNil: [:as |
		"Typically we send #future to the audio-session, but here 
		 we want it to immediately get the avatar's localTransform." 
		audioSession aboutToRender 
	].! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 6/8/2008 18:52'!
absUrlForPostcard: aTPostcard inService: aService

	"Given a tframe's #postcard - which is relative to an island -
	 return the absolute URL of that TFrame in the given service."

	^TPostcard makeUrl: aTPostcard urlString
				relativeTo: (aService url, '/dummy').! !

!QWindowedSession methodsFor: 'accessing' stamp: 'ar 10/22/2009 12:49'!
activeIsland
	"Conversion leftover - the island into which imports, etc, go by default."
	^ [owner avatar replica island] on: Error do: [ nil ]! !

!QWindowedSession methodsFor: 'islands' stamp: 'bgf 9/10/2009 13:00'!
addIsland: isl postcard: pc

	super addIsland: isl postcard: pc.

	"Make this island's URL string available from within-space."

	isl future at: #IslandURLString put: pc urlString.
	isl future at: #ServiceURLString put: (pc urlString copyUpToLast: $/).
	isl future clientVersion: (QwaqVersion current infoString).

	owner ifNotNil: [ owner islandAdded: isl postcard: pc ].! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/28/2009 12:39'!
addResource: resFarRef type: resourceType in: anIsland

	resourceType caseOf: {
		[QSharedDocument resourceType] -> [ self updateSharedDoc: resFarRef signalTools: true].
	} otherwise:[
		super addResource: resFarRef type: resourceType in: anIsland
	].
! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
audioSession
	^audioSession! !

!QWindowedSession methodsFor: 'accessing' stamp: 'jcg 4/30/2010 11:22'!
audioSession: aQwaqAudioSession

	(audioSession = aQwaqAudioSession) ifFalse: [
		audioSession ifNotNil: [audioSession future endVoiceChat wait].
		audioSession := aQwaqAudioSession.	
		audioSession ifNotNil: [audioSession view: self].
	].! !

!QWindowedSession methodsFor: 'render' stamp: 'bgf 6/8/2008 12:24'!
bounds: bnds 
	
	"Create or reset the openGL window 
	 to occupy the given screen rectangle
	 and use the given camera viewing angle."

	owner doOpenGL ifTrue: [
		^ super bounds: bnds
	].

	"Below is a partial version of the inherited method,
       allowing the participant basics to work headless, without an OpenGL window."

	globalBounds := bnds.
	doRender ifFalse:[^nil].

! !

!QWindowedSession methodsFor: 'private' stamp: 'ar 4/16/2011 17:26'!
checkDropSize: aFile tag: aString size: maxSize

	"If the prospective drop file is very big, get permission first
	 (and help the user understand long delays from dropping huge things.)"

	^ (aFile size < maxSize)
	   or:[
		self confirm: ('This ', aString,' is ', (aFile size / 1024) rounded, ' KB.\',
			'\Large objects may create performance\',
			'and voice issues during import activity.\',
			'\Upload ', aString,
			'\to Forum host ', 
			owner serviceMgr host, '?\') withCRs]
! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bgf 9/8/2009 12:37'!
checkDropSlides: droppedFiles event: anEvent initialName: aString
	"Part of desktop drop-into-forums support.
	 If the user has dropped a group of images,
	 ask if they want to make a slide show out of them.
	 Returns true if the drop processing is handled here by making a show; false if not."
	| doc card files |

	(QSlideShowDocument checkForSlideShowDrop: droppedFiles) caseOf: {
		[#oneByOne] -> [ ^ false ].
		[#cancel ]	  -> [ ^ true ]. 	"Do nothing, claim to have finished processing."
	} otherwise: [ "carry on" ].

	owner readyForEvents ifFalse: [ ^ true ].
	doc := QSlideShowDocument new.
	doc title: (aString ifNil: ['Untitled']).
	doc ctime: (documentManager newTimeStamp: 'Import').
	files := droppedFiles 
			select: [ : strm |  (ImageReadWriter readerClass: strm) notNil ].

	"To get powerpoint-named sequences, we need a  custom comparison."
	files := files sort: [ : s1 : s2 | QSlideShowDocument filenameCompare: s1 name with: s2 name ].

	[
	CProgressBar 
		displayProgress: 'Importing Slide Show...'
		at: owner bounds center
		from: 1 to: (files size+1) 
		during: [  :bar | | i |
			owner waitTick.
			i := 0.
			documentManager
				useUploadProgress: [ :done| bar value: ( i + done )]
				during: [
					files do: [ : strm | 
						bar value: i.
						owner readyForEvents ifFalse: [ ^ true ].
						doc add: (self importImageAsTexture: strm).
						i := i + 1 ].	
					owner readyForEvents ifFalse: [ ^ true ].
					card := QFileCard forDocument: doc.
					card isTransient: true.
					card := TIslandCopier new import:(TIslandCopier new export: card) 
												  to: self activeIsland.
					bar value: (i + 0.5).
					self dropCard: card event: anEvent.
				]].
	] on: CProgressCancel do: [
		QLogger logInfo: 'Canceled slides upload'.
		^ true "I.e. handled, nothing further to do."
	].
	^ true! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 2/14/2009 15:32'!
colorPreviewForService: aServiceUrl
	"Express interest so the main manager knows to fetch when needed.
	 This method is called from render (thread) so we signal across to tweak.
	 If the preview must be fetched, it'll arrive in our cache sometime later."
	^ owner infoCache colorPortalTexture: aServiceUrl! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 5/29/2009 11:10'!
colorPreviewForUrl: aUrl

	| shorten |
	aUrl isEmptyOrNil ifTrue: [ ^ nil ].
	(aUrl beginsWith: '//') ifFalse: [ ^ nil   "Relative url, no preview" ].

	shorten := [owner quickServiceUrlForUrl: aUrl]
				on: Error
				do: [ 
					QLogger logWarning: 'Not a service URL: ', aUrl. 
					^ nil ].
	
	^ self colorPreviewForService: shorten! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 4/14/2010 17:51'!
copyResourceToUser: fileName 
	"Copy one servers-side resource to the user's public storage area on the server, if not already in a globally scoped location.
	Answer the new fileName (so that senders can change the referencing object), or nil if already globally scoped.
	Signals error on failure."

	| oldName baseName userDocName response |
	(fileName beginsWith: '/') ifTrue: [
		"Not relative to the forum; already in an org-scoped location;  leave it alone."
		^nil
		].

	oldName := '/services/', owner currentForumName, '/', fileName.

	baseName := (fileName copyAfterLast:$/) ifEmpty: [ fileName ].
	"Copy from sharedName to /users/<me>/public/resources/<shared-name> -
	  This is a spot from which others can download, no matter what forum they meet me in."

	userDocName := '/users/', owner serviceMgr user, '/public/resources/', baseName.

			"XXX To-Do: this should use the SHA to name the user-side resource file,
	 		 and do the same sort of redundancy optimizations as we do for mesh uploads."

	owner statusInfo: 'Processing ', baseName; waitTick; waitTick.
		
	"Copy from current location to the user's public storage area "
	response := owner serviceMgr 
					copyFiles: oldName
					to: userDocName.
	response name == #ok 
		ifFalse: [ self error: ('Cannot copy resource\', baseName, '\ to your inventory:\\', response contentString) withCRs.].	

	^userDocName.! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 3/21/2009 13:06'!
copyResourcesToUserForFrame: tframeRef

	"Take all the server-side textures used in tframe,
	 and copy them to the user's public storage area on the server.
	 This is done when we capture and save an avatar model from the space,
	 so that the textures required for our avatar appearance
	 are available from the user-public-documents path.

	 This is analogous to model upload (or to 'convert to lazy textures)
	 but for the case of copying a model out of a forum into the user's general storage,
	 for later acquisition by other users in different forums."

	| fileName thumbs |

	thumbs := tframeRef get:#allThumbs.
	(thumbs value class = Array) ifFalse: [ ^ true ].

	thumbs do: [ : thumb |

		fileName := thumb get:#fileName.

		(TContentManager isBundledFileName: fileName) ifFalse: [
			[
				(self copyResourceToUser: fileName) ifNotNilDo: [:newName |
					thumb future fileName: newName.
					QLogger logInfo: 'Thumb transferred to: ', newName.
					].
			] on: Error do: [:exc |
				owner inform: exc description.
				^ false.
				].
			]].

	owner statusInfo: 'Resources done. '; waitTick; waitTick.  "Allow the last progress bar to clear."
	^ true.
! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bgf 9/19/2008 14:10'!
curDropDistanceFor: aCard atPosition: aPosition

	"Return the 'forward range' to use right now to handle a document-card drop into the space,
	 based on where I am and what I'm doing.  
	 A result of 0 means the incoming file card will always become carried by the avatar,
	 and will not auto-appear anywhere.  A positive result is the distance outward from the avatar
	 within which a picked drop-target (that can receive the card) will be allowed to do so.
	 Short ranges prevent a drop from acting on something too far away.

	 The rules are attempts to anticipate user intentions and do the 'right thing'.
	 For instance, if you import a document or URL while you're focal on a display-panel,
	 you probably really mean to put that right into the panel.
	 But even if you're aimed at a panel, if you're further away from it, you probably don't mean that."

	self avatar ifNil: [ ^ 0 ].

	"If I'm not in a defined place, just free-wandering,
	 allow logical drops at medium range for pictures,
	 but much lower range for other content.
	 This is the case that gets posters onto walls, for instance;
 	 we need to keep behavior compatible with earlier forums behaviors
	 (when such drops -always- fired) 
       enough so that people aren't too disoriented by the transition to smarter rules."
	self avatar currentPlace ifNil: [ 
		aPosition ifNotNil: [			
			"I.e the user did 'point' at where this is going..."
			(aCard waitTimeoutSeconds: 2).
			^ ((aCard get:#kind) = #image) ifTrue: [ 128 ] ifFalse: [ 64 ] 
		] ifNil: [
			 "Command-originated, so don't drop unless necessary to avoid losing the card."
			^ 4
		].
	].

	"If we're focal on a display panel, assume we mean it"
	(self avatar currentPlace category = 'Panels') ifTrue: [
		^ 72
	].

	"Default - we are at a defined place (but not a Panel).
	 In this case, we only autodrop at short range
	 so we don't lose cards behind obstacles and so on...
	 Allow targeted drop at intermediate range."

	^ (aPosition ifNil: [ 4 ] ifNotNil: [ 48 ]).! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 6/21/2009 09:29'!
currentForumURL
	"The service url of the forum my user currently occupies."
	^ owner ifNotNil: [ owner currentForumURL ]! !

!QWindowedSession methodsFor: 'initialize' stamp: 'bgf 11/14/2009 16:36'!
destroy

	self stopScript: #onFetchComplete:.
	
	QLogger logInfo: 'Ending windowed session.'.
	jobHandler ifNotNil:[ 
		jobHandler destroy.
		jobHandler := nil.
		QLogger logInfo: 'Job handler cleared.'.
	].

	self audioSession: nil.
	self videoSession: nil.

	animationSession destroy.
	
	super destroy.

	sharedDesktopManager ifNotNil: [
		sharedDesktopManager destroy.
		sharedDesktopManager := nil.
		QLogger logInfo: 'Desktop manager cleared.'.
	].

	self documentManager: nil. 	"Owner manages lifecycle."! !

!QWindowedSession methodsFor: 'initialize' stamp: 'jcg 4/9/2010 17:29'!
discardIsland: oldIsland

	super discardIsland: oldIsland.
	self ogl ifNotNilDo: [ :opengl | opengl  discardIsland: oldIsland ].
	self jobHandler ifNotNilDo: [ :jh| jh discardIsland: oldIsland ].
	self audioSession ifNotNilDo: [ : as | as future discardIsland: oldIsland ].
	self videoSession ifNotNilDo: [ : vs | vs discardIsland: oldIsland ].

	animationSession ifNotNilDo: [ :as | as discardIsland: oldIsland ].
	contentManager discardIsland: oldIsland.! !

!QWindowedSession methodsFor: 'resource-fetch' stamp: 'bgf 9/15/2009 14:05'!
discardIslandResources: anIsland

	super discardIslandResources: anIsland.
	documentManager ifNotNil: [ 
		documentManager discardIslandResources: anIsland
	].! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/8/2008 18:53'!
documentManager
	^ documentManager! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 3/1/2009 09:59'!
documentManager: docMgr
	documentManager := docMgr.

	"The document manager delivers resource data from the server..."
	self stopScript: #onFetchComplete:.
	documentManager ifNotNil: [
		self startScript: #onFetchComplete: when: { documentManager . #fetchComplete }.
	].! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'das 4/10/2009 14:13'!
drop3DS: aFile event: anEvent2D editBox: inEditBox

	"3DS is the suffix for old (pre-Max) 3D Studio files"

	| loader model rval |
	
	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: '3D Studio file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:100000) ifFalse: [ ^ false ].
	].
	owner statusInfo: 'Importing AutoDesk 3D Studio file...'; waitTick.

	loader := TThreeDSParser new.
	loader baseDir: (FileDirectory forFileName: aFile name).
	loader addSearchPath: '.'.
	loader addSearchPath: 'textures'.
	loader classSpec: TThreeDSParser croquetSpec.
	
	self runBlock: [ model := loader parse: aFile. ]
			withActivityText: 'Importing AutoDesk 3D Studio scene '
			onSignal: #loaderActivity
			from: loader.

	model ifNotNil: [
		(self importConfirm: aFile name missingResources: loader missingResources)
		ifFalse: [ 
			model obliterate. 
			^ self
		].
		self importResourcesFrom: loader.
		model objectName: aFile localName.
		rval := self persistModel: model fileName: aFile localName event: anEvent2D editBox: inEditBox.
	].
	loader reset.
	^ rval.! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'bgf 6/7/2008 13:35'!
dropASC: aFile event: anEvent2D  editBox: inEditBox

" ASE and ASC are the same format"
	^self dropASE: aFile event: anEvent2D editBox: inEditBox.
	! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'das 4/10/2009 14:13'!
dropASE: aFile event: anEvent2D editBox: inEditBox

	| loader model rval |
	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: '3D ASE/ASC file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:200000) ifFalse: [ 	^ false ].
	].

	owner statusInfo: 'Importing ASE Scene ...'; waitTick.

	loader := TLoad3DSMax new.
	loader baseDir: (FileDirectory forFileName: aFile name).
	loader addSearchPath: '.'.
	loader addSearchPath: 'textures'.

	self runBlock: [ model := loader loadFile: aFile scale: 1.0 shadeAngle: 90.1 textureMode: GLModulate ]
		withActivityText: 'Importing ASE scene'
		onSignal: #loaderActivity
		from: loader.

	model ifNotNil: [
		(self importConfirm: aFile name missingResources: loader missingResources)
		ifFalse: [ 
			model obliterate. 
			^ self
		].
		self importResourcesFrom: loader.
		rval := self persistModel: model fileName: aFile localName event: anEvent2D editBox: inEditBox.
	].
	loader reset.
	^ rval.! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'hrs 5/11/2009 12:10'!
dropBVH: aFile event: anEvent2D  editBox: inEditBox
	| loader model anim aFrame thumb bonemapFilename map |

	owner statusInfo: 'Importing BVH skeleton/animation file...'; waitTick.
	loader := BVHParser new rootName: nil.

	aFile ascii; wantsLineEndConversion: true.
	loader parseStream: aFile.
	"There could be more than one motion in a BVH, but there won't be."
	loader motion motionName: (aFile localName upTo: $.).

	model := loader buildSkeleton.

	bonemapFilename := loader motion motionName, '.bonemap.xml'.
	(aFile directory fileExists: bonemapFilename) ifTrue: [
		map := TBoneMap initializeWithFileName: (aFile directory fullPathFor: bonemapFilename).
		model boneMap: map.
	].

	model showSkeleton: true.

	"Deliberately do this before any conversion to thumbnail.
	This will cache the real animation on-island for demonstration purposes.
	Yes, this takes more space on island, but the whole point of stand-alone skeletons
	is for demo and manipulation, so we really want it on island.
	Meanwhile, the animations that get passed around can be thumbnails, below."
	model demonstrateAnimation: model name.

	dropModelThumb ifTrue: [ "convert animation to a thumbnail"
		"BVH files are pretty well debugged, especially the SL ones, so let's go ahead and
		do this now so that we only have canonicalized data stored off-island."
		model canonicalize. 

		anim := model animations at: model name.
		aFrame := TAnimationFrame new data: anim; length: anim length.
		thumb := self persistAndThumbnailModel: aFrame fileName: aFile localName.
		thumb ifNil: [
			self error: 'The imported animation was empty.'.
		].

		"thumbnail default name includes .c3z. We don't want that, as there is a lot of machinery
		that depends on asking the animation (thumbnail or otherwise) what it's name is."
		thumb objectName: model name.
		model animations at: model name put: thumb.
		].
	^ self simplePersistModel: model fileName: aFile localName event: anEvent2D editBox: inEditBox.
! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'ar 4/16/2011 17:26'!
dropC3X: aFile event: anEvent2D editBox: inEditBox

	"C3X is the suffix for objects saved as templates using QInspector;
	 we may way to change that for general use.

	TODO: this should use a TImporter wrapped around the scene-reader,
	to handle textures consistently with other importers.
	(Or scene reader should be a TImporter, perhaps.)"

	| loader model writer xml |

	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: 'OpenQwaq C3X file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:200000) ifFalse: [ ^ false ].
	].
	self mainView ifNil: [ ^ nil ].
	owner statusInfo: 'Importing Scene...'; waitTick.

	"First read the file into a model. Better that it break here non-replicated than take everyone else down with us"
	loader := TSceneReader new.
	aFile ascii.
	model := loader readScene: aFile.
	
	(model isKindOf: TFrame) ifFalse: [
		self error: ('The scene file ', aFile localName, '\does not contain a stored object.' withCRs).
		^ nil
	].

	"Now do it properly, building things on the island.  
	 We cannot always import the model we just unpacked,
	 because internal startScript:s and the like don't cross into the island.
	 Instead... write it out again (also making the persistence latest&greatest) 
	 and send the xml into the island for unpacking."

	writer := TSceneWriter new.
	xml := writer storeScene: model.
	writer clear.
	model obliterate.
	model := nil.
	[ 
		| reader |
		reader := self activeIsland future new: TSceneReader.
		model := reader future readScene: xml.
		(model waitTimeoutSeconds: 90) ifFalse: [
			self error: 'Scene import timed out.'.
		].
		self mainView 
				insertModel: model
				forPosition: (
					anEvent2D ifNotNil: [ anEvent2D position - (self mainView origin) ] ifNil: [ nil ])
				editBox: inEditBox
				doScaling: (Sensor shiftPressed not).

		[ reader future clear ] on: Error do: [ : ex | QLogger logWarning: 'Reader cleanup: ', ex ].
	] on: Error do: [ : ex |
		self error: ('The scene file ', aFile localName, '\could not be imported.\\' withCRs, (ex description)).
	].

	^ model.! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'ar 4/16/2011 17:26'!
dropC3Z: aFile event: anEvent2D editBox: inEditBox

	"C3Z is the suffix for objects saved as templates using QInspector;
	 we may want to change that for general use."

	| pos archive  targetView rval |

	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: 'OpenQwaq C3Z file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:100000) ifFalse: [ ^ false ].
	].
	[
		owner statusInfo: 'Importing C3Z file...'; waitTick.
		archive := ZipArchive new.
		archive readFrom: aFile.
		targetView := self mainView ifNil: [ ^ nil ].

		rval := self importPersistedScene: archive named: aFile localName toIsland: (targetView activeIsland).
		rval ifNotNil: [

			"Pose the object based on the pointer drop position..."
			pos := anEvent2D ifNotNil: [ anEvent2D position - (targetView  origin) ]
							 ifNil: [ targetView center ].
			targetView insertModel: rval
					    forPosition: pos 
					    editBox: true 
					    doScaling: (Sensor shiftPressed not).
		].

	]on: Error do:[:exc |
		owner inform: 'This C3Z file could not be imported.\\' withCRs, exc description.
		QLogger logError: exc from: self.
	].
	^ rval.! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bgf 7/13/2008 18:16'!
dropCard: card event: anEvent2D

	"Drop the given QFileCard (far ref) into the space, attempting to
	 perform a logical drop to any target under the given (2D ui) position.
	 If no drop target is present, float the card in front of the avatar and 
	 make the avatar carry it."

	| distance pos |
	self mainView ifNil: [ ^ nil ].

	pos := (anEvent2D ifNotNil: [anEvent2D position] ifNil: [ nil ]).
	distance := self curDropDistanceFor: card atPosition: pos.

	self readyForEvents ifFalse: [ ^ nil ].

	self mainView
			dropCard: card 
			position: pos
			maxDropDistance: distance.

	! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'das 1/7/2010 00:22'!
dropDAE: aFile event: anEvent2D  editBox: inEditBox

	"DAE is the suffix for Collada files"

	| loader model rval |

	"Check for unusually hefty drops..."
	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: '3D Collada file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:100000) ifFalse: [ ^ false ].
	].
	[
		CProgressBar
			displayProgress:  'Importing ', (FileDirectory localNameFor: aFile name) 
			at: owner bounds center
			from: 1 to: 5
			during: [:bar |
				owner waitTick.
				bar value: 1.
				loader := TLoadDAE new.
				model := loader parseStream: aFile informing: [:done | bar value: 1.0 + done].
				bar value: 2.
				model flattenMultiChildrenInforming: [:done | bar value: 2.0 + done].
				model := model flattenGroups.
				bar value: 3.
				model ifNotNil: [
					(self importConfirm: aFile name missingResources: loader importer missingResources)
					ifFalse: [ 
						model obliterate. 
						^ self
					].
					self importResourcesFrom: loader importer informing: [:done | bar value: 3.0 + done].
					bar value: 4.
					documentManager
						useUploadProgress: [:done | bar value: 4.0 + done]
						during: [rval := self persistModel: model fileName: aFile localName event: anEvent2D editBox: inEditBox].
					bar value: 5.
				].
		]
	] on: Error do: [ :exc|
		owner inform: 'This Collada DAE file could not be imported.\\' withCRs, exc description.
		QLogger logError: exc from: self.
	].
	^ rval.! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'jcg 12/11/2009 13:06'!
dropDocument: aFile event: anEvent2D

	"If the file is a recognized document type, 
	 absorb it, attach it to a FileCard, and insert it into the shared space.
	 Returns true if the document is recognized and handled,
	 false if the file is not a document we can handle this way."

	| document card type shortName |

	"Document type recognition is simply a lookup based on the file suffix."
	type := (QDocumentType forFile: aFile) 	
		 ifNil: [^ false].

	(self droppedDocumentIsValid: aFile forType: type) ifFalse: [
		^ true  	"Get out with no further action."	
	].

	"Normal case - we've recognized  a SharedDocument.
	 Read the document data, and use the document manager
	 to upload it for sharing; this produces a QSharedDocument for insertion into the space."

	owner statusInfo: 'Importing ', type description, '...'.
	shortName := (aFile name copyAfterLast: (FileDirectory pathNameDelimiter)).
	QActivityCapture
			action: #DropDocument
			details:  (QActivityCapture fileDetails: shortName).

	"Ship it to the server."
	[
		"We have some restrictions about what MP4 files we handle...
		 point the user at some more information."
		type isMP4 ifTrue: [self popupMP4Info].

		document := (type docClass)
					uploadForImport: aFile docType: type session: (self).
		document ifNil: [^ false].
		document ctime: (Time utcSeconds) who: (self userLogin) what: 'Import'.

	] on: Error, CProgressCancel do: [ :exc |	
		(exc isKindOf: CProgressCancel) ifFalse: [
			QLogger logError: exc.
			self inform: ('Cannot import ', shortName,'\\(', exc description, ')') withCRs.
		].
		^ true
	].

	"Give the user a shared-document file card representing the imported content"
	card := QFileCard forDocument: document.
	card isTransient: true.
	card := TIslandCopier copy: card to: self activeIsland.
	self dropCard:card event: anEvent2D.
	^ true! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'ar 10/22/2009 12:50'!
dropFile: aFile event: anEvent2D editBox: inEditBox

	"Handle a file dropped via the OS"
	| mssg suffix |

	"We use the file type to construct a method symbol;
	if we have such a method, we use that to load the file.
	If the file name is of the form fileName.xyz, 
	then we construct the message to be  #dropXYZ:event:. 
	(The file type is capitalized.)

	NOTE- the handler methods must be able to deal with nil event2D
	this is the case from the File...Import... command."

	self readyForEvents ifFalse: [ ^ self ].
	self updateLastEventTime.

	"First allow the general-purpose SharedDocument loader a chance;
	 this will handle the drop if it recognizes the file type."
	(self dropDocument: aFile event: anEvent2D) ifTrue: [ ^self].

	suffix := (aFile name findTokens:'.') last asUppercase.
	suffix size = aFile name size ifTrue: [
		"No suffix found - just say no."
		QActivityCapture countAction: #DropDocumentUnknown details: ''.
		^ self inform: 'This file is of unknown type.\It cannot be imported.\(' 
					withCRs, aFile name, ')'.
	].

	"Fell through the shared-document check.
	 See if we have a specific selector-pattern for this type.
	 These selectors (see category 'drops-by-type' handle the drop process
	 for files whose suffix match their name."

	mssg := ('drop', suffix, ':event:editBox:') asSymbol.
	owner playSoundEffect: #acceptDrop.
	(self respondsTo: mssg) ifTrue:[
		[ [
			owner mouseCursor: Cursor wait.
			owner statusInfo: 'Importing...'; waitTick.
			^ self perform: mssg withArguments:{aFile. anEvent2D. inEditBox.}.
		] on:  Error do:  [ :ex|
			QLogger logError: ex from: self.

			owner mouseCursor: nil.
			self inform: 'This file could not be imported.\\' withCRs, ex description.
			^ nil.
		] ] ensure: [ 
			owner mouseCursor: nil.
		].
	].

	"Final fallthrough... no importer for this type."
	QActivityCapture countAction: #DropDocumentUnknown details: suffix.
	self inform: 'Files of type ', suffix, ' cannot be imported.'.
	^ nil.! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'das 1/7/2010 15:14'!
dropKMZ: aFile event: anEvent2D  editBox: inEditBox

	"KMZ is the suffix for Google Earth files (zipped collada)"

	| loader model rval |

	"Check for unusually hefty drops..."
	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: '3D KMZ file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:100000) ifFalse: [ ^ false ].
	].
	[
		CProgressBar
			displayProgress:  'Importing ', (FileDirectory localNameFor: aFile name) 
			at: owner bounds center
			from: 1 to: 5
			during: [:bar |
				owner waitTick.
				bar value: 1.
				loader := TLoadKMZ new.
				model := loader parseStream: aFile informing: [:done | bar value: 1.0 + done].
				bar value: 2.
				model flattenMultiChildrenInforming: [:done | bar value: 2.0 + done].
				model := model flattenGroups.
				bar value: 3.
				model ifNotNil: [
					(self importConfirm: aFile name missingResources: loader importer missingResources)
					ifFalse: [ 
						model obliterate. 
						^ self
					].
					self importResourcesFrom: loader importer informing: [:done | bar value: 3.0 + done].
					bar value: 4.
					documentManager
						useUploadProgress: [:done | bar value: 4.0 + done]
						during: [rval := self persistModel: model fileName: aFile localName event: anEvent2D editBox: inEditBox].
					bar value: 5.
				].
		]
	] on: Error do: [ :exc|
		owner inform: 'This KMZ file could not be imported.\\SketchUp models must be exported for ''Google Earth(4)''\\' withCRs, exc description.
		"owner isSealedApplication ifFalse: [self halt]."
		QLogger logError: exc from: self.
	].
	^ rval.! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
dropModelThumb

	^ dropModelThumb.! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
dropModelThumb: bool

	dropModelThumb := bool.! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'das 4/10/2009 14:14'!
dropOBJ: aFile event: anEvent2D editBox: inEditBox
	| loader model rval |

	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: '3D Wavefront file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),'''' size:200000) ifFalse: [ ^ false ].
	].
	owner statusInfo: 'Importing Wavefront OBJ file...'; waitTick.
	loader := TLoadOBJ new.
	loader addSearchPath: 'textures'.

	self runBlock: [ loader parseStream: aFile ]
		withActivityText: 'Importing Wavefront OBJ file'
		onSignal: #loaderActivity
		from: loader.
	model := loader frame.

	model ifNotNil: [
		(self importConfirm: aFile name missingResources: loader missingResources)
		ifFalse: [ 
			model obliterate. 
			^ self
		].
		self importResourcesFrom: loader.
		rval := self persistModel: model fileName: aFile localName event: anEvent2D editBox: inEditBox.
	].
	loader reset.
	^ rval.! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'ar 1/26/2009 18:17'!
dropPersonalShare: aFile event: anEvent2D

	"Handle a drop of a file which should be treated as personal share."

	| document card |

	document := QPersonalShareDocument new.
	document owner: self userLogin.
	document uri: aFile uri asString.
	card := QFileCard forDocument: document.
	card isTransient: true.
	card := TIslandCopier new import:(TIslandCopier new export:card)
									  to: self activeIsland.
	self dropCard: card event: anEvent2D.
	^true! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'bgf 6/7/2008 13:35'!
dropSKP: aFile event: anEvent2D editBox: ignore

	"SKP is the suffix for Sketchup Documents - we don' t handle these."

	owner inform: 'This is a SketchUp (.SKP) document.\\Use SketchUp to Export it as a 3D-Model,\and choose Google Earth(4) as the format.\\You can import the resulting .KMZ file.' withCRs
! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'das 9/16/2009 12:49'!
dropWRL: aFile event: anEvent2D editBox: inEditBox

	| name frame lights builder rval |

	inEditBox ifTrue:[
		(self checkDropSize: aFile tag: '3D VRML file ''', 
			(aFile name copyAfterLast:FileDirectory pathNameDelimiter),''''  size: 400000) ifFalse: [ ^ false ].
	].
	owner statusInfo: 'Importing VRML scene...'; waitTick.

	name := aFile name.
	aFile close.
	builder := VRMLCroquetBuilder new.
	builder importer baseDir: (FileDirectory forFileName: aFile name).
	builder importer addSearchPath: '.'.
	builder importer addSearchPath: 'textures'.

	self runBlock: [ frame := builder makeFrameFromVRMLFile: name ]
			withActivityText: 'Importing VRML Scene'
			onSignal: #loaderActivity
			from: builder.

	frame ifNotNil: [
		(self importConfirm: aFile localName missingResources: builder importer missingResources)
		ifFalse: [ 
			frame obliterate. 
			^ self
		].
		"Relocate textures"
		self importResourcesFrom: builder importer.
		 
		"Accumulating lights causes problems due to ogl strict limits;
		 render sorts them so it isn't fatal, but the visuals can be appalling.
		 For now, simply discard lights from the import and allow our lighting to prevail."
		lights := OrderedCollection new.
		frame allFramesDo: [:f|
			f isLight ifTrue: [ lights add: f ].
		].
		lights do: [:f | f destroy ].
		frame := frame collapse.
		rval := self persistModel: frame fileName: aFile localName event: anEvent2D editBox: inEditBox.
	].
	builder importer reset.
	^ rval.! !

!QWindowedSession methodsFor: 'drops-by-type' stamp: 'bgf 5/12/2010 20:49'!
dropXML: aFile event: anEvent2D editBox: inEditBox

	| name frame loader archive motionsSkel animatedMeshClass hasNoVertexAnimations newFrame |

	"A quick and dirty development-time test for dropping ogre xml files.
	 Drop the mesh.xml file, not the skeletion.xml file !!"

	owner statusInfo: 'Importing Ogre Scene...'; waitTick.
	name := aFile name.
	aFile close.
	loader := TLoadOgreMesh new.
	animatedMeshClass :=  TReplicatedAnimatingFigure.
	self runBlock: [ 
				loader initializeWithFileName: name scale: 1.0.
				frame := loader frame: animatedMeshClass new 
				]
		withActivityText: 'Importing Ogre Scene'
		onSignal: #loaderActivity
		from: self.

	frame ifNotNil: [
		(self importConfirm: aFile localName missingResources: loader missingResources)
			ifFalse: [ 
				frame obliterate. 
				^ self
			].
		self importResourcesFrom: loader. "Relocate textures"
		hasNoVertexAnimations := frame vertexPoses isEmpty.

		TAnimationManager useFigures ifTrue: [
			hasNoVertexAnimations ifTrue: ["we don't canonicalize vertex right yet"
			"Sharing animations is not going to work unless the shared figure data is canonical."
				frame figure scale: frame scaleForAvatar.
				frame figure: frame figure. "just to reset skeleton and such"
				frame canonicalize.  "standard bone names and such."
				].

			"Now shift over to a lightweight model. We couldn't do that above because such
			models don't know how to canonicalize (which requires a built-out skeleton)."
			dropModelThumb ifTrue: [
				newFrame := TAnimatedFigureModel new.
				newFrame figure: frame figure.
				newFrame boneMap: frame boneMap.
				newFrame head localTransform: frame head localTransform.
				newFrame head index: frame head index.
				newFrame head parent localTransform: frame head parent localTransform.
				newFrame laserSite localTransform: frame laserSite localTransform.
				newFrame laserSite parent localTransform: frame laserSite parent localTransform.
				frame := newFrame.
				].
			].

		dropModelThumb ifTrue: [ "convert skin (mesh) to a thumbnail"

			"Turn the (already canonical) figure into a thumbnail, so that it can be 
			shared among all instances, on-island and off."
			frame figure: 
				(self persistAndThumbnailModel: frame figure fileName: aFile localName).

			"Add the Standard-Motions. We could get arbitrarilly fancy here. Let's not, yet."
			hasNoVertexAnimations ifTrue: ["Cannot yet canonicalize with vertex animations"

				"Here's a fat zip with a skeleton that has all the motions we want."
				archive := ZipArchive new readFrom: 
					(FileDirectory default pathFromURI: 
						'Objects/Qwaq/Avatars/Standard-Motions.c3z').
				"Get the skeleton. We're still harness-side/ocean at this point."
				motionsSkel := self extractPersistedScene: archive 
									named: 'Standard-Motions'
									asModel: true.
				motionsSkel ifNil: [ ^ nil ].

				"Add a list of thumbnails to our main imported frame, 
				which refer to the motions we just added."
				motionsSkel  giveAnimationsTo: frame.
				motionsSkel obliterate.  "No need to actually import the dummy motionsSkel."
			]].
		^ self simplePersistModel: frame fileName: aFile localName event: anEvent2D editBox: inEditBox.
		].
	^ nil.! !

!QWindowedSession methodsFor: 'private' stamp: 'ar 4/16/2011 17:27'!
droppedDocumentIsValid: aFile forType: aDocType
	"Check for reasons we might reject the given document file.
	 Inform the user, and return false, if there is problem."
	| valid |

	"Type-specific checks for file validity."
	valid := true.
	true caseOf: {
			[ aDocType isSampledSound ] -> [ valid := self soundFileIsValid: aFile ].
			[ aDocType isMpegTwo ] -> [ valid := self mp2FileIsValid: aFile ].
		} otherwise: [].
	valid ifFalse: [ ^ false ].

	"Thanks Bill!!  We really appreciate your fine work."
	aDocType isMicrosoftXML ifTrue: [
		^ self confirm: '', aFile localName, (String cr), (String cr),
				'is in Microsoft Office Open XML Format.', (String cr), (String cr),
				'You can open this document with OpenQwaq MultiShare,', (String cr),
				'but any changes you make will not be saved.'.
	].

	"Check for unusually hefty drops..."
	(self checkDropSize: aFile tag: (aDocType description) size: 400000) ifFalse: [ ^ false ].

	^ true
	! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
endBenchTest

	endBenchTest := true.! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/7/2010 11:38'!
ensureCopy: resRef fromService: sourceSvc toService: targetSvc
	"Ensure that the target forum has available a copy of the given resource,
	 copying it from the source service if necessary.
	 The resRef argument is a QAbstractDocument, or file-texture, or frame-thumb...
	 all of which satisfy some parts of the document sharedName protocols."

	(documentManager serverFileExists: '/services/', (sourceSvc name), '/', (resRef sharedName))
	ifFalse: [
		QLogger logWarning: 'Resource: ', (resRef sharedName), ' not found in ', (sourceSvc name).
		^ false
	].

	(documentManager serverFileExists: '/services/', (targetSvc name), '/', (resRef sharedName))
	ifTrue: [
		^ true
	].

	[
		QLogger logInfo: 'Copying ', (resRef sharedName), ' to ', (targetSvc name).
		owner statusInfo: 'Paste: Copying ', (resRef sharedName), '...'.
		owner serviceMgr
			copyFiles: '/services/', (sourceSvc name), '/', (resRef sharedName)
			to: '/services/', (targetSvc name), '/', (resRef sharedName).
		^ true
	] on: Error do: [  :exc |
		QLogger logWarning: 'Cannot copy ', (resRef shortName), ': ', (exc description).
		^ false
	].
			 ! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/7/2010 12:24'!
ensureCopyResources: resList fromService: sourceSvc toService: targetSvc
	"Copy-Paste support.
	 Ensure that targetService has all the resources referred to in the resource list,
	 copying from the srcService if necessary.
	 The resList consists of abstract-documents, file textures, and frame-thumbs
	 (resources that carry sharedNames identifying server-side resource backing files.)
	 Return a list of resources that could not be copied."

	| absentList doneSet fname |
	self readyForEvents ifFalse: [ 
			QLogger logInfo: 'Cannot copy resources - session not ready.'.
			^ resList ].

	absentList := OrderedCollection new.
	doneSet := Set new.
	resList 
		select: [ : res | (res isRenamableContent not) and: [ res isDocumentInForumStorage]  ]
		thenDo:[:resRef|
			fname := resRef fileName.
			"The same textures may be referenced multiple times, but we only read it once"
			(fname notNil and: [ (doneSet includes: fname) not ])
			 ifTrue: [
				doneSet add: fname.
				(self ensureCopy: resRef fromService: sourceSvc  toService: targetSvc) ifFalse: [
					absentList add: resRef
				].
			].
		].
	^absentList ! !

!QWindowedSession methodsFor: 'resources' stamp: 'ar 4/16/2011 17:47'!
ensureDocument: docRef fromService: sourceSvc toService: targetSvc
	"Private worker method for ensureDocuments....
	 Make sure the target forum has a fresh copy of the given document,
	 copying it from the source service.
	 The docRef argument is a QAbstractDocument.
	 The result is nil if the document is placed into the new service under its original shared name;	
 	   if the new document has a different name, then a pair
		{ original-shared-name . new-shared-name }.
	   is returned."

	| clone |
	(documentManager serverFileExists: '/services/', (sourceSvc name), '/', (docRef sharedName))
	ifFalse: [
		QLogger logWarning: 'Document: ', (docRef sharedName), ' not found in ', (sourceSvc name).
		(FileDoesNotExistException fileName: (docRef shortName)) signal
	].

	clone := (TSceneReader new duplicateScene: docRef).
	clone serviceURL: (sourceSvc url).
	clone class copy: clone toActiveForum: (owner).
	QLogger logInfo: 'Copied ', (docRef sharedName), ' to ', (targetSvc name), ' as ', (clone sharedName).
	(clone sharedName = docRef sharedName) ifFalse: [
		^ { (docRef sharedName) . (clone sharedName) }
	].
	^ nil
! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/15/2010 18:59'!
ensureDocuments: resList fromService: sourceSvc toService: targetSvc
	"Copy-Paste support.
	 Ensure that targetService has all the documents referred to in the resource list,
	 copying from the srcService if necessary.
	 The resList consists of abstract-documents, file textures, and frame-thumbs
	 (resources that carry sharedNames identifying server-side resource backing files.)
	 Return a list string pairs, giving the original and result sharedNames for documents
	 that were renamed to give the result."

	| renames doneSet fname |
	self readyForEvents ifFalse: [ 
			QLogger logInfo: 'Cannot copy documents - session not ready.'.
			^ resList ].

	doneSet := Set new.
	renames := Array streamContents: [ : strm |
		resList 
			select: [ :  res | res isRenamableContent and: [ res isDocumentInForumStorage]  ]
			thenDo:[:resRef|
				fname := resRef fileName.
				"If the document was repeated in the source material, we rename it once only."
				(fname notNil and: [ (doneSet includes: fname) not ])
				 ifTrue: [
				doneSet add: fname.
				[ 
					(self ensureDocument: resRef fromService: sourceSvc  toService: targetSvc) 
					 ifNotNilDo: [ : rename |
						strm nextPutAll: rename
					].
				] on: Error do: [ : exc |
					QLogger logWarning: 'Cannot copy ', resRef sharedName, ': ', (exc description).
				].
			]]].
	^renames ! !

!QWindowedSession methodsFor: 'initialize' stamp: 'bgf 6/7/2008 13:35'!
eventClass

	^QCroquetEvent! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 3/21/2009 16:17'!
existingResourceFile: sha forIsland: aServiceIsland

	"If the given sha corresponds to a resource file in the active service,
	 that is present and in use in the given island,
	 return its service-side file name (typically resources/<sha>.suffix)
	 If not, return nil."

	| satisfied |

	"Special case to reduce needless traffic when loading presupplied avatars;
	 if we recognize the sha as a bundled avatar model, we know its filename."
	(QAnimatedFigureDetails bundledModelFileForSHA: sha) ifNotNilDo: [ : uri | ^ uri ].

	satisfied := (self allContentInIsland: aServiceIsland).
	(satisfied at: sha ifAbsent: [nil]) 	
	 ifNotNilDo: [ : existingName |  "I.e. we have such a resource."
		"Confirm that the file for this resource is available in the relevant forum...
		 we may have cached it from a download from a different forum."
		(TContentManager isBundledFileName: existingName) ifTrue: [
			^ existingName
		].
		 (self documentManager serviceFileExists: existingName)  ifTrue: [
			^ existingName
		]].

	^ nil! !

!QWindowedSession methodsFor: 'resources' stamp: 'hrs 6/10/2008 12:32'!
extractPersistedScene: archive named: aName asModel: asModel

	"Given the zip-archive loaded from a .c3z file
 	 (whose members are a scene.xml, and the associated resource files)
	 extract the TFrame model.  Answers either the model (if asModel is true)
	 or the xml String (suitable for importing).

	 Extract and upload all the payload resources into the current forum.

	 Returns nil if there is are missing resources and the user cancels;
	 may raise errors for other problems."

	| loader model writer missingResources xml member |

	self mainView ifNil: [ 
		self error: 'Not ready for scene import.'.
	].

	member := ((archive memberNamed: 'tframe.xml') 		"Save-As-Template files"
			 	ifNil: [archive memberNamed: 'scene.xml']).	"Most others (e.g. inventory)"

	member ifNil: [ self error: 'The scene file ', aName, '\does not contain a scene record.' ].

	"Read in the XML as a harness-side object first. If it is OK, then we can replicate it.
 	 (We don't want to expose the replicated space to arbitrary errors from bad files.)"
	loader := TSceneReader new.
	model := loader readScene: member contents asString.
	(model isKindOf: TFrame) ifFalse: [
		self error: ('The scene file ', aName, '\does not contain a Croquet object.' withCRs).
	].

	"Re-encode as XML (which extracts the resource list) "
	writer := TSceneWriter new.
	writer initResourceList.
	xml := writer storeScene: model.
	
	"Import all the bundled resources."
	missingResources := self loadResources: writer resourceList zip: archive.

	"Don't complain about bundled content, since it is not packed into the archive.
	 And collect the list of names, rather than the kind.name pairs,
	 as that is what importConfirm: expects."
	 missingResources := (missingResources 
		reject: [:entry | TContentManager isBundledFileName:(entry second)])
		collect: [:entry | entry second].

	(self importConfirm: aName missingResources: missingResources) ifFalse: [
		^nil
	].

	^asModel ifTrue: [model] ifFalse: [xml]! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 5/27/2010 09:13'!
focalDocumentName
	^ focalDocumentName! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 5/27/2010 09:43'!
focalDocumentName: aName

	"The name of a document that the user is currently input-focused upon in an application,
	 or nil the rest of the time.  
	 The focal dispatch rules set this, and the flags that set us back to keyboard driving clear it."

	(aName = focalDocumentName) ifFalse: [
		aName ifNil: [ 
			QActivityCapture action: #DocumentUnfocus details: focalDocumentName.
			QLogger logInfo: 'Releasing focus from ', focalDocumentName.
		] ifNotNil: [
			focalDocumentName ifNotNil: [
				QActivityCapture action: #DocumentUnfocus details: focalDocumentName.
			].
			QLogger logInfo: 'Focusing on ', aName.
			QActivityCapture action: #DocumentFocus details: aName.
		].
		focalDocumentName := aName.
	].

! !

!QWindowedSession methodsFor: 'resource-fetch' stamp: 'bgf 5/26/2010 10:50'!
forgetThumb: thumbRef
	"The thumb (texture or model) has been removed from the resource table of its island."
	super forgetThumb: thumbRef.

	(thumbRef get: #sha) ifNotNilDo: [ : sha |
		contentManager ifNotNil: [ : cm |
			(cm entryForSHA: sha) ifNil: [
				"If no pending content-manager entry, forget any downloads.
			 	 This is stub code... the content manager does not yet expunge entries
	 			 since there is no requirement to include the referrers when adding content."
				documentManager ifNotNil: [ : dm | dm unqueueSHA: sha forThumb: thumbRef ]
			]]].! !

!QWindowedSession methodsFor: 'render' stamp: 'bgf 2/21/2009 13:23'!
globalBounds: bnds
	"Create or reset the OpenGL window.
	Here we do Qwaq-specfic initialization that requires a valid ogl."

	owner doOpenGL ifFalse: [ ^ self ].
	super globalBounds: bnds.
	ogl ifNil: [^self].
	ogl hasVBO ifFalse: [
		QLogger logInfo: 'No VBO. Using software animation.'.
		animationSession class accelerationManager hardwareAnimationType: #softwareAnimation
			 reset: false.
		].! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/8/2008 18:54'!
hubID
	"Helper to abstract away the fact that the hub is currently called 'mediaController'"
	^mediaController sessionID! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bgf 4/1/2010 19:03'!
importConfirm: filename missingResources: set

	"Given a file and a list of missing resource names,
	 confirm that the import should proceed."

	| msg array count cursor |

	set isEmpty ifTrue: [ owner waitTick. ^ true ].		"Nothing missing... implicit OK"

	cursor := owner mouseCursor.
	array :=  set asArray. 		"So we can get the first 10"
	count := array size.

	"Fussy because we don't apply withCRs to the filenames themselves.."
	msg := 'Missing resources in model ', 
			(filename copyAfterLast: FileDirectory pathNameDelimiter),
			('\\The following resources were not found:' withCRs).
	(array first: (count min:10)) do: [ :res| 
		(res includes: FileDirectory pathNameDelimiter) ifTrue: [
			msg := msg, ('\   ' withCRs), (res copyAfterLast: FileDirectory pathNameDelimiter).

		] ifFalse: [
			msg := msg, String cr, String tab, res		
		].
	].
	array size > 10 ifTrue: [ msg := msg,  ('\   ...' withCRs)].
	msg := msg, ('\\Proceed with import?' withCRs).

	[
		owner  mouseCursor: nil.
		^ self confirm: msg
	] ensure: [
		owner mouseCursor: cursor.
		owner waitTick.		"Clear the confirm dialog after ok or cancel."
	].! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'ar 10/22/2009 12:48'!
importDocument: aFile island: targetIsland

	"If the file is a recognized document type, 
	 import it to the target island, attach it to a FileCard, and return the file card.
	If anything goes wrong, return nil"
	| document card type shortName |

	"Document type recognition is simply a lookup based on the file suffix."
	type := (QDocumentType forFile: aFile) ifNil: [^nil].

	"Normal case - we've recognized  a SharedDocument.
	 Read the document data, and use the document manager
	 to upload it for sharing; this produces a QSharedDocument for insertion into the space."

	owner statusInfo: 'Importing ', type description, '...'.
	shortName := (aFile name copyAfterLast: (FileDirectory pathNameDelimiter)).
	QActivityCapture
			action: #ImportDocument
			details:  (QActivityCapture fileDetails: shortName).

	"Ship it to the server."
	[
		document := (type docClass)
					uploadForImport: aFile docType: type session: self.
		document ifNil: [^nil].
		document ctime: (Time utcSeconds) who: (self userLogin) what: 'Import'.

	] on: Error, CProgressCancel do: [ :exc |	^nil].

	"Create the file card"
	card := QFileCard forDocument: document.
	^TIslandCopier copy: card to: targetIsland.
! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 12/7/2008 19:41'!
importImageAsTexture: aFile
	"Attempt to read the file as a texture graphic;
	 import the graphic file to shared-document storage for the current forum,
	 and return a TFileTexture referring to that imported copy of it.   Raises errors."

	^ self importImageAsTexture: aFile shrinkFit: true! !

!QWindowedSession methodsFor: 'resources' stamp: 'ar 10/22/2009 14:28'!
importImageAsTexture: aFile shrinkFit: bShrink

	^ self importImageAsTexture: aFile 
			shrinkFit: bShrink 
			extension: (TTextures extensionForFile: aFile name)
			island: self activeIsland
! !

!QWindowedSession methodsFor: 'resources' stamp: 'ar 10/22/2009 14:28'!
importImageAsTexture: aFile shrinkFit: bShrink extension: extension island: targetIsland

	"Attempt to read the file as a texture graphic;
	 import the graphic file to shared-document storage for the current forum,
	 and return a TFileTexture referring to that imported copy of it.   Raises errors."
	| texture doc existing |

	"Import the texture to trap any errors (broken files etc)."
	texture := TContentManager default 
		loadTexture: aFile name 
		mipmap: true 
		shrinkFit: bShrink 
		extension: extension.
	texture ifNil: [
		self error:'Cannot import that file type as an image.'.
	].

	"If we already use this texture in per-forum storage, use that."
	existing := self existingResourceFile: (texture sha) forIsland: targetIsland.
	(existing notNil and: [ (existing beginsWith: '/') not])  ifTrue: [
		texture fileName: existing.
		^ texture
	].

	"Upload file and assign shared name.
	 Note that these are textures, but we put them into the sharedDocuments area,
	 since they are user-supplied images, rather than model-support resources."
	doc := self documentManager 
				uploadDocumentFile: aFile collisions: #makeUnique.
	doc ifNil: [
		self error: 'The service host did not accept the file.'
	].
	texture fileName: doc sharedName.
	^ texture! !

!QWindowedSession methodsFor: 'resources' stamp: 'hrs 6/10/2008 12:37'!
importPersistedScene: archive named: aName

	"Given the zip-archive loaded from a .c3z file
 	 (whose members are a scene.xml, and the associated resource files)
	 import it into the given space as a TFrame, returning a far ref to the result.
	 Extract and upload all the payload resources into the current forum.
	 Returns nil if there is are missing resources and the user cancels;
	 may raise errors for other problems."

	^self importPersistedScene: archive named: aName toIsland: self mainView space island.
! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 5/12/2010 20:48'!
importPersistedScene: archive named: aName toIsland: anIsland
	"Given the zip-archive loaded from a .c3z file
 	 (whose members are a scene.xml, and the associated resource files)
	 import it into the given space as a TFrame, returning a far ref to the result.
	 Extract and upload all the payload resources into the current forum.
	 Returns nil if there is are missing resources and the user cancels;
	 may raise errors for other problems."

	| reader obj xml |

	xml := self extractPersistedScene: archive named: aName asModel: false.
	xml ifNil: [ ^ nil ].

	 "Finally read the scene within the space (its resources are waiting to be used.)"
	reader := anIsland future new: TSceneReader.
	reader future sessionID: (self sessionID).
	obj := reader future readScene: xml.
	obj wait.
	(obj refIsKindOf: TGeometry) ifTrue:[ obj future fileName: aName ].
	^ obj! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bf 11/20/2008 18:13'!
importResourcesFrom: importer

	^self importResourcesFrom: importer informing: nil! !

!QWindowedSession methodsFor: 'resources' stamp: 'das 3/27/2009 19:01'!
importResourcesFrom: importer forIsland: aServiceIsland informing: progressBar
	"Upload all the loaded resources (textures) from the importer.
	Ensure that the new textures get relocated to the import location 
	by any means necessary.

	Detect textures we already have present as resources,
	(verifying both sha-match, and that the sha we have cached is actually present
      in service/resources as a file that matches the name we want to use.)
	We don't upload those, we create references to them.
	TODO: This is the place where we *REALLY* want to convert BMP files etc."

	| localName usedNames file data satisfied gotIt total done nm |

	usedNames := Set new.
	satisfied := (self allContentInIsland: aServiceIsland).
	total := importer loadedResources size asFloat.
	done := 0.

	importer loadedResources keysAndValuesDo: [:filePath :thumbRef |
		((TContentManager isBundledFileName: filePath) or:[thumbRef == nil])
		  ifFalse:[
			"We only upload each unit of resource data once per island,
			 correcting the names in the in-island referencers  if necessary.
			 No need to upload again if we already have it in for this service."
			gotIt := false.
			(satisfied at: (thumbRef get:#sha) ifAbsent: [nil]) 	"I.e. we have such a resource."
				ifNotNilDo: [ : existingName |
					"Confirm that the file for this resource is available in the relevant forum...
					 we may have cached it from a download from a different forum."
					(self documentManager serviceFileExists: existingName)  ifTrue: [
						thumbRef class = TFileTexture ifTrue:[
							thumbRef fileName: existingName.
							gotIt := true ]]].
			gotIt ifFalse: [
				"Note: file mustn't be nil here;  after all, we just imported the whole model..."
				file := FileDirectory default readOnlyFileNamed: filePath.
				[file binary. data := file contents] ensure:[file close].
				"Grab the local name from the file"
				localName := FileDirectory localNameFor: filePath.
				usedNames add: localName.

				"Upload file and assign the resulting shared resouce name
 				 (the manager assigns a sha-based name to the file it uploads;
				 the original name is supplied for interactive feedback.)"
				self documentManager
					useUploadProgress: (progressBar ifNotNil: [
						[:fraction | progressBar value: (done+fraction) / total]])
					during: [ nm := self documentManager
						   				uploadResource: data 
										shortName: localName 
										sha: (thumbRef sha) ].
				nm ifNil: [ self error: 'Could not upload ', localName].
				thumbRef class = TFileTexture ifTrue:[
					thumbRef fileName: nm.
				].
				"Note it , so we do one upload for items repeatedly used in the model."
				satisfied at: (thumbRef get: #sha) put: nm.
			].
		].
		progressBar ifNotNil: [progressBar value: (done := done + 1) / total].
		].	


	! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bf 11/20/2008 18:12'!
importResourcesFrom: importer informing: progressBar

	^ self importResourcesFrom: importer forIsland: (self activeIsland) informing: progressBar! !

!QWindowedSession methodsFor: 'initialize' stamp: 'jcg 1/4/2010 16:15'!
initialize
	super initialize.

	reconnecting := false.
	appLaunchMutex := Semaphore forMutualExclusion.
	self class preloadCachePaths do: [ : path |
		cacheMgr addExtraDirectory: path
	].
	dropModelThumb := true.
	keyboardDriving := true.
	jobHandler := QJobHandler new.

	animationSession := TAnimationManager new.
! !

!QWindowedSession methodsFor: 'initialize' stamp: 'bgf 6/8/2008 19:05'!
initializeRandomStream
	"Don't complain in Qwaq products"
	| seed |
	seed := ByteArray new: 128.
	TCryptoRandom gatherEntropy: seed.
	Island default randomStream seedVector: seed.
! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 7/31/2008 14:06'!
isAvatarSpecificResource: resRef

	"Qwaq policy - resources like user faces are kept in /users/public/, etc.
	 and those are not to be uploaded as part of fat island.zips when saving templates."

	^ ((resRef get:#fileName) ifNil: ['']) beginsWith: '/users/'! !

!QWindowedSession methodsFor: 'testing' stamp: 'jcg 10/17/2009 08:54'!
isDemon
	"#isDemon asks whether the avatar is a demon, and #isDemonic asks whether it is currently acting demonically (i.e. not idle)."
	^owner isDemon! !

!QWindowedSession methodsFor: 'testing' stamp: 'jcg 10/17/2009 08:55'!
isDemonic
	"#isDemon asks whether the avatar is a demon, and #isDemonic asks whether it is currently acting demonically (i.e. not idle)."
	^owner isDemonic! !

!QWindowedSession methodsFor: 'event handling' stamp: 'bgf 6/19/2008 12:36'!
isReadonlyUser

	"Indicates a user who cannot do -anything- to interact with stuff in the space.
	 This is (currently) the only implementation of action-access-control
	 for inspace objects and delivery of events into the space.
	 When the event-handling regime and all the in-space objects are enhanced
	 to honor user roles in there, then this will unfold into more specific capabilities."

	^ owner isReadonlyUser! !

!QWindowedSession methodsFor: 'islands' stamp: 'bgf 6/8/2008 19:04'!
islandsInServiceUrl: svcURL

	"Given a normalized service url, 
	 return all TIslands that are part of that service.
	 (Usually there is only one.)"	
	| len |
	len := svcURL size.
	^ islandsByUrl keys 
			select: [ :key | (key beginsWith: svcURL) 
					 	   and: [ (key size = len) or: [(key at: (len+1)) = $/]]] 
			thenCollect: [ : key | islandsByUrl at: key ].! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
jobHandler
	^ jobHandler! !

!QWindowedSession methodsFor: 'event handling' stamp: 'bgf 6/17/2008 19:20'!
keyboardDriving	
	"Readonly users are always driving; they cannot enter input mode
	 because they are not allowed to input into apps in the space."

	^ keyboardDriving or: [ owner isReadonlyUser ].! !

!QWindowedSession methodsFor: 'event handling' stamp: 'bgf 5/27/2010 09:14'!
keyboardDriving: aBool
	"True when the keyboard is free for use as a driving console,
	 false when key strokes are focused into an application in the space."

	"Readonly users are always driving; they cannot enter input mode
	 because they are not allowed to input into apps in the space."

	self isReadonlyUser ifFalse: [
		keyboardDriving := aBool.
		self signal: #keyboardDrivingChanged.
	].

	aBool ifTrue: [ self focalDocumentName: nil ].! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'RJT 5/13/2010 13:40'!
load3DFile: aFileName
	"Handle a file loaded via a script"
	| mssg suffix aFile rval requiresEditBox |

	"We use the file type to construct a method symbol;
	if we have such a method, we use that to load the file.
	If the file name is of the form fileName.xyz, 
	then we construct the message to be  #dropXYZ:event:. 
	(The file type is capitalized.)

	NOTE- the handler methods should be able to handle use
	with anEvent2D = nil; this is the case from the File...Import... command."

	suffix := (aFileName findTokens:'.') last asUppercase.
	suffix size = aFileName size ifTrue: [
		"No suffix found - just say no."
		^ nil.
	].
	"See if we have a specific selector for this type"
	mssg := ('drop', suffix, ':event:editBox:') asSymbol.
	(self respondsTo: mssg) ifTrue:[
		[ [
			"owner mouseCursor: Cursor wait.
			owner statusInfo: 'Importing...'; waitTick."
			(FileDirectory default fileExists:aFileName)ifFalse:[^nil].
			aFile := FileDirectory default readOnlyFileNamed: aFileName.
			requiresEditBox := #('XML') includes: suffix.
			rval := self perform: mssg withArguments:{aFile. nil. requiresEditBox.}.
		] on:  Error do:  [ :ex|
			QLogger logError: ex  from: self.
			"owner mouseCursor: nil."
			^ nil.
		] ] ensure: [ 
			"owner mouseCursor: nil."
		].
	].
	^ rval.
! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/5/2010 20:43'!
loadResources: resList zip: archive
	"Import the given list of resouces from the given (c3z) archive,
	 uploading them to the server as resources in this forum.
	 Returns a list of pairs (type, name) for resources that were not found in the zip."

	| absentList doneSet fname mbr data |

	absentList := OrderedCollection new.
	doneSet := Set new.
	"Repeatedly wait for the resources since this is the place that blows up if
	we get a disconnect during the scene construction above. Reconnect
	will work fine but we need to make sure we pick up the right future send.
	Of course, the user can always cancel this loop by hitting the cancel button."
	resList do:[:resRef|
		fname := resRef fileName.
		"The same textures may be referenced multiple times, but we only read it once"
		(fname notNil and: [ (doneSet includes: fname) not ])
		 ifTrue: [
			doneSet add: fname.
			mbr := archive memberNamed: fname.
			(mbr isNil and:[fname beginsWith: '/']) ifTrue:[
				"The zip archive doesn't like absolute path names for some reason.
				Try to find the texture again stripping of the leading slash"
				mbr := archive memberNamed: fname allButFirst.
			].

			"We cannot update names in the model
	 		 (as for all other resource-loading paths) as we have no access to the original.
			 The best we can do right now is make sure paths aren't changed
	 		 for all the resources the model depends on.
			 (This is OK as all those resource uploads were sha-named in the first place.)"
			(fname beginsWith: 'resources/') ifTrue: [ 
				 fname := fname copyAfter: $/.
			].
			mbr 
				ifNil:[absentList add: {resRef resourceType . fname} ]
				ifNotNil:[
					data := mbr contents.
					self documentManager 
						uploadResource: data 
						shortName: fname 
						sha: fname.   "Forces literal use of this name."
					resRef sha ifNotNil: [ :  sha |
						TContentManager default cacheManager cacheIfAbsent: data to: sha.
					].
				].
		].
	].
	^absentList! !

!QWindowedSession methodsFor: 'resource-fetch' stamp: 'bgf 4/7/2009 17:25'!
markNeededDownloads
	"Pass on the 'needed' pending content from the content manager to 
	 the document manager, 
	 so that the download queue fetches visible things first."
	contentManager entriesAwaitingData do: [ : entry |
		documentManager ifNotNil: [ 
			"I.e. guarded against really sudden dis-reconect oddities."
			documentManager noteNeededThumbs: (entry referringThumbs).
		].
	].! !

!QWindowedSession methodsFor: 'private' stamp: 'ar 4/16/2011 17:27'!
mp2FileIsValid: aFile 
	"We no longer support Mpeg-2 import - it creates hard-to-load forums.  Just say no."
	| dlog |
	dlog := CUserDialog new.
	dlog label: 'MPEG-2 Notice'.
	dlog message:
'
MPEG-2 files cannot be imported into OpenQwaq.

If you convert the file to an MP4 (H.264) format, 
the result can be imported and shared efficiently.
'.
	dlog messageText allowEditing: false.
	dlog messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	dlog messageText allowEditing: false.

	dlog addButton: 'Cancel'				value: #ok.
	(dlog addButton: 'More Information...'	value: #moreInfo) width: 150.

	(dlog run: nil centerAt: nil) caseOf: {
		[#moreInfo]-> [ owner openMP4InfoPage ].
		[#ok ] -> [].
	} otherwise: [].

	^ false

! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 9/8/2009 14:58'!
numRemoteApps
	"The number of remote applications in use."
	^ (self embeddedApps values count: [ :app | app countsAsRemoteApp ])! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 3/21/2009 17:32'!
onFetchComplete: docRequest
	"A document-manager fetch request completed;
	 if the requestor was a thumbnail, complete the load process for it."
	
	docRequest requestType caseOf:{
		[#thumb] -> [
			owner flowControlFlag: true.
			self processOneDelivery: docRequest.
			documentManager ifNil: [ ^ self].
			owner flowControlFlag: (documentManager countToFetch > 0).
		].
	} otherwise:[].
! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 19:01'!
openAL

	^ audioSession ifNotNil:  [ audioSession openAL ]! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
openAL: anOpenAL
	"Should not implement; OpenAL object is created by our 'audioSession'. "
	self shouldNotImplement! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
owner 
	"The participant UI."
	^ owner! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
owner: aParticipantUI
	owner := aParticipantUI.
	jobHandler owner: aParticipantUI.! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bgf 3/14/2009 00:30'!
persistAndThumbnailModel: import fileName: fname 

	"Save the given imported model to active-service storage as a c3z resource, 
	 generate and return a TFrameThumb representing it.

	 Note: this method assumes ownership of the 'import' frame,
	 which is assumed to be un-referenced and forgotten by callers.
	 (E.g. import may be obliterated by this method.)
	 Send in a copy if you have other users for import.

	 (This is for efficiency... so we're not duplicated big meshes unnecessarily;
	 the content manager needs an unencumbered copy of the import mesh,
	 so we assume that the one we're given is OK to take.)"

	| c3zName masterFrame data sha myFrameHolder thumb cube resFile |

	"Set to canonical identity stance so persistent-data comparisons can work."
	import localTransform: Matrix4x4 identity.
	import allUpdateShape.
	import boundingBox extent max  > 0 ifTrue:[

		c3zName := ((fname copyUpToLast: $.) ifEmpty: [fname]), '.c3z'.
		masterFrame := import.
		masterFrame objectName: (c3zName copyUpToLast: $.).
		data := TSceneWriter new storeScene: masterFrame.

	"Compute the sha hash, without zipping (which has a timestamp),
	 (variation in the timestamp won't affect the usability of a cached c3z for a thumb,
	 if everything else about the model-persistence matches.)"

		sha := TContentManager generateSHA: data.
		myFrameHolder := TContentManager default lookup: sha.

	"If it does not already exist, then persist it"
		myFrameHolder ifNil:[
			data := (TSceneWriter new storeSceneAsZip: masterFrame) contents. 
					 "zipped for upload"
			myFrameHolder := TFrameHolder hold: masterFrame.
			self contentManager at: sha ifAbsentPut: myFrameHolder.
			self contentManager cacheManager cache: data to: sha.

		] ifNotNil: [ "Make the thumb a new instance of the match we found."
			masterFrame obliterate.
			masterFrame := myFrameHolder tframe.
			masterFrame allUpdateShape. 
						"hasn't been done for this branch (maybe no longer needed)"
			data := (TSceneWriter new storeSceneAsZip: masterFrame) contents. 	
						"zipped for upload"	
		]. 

	"If this is a new resource (in the context of the current service-island) then upload it,
 	 otherwise just re-use what's already uploaded."
		resFile := self existingResourceFile: sha forIsland: (self activeIsland).
		resFile ifNil: [
			resFile := self documentManager 
							uploadResource: data 
							shortName: c3zName
							sha: sha.
			resFile ifNil: [ 
				self error: 'Could not upload model to service host.'.
			].
		] ifNotNil: [
			QLogger logInfo: 'Model ', fname, ' is a new instance of ', resFile.
		].

	"Get thumbnail from mesh"
		thumb := TFrameThumb new.
		thumb tFileTextures: masterFrame allFileTextures.
		cube := masterFrame asCube.
		cube localTransform: Matrix4x4 new.
		thumb sha: sha.
		thumb addChild: cube.
		thumb fileName: resFile.
		thumb objectName: masterFrame objectName.
		^ thumb.
	] ifFalse: [
	 	import obliterate.	"Preserving the contract; I own import now, so clean it up."
		^ nil
	].

! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'bgf 6/7/2008 13:35'!
persistModel: model fileName: fname event: anEvent2D editBox: inEditBox

	"Place the given TFrame tree in front of the avatar in the active space,
	 in an edit box."

	| thumb rval |
	model nonOrthoCollapse.

	"Two paths - thumbnailed off-island mesh if desired (and the model is suitable),
	 or direct import into the island for all others."

	(dropModelThumb and:[model isMeshTree]) ifTrue: [

		owner statusInfo: 'Saving the model to the server...'.
		thumb := self persistAndThumbnailModel: model fileName: fname.
		thumb ifNil: [
			anEvent2D ifNotNil:[self inform: 'The imported model was empty.'].
		] ifNotNil: [
			rval := self placeThumb: thumb event: anEvent2D editBox: inEditBox.
			thumb obliterate.
			^ rval.
		].

	] ifFalse:[
		^ self simplePersistModel: model fileName: fname event: anEvent2D editBox: inEditBox.
	].
	^ nil.! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'ar 2/13/2009 14:43'!
personalShareVNC

	"Launch a personal VNC share"

	| document card |
	document := QPersonalVNCDocument new.
	card := QFileCard forDocument: document.
	card isTransient: true.
	card := TIslandCopier new import:(TIslandCopier new export:card)
									  to: self activeIsland.
	self dropCard: card event: nil.
	^true! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'ar 2/12/2009 11:19'!
personalShareWindow: windowID

	"Launch a personal share for the window with the given ID"

	| document card |

	document := QPersonalShareDocument new.
	document owner: self userLogin.
	document windowID: windowID.
	document processID: (QPersonalShare windowPidFor: windowID).
	card := QFileCard forDocument: document.
	card isTransient: true.
	card := TIslandCopier new import:(TIslandCopier new export:card)
									  to: self activeIsland.
	self dropCard: card event: nil.
	^true! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'hrs 2/7/2009 14:57'!
placeThumb: thumb event: anEvent2D editBox: inEditBox

	| modelData copy pos camT xform targetView |

	targetView := self mainView ifNil: [ ^ self ].
	owner statusInfo: 'Inserting the thumbnail model into the space...'.

	modelData := TIslandCopier new export: thumb.
	thumb destroy.

	targetView space isValidFarRef ifFalse: [ ^ nil ].

	copy := TIslandCopier new import: modelData to: targetView space island.
	copy wait.
	owner playSoundEffect: #tingle.
	"We honor the left-right position of the drop,
 	 leaving the vertical placement to the in-space sizing/scaling rules."

	inEditBox ifTrue:[
		pos := (anEvent2D ifNotNil: [ (anEvent2D position x) - targetView origin x] 
						    ifNil: [ targetView extent x // 2 ])
			@ (targetView extent y // 2).

		camT := targetView camera pointerXY: pos.
		xform := (targetView camera localTransform * camT).
		targetView space future 
			addModelInEditor: copy
			inFrontOf: xform
			for: (targetView avatar replica)
			doScale: Sensor shiftPressed not.
	].
	^ copy.! !

!QWindowedSession methodsFor: 'event handling' stamp: 'bgf 6/7/2008 13:35'!
popTask
	super popTask.

	"Don't clear the cursor, since the task doesn't correspond the cursor mode;
	 most tasks exist for the duration of pointer-down cycle,
	 but the mode that triggers them may remain in place until cleared."! !

!QWindowedSession methodsFor: 'netvid' stamp: 'ar 4/16/2011 17:27'!
popupMP4Info
	| dialog props msg wwwButton |
	
	dialog := CUserDialog new.
	dialog label: 'Video streaming information'.
	dialog addButton: 'OK' value: true.
	wwwButton := dialog addButton: 'More Information...' value: true.
	wwwButton width: 150.

	owner startScript: #openMP4InfoPage when: { wwwButton . #fire }.

	msg := 'Only some MP4 videos are suitable for streaming...',
				 '\please see the OpenQwaq support site for more information.'.

	props := owner serviceMgr 
						matchProperty: 'Limits.StreamingVideo*' 
						scopes: #(organization tier global).
	props ifNotEmpty: [
		| width height bitrate |
		width := props at: 'Limits.StreamingVideoWidth' ifAbsent: [nil].
		height := props at: 'Limits.StreamingVideoHeight' ifAbsent: [nil].
		bitrate := props at: 'Limits.StreamingVideoBitrate' ifAbsent: [nil].

		msg := msg, 
				 '\\This organization places additional', 
				 '\restrictions on streaming video: '.
		width ifNotNil: [msg := msg, '\     Width must be <= ', width].
		height ifNotNil: [msg := msg, '\     Height must be <= ', height].
		bitrate ifNotNil: [msg := msg, '\     Bitrate must be <= ', bitrate].
	].
	
	dialog message: msg withCRs.
	dialog messageText
					allowEditing: false;
					hResizing: #shrinkWrap;
					vResizing: #shrinkWrap.
	dialog run: nil centerAt: owner bounds center.
! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 2/14/2009 15:52'!
previewForService: aServiceUrl
	"Returns a TTexture containing the saved preview image for the given service,
	 if we've already gathered and cached it.   The TTexture is based on a slice out of 	
 	 the original (letterbox format) preview; the slice is higher than wide, to fit into QGate portals."

	"Asking for the texture also tells the infoCache to fetch if not yet present
	 (so we may get nil for a while, and then a texture after a  few cycles.)"

	(owner servicesPane entryForUrl: aServiceUrl) ifNotNilDo: [ : entry |
		entry showFullColor ifTrue: [
			^ owner infoCache colorPortalTexture: aServiceUrl
		]].

	^ owner infoCache grayPortalTexture: aServiceUrl.! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 5/29/2009 11:11'!
previewForUrl: aUrl

	"Returns a TTexture containing the saved preview image for the given service,
	 if we've already gathered and cached it.   The TTexture is based on a slice out of 	
 	 the original (letterbox format) preview; the slice is higher than wide, to fit into typical QGate portals."

	| shorten |
	aUrl isEmptyOrNil ifTrue: [ ^ nil ].
	(aUrl beginsWith: '//') ifFalse: [ ^ nil   "Relative url, no preview" ].


	shorten := [owner quickServiceUrlForUrl: aUrl]
				on: Error
				do: [ 
					QLogger logWarning: 'Not a service URL: ', aUrl. 
					^ nil ].
	
	^ self previewForService: shorten! !

!QWindowedSession methodsFor: 'resource-fetch' stamp: 'bgf 2/8/2010 12:06'!
processOneDelivery: docRequest
	"Receiver for thumb-resource data requested from the document manager."
	| thumb data |
	
	owner session ifNil: [ ^ nil ].
	contentManager ifNil: [ 
		QLogger logWarning: 'Received delivery but CM is nil.'.
		^ nil ].	"Destroyed "
	(data := docRequest fetchedData) ifNil: [ 
		docRequest isAppPreview ifFalse: [
			QLogger logInfo: 'Empty delivery: ', (docRequest shortName).
		].
		^ nil ].

	"Guard against invalid requestors e.g. after reconnect."
	(thumb := docRequest requestor) isValidFarRef ifFalse:[
		thumb ifNotNil: [	"I.e. not a prefetch"
			QLogger logWarning: 'Requestor of ', (docRequest serviceManagerPath), ' is invalid'.
		].
		^nil].

	"Guard against resources whose users have been destroyed."
	((thumb class = TFarRef) and: [ thumb isStale ]) ifTrue: [
		QLogger logWarning: 'Requestor of ', (docRequest serviceManagerPath), ' is stale'.
		^nil].

	docRequest refetch ifTrue: [
		"Refetch -> Use the new even if we've already got something."
		 contentManager flushTextureSHA: (docRequest cacheFileName).
	].
	contentManager dataArrived: data forThumb: thumb.
! !

!QWindowedSession methodsFor: 'event handling' stamp: 'bgf 8/24/2009 04:34'!
readyForEvents
	"Am I fully set-up and rendering, with a valid avatar replica connected?
	 We ignore interaction events from the UI,
	 and refrain from sending messages to the replica, when this is false.

	 This method should be used liberally in client-side code 
	 as a preflight readiness/health check before operating on far-refs 
	 in user-driven or routine step-based client side activities."

	^ reconnecting not 
	    and: [ errorTripped not ]
	    and: [ globalBounds notNil or: [owner doOpenGL not ] ]
	    and: [ viewpanes notEmpty ]
	! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/8/2008 18:57'!
reconnecting
	"Boolean, describing whether we are in the midst of a reconnect. Can be useful to know for certain error handling conditions to find out whether an error was due to reconnect (and will be fixed onces reconnect completes) or a genuine problem."
	^reconnecting! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/8/2008 18:57'!
reconnecting: aBool
	"Boolean, describing whether we are in the midst of a reconnect. Can be useful to know for certain error handling conditions to find out whether an error was due to reconnect (and will be fixed onces reconnect completes) or a genuine problem."
	reconnecting := aBool.! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 1/18/2009 01:08'!
refetchSHA: sha forIsland: anIsland fromSession: aUserID

	"If I already have cached content in use for the given SHA,
	 then fetch it anew from the server;
	 if we have no live content, it suffices to flush the sha from the disk cache.

	 This action can be initiated across all participants for an island,
	 by signalling #refetchSHA with: sha with island to the island."

	"This guard means don't re-download what I just uploaded."
	(aUserID = self userID) ifTrue: [ ^ self ].

	"Prepare for later access to this resource... dump it if cached on disk."
	[ self cacheManager directory deleteFileNamed: sha ]
		ifError: [:msg | QLogger logInfo: 'SHA refetch (', sha, '): ', msg].
						"Nothing stronger since they'll often be absent...
						 but log it in case its something more problematic."

	owner exitTripped ifFalse: [
		"This will provoke the fetch/update chain if the resource is in use."
		contentManager flushTextureSHA: sha.
		anIsland isValidFarRef ifTrue: [	
			self updateTexturesIn: anIsland.
		]].
! !

!QWindowedSession methodsFor: 'resource-fetch' stamp: 'bgf 6/8/2008 19:03'!
refetchThumb: tThumb fromSession: aUserID

	"If I already have cached content in use for the given SHA,
	 then fetch it anew from the server;
	 if we have no live content, it suffices to flush the sha from the disk cache.
	 This action can be initiated across all participants of an island,
	 by signalling #refetchThumb {thumb, userID} on the island."

	"This guard means don't re-download what I just uploaded."
	(aUserID = self userID) ifTrue: [ ^ self ].
	tThumb isValidFarRef ifFalse: [ ^ self ].
	owner serviceMgr ifNil: [ ^ self ]. 

	documentManager refetchThumb: tThumb
	
	! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/28/2009 12:45'!
removeResource: resFarRef type: resType in: anIsland

	"The given resource, of the given type, has been unregistered from the island."

	(resType == QSharedDocument resourceType) ifTrue: [ 
		"We keep no client-state for these, so we needn't a broad update when one is removed."
		owner ifNotNil: [ owner documentsPane signal: #update ].
		^ self ].
			
	"Fallthrough...."
	super removeResource: resFarRef type: resType in: anIsland! !

!QWindowedSession methodsFor: 'menus' stamp: 'bgf 7/13/2008 11:57'!
render360: seconds

	"A quick speed test available from the participant extras menu.
	 Turn the avatar in a circle for n seconds, report the frame rate."

	| saveIFPS delay endTime stats rate |

	owner viewPanel mouseCursor: (Cursor wait).
	delay := Delay forMilliseconds: 75.
	saveIFPS := idleFramesPerSecond.
	idleFramesPerSecond := 1000.  "Maximum"
	endTime := Time millisecondClockValue + (seconds*1000).
	stats := QFrameRate newForView: self.
	[Time millisecondClockValue < endTime] whileTrue: [
		delay wait.
		(self mainView avatar) addYaw: 3; moveForward: 0.5.
	].
	rate := stats frameRate.
	idleFramesPerSecond := saveIFPS.
	owner viewPanel mouseCursor: nil.

	self inform: ('Rendering Test Result:\\Rendered Frames: ', stats lastFrameCount, 
				'\Elapsed: ', (stats lastPeriodInSeconds roundTo: 0.1), ' seconds',
				'\Net Rate: ', (rate rounded))
			  withCRs.
	! !

!QWindowedSession methodsFor: 'menus' stamp: 'bgf 8/19/2008 12:24'!
renderTest

	| avatar saveSpace saveTransform saveIFPS string newSpace island filename file |
	endBenchTest := false.
	owner inBenchTest: true.
	Transcript show:'Begin renderTest'.

	saveSpace := self mainView space.
	avatar := self mainView avatar.

	saveTransform := avatar localTransform.
	owner viewPanel mouseCursor: (Cursor wait).
	island := self setupTestIsland.
	saveIFPS := idleFramesPerSecond.
	idleFramesPerSecond := 1000.
	newSpace := island future new: QSpace.
	newSpace future assuredDefaultEntry.
	newSpace wait.
	avatar newTransform: (Matrix4x4 identity translation:0@0@20).
	avatar newSpace: newSpace.

	[ 
		owner placesPane otherPlaceView visible: false.
		owner placesPane disable.
		owner usersPane disable.
		string := self renderTestLoop: newSpace.
	] ensure: [
		owner exitTripped ifFalse: [
			owner usersPane enable.
			owner placesPane enable.
			owner placesPane otherPlaceView visible: true.
		].
	].

	Transcript cr; show: string;cr.

	idleFramesPerSecond := saveIFPS.
	owner viewPanel mouseCursor: nil.
	avatar newSpace: saveSpace.
	avatar newTransform: saveTransform.
	owner disconnectIslandController: island controller.
	filename := 'Benchmark.txt'.
	file := FileDirectory default forceNewFileNamed: filename.

	file ascii.
	[file nextPutAll: string ] ensure: [file close].

	self inform: 'File saved as ''Benchmark.txt'''.
	owner inBenchTest: false.
	endBenchTest := false.
! !

!QWindowedSession methodsFor: 'menus' stamp: 'bgf 8/19/2008 12:25'!
renderTest: space type: type label: lbl delay: time

	| delay endTime stats rate string rval lastRender |

	owner exitTripped ifTrue: [ endBenchTest := true.  ^ 'Quit'].
	owner statusInfo: 'Computing ', lbl, '; please wait...'; waitTick.
	Transcript cr; show:'Computing ', lbl.
"Time to set up and display the world"
	rval := space future install: type.
	rval wait.
	ogl resetMeshCache.
	lastRender := lastRenderTime.
	delay := Delay forMilliseconds: 75.
	endTime := Time millisecondClockValue + (time*1000).
	[Time millisecondClockValue < endTime | lastRender = lastRenderTime] whileTrue: [
		delay wait.
		endBenchTest ifTrue:[^''].
	].
"Time to measure the world"
	owner statusInfo: 'Measuring ', lbl, '; please wait...'; waitTick.
	Transcript cr; show: 'Measuring ', lbl.
	delay := Delay forMilliseconds: 75.
	endTime := Time millisecondClockValue + (15*1000).
	stats := QFrameRate newForView: self.
	[Time millisecondClockValue < endTime] whileTrue: [
		delay wait.
		endBenchTest ifTrue:[^''].
	].
	space future removeAll.
	rate := stats frameRate.
	string := type asString,' ',lbl,'  frames=', stats lastFrameCount,'; seconds=',(stats lastPeriodInSeconds roundTo:0.1),'; rate=',(rate rounded),Character cr.
	Transcript cr; show: string.
	^ string.
! !

!QWindowedSession methodsFor: 'menus' stamp: 'bgf 8/19/2008 12:12'!
renderTestLoop: newSpace

	| string |

	string := 'Benchmark;Type;#Frames;Elapsed Time;Net rate', Character cr.
	string := string,(self renderTest: newSpace type: BenchmarkWorld label: 'Empty' delay:6).
	endBenchTest ifTrue:[^string].

	{TRectangle. TCube. TSphere.} 
	 do:[:type |
		BenchmarkWorld primitive: type.
		string := string, (self renderTest: newSpace type: Bench5World label: '5 ', type,'s' delay: 1).
		endBenchTest ifTrue:[^string].
		string := string,(self renderTest: newSpace type: Bench125World label: '125 ', type,'s' delay: 2.5).
		endBenchTest ifTrue:[^string].
		string := string,(self renderTest: newSpace type: Bench625World label: '625 ', type,'s' delay: 5).
		endBenchTest ifTrue:[^string].
	].
	string := string,(self renderTest: newSpace type: BenchTreeWorld label: '1 ', QTree delay: 5).
	endBenchTest ifTrue:[^string].
	string := string,(self renderTest: newSpace type: Bench5TreeWorld label: '5 ', QTree,'s' delay: 5).
	endBenchTest ifTrue:[^string].
	string := string,(self renderTest: newSpace type: Bench25TreeWorld label: '25 ', QTree,'s' delay:10).

	^ string.! !

!QWindowedSession methodsFor: 'embedded apps' stamp: 'bgf 2/22/2009 10:38'!
requestQMSLaunch: aRemoteApp byUser: aUserID

	"Called for each QMS application launch... permits only one at a time
	 to be in process; the others wait here.  We reset the launch timeout
	 here so that it doesn't begin couting until we've really begun the launch."
	[
		| queued |
		queued := true.
		[queued and: [aRemoteApp isDestroyed not]] whileTrue: [
			appLaunchMutex critical: [
				queued := false.
				aRemoteApp isDestroyed ifFalse: [
					aRemoteApp launchAppByUser: aUserID.
					"Even once that returns, the server side still has plenty to do.
				       We add a bit more delay by holding the mutex...	
				       this just creates space between the app-startup loads."
					(Delay forMilliseconds: 5000) wait.
				].
			] ifLocked: [
				"Serialize the launch attempt."
				(Delay forMilliseconds: 1000) wait.
			].
		].
	] fork.! !

!QWindowedSession methodsFor: 'resources' stamp: 'das 7/30/2009 19:53'!
resourceFilesForCopy: anlsland

	"All the server-side resource files that should be copied on the server
	 to copy the given forum island.    
	 We need an selective service-provider operation to make use of this efficiently.   
	 The intention is to make copies that do not inherit all the detritus of their originals;
	 this is very important for templates where stray document copies
	 copying around the server may be unwelcome."

	| names  fname |
	names := Set new.

	{ TFileTexture resourceType . 
	   TFrameThumb resourceType . 
 	   QSharedDocument resourceType} 
	do:  [ : restype |
		| resList |
		resList := anlsland future resourcesAt: restype.
		resList wait.
		resList value select: [:resRef|
			(resRef get:#isBundled) not   "Don't upload prepackaged or user-owned resources." 
			and: [ (self isAvatarSpecificResource: resRef) not ]
		] thenDo:[:resRef|
			fname := resRef get: #fileName.
			fname ifNotNil: [
				names add: fname
			].
		restype = QSharedDocument resourceType ifTrue:[
			(resRef get: #isPythonDoc) ifTrue:[
				fname := resRef get: #docName.
				fname ifNotNil:[
					names add: 'sharedDocuments/', fname encodeForHTTP.
					].
				].
			].
		].
		resList discardRef.
	].
	^ names asArray.
! !

!QWindowedSession methodsFor: 'private' stamp: 'bgf 6/7/2008 13:35'!
runBlock: aBlock withActivityText: aString onSignal: aSymbol from: anObject

	"Run the given block, appending a rotating bar to the given message 
	 as a sort of cheap activity indicator."

	| indicators index progressBlock clockTime |
	indicators := #(':...' '.:..' '..:.' '...:').
	index := 1.
	clockTime := Time millisecondClockValue.
	progressBlock := [:node | 
						| t |
						((t := Time millisecondClockValue) - clockTime) > 250 ifTrue: [
							owner statusInfo: aString, ' ', (indicators at: index).
							index := (index \\ indicators size) + 1.
							clockTime := t.
							owner waitTick.
						].
					].

	progressBlock runScript: #value: when: {anObject . aSymbol}.
	aBlock value.
	progressBlock stopScript: #value:.
	owner statusInfo: aString; waitTick.! !

!QWindowedSession methodsFor: 'embedded apps' stamp: 'bgf 3/12/2009 10:55'!
saveEmbeddedAppsIn: anIsland notifying: bar
	
	"Kick the documunt tool to recheck change times."
	super saveEmbeddedAppsIn: anIsland notifying: bar.
	owner documentsPane signal: #update.! !

!QWindowedSession methodsFor: 'islands' stamp: 'bgf 6/1/2009 21:43'!
saveIsland: anIsland asZip: archive notifying: bar
	"Qwaq replacement: 
	 We manage resources through service-manager file operations,
	 and do not include them in island zips."

	| data |

	"First notify the embedded applications so they can save off-island state."
	self saveEmbeddedAppsIn: anIsland notifying: [ :v | bar value: (v * 0.6) ].

	data := anIsland future persistString.
	bar value: 0.75.
	data isValidFarRef ifFalse: [
		self error: 'Save abandoned (disconnected).'
	].
	data wait.
	data isValidFarRef ifFalse: [
		self error: 'Save abandoned (disconnected).'
	].
	bar value: 0.9.
	archive addDeflateString: data value as: 'island.xml'.
	bar value: 1.0.
	data discard.


! !

!QWindowedSession methodsFor: 'render' stamp: 'bgf 3/16/2009 22:34'!
screenShot
	"Returns a Form capturing the entire session window.
	 This currently returns occasional miniscene-in-corner results,
	 because of the way OpenGL handles dual-buffering versus pixel-capture...
	 I'm not sure how to resolve that. "

	^ ogl ifNotNil: [
			self renderLock critical:  [ ogl glFinish; screenShot ]
		].! !

!QWindowedSession methodsFor: 'services' stamp: 'bgf 6/8/2008 18:57'!
serviceUrlForIsland: islandRef
	| url |
	url := self findIslandUrl: islandRef.
	^url ifNotNil: [url copyUpToLast: $/].! !

!QWindowedSession methodsFor: 'islands' stamp: 'bgf 5/5/2009 15:44'!
sessionIslandControllers
	"The org services controller manages only the org services island,
	 which we do not want to lock in allIslandsCritical:,
	 as it never contains rendered content, or speakers or other resource-managing objects;
	 we exclude it from the allIslandsCritical lock
	 specifically to eliminate one possible source of lock contention during render."

	(owner notNil and: [ owner orgServices notNil ]) ifTrue: [
		^ super sessionIslandControllers copyWithout: (owner orgServices controller).
	] ifFalse: [
		^ super sessionIslandControllers.
	].! !

!QWindowedSession methodsFor: 'initialize' stamp: 'bgf 6/9/2009 12:07'!
setupSystemIsland
	"We don't use the system island,
	 which is only touched by the (unused) legacy CroquetHarness setup,
	 don't want its lock and local ticking process contributing to contention."! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/8/2008 18:46'!
show3DPointerFor: anAvatarUserID
	"Shorthand query - given current preferences,
	 should the 3d-cursor be drawn for the avatar with the given ID?"
	
	(anAvatarUserID = self sessionID) ifTrue: [
		^ owner preferences showMy3DPointer
	] ifFalse: [
		^ true
	]! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 11/21/2009 14:16'!
showNameFor: anAvatarUserID whenSpeaking: bSpeaking
	"Shorthand query - given current preferences,
	 should the name be drawn for the avatar with the given ID?"

	"Under all-follow-me, when we are being dragged around, 
	 we automatically suppress labels except for the speakers.
	 (the speakers include the leader).   This declutters the posse,
	 and gives some feel that you have 'lost independence' in this situation."
	(owner meetingState notNil and: [ owner meetingState isAllFollowMe ]) ifTrue: [
		^ bSpeaking
	].
	owner preferences showOnlySpeakers ifTrue: [
		^ bSpeaking
	].
	(anAvatarUserID = self sessionID) ifTrue: [
		^ owner preferences showMyName
	] ifFalse: [
		^ owner preferences showUserNames
	]! !

!QWindowedSession methodsFor: 'drag-drop' stamp: 'hrs 2/7/2009 14:57'!
simplePersistModel: model fileName: fname event: anEvent2D editBox: inEditBox

	"Place the given TFrame tree in front of the avatar in the active space,
	 in an edit box."

	| import modelData space copy pos camT targetView |

	targetView := self mainView ifNil: [ ^ nil ].
	targetView space isValidFarRef ifFalse: [ ^ nil ].
	owner statusInfo: 'Saving the model to the server...'.

	model objectName: fname.
	import := model.
	import translation:0@0@0.
	import allUpdateShape.
	import boundingBox extent max  > 0 ifTrue:[
		modelData := TIslandCopier new export: import.

		space := targetView space.
		space wait.
		copy := TIslandCopier new import: modelData to: space island.
		copy wait.
		owner playSoundEffect: #tingle.
		inEditBox ifTrue:[
			pos := anEvent2D ifNotNil: [ (anEvent2D position) - targetView origin ] 
							    ifNil: [ targetView extent // 2 ].
			camT := targetView camera pointerXY: pos.
			space future 
				addModelInEditor: copy
				inFrontOf: (targetView camera localTransform * camT)
				for: targetView avatar replica
				doScale: Sensor shiftPressed not.
		].
		import obliterate.
		^ copy.
	] ifFalse: [
		self inform: 'The imported model was empty.' 
	].
	import obliterate.
	^ nil.! !

!QWindowedSession methodsFor: 'private' stamp: 'bgf 6/24/2009 03:53'!
soundFileIsValid: aFile 
	"The sound-file readers fail to recognize various newer format options,
	 so we preflight-check sound files to make sure we can handle'em."
	[
		aFile close.	"For windows - don't hold it open twice
				       (we have to use filename-based loaders, and they want to open the file.)"
		QSounds current loadSoundFromPath: (aFile fullName).
		aFile open.
		
	] on: Error do: [ :ex|
		self inform: 'That sound file cannot be loaded.\\' withCRs, 
					'Reason: ', (ex messageText ifNil: [ ex description]).
		^ false
	].
	
	"Arbitrary cap on sound file size... 1 meg ... 
	 since the sampled-sound reader is a dog,
	 and these like images must be downloaded by everyone."

	aFile size > (1024*1024) ifTrue: [
		self inform: ('This sound file is too large to import as a sound effect.\\',
					'', aFile localName, ' is ', ((aFile size) // 1024), ' kilobytes.\',
					'Sound effects must be less than 1024 kilobytes.') withCRs.
		^ false
	].
	^ true! !

!QWindowedSession methodsFor: 'messages' stamp: 'bgf 3/25/2009 13:57'!
step
	self markNeededDownloads.
	super step.
	self audioSession ifNotNilDo: [ : as | as step ].
	owner statisticsPane countFrame.
! !

!QWindowedSession methodsFor: 'embedded apps' stamp: 'bgf 6/8/2008 19:08'!
trafficColorForApp: appID
	"Find the TEmbeddableApp corresponding to the given embedded app,
	 and return its current trafficColor.   Will return nil for non-remote apps."

	^ (self embeddedApps at: appID ifAbsent: [ ^ nil ]) 
		trafficLevelColor! !

!QWindowedSession methodsFor: 'embedded apps' stamp: 'bgf 11/19/2009 14:32'!
updateAppPreview: form data: data sha: sha island: anIsland

	| service location txtr |

	owner exitTripped ifTrue: [  ^ self ].
	anIsland isValidFarRef ifFalse: [ ^ self ].

	"Go straight to the preview texture... it is managed directly by this process."
	self cacheManager cache: data to: sha.
	txtr := contentManager contentWithSHA: sha.
	txtr ifNotNil: [
		txtr setupForm: form asStatic: true. 		"We update manually, remaking the whole."
		contentManager substituteForm: txtr.
	] ifNil: [
		contentManager flushTextureSHA: sha.	"Remove any not-found markers."
		self updateTexturesIn: anIsland.
	].
	
	"And share it on the server..."
	service := self owner 
				urlCacheAt: (self serviceUrlForIsland: (anIsland)) ifAbsent: [nil].
	service ifNil: [ ^ self ].
	location := 'services/', service name, '/', (QEmbeddedApp previewPath), sha.
	[[ 
		self owner serviceMgr 
				put: data
				name: location 
				backup: false
				informing: nil.
		self owner exitTripped ifFalse: [	"I.e. if we're not in exit-time autosave."
			anIsland isValidFarRef ifTrue: [
				"Signal the update to all sessions."
		 		anIsland future signal: #refetchSHA with: sha with: anIsland with: sessionID.
			]].
	  ] on: Error do: [ :ex |
		QLogger logError: ex from: self.
	  ]] fork.
! !

!QWindowedSession methodsFor: 'resources' stamp: 'jcg 4/9/2010 17:21'!
updateResourceType: resourceType in: anIsland

	resourceType caseOf: {
		[QSharedDocument resourceType] -> [ self updateSharedDocsIn: anIsland].
	} otherwise:[
		super updateResourceType: resourceType in: anIsland
	].

	audioSession ifNotNil: [audioSession future updateResourceType: resourceType in: anIsland].
	videoSession ifNotNil: [videoSession updateResourceType: resourceType in: anIsland].
	animationSession ifNotNil: [ animationSession updateResourceType: resourceType in: anIsland].! !

!QWindowedSession methodsFor: 'resources' stamp: 'crl 2/12/2010 11:51'!
updateResourcesIn: anIsland newSession: isNewSession notifying: bar

	"Resource initialization / update broadcast;
 	 set up or rebuild the harness-side resource handling for all resources.
	 (Called during island-load, and also after a reconnect re-syncs the island.)
	 The flag 'isNewSession' is true when is the first user into the space."

	self updateResourceType: QSpeaker resourceType in: anIsland.
	self updateResourceType: QVideoCamera resourceType in: anIsland.
	self updateResourceType: TAnimatedFigure resourceType in: anIsland.

	super updateResourcesIn: anIsland newSession: isNewSession notifying: bar.

	self updateResourceType: QSoundFrame resourceType in: anIsland.
	self updateResourceType: QVideoTexture resourceType in: anIsland.
	self updateSharedDocsIn: anIsland.
	
	"Cannot do stopScript here because we listen to all the islands in the session;
 	 so this method is only appropriate for initialization or reset-after-reconnect,
	 both of which give us a new island.   IF that changes,
	 we'll need to stopScript, and then run these startScripts for each island
	 in the session at this point."
	self startScript: #refetchThumb:fromSession: when: { anIsland . #refetchThumb }.
	self startScript: #refetchSHA:forIsland:fromSession: when: { anIsland . #refetchSHA }.! !

!QWindowedSession methodsFor: 'resources' stamp: 'bgf 2/28/2009 12:46'!
updateSharedDoc: docRef signalTools: bSignal

	"A shared document has been added to the island."
	docRef whenResolved: [
		((docRef get:#cacheLocally) 
		 and: [documentManager isAccessibleDocument: docRef]) ifTrue: [
			documentManager fetchDocument: docRef
		 ].
	].

	(bSignal and: [ owner notNil]) ifTrue: [ owner documentsPane signal: #update ].! !

!QWindowedSession methodsFor: 'resources' stamp: 'jcg 6/16/2010 18:47'!
updateSharedDocsIn:anIsland
	| ref |
	ref := (anIsland future resourcesAt: QSharedDocument resourceType).
	ref isValidFarRef ifFalse: [ ^ self ].
	ref whenResolved: [:list|
		list ifNotNil:[list do:[:docRef |
			self updateSharedDoc: docRef signalTools: false.
			]].
		ref discardRef 
		].

	owner ifNotNil: [ owner  documentsPane signal: #update ].! !

!QWindowedSession methodsFor: 'resource-fetch' stamp: 'bgf 8/24/2010 16:51'!
updateThumb: thumbRef

	"Qwaq resource-management implementation
	 for reading (from cache) or retrieving (from service provider)
	 the resource backing a texture or model."

	"Load or fetch one thumbnail that has been requested by the session.
	 	- Check the local cache for the data
		- Failing that, issuing a fetch to the service provider;
		  (the result will be handled by the delivery processor.) "

	thumbRef isValidRef ifFalse: [ ^ nil ].	"Stale or disconnected."
	contentManager ifNil: [ ^ nil ].	"Destroyed "
	owner ifNil: [  ^nil ].

	owner flowControlFlag: true.
	[
		(contentManager loadFromCache: thumbRef forOwner: owner)
		ifFalse:[
			documentManager ifNil: [ ^ nil ].	"Destroyed "
			owner ifNil:  [ ^ nil ].
			thumbRef isValidRef ifFalse: [ ^ nil ].	"Stale or disconnected."

			"Don't request files that are not available to us... this prevents
			 bursts of resource-to-load requests from textures in other user inventories."
			(documentManager isAccessiblePath: (thumbRef get: #fileName))
			ifTrue: [
				documentManager fetchThumb: thumbRef.
			] ifFalse: [
				"Remove the nil-pending marker if we're not going to try to fetch."
				contentManager flushTextureSHA: (thumbRef get: #sha).
			]]
	] ensure: [
		(owner notNil and: [ documentManager notNil ])
		 ifTrue: [
			owner flowControlFlag: (documentManager countToFetch > 0) 
		]
	].! !

!QWindowedSession methodsFor: 'accessing' stamp: 'jcg 7/11/2008 12:39'!
userLogin
	^ self owner ifNotNilDo: [:o | o userLogin]! !

!QWindowedSession methodsFor: 'initialize' stamp: 'bgf 8/21/2008 16:58'!
userLogin: aString

	"Initialization step - the session is being created on behalf of the given user-login.
	 Set up whatever session context that relates to the login user name."

	"Make a cache directory for this user,
	 so that distinct clients on the same machine use separate caches;
	 the login is sufficient... we only want to ensure that two clients 
	 that might enter the same forum use separate cache areas."

	[
		contentManager cacheManager directory: (
			(FileDirectory cacheDirectory directoryNamed: aString)
				assureExistence).
	] on: Error do: [ :ex|
		QLogger logError: ex from: self.
	].

	owner crashRecovery ifTrue: [
		[ contentManager cacheManager directory deleteLocalFiles ]
		on: Error do: [ :err |
			QLogger logError:  err from: self.
		].
	].! !

!QWindowedSession methodsFor: 'accessing' stamp: 'bgf 6/7/2008 13:35'!
videoSession
	^videoSession! !

!QWindowedSession methodsFor: 'accessing' stamp: 'jcg 4/8/2009 23:38'!
videoSession: aQwaqVideoSession
	videoSession ifNotNil: [videoSession destroy].
	videoSession := aQwaqVideoSession.	
	videoSession ifNotNil: [
		videoSession view: self.
		self owner isVideoSessionEnabled ifFalse: [videoSession suspendVideo].
	].
! !

!QWindowedSession methodsFor: 'initialize' stamp: 'bgf 6/8/2008 14:04'!
viewPaneClass
	^ QViewPane! !

Exception subclass: #QwaqRestartException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Exceptions'!

!QwaqRestartException commentStamp: '<historical>' prior: 0!
A restart exception is signaled when we log off from a participant UI.!

!QwaqRestartException class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QwaqRestartException methodsFor: 'testing' stamp: 'ar 12/21/2006 11:23'!
defaultAction
	"By deault, ignore restart exceptions"
	^self resume! !

!QwaqRestartException methodsFor: 'testing' stamp: 'ar 12/21/2006 11:23'!
isResumable
	^true! !

CObject subclass: #QClientMeetingState
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Meeting'!

!QClientMeetingState commentStamp: '<historical>' prior: 0!
Client-side mirror of QMeetingState.

Holds the client's cache of the QMeetingState in effect for the current forum,
with set-methods to insert changes into the island-side master.

Uses tweak scripted listeners to catch all the changes in the island-side master.
and to catch space changes in the participant.

Re-raises all the island-side value changes as normal tweak-side change signals on me.

Like the TClientObject convention, 'ref' is the far-ref to the island-side QMeetingState.
Owner is the participant ui (I'm really just an accessory for the main ui.)
The rest of the instance variables are client-side caches of the ref's island-side state.
!

QClientMeetingState defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="owner" toSet="owner:" changeEvent="ownerChanged">owner</field>
	<field toGet="ref" toSet="ref:" changeEvent="refChanged">ref</field>
	<field toGet="leaderLogin" toSet="leaderLogin:" changeEvent="leaderLoginChanged">leaderLogin</field>
	<field toGet="audienceIsMuted" toSet="audienceIsMuted:" changeEvent="audienceIsMutedChanged">audienceIsMuted</field>
	<field toGet="audienceStayPut" toSet="audienceStayPut:" changeEvent="audienceStayPutChanged">audienceStayPut</field>
	<field toGet="isEditBlocked" toSet="isEditBlocked:" changeEvent="isEditBlockedChanged">isEditBlocked</field>
	<field toGet="isPrivateTextBlocked" toSet="isPrivateTextBlocked:" changeEvent="isPrivateTextBlockedChanged">isPrivateTextBlocked</field>
	<field toGet="meetingPlace" toSet="meetingPlace:" changeEvent="meetingPlaceChanged">meetingPlace</field>
	<field toGet="wantsToSpeak" toSet="wantsToSpeak:" changeEvent="wantsToSpeakChanged">wantsToSpeak</field>
	<field toGet="allowedToSpeak" toSet="allowedToSpeak:" changeEvent="allowedToSpeakChanged">allowedToSpeak</field>
</fields>'!

!QClientMeetingState class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:33'!
allowedToSpeak
	"Answer the allowedToSpeak of the receiver - - this is the SET of LOGINS"
	<bewareOf: #allowedToSpeakChanged>
	^self propertyValueAt: #allowedToSpeak! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:33'!
allowedToSpeak: aValue
	"Modify the receiver's allowedToSpeak - - this is the SET of LOGINS"
	^self propertyValueAt: #allowedToSpeak put: aValue with: #allowedToSpeakChanged! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
audienceIsMuted
	"Answer the audienceIsMuted of the receiver"
	<bewareOf: #audienceIsMutedChanged>
	^self propertyValueAt: #audienceIsMuted! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
audienceIsMuted: aValue
	"Modify the receiver's audienceIsMuted"
	^self propertyValueAt: #audienceIsMuted put: aValue with: #audienceIsMutedChanged! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 22:10'!
audienceStayPut
	"Answer the audienceStayPut of the receiver"
	<bewareOf: #audienceStayPutChanged>
	^self propertyValueAt: #audienceStayPut! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 22:10'!
audienceStayPut: aValue
	"Modify the receiver's audienceStayPut"
	^self propertyValueAt: #audienceStayPut put: aValue with: #audienceStayPutChanged! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 6/26/2009 08:22'!
clearReference
	self ref: nil; onRefChanged.		"Foreground, for immediate release of triggers."! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/25/2009 21:49'!
hasLoginPermissionToSpeak: login
	"Has this login specific leader-permission to speak now?"
	^ audienceIsMuted 
		and: [ allowedToSpeak includes: login ]! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 8/8/2009 19:57'!
initialize
	super initialize.
	self define: #audienceIsMuted as: false.	
	self define: #audienceStayPut as: false.
	self define: #isEditBlocked as: false.
	self define: #isPrivateTextBlocked as: false.
	self define: #wantsToSpeak as: Set new.
	self define: #allowedToSpeak as: Set new.! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 11/21/2009 14:07'!
isAllFollowMe
	^ meetingPlace notNil and: [ meetingPlace isFollowMePlace ]! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 23:47'!
isEditBlocked
	"Answer the isEditBlocked of the receiver"
	<bewareOf: #isEditBlockedChanged>
	^self propertyValueAt: #isEditBlocked! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 23:47'!
isEditBlocked: aValue
	"Modify the receiver's isEditBlocked"
	^self propertyValueAt: #isEditBlocked put: aValue with: #isEditBlockedChanged! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 8/2/2009 12:35'!
isLoginAllowedToMove: login
	"Can move if leader,
	  or if not in follow-me and stayput is off.
	  (I.e. a follow-me meeting place disables movement.)"
	^ (login = leaderLogin)
	   or: [ audienceStayPut not 
			and: [ meetingPlace isNil or: [ meetingPlace isFollowMePlace not ] ]]
! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:23'!
isLoginAllowedToSpeak: login

	^ (audienceIsMuted not) 
		or: [ login = leaderLogin ]
		or: [ allowedToSpeak includes: login ]! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:17'!
isLoginWantingToSpeak: login

	^ wantsToSpeak includes: login! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 8/8/2009 00:52'!
isPrivateTextBlocked
	"Answer the isPrivateTextBlocked of the receiver"
	<bewareOf: #isPrivateTextBlockedChanged>
	^self propertyValueAt: #isPrivateTextBlocked! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 8/8/2009 00:52'!
isPrivateTextBlocked: aValue
	"Modify the receiver's isPrivateTextBlocked"
	^self propertyValueAt: #isPrivateTextBlocked put: aValue with: #isPrivateTextBlockedChanged! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
leaderLogin
	"Answer the leaderLogin of the receiver"
	<bewareOf: #leaderLoginChanged>
	^self propertyValueAt: #leaderLogin! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
leaderLogin: aValue
	"Modify the receiver's leaderLogin"
	^self propertyValueAt: #leaderLogin put: aValue with: #leaderLoginChanged! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/4/2009 15:03'!
meetingPlace
	"Answer the meetingPlace of the receiver"
	<bewareOf: #meetingPlaceChanged>
	^self propertyValueAt: #meetingPlace! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/4/2009 15:03'!
meetingPlace: aValue
	"Modify the receiver's meetingPlace"
	^self propertyValueAt: #meetingPlace put: aValue with: #meetingPlaceChanged! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/1/2009 00:14'!
onAddAllowedToSpeak: aLogin
	<on: addAllowedToSpeak in: ref>

	allowedToSpeak add: aLogin.		
	self signal: #addAllowedToSpeak with: aLogin! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/1/2009 00:14'!
onAddWantsToSpeak: aLogin
	<on: addWantsToSpeak in: ref>

	wantsToSpeak add: aLogin.		
	self signal: #addWantsToSpeak with: aLogin! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/1/2009 00:22'!
onAudienceIsMutedChanged: bool
	<on: audienceIsMutedChanged in: ref>

	audienceIsMuted := bool.
	audienceIsMuted ifFalse: [
		allowedToSpeak := Set new.
	].! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 6/26/2009 22:32'!
onAudienceStayPutChanged: bool
	<on: audienceStayPutChanged in: ref>

	audienceStayPut := bool.! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 6/26/2009 23:53'!
onIsEditBlockedChanged: bool
	<on: isEditBlockedChanged in: ref>

	isEditBlocked := bool.! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 8/8/2009 00:52'!
onIsPrivateTextBlockedChanged: bool
	<on: isPrivateTextBlockedChanged in: ref>

	isPrivateTextBlocked := bool.! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 6/26/2009 08:07'!
onLeaderLoginChanged: loginString
	<on: leaderLoginChanged in: ref>

	leaderLogin := loginString.! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/4/2009 15:11'!
onMeetingPlaceChanged: place
	<on: meetingPlaceChanged in: ref>

	meetingPlace := place.! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 8/8/2009 00:51'!
onRefChanged
	"Read the meeting state out of the new meeting-state reference."
	"We would suppress this if I'm the leader..."

	<on: refChanged>

	(owner readyForEvents )  ifTrue: [
		ref isValidFarRef ifTrue: [
			self leaderLogin: (ref get: #leaderLogin).
			self audienceIsMuted: (ref get: #audienceIsMuted).
			self wantsToSpeak: ((ref get: #wantsToSpeak) asSet).
			self allowedToSpeak: ((ref get: #allowedToSpeak) asSet).
			self audienceStayPut: (ref get: #audienceStayPut).
			self isEditBlocked: (ref get: #isEditBlocked).
			self isPrivateTextBlocked: (ref get: #isPrivateTextBlocked).
			self meetingPlace: (ref get: #meetingPlace).

		] ifFalse: [
			self audienceIsMuted: false;
				audienceStayPut: false;
				isEditBlocked: false;
				isPrivateTextBlocked: false;
				meetingPlace: nil;
				allowedToSpeak: Set new;
				wantsToSpeak: Set new;
				leaderLogin: nil.
		]].
! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/1/2009 00:15'!
onRemoveAllowedToSpeak: aLogin
	<on: removeAllowedToSpeak in: ref>

	allowedToSpeak remove: aLogin ifAbsent: [
		^ QLogger logWarning: 'ATS missing ', aLogin.
	].
	self signal: #removeAllowedToSpeak with: aLogin! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/1/2009 00:15'!
onRemoveWantsToSpeak: aLogin
	<on: removeWantsToSpeak in: ref>

	wantsToSpeak remove: aLogin ifAbsent: [
		^ QLogger logWarning: 'WTS missing ', aLogin.
	].
	self signal: #removeWantsToSpeak with: aLogin! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 6/26/2009 10:48'!
onSpaceChange
	<on: activeSpaceChanged in: owner>

	owner activeSpace isValidFarRef ifTrue: [
		| ms | 
		ms := owner activeSpace future meetingState wait.
		owner activeSpace island = ms island ifTrue: [	"Overlap-check."
			self ref: ms.
		].
	] ifFalse: [
		ref := nil.
	].! !

!QClientMeetingState methodsFor: 'events' stamp: 'bgf 7/5/2009 10:58'!
onSpeakerListsCleared
	<on: speakerListsCleared in: ref>
	wantsToSpeak := Set new.
	allowedToSpeak := Set new.
	self signal: #speakerListsCleared.! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
owner
	"Answer the owner of the receiver"
	<bewareOf: #ownerChanged>
	^self propertyValueAt: #owner! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
owner: aValue
	"Modify the receiver's owner"
	^self propertyValueAt: #owner put: aValue with: #ownerChanged! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
ref
	"Answer the ref of the receiver"
	<bewareOf: #refChanged>
	^self propertyValueAt: #ref! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 6/26/2009 07:54'!
ref: aValue
	"Modify the receiver's ref"
	^self propertyValueAt: #ref put: aValue with: #refChanged! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 6/26/2009 08:01'!
setAudienceIsMuted: bool
	ref isValidFarRef ifTrue: [
		ref future audienceIsMuted: bool
	].! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 6/26/2009 22:13'!
setAudienceStayPut: bool
	ref isValidFarRef ifTrue: [
		ref future audienceStayPut: bool
	].! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 6/26/2009 23:48'!
setEditBlocked: bool
	ref isValidFarRef ifTrue: [
		ref future isEditBlocked: bool
	].! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 7/1/2009 00:20'!
setLogin: login allowedToSpeak: bAllowed
	ref isValidFarRef ifTrue: [
		bAllowed ifTrue: [
			ref future addAllowedToSpeak: login 
		] ifFalse: [
			ref future removeAllowedToSpeak: login
		]]! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 7/1/2009 00:19'!
setLogin: login wantsToSpeak: bWants
	ref isValidFarRef ifTrue: [
		bWants ifTrue: [
			ref future addWantsToSpeak: login 
		] ifFalse: [
			ref future removeWantsToSpeak: login
		]]! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 7/1/2009 00:18'!
setLoginWantsToSpeak: login
	ref isValidFarRef ifTrue: [
		ref future addWantsToSpeak: login 
	]! !

!QClientMeetingState methodsFor: 'control' stamp: 'bgf 8/8/2009 02:15'!
setPrivateTextBlocked: bool
	ref isValidFarRef ifTrue: [
		ref future isPrivateTextBlocked: bool
	].
	owner orgServices ifNotNil: [
		owner orgServices setMeetingPrivateTextBlocked: bool
	].! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:32'!
wantsToSpeak
	"Answer the wantsToSpeak of the receiver - - this is the SET of LOGINS"
	<bewareOf: #wantsToSpeakChanged>
	^self propertyValueAt: #wantsToSpeak! !

!QClientMeetingState methodsFor: 'accessing' stamp: 'bgf 7/1/2009 00:32'!
wantsToSpeak: aValue
	"Modify the receiver's wantsToSpeak - this is the SET of LOGINS"
	^self propertyValueAt: #wantsToSpeak put: aValue with: #wantsToSpeakChanged! !

CObject subclass: #QStatusLine
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QStatusLine commentStamp: '<historical>' prior: 0!
Manages transient status messages for the Qwaq participant UI.

Messages are handled in three categories:
	Alerts - always appear, persist longest.
	Info	- appear unless an alert is present, persist less than alerts.
	Status - appear unless an info or alert is present; persist less than info messages.

Info messages are displayed in two places - both in the status line,
and as a 'floater' hovering above the pointer.  The hover fades more
quickly than the status line message.

When a message is displayed (via #alert:, #info, or #status)
it appears in the (client-provided) label for a short amount of time 
(currently 5, 2, and 2 seconds respectively) and then vanishes.   
This mechanism is not meant for dialog interactions or system logging;
it is merely a gentle way to communicate.

The visuals are weak for status, stronger for messages, and distracting for alerts.

Also manages the floating yellow-tag messages;
these are set with 'floatingMessage: ' and its variants.
They start to fade a short time after the mouse next moves, after they appear;
they can be made to last longer using the full variant
	#floatingMessage: string at: aPoint extraSeconds: nExtraPersistSeconds!

QStatusLine defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="owner" toSet="owner:" changeEvent="ownerChanged">owner</field>
	<field toGet="label" toSet="label:" changeEvent="labelChanged">label</field>
	<field toGet="currentMessage" toSet="currentMessage:" changeEvent="currentMessageChanged">currentMessage</field>
	<field toGet="category" toSet="category:" changeEvent="categoryChanged">category</field>
	<field toGet="timestamp" toSet="timestamp:" changeEvent="timestampChanged">timestamp</field>
	<field toGet="alertForeground" toSet="alertForeground:" changeEvent="alertForegroundChanged">alertForeground</field>
	<field toGet="infoForeground" toSet="infoForeground:" changeEvent="infoForegroundChanged">infoForeground</field>
	<field toGet="statusForeground" toSet="statusForeground:" changeEvent="statusForegroundChanged">statusForeground</field>
	<field toGet="infoStyle" toSet="infoStyle:" changeEvent="infoStyleChanged">infoStyle</field>
	<field toGet="alertStyle" toSet="alertStyle:" changeEvent="alertStyleChanged">alertStyle</field>
	<field toGet="statusStyle" toSet="statusStyle:" changeEvent="statusStyleChanged">statusStyle</field>
	<field toGet="smallStyle" toSet="smallStyle:" changeEvent="smallStyleChanged">smallStyle</field>
	<field toGet="floaterTimestamp" toSet="floaterTimestamp:" changeEvent="floaterTimestampChanged">floaterTimestamp</field>
	<field toGet="floater" toSet="floater:" changeEvent="floaterChanged">floater</field>
</fields>'!

!QStatusLine class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStatusLine methodsFor: 'messages' stamp: 'jcg 4/28/2010 15:07'!
alert: aString
	 "Present the string as an alert message, 
	  overriding any current status or info message."

	aString ifNotNil: [
		"Don't clobber a displayed alert with another unless the user is idle.
	       (E.g. if you awake after a long idle, the alert you'll see is the most recent)"
		(floater visible not 						" -- no message"
		 or: [self floaterAlpha < 0.6 ]				" -- enough fade to have been seen "
		 or: [self floaterTimestamp <= 0 ]  ) 		" -- no countdown yet, no user activity"
		ifTrue: [
			self floatingMessage: aString.
		]].
	
	((category == #alert) and: [self message = aString]) ifFalse: [
		"So it shows up in the transcript in bug reports, but not as a flood."
		aString ifNotNil: [
			QLogger logWarning: '[Status-Line Alert] ', aString from: thisContext sender receiver.
		]].

	"And for the status line (the status line fades more slowly so it is a useful extra here.)"
	self prepareMessage: aString category: #alert.

! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
alertForeground
	"Answer the alertForeground of the receiver"
	<bewareOf: #alertForegroundChanged>
	^self propertyValueAt: #alertForeground! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
alertForeground: aValue
	"Modify the receiver's alertForeground"
	^self propertyValueAt: #alertForeground put: aValue with: #alertForegroundChanged! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
alertStyle
	"Answer the alertStyle of the receiver"
	<bewareOf: #alertStyleChanged>
	^self propertyValueAt: #alertStyle! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
alertStyle: aValue
	"Modify the receiver's alertStyle"
	^self propertyValueAt: #alertStyle put: aValue with: #alertStyleChanged! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
category
	"Answer the category of the receiver"
	<bewareOf: #categoryChanged>
	^self propertyValueAt: #category! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
category: aValue
	"Modify the receiver's category"
	^self propertyValueAt: #category put: aValue with: #categoryChanged! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 3/6/2009 01:51'!
clearFloat
	floater visible: false.! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 6/25/2009 11:09'!
currentMessage
	"Answer the currentMessage of the receiver"
	<bewareOf: #currentMessageChanged>
	^self propertyValueAt: #currentMessage! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 6/25/2009 11:09'!
currentMessage: aValue
	"Modify the receiver's currentMessage"
	^self propertyValueAt: #currentMessage put: aValue with: #currentMessageChanged! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
floater
	"Answer the floater of the receiver"
	<bewareOf: #floaterChanged>
	^self propertyValueAt: #floater! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
floater: aValue
	"Modify the receiver's floater"
	^self propertyValueAt: #floater put: aValue with: #floaterChanged! !

!QStatusLine methodsFor: 'events' stamp: 'bgf 8/18/2009 16:37'!
floaterAlpha
	^ floater textColor alpha
	! !

!QStatusLine methodsFor: 'events' stamp: 'bgf 3/8/2009 12:22'!
floaterAlpha: a
	floater 
		textColor: (Color r: 0.05 g:0.05 b:0.04 alpha: a); 
		borderColor: ((Color gray: 0.4) alpha: (a* 0.9));
		fillColor: (Color r: 0.96 g: 0.96 b: 0.65 alpha: (a * 0.9)).
	floater dropShadow 
		color: (floater dropShadow color alpha: (a min: 0.5) squared).
		! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
floaterTimestamp
	"Answer the floaterTimestamp of the receiver"
	<bewareOf: #floaterTimestampChanged>
	^self propertyValueAt: #floaterTimestamp! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
floaterTimestamp: aValue
	"Modify the receiver's floaterTimestamp"
	^self propertyValueAt: #floaterTimestamp put: aValue with: #floaterTimestampChanged! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 2/27/2008 21:11'!
floatingMessage: aString
	"Set a new message for the over-the-pointer float.
	 It will appear near the mouse-pointer, start to fade when the mouse moves,
	 and vanish after a few seconds, unless reinforced by another call to this method."

	^ self floatingMessage: aString at: nil extraSeconds: 0! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 2/27/2008 20:45'!
floatingMessage: aString at: aPoint
	
	"A floating message that appears immediately
	 and lasts until the mouse moves, plus the fade time."

	^ self floatingMessage: aString at: aPoint extraSeconds: 0! !

!QStatusLine methodsFor: 'messages' stamp: 'bgf 3/8/2009 12:04'!
floatingMessage: aString at: aPoint extraSeconds: nSecs
	"Set a new message for the over-the-pointer float.
	 It will appear near the mouse-pointer, 
	 start to fade: when the mouse moves, plus extraSeconds,
	 and vanish after a few seconds
	 unless reinforced (by another call to this method.)

	 The string is de-formatted (interior whitespace removed)
	 but that if it contains backslash characters,
	 those will become linefeeds -after- stripping out surplus white space."
	| pos msg |
	aString isEmptyOrNil ifTrue: [
		^ self clearFloat.
	].
	msg := (aString contractTo: 128) withSeparatorsCompacted withCRs.

	"If we're currently showing the same text, we don't move the tag position,
	 so that for instance mouseover triggered tags can stay steady
	 while the software just shoots us a wobbly mouse position."
	(floater label ~= msg  
	 or: [floater fillColor alpha  < 0.3] 
	 or: [ floater visible not ]) ifTrue: [
		pos := aPoint ifNil: [ Processor activeProcess hand position].
		floater label: msg.
	].
	floaterTimestamp := (-1000*nSecs).
	pos ifNotNil: [  self placeFloat: pos ].! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 7/20/2006 10:38'!
foregroundFor: catSymbol

	^ catSymbol caseOf: {
		[#alert] -> [alertForeground] .
		[#info] -> [infoForeground].
		[#status] -> [statusForeground].
	}.
! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 4/27/2007 14:05'!
info: aString
	"Present the string as a new information message,
	 overriding any current status message."

	(category = #alert) ifTrue:	[
		^ self.
	].
	self prepareMessage: aString category: #info.! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
infoForeground
	"Answer the infoForeground of the receiver"
	<bewareOf: #infoForegroundChanged>
	^self propertyValueAt: #infoForeground! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
infoForeground: aValue
	"Modify the receiver's infoForeground"
	^self propertyValueAt: #infoForeground put: aValue with: #infoForegroundChanged! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
infoStyle
	"Answer the infoStyle of the receiver"
	<bewareOf: #infoStyleChanged>
	^self propertyValueAt: #infoStyle! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
infoStyle: aValue
	"Modify the receiver's infoStyle"
	^self propertyValueAt: #infoStyle put: aValue with: #infoStyleChanged! !

!QStatusLine methodsFor: 'initialize' stamp: 'bgf 8/21/2009 15:15'!
initialize
	super initialize.

	category := #status.
	timestamp := 0.
	floaterTimestamp := 0.
		
	self setStyleForWindowWidth: (QwaqParticipantUI defaultWindowSize x).

	alertForeground := Color black.
	infoForeground := Color gray: 0.08.
	statusForeground := Color gray: 0.2.
	
	floater := (QInertLabel new)
		label: '';
		textAnchorPoint: #center;
		textInset: 4@3;
		borderStyle: #simple;  borderWidth: 1;
		dropShadowEnabled: true;
		font: (CWidgetLibrary default balloonFont);
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;	
		visible: false;
		yourself.
	self floaterAlpha: 0.
! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 02:47'!
label
	"Answer the label of the receiver"
	<bewareOf: #labelChanged>
	^self propertyValueAt: #label! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 02:47'!
label: aValue
	"Modify the receiver's label"
	^self propertyValueAt: #label put: aValue with: #labelChanged! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 6/25/2009 11:11'!
message
	"The status/info/alert message currently being displayed; 
	  nil when the status line is free to show the default message."
	^ currentMessage! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 6/25/2009 11:59'!
message: aString
	"The message currently being displayed; 
	 nil when the status line is free to show the owner's defaultMessage."
	currentMessage := aString.
	label label: (aString ifNotNil: [ aString copyReplaceAll: '\' with: ' ']).! !

!QStatusLine methodsFor: 'events' stamp: 'bgf 7/3/2009 23:41'!
onLabelChanged
	<on: labelChanged>
	"When the UI designates the receiving label for messages."

	self stopScript: #onTick.
	label ifNotNil: [
		label label: ''.
		self startTicking: #onTick at: 10.
	].! !

!QStatusLine methodsFor: 'events' stamp: 'bgf 7/3/2009 23:50'!
onTick
	"Check the age of the current message, 
	and fade it out it if it has been up long enough."
	| age threshold alpha fadeTime  a |

	self message ifNotNil: [
		fadeTime := 1500.0.
		age := (Time coarseMillisecondsSince: timestamp).
		threshold := category caseOf: {
			[#alert] -> [2500]. 	"duration an alert message remains before fading"
			[#info] ->  [1000].	"duration an info message remains before fading"
			[#status]-> [500].	"duration a status message remains before fading"
		}.
		alpha := ((fadeTime+threshold)-age)/fadeTime.
		alpha > 0.2 ifTrue: [
			label textColor: ((self foregroundFor:category) alpha: alpha).
		] ifFalse: [
			self message: nil.
			category := #status.
		].
	] ifNil: [
		"Reset it if we're not showing anything."
		category := #status.
		label label = (owner defaultStatusMessage) ifFalse: [
			label textFont: (self styleFor: category) defaultFont;
				  textColor: (self foregroundFor: category);
				  label: (owner defaultStatusMessage).
		].
	].

	"The floater starts to fade once a motion-timestamp has been set on it."
	(floater label isEmptyOrNil) ifFalse: [
		fadeTime := 1000.0.
		age := (floaterTimestamp <= 0) 
				ifTrue: [ 0 ] 
				ifFalse: [ Time coarseClockValue - floaterTimestamp ].
						"Yes, this has rare rollover potential. 
						 But we use the size of the difference for information,
						 since the floater timestamp can be set into the future to delay clearing it.
						 A new message will clear it up no matter what."
		age > fadeTime ifTrue: [
			floater visible: false.
		] ifFalse: [
			(age > 400)
				ifTrue: [	 a := (fadeTime-age)/fadeTime ] "Prevent flicker"
				 ifFalse: [ a := 0.95 ].
			self floaterAlpha: a.
			]].! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
owner
	"Answer the owner of the receiver"
	<bewareOf: #ownerChanged>
	^self propertyValueAt: #owner! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
owner: aValue
	"Modify the receiver's owner"
	^self propertyValueAt: #owner put: aValue with: #ownerChanged! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 8/21/2009 15:09'!
placeFloat: target
	| pos maxArea result usableArea focalArea rectPoint handPoint rect box a | 

	usableArea := (2@26) extent: (owner extent - (4@56)).   "Dodge menu bar, control bar."
	pos := target.
	(usableArea containsPoint: pos) ifFalse: [ 
		pos := usableArea pointNearestTo: pos.	"So some case in the loop below will find overlap."
	].

	floater waitTick; fullBounds.
	floater visible: false.
	floater position: pos; fullBounds.

	"Lifted straight out of balloon-help"
	maxArea := -1.
	result := (box := floater bounds).
	owner fullBounds.
	usableArea := (2@26) extent: (owner extent - (4@54)).
	focalArea := Rectangle center: pos extent: 8@32.
	1 to: 4 do: [:i |
		"Corresponding points to align, from the balloon rect and the hand position."
		rectPoint := #(bottomCenter topCenter leftCenter rightCenter) at: i.
		handPoint := #(topCenter bottomCenter rightCenter leftCenter) at: i.
		rect := box align: (box perform: rectPoint) with: (focalArea perform: handPoint).
		a := (rect intersect: usableArea) area.
	
		(a > maxArea or: [a = box area and: [rectPoint = #bottomCenter]]) ifTrue:
			[maxArea := a.
			 result := rect]
	].

	floater bounds: result.
	floater comeToFront; visible: true.! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 3/6/2009 02:42'!
prepareMessage: aString category: aSymbol

	"Composite the message into my label, to prepare for drawing and fading."

	(aString isNil or: [aString = '']) ifTrue: [
		self message: nil.
		category := #status.
		^ self
	].
	timestamp := Time millisecondClockValue.
	category := aSymbol.
	label textFont: (self styleFor: category) defaultFont;
		   textColor: (self foregroundFor: category).
	self message: (aString contractTo: 128) withSeparatorsCompacted.
	label fullBounds.


! !

!QStatusLine methodsFor: 'initialize' stamp: 'bgf 7/16/2009 15:44'!
setStyleForWindowWidth: pixels
	"We make the status text a bit smaller when the window is narrower."
	| style sz |

	"We get the notification when the main-window is resized,
	 but base the decision on the label that we put messages into."
	label ifNotNil: [
		label fullBounds.
		sz :=   (true caseOf: {
					[label width < 300] -> [ 10 ] .
					[label width < 500] -> [ 11 ] .
				} otherwise: [12]).
	] ifNil: [
		sz := 11.
	].
	style := TextStyle named:(QStyles controlFont familyName).
	alertStyle := TextStyle fontArray: {style fontOfPointSize: sz}.
	infoStyle := TextStyle  fontArray: {style fontOfPointSize: sz}.
	statusStyle := TextStyle fontArray: {style fontOfPointSize: sz}.
	smallStyle := TextStyle fontArray: { QStyles smallTextFont }.! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
smallStyle
	"Answer the smallStyle of the receiver"
	<bewareOf: #smallStyleChanged>
	^self propertyValueAt: #smallStyle! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
smallStyle: aValue
	"Modify the receiver's smallStyle"
	^self propertyValueAt: #smallStyle put: aValue with: #smallStyleChanged! !

!QStatusLine methodsFor: 'events' stamp: 'bgf 3/14/2009 17:30'!
startCounting
	"Called for all user input in the application...
	 trigger to start the clock on the current floating message, if any."

	"A floating timestamp <= 0 is an indication that we should 
	 set the timestamp on next mouse move.  
	 When less than zero, the  (abs) value holds
	 the number of extra milliseconds to show the message before fading;
	 this is staged by the floatingMessage:... setter and honored by #tick."

	(floater label notNil and: [floaterTimestamp <= 0]) ifTrue: [ 
		floaterTimestamp := Time millisecondClockValue + (floaterTimestamp negated).
	].! !

!QStatusLine methodsFor: 'messages' stamp: 'bgf 3/6/2009 01:37'!
status: aString
	"Present the string as a low-importance status message.
	Status messages supersede only other status messages."

	(category ~= #status) ifTrue: [
		^ self.
	].
	self prepareMessage: aString category: #status.
! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
statusForeground
	"Answer the statusForeground of the receiver"
	<bewareOf: #statusForegroundChanged>
	^self propertyValueAt: #statusForeground! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
statusForeground: aValue
	"Modify the receiver's statusForeground"
	^self propertyValueAt: #statusForeground put: aValue with: #statusForegroundChanged! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
statusStyle
	"Answer the statusStyle of the receiver"
	<bewareOf: #statusStyleChanged>
	^self propertyValueAt: #statusStyle! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
statusStyle: aValue
	"Modify the receiver's statusStyle"
	^self propertyValueAt: #statusStyle put: aValue with: #statusStyleChanged! !

!QStatusLine methodsFor: 'private' stamp: 'bgf 3/6/2009 01:38'!
styleFor: catSymbol
	"Table lookup the message-area text style for the category of message"
	(label width < 360) ifTrue: [ ^ smallStyle ].
	^ catSymbol caseOf: {
		[#alert] -> [alertStyle] .
		[#info] -> [infoStyle].
		[#status] -> [statusStyle].
	}.
! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
timestamp
	"Answer the timestamp of the receiver"
	<bewareOf: #timestampChanged>
	^self propertyValueAt: #timestamp! !

!QStatusLine methodsFor: 'accessing' stamp: 'bgf 3/6/2009 01:44'!
timestamp: aValue
	"Modify the receiver's timestamp"
	^self propertyValueAt: #timestamp put: aValue with: #timestampChanged! !

CObject subclass: #QUndoRedoList
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Undo'!

!QUndoRedoList commentStamp: '<historical>' prior: 0!
Manages the stacks of user actions that are available to be undone
(or, if some have been undone, available for redoing.)

The main logic is in the methods
	redoOneForClient:
	undoOneForClient:
which is where the invocation of available undo/redo is managed.

Actions arrive in the system via 
	recordAction: aQUndoableAction

!

QUndoRedoList defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="undoList" toSet="undoList:" changeEvent="undoListChanged">undoList</field>
	<field toGet="redoList" toSet="redoList:" changeEvent="redoListChanged">redoList</field>
</fields>'!

!QUndoRedoList class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUndoRedoList class methodsFor: 'defaults' stamp: 'bgf 3/13/2010 09:09'!
undoListSize
	"The maximum number of undoable operations recorded.
	 Beyond this limit we'll forget the oldest actions as new ones arrive."
	^ 32! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:10'!
canRedo
	^ redoList size > 0 and: [ redoList last canBePerformed ]! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:11'!
canUndo
	^ undoList size > 0 and: [ undoList last canBePerformed ]! !

!QUndoRedoList methodsFor: 'initialize' stamp: 'bgf 3/13/2010 09:09'!
initialize
	self define: #undoList as: OrderedCollection new.
	self define: #redoList as: OrderedCollection new.! !

!QUndoRedoList methodsFor: 'actions' stamp: 'bgf 3/13/2010 09:10'!
recordAction: newAction	
	"A new intervention is being queued. "

	(undoList size >= self class undoListSize) ifTrue: [
		undoList removeFirst.
	].
	undoList addLast: newAction.
	redoList := OrderedCollection new.
! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:10'!
redoLabel
	"Label for a Redo menu entry at this moment."
	^  ( redoList size > 0 )
			ifTrue: [ 'Redo ', (redoList last description) ]
			ifFalse: [ 'Redo' ].! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:09'!
redoList
	"Answer the redoList of the receiver"
	<bewareOf: #redoListChanged>
	^self propertyValueAt: #redoList! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:09'!
redoList: aValue
	"Modify the receiver's redoList"
	^self propertyValueAt: #redoList put: aValue with: #redoListChanged! !

!QUndoRedoList methodsFor: 'actions' stamp: 'bgf 3/13/2010 11:11'!
redoOneForClient: aClient
	"Redo the most-recently Undone action."

	aClient readyForEvents ifFalse: [ ^ self ].

	redoList isEmptyOrNil ifFalse: [
		| action |
		action := (redoList removeLast).
		action canBePerformed ifFalse: [ 
			aClient statusAlert: 'Cannot redo ', (action description).
			^ false
		].
		undoList addLast: action.
		aClient statusAlert: 'Redo ', (action description); waitTick.
		[ 
			action redo: aClient 
		] on: Error do: [ : ex |
			aClient inform: ('Unexpected problem in redo.\\', ex description, '\') withCRs.
			self reset.
			^ false
		].
		^ true
	].
	aClient statusAlert: 'Nothing to redo.'.
	^ false
! !

!QUndoRedoList methodsFor: 'initialize' stamp: 'bgf 3/13/2010 09:11'!
reset
	QLogger logInfo: 'Undo-Redo list reset.'.
	undoList := OrderedCollection new.
	redoList := OrderedCollection new.! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:11'!
undoLabel
	^  ( undoList size > 0 )
			ifTrue: [ 'Undo ', (undoList last description) ]
			ifFalse: [ 'Undo' ].! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:09'!
undoList
	"Answer the undoList of the receiver"
	<bewareOf: #undoListChanged>
	^self propertyValueAt: #undoList! !

!QUndoRedoList methodsFor: 'accessing' stamp: 'bgf 3/13/2010 09:09'!
undoList: aValue
	"Modify the receiver's undoList"
	^self propertyValueAt: #undoList put: aValue with: #undoListChanged! !

!QUndoRedoList methodsFor: 'actions' stamp: 'bgf 3/13/2010 11:10'!
undoOneForClient: aClient
	"Undo the most recent action that has not already been undone."

	aClient readyForEvents ifFalse: [ ^ self ].
	undoList isEmptyOrNil ifFalse: [
		| action |
		action := (undoList removeLast).
		action canBePerformed ifFalse: [ 
			aClient statusAlert: 'Cannot undo ', (action description).
			^ false
		].
		redoList addLast: action.
		aClient statusAlert: 'Undo ', (action description); waitTick.
		[ 
			action undo: aClient.
		] on: Error do: [ : ex |
			aClient inform: ('Unexpected problem in undo.\\', ex description, '\') withCRs.
			self reset.
			^ false
		].
		^ true
	].
	aClient statusAlert: 'Nothing to undo.'.
	^ false
! !

CPanel subclass: #QwaqParticipantUI
	instanceVariableNames: ''
	classVariableNames: 'InterruptPassword KillRenderOnError ProfilerPreference UpdateSemaphore'
	poolDictionaries: 'OpenGLConstants'
	category: 'Qwaq-Client'!

!QwaqParticipantUI commentStamp: '<historical>' prior: 0!
Qwaq's client application - our Croquet Browser.

This is the top-level owner of all the parts of a running client;
the login process and view/session setup are orchestrated from here.

When running 'Local' ad-hoc services, the master's ServiceManager
is also run and supervised here.

Instance variables include all the UI constituents,
as well as common tables of resources,
and the service-manager and CroquetSession for this application.

Subclassing is mainly done to 'skin' the client as a named application;
subclasses often override appName, for instance.
Some aspects of service-listing and discovery rely on matching
the client class to a 'clientClass' property in the service.
So some subclasses are made simply to serve as marker classes
to associate the client with a particular set of services
(e.g. the QwaqForumsUI is the client for Forums services.)!

QwaqParticipantUI subclass: #QwaqForumsUI
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

QwaqForumsUI defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!QwaqForumsUI class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QwaqForumsUI methodsFor: 'initialize' stamp: 'ar 4/16/2011 17:27'!
appName
	^ 'OpenQwaq'! !

!QwaqForumsUI methodsFor: 'menu' stamp: 'bgf 10/22/2008 21:31'!
objectsDirectory

	"the templates for Forums applications are in 'objects/Forums'"

	^FileDirectory default directoryNamed: (FileDirectory default pathFromURI: 'objects-menu/Forums').! !

!QwaqForumsUI methodsFor: 'events' stamp: 'bgf 7/17/2009 22:41'!
onDropFiles: anEvent
	"Handle a number of dropped files from the OS."
	| numFiles files path |
	<on: dropFiles>

	(uiStyle = #Miramar) ifFalse:[ ^super onDropFiles: anEvent ].
	self mainView alternateDropPolicy ifTrue:[ ^super onDropFiles: anEvent ].
	self readyForEvents ifFalse: [ ^ self ].
	(meetingState isEditBlocked and: [ self isMeetingLeader not]) ifTrue: [ 
		self statusAlert: 'Imports are not permitted by the meeting leader.'.
		^ self ].
	self isReadonlyUser ifTrue: [
		self statusAlert: 'As a visitor, you cannot add documents.'.
		^ self ].

	numFiles := anEvent contents.
	(numFiles = 1) ifTrue:[
		path := (FileStream requestDropName: 1) unquoted.
		(FileDirectory isDirectory: path)
			ifTrue: [session dropDirectory: path]
			ifFalse:[session dropFile: path].
	] ifFalse:[
		"If multiple items are dropped, they will be put in a bulletin board,"
		files := OrderedCollection new.

		1 to: numFiles do: [:i |
			files addLast: (FileStream requestDropName: i).
		].

		(files size = 0) ifFalse:[
			session dropFiles: files asArray.
		].
	].! !

QwaqParticipantUI defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="exitTripped" toSet="exitTripped:" changeEvent="exitTrippedChanged">exitTripped</field>
	<field toGet="session" toSet="session:" changeEvent="sessionChanged">session</field>
	<field toGet="mainView" toSet="mainView:" changeEvent="mainViewChanged">mainView</field>
	<field toGet="viewPanel" toSet="viewPanel:" changeEvent="viewPanelChanged">viewPanel</field>
	<field toGet="controlBar" toSet="controlBar:" changeEvent="controlBarChanged">controlBar</field>
	<field toGet="preferences" toSet="preferences:" changeEvent="preferencesChanged">preferences</field>
	<field toGet="undoRedoList" toSet="undoRedoList:" changeEvent="undoRedoListChanged">undoRedoList</field>
	<field toGet="avatarPane" toSet="avatarPane:" changeEvent="avatarPaneChanged">avatarPane</field>
	<field toGet="catalogPane" toSet="catalogPane:" changeEvent="catalogPaneChanged">catalogPane</field>
	<field toGet="chatPane" toSet="chatPane:" changeEvent="chatPaneChanged">chatPane</field>
	<field toGet="lastBounds" toSet="lastBounds:" changeEvent="lastBoundsChanged">lastBounds</field>
	<field toGet="oglCache" toSet="oglCache:" changeEvent="oglCacheChanged">oglCache</field>
	<field toGet="navPanel" toSet="navPanel:" changeEvent="navPanelChanged">navPanel</field>
	<field toGet="avatar" toSet="avatar:" changeEvent="avatarChanged">avatar</field>
	<field toGet="login" toSet="login:" changeEvent="loginChanged">login</field>
	<field toGet="proxyDlg" toSet="proxyDlg:" changeEvent="proxyDlgChanged">proxyDlg</field>
	<field toGet="serviceChooser" toSet="serviceChooser:" changeEvent="serviceChooserChanged">serviceChooser</field>
	<field toGet="serviceHost" toSet="serviceHost:" changeEvent="serviceHostChanged">serviceHost</field>
	<field toGet="prefsPanel" toSet="prefsPanel:" changeEvent="prefsPanelChanged">prefsPanel</field>
	<field toGet="aboutBox" toSet="aboutBox:" changeEvent="aboutBoxChanged">aboutBox</field>
	<field toGet="activeSpace" toSet="activeSpace:" changeEvent="activeSpaceChanged">activeSpace</field>
	<field toGet="treeInspector" toSet="treeInspector:" changeEvent="treeInspectorChanged">treeInspector</field>
	<field toGet="materialCatalog" toSet="materialCatalog:" changeEvent="materialCatalogChanged">materialCatalog</field>
	<field toGet="sqScriptEditor" toSet="sqScriptEditor:" changeEvent="sqScriptEditorChanged">sqScriptEditor</field>
	<field toGet="statusLine" toSet="statusLine:" changeEvent="statusLineChanged">statusLine</field>
	<field toGet="flowControlFlag" toSet="flowControlFlag:" changeEvent="flowControlFlagChanged">flowControlFlag</field>
	<field toGet="logView" toSet="logView:" changeEvent="logViewChanged">logView</field>
	<field toGet="logFile" toSet="logFile:" changeEvent="logFileChanged">logFile</field>
	<field toGet="keyboardIcon" toSet="keyboardIcon:" changeEvent="keyboardIconChanged">keyboardIcon</field>
	<field toGet="posReadout" toSet="posReadout:" changeEvent="posReadoutChanged">posReadout</field>
	<field toGet="serviceMgr" toSet="serviceMgr:" changeEvent="serviceMgrChanged">serviceMgr</field>
	<field toGet="serviceSvc" toSet="serviceSvc:" changeEvent="serviceSvcChanged">serviceSvc</field>
	<field toGet="usersPane" toSet="usersPane:" changeEvent="usersPaneChanged">usersPane</field>
	<field toGet="placesPane" toSet="placesPane:" changeEvent="placesPaneChanged">placesPane</field>
	<field toGet="peoplePlacesSash" toSet="peoplePlacesSash:" changeEvent="peoplePlacesSashChanged">peoplePlacesSash</field>
	<field toGet="viewAreaSash" toSet="viewAreaSash:" changeEvent="viewAreaSashChanged">viewAreaSash</field>
	<field toGet="ppToggle" toSet="ppToggle:" changeEvent="ppToggleChanged">ppToggle</field>
	<field toGet="videoAreaSash" toSet="videoAreaSash:" changeEvent="videoAreaSashChanged">videoAreaSash</field>
	<field toGet="documentsPane" toSet="documentsPane:" changeEvent="documentsPaneChanged">documentsPane</field>
	<field toGet="servicesPane" toSet="servicesPane:" changeEvent="servicesPaneChanged">servicesPane</field>
	<field toGet="statisticsPane" toSet="statisticsPane:" changeEvent="statisticsPaneChanged">statisticsPane</field>
	<field toGet="trafficPane" toSet="trafficPane:" changeEvent="trafficPaneChanged">trafficPane</field>
	<field toGet="latencyPane" toSet="latencyPane:" changeEvent="latencyPaneChanged">latencyPane</field>
	<field toGet="recorderPane" toSet="recorderPane:" changeEvent="recorderPaneChanged">recorderPane</field>
	<field toGet="webcastPane" toSet="webcastPane:" changeEvent="webcastPaneChanged">webcastPane</field>
	<field toGet="captureProxy" toSet="captureProxy:" changeEvent="captureProxyChanged">captureProxy</field>
	<field toGet="quickReference" toSet="quickReference:" changeEvent="quickReferenceChanged">quickReference</field>
	<field toGet="svcMgrCache" toSet="svcMgrCache:" changeEvent="svcMgrCacheChanged">svcMgrCache</field>
	<field toGet="urlCache" toSet="urlCache:" changeEvent="urlCacheChanged">urlCache</field>
	<field toGet="firstArrivalPending" toSet="firstArrivalPending:" changeEvent="firstArrivalPendingChanged">firstArrivalPending</field>
	<field toGet="historyPane" toSet="historyPane:" changeEvent="historyPaneChanged">historyPane</field>
	<field toGet="webcamPane" toSet="webcamPane:" changeEvent="webcamPaneChanged">webcamPane</field>
	<field toGet="phonePane" toSet="phonePane:" changeEvent="phonePaneChanged">phonePane</field>
	<field toGet="customsPane" toSet="customsPane:" changeEvent="customsPaneChanged">customsPane</field>
	<field toGet="backButton" toSet="backButton:" changeEvent="backButtonChanged">backButton</field>
	<field toGet="forwardButton" toSet="forwardButton:" changeEvent="forwardButtonChanged">forwardButton</field>
	<field toGet="webcastIndicator" toSet="webcastIndicator:" changeEvent="webcastIndicatorChanged">webcastIndicator</field>
	<field toGet="recordingIndicator" toSet="recordingIndicator:" changeEvent="recordingIndicatorChanged">recordingIndicator</field>
	<field toGet="myDemon" toSet="myDemon:" changeEvent="myDemonChanged">myDemon</field>
	<field toGet="sanityChecker" toSet="sanityChecker:" changeEvent="sanityCheckerChanged">sanityChecker</field>
	<field toGet="isMeetingLeader" toSet="isMeetingLeader:" changeEvent="isMeetingLeaderChanged">isMeetingLeader</field>
	<field toGet="meetingState" toSet="meetingState:" changeEvent="meetingStateChanged">meetingState</field>
	<field toGet="stereoOn" toSet="stereoOn:" changeEvent="stereoOnChanged">stereoOn</field>
	<field toGet="proxy" toSet="proxy:" changeEvent="proxyChanged">proxy</field>
	<field toGet="hostCache" toSet="hostCache:" changeEvent="hostCacheChanged">hostCache</field>
	<field toGet="projectMorph" toSet="projectMorph:" changeEvent="projectMorphChanged">projectMorph</field>
	<field toGet="firstLaunch" toSet="firstLaunch:" changeEvent="firstLaunchChanged">firstLaunch</field>
	<field toGet="crashRecovery" toSet="crashRecovery:" changeEvent="crashRecoveryChanged">crashRecovery</field>
	<field toGet="authCache" toSet="authCache:" changeEvent="authCacheChanged">authCache</field>
	<field toGet="embeddedOverlay" toSet="embeddedOverlay:" changeEvent="embeddedOverlayChanged">embeddedOverlay</field>
	<field toGet="doOpenGL" toSet="doOpenGL:" changeEvent="doOpenGLChanged">doOpenGL</field>
	<field toGet="quindiLogging" toSet="quindiLogging:" changeEvent="quindiLoggingChanged">quindiLogging</field>
	<field toGet="roundTripSound" toSet="roundTripSound:" changeEvent="roundTripSoundChanged">roundTripSound</field>
	<field toGet="orgServices" toSet="orgServices:" changeEvent="orgServicesChanged">orgServices</field>
	<field toGet="orgSetupMutex" toSet="orgSetupMutex:" changeEvent="orgSetupMutexChanged">orgSetupMutex</field>
	<field toGet="inBenchTest" toSet="inBenchTest:" changeEvent="inBenchTestChanged">inBenchTest</field>
	<field toGet="awayOverlay" toSet="awayOverlay:" changeEvent="awayOverlayChanged">awayOverlay</field>
	<field toGet="okToQuit" toSet="okToQuit:" changeEvent="okToQuitChanged">okToQuit</field>
	<field toGet="visiblePanes" toSet="visiblePanes:" changeEvent="visiblePanesChanged">visiblePanes</field>
	<field toGet="uiStyle" toSet="uiStyle:" changeEvent="uiStyleChanged">uiStyle</field>
	<field toGet="snapshotBool" toSet="snapshotBool:" changeEvent="snapshotBoolChanged">snapshotBool</field>
	<field toGet="featureProperties" toSet="featureProperties:" changeEvent="featurePropertiesChanged">featureProperties</field>
	<field toGet="capabilities" toSet="capabilities:" changeEvent="capabilitiesChanged">capabilities</field>
	<field toGet="windowSharePane" toSet="windowSharePane:" changeEvent="windowSharePaneChanged">windowSharePane</field>
	<field toGet="infoCache" toSet="infoCache:" changeEvent="infoCacheChanged">infoCache</field>
	<field toGet="uiLatencyStats" toSet="uiLatencyStats:" changeEvent="uiLatencyStatsChanged">uiLatencyStats</field>
	<field toGet="fpsStats" toSet="fpsStats:" changeEvent="fpsStatsChanged">fpsStats</field>
	<field toGet="icmpSocket" toSet="icmpSocket:" changeEvent="icmpSocketChanged">icmpSocket</field>
	<field toGet="audioSession" toSet="audioSession:" changeEvent="audioSessionChanged">audioSession</field>
	<field toGet="logViewCharacterLimit" toSet="logViewCharacterLimit:" changeEvent="logViewCharacterLimitChanged">logViewCharacterLimit</field>
	<field toGet="recordingUsers" toSet="recordingUsers:" changeEvent="recordingUsersChanged">recordingUsers</field>
	<field toGet="broadcastingUsers" toSet="broadcastingUsers:" changeEvent="broadcastingUsersChanged">broadcastingUsers</field>
	<field toGet="mobileProxy" toSet="mobileProxy:" changeEvent="mobileProxyChanged">mobileProxy</field>
	<field toGet="crowdPane" toSet="crowdPane:" changeEvent="crowdPaneChanged">crowdPane</field>
	<field toGet="repoBrowser" toSet="repoBrowser:" changeEvent="repoBrowserChanged">repoBrowser</field>
</fields>'!

!QwaqParticipantUI class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QwaqParticipantUI class methodsFor: 'roles&capabilities' stamp: 'bgf 6/19/2008 14:23'!
allCapabilities

	"This is for documentation and development-time consistency checking
	 All the capability tokens for which we have implementation in the client."

	^ #(
			modify			"Not a guest - can operate on things in the space, modify and save."
			administer		"Can manage users, create forums from templates"
			invite			"Can invite new users."
		)! !

!QwaqParticipantUI class methodsFor: 'roles&capabilities' stamp: 'ar 5/3/2010 11:55'!
allFeatureKeys

	"This is for documentation and development-time consistency checking
	 All the EnabledFeatures keys for which we have implementation in the client."

	^ #(
			AnyUserCanInvite		"Members can invite members (not just admins)"
			DesktopEdition			"Can see and enter miramar spaces"
			InviteByRole				"Invitation dialog offers guest/member/admin choice"
			OfflineUploads			"Can files be uploaded without entering, and later retrieved."
			PersonalShare			"Create Miramar-style personal-shares by importing documents."
			PersonalShareInput		"In-space input to personal share"
			PersonalShareVNC		"Connect-to-VNC"
			PersonalShareColorKey	"Color keying in personal share"
			Phone					"Telephone call in/out stuff"
			ForumPages				"Forum Pages corresponding to forums. NOT Landmark Links."
			WebCam					"Web cameras - which have license/royalty constraints."
			Webcast				"Can webcast (i.e. stream video+audio)."
			"XXXXX: SessionBroadcast is deprecated... remove it in 2010."
			SessionBroadcast		"Can webcast (i.e. stream video+audio)."
			SessionRecording		"Can record sessions inside forums"
			PublicChat				"If this is false, chat panes and associated menus are unavailable"
			PrivateChat				"If this is false, private chat is unavailable"
			SecureVideo				"If true, secure-video feeds are available from the Video Feeds pane."
		)! !

!QwaqParticipantUI class methodsFor: 'utilities' stamp: 'bgf 10/10/2008 10:08'!
asMutedColor: aForm

	"Convert the form to a slighlty washed-out version of itself,
	 with a strong suggestion of the original color.
	 This is used wherever we're showing a forum preview 
	 that has activity going on inside it (ie. is occupied at the moment by others)"

	| result base |
	aForm ifNotNil: [
		"This is a bit costly, but the results are cached by callers..."

		base := aForm asGrayScale.

		"Flatten the contrast"
		base colors: (base colors collect: [:c | 
						| v | 
						v := c brightness.
						(Color r: (0.16 + (0.7 * v)) 
							    g: (0.16 + (0.7 * v)) 
							    b: (0.16 + (0.7 * v)))]).

		result := aForm deepCopy copyBits: base at: 0@0 translucent: 0.6.
	].
	^ result! !

!QwaqParticipantUI class methodsFor: 'utilities' stamp: 'bgf 10/9/2008 20:46'!
asTonedGray: aForm

	"Convert the form to a washed-out (grayed) version of itself,
	 with just a suggestion of the original color.
	 This is used wherever we're showing a forum preview that is not yet logged in."

	| result base |
	aForm ifNotNil: [
		"This is a bit costly, but the results are cached by callers..."

		base := aForm asGrayScale.

		"Flatten the contrast"
		base colors: (base colors collect: [:c | 
						| v | 
						v := c brightness.
						(Color r: (0.14 + (0.7 * v)) 
							   g: (0.14 + (0.7 * v)) 
							   b: (0.14 + (0.7 * v)))]).

		result := aForm deepCopy copyBits: base at: 0@0 translucent: 0.9.
	].
	^ result! !

!QwaqParticipantUI class methodsFor: 'roles&capabilities' stamp: 'bgf 8/15/2008 01:05'!
capabilitiesForRole: aRole

	"Hardcoded role-definitions.
 	 In the full tiered-service implementation, these will be derived
	 from authoritative data from the service provider."

	| result |
	result := aRole asLowercase caseOf: {

				['visitor'] -> [ #() ] .
				['guest'] -> [ #() ] .
				['user'] -> [ #(modify) ] .
				['admin'] -> [ #(administer invite modify) ] .

			} otherwise: [
				#(modify)
			].

	^ result asIdentitySet! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'sn 5/13/2011 23:12'!
cleanup
	"Do various bits of cleanup to ensure a safe restart"
	ScriptProcess allInstancesDo:[:sc| sc terminate].
	QStyles clear.
	QForms clear.
	QServiceProvider reset.
	TConnectionDispatcher reset.
	TContentManager default:nil.
	TIslandController initialize.
	TFarRef initialize.
	QClientSpeakerOutput allSubInstancesDo: [:v|v stop].
	QwaqSanityChecker allInstancesDo: [:i| i destroy ].
	TSpace allSubInstancesDo: [:tf | tf obliterate].
	QServiceDescription initialize.
	TObjectID resetLogs.
	Island classPool at:#Default put: nil.
	Smalltalk garbageCollect.
	TIslandController allSubInstancesDo: [:t|t destroy].
	TContactPoint allInstancesDo: [:tp | tp destroy].
	MCMethodDefinition shutDown.
	MCFileBasedRepository flushAllCaches.
	MCDefinition clearInstances.
	Smalltalk garbageCollect.
	ChangeSet allInstances do:[:cs|
		cs == ChangeSet current
			ifFalse:[ChangeSet removeChangeSet: cs]].
! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 6/22/2008 03:26'!
defaultExtent
	"Initial size of the unsealed development app window"
	"So the CPanel, the Window, and the partsbin morph agree."
	^ 920@600! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 2/5/2009 22:41'!
defaultFullWindowSize
	"Initial size of the sealed application's full window (once we're in a space)"
	^ 1280@840.	

	"Typical 'widescreen' midsize laptop is 1440@900" ! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 2/13/2009 22:59'!
defaultSmallWindowSize
	"Initial size of the unsealed main window (once in a space, in the unsealed case.)"
	^ 820 @ 560.	
! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 10/22/2008 22:32'!
defaultSplashForm

	"This is the 'splash' background that appears before login,
	 before any preview graphic has been identified."

	^ (QForms current find: 'icons/Qwaq/splash/Sky.jpg' ifAbsent: [
			QForms current import: 'Sky.jpg' from: 'icons/Qwaq/splash'
			]).

! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'ar 6/27/2007 12:40'!
defaultSystemErrorHandler
	"Needs to be created in a separate method to avoid sharing the home context"
	^(SystemErrorHandler new handlerBlock:[:ignored]).! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 2/5/2009 20:14'!
defaultWindowSize
	"Initial size of the sealed application's initial window"
	^ QwaqLogin lessExtent! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'bgf 7/11/2009 17:02'!
flushSurplusPayload

	"CAUTION.
	 Drop the versioning information in the main registry.
	 DO NOT EVALUATE THIS IN A WORKING IMAGE"

	MCWorkingCopy allInstancesDo: [ : wc |  
		wc ancestry ifNotNil: [
			wc ancestry trimAfterVersionInfo: (wc ancestry).
		]].
	MCFileBasedRepository flushAllCaches.

	PartsBin initialize.		"I.e. without all the thumbnailing registries."
	PaintBoxMorph initialize.
	(PaintBoxMorph bindingOf: #Prototype) value: nil.	"Hangs on to a bunch of color forms"

	TextConstants removeKey: 'Accuve' ifAbsent: [ ].
	TextConstants removeKey: 'Accumon' ifAbsent: [].
	TextConstants removeKey: 'Accuat' ifAbsent: [].
	TextConstants removeKey: 'Accusf' ifAbsent: [].
	TextConstants removeKey: 'Atlanta' ifAbsent: [].
	TextConstants removeKey: 'BitstreamVeraSerif' ifAbsent: [].
	Smalltalk garbageCollect.

! !

!QwaqParticipantUI class methodsFor: 'utilities' stamp: 'ar 1/22/2008 11:02'!
generatePassword: length
	"Generate a random password of the given length"
	"QwaqParticipantUI generatePassword: 8"
	| chars idx rnd |
	rnd := Random new.
	^String streamContents:[:s|
		chars := #('aeiou' 'bcdfghkmnprstvwxz').
		1 to: length // 2 do:[:i|
			idx := (1 to: 2) atRandom.
			s nextPut: ((chars atWrap: idx) atRandom: rnd).
			s nextPut: ((chars atWrap: idx+1) atRandom: rnd)]].
! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 6/22/2008 14:47'!
imageVersionsFileName
	"Subclasses can override to for app-specific name"
	^'ImageVersions.xml'! !

!QwaqParticipantUI class methodsFor: 'class initialization' stamp: 'bgf 3/11/2009 14:11'!
initialize
	"QwaqParticipantUI initialize"
	Smalltalk addToShutDownList: self.
	KillRenderOnError := true.! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'ar 6/21/2010 12:29'!
interruptPassword: password
	"Sets the interrupt password"

	InterruptPassword := TCryptoRandom md5HashMessage: password.
	password atAllPut: $*.
! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'ar 6/17/2010 15:04'!
launchAppFrame
	"Launch the application frame"
	| seed root |
	"Errors that happen during launch will just have us die. 
	Not much we can do about it since there is no UI yet."
	SystemErrorHandler default: (SystemErrorHandler handlerBlock:[:err|
		[Smalltalk logError: err description
			inContext: err signalerContext 
			to: (FileDirectory logDirectory fullNameFor: self name,'Debug.log')] on: Error do:[].
		Smalltalk quitPrimitive]).
	"Set the VM log directory for crashes"
	Smalltalk vmLogDirectory: FileDirectory logDirectory pathName.
	"Eeek. Reseed islands manually (this should happen automatically!!!!!!)"
	Island allSubInstancesDo:[:each|
		seed := ByteArray new: 128.
		TCryptoRandom gatherEntropy: seed.
		each randomStream key: seed.
	].
	root := CProjectMorph new.
	root bounds: World bounds.
	root noRedraw: true.
	root borderWidth: 0.
	root newProject: self.
	root color: QColors paneDarkerColor.
	root projectWorld projectMorph: root.
	root openInWorld.

	^root! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'bgf 7/6/2008 12:21'!
launchRestart: bSave
	"Tricky. Fire off restartAsApp in a separate context so that we don't hold on to the previous UI and can GC it properly.
	If bSave is true, the image is saved just prior to restart."
	WorldState addDeferredUIMessage:[self restartAsApp: bSave].! !

!QwaqParticipantUI class methodsFor: 'defaults' stamp: 'bgf 2/19/2009 12:22'!
minDisplaySize
	"Large enough to keep the menu available."
	^ 580@240! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'bgf 7/5/2008 11:52'!
restartAsApp: bSave

	"Restart me after an update cycle completed"
	"Does various cleanups before we restart.
	 If bSave is true, save the image before the restart."

	SystemErrorHandler default: self defaultSystemErrorHandler.
	QwaqVersion current isSealedApplication ifTrue:[
		Sensor installInterruptWatcher:[].
		Smalltalk installLowSpaceWatcher:[].
	].
	Cursor wait showWhile:[
		self cleanup.
		World doOneCycleNow.
		World primaryHand keyboardFocus: nil.
		World primaryHand mouseFocus: nil.
		Smalltalk garbageCollect. 
		Symbol compactSymbolTable.
	].
false ifTrue:["this is useful for testing"
	self someInstance ifNotNilDo:[:prior| ^PointerFinder on: prior].
].
	QwaqVersion current isSealedApplication ifTrue:[
		bSave ifTrue: [ Smalltalk snapshot: true andQuit: false ].
		self launchAppFrame.
	] ifFalse:[
		CProjectMorph open: self.
	].! !

!QwaqParticipantUI class methodsFor: 'roles&capabilities' stamp: 'bgf 8/15/2008 01:16'!
roleLabelToRole: roleLabel

	"Conversions between service-provider role string and user-visible equivalent"
	^ roleLabel caseOf: {
		[ 'Admin' ] -> [ 'admin' ].
		[ 'Member' ] -> [ 'user' ].
		[ 'Guest' ] -> ['guest'].
		[ 'Visitor' ] -> [ 'guest' ]. 		"TBD - at some point the SP will switch!!"
	} otherwise: [
		"Don't know - pass through "
		roleLabel
	].
		! !

!QwaqParticipantUI class methodsFor: 'roles&capabilities' stamp: 'bgf 8/15/2008 01:15'!
roleToRoleLabel: role

	"Conversions between service-provider role string and user-visible equivalent"
	^ role caseOf: {
		[ 'admin' ] -> [ 'Admin' ].
		[ 'user' ] -> [ 'Member' ].
		[ 'guest' ] -> ['Visitor'].
		[ 'visitor' ] -> [ 'Visitor' ].
	} otherwise: [
		"Don't know - pass through "
		role
	].
		! !

!QwaqParticipantUI class methodsFor: 'tests' stamp: 'ar 4/16/2007 18:03'!
runAllTests: quitWhenDone
	"Run all the tests"
	| runner |
	Utilities authorInitialsPerSe isEmptyOrNil 
		ifTrue:[Utilities setAuthorInitials: 'qwaq'].
	runner := QTestRunner new.
	runner openLog: 'testResults.log'.
	[runner runAllTests.
	runner printSummary.
	] ensure:[runner closeLog].
	quitWhenDone ifTrue:[Smalltalk quitPrimitive].! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'ar 6/21/2010 12:26'!
runAsApp
	"Run me as a standalone app, sealed without development information."

	^self runAsApp: true.! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'ar 4/16/2011 19:00'!
runAsApp: interactive
	"Run me as a standalone app, sealed without development information.
	When running non-interactive (i.e., as ant task) it is assumed that:
		* The interrupt password has been set
		* The image has the right name already.
	"

	| version passA passB |
	interactive ifTrue:[
		version := UIManager default 
			request: 'Enter the version string for the sealed image.' 
			initialAnswer: QwaqVersion current version.
		(version isNil or: [ version withBlanksTrimmed = ''] )  ifTrue: [ 
			self inform: 'Sealing canceled.\(Must give a version)' withCRs.
			^ self.
		].
		QwaqVersion newVersion: version.

		passA := UIManager default requestPassword: 'Please set the interrupt password'.
		passB := UIManager default requestPassword: 'Repeat password to confirm'.
		passA = passB ifFalse:[^self error: 'Password mismatch'].
		self interruptPassword: passA.
		passA := passB := nil.
	].

	InterruptPassword isEmptyOrNil ifTrue:[self error: 'Interrupt password not set'].

	"Save the world with the same background as the login-level of Forums;
	 reduces flicker during startup."
	World color: (QColors paneDarkerColor).
	Smalltalk windowSize: (self defaultWindowSize + (12@36)).	"Frame + slop"	
	QwaqVersion current seal.

	"Drop the versioning trails and caches."
	self flushSurplusPayload.
	BitBlt initialize.		"Dump native font colormaps"
	KillRenderOnError := true.

	WorldState addDeferredUIMessage:[
		World submorphs do:[:each| each delete].
		Preferences disable: #warnIfNoSourcesFile.
		Preferences disable: #warnIfNoChangesFile.
		(interactive not or:[self confirm: 'Save the sealed image?']) ifTrue:[
			AutoStart deinstall. "no autostart please"
			self setVMParameters.
			interactive 
				ifTrue:[Smalltalk saveAsSuggesting: 'OpenQwaq']
				ifFalse:[Smalltalk snapshot: true andQuit: true].
		].
		self launchAppFrame.
	].! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'bgf 12/5/2009 16:53'!
setVMParameters
	"Set the VM parameters that only take effect after restarting the image"

	"Desired number of stack pages"
	Smalltalk vmParameterAt: 43 put: 128.

	"Desired eden size"
	Smalltalk vmParameterAt: 45 put: 4 * 1024 * 1024.

	"Compiled code space size"
	Smalltalk vmParameterAt: 47 put: 2 * 1024 * 1024.
! !

!QwaqParticipantUI class methodsFor: 'shut down' stamp: 'jcg 5/20/2008 08:52'!
shutDown
	"Close existing Qwaq Forums instances before saving, so that the image isn't borked when it next starts."	

	QwaqVersion current isSealedApplication ifTrue: [
		"We must be in the process of sealing the application, so we really do want to save with the QwaqParticipantUI window open."
		^self
	].

	SystemWindow allSubInstancesDo: [:win |
		[
			((win model class == CProjectMorph) and: 
			[(win model project world get: #class) inheritsFrom: QwaqParticipantUI])
				ifTrue: [win delete]
		] on: Error do: [nil "just to be safe"].
	]
							! !

!QwaqParticipantUI class methodsFor: 'roles&capabilities' stamp: 'jcg 10/8/2008 00:14'!
tosLabelForOrg: anOrg

	"Given an org record, 
	 return the user-visible label to describe it's tier-of-service."
	| tos dtd dts days |

	anOrg ifNil: [ ^ '' ].
	anOrg expiryDate ifNotNilDo: [ : dt | 
		dtd := dt asDate.
		dts := true caseOf: {
				[ dtd = (Date today) ] -> [ 'Expires Today' ] .
				[ dtd = (Date tomorrow) ] -> [ 'Expires Tomorrow' ] .
			} otherwise: [ 
				days := ((dtd - (Date today)) duration days).
				days < 2 ifTrue: [
					'Expires in One Day'
				] ifFalse: [
					'Expires in ', days, ' Days'
				]]].
		

	"This is used to decorate the UI... only show nice names we recognize."
	tos := anOrg tierName.
	^ tos caseOf: {

			[ 'PersonalTier' ]		-> [ 'Personal Edition' ] .
			[ 'PersonalEdition' ]  -> [ 'Personal Edition' ] .

			[ 'GroupTier' ]	 	-> [ 'Group Edition' ] .
			[ 'GroupEdition' ]	-> [ 'Group Edition' ] .

			[ 'EnterpriseTier' ]	-> [ 'Enterprise Edition'  ].
			[ 'EnterpriseEdition' ] -> [ 'Enterprise Edition'  ].

		} otherwise: [ 
			anOrg isFreeTrial ifTrue: [ ^ 'Trial', (dts ifNotNil: [ ' - ', dts ] ifNil: ['']) ].
			'' 
		].
! !

!QwaqParticipantUI class methodsFor: 'deployment' stamp: 'ar 4/16/2007 14:05'!
upgrade: urlString restart: aBool
	"Upgrade the system from a url referring to an MC configuration map"
	"QwaqParticipantUI upgrade: 'http://user:pass@foo.bar.com/updates/update.mcm'"
	| url username password repository mcm |
	url := urlString asUrl.
	username := url username.
	password := url password.
	repository := MCHttpRepository
		location: (url scheme,'://', url authority, (url path allButLast inject: '' into:[:s :p| s,'/',p]))
		user: username
		password: password.
	repository readStreamForFileNamed: url path last do:[:s|
		mcm := MCConfiguration fromArray: (MCScanner scan: s).
	].
	mcm repositories do:[:repo| repo user: username; password: password].
	mcm updateFromRepositories.
	[mcm upgrade] on: Warning do:[:ex| ex resume].

	aBool ifTrue:[
		MCFileBasedRepository flushAllCaches.
		World submorphs do:[:m| m delete].
		Smalltalk snapshot: true andQuit: true.
		self launchAppFrame.
	].
! !

!QwaqParticipantUI class methodsFor: 'utilities' stamp: 'ar 4/16/2011 19:00'!
versionedUserGuide: topic
	^'wwhelp/wwhimpl/api.htm?context=openqwaq_3_0_user_guide_ww&topic=', topic.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 8/29/2006 00:24'!
aboutBox
	"Answer the aboutBox of the receiver"
	<bewareOf: #aboutBoxChanged>
	^self propertyValueAt: #aboutBox! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 8/29/2006 00:24'!
aboutBox: aValue
	"Modify the receiver's aboutBox"
	^self propertyValueAt: #aboutBox put: aValue with: #aboutBoxChanged! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:45'!
actionsMenu
	| menu |
	"The empty starting point for the space verbs (actions) menu."

	menu := CMenu new.
	menu add: 'Action...' translateMe target: self action: #noSuchAction.
	self runScript: #updateActionsMenu when: {menu . #open}.

	^menu! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 12/4/2007 23:40'!
activateService: serviceUrl
	"Activate (that is resolve and join) the service with the given service url.
	Service url must be fully-formed absolute url of the form:
		//hostname:port/realm/service
	Returns the resulting QServiceRecord."
	
	[
		self flowControlFlag: true.
		^ self activateService: serviceUrl fromTemplate: nil usingIslandFiles: nil
	] ensure: [
		self flowControlFlag: false.
	].! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 4/16/2011 17:28'!
activateService: serviceUrl fromTemplate: aTemplateOrNil usingIslandFiles: aDictOrNil

	"Activate (that is resolve and join) the service with the given service url.
	Service url must be fully-formed absolute url of the form:
		//hostname/realm/service
	Returns the resulting QServiceRecord.

	The template, if supplied, is the QServiceRecord of a service template;
 	for each island in the service that doesn't yet have a snapshot file,
	we'll look for  a snapshot in the template.  (If the template doesn't
	have a snapshot, then we generate from the bootstrap class.)
	QServiceLink handles the nitty-gritty details of that.

	If islandFiles (a dictionary mapping island names to .c3z file names)
      is supplied (for reverting to older versions) then it
	overrides the default latest-file lookup for those islands.  
	If for any reason a specified file cannot be used, the operation fails."

	|  normalizedUrl svcMgr response svcRecord link svcName realm |
	
	firstArrivalPending := false.	"Until we learn otherwise."
	normalizedUrl := QServiceLink normalizedServiceUrl: serviceUrl.

	"Do we have the service already?"
	svcRecord := urlCache at: normalizedUrl ifAbsent:[nil].
	svcRecord ifNotNil:[^svcRecord].

	"We do not allow cross-provider links."
	self statusInfo: 'Checking access: ', normalizedUrl.
	self checkLinkHost: normalizedUrl.

	svcName := QServiceLink serviceNameFromServiceUrl: serviceUrl.
	realm := QServiceLink realmFromServiceUrl: serviceUrl.

	"XXXX: This needs to be looked up from the set of cached svc managers"
	svcMgr := svcMgrCache at: realm ifAbsent:[nil].
	svcMgr ifNil:[^self error: 'Cross-org links are not supported'].

	"Ask the service manager for the service record (which verifies it is still present)"
	response := svcMgr queryService: svcName.
	response name == #ok ifFalse:[
		QActivityCapture action: #LoginForumNoForum details: serviceUrl.
		self error: 'Cannot enter ', svcName, String cr, String cr, response contentString].

	svcRecord := QServiceRecord new readFullFrom: (response elementAt: 'service').
	svcRecord baseUrl: svcMgr baseUrl.

	self checkEntryEntitlement: svcRecord.
	self checkDuplicateLogin: svcRecord urlString: serviceUrl serviceManager: svcMgr.

	"If it's a template, we don't activate it directly; 
	 rather we divert and copy it to a newly-named instance.
	 (Note that the tail end of instantiateTemplate: will call this method again,
	 on a real service instance.)"
	svcRecord isTemplate ifTrue: [ ^ self instantiateTemplate: svcRecord ].

	QActivityCapture action: #LoginForum details: serviceUrl.

	self informIfEnteringMeeting: serviceUrl.
	firstArrivalPending := true.
	link := QServiceLink forClient: self.
	link session: session.
	link substituteFiles: aDictOrNil.
	link template: aTemplateOrNil.
	link serviceManager: svcMgr.
	link service: svcRecord.

	"At login, set a serviceSvc here, just before going into low-level activate,
	 since some resource-setup methods seem to want serviceSvc."
	serviceSvc ifNil: [ serviceSvc := svcRecord ].
	[ 
		link joinService ifNil:[
			"If we got nil rather than error-ing, assume the user's been informed already."
			QActivityCapture action: #LoginForumJoinRejected details: serviceUrl.
			CProgressCancel signal.
		]
	] on: Error do: [ : ex |
		"Forgetting prefetches is easier for us and for everyone... reduces dl traffic."
		(session notNil and: [ session documentManager notNil ]) ifTrue: [
			session documentManager discardPrefetches.
		].
 		((ex isKindOf: CProgressCancel) 
		 or: [ex description includesSubString: 'occupants are using' ]
		 or: [ex description includesSubString: 'Access to the Forum:' ]
		 or: [ex description includesSubString: 'Access denied']) 
		 ifFalse: [
				self reportIncident: 'Forum entry unsuccessful'
					error: ex
					stack: nil ].
		ex pass.
	].
	
	urlCache at: normalizedUrl put: svcRecord.
	session flushLinksMatching: normalizedUrl.
	controlBar updateServiceCount.	
			"(Could be a trigger if there were an observable collection of services.)"

	self invalidate; waitTick. "refresh display"
	^svcRecord! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 2/15/2008 23:17'!
activateServiceFromUrl: url

	"Activate an existing forum or template;
	 this method may raise Error or QwaqRestartException."
	^self activateServiceFromUrl: url usingIslandFiles: nil! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 2/15/2008 23:16'!
activateServiceFromUrl: url usingIslandFiles: aDictOrNil

	"Activate a service from an absolute url.

	 If the dictionary is supplied, it contains a map from island names to .c3z file names.
	 The startup process will require that those precise c3z files be used to activate.
	 If that cannot be done (for instance if the forum has been activated elsewhere)
	 then the operation fails with an Error."
	
	| svcUrl result |
	"Tolerate network errors."

	svcUrl := [ self serviceUrlForUrl: url ] on: Error do: [ nil ].
	svcUrl ifNotNil: [
		self flowControlFlag: true.
		[ result := self activateService: svcUrl fromTemplate: nil usingIslandFiles: aDictOrNil ]
		  ensure: [self flowControlFlag: false].
	].
	result ifNil: [
		self error: 'Service not available: ', url.
	].

	^result! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 6/20/2009 10:26'!
activeServices
	"Return an array of QServiceRecords denoting the forums I am logged into."
	^ (self urlCache select: [ : item | item class includesBehavior: QServiceRecord ])
		values asArray
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:19'!
activeSpace
	"Answer the activeSpace of the receiver"
	<bewareOf: #activeSpaceChanged>
	^self propertyValueAt: #activeSpace! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:19'!
activeSpace: aValue
	"Modify the receiver's activeSpace"
	^self propertyValueAt: #activeSpace put: aValue with: #activeSpaceChanged! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 8/28/2009 08:41'!
addAnnotation: aClass asDrop: bDrop
	"Add a new QAnnotation  (of the given concrete class, aClass)
	 at the point where the user clicked."
	| selection target |

	self readyForEvents ifFalse: [ ^ false ].

	"Force render w/ picking"
	mainView pushTask: TPickAll new.
	selection := mainView fastPickSelection.
	mainView popTask.
	target := selection ifNotNil: [ selection target ].
	target ifNil:[ 
		self statusAlert: 'There is nothing there to put a note on.'.
		^ false.
	].
	selection normal ifNil: [
		self statusAlert: 'You cannot put a note on that object.'.
		^ false
	].
	self readyForEvents ifFalse: [ ^false ].

	bDrop ifFalse: [
		"Special-case; if you click a topic card, you open its topic-discussion editor."
		((target get:#class) includesBehavior: QTopicCard)  ifTrue:[
			target future editByUser: (avatar replica).
			^ false
		]].
	(mainView avatar replica send: [ :rep | rep isOrContains: target ])  ifTrue:[ 
		"Could be reconnect or other nonsense that the system will deal with;
	 	 for this activity, just let the user try again."
		self statusAlert: 'You cannot put a note on your avatar.'.
		^ false
	].
	avatar replica future addAnnotation: aClass toFrame: target selection: selection. 
	QLogger logInfo: 'Sticky note inserted.'.
	^ true! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 12/30/2012 18:54'!
addFixedItemsToActionsMenu: menu
	"Add the predefined final items (provided regardless of what space we're in)
	 to the actions menu. (This forms the entire actions menu for visitors.) "
	|  item |

	(menu add: 'Reset Tool Layout' translateMe target: self action: #restoreDefaultLayout)		
		enabled: (
			 self floatingPanes anySatisfy: [ : fp | fp hasFactoryPlacement not]).

	menu menuItems last separator: true.
	"Happens frequently enough that there's really no point in confusing people with more choices.
	self canModifyForum ifTrue: [
		self hasOfflineUploadSupport ifTrue: [
			menu add: 'Check for Offline Uploads' target: customsPane 
				action: #processOfflineUploadsWithNotice: arguments: {true}.
		]]. Unmatched comment quote ->"
	menu add: 'Compose Mail' translateMe target: self action: #mailUsers 
		disabled: self hasForumPagesSupport not.
	menu add: 'Copy Landmark Link' translateMe target: self action: #copyQRLForCurrentPlace.
	menu add: 'Open Forum Page' translateMe target: self action: #openQrlForCurrentForum
		disabled: self hasForumPagesSupport not.
	menu menuItems last separator: true.
	self hasSessionRecordingSupport ifTrue: [
		self canModifyForum ifTrue: [
			item := menu add: 'Record Session...' translateMe target: recorderPane action: #toggle.
			self linkItem: item toTool: recorderPane.
		].
	].
	self hasWebcast ifTrue: [
		self canModifyForum ifTrue: [
			item := menu add: 'Webcast Session...' translateMe  target: webcastPane action: #toggle.
			self linkItem: item toTool: webcastPane.
		].
	].
	self hasSecureVideo ifTrue: [
		(
			menu
				add: 'Secure Broadcast'
				target: self
				action: #toggleSecureBroadcast
		)
			checked: ((avatar replica future get: #cameraEnabled) wait; value);
			enabled: self canModifyForum].

	(menu add: 'Take Screen Shot' translateMe  target: self action: #takeScreenShot)
		enabled: (session notNil and: [session ogl notNil]).

	QStyles installMenuFontInMenu: menu.
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 2/23/2008 11:03'!
addFloatingPane: aPane
	"Add a QFloatingPane to the UI."
	self add:aPane.
	aPane owner: self.
	aPane hide.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:43'!
addSubMenusTo: aMenu
	{
		{'File' translateMe. self fileMenu}.	
		{ 'Edit' translateMe . self editMenu }.
		{'View' translateMe. self viewMenu }.
	   	{'Meeting' translateMe. self meetingMenu } . 
		{'Tools' translateMe. self toolsMenu}.
		{'Actions' translateMe. self actionsMenu}.
		{'Forums' translateMe. self forumsMenu}.

		{'Extras' translateMe. self extrasMenu}.
		{'Developer' translateMe. self magicDeveloperMenu}.
		{'Help' translateMe. self helpMenu }.

	} do:[:spec|
		spec last ifNotNil:[
			aMenu add: spec first subMenu: spec last.
		].
	].! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 1/6/2010 11:26'!
addToInventory: inventoryCardZip
	"Relay a new inventory-card (its persistent XML zip) 
	 to the catalog for inclusion in the user's inventory."

	catalogPane addToInventory: inventoryCardZip.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 12/31/2006 15:41'!
alertFromAvatar: aString
	"Private handler; see updateStatusRelay."
	statusLine alert: aString.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 12/31/2006 15:41'!
alertFromSpace: aString
	"Private handler; see updateStatusRelay."
	statusLine alert: aString.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 8/2/2009 12:36'!
allowedToMove
	"True when I am permitted to move about  in the current meeting.
	 (False only if the leader has stay-put the audience.)"
	^ self meetingState ifNotNilDo: [ : meeting |
			^ meetingState isLoginAllowedToMove: self userLogin
		] ifNil: [ 
			true
		].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 4/16/2011 17:28'!
appName
	"Bare default. This is used to build various title strings starting in #initialize,
	 so for subclasses to customize their names, simply override this method."
	^ 'OpenQwaq'! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 4/14/2010 17:54'!
appServiceLabelString
	"The string describing where I am, for window title. Add a note if I'm a demon."
	| dirty |
	serviceMgr ifNil: [ ^ (self appName)].

	dirty := false.
	serviceSvc ifNotNil: [ 
		( servicesPane entryForUrl: serviceSvc url ) ifNotNil: [ : e | dirty := (self serviceEntryIsDirty: e) ] ].

	^ (String streamContents: [ : str |
		str nextPutAll: (self appName).
		serviceSvc ifNotNil: [
			str nextPutAll: '  |  ';
				nextPutAll: (self currentOrgName);
				nextPutAll: '  |  ';  
				nextPutAll: ((self currentForumName)).
			dirty ifTrue: [
				str nextPutAll: '  *'
			].	
		] ifNil: [
			str nextPutAll: '  |  Lobby'
		]]).
! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'sn 5/13/2011 14:48'!
applyUserRole
	"A new user role has been put into play for this user...
	 update all the non-self-regulating menu items, etc, in the UI "

	self repopulateToolsMenu: (self menu itemNamed: 'Tools' translateMe) subMenu.

	controlBar applyUserRole.
	usersPane applyUserRole.
	servicesPane applyUserRole.

	self updateTitle.

	"Pass along into the avatar..."
	self readyForEvents ifTrue: [
		avatar replica
			future capabilities: (self capabilities asArray)
			features: (self currentFeatures asArray).	
	].

	self updateMobility.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 8/30/2009 01:29'!
askEveryoneAQuestion: evt
	self readyForEvents ifTrue: [
		QPollBarChart prepareQuestionForClient: self
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 3/19/2009 17:00'!
assureDisplaySize: anExtent
	"Make the window big enough to accomodate the given rectangle."
	| size |
	size := self windowDisplaySize.
	((size x < anExtent x) 	
	 or: [size y < anExtent y])
	 ifTrue: [
		self setWindowDisplaySize: (anExtent max: size) ].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 7/22/2009 01:05'!
assureOrgServices: aSvcManager
	
	"Ensure that I have an org services connected 
	 to the organization for which the service manager is configured."

	"In the check below, testing isConnected gives us some reconnect-support
	 in the early phases.  Automatic reconnect doesn't come into play
 	 until there is a full CroquetSession running, and we don't have that during login."

	"Don't let the mutex prevent us quitting!!
	 The error-capture of the stuff around the connection clause below protects us."
	(aSvcManager isNil and: [orgServices notNil]) ifTrue: [ 
		self stopScript: #robot:mode:.
		orgServices destroy.
		orgServices := nil.
		^ nil
	].

	orgSetupMutex critical: [
		orgServices ifNotNil: [
			((orgServices isConnected not) 						"Need to reconnect"
			 or: [ aSvcManager isNil ]
			 or: [ orgServices orgRecord ~= aSvcManager currentOrg ]	"Org change."
			) ifTrue: [
				self stopScript: #robot:mode:.
				orgServices destroy.
				orgServices := nil.
			].
		].
		(orgServices isNil 
		  and: [aSvcManager notNil] 
		  and: [aSvcManager currentOrg notNil]) ifTrue: [
			[
				orgServices := QOrgServices forClient: self serviceManager: aSvcManager.
				self startScript: #robot:mode: when: {orgServices . #robot:mode:}.
			] on: Error do: [ :ex |
				"This  is unsatisfying.    
				  If the org services login fails but the service manager is ok, what to do?   
				 For now,  raise a notification so it doesn't go unnoticed internally.
				 We'd have to restructure the service chooser org+org choices regime
	 			 to do better this, since the org choice page is up and needs -something- in it."

				self startScript: #inform:
				       withArguments: {'There was a problem connecting.\Location services are not available.' withCRs }.
				QLogger logWarning: ('No org services: ', ex) from: self.
				orgServices ifNotNil: [ orgServices destroy ].
				orgServices := nil.
			]].
	].
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/6/2009 20:35'!
audioSession
	"Answer the audioSession of the receiver"
	<bewareOf: #audioSessionChanged>
	^self propertyValueAt: #audioSession! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/6/2009 20:35'!
audioSession: aValue
	"Modify the receiver's audioSession"
	^self propertyValueAt: #audioSession put: aValue with: #audioSessionChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 8/15/2007 16:32'!
authCache
	"Answer the authCache of the receiver"
	<bewareOf: #authCacheChanged>
	^self propertyValueAt: #authCache! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 8/15/2007 16:32'!
authCache: aValue
	"Modify the receiver's authCache"
	^self propertyValueAt: #authCache put: aValue with: #authCacheChanged! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 5/6/2009 15:24'!
authenticateConnection: info
	"Determine whether we can connect using the given connection info.
	If (proxy) authentication is required handle the authentication request."
	| ok dialog pw usr realm authInfo wasSaved showDialog |
	"Repeatedly attempt to ping the connection if we have new auth info"
	[true] whileTrue:[
		[ok := self pingConnectionInfo: info] on: QProxyUnauthorized do:[:ex| ex return].
		ok ifNotNil:[^ok].
		proxy := info proxy ifNil:[^false].
		"Note: Realm may not be used with basic auth so don't insist on it"
		realm := proxy authParams at: 'realm' ifAbsent:['Unknown Proxy Realm'].

		"The first attempt may use auth info from the cache without user interaction"
		authInfo := authCache at: realm ifAbsentPut:[self loadAuthInfo: realm].
		"Show the dialog if either the prior credentials were incorrect,
		or we don't have default authentication info for this realm."
		showDialog := proxy username notNil or:[authInfo == nil].
		authInfo ifNotNil:[
			usr := authInfo base64Decoded copyUpTo: $:.
			pw := QPasswordHolder with: (authInfo base64Decoded copyAfter: $:).
		].
		showDialog ifTrue:[
			"Offer the login dialog"
			wasSaved :=  (self loadAuthInfo: realm) notNil.
			dialog := QUserLoginDialog new.
			dialog title value: 'Proxy Authentication Required '.
			dialog explanation value: String cr, 
				'Enter username and password for "', realm,'"'.
			dialog userName value: (usr ifNil:[login user value]).
			dialog password value: (pw value ifNil:['']).
			dialog savePassword value: wasSaved.
			dialog acceptAction:[pw := QPasswordHolder with: dialog password value. dialog close].
			dialog cancelAction:[pw := nil. dialog close].
			dialog openCenteredAt: self dialogCenter.
			login disable.
			dialog waitUntil: #closed.
			login enable.
			pw ifNil:["user cancelled"
				QProxyHandler log: 'Login canceled' from: self class.
				^false].
			usr := dialog userName value withBlanksTrimmed.
			authInfo := (usr,':',pw value) base64Encoded.
			wasSaved = dialog savePassword value ifFalse:[
				self saveAuthInfo: realm value: (wasSaved ifTrue:[nil] ifFalse:[authInfo]).
			].
		].
		proxy username: usr.
		proxy password: pw value.
	].
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'ar 4/16/2011 17:28'!
autoLogin2: orgOrEmpty forum: forum location: placeName robotic: robot
	"Phase 2 of autoLogin, after we're logged in successfully to the serviceChooser."
	|  forumToLoad org |
	forumToLoad := forum.

	self autoLoginCheckServiceList: robot.

	org := orgOrEmpty ifEmpty: [serviceChooser orgChoices items first].
	
	(robot or: [QWebServices match: org against: self serviceMgr currentOrg name caseSensitive: false]) ifFalse: [
		"The service chooser picks the first one in the list if it can't find what you ask for."
		CDialog inform: 'Your organizations do not include ', org, '.' label: 'Cannot Log In'.
		login forumGoal: nil. "allow more tries"
		^self stopScript: #autoLogin2:forum:location:robotic:. "but let them finish themselves"
	].

		forumToLoad ifNil: [
			robot ifFalse: [^self].  "Unless we're a robot, just stop here and let the user pick."
			"Default robot case... just get into the first forum, if we can,
	  		 or bail, if we cannot; we don't want robots stuck at login."
			[
				forumToLoad := serviceChooser serviceList first name.
			] on: Error do: [
				QLogger logError: 'Robot finds no Forums in ', org.
				self stopScript: #startAsDemon:.
				^ self onExit.
			].
		] ifNotNil: [
			forumToLoad :=  self matchForum: forumToLoad.
		].
		robot & StafProxy isRunning not  ifTrue: [
			"For demons on Unix,  openAL setup is more likely to cause problems."
			self preferences ifNotNilDo: [ : prefs | prefs useOpenAL: false ].
		].
		(self serviceChooser selectServiceByName: forumToLoad) ifNotNilDo: [ : svc |

			QLogger logInfo: 'Logging in to ', forumToLoad, ' in ', org.
			placeName isEmptyOrNil ifFalse: [
				self startScript: #goToPlaceName: withArguments: {placeName} when: {self. #ready}. 
			].
			(Delay forMilliseconds: 500) wait.		"Long waitTick... let the selection propagate."
			self onServiceLogin: svc.

		] ifNil: [
			QLogger logError: 'Cannot find Forum ', forumToLoad, ' in ', org.
			robot ifTrue: [
				^ self onExit.
			] ifFalse: [
				CDialog inform: 'Your Forums do not include ', forumToLoad, '.' 
					label: 'Cannot Log In'.
				login forumGoal: nil. "allow more tries"
			].
		].
	self stopScript: #autoLogin2:forum:location:robotic:.
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 3/26/2009 17:06'!
autoLogin: host organization: org forum: forum location: placeName robotic: robot
	"If a forum name was provided, or we're starting robotic, log in all the way."
	login serverList value: host.
	login organization value: org.

	self startScript: #autoLogin2:forum:location:robotic: 
		withArguments: {org. forum. placeName. robot}
		when: {self. #unready}.

		self login loginButton enable.		"Go around accelerator-key guards."
		self waitTick.	"The input fields need to absorb their values."
		self login onLogin.
		self waitTick.

		"robots don't wait for someone to push the button.  
		QRL users might have login failure, so we'll wait for the signal before checking them."
		robot ifTrue: [self autoLoginCheckServiceList: robot].  

! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 9/2/2009 12:00'!
autoLoginCheckServiceList: robot
	"Wait for the serviceList, but not indefinitely."
	| count  |

		count := 0. 
		[serviceChooser serviceManager] whileNil: [
			count := count + 1.
			count > 450 ifTrue: [  "Give time to load patches, if any"
			QLogger logError: 'Robot cannot login to ',  self currentOrgName.
			self stopScript: #startAsDemon:.
			robot ifTrue: [^ self onExit].
			].
			(Delay forMilliseconds: 200) wait.
		].

		count := 0.
		[ serviceChooser serviceList isEmpty | (self preferences isNil) ] whileTrue: [
			count := count + 1.
			count > 450 ifTrue: [
				QLogger logError: 'Cannot see service list after 90 seconds ', 
									(serviceChooser orgChoices value).
				self stopScript: #startAsDemon:.
				self onExit.
			].
			(Delay forMilliseconds: 200) wait.
		].
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 4/8/2009 13:17'!
autoSave: anIsland in: qService notifying: bar
	"Auto-save the given island."
	^ self saveIsland: anIsland in: qService notifying: bar forAutoSave: true.

! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
avatar
	"Answer the avatar of the receiver"
	<bewareOf: #avatarChanged>
	^self propertyValueAt: #avatar! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
avatar: aValue
	"Modify the receiver's avatar"
	^self propertyValueAt: #avatar put: aValue with: #avatarChanged! !

!QwaqParticipantUI methodsFor: 'menu-handling' stamp: 'bgf 11/15/2009 00:12'!
avatarMenuPosted: tMenu itemData: itemData
	| cmenu |
	"The avatar replica has been given a t-menu in the space.  Present a CMenu to represent it."

	self readyForEvents ifFalse: [ ^ self ].
	itemData isEmptyOrNil ifTrue: [ ^ self ].
	(tMenu island = mainView activeIsland) ifFalse: [ ^ self ].

	self playSoundEffect: #softClick.

	"Assemble the CMenu to represent this to the user."
	cmenu := CMenu new.
	itemData doWithIndex: [ : item : idx |
		(cmenu add: item first target: self action: #fireTMenu:item: arguments: { tMenu . idx })
			separator: (item second);
			enabled: (item third);
			checked: (item fourth)
		].
	cmenu fillAlpha: 0.9.	"Gives the menu a wee bit of 'in-the-space' feeling (from the hue.)"

	"If there's a menu task in force, re-use it for the followon  menu.
	 This can happen when a menu action raises a menu;
	 we can get the new menu before the old menu is closed.
	 We don't want to pop the task in that case, since popping includes a replica->cancelMenu. 
	 Timing overlaps can mean that would happen after the new one is supposed to have taken over."
	mainView task menuMode ifTrue: [
		mainView task menu: cmenu
	] ifFalse: [
		mainView pushTask: ((QMenuTask new) view: mainView; menu: cmenu; yourself).
	].

	"Various things can happen to take the menu away..."
	cmenu startScript: #close when: { self . #unready }.
	cmenu startScript: #close when: { self . #userAway }.
	cmenu startScript: #close when: { tMenu .  #destroyed }.
	itemData first third ifTrue: [ 
		"The first menu item is enabled, so start relative to it."
		cmenu openAt: hand position + (2@-16).
	] ifFalse: [
		"The first menu item is disabled. These are often titles.  Shift up a bit."
		cmenu openAt: hand position + (2@-42).
	].

	"Clear any lingering status floater, let the menu get attention."
	statusLine ifNotNil: [
		statusLine floatingMessage: nil.	
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/29/2007 20:35'!
avatarPane
	"Answer the avatarPane of the receiver"
	<bewareOf: #avatarPaneChanged>
	^self propertyValueAt: #avatarPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/29/2007 20:35'!
avatarPane: aValue
	"Modify the receiver's avatarPane"
	^self propertyValueAt: #avatarPane put: aValue with: #avatarPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 7/28/2008 21:13'!
avatarReplica
	^self avatar ifNotNilDo: [:a | a replica]! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'jjl 4/28/2008 11:10'!
avatarReplicaClass

	"Uncomment the following two lines to run Forums without Miramar packages.
	 Search for ***MIRAMAR IS INTEGRATED*** for the other places"
	"(self respondsTo:#alternateAvatarReplicaClass) ifTrue:[ ^self alternateAvatarReplicaClass ].
	^ QAvatarReplica"

	^MAvatarReplica! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'jjl 4/28/2008 11:10'!
avatarUserClass

	"Uncomment the following two lines to run Forums without Miramar packages.
	 Search for ***MIRAMAR IS INTEGRATED*** for the other places"
	"(self respondsTo:#alternateAvatarUserClass) ifTrue:[ ^self alternateAvatarUserClass ].
	^ QAvatarUser"

	^MiramarAvatarUser! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/1/2008 23:12'!
awayOverlay
	"Answer the awayOverlay of the receiver"
	<bewareOf: #awayOverlayChanged>
	^self propertyValueAt: #awayOverlay! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/1/2008 23:12'!
awayOverlay: aValue
	"Modify the receiver's awayOverlay"
	^self propertyValueAt: #awayOverlay put: aValue with: #awayOverlayChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/7/2007 03:37'!
backButton
	"Answer the backButton of the receiver"
	<bewareOf: #backButtonChanged>
	^self propertyValueAt: #backButton! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/7/2007 03:37'!
backButton: aValue
	"Modify the receiver's backButton"
	^self propertyValueAt: #backButton put: aValue with: #backButtonChanged! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 3/25/2010 13:33'!
beADemon: demonMode

	"Menu callback for Developers/Demon menu commands.
	 Start or switch the demon to the given actionSuite.
	 If the action suite is nil, the demon is made idle."

	myDemon 
		ifNil: [ myDemon := QDemon new participant:self ]
		ifNotNil: [
			(myDemon mode = demonMode) ifTrue: [
				^ self
			]].

	self stopScript: #demonAction:.
	myDemon isDemonic ifTrue: [
		myDemon stop.
		myDemon stop.
	].
	myDemon start: demonMode.
	self startScript: #demonAction: when: {myDemon . #action}.
	self updateTitle.
	self setAsRobot: true mode: demonMode.! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'jcg 3/30/2008 11:02'!
beginOverlay: ogl
	"Set up gl to provide overlay drawing - this is done in pixel coordinates"
	ogl glMatrixMode: GLProjection.
	ogl glPushMatrix.
	ogl glLoadIdentity.
	ogl glMatrixMode: GLModelview.
	ogl glPushMatrix.
	ogl glLoadIdentity.
	ogl clearTextureMatrix: 'not nil'.
	ogl glTranslated: -1 with: 1 with: 0.0.
	ogl glScaled: (2.0 / bounds width) with: (-2.0 / bounds height) with: 1.0.
	ogl glDisable: GLDepthTest.
	ogl pushBlend: GLOne  with: GLOneMinusSrcAlpha.

! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 5/28/2007 15:57'!
benchmarkEnd: aController
	"Notify the receiver that the given controller just finished its server benchmark."
	flowControlFlag := false.! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 5/28/2007 16:35'!
benchmarkProgress: amount
	"Notify the receiver that we are making progress with the benchmark"! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 5/28/2007 15:57'!
benchmarkStart: aController
	"Notify the receiver that the given controller just started its server benchmark."
	flowControlFlag := true.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:25'!
bringEveryoneWithMe: bBring
	"Meeting-menu command - have everyone play follow-the-leader (or not)."
	self readyForEvents ifFalse: [ ^ self ].
	bBring ifTrue: [ 
		avatar replica future setMeetingPlaceWithMe
	] ifFalse: [
		avatar replica future setMeetingPlace: nil 
	].
	QActivityCapture action: #MeetingAllFollowMe details: bBring asString.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 6/27/2009 00:26'!
broadcastResetSound
	"Broadcast a reset-sound command request to all participants in this forum;
	 the signal goes through org services and is caught by the users-panes at the other end."
	orgServices ifNotNilDo: [ :os|
		os sendSignal: #forum:resetSound:
			forForum: (self currentForumURL) 
			argument: self userLogin
	] ifNil: [
		self statusAlert: 'Cannot reset all sound.\Those services are not available right now.'.
	].	! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'crl 4/14/2010 16:48'!
broadcastTestClientExperience
	"Broadcast a run-session-experience-test command request to all participants in this forum;
	 the signal goes through org services and is caught by the users-panes at the other end."
	QLogger logInfo: 'broadcasting a client-experience-test command...'.
	orgServices ifNotNilDo: [ :os|
		os sendSignal: #forum:testClientExperience:
			forForum: (self currentForumURL) 
			argument: self userLogin
	] ifNil: [
		self statusAlert: 'Cannot broadcast a client-experience-test command.\Those services are not available right now.'.
	].	! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 9/7/2009 00:13'!
broadcastingUsers
	"Answer the broadcastingUsers of the receiver"
	<bewareOf: #broadcastingUsersChanged>
	^self propertyValueAt: #broadcastingUsers! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 9/7/2009 00:13'!
broadcastingUsers: aValue
	"Modify the receiver's broadcastingUsers"
	^self propertyValueAt: #broadcastingUsers put: aValue with: #broadcastingUsersChanged! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 2/16/2007 11:10'!
cachedPasswordForRealm: normalizedRealm
	"Answer the password for the given (normalized) realm if we have cached this information.
	If no information is available, the user will be prompted for her credentials."
	^serviceMgr password! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 2/16/2007 11:04'!
cachedUserForRealm: normalizedRealm
	"Answer the user for the given (normalized) realm if we have cached this information.
	If no information is available, the user will be prompted for her credentials."
	^serviceMgr user! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/23/2008 22:04'!
canInviteByRole
	"A switchable feature 'InviteByRole - if true,
	 the invite user dialog offers a Guest / Member / Administrator choice. "

	^ self isSealedApplication not or: [
		self hasFeatureEntitlement: #InviteByRole ]
	! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/27/2009 00:04'!
canModifyForum
	"Return true if the user can do in-forum editing,
	 adding models, dropping links, etc."
	^ (capabilities includes: #modify)
		and: [ self isMeetingLeader or: [ meetingState isEditBlocked not ] ]! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/1/2009 00:30'!
canSpeakInMeeting
	"True if there is no meeting-imposed block on chat,
	 or if I'm the leader
	 or if the leader has given me permission, TBD"
	^ meetingState isLoginAllowedToSpeak: (self userLogin)! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 3/12/2009 17:44'!
cancelLocalProvider

	"Take down the local service provider if we're using one;
	 this is done after nonlocal login so I'm not left running a router."

	login hasLocalServices ifTrue: [
		QLogger logInfo: 'Disabling local provider.'.
		QLocalServiceProvider reset.
		TConnectionDispatcher reset.
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/18/2008 21:46'!
capabilities
	"Answer the capabilities of the receiver"
	<bewareOf: #capabilitiesChanged>
	^self propertyValueAt: #capabilities! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/18/2008 21:46'!
capabilities: aValue
	"Modify the receiver's capabilities"
	^self propertyValueAt: #capabilities put: aValue with: #capabilitiesChanged! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 6/18/2008 22:22'!
capabilitiesForRole: aRoleName

	"Return the (IdentitySet) capability symbol collection
	 that the official policy defines as those available to the given role."

	^ self class capabilitiesForRole: aRoleName! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 11/9/2009 14:39'!
captureProxy
	"Answer the captureProxy of the receiver"
	<bewareOf: #captureProxyChanged>
	^self propertyValueAt: #captureProxy! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 11/9/2009 14:39'!
captureProxy: aValue
	"Modify the receiver's captureProxy"
	^self propertyValueAt: #captureProxy put: aValue with: #captureProxyChanged! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'bgf 1/9/2010 12:14'!
catalogDrop: catEntry at: aPoint 

	"Dispatcher for drag-drop of QCatalogEntry  from the object catalog."

	self floatingPanes do: [ : pane | 
		(pane visible and: [ pane respondsTo: #catalogDrop:at:]) ifTrue: [
			(self isPoint: aPoint inPlayer: pane ignoring: #() ) ifTrue: [
				^ pane catalogDrop: catEntry at: aPoint 
			]]].

	"OK, its probably for dropping into the forum."
	self readyForEvents ifFalse: [
		self statusAlert: 'Not ready to receive objects.'.
		^ false
	].
	(meetingState isEditBlocked and: [ self isMeetingLeader not]) ifTrue: [ 
		self statusAlert: 'Changes are not permitted by the meeting leader.'.
		^ false ].
	self canModifyForum ifFalse: [
		self statusAlert: 'As a visitor, you cannot add objects.'.
		^ false
	].

	(self isPointInMainView: aPoint)  ifFalse: [
		self statusAlert:  'Cannot be dropped here.\Drop only in the space.'.
		^ false
	].

	"Do the main work as followup script so the return is immediate
	 The caller uses true result to -start- the drop-in icon animation."
	catalogPane startScript: #dropObject:at: withArguments: {catEntry . aPoint}.

	^ true
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 1/5/2010 13:15'!
catalogPane
	"Answer the catalogPane of the receiver"
	<bewareOf: #catalogPaneChanged>
	^self propertyValueAt: #catalogPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 1/5/2010 13:15'!
catalogPane: aValue
	"Modify the receiver's catalogPane"
	^self propertyValueAt: #catalogPane put: aValue with: #catalogPaneChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'ar 11/24/2009 17:41'!
causeSynchronizationLoss
	"Cause synchronization loss in the active space's island"
	self activeSpace controller causeSynchronizationLoss.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/17/2009 14:18'!
changeUserPassword
	"Change the local user's password"
	| dlg resp newPass |

	QActivityCapture countAction: #ChangePassword.

	dlg := QChangePasswordDialog new.
	dlg userName value:  self userLogin.
	dlg acceptAction:[
		newPass := QPasswordHolder with: (dlg password contents asString).
		resp := serviceMgr changePassword: newPass user: self userLogin.
		resp name == #ok
			ifTrue:[dlg close.  self updatePassword: newPass forUser: self userLogin]
			ifFalse:[CDialog inform: 'Password change denied.', String cr, resp contentString]].
	dlg openCenteredIn: self.
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'sn 5/13/2011 14:55'!
changeViewpoint: aViewpointOption

	|aScalar|
	aScalar := QUserPreferences viewpointTable at: aViewpointOption ifAbsent: [1.5].

	self changeViewpointValue: aScalar.

	"Make the on-the-fly PoV menu show the right checkmark."
	(self menu itemNamed: 'View' translateMe) subMenu menuItems asArray
		from: 1 to: (QUserPreferences viewpointTable size)
		do: [ :item|
			item checked: (item label = aViewpointOption).
		].
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 7/4/2009 00:06'!
changeViewpointValue: aScalar

	"Directly change the avatar point-of-view setting."
	avatar ifNil: [ ^ self ].

	"If you assume control of viewpoint while in sticky, you stay sticky... 
	 but your viewpoint setting takes up the adjustments you make;
	 this lets you adjust viewpoint without 'falling' out of a focus.
	 Consistent with newer zoomIn/zoomOut mousewheel handling in Croquet-Objects.1045"

	avatar driver stickyViewpoint ifNotNil: [
		avatar driver stickyViewpoint: aScalar 
	] ifNil: [
		avatar thirdPerson: aScalar.
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 11/9/2006 10:36'!
chatPane
	"Answer the chatPane of the receiver"
	<bewareOf: #chatPaneChanged>
	^self propertyValueAt: #chatPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 11/9/2006 10:36'!
chatPane: aValue
	"Modify the receiver's chatPane"
	^self propertyValueAt: #chatPane put: aValue with: #chatPaneChanged! !

!QwaqParticipantUI methodsFor: 'proxy' stamp: 'ar 7/12/2011 12:53'!
checkConnectivity
	"Estabilish communication with the current forums host, via proxy if necessary.
	Return connection info on success, nil on failure."

	| info address |
	self statusInfo: 'Resolving server ', self serviceHost.

	"Try resolving the hostName first - we have seen situations in which customers
	didn't have DNS and we'd complain below about wrong proxy settings when all
	that was wrong was DNS."
	address := [NetNameResolver addressForName: self serviceHost] 
		on: NameLookupFailure do:[:ex| ex return: nil].

	"Only remember DNS failure but don't actually report it yet.
	That's because in some proxied environments DNS is disabled but
	the proxy server will establish the connection just fine. Unfortunately,
	this makes it impossible to fail quickly here since we need to go through
	the full auto-detect cycle."

	QProxyHandler log: '------------------- Proxy Detect Start ---------------------' from: self class.
	QProxyHandler log: QwaqVersion current infoString from: self class.
	QProxyHandler log: 'Proxy setting: ', proxyDlg actualProxyType from: self class.
	info := self checkForConnectivityTo: self serviceHost.
	QProxyHandler log: '------------------- Proxy Detect End ---------------------' from: self class.
	info ifNil:[
		address ifNil:[
			"Don't give people the proxy settings if we have a DNS failure."
			self inform: 'Unable to resolve the server named ', self serviceHost, 
				'\Please check your network configuration' withCRs.
			^nil
		].
		self inform: 'Unable to connect to ', self serviceHost,
				'\Please check your proxy configuration.' withCRs.
		self editProxyConfiguration.
		^ nil 	"Not reachable"
	].
	^ info 	"Continue.. connectivity OK"
! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'sn 3/10/2013 15:26'!
checkDuplicateLogin: svcRecord urlString: serviceUrl serviceManager: svcMgr
	"If this login is already present in that forum, then that login will be kicked out
	 if we login from here... warn about this.  Raises a progress-cancel if necessary."

	orgServices ifNotNil: [ 
		"First check that we'd be able to get in at all with this version."
		(orgServices isForum: svcRecord url okForVersion: (QwaqVersion current infoString))
		ifFalse: [
			^ nil		"We don't warn about duplicate login, because the version-mismatch intercept			
					 is going to fire and not let us log in,  so just let that do its thing.
					 (If we ask the user about duplicates, and they say go ahead, 
					  and then we immediately  say 'aha, but the version is wrong' ... that's annoying."
		]].
	"... And the same check using the service record, if activeVersion is provided there.."
	svcRecord activeVersion ifNotNilDo: [ : ver | 
		ver = (QwaqVersion current infoString) ifFalse: [  ^ nil ].
		].

	((orgServices notNil and: [ orgServices isLogin: self userLogin inForum: serviceUrl] )	
	or: [ svcMgr service: svcRecord hasLiveUser: (svcMgr user) ])
	ifTrue:[
		| msg |
		QActivityCapture action: #LoginForumAlreadyPresent details: serviceUrl.

		"If we are demon or client proxy or anything else that is just going to exit on 
		failure, then there is no point in asking. Just continue without cancelling progress,
		so the old session gets killed."
		(self handlesEvent: #serviceLoginFailed) ifTrue: [^self].

		msg := '    This account ({1}) is already logged into {2}.

    If you proceed into the Forum,
    the other session may be disconnected without warning.' format: {
			svcMgr user storeString.
			svcRecord name storeString.
		}.
		(self confirm: msg) ifFalse:[
			(CProgressCancel new) 
				messageText: 'Forum entry canceled.'translateMe;
				signal.
		].
	].! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 6/23/2008 19:28'!
checkEntryEntitlement: svcRecord

	"We are about to activate a forum from a URL/link that the user has encountered. 

	 The context is such that error: is the delivery vehicle for rejection;
	 this is called on the activateService: path, which is handler-wrapped above us."

	svcRecord isMiramarDesktopService ifTrue: [
		self hasMiramarDesktopEdition ifTrue: [
			self statusInfo: 'Entering a Desktop Edition Workspace -- Beta'.
		] ifFalse: [
			self error: (
				'This forum is a Desktop Edition workspace.',
				'\It cannot be entered from this account.')
				withCRs.
		]].! !

!QwaqParticipantUI methodsFor: 'proxy' stamp: 'ar 7/12/2011 12:53'!
checkForConnectivityTo: hostName
	"Check for connectivity, determining the proxy server if needed.
	If no connectivity, open proxy dialog for the user to figure it out"

	| info proxyMethods url proxyString proxyList proxyName proxyUrl |
	self statusAlert: nil.
	self statusInfo: 'Determining proxy settings... '.

	"Make sure the defaults are set properly"
	QProxyHandler defaultProxy: proxyDlg proxyServerString.
	QProxyHandler pacDefaultUrl: proxyDlg proxyConfigUrl contents asString.
	proxyMethods := QProxyHandler proxyMethodsFor: proxyDlg actualProxyType.

	"We need an actual url below so create one"
	url := 'http://', hostName,'/'.
	proxyMethods do:[:aSymbol|
		"Get the list of proxies (separated by commas)"
		proxyString := [QProxyHandler perform: aSymbol with: url] on: Error do:[:ex| 
			QProxyHandler log: aSymbol, ' - ', ex from: self class.
			ex return: ''].
		"Parse the proxy string"
		proxyList := (proxyString findTokens:';') collect:[:each| 
			proxyName := each withBlanksTrimmed.
			"Windows Internet allows proxies of the form
				http=http://host.domain:8080/
			for various protocols. Try to decode them properly."
			[(proxyName includes: $=) ifTrue:[
				proxyUrl := (proxyName copyAfter: $=) asUrl.
				proxyName := proxyUrl authority.
				proxyUrl port ifNotNil:[proxyName := proxyName,':', proxyUrl port asString].
			]] on: Error do:[:ex| 
				QProxyHandler log: 'Error parsing proxy: ', each from: nil.
				proxyName := each withBlanksTrimmed. "make sure we got something"
			].
			proxyName].

		proxyList do:[:each|
			info := [self tryProxyServer: each forHost: hostName] on: Error do:[:ex|
				QProxyHandler log: each, ' - ', ex from: self class.
				QProxyHandler logRequest: (ex signalerContext stackOfSize: 100).
				ex return: nil].
			info ifNotNil:[
				QProxyHandler lastUsedProxy: each. "remember the last used proxy"
				^info].
		].
	].
	^nil
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 5/28/2010 15:09'!
checkForValidEnvironment
	"Check to see if we're running in a valid environment"

	"Check for writability of logs, cache etc. directory"
	{FileDirectory cacheDirectory. FileDirectory logDirectory} do:[:fd|
		[	| file |
			file := fd forceNewFileNamed: 'file-access-test'.
			file nextPutAll: 'Hello'.
			file close.
			fd deleteFileNamed: 'file-access-test' ifAbsent:[self error: 'Test file does not exist'].
		] on: Error do:[:ex|
			QLogger logError: ex  from: self.
			^self inform: 'Error accessing ', fd pathName,
				'\Please check permissions and restart.' withCRs.
		].
	].

	"Further environment tests should go here...."
	[ self initializePyScript ] on: Error do:[:ex|
		QLogger logError: ex  from: self.
		self inform: 'Python initialization failed:\' withCRs, ex description.
	].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 4/16/2011 18:59'!
checkForValidFullEnvironment
	"Check to see if we're running in a valid Full environment (graphics, python, etc.).
	Answer true if we're good to go, otherwise false."
	| ogl err |

	"Check to see if we are running on intel macs of the right version"
	((Smalltalk platformName = 'Mac OS')  and:[self isSealedApplication]) ifTrue:[
		Smalltalk platformSubtype = 'intel' ifFalse: [
			self inform: 'OpenQwaq is only supported on Intel processors.'
				helpUrl: '/support/hardware.php'.
			^false.
		].
		Smalltalk osVersion >= '1050' ifFalse: [
			self inform: 'OpenQwaq does not officially support Mac OSX below version 10.5.\You may proceed, but you may experience issues with sound or other features.' withCRs
				helpUrl: '/support/hardware.php'.
			"but don't force them away by answering false"
		].
	].

	doOpenGL ifTrue: [
		"Verify OpenGL existence.
		Note: The use of globalBounds is meant to ensure we also test for low-memory 
		conditions (if the window we're running in is huge for example)."
		["Does not use 'OpenGL newIn: to be able to handle errors after context creation"
		ogl := OpenGL new.
		ogl initializeIn: self globalBounds] on: Error do:[:ex| 
			ogl ifNotNil:[ogl destroy. ogl := nil].
			err := ex description. 
			ex return: nil].
		ogl ifNil:[
			self inform: ('Your computer''s OpenGL support appears to be insufficient.\',  
			err printString,'\\OpenGL is a 3D graphics capability that your computer must have in order to use OpenQwaq.\\You should update your graphics driver and try OpenQwaq again. In some cases, Microsoft\Windows Update improves DirectX performance at the expense of OpenGL. Getting the latest\graphics driver from your card manufacturer ensures the best support for all applications.\\The "More" button below will open your browser to a Web page with possible solutions.') withCRs
				helpUrl: '/docs/userguide/', (QwaqParticipantUI versionedUserGuide: 'graphics_card').
			^false
		].

		"Verify required OpenGL extensions"
		#(
			'GL_ARB_transpose_matrix'
		) do:[:ext|
			err := nil.
			[ogl loadExtension: ext asSymbol] on: Error do:[:ex| 
				ogl destroy. ogl := nil.  "don't let it interfere with the dialog"
				err := ex description. ex return].
			err ifNotNil:[
				self inform: ('Your computer does not support a required OpenGL feature: ',  ext, '\\OpenGL is a 3D graphics capability that your computer must have in order to use OpenQwaq.\The "More" button below will open your browser to a Web page with possible solutions.') withCRs
					helpUrl:  '/docs/userguide/', (QwaqParticipantUI versionedUserGuide: 'graphics_card').
				^false
			].
		].

		"Done with OpenGL tests"
		ogl destroy.
		Display forceToScreen. "clear any left-over bits"
	].
	^true! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 6/15/2010 11:22'!
checkHeartbeatLag

	"Check how far behind the avatar is in receiving acknowledgement of its heartbeats.
	 The path is 
		Avatar - driveStep -> Replica  -signal-acknowledgement-> Avatar 
	 The avatar's heartbeatLag measures how long that roundtrip is taking.
       So if my connection to t-time is delayed, we see that here.

	 This is the clients-own-avatar's local equivalent of TAvatarReplica selfDestruct.
	 We can't rely on the in-space replica to detect timeouts in our own avatar,
	 because it requires that t-space time be advancing, 
	 which doesn't happen when the network is miscreant."

	| totalSecondsBehind |

	self inBenchTest ifTrue: [ ^ self ].

	"This is one case where we don't use #readyForEvents,
	 because we want these guards to apply when we're in an error dialog.
	 But if we've lost the replica altogether, there's no need, as we're either leaving or reconnecting."
	mainView ifNil: [ ^ self ].
	avatar ifNil: [ ^ self ].
	avatar isReplicaConnected ifFalse: [ ^ self ].

	totalSecondsBehind := avatar heartbeatLag.
	(totalSecondsBehind > (5*60)) ifTrue: [
		^ self logoffConfirming: false.		"Don't linger in this sort of limbo forever."
	].

	"Below are separate cases for different degrees of delay."

	"Worst case - past the visible timeout, at which the replica hides itself 
	 - lag is too much to pretend we're participating."
	(totalSecondsBehind > (TAvatarReplica visibleTimeout // 1000))  ifTrue: [	
		session errorTripped ifFalse: [
			"(Don't need these notices if we've fallen into error-dialog limbo.)"
			self audioSession isRecording ifTrue: [
				statusLine alert: 'High network delay. Voice chat suspended.'.
			] ifFalse: [
				statusLine info: 'High network delay.  Voice chat suspended.'.
			].
			usersPane transientNotice: 'Delay' forSession: (self sessionID) forMillis: 500.
			self isAudioSuspended ifFalse: [
				"I.e. only record the incident when it arises, not for every tick."
				QActivityCapture countAction: #DelaySelfAudioSuspended.
			].
		].
		self suspendVideoBecauseOfLag.		"We're far beyond the threshold for this."
		self suspendAudio.	"Suppress audio, even (especially) when in error."
		^ self
	].

	session errorTripped ifTrue: [ ^ self ].

	"Fallthrough - we may have delay, but at a lower level than the visibleTimeout."

	self resumeAudio.	"In case delayed earlier by the visibleTimeout-case (we're recovering)"
						"Note that we don't auto-recover video."

	(totalSecondsBehind > (TAvatarReplica warningTimeout // 1000)) ifTrue: [
		"Lag justifies letting the user know something is wrong."
		usersPane isSelfDelayed ifFalse: [
			QActivityCapture countAction: #DelaySelfWarning
		].
		usersPane transientNotice: 'Delay' forSession: (self sessionID) forMillis: 500.
		statusLine info: 'Activity is delayed. There might be a network problem.'.
	].
		
	(totalSecondsBehind > (TAvatarReplica dropOptionalLoadTimeout // 1000)) ifTrue: [
		"The level of lag justifies taking some remedial steps."
		self suspendVideoBecauseOfLag ifTrue: [
			statusLine info: 'There is high delay. Disabling video streaming.'.
			QActivityCapture countAction: #DelaySelfVideoSuspended
		].	
	].



	! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'ar 8/18/2009 18:41'!
checkLinkHost: svcUrl

	"Given a service URL , verify that its host name matches the service manager in force;
	 raise an error if the svcUrl is not in the same host as we are connected to."

	"ar 8/18/2009: It is not clear to me if we should have this test at all. Since the server can 
	respond with a redirect to the host in question,  the information about the host is really 
	derived from the organization we're trying to join."

	| host |
	serviceMgr ifNil: [ ^ self ].
	host := QServiceLink hostNameFromServiceUrl: svcUrl.
	(host = serviceMgr originalHost or:[host = serviceMgr host]) ifFalse: [
		self error: ('Cannot open links to other hosts.\Currently connected to ', (serviceMgr host), '.\The link leads to ', host, '.') withCRs.
	].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'hrs 12/8/2008 16:45'!
checkMicrophoneKeyEvent: kbdEvt onOff: yesNo

	"If the given key up/down event is the control key, 
	 handle press-to-talk press-to-talk."

	| ptt |

	(Smalltalk platformName = 'Mac OS') ifTrue: [ 
		"Mac sends meta-key changes as mouse move."
		^self
	].
	self audioSession ifNil: [ ^ self ].
	self audioSession isPressToTalk ifTrue: [
		"In press to talk, talk as long as control is down."
		ptt := false.
		kbdEvt controlKeyPressed ifTrue: [
			ptt := true.
		].
		controlBar microphone pressToTalkPressed: ptt.	
		self audioSession pressTalkDown: ptt.
	].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'hrs 12/8/2008 16:45'!
checkMicrophoneMouseEvent: mouseEvent
	"Boolean - 
	 If the given event reflects a control-button transition, do microphone control.
	 The Mac uses mouse-over events to communicate changes in ctl-key state."

	| pressed |

	self audioSession ifNil: [ ^self ]. 
	(Smalltalk platformName = 'Mac OS') ifFalse: [ ^ self ].

	self audioSession isPressToTalk ifTrue: [
		
		"In pressToTalk, the control-key is hotkey for the microphone."
		pressed := mouseEvent controlKeyPressed.
		pressed ifTrue: [ self controlBar awayButton selected: false ].
		controlBar microphone pressToTalkPressed: pressed.
		self audioSession pressTalkDown: pressed.
	].

! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 9/6/2009 02:03'!
checkReturnLinkFor: originGate linkedTo: postcard url: url atService: aService

	"The user has activated the given QGate link,
 	 and the forum to which it points is logged in as 'aService'.
	 This is our opportunity to ensure two-way integrity of gate-gate links.

	 Returns true if the linkage seems complete, false otherwise."

	|  abspostcard target space returnLink gate |

	"Resolve the target frame (may be any TFrame in the destination space).
	 We need an absolute form of the original url to do this."

	abspostcard := TPostcard urlString: url.
	target := self resolvePostcard: abspostcard within: 5000 info: 'Connecting...'.
	target ifNil: [
		| entryUrl |
		"We cannot resolve the specific destination frame.
	 	 Assume there's been save-mismatch mischief, and start with a link to the default entry,
	 	 so there is something to show in the gateway."
		entryUrl := (TPostcard makeUrl: aService entryUrl relativeTo: aService url).
		target := self resolvePostcard: (TPostcard urlString: entryUrl) within: 2000 info: 'Connecting...'.
		target ifNotNil: [
			self statusAlert: 'Cannot find destination. Repairing gate link.'.
		].
	].  
	(target isNil or: [ (space := target get:#root) isNil]) ifTrue: [
		self statusAlert: 'The destination has no entry portal.'.
		^ false
	].

	"Now that we have the destination TSpace, let the history pane know,
	 so it can manage the connection as part of the service cache."
	historyPane noteSpaceActivate: space.

	((target get:#class) includesBehavior: TPortal) ifTrue: [
		^ true	"Already linked out to a TPortal."
	].
	(((target get:#parent) get:#class) includesBehavior: QGate) ifTrue:[
		^ true	"Already linked out to a gate."
	].

	"Now the sport... two-way link setup across the gates. ".
	self statusInfo: 'Finishing connection...'; waitTick.
	returnLink := TPostcard urlString: 
					(session absUrlForPostcard: (originGate get:#postcard)
							inService: serviceSvc).

	"Let the destination space find a gate."
	gate := space future linkInboundPortal: returnLink.
	(gate waitTimeoutSeconds:15) ifTrue: [
		gate := gate value.
	] ifFalse: [
		gate := nil
	].
	gate ifNotNil: [
		originGate future 
			updatePostcard: (TPostcard urlString: 
								(session absUrlForPostcard:(gate get:#postcard)
					    					inService: aService))
			originalUrl: postcard urlString.
		^ true
	] ifNil: [
		^ false
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 8/6/2009 00:39'!
clearDocumentManager
	session ifNotNil: [ session documentManager: nil ].
	(self propertyValueAt: #documentManager ifAbsent: [ nil ]) ifNotNilDo: [ :dm |
		dm destroy.
	].
	self removePropertyNamed: #documentManager.
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 3/4/2009 13:51'!
clearReturnQwqFile
	"Delete any auto-return file, as we won't be using it anymore this session."
	FileDirectory deleteFilePath: self returnQwqPath.! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/13/2010 13:58'!
clientAuth: webClient realm: authRealm fail: failBlock
	"Ask the user for authentication credentials.

	authRealm is expected to be the user visible description for the authentication request.
	For Sharepoint (and other repositories) this should be the repository URL, NOT the realm
	used by WebClient (which is not available for NTLM and friends)."

	| dialog pw |
	dialog := QUserLoginDialog new.
	dialog title value: 'Authentication Required '.
	dialog explanation value:
		'Enter username and password to access', String cr,  authRealm.
	dialog userName value: (webClient username ifNil:[login user value]).
	dialog password value: (webClient  password value ifNil:['']).
	dialog savePassword visible: false.
	dialog acceptAction:[pw := (QPasswordHolder with: dialog password value).  dialog close].
	dialog cancelAction:[pw := nil. dialog close].
	dialog openCenteredAt: self dialogCenter.
	dialog waitUntil: #closed.
	pw ifNil:[^failBlock value].
	webClient username: dialog userName value.
	webClient password: pw.
! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 5/10/2010 12:47'!
clientCheckin: serverPath repository: repoUrl path: repoPath type: repoType

	"Performs a client-side checkin of a given document.
	May require user credentials to be supplied for authentication.

	Returns an error string if anything goes wrong or nil for success."

	| repository entries data resp |
	"Assume sender added repository to cache"
	repository := QDocumentRepository cache at: repoUrl.

	"Get the document from the server"
	entries := serviceMgr listFiles: serverPath.
	entries ifEmpty:[^'The document has already been checked in'].

	"Scale progress to accomodate both download and upload.
	It would be better if we could interleave this in the same way
	we do in checkIn but currently the repository operations do 
	not support this out of the box."

	CProgressBar displayProgress: 'Checking in ...' 
		at: self dialogCenter from: 0.0 to: 2.0  during: [:bar|
			data := serviceMgr get: serverPath informing: bar.
			[[resp := repository upload: repoPath stream: data readStream do: nil] 
				on: HTTPProgress do:[:ex| bar value: 1.0 + (ex amount / ex total asFloat). ex resume].
			] ensure:[repository close].
	].
	resp isSuccess ifFalse:[^'Could not check in document.\\' withCRs, resp status].

	^nil "meaning: no error, i.e., success"! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 7/27/2010 17:31'!
clientCheckout: serverPath repository: repoUrl path: repoPath type: repoType

	"Performs a client-side checkout of a given document.
	May require user credentials to be supplied for authentication.

	Returns an error string if anything goes wrong or nil for success."

	| repository resp amount count debug |
	debug := true. "for extra logging"

	repository := (QDocumentRepository cache)
		at: repoUrl 
		ifAbsentPut:[QDocumentRepository type: repoType url: repoUrl owner: self].
	repository ifNil:[^'Could not check out document.\\Unknown repository: ', repoType].


	"Get the document via repository stream"
	debug ifTrue:[QLogger logInfo: 'Downloading ', repoPath].
	[resp := repository download: repoPath.
	resp isSuccess ifFalse:[^'Unable to download document:\' withCRs, resp status].

	"Guard against zero content-length reply observed at LMCO"
	resp contentLength = 0 ifTrue:[
		QLogger logInfo: 'Sharepoint response: ', String cr, resp asString.
		^'Unable to download document:\Server responded with empty document'.
	].
	debug ifTrue:[QLogger logInfo: 'Sharepoint response: ', String cr, resp asString].

	"upload the document to the server"
	amount := 0.
	resp contentLength ifNil:[
		debug ifTrue:[
			QLogger logInfo: 'No content-length header'.
			QLogger logInfo: 'Starting download'.
			resp contentStream isConnected
				ifFalse:[QLogger logWarning: 'Socket lost connection'].
		].
		CProgressBar displayProgress: 'Checking out document ...' 
			at: self dialogCenter from: 0 to: 1.0 during:[
				resp := serviceMgr putStreaming:[ | buf |
					buf := resp contentStream next: 4096.
					amount :=  amount + buf size.
					debug ifTrue:[QLogger logInfo: 'Received ', buf size, ' bytes'].
					buf
				] name: serverPath backup: false.
		].
	] ifNotNil:[:max|
		debug ifTrue:[
			QLogger logInfo: 'Content-length is ', max.
			resp contentStream isConnected 
				ifFalse:[QLogger logWarning: 'Socket lost connection'].
		].
		CProgressBar displayProgress: 'Checking out document ...' 
			at: self dialogCenter from: 0 to: max
			during: [:bar|
				resp := serviceMgr putStreaming:[ | buf |
					bar value: amount.
					count := max - amount min: 4096.
					buf := resp contentStream next: count..
					amount := amount + buf size.
					debug ifTrue:[QLogger logInfo: 'Received ', buf size, ' bytes'].
					buf
				] name: serverPath backup: false.
		].
		amount = max ifFalse:[
			"If we failed to copy the right amount, signal a problem"
			serviceMgr deleteFiles: serverPath forReal: true.
			^('Checkout failed.\\Copied ', amount,' / ', max,' bytes') withCRs, resp contentString.
		].
	].
	resp name == #ok ifFalse:[^'Checkout failed.\\' withCRs, resp contentString].
	] ensure:[repository close].

	QLogger logInfo: 'Checked out ', (serverPath copyAfterLast: $/), ': ', amount, ' bytes'.

	^nil "meaning: no error, i.e., success"! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 5/17/2010 13:26'!
clientSpeakersWithStreamID: streamID
	"Answer a collection containing all QClientSpeakers that are subscribed to the specified stream-id."
	self audioSession ifNotNilDo: [:sess |
		sess speakerManager ifNotNilDo: [:mgr | 
			| promise |
			promise := mgr future clientSpeakersWithStreamID: streamID.
			(promise waitTimeoutSeconds: 1)
				ifTrue: [^promise result]
				ifFalse: [^#()].
		]
	].
	^#() "coudn't find a speaker-manager to ask"! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 4/13/2010 15:41'!
closeApps: appList

	"Close all the embeddable apps in appList.
	Return true if things went successful, false if something went wrong 
	and the save / quit / logout should be canceled."

	| pb time |
	appList ifEmpty:[^true].
	time := [
		pb := CProgressBar new.
		pb displayProgress: 'Saving...' at: self dialogCenter from: 1 to: appList size during:[:bar|
			1 to: appList size do:[:i| | embeddable |
				bar value: i.
				embeddable := appList at: i.
				pb label: embeddable saveAndCloseString.
				self waitTick. "refresh label"
				(embeddable closeApp) ifFalse:[^false].
			].
		].
	] timeToRun.
	time < 2000 ifTrue:[(Delay forMilliseconds: 2000 - time) wait].
	^true
! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 4/13/2010 15:38'!
closeAppsInAllIslands

	"Close all the apps in all islands that I am the principal for.
	Return true if things went successful, false if something went wrong 
	and the save / quit / logout should be canceled."

	^self closeApps:
		(session embeddedApps asArray select:[:embeddable|
			embeddable principal == session sessionID]).
! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 4/13/2010 15:38'!
closeAppsInIsland: anIsland

	"Close all the apps in the given island that I am the principal for.
	Return true if things went successful, false if something went wrong 
	and the save / quit / logout should be canceled."

	^self closeApps:
		(session embeddedApps asArray select:[:embeddable|
			embeddable app island == anIsland 
				and:[embeddable principal == session sessionID]]).
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 11/24/2009 13:41'!
commitPreferencesInBackground
	"Can be called from any thread context.
	 Issues a self-signal to trigger a background commit of preferences to the server."
	self signal: #commitPreferences.
	! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 10/22/2008 21:31'!
commonObjectsDirectory
	"The directory for objects common to all applications."
	^FileDirectory default directoryNamed: (FileDirectory default pathFromURI: 'objects-menu/Common').! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'ar 4/16/2011 17:29'!
concludeForumCopy: newName serviceRecord: svc
	| ans |
	ans := CDialog	
			confirm: ('\Forum copied as\	', newName, 
					'\\You are still in the original, not the copy.\') withCRs
			yes: 'Go To Copy'
			no: 'Stay Here'
			label: 'Copy Complete'
			centerAt: (self dialogCenter).
	ans ifTrue: [
		self teleportToUrl: svc url
	].
				! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 2/21/2008 17:46'!
confirmExit

	^ self confirmQuit: false! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'sn 5/13/2011 22:21'!
confirmQuit: logout
	"Handle normal logout and exit requests...
	 check unsaved states and warn if we're past due,
	 offering to save-all in that case.
	 The logout arg is just to have the labeling suit the action."
	| msg time unsaved mins needSave dlog result shortLabel stayingOpen |

	needSave := false.
	stayingOpen := (logout ~= false).

	"If autosave is enabled, then there's already a save underway since we're quitting.
	 If we're errored-out or otherwise disconnected, don't offer the option of saving."
	(preferences doAutoSave not 
	    and: [self readyForEvents]
	    and: [self canModifyForum]) 
	 ifTrue: [
		time := Time utcSeconds.
		unsaved := servicesPane entriesUnsaved.
		unsaved size > 0 ifTrue: [
			needSave := true.
			msg := (self longQuitDescription: logout),  '? '.
			msg := msg, String cr, String cr, 'There may be unsaved changes:', String cr, String cr.
			unsaved do: [ :e|
				mins := ((time - e lastSaveTime) // 60) max: 1.
				mins < 60 ifTrue: [
					msg := msg, 
						String tab, e name, ' (last saved ', mins, ' minute', 
									((mins == 1) ifTrue: [' ago)'] ifFalse: ['s ago)']),
						String cr.
				] ifFalse: [
					msg := msg, String tab, e name, ' (not saved for over an hour)', String cr.
				].
			].
			msg := msg, String cr,
					'Auto-save is disabled in your preferences.', String cr.
		].
	].

	msg ifNil: [
		"Nothing needs saving"
		self useQuickTimeouts.
		^ self finalQuitConfirm: logout afterSaveWasChecked: false.
	].

	dlog := CUserDialog new.
	dlog label: 'Confirm' translateMe.
	dlog message: msg.
	dlog messageText allowEditing: false.
	dlog messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	shortLabel :=  (self shortQuitLabel: logout).
	needSave ifTrue: [
		(dlog addButton: 'Save and' translateMe,' ', shortLabel value: #save)
			width: (stayingOpen ifTrue: [130] ifFalse: [112]).
		(dlog addButton:  shortLabel, ', Don''t Save' translateMe value: #exit)
			width: (stayingOpen ifTrue: [142] ifFalse: [122]).
	] ifFalse: [
		(dlog addButton: shortLabel value: #exit)
	].
	dlog addButton: 'Cancel' translateMe value: #cancel.

	okToQuit := true.	"Set to false if an error occurs in save."

	result := (dlog run: #save centerAt: (self dialogCenter)) 
		caseOf: {
			[#save] -> [ self saveAllForums.  okToQuit ].
			[#exit ] -> [ true ]
		} otherwise: [ false ].

	^ result ifTrue: [ 
				self useQuickTimeouts.
				self finalQuitConfirm: logout afterSaveWasChecked: true ] 
			 ifFalse: [ 
				self useNormalTimeouts.
				false ].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 7/23/2009 18:18'!
connectAgainToLastNetVidID
	"Just for development:  connect again to the last NetVid stream that we launched."
	| streamID document card |
	streamID := QJoshHacks lastNetVidID.
	streamID ifNil: [^self inform: 'No NetVid stream-ID available; launch one first.'].

	document := QExistingStreamingVideoDocument new.
	document streamID: streamID.
	card := QFileCard forDocument: document.
	card isTransient: true.
	card := TIslandCopier new import:(TIslandCopier new export:card)
									  to: session activeIsland.
	session dropCard: card event: nil.
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 8/31/2010 11:37'!
connectionError: error from: aController
	
	"Notify the participant about a connection error in the controller.
	We only signal an event here as the method is called from a different thread."

	self signal: #reconnect! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'bgf 8/21/2008 15:26'!
connectionInfoFor: host
	"Answer the connection info for the given host."
	| info address |
	info := hostCache at: host asLowercase ifAbsent:[nil].
	info ifNotNil:[
		"If the cached info is young enough, just use it."
		((Time totalSeconds) - info timestamp) < (15*60) ifTrue: [
			^ info
		].

		statusLine info: 'Verifying connection to ', host,' ...'.
		(self authenticateConnection: info) ifTrue:[
			info timestamp: (Time totalSeconds).
			^info
		] ifFalse: [ 
			"Something's probably changed in the network; forget the old info."
			hostCache removeKey: host asLowercase ifAbsent: [].
		].
	].

	info := TConnectionInfo new.
	info hostName: host.
	info proxy: proxy.

	"Look up the address for the host"
	address := [NetNameResolver addressForName: host] 
		on: NameLookupFailure do:[:ex| ex return: nil].
	info address: address.

	"In proxied environments it's okay to have no address but unless we have proxy it's not"
	(proxy == nil and:[address == nil]) 
		ifTrue:[^self error: 'Unable to resolve ', host printString].

	"Try connecting via various ports"
	(self connectionPortsFor: host) do:[:pNum|
		proxy 
			ifNil:[statusLine info: 'Trying ', host,':', pNum,' ...']
			ifNotNil:[statusLine info: 'Trying ', host,':', pNum,' (via ', proxy proxyString, ') ...'].
		info port: pNum.
		(self authenticateConnection: info) ifTrue:[
			"This port/proxy combination works; remember it"
			QLogger logInfo: 'Ping: ', host, ':', pNum, ' -- Succeeded.'. 
			hostCache at: host asLowercase put: info.
			info timestamp: (Time totalSeconds).
			^info.
		].
	].

	"We have found no connection to the host in question; raise an error and give up"
	^self error: 'No connection to ', host! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 8/5/2010 14:20'!
connectionPortsFor: host
	"Answer the ports that we should be using for the given host.
	We'll look at the file ports.txt (but maybe we also should look at the host)"
	| file data portList |

	QServiceProvider fixedPortList ifNotEmpty:[:list| ^list].

	file := [FileStream readOnlyFileNamed: 'ports.txt'] 
			on: Error do:[:ex| ex return: nil].
	file ifNil:[^self error: 'No ports.txt file'].
	[data := file contents] ensure:[file close].
	portList := (data findTokens: Character separators) collect:[:each|
		[Integer readFrom: each readStream] on: Error do:[:ex| ex return: 0].
	] thenSelect:[:p| p > 0].
	^portList
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/17/2006 11:02'!
controlBar
	"Answer the controlBar of the receiver"
	<bewareOf: #controlBarChanged>
	^self propertyValueAt: #controlBar! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/17/2006 11:02'!
controlBar: aValue
	"Modify the receiver's controlBar"
	^self propertyValueAt: #controlBar put: aValue with: #controlBarChanged! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 2/14/2008 09:58'!
controlStreamLost: aController
	"Notify the participant that we lost the control stream in a streaming controller.
	There are two possible causes:
		a) A connection failure (in which case the hub will have reported a connection error)
		b) We got kicked off the router (most likely due to a login by the same user)
	We only signal an event here as the method is called from a different thread."
	self signal: #controlStreamLost with: aController.
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 3/14/2009 00:33'!
convertToLazyTextures: anIsland
	"Convert the existing TFileTextures in anIsland to be lazy (external)"
	| txtrList fName data lastRef promise |
	promise := anIsland future resourcesAt: TFileTexture resourceType.

	"In case we're disconnected while quit-saving; this step is legacy, mostly optional,
	 so just don't bother if we can't do it quickly.  The body of save will catch the real errors."
	(promise waitTimeoutSeconds: 3) ifFalse: [ ^ nil ].
	promise value ifNil: [ ^ nil ].

	txtrList := promise value reject: [:txtrRef|
		| fn | 
		fn := txtrRef get:#fileName.
		(txtrRef get:#isBundled)   "Don't upload prepackaged textures." 
			or:[fn beginsWith: 'sharedDocuments/']	"or ones already up (old style)"
			or:[fn beginsWith: 'resources/']			"or ones already up"
			or:[fn beginsWith: '/users/']].		"or ones in per-user storage"
	promise discardRef.

	txtrList do:[:txtrRef| |sha|
		fName := txtrRef get: #fileName.
		sha := txtrRef get: #sha.
 		(fName indexOf: $/) > 0 ifTrue:[fName := fName copyAfterLast: $/].
		data := session cacheManager find: sha.
		data ifNotNil:[
			|nm|
			nm := session documentManager 
					uploadResource: data shortName: fName sha: sha.
			nm ifNotNil:[ lastRef := (txtrRef future fileName: nm )].
		]].
	lastRef ifNotNil: [ lastRef waitTimeoutSeconds:15 ].	
		"Ensure all updates are done before the scene-save."
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 3/14/2009 00:32'!
convertToLazyTexturesSpecial: anIsland
	"Upload (to latest resource storage practice) all content resources in the island.
	 If substituted for convertToLazyTextures, 
	 this can force optimization of the resource-references in a forum,
 	 so all referenced resources use ./resource/ paths based on the resource SHA.
	 This method is uncalled... but LEAVE IT HERE.  
	 It's useful for converting forums to production-ready templates. "

	| txtrList fName data lastRef |
	txtrList := anIsland future contentResources.
	txtrList wait.

	txtrList := txtrList value reject: [:txtrRef|
		| fn | 
		fn := txtrRef get:#fileName.
		(txtrRef get:#isBundled)   "Don't upload prepackaged textures." 
			or:[fn beginsWith: '/users/']].		"or ones in per-user storage"

	txtrList do:[:txtrRef| |sha|
		fName := txtrRef get:#fileName.
		sha := txtrRef get: #sha.
 		(fName indexOf: $/) > 0 ifTrue:[fName := fName copyAfterLast: $/].
		data := session cacheManager find: sha.
		data ifNotNil:[
			|nm|
			nm := session documentManager 
					uploadResource: data shortName: fName sha: sha.
			nm ifNotNil:[ lastRef := (txtrRef future fileName: nm )].
		].
	].

	"We want the new data in the textures before we walk the scene to save it."
	lastRef ifNotNil: [ lastRef wait ].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 4/14/2010 17:54'!
copyQRLForCurrentPlace
	"Copy a QRL (go) page for the current form/place into the operating system clipboard."
	| qrl place base |
	place := placesPane currentStablePlaceName ifNil: ['Home'].
	base := serviceMgr qrlFromForum: (self currentForumName).
	qrl := QWebServices qrl: base go: place.
	QActivityCapture action: #CopyQRL details: qrl.
	Clipboard clipboardText: qrl.
	self inform: 'The URL 

	', qrl, '

has been copied to the ', CPlatform current commonParlancePlatformName, ' clipboard.'.! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 9/4/2008 11:26'!
copyResources: oldService to: newService informing: bar

	"Copy the shared documents and resources directories from the template to the new forum.
	 Transcripts any errors (does not fail)"

	| response dirs |

	dirs :=  {'/sharedDocuments/' . '/resources/' }.

	dirs doWithIndex: [ :subdir :idx |
		response :=
			serviceMgr copyFiles: ('services/', oldService name, subdir)
						         to: ('services/', newService name, subdir).
		response name == #ok 
			ifFalse: [ QLogger logWarning: 'Cannot copy shared files to ', 
					   newService name, ': ', response contentString ].

		bar value: ( idx / (dirs size)).
	].

	[
		"If there's a preview, grab that too.  But don't sweat errors."
		serviceMgr copyFiles: 'services/', oldService name, '/preview.jpg'
						to: 'services/', newService name, '/preview.jpg'
	] on: Error do: [ :ex |
		QLogger logWarning: 'Could not copy template preview.jpg: ', (ex description).
	].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 11/7/2009 20:29'!
cornerMessageFromAvatar: aString

	"Private handler; see updateStatusRelay....
	 this relays in-space corner-message signals to the status line."

	mainView ifNotNil: [
		statusLine floatingMessage: aString at: ((viewPanel topRight) * (0.9 @ 1)) rounded.
	].! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'ar 4/16/2011 17:29'!
crashDumpPathOrNil
	"Answer the full path of the crash dump file if it exists, or nil if it does not."
	| directory fileName entries |
	^Smalltalk platformName caseOf:{
		['Win32'] ->
			[directory := FileDirectory logDirectory.
			 (directory fileExists: (fileName := 'crash.dmp')) ifTrue:
				[directory fullPathFor: fileName]].
		['Mac OS'] ->
			[directory := FileDirectory on: (MacOSShell new getenv: 'HOME')
										, '/Library/Logs/'
										, (Smalltalk osVersion >= '1060'
											ifTrue: ['DiagnosticReports']
											ifFalse: ['CrashReporter']).
			 entries := directory entries select: [:e| e name beginsWith: 'OpenQwaq'].
			 entries notEmpty ifTrue:
				[directory fullPathFor: (entries  asSortedCollection: [:e1 :e2| e1 name <= e2 name]) last name]]. 
		['unix'] -> [nil]. "don't know how"
		} otherwise: [nil]. "don't know how"! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 6/27/2007 15:07'!
crashRecovery
	"Answer the crashRecovery of the receiver"
	<bewareOf: #crashRecoveryChanged>
	^self propertyValueAt: #crashRecovery! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 6/27/2007 15:07'!
crashRecovery: aValue
	"Modify the receiver's crashRecovery"
	^self propertyValueAt: #crashRecovery put: aValue with: #crashRecoveryChanged! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'ar 4/16/2011 17:29'!
createC3XMenu: appName

	"Check the server installed QPZ files and generate a c3x menu file for them if they don't
	yet exist. Called from QwaqParticipantUI>>initServerQPZMenu"

	|  cacheDir menuDir menuFile menuString fileName|

		cacheDir := FileDirectory cacheDirectory.
		cacheDir assureExistenceOfPath: 'Server Applications'.
		menuDir := cacheDir directoryNamed:'Server Applications'.
		fileName := appName, '.c3x'.
		(menuDir fileExists: fileName)ifFalse:[ "only if the file does not exist"
			menuFile := menuDir forceNewFileNamed: fileName.
			menuString  := 
'<group id="1" class="QFileCard">
	<name>', appName, '</name>
	<kind>panelApp</kind>
	<content>
		<object id="2" class="PyPanelDocCS">
			<appName>OpenQwaqPythonApps.', appName, '</appName>
		</object>
	</content>
</group>'.
	
		[menuFile nextPutAll: menuString] ensure: [menuFile close].
	].! !

!QwaqParticipantUI methodsFor: 'testing' stamp: 'ar 7/12/2011 12:52'!
createSelfContainedClientConnection
	"Useful for debugging stream stuff.  If you want to test something about end-to-end pub/sub, then create a couple of these, and publish with one / subscribe with the other."
	| con |
	con := QSelfContainedClientConnection 
				host: self serviceHost
				port: login servicePort
				user: (login isLocal ifTrue: ['admin'] ifFalse: [login userInput value])
				password: login password value
				org: login organization value.
	con connect.
	^con	
	! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'crl 9/1/2009 14:03'!
crowdPane
	"Answer the crowdPane of the receiver"
	<bewareOf: #crowdPaneChanged>
	^self propertyValueAt: #crowdPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'crl 9/1/2009 14:03'!
crowdPane: aValue
	"Modify the receiver's crowdPane"
	^self propertyValueAt: #crowdPane put: aValue with: #crowdPaneChanged! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 6/10/2009 11:27'!
currentFeatures
	"The array of symbols denoting enabled features for me in the current forum"
	^ (self featureProperties select: [ : v | v ]) keys! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 4/14/2010 17:46'!
currentForumName
	"If in-forum, the nice name of the forum we occupy right now; else nil."
	^ serviceSvc ifNotNil: [ serviceSvc name ]
! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 6/20/2009 21:02'!
currentForumURL
	"If in-forum, the url of the forum we occupy right now; else nil."
	^ serviceSvc ifNotNil: [ serviceSvc url ]
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/19/2009 14:24'!
currentOrgName
	"The organization name I'm signed into, or poised to sign into...
	  the current choice from the login pane."
	^ login organization value asString! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 2/15/2009 18:23'!
customsPane
	"Answer the customsPane of the receiver"
	<bewareOf: #customsPaneChanged>
	^self propertyValueAt: #customsPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 2/15/2009 18:23'!
customsPane: aValue
	"Modify the receiver's customsPane"
	^self propertyValueAt: #customsPane put: aValue with: #customsPaneChanged! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 4/13/2010 16:29'!
deactivateService: serviceUrl 

	"Experimental: Disconnect  (log out) of this one service,
	 while preserving the rest of the user's session.
	 See also reconnect:, which has much code in common."

	| normalizedUrl svcRecord islands |

	serviceUrl ifNil: [ 
		^ false 	"Already deactivated, we'll assume, since url-cache lookup probably failed."
	].
	"Check for async logout, error, etc."
	self readyForEvents ifFalse: [ ^ false ].

	normalizedUrl := QServiceLink normalizedServiceUrl: serviceUrl.
	"Check if we have the service."
	svcRecord := urlCache at: normalizedUrl ifAbsent:[nil].
	svcRecord ifNil: [ 
		self inform: 'Cannot find the service to disconnect.'.
		^ false.
	].
	"Check for async logout, error, etc."
	self readyForEvents ifFalse: [ ^ false ].

	QActivityCapture action: #LogoutForum details: serviceUrl.

	"Ensure we're not going to try to log out of the room we're in."
	islands := session islandsInServiceUrl: normalizedUrl.
	islands do: [:island |
		mainView activeSpace island == island ifTrue:[
			self inform: 'Cannot close the space you currently occupy.'.
			^ false.
		].
	].

	"Save and quite all the apps in the island"
	islands do:[:island|
		(self closeAppsInIsland: island) ifFalse:[^false].
	].

	self statusInfo: 'Disconnecting ', normalizedUrl.

	"OK - begin forgetting."
	urlCache keys copy do: [ : k |
		((TPostcard url: k hasBaseUrl: normalizedUrl) 
		 or: [TPostcard url: k hasBaseUrl: serviceUrl])
		ifTrue: [
			urlCache removeKey: k.
		].
	].

	"Drop all the islands... this is the meat of the operation."
	islands do: [ :island|
		self disconnectIslandController: island controller.
	].

	"Clean up various tables in the UI."
	servicesPane signal: #updateServiceList.
	historyPane discardService: serviceUrl.
	controlBar updateServiceCount.

	^ true.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 11/21/2007 18:28'!
defaultAvatarCostume
	"Zipped scene containing 'avatar.xml', with the xml for the user's avatar.
	 This is the costume used for a new user who hasn't made any preferences changes."
	^ QUserPreferences 
			xmlToZip: QPlainAvatar new persistString 
			unitName: 'avatar'! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 9/6/2008 00:03'!
defaultFeaturesForTier: tierName intoDictionary: orgDict

	"Bridging code until the tiering-to-features relationship is worked out.
	 If the current tier specifies defaults for features, 
	 seed them into the organization-level feature dictionary.
	 This is positioned in the feature-checking process 
	 such that one can edit an org's forums.properties to override the hardcoded tier default."

	QLogger logInfo: 'Tier ', (tierName).

	"Web Cam for Enterprise and Group."
	(orgDict includesKey: 'EnabledFeatures.WebCam') ifFalse: [
		({'EnterpriseEdition' . 'EnterpriseTier' . 'GroupEdition' . 'GroupTier'}
		includes: tierName)
		ifTrue: [
			orgDict at: 'EnabledFeatures.WebCam' put: 'true'.
		]].! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 1/4/2009 20:07'!
defaultPPSashFraction
	"Where we split people / places by default..."
	^ 0.46! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 9/9/2009 01:54'!
defaultPPWidth
	"The default target width for people&places.  See also control-bar layoutButtons"
	^ 221! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/7/2007 15:54'!
defaultServicePort
	"Answer the default port to use for services"
	^login servicePort! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'sn 5/13/2011 22:55'!
defaultStatusMessage
	| nPeople |

	self readyForEvents ifFalse: [ ^ nil ].
	nPeople := usersPane nPeopleHere.
	nPeople = 0 ifTrue: [  ^ 'There may be a connection problem.' ].
	
	self isMeetingInProgress ifFalse: [ ^ self idlingMessage ].

	(self isMeetingLeader not 
	 and: [meetingState audienceIsMuted ]
	 and: [ meetingState isLoginAllowedToSpeak: self userLogin ]) ifTrue: [
		^ 'You may speak' translateMe, 
		    (controlBar microphone selected
			ifFalse: [' - but your microphone is off. ']
		 	ifTrue: [ ' now. ']).
	].

	^ controlBar meetingStatusString, ' ', (self idlingMessage)! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/14/2010 11:14'!
defaultVidPaneHeight
	"The default height for the video strip.  ."
	^ 110! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 10/10/2006 13:11'!
delete
	self destroy.
! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'ar 4/16/2011 17:29'!
deleteServerPyLib
	"Delete the server based Python API libraries if there is no library on the server or if an error occurs in downloading it. "
	(FileDirectory cacheDirectory directoryExists:'OpenQwaqAPI')ifTrue:[	
		(FileDirectory cacheDirectory directoryNamed:'OpenQwaqAPI') recursiveDelete.
	].
! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 4/5/2007 01:20'!
demonAction: anAction

	"We execute demonic test actions in a startScript: handler 
	 to ensure the proper execution context.

	 We get here from the demon test engine...
		1) from the UI, we start the demon (beADemon);
	 	    this sets up a startScript: to run this method.
		2) the demon, from a background thread,
	 	    emits a signal (#action) with: an action selector.

		and our role here is to invoke that action on the demon,
		so the testing activity happens in my project / process context,
	 	as if part of normal UI activity.
	"

	(myDemon notNil and: [myDemon isDemonic])  ifTrue: [
		myDemon doAction: anAction
	].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'jcg 3/5/2010 14:40'!
destroy

	| x |
	icmpSocket ifNotNil:[icmpSocket destroy. icmpSocket := nil].

	self exitTripped: true.	"Just to be sure.. this is also set by quitting."
	Transcript removeDependent: self.
	SystemErrorHandler default: nil.
	myDemon ifNotNil: [myDemon stop].

	mobileProxy ifNotNil: [x := mobileProxy. mobileProxy := nil. x destroy. ].
	self serviceManager: nil.
	login hasLocalServices ifTrue: [
		QLocalServiceProvider reset.
		TConnectionDispatcher reset.
	].
	login destroy.
	self destroySession.
	super destroy.
	statusLine destroy.
	"Guarded against destruction during startup."
	controlBar ifNotNilDo: [:cb | cb soundDialog ifNotNilDo: [:sd | sd destroy]]. 
	audioSession ifNotNil: [:as |
		audioSession := nil.
		as destroy
	].

	"Clear out any dangling global state."
	TestingScriptRegistry keys do: [ :k|
		TestingScriptRegistry removeKey: k.].

	Smalltalk windowLabel:''. "restore default"

	Processor activeProject ifNotNil: [ : ap | 
		ap displayHook = self ifTrue: [ ap displayHook: nil ]].	 "Aid to garbage collection."

	"Not terribly important, but allows it to free up some references for garbage collection."
	TAnimationManager shutdownAnimations.

	"Stop background logging."
	[
		TTrafficCounter current logger: nil.
		QLogger reset.
	] on: Error do: [].

	
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 4/16/2011 17:29'!
destroyServiceSession

	"Take down the service provider session; close and discard all service managers."
	serviceMgr ifNotNil: [ QLogger logInfo: 'Closing OpenQwaq session.' ].

	self serviceManager: nil forceChanges: true.
	(Delay forMilliseconds: 50) wait.
	self installServiceManager: nil.

	"Service manager destroy is idempotent... this gets rid of stray recv-loops."
	svcMgrCache valuesDo: [ : v | v destroy ].
	svcMgrCache := Dictionary new.

	! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 3/25/2010 12:50'!
destroySession
	"Take down the Croquet session and all its dependent state."
	session ifNotNil: [ QLogger logInfo: 'Closing Croquet session.' ].

	"Take us out of the entire Croquet session and its supporting session-state."
	sanityChecker stopSpeakerChecker.
	self endVoiceChat.

	self stopScript: #onIslandSaved:postcard:author:.
	self stopScript: #onIslandAction:url:author:.
	self stopScript: #videoSessionTick.
	TAnimationManager shutdownAnimations.
	session ifNotNil: [
		"The org services controller is listed in the session to support reconnect,
		 but it is separately created and destroyed."
		orgServices ifNotNil: [
			session removeController: orgServices controller
		].
		(session islandControllers ifNil: [ #() ]) do: [ : c | c notifyTarget: nil ].
	].

	"Destroying the session discards all this stuff:"
	preferences ifNotNilDo: [:prefs | prefs setUserInfoRef: nil inClient: self ].
	audioSession ifNotNilDo: [:sess | sess view: nil ].
	mainView := nil.
	viewPanel view: nil.
	navPanel view: nil.
	activeSpace := nil.
	meetingState clearReference.
	session ifNotNil: [
		session destroy.
		session := nil.
		QLogger logInfo: 'Session closed.'.
	].
	avatar ifNotNil: [ avatar destroy. avatar := nil ].
	controlBar camera stopScript: #selected:.  "See setupVideoSessionInView:"
	lastBounds := nil.	"So we recreate OGL on re-entry."
	urlCache := Dictionary new.
	! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 3/31/2010 11:33'!
developerMenu
	"Not in sealed app; only in developer images, unless activated by alt-."
	| menu item subMenu |
	menu := CMenu new.

"Dumping snapshots."
	menu add: 'Dump to XML' target: self action: #dumpSpaceToXML.
	menu add: 'Dump to C3D' target: self action: #dumpSpaceToSnapshot.

"Demon testing"
	menu addSeparator.
	subMenu := CMenu new.
		(subMenu add: 'Not a Demon' target: self action: #noDemon) tag: #noDemon.
		QDemon modes do: [ :mode|
			(subMenu add: (mode capitalized) target: self action: #beADemon: argument: mode) 
			  tag: mode.
		].
	menu add: 'Demon' subMenu: subMenu.

	menu addSeparator.
	item := menu add: 'Experimental Features' target: self action: #toggleExperimentalFeatures.
	item checked: (preferences notNil and: [ preferences enableExperimentalFeatures ]).
	item tag: #experimentalFeatures.

	subMenu := CMenu new.
	menu add: 'Logging' subMenu: subMenu.
		item := subMenu add: 'Object-ID Log' target: self action: #toggleIDLog.
		item tag: #logIDs.
		item :=subMenu add: 'Panel Logs' target: self action: #togglePanelLog.
		item tag: #panelLog.
		item := subMenu add: 'Quindi Log' target: self action: #toggleQuindiLogging.
		item tag: #quindiLogging.
		item := subMenu add: 'Long Transcript' target: self action: #toggleLongTranscript.
		item tag: #longTranscript.
		item separator: true.
		item := subMenu add: 'Current Log Directory?' target: self action: #displayLogDirectory.

	subMenu := CMenu new.
	menu add: 'Operating Options' subMenu: subMenu.
		item:= subMenu add: 'Disable Materials' target: self action: #toggleMaterials.
		item tag: #materialsEnabled.
		item := subMenu add: 'Foreign Client Proxy' target: self action: #toggleForeignClientProxy.
		item tag: #foreignClientProxy. 
		item := subMenu add: 'ICMP Pinging' target: self action: #toggleICMP.
		item tag: #icmpPing.
		item := subMenu add: 'Import to Thumbnail' target: self action: #toggleDropModelThumbs.
		item tag: #thumbToggle.
		item := subMenu add: 'QMS Auto-Pause' target: self action: #toggleQMSAutoPause.
		item tag: #qmsAutoPause.
		item := subMenu add: 'Readonly Visitor Status' target: self action: #toggleReadonlyBehavior.
		item tag: #devGuestToggle.
		item := subMenu add: 'Zoom Embedded Applications' target: self action: #toggleEmbeddedApps.
		item tag: #zoomApps. 

"Graphics"
	menu addSeparator.
	TAnimationManager addAnimationChoicesMenuTo: menu.
	(menu add: 'Do Rendering' target: self action: #toggleOpenGL) tag: #toggleOpenGL.

"Profiling"
	menu addSeparator.
	subMenu := CMenu new.
	menu add: 'Profiling/Verification' subMenu: subMenu.
		subMenu add: 'Profile for 30' target: self action: #systemProfiler: argument: 30.
		subMenu add: 'Turn and Profile for 60' target: self action: #renderProfile: argument: 60.
		"Only offer choice of which profiler to use if more than one is loaded ;-)"
		(Smalltalk includesKey: #QVMProfiler) ifTrue: [
			subMenu
				add: 'Choose Profiler'
				subMenu: (CMenu new
							add: 'QSystemProfiler'
								target: self
									action: #profilerPreference:
										argument: #QSystemProfiler
											tag: #QSystemProfiler;
							add: 'QVMProfiler'
								target: self
									action: #profilerPreference:
										argument: #QVMProfiler
											tag: #QVMProfiler;
							yourself).
		].

		item := subMenu add: 'Replicated Random Test' target: self action: #toggleReplicatedRandom.
		item tag: #replicatedRandom.

		item := subMenu 
					add: 'Latency of packets in JB' 
					target: (Smalltalk at: #QAudioSinkEventProfiler) 
					action: #toggleGlobalProfiling.
		item tag: #jitterBufferLatencyProfile.

		item := subMenu add: 'Snapshot Every 5s' target: self action: #toggleSnapshotLoad.
		item tag: #snapshotToggle.

	subMenu := CMenu new.
	menu add: 'Special Tests' subMenu: subMenu.
		subMenu add: 'Shoot Media Controller' target: self action: #shootTheMediaController.
		subMenu add: 'Shoot Org Services' target: self action: #shootOrgServicesController. 
	subMenu addSeparator.
	     	subMenu add: 'Test Voice Extrapolation' target: self action: #voiceTest.
		subMenu add: 'Throttle Inbound Traffic' target: self action: #throttleInboundTraffic.
	subMenu addSeparator.
		subMenu add: 'Cause Synchronization Loss' target: self action: #causeSynchronizationLoss.
 
	item := menu add: 'System Timings...' target: latencyPane action: #toggle.
		self linkItem: item toTool: latencyPane.

"Video"
	menu addSeparator.
	menu add: 'Connect again to last NetVid video' target: self action: #connectAgainToLastNetVidID.

"Audio"
	menu addSeparator.

	subMenu := CMenu new.
	{ 0. 5. 10. 15. 25. 40. 60. 80 } do: [:activity |
		subMenu
			add: 'Activity Threshold ', activity
			target: self
			action: #setJitterBufferActivityThreshold:
			argument: activity
	].
	menu add: 'Jitter Buffer' subMenu: subMenu.

	subMenu := CMenu new.
	subMenu 
		add: '0 (lowest, default)' 
		target: QAudioSession
		action: #primSetPluginVerbosity: 
		argument: 0.
	#(1 2 3 4 5 6 7 8 9 ) do: [:verbosity | 
		subMenu 
			add: verbosity printString
			target: QAudioSession
			action: #primSetPluginVerbosity: 
			argument: verbosity.
	].
	subMenu 
		add: '10 (highest)' 
		target: QAudioSession
		action: #primSetPluginVerbosity: 
		argument: 10.
	menu add: 'QAudioPlugin verbosity' subMenu: subMenu.
	menu add: 'Record microphone to disk' target: self action: #recordRawMicrophoneData.

"Squeak stuff."
	menu addSeparator.	
	item := menu add: 'Squeak...' target: self action: (
		(Smalltalk includesKey: #CDemoScriptEditor) 
				ifTrue: [ #showSqScriptEditor ] 
				ifFalse: [ #noSuchMethod ]).
	self linkItem: item toTool: sqScriptEditor.
	menu add: 'Explore UI...' target: self action: #explore.

	self runScript: #updateDeveloperMenu when: {menu . #open}.
	^ menu! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/23/2010 14:08'!
dialogCenter
	"We want dialogs to materialize centered over the space-view, 
	 rather than the outer window... it sometimes looks more natural."
	placesPane visible ifTrue: [
		viewPanel width > 600 ifTrue: [ ^ viewPanel center x@ (height //2). ].
	].
	^ self center
! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'jcg 4/8/2009 17:47'!
disableVideoSession
	"Suppress video in this session
	 without changing preferences-videoDisabled. 

	 See also suspendVideoSession, which does the same thing 
	  but conditioned on whether there is any video set up."

	[ controlBar camera selected: false ] on: Error do: [ ].
	self suspendVideo.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 3/23/2009 17:34'!
discardIsland: oldIsland

	"Discard the #resourceChanged and other scripts signaled by the old island"
	self myScripts select:[:sc | (sc eventSpec size > 0) and: [(sc eventSpec at: 1) == oldIsland]]
				   thenDo:[:sc| self stopScript: sc].

	session ifNotNil: [ session discardIsland: oldIsland ].
	! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'bgf 3/23/2009 17:34'!
disconnectIslandController: aController

	"Close the connection to the given island controller -
	 which is assumed -NOT- to be the controller of the active island.
	 This is (experimental) support for logging out of individual services mid-session.

	Note that code fixes here may indicate a need
	for corresponding changes in reconnectIslandController."

	| url oldIsland |

	oldIsland := aController island.
	session islandControllers remove: aController ifAbsent: [
		self error: 'Closing an unknown controller.'.
		^ false.
	].
	aController disableMessages: true.
	session discardIslandResources: oldIsland.
	aController destroy.

	url := session findIslandUrl: oldIsland.
	self audioSession ifNotNilDo: [ :as | as flushForIslandReconnect: oldIsland ].
	session flushLinksMatching: url.

	"Discard stale state that might still have references or triggers in the old island."
	self discardIsland: oldIsland.

	"Retire the old island.  HAS to be syncSend, there is no message queue."
	[ 
		oldIsland syncSend: #discard withArguments:#().
	] on: Error do: [ :ex |
		QLogger logWarning: 'Exception retiring island: ', ex 
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 7/12/2011 12:54'!
displayInitialScreen
	"Display the initial screen without flickering or other garbage"

	projectMorph ifNotNil:[
		WorldState addDeferredUIMessage:[
			projectMorph noRedraw: false.
			projectMorph changed.
			self signal: #completeAfterShowingScreen].
		self waitUntil: #completeAfterShowingScreen.
		Display forceToScreen.].
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 9/26/2008 09:52'!
displayLogDirectory
	"Confused about where the log directory is on OSX/Vista/XP/etc.?  Thats what this method is for.  Of course, whether all of the logs are placed in this directory is a different story."
	CDialog inform: 'The current log directory is: ', 
						String cr, FileDirectory logDirectory fullName! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 17:30'!
disposeScreenShot: form 
	| fname choice tmpFile |
	"We've generated a screen shot... ask the user where to put it.
	 Expects the caller to do friendly error-handling."

	fname := 'OpenQwaq-', (DateAndTime now asFileNameString).
	choice := self disposeScreenShotDialog: form forName: fname.
	self readyForEvents ifFalse: [ ^ self ].   "Logout or somesuch."

	self mouseCursor: Cursor wait.
	[
	  choice caseOf: {
		[#desktop] -> [
			"Write it out as full PNG in all its glory."
			| out writer |
			out := (FileDirectory desktopDirectory forceNewFileNamed: (fname,'.png')).
			writer :=PNGReadWriter on: (out binary).
			writer nextPutImage: form.
			writer close.
			].		
		[#drop] -> [
			"Cap the size for memory-conservation, and send JPG into the forum."
			| writer outform |
			outform := form.
			form extent maxDimension > 1024 ifTrue: [
				outform := form magnifyBy: (1024.5 / form extent maxDimension)
			].
			tmpFile := FileDirectory tempDirectory forceNewFileNamed:(fname, '.jpg').
			writer := JPEGReadWriter2 on: (tmpFile binary).
			writer nextPutImage: outform quality: 80 progressiveJPEG: false.
			writer close.
			tmpFile open.
			self readyForEvents ifTrue: [
				session dropDocument: tmpFile event: nil.
			].
		].
		[#mail] -> [
			"Cap the size for politeness, and mail JPG around."
			| writer outform |
			outform := form.
			form extent maxDimension > 1024 ifTrue: [
				outform := form magnifyBy: (1024.5 / form extent maxDimension)
			].
			tmpFile := FileDirectory tempDirectory forceNewFileNamed: (fname, '.jpg').
			writer := JPEGReadWriter2 on: (tmpFile binary).
			writer nextPutImage: outform quality: 90 progressiveJPEG: false.
			writer close.
			self readyForEvents ifTrue: [
				self mailUsersAttaching: tmpFile name.
				tmpFile := nil.	"So we don't delete it... since the mailer will want it."
			].
		]
		} otherwise: [ ]
	] ensure: [ 
		[ tmpFile ifNotNil: [
			tmpFile close.
			FileDirectory tempDirectory deleteFileNamed: (tmpFile localName) 
			].
		] on: Error do: [ 	:ex|
			QLogger logWarning: 'Screen shot cleanup: ', ex description.
		].
		self mouseCursor: nil.
	].
			

! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 17:29'!
disposeScreenShotDialog: form forName: fname
	| dlog   img |
	"Canned dialog for asking how to use a screen shot.
	 Returns a symbol (drop, desktop, mail, nothing)"

	dlog := CUserDialog new.
	dlog label: 'Screenshot: ', fname.
	
	img := CPlayer new.
	img extent:  (form extent scaledWithin: (400@180));
		graphicForm: form; graphicFit: #autoScale;
		borderStyle: #simple; borderWidth: 1; borderColor: (Color black).
	dlog group add: img after: dlog messageText.

	dlog message: 'Actual size is ', (form extent x), ' x ', (form extent y), 
					'\\Where do you want to send the screen shot?' withCRs.
	(dlog messageText) allowEditing: false;
		hResizing: #shrinkWrap; 
		vResizing: #shrinkWrap.

	self canModifyForum ifTrue: [
		(dlog addButton:  'Into Forum' value: #drop) 
			width: 100.
	].

false ifTrue: [
		(dlog addButton: 'Mail ...' value: #mail) 
			width: 100;
			enabled: (false "Plumbing doesn't work").
].
	(dlog addButton: 'My Desktop' value: #desktop)
		  	width: 100.
	(dlog addButton: 'Cancel' value: #nothing)
		 	width: 75.

	[	"We don't have modal dialogs, but they'd be handy sometimes."
		self menu disable. 
		^ dlog run: nil centerAt: self dialogCenter.
	] ensure: [
		self menu enable.
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 11/19/2007 16:22'!
doOpenGL

	"If false, suppresses the openGL viewport creation and the rendering process,
	 for a low-CPU-load test client (that flies blind; the UI is present and active,
	 but the scene is not shown and no render-based picking is done."

	<bewareOf: #doOpenGLChanged>
	^self propertyValueAt: #doOpenGL! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 11/19/2007 15:53'!
doOpenGL: aValue
	"Modify the receiver's doOpenGL"
	^self propertyValueAt: #doOpenGL put: aValue with: #doOpenGLChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'ar 4/16/2011 17:30'!
doRenovateWithTemplate: description

	| converter url |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	((activeSpace get: #class) includesBehavior: QConferenceRoom) ifFalse: [
		^ self inform: 'Only room-style Forums can be renovated.'.
	].

	(self confirm: ('Experimental.  This will reconstruct this forum\using the current layout for a ', description name, '.\\Displays will be reloaded where possible.\\Surplus documents will be moved to an overflow rolodex.\Annotations and signs will be piled on the floor.\\Things created from the Objects menu may be removed.\\OK to proceed with rebuilding?\') withCRs)
	ifFalse: [ ^ self ].

	self mouseCursor: Cursor wait; waitTick.
	[
		url := session findIslandUrl: activeSpace island.
		converter := activeSpace island future new: (QForumConverter).
		converter wait.
		(converter future upgradeSpace: activeSpace toTemplate: description entryWorld)
			wait.
		session flushLinksMatching: url.
		converter island future obliterate: converter.

	] ensure: [
		self mouseCursor: nil.
	].! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'bgf 1/7/2010 17:49'!
docToolDrop: document at: aPoint fromService: svcRecord

	"Dispatcher for drag-drop of QAbstractDocuments from the document tool;
	 relayed here to decide whether to go to the service chooser, 
	 or a document-drop aware panel,
	 or into a logged in space."

	((self floatingPanes asArray), {catalogPane})  do: [ : pane | 
		(pane visible and: [ pane respondsTo: #docToolDrop:at:fromService: ]) ifTrue: [
			(self isPoint: aPoint inPlayer: pane ignoring: #() ) ifTrue: [
				^ pane docToolDrop: document at: aPoint fromService: svcRecord
			]]].

	serviceChooser visible ifTrue: [
		^ serviceChooser docToolDrop: document at: aPoint fromService: svcRecord
	].
	
	"OK, its for the logged-in case, and probably for dropping into the forum."
	self readyForEvents ifFalse: [
		self statusAlert: 'Not ready to receive document transfer.'.
		^ false
	].
	(meetingState isEditBlocked and: [ self isMeetingLeader not]) ifTrue: [ 
		self statusAlert: 'Imports are not permitted by the meeting leader.'.
		^ false ].
	self canModifyForum ifFalse: [
		self statusAlert: 'As a visitor, you cannot add documents.'.
		^ false
	].

	(self isPointInMainView: aPoint)  ifFalse: [
		self statusAlert:  'Cannot be dropped here.\Drop only in the space.'.
		^ false
	].

	"Do the main work as followup script so the return is immeidate; it starts the drop-in icon animation."
	documentsPane startScript: #dropDocument:at: withArguments: {document . aPoint}.

	^ true
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/1/2009 09:53'!
documentManager
	"Lazy on-demand creation of a document manager, if we have a session;
	 see also #clearDocumentManager."
	^ serviceMgr ifNotNil: [
		self propertyValueAt: #documentManager
			ifAbsentPut: [ 
				| newdm | 
				newdm := (QDocumentManager new owner: self; yourself).
				session ifNotNil: [ session documentManager: newdm ].
				newdm 
				]].
	! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/28/2009 12:32'!
documentsPane
	"Answer the documentsPane of the receiver"
	<bewareOf: #documentsPaneChanged>
	^self propertyValueAt: #documentsPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/28/2009 12:32'!
documentsPane: aValue
	"Modify the receiver's documentsPane"
	^self propertyValueAt: #documentsPane put: aValue with: #documentsPaneChanged! !

!QwaqParticipantUI methodsFor: 'pshare-utils' stamp: 'bgf 9/8/2009 12:56'!
downloadPShareDocument: document andLaunch: embeddableApp
	"Download a document and launch an associated embeddable app when completed.
	We need to do this in the UI since there are some interactions with the user that
	may be required."
	| localName dlg result shortName |
	<on: downloadPShareDocument>
	self ifTweak:[
		shortName := document get: #shortName.
		localName := FileDirectory pathFromURI: (document get: #asLocalFileUri).
		(FileDirectory on: (FileDirectory dirPathFor: localName)) assureExistence.
		(FileDirectory default fileExists: localName) ifTrue:[
			"XXXX: Need to be more clever here. Should give additional info about file size, modification date etc."
			dlg := CUserDialog new.
			dlg label: 'Confirmation'.
			dlg message: 'This document ', shortName, ' is already downloaded.' withCRs.
			dlg messageText allowEditing: false.
			dlg messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
			(dlg addButton: 'Use Local Version' value: #useLocal) width: 125.
			(dlg addButton: 'Use Server Version' value: #useServer) width: 125.
			(dlg addButton: 'Cancel Launch' value: #cancel) width: 125.
			result := dlg run: #cancel centerAt: self dialogCenter.
			result == #cancel ifTrue:[^embeddableApp app future stopRunning: (session userID)].
			result == #useLocal ifTrue:[^embeddableApp launchDocument: document].
		].
		"Download document into cache"
		[[
			"Unproven"
			((document get: #class) saveDocument: document asFile: localName forClient: self)
			 ifNotNil: [
				embeddableApp launchDocument: document.
			].
		 ] on: CProgressCancel do: [
			QLogger logInfo: 'Canceled PShare Download'.
			^ self
		 ]

		] on: Error do: [ : ex |
			self inform: 'Cannot download the shared document to open.\\' withCRs, ex description.
		].

	] ifNot:[
		self signal: #downloadPShareDocument with: document with: embeddableApp
	].! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 3/16/2009 22:33'!
drawWorldOn: aCanvas
	"This is the main world-drawing hook method (the top of the redraw loop.)
	 If we have a view, signal render, to trigger a simulation+render step.
       If the view has an ogl window, then render is all we need - tweak is drawn that way.
	 If there's no openGL, then let tweak draw the world."

	 self signal: #render.
	(mainView isNil or: [mainView ogl isNil]) 
		ifTrue: [
			self resetTweak.
			^ Processor activeProject drawWorldOn: aCanvas].
	^#()! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 8/20/2010 13:54'!
dropMaterial: aMaterialFarRef

	"Drag-drop from a material catalog onto the main view."
	| target selection |
	self readyForEvents ifFalse: [ ^ false ].
	(activeSpace island = aMaterialFarRef island) ifFalse: [ ^ false ].

	self viewPanel pickTarget.
	mainView pushTask: TPickAll new.
	selection := mainView fastPickSelection.
	mainView popTask.
	target := selection ifNotNil: [ selection target ].
	target ifNil:[ 
		self statusAlert: 'There is nothing there to paint.'.
		^ false.
	].
	selection normal ifNil: [
		self statusAlert: 'You cannot paint that object.'.
		^ false
	].
	self readyForEvents ifFalse: [ ^false ].
	
	avatar replica future paintFrame: target withMaterial: aMaterialFarRef. 
	
	^ true! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 8/20/2010 14:05'!
dropMaterial: materialRef at: aPoint
	"Drag-drop relay protocol.
	 Used when dragging a material-swatch from a material-pane.
	 MaterialRef is a far-ref to a TMaterial. "
	
	(self floatingPanes asArray)  do: [ : pane | 
		(pane visible and: [ pane respondsTo: #dropMaterial:at: ]) ifTrue: [
			(self isPoint: aPoint inPlayer: pane ignoring: #() ) ifTrue: [
				^ pane dropMaterial: materialRef at: aPoint
			]]].

	(self isPointInMainView: aPoint) ifTrue: [
		^ self dropMaterial: materialRef	
	].
	
	^ false! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 7/13/2008 18:00'!
dropPortalLink: url at: aPoint
	"Complete a DnD of a portal link onto some target in the replicated space"

	| card |
	mainView ifNil: [ ^ self ].

	"We just make a link-card, and hand it into the space 
	 for the usual in-space drag-drop processing."

	card := TIslandCopier new 
				import: (TIslandCopier new export:(QFileCard forServiceURL:url)) 
				to: mainView activeIsland.
	
	mainView dropCard: card 
			   position: (aPoint - mainView origin)
	 		   maxDropDistance: 128.	"Long-range - these drops are deliberate in-space drops."
! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 2/12/2009 11:19'!
dropWindowShare: windowID at: aPoint
	"Complete a DnD of a portal link onto some target in the replicated space"

	| card document |

	mainView ifNil: [ ^ self ].

	document := QPersonalShareDocument new.
	document owner: self userLogin.
	document windowID: windowID.
	document processID: (QPersonalShare windowPidFor: windowID).
	card := QFileCard forDocument: document.
	card isTransient: true.
	card := TIslandCopier new import:(TIslandCopier new export: card)
									  to: mainView activeIsland.

	mainView dropCard: card 
			   position: (aPoint - mainView origin)
	 		   maxDropDistance: 128.	"Long-range - these drops are deliberate in-space drops."
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 7/7/2007 12:28'!
dumpSpaceToSnapshot

	"For testing: dump the current space to the working directory;
	 the resulting file is 'dump-<island name>.c3d"

	| filename file controller data |
	controller := mainView activeSpace controller.
	data := controller snapshot.

	filename := 'dump-', ((mainView activeSpace island get:#name) onlyLetters), '.c3d'.
	file := FileDirectory default forceNewFileNamed: filename.
	file binary.
	[file nextPutAll: data ] ensure: [file close].

	
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 12/4/2006 16:12'!
dumpSpaceToXML

	"For testing: dump the current space to the working directory;
	 the resulting file is 'dump-<island name>.xml"


	| filename file xml writer |
	writer := mainView activeSpace island future new: TSceneWriter.
	xml := writer future storeScene: (mainView activeSpace island).
	xml wait.

	filename := 'dump-', ((mainView activeSpace island get:#name) onlyLetters), '.xml'.
	file := FileDirectory default forceNewFileNamed: filename.

	file ascii.
	[file nextPutAll: (xml value) ] ensure: [file close].

	
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:35'!
editCopy: evt
	"Edit menu action - Copy"
	QActivityCapture action: #EditCopy.
	avatar copySelection: evt! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:35'!
editCut: evt
	"Edit menu action - Cut"
	QActivityCapture action: #EditCut.
	avatar cutSelection: evt! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:35'!
editDelete: evt
	"Edit menu action - Delete"
	QActivityCapture action: #EditDelete.
	avatar deleteSelection: evt! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/11/2009 16:13'!
editForumDetails
	"The extras menu 'Forum Details...' command...
	 simply opens a Property sheet on the Space,
	 where motion limits and so on can be controlled.
	 The menu item just makes this more discoverable."
	self readyForEvents ifFalse: [
		^ self inform: 'The space is not available for editing.'.
	].
	self treeInspector editProperties: activeSpace.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:35'!
editLocate: evt
	"Edit menu action - Locate"
	QActivityCapture action: #EditLocate.
	self readyForEvents ifTrue: [
		avatar replica future locateSelection	
	]! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 14:19'!
editMenu
	| menu |
	menu := CMenu new.
	self runScript: #updateEditMenu when: {menu . #open}.
	^menu! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:35'!
editPaste: evt
	"Edit menu action - Paste"
	QActivityCapture action: #EditPaste.
	avatar pasteSelection: evt! !

!QwaqParticipantUI methodsFor: 'proxy' stamp: 'ar 7/12/2011 12:53'!
editProxyConfiguration
	"Edit the current proxy configuration"

	login disable.
	proxyDlg openCenteredAt: self bounds center.
	proxyDlg waitUntil: #closed.
	login enable.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:34'!
editRedo: evt
	"Edit menu action - Redo"
	QActivityCapture action: #EditRedo.
	avatar redo: evt! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/5/2010 17:35'!
editUndo: evt
	"Edit menu action - Undo"
	QActivityCapture action: #EditUndo.
	avatar undo: evt! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 9/17/2007 12:17'!
embeddedOverlay
	"Answer the embeddedOverlay of the receiver"
	<bewareOf: #embeddedOverlayChanged>
	^self propertyValueAt: #embeddedOverlay! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 9/17/2007 12:17'!
embeddedOverlay: aValue
	"Modify the receiver's embeddedOverlay"
	^self propertyValueAt: #embeddedOverlay put: aValue with: #embeddedOverlayChanged! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 17:30'!
emptyCache

	| emptyDirs msg |
	QActivityCapture action: #EmptyCacheExtrasMenu.
	"Implementation for Extras menu -> Clear Resource Cache"

	(self confirm: 
'Clears all cached resources from disk.
Resources will be reloaded when you next log in.

Proceed?') ifFalse: [ ^self ].

	"Don't do this out from underneath a cache operation."
	TContentManager default cacheManager mutex critical: [
		[ 	"don't touch the Python directories - especially while we are running"
			emptyDirs := FileDirectory cacheDirectory directories select:[:dir | (#('Installed Applications' 'Server Applications' 'OpenQwaqAPI' 'OpenQwaqPythonApps' 'OpenQwaqPythonUserApps') includes: dir localName) not].
			emptyDirs do: [ :dir | dir deleteLocalFiles].	"Per-login caches"
			FileDirectory cacheDirectory  deleteLocalFiles
		]  on: Error do:[:ex| msg := ex description. ex return.].
	].
	msg 
		ifNil:[self statusInfo: 'Disk cache has been emptied.'] 
		ifNotNil:[self inform: 'Unable to clear cache: \\' withCRs, msg].
! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'bgf 11/24/2009 13:43'!
enableVideoSession
	"Enable or re-enable video after it was disabled; 
	 enable the preference if it was disabled,
	 and enable live video in any case."
	(preferences notNil and: [preferences videoDisabled]) ifTrue: [
		preferences videoDisabled: false.		"Normal updating will restore the manager."
		self commitPreferencesInBackground.
	]  ifFalse: [
		self resumeVideo
	].! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 12/24/2006 12:00'!
endOverlay: ogl
	"Set up gl to provide overlay drawing - this is done in pixel coordinates"
	ogl popBlend.
	ogl glMatrixMode: GLModelview.
	ogl glPopMatrix.
	ogl glMatrixMode: GLProjection.
	ogl glPopMatrix.
	ogl glMatrixMode: GLModelview.
! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'jcg 9/11/2007 19:21'!
endVideoChat
	self videoSession ifNotNilDo: [:vs | vs cameraManager endVideoChat]
! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/30/2010 11:22'!
endVoiceChat
	self stopScript: #micActivity:.
	audioSession ifNotNil: [
		audioSession future endVoiceChat.
		self statusInfo: 'Microphone Off.'
	].
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 8/18/2010 14:59'!
ensureDisplay: cardName
	self stopScript: #ensureDisplay:.
	self readyForEvents ifFalse: [^self statusAlert: 'Unable to display ', cardName].
	avatar replica future ensureDisplay: cardName.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'jcg 8/30/2008 00:09'!
entryZoom 

	"Once the scene is ready, pull the camera in from the far third person state
	 to the user's preferred viewpoint.  This happens on first entry into a session
	 to orient the user and make them aware of their current avatar."

	| vp |

	"This prevents the zoom from starting until the avatar addToNewSpace activity
	 has completed, and the avatar is posed in the space."
	(Delay forMilliseconds: 50) wait.

	avatar ifNil: [ ^ self].
	avatar replica isValidFarRef ifFalse: [ ^ self].
	avatar replica wait.	"Forces a mutex rendezvous."

	avatar ifNil: [ ^ self ]. 	"Outta here"
	"Zoom inwards to the user's preferred viewpoint"
	vp := avatar thirdPerson.
	[vp > preferences povValue] whileTrue: [
		exitTripped ifTrue: [ ^ self ].
		(Delay forMilliseconds:25) wait.
		self changeViewpointValue:(vp := ((vp - 0.05) max:(preferences povValue))).
	].
	
	! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 2/23/2010 15:09'!
errorTripped
	"Has the session indicated that an error situation has arisen, or if no session, 
	has orgServices so indicated (or is orgServices nuked altogether)?"
	session ifNotNil: [^session errorTripped].
	orgServices ifNotNil: [^orgServices errorTripped].
	^true.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 6/8/2008 23:44'!
exit
	"We run this through a signal in case it's being called from
	 the system window setup... we want this to happen in our own project context."
	self signal: #exit.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 7/1/2010 15:45'!
exitConfirming: bConfirm
	"Confirm and exit the application."

	"If we're trying to quit after there's already error state in the app,
	 the save and/or replica-cleanup can lock up getting island mutexes.
	 Don't make that an infinite struggle, only try it once."

	self mouseCursor: Cursor wait.

	"Confirm exit if we're up and healthy"
	(session notNil and: [ exitTripped not ] and: [ session errorTripped not ])
	 ifTrue: [
		"Trigger autosave before confirm so it happens while the dialog is taking
		 the user's attention.  (A little finesse so there's less lag in real exit,
		 and so hard quit is less likely to fire until we're done.)"
		(bConfirm and: [self readyForEvents]) ifTrue: [ 
			preferences doAutoSave ifTrue: [ 
				[ activeSpace critical: [ activeSpace future aboutToQuit ] ifLocked: []]
				 on: Error do: [] ].
			(self confirmExit) ifFalse: [
				exitTripped := false.
				self mouseCursor: nil.
				self useNormalTimeouts.
				 ^ self
			].
			QLogger logInfo: 'Exit confirmed.'.
			preferences doAutoSave ifTrue: [ (Delay forMilliseconds: 300) wait. ].
		]. 
		"This makes the exit appearance quicker..."
		[	
			self useQuickTimeouts.
			self retractWindow.
			self saveLayout.
		]  on: Error do: [ :ex | QLogger logError: ex ].

		"This reduces island contention during the takedown processes..."
		session killRender.
		QLogger logInfo: 'Rendering stopped.'.
		"Extension point for shutting stuff down."
		[ self loggingOut: true ] on: Error do: [ :ex |
			QLogger logError: ex.
		].
		exitTripped := true.

		QLogger logInfo: 'Exit tripped.'.
		"Make an effort to get our avatar out of the space as we leave."
		[	orgServices ifNotNil: [ 
				orgServices clearPresence. 
				(Delay forMilliseconds: 100) wait. ].
			QLogger logInfo: 'Location cleared.'.
			session videoSession: nil. 
			session audioSession: nil.  "So the destroy doesn't trigger events."
			audioSession ifNotNilDo: [:as |
				"Josh isn't sure whether this is necessary.  It will happen later
				 in #destroy, but he's not sure whether #destroy is called in
				 a sealed application."
				audioSession := nil.
				as destroy.
			].
			 QLogger logInfo: 'Media handlers cleared.'.
			 mainView avatar replica future destroyAtClientTime: 
													(QChatLog clientUtcTime).
			 mainView avatar replica: nil.
			 QLogger logInfo: 'Replica retired.'.
		]  on: Error 
			do: [:ex|
				self isSealedApplication ifFalse: [ 	
					QLogger logWarning: 'Exit-confirming minor error: ', (ex description)]].
		(Delay forMilliseconds: 100) wait.  "Help the destroy message on its way."
		TTrafficCounter current flushLog.
	].

	"Try to catch case: open to chooser, move it around, close with x-box."
	[serviceChooser saveShape] ifError: [ :msg | QLogger logWarning: 'Save-shape: ', msg ].

	"More room for cleanup msgs to process, before quitPrimitive."
	(Delay forMilliseconds: 150) wait.
	exitTripped := true.

	QLogger logInfo: 'Cleanup complete. '.

	self isSealedApplication ifTrue:[
		"This is the sealed application case."
		Smalltalk quitPrimitive.
	] ifFalse: [
		"This is the unsealed case... trying to take down Forums from
		 a possibly debugged or otherwise unpredictable state."

		"Last effort to kill off broken render that might be holding locks."
		[ session errorTripped ifTrue: [ session killRender ] ] 
			on: Error do: [].

		"This should work and if it doesn't we need to be aware."
		self destroySession.
		self destroyServiceSession.
		QSharePointRepository resetCache.
		Python stopCallbackProcess.	"Releases CProject/island references."

		"Workaround until this can be fixed properly... File/exit from fullScreen
             sometimes leaves a zombie UI appearance covering the Squeak desktop.
		 See Qwaq Mantis 0084."
		window fullScreenMode ifTrue: [
			self toggleFullScreen.
		].
		audioSession ifNotNil: [audioSession  destroy. self audioSession: nil].
		"Attempting to close or delete the containing system window from 
	 	within this menu callback causes grief; the CProjectMorph 
	 	does not like to be dismantled from within itself."
		WorldState addDeferredUIMessage:(MessageSend
			receiver: self window window selector: #closeBoxHit).
		WorldState addDeferredUIMessage:(MessageSend
			receiver: World selector: #restoreMorphicDisplay).
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/6/2007 16:44'!
exitTripped
	"Answer the exitTripped of the receiver"
	<bewareOf: #exitTrippedChanged>
	^self propertyValueAt: #exitTripped! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/6/2007 16:44'!
exitTripped: aValue
	"Modify the receiver's exitTripped"
	^self propertyValueAt: #exitTripped put: aValue with: #exitTrippedChanged! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 10/22/2008 21:31'!
experimentalObjectsDirectory
	"The directory for objects common to all applications."
	^FileDirectory default directoryNamed: (FileDirectory default pathFromURI: 'objects-menu/Experimental').! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:34'!
extrasMenu
	| menu item subMenu |

	menu := CMenu new.

	(menu add: 'Ignore Gravity (Fly)' translateMe target: self action: #toggleRespectForTheLaw) 
		checked: false;
		separator: true;
		tag: #ignoreMotionRules.

	menu addSeparator.
		(menu add: 'Forum Details...' translateMe target: self action: #editForumDetails) tag: #extrasForumDetails.

	(menu add: 'Set Home Position...' translateMe target: self action: #makeThisTheHomePosition) tag: #extrasEntryPosition.

	"The following are both less useful now that we're using remote XML all the time,
	 and not up to date with current controller/session connectivity."
	"menu add: 'Test Persistence' target: self action: #persistenceTest."
	"menu add: 'Test Reconnect' target: self action: #reconnectTest."

	menu addSeparator.
	(menu add: 'Reset Sound' translateMe target: self action: #extrasResetSound)
		separator: true.

	subMenu := CMenu new.
	subMenu add: 'Mono Sound' translateMe target: self action: #testBuiltinSound.
	subMenu add: 'Spatial Sound' translateMe target: self action: #testOpenALSound.
	subMenu add: 'Voice Transport' translateMe target: self action: #testVoiceTransportRepeatedly.
	subMenu add: 'Voice Transport with .wav file' translateMe target: self action: #testVoiceTransportRepeatedlyWithUserFile.
	subMenu add: 'Client Experience' translateMe target: self action: #testClientExperience.

	subMenu addSeparator.
	subMenu add: 'OpenQwaq Multi-Share' translateMe target: self action: #placeXSetRootCard.
	subMenu addSeparator.
	subMenu add: '360 Rendering' translateMe target: self action: #render360.
	(menu add: 'Test' translateMe subMenu: subMenu) tag: #extrasTestMenu.

	menu addSeparator.
	item := menu add: 'Clear Resource Cache' translateMe target: self action: #emptyCache.
	item tag: #emptyCache.

	menu addSeparator.
	(menu add: 'Change Password...' translateMe target: self action: #changeUserPassword) tag: #extrasChangePassword.

	self runScript: #updateExtrasMenu when: {menu . #open}.

	^menu
! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 2/17/2010 18:42'!
extrasResetSound
	"Reset sound - called from the extras menu"
	QActivityCapture action: #ResetSoundExtrasMenu.
	self resetSound: 'extras menu'.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/14/2009 13:47'!
faceCache
	"My QUserFaceCache - where we stow avatar face-pictures."
	^ infoCache faces! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 5/27/2008 19:49'!
featureProperties
	"Answer the featureProperties of the receiver"
	<bewareOf: #featurePropertiesChanged>
	^self propertyValueAt: #featureProperties! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 5/27/2008 19:49'!
featureProperties: aValue
	"Modify the receiver's featureProperties"
	^self propertyValueAt: #featureProperties put: aValue with: #featurePropertiesChanged! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 3/31/2009 15:48'!
fetchPreviewUpdate: loginUrl
	"Update the background preview we're using during login"
	| e |
	(infoCache previewForUrl: loginUrl) ifNotNilDo: [ : form |
		
		e := servicesPane entryForUrl: loginUrl.
		self graphic: ((e notNil and: [ e isActive ])
						ifTrue: [ self class asMutedColor: form] 
						ifFalse: [ self class asTonedGray: form ] ).
		self graphicFit: #stretch.
		].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:34'!
fileImport

	| fileToImport |

	self readyForEvents ifFalse: [ ^ self ].
	self canModifyForum ifFalse: [ ^ self ].

	[
		"Native dialog if available.  The file types supported are an open-ended set,
	       so we allow anything at the chooser and then get more specific in the dropFile: method."
		fileToImport := QFileDialog openFile: {
				{'Documents' . '*.*'} }
			label: 'Choose a File to Import' translateMe.
	] on: Error do: [
		^self inform: 'Could not open file chooser.'
	].

	fileToImport ifNotNil: [
		| f | 
		[
			QActivityCapture action: #ImportFileMenu 
						       details: (QActivityCapture fileDetails: fileToImport).
			f := FileDirectory default readOnlyFileNamed: fileToImport.
		] on: Error do: [ :ex |
			^ self inform: ('Cannot read file\  ', fileToImport) withCRs
		].
		f ifNotNil: [ session dropFile: f event: nil editBox: true]
	] ifNil: [
		QActivityCapture action: #ImportFileMenuAbandoned
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:33'!
fileMenu
	| menu |
	menu := CMenu new.
	menu add: 'Exit' translateMe target: self action: #onExit.

	self runScript: #updateFileMenu when: {menu . #open}.
	^menu! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 5/7/2010 10:06'!
finalQuitConfirm: logout afterSaveWasChecked: bSaveChecked

	"Final chance to offer exit-time queries or confirm logout.
	  logout - true if this is a 'Log Out' (back to service chooser) departure, false for full Exit,
			or a string to indicate where you're going next (e.g., a URL).
	  bSaveChecked  - true if the user has already been asked whether or not to save forums,
					   (and the saving has been done if requested.)
	Returns true if the departure should proceed."

	"Attempt to close all owned applications in the space"
	(self closeAppsInAllIslands) ifFalse:[^false].

	mobileProxy ifNotNil: [^true].  "Mobile user has requested a different org in a warm proxy."

	"Exit-time survey stuff.
	 Should we ask for feedback? 
	 Do this only if we've been in a session for a non-trivial time.
	 Probably the time threshold should be longer (greater than one minute)
	 but we'll let people discover that for themselves."
	(QExitQuizDialog surveyQuestions: self) ifNotNilDo: [:questions |
		^ QExitQuizDialog runFor: self 
				quitLabel: (self shortQuitLabel: logout)
				questions: questions
				name: 'exit_survey'
		].

	"Fallthrough - this is normal exit with no survey popup.
	 Just confirm logout if we haven't yet.
	 (If bSaveChecked was true, 
	 then the user has already been given a dialog about quitting,
	 so we don't bother them further.) "
	bSaveChecked ifFalse: [
		^ self confirm: (self longQuitDescription: logout), '?'
	].

	^ true
	! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 4/22/2007 12:45'!
findPreviewSpaceFor: aService

	"Return the TSpace far ref we should use to render a preview of this service,
	 or nil if we haven't yet resolved and activated that space."

	| eurl vp |

	eurl := TPostcard makeUrl: (aService entryUrl) relativeTo: aService url.
	vp := session findViewpointByPostcard: (TPostcard urlString: eurl).

	"We don't wait or verify... we return nil in all unready or not-found cases."
	^ (vp get: #root)! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'jcg 5/15/2010 01:18'!
finishMediaReconnection
	"We do this after the media router connection has broken and been repaired."
	| islands appsToDestroy |
	"Re-establish voice connections (both incoming and out-bound)."
	session ifNil: [ ^ self ].
	QLogger log: '#finishMediaReconnection start' from: self.
	islands := session allSessionIslands.
	self audioSession ifNotNilDo: [:aSess |
		aSess speakerManager future recreateAllSpeakers.

		"XXXXX: maybe this should go in the QAudioSession event-loop,
		 but it's too risky a change right now, and there's no evidence that
		 it causes any problems.  After a quick inspection of the code, it 
		 seems better to leave it alone for now."
		self updateVoiceRelay.
	].
	session ifNil: [ ^ self ].

	"Re-establish video connections (both incoming and out-bound)."
	self videoSession ifNotNilDo: [:aSess |
		aSess webcastManager destroyAllTexturesAndDecoders.
		islands do: [:island | aSess webcastManager updateVideoTexturesInIsland: island].
		aSess cameraManager republishVideo.
	].
	session ifNil: [ ^ self ].

	"Destroy all apps that make their own network connections, so that they will be recreated, and reconnect themselves."
	 appsToDestroy := session embeddedApps "embeddable, actually"
									select: [:app | app shouldDestroyUponReconnect].
	appsToDestroy do: [:app |
		session ifNil: [ ^ self ].
		QLogger log: 'Resetting reconnected app: ', app from: self.
		session embeddedApps removeKey: (app app get: #appID) ifAbsent: [].
		app destroy.
	].
	session ifNil: [ ^ self ].

	"Recreate QMS/video/etc. apps that were destroyed above.  Also, if we resynced islands from others (eg: if we were not alone in a Forum) then we may have some embeddable-apps whose far-ref is to an object on the discarded island, not the newly-synced replica of the island.  Updating the embedded-apps for each island fixes this."
	islands do: [:isl | session updateEmbeddedAppsIn: isl newSession: false].

	QLogger log: '#finishMediaReconnection end' from: self.! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'bgf 11/15/2009 10:07'!
finishOrgReconnection

	"We do this after any reconnect, to ensure we didn't miss  presence signals
	 while the relevant island was unconnected."
	usersPane signal: #updateUserList.

	"Users list catches changes in the leadership status of others...
	 but if I'm the leader, or I've lost leadership, make sure I know it."

	self isMeetingLeader: (
			self currentForumURL notNil
			and: [ orgServices notNil ] 
			and: [orgServices meetingLeaderSessionID = self sessionID]).
	! !

!QwaqParticipantUI methodsFor: 'menu-handling' stamp: 'bgf 11/14/2009 23:58'!
fireTMenu: tMenu item: idx

	self readyForEvents ifFalse: [ ^ self ].

	mainView task menuMode ifTrue: [ mainView task menu: nil; popSelf ].

	( tMenu isValidFarRef  and: [ tMenu island = mainView activeIsland] )  
	ifTrue: [
		mainView event utcSeconds: (Time utcSeconds).
		tMenu future 
			fireItem: idx 
			forEvent: (mainView event) 
			fromAvatar: (mainView avatar replica).
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/20/2008 13:17'!
firstArrivalPending
	"Answer the firstArrivalPending of the receiver"
	<bewareOf: #firstArrivalPendingChanged>
	^self propertyValueAt: #firstArrivalPending! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/20/2008 13:17'!
firstArrivalPending: aValue
	"Modify the receiver's firstArrivalPending"
	^self propertyValueAt: #firstArrivalPending put: aValue with: #firstArrivalPendingChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 6/27/2007 10:26'!
firstLaunch
	"Answer the firstLaunch of the receiver"
	<bewareOf: #firstLaunchChanged>
	^self propertyValueAt: #firstLaunch! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 6/27/2007 10:26'!
firstLaunch: aValue
	"Modify the receiver's firstLaunch"
	^self propertyValueAt: #firstLaunch put: aValue with: #firstLaunchChanged! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 3/23/2009 23:59'!
flickerRecordingIndicator
	"A routine ticker for the recording indicator.
	 It has two slightly different images that we swap to make it 'glow',
	 very quietly so it is not annoying... just slightly alive."
	recordingIndicator visible ifTrue: [
		recordingIndicator selected: recordingIndicator selected not.
	].! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 12/11/2009 18:02'!
flickerWebcastIndicator
	"A routine ticker for the recording indicator.
	 It has two slightly different images that we swap to make it 'glow',
	 very quietly so it is not annoying... just slightly alive."
	webcastIndicator visible ifTrue: [
		webcastIndicator selected: webcastIndicator selected not.
	].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 4/5/2008 02:12'!
floatFromAvatar: aString

	"Private handler; see updateStatusRelay....
	 this relays in-space float-messages signals to the status line.
	 Note that we don't do this - we don't post or reinforce float-messages -
 	 while doing mouse driving, as no picking is happening while that is going on,
	 so the in-space objects believe the pointer-over state from before driving started."

	(mainView isNil or: [ mainView mouseDriving]) ifFalse: [
		statusLine floatingMessage: aString.
	].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 3/24/2009 00:43'!
floatMessage: message
	"Deliver the message as a yellow-tag soft alert.
	 This is done through a signal so it is safe from any Process."
	self signal: #floatMessage with: message! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 1/5/2009 17:42'!
floatingPanelArea
	"The bounds that QFloatingPanels should be confined to -
	 corresponds to the area inside the menu bar and bottom status panel.
	 26 is the menu's effective height, but the menu is rarely reliable in telling us that."

	^ ppToggle selected ifTrue: [
			Rectangle 
				origin: (0 @ 26) 
			 	corner: (viewPanel width) @ (controlBar top)
		] ifFalse: [
			Rectangle
				origin: (0@26)
				corner: (self width-1)@(controlBar top).
		]! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 5/1/2007 17:00'!
floatingPanes
	"All the QFloatingPanes used in this UI."
	^ self elements select: [ : e | e isKindOf: QFloatingPane ]! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 11/4/2006 09:05'!
flowControlFlag
	"When true, the rendering (and driveStep) rate held below 10FPS."
	<bewareOf: #flowControlFlagChanged>
	^self propertyValueAt: #flowControlFlag! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/4/2007 23:58'!
flowControlFlag: aValue
	"When true, the rendering rate forced below 25 FPS with 40ms delays to free cycles,
	 and the IO arrow indicators appear in the status bar."
	controlBar ioIcon visible: aValue.
	^self propertyValueAt: #flowControlFlag put: aValue with: #flowControlFlagChanged.
	! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 8/22/2009 18:49'!
forkEntryZoom
	"Bridge between startScript: and the forked ceremony 'entryZoom'
	 which does scripted delays and needs to run in the background."

	self readyForEvents ifFalse: [ ^ self ].
	avatar thirdPerson: ((preferences povValue + 2.5)max: 3.25).

	[self entryZoom] fork.
	self stopScript: #forkEntryZoom.	
		"We don't do this on every space transition; just on sign-in and teleports."! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'sn 5/13/2011 15:26'!
forumsMenu

	"To seed the menu; just has a default command."
	| menu |
	menu := CMenu new.
	menu add: 'All Forums...' translateMe target:  self action: #showServicesPane.
	self runScript: #updateForumsMenu when: {menu . #open}.
	^menu

! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/7/2007 03:37'!
forwardButton
	"Answer the forwardButton of the receiver"
	<bewareOf: #forwardButtonChanged>
	^self propertyValueAt: #forwardButton! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/7/2007 03:37'!
forwardButton: aValue
	"Modify the receiver's forwardButton"
	^self propertyValueAt: #forwardButton put: aValue with: #forwardButtonChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/4/2008 13:19'!
fpsStats
	"Answer the fpsStats of the receiver"
	<bewareOf: #fpsStatsChanged>
	^self propertyValueAt: #fpsStats! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/4/2008 13:19'!
fpsStats: aValue
	"Modify the receiver's fpsStats"
	^self propertyValueAt: #fpsStats put: aValue with: #fpsStatsChanged! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'hrs 7/21/2010 16:39'!
fullReportFor: incidentString error: anErrorOrNil stack: preStackOrNil on: strm
	"Print the given incident string at the front of a full standardized error report payload."

	QLogger drainLogger.
	self waitTick.				"Allow logView to receive the drained log updates."
	strm nextPutAll: incidentString; cr.
	anErrorOrNil ifNotNil: [strm nextPutAll: anErrorOrNil description; cr].
	strm nextPutAll: ('Version: ', QwaqVersion current infoString).
	mobileProxy ifNotNil: [strm nextPutAll: '   (mobile proxy)'].
	strm cr.
	
	[  "Guard the following extraction of client details so we don't miss reports..."
		(self isDemonic) ifTrue: [
			strm nextPutAll: myDemon infoString; cr; cr.
		].
		strm
		  nextPutAll: (self login summaryString);
		  cr;
		  nextPutAll: 'Session duration: ', (session 
												ifNil: ['(no session)'] 
												ifNotNil: [session timeOnline]);
		  cr; cr;
		  nextPutAll: (statisticsPane reportString);
		  cr; cr;
		  nextPutAll: (self usersPane loggedInUserSummaryString); 
	        cr; cr;
		  nextPutAll: (Python statusInformationString);
		  cr; cr;
		nextPutAll: 'VM statistics'; cr; nextPutAll: (Utilities vmStatisticsReportString);
		cr; cr.
	] on: Error do: [ :ex |
		strm cr; 
			  nextPutAll: 'Session details truncated: ', (ex description);
			  cr
	].
	preStackOrNil ifNotNil: [
		  self printStackReport: preStackOrNil on: strm.
		  strm cr.
	].
	strm
	        nextPutAll: '------- Transcript -------'; cr;
		  nextPutAll: logView panel contents;
		  cr; cr;
		  nextPutAll: '------- Five Second ObjectID Log -------'; cr.
	TObjectID fiveSecondLogReport: strm.
"	strm
		  cr; cr;
		  nextPutAll: '------- Five Second Croquet Log -------'; cr.
	TIslandController fiveSecondLogReport: strm.
	TIsland fiveSecondLogReport: strm."
	strm cr; cr;
		  nextPutAll: '------- System Profile -------'; cr;
		  nextPutAll: (QwaqVersion sysProfile);
		  cr.
! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 8/7/2008 15:14'!
fullReportStringFor: incidentString error: anError stack: preStackOrNil

	"Put the given incident string at the front of a full standardized error report payload,
	 and return it as a string."

	| strm |
	strm := WriteStream on: (String new: 1024).
	self fullReportFor: incidentString error: anError stack: preStackOrNil on: strm.
	^ strm contents! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 1/29/2009 15:44'!
getCurrentViewAngle
	"The user's current main viewing angle (the basis of the scene projection),
 	 or the preferred angle if the mainView isn't fully set up yet."

	^ mainView currentViewAngle ifNil:[ preferences viewAngleValue ] ! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'jjl 10/30/2008 18:07'!
getNominalViewAngle

	^preferences viewAngleValue! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 3/9/2009 10:18'!
getPanelNameFor: aQPlace
	"Answer the user readable/settable objectName for the panel at the given place."
	^(self activeSpace island get: #lookup: with: aQPlace destinationID) get: #objectName! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'das 11/15/2007 13:55'!
glRenderCostume: aCostume on: ogl
	"Render a Tweak costume in OpenGL"
	| box formCache material tfm m44 tw th |

	aCostume visible ifFalse:[^self].
	box := aCostume drawingBounds.
	box := box origin asIntegerPoint corner: box corner asIntegerPoint.
	formCache := aCostume localCache ifNil:[^nil]. "no cache, no render"
	"grab the texture cache"
	material := oglCache at: aCostume ifAbsent:[nil].
	"make sure it hasn't changed"
	(material == nil or:[material textureForm ~~ formCache form]) ifTrue:[
		material ifNotNil:[ogl deleteTexture: material texture sha].
		material := formCache material.
		material ifNotNil:[oglCache at: aCostume put: material].
	].

	material ifNil:[^nil]. "suppress render for this frame if no material"

	"draw the costume"
	material enable: ogl.
	tfm := aCostume transform.
	ogl glPushMatrix.
	ogl glTranslatef: tfm offset x with: tfm offset y with: 0.0.
	tfm isPureTranslation ifFalse:[
		tfm := tfm asMatrixTransform2x3.
		m44 := Matrix4x4 new.
		m44 a11: tfm a11; a12: tfm a21; a21: tfm a12; a22: tfm a22.
		ogl glMultMatrixf: m44.
	].
	tw := box width asFloat / material textureForm width asFloat.
	th := box height asFloat / material textureForm height asFloat.
	ogl glBegin: GLQuads;
		glTexCoord2f: 0.0 with: th;
		glVertex2f: box left with: box bottom;

		glTexCoord2f: tw with: th;
		glVertex2f: box right with: box bottom;

		glTexCoord2f: tw with: 0.0;
		glVertex2f: box right with: box top;

		glTexCoord2f: 0.0 with: 0.0;
		glVertex2f: box left with: box top;

	glEnd.
	ogl glPopMatrix.
	material disable: ogl.
! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 3/6/2009 02:11'!
glRenderOverlayOn: ogl
	"Display all overlapping players"
	| material viewPanelCostume cc |

	viewPanelCostume := viewPanel asPrimCostume.
	oglCache ifNil:[oglCache := IdentityDictionary new].
	ogl glColor4f: 1.0 with: 1.0 with: 1.0 with: 1.0.
	self asPrimCostume do:[:aCostume|
		aCostume == viewPanelCostume ifTrue:[
			ogl glPushMatrix.
			ogl glTranslatef: aCostume left with: aCostume top with: 0.0.
			aCostume do:[:each| self glRenderCostume: each on: ogl].
			ogl glPopMatrix.
		] ifFalse:[self glRenderCostume: aCostume on: ogl].
	].
	"Also draw the hands if any of them are software rendering"
	Processor activeProject ifNotNilDo:[:proj|
		proj hands do:[:h|
			h needsSoftwareCursor
				ifTrue:[self glRenderCostume: h asPrimCostume on: ogl]]].
	"remove obsolete costumes from texture dictionary"
	oglCache keys "need a copy" do:[:aCostume|
		cc := aCostume container.
		(cc == self costume or:[cc == viewPanelCostume]) ifFalse:[
			"it got removed"
			aCostume localCache: nil.
			material := oglCache removeKey: aCostume.
			ogl deleteTexture: material texture sha.
		].
	].


! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 7/18/2009 10:42'!
goToNextPlace: bForward
	"Generic tab-event relay.  The places pane handles tabbing."
	self ifAllowedToMove: [
		placesPane goToNextPlace: bForward
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/8/2008 14:43'!
goToPlace: aPlace

	"Send the avatar to the given QPlace (usually a landmark from the places menu)"
	self readyForEvents ifFalse: [ ^ self ].
	QActivityCapture action: #GoToPlacesMenu details: (aPlace label).
	avatar goToPlace: aPlace
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'ar 4/16/2011 17:30'!
goToPlaceName: placeName
	"Go to the place matching placeName, or notify user that it cannot be done.

	We use a general matching definition of 'place', such that:
		* The current position is found, even if it has moved within the forum.
		* Positioning is like for the UI, e.g., 'Go To' or 'Locate' at panels and such.
		* File card resources are available (just as they are through the inspector).
	Because display panels are quite dynamic, the best way to achieve all the above is to specify
	resources by something stable: E.g., QSharedDocument>>contentSummary (which has author), 
	rather than the display panel title used in the Place name, which comes from 
	QFileCard>>shortSummary (which is not unique because it doesn't have author, and which does
	have extra text such as the kindSummary that does not appear in the QSharedDocument attributes).

	While the exact match of a resource should always work (as long as the resource is present),
			some flexibility is allowed:
			+ a copy now on a panel is preferrable to a card on shelf, even if the reference was to a precise 
				name (e.g., including author) that is not part of the 'Places' designation.
			+ slighly mangled names should be allowed (e.g., so that email with QRLs have neither
				spaces nor %20.

	The current rules are:
	1. Exact match in the Places listing (e.g., #currentLandmarks).
	2. (Reserved: Exact match in People listing).
	3. Examine all file cards in space, unless isInInventory, matching identifier and preferring isOpenInOwner.
	4. Repeat the above with wildcard matching (e.g., for whitespace).

	The above assumes a correct and readable placeName, but one that has enough info in it that
	it a user probably wants it generated for them. This method is NOT currently extended to 'do what 
	I mean' (i.e., to search):
	- Repeat the above for a wildcard placeName but ignoring:
		a document placeName's leading importing username, 
		a python placeName's leading GUID,
		trailing 'kind' labels in places,
		leading python host labels in places,
	- Repeat ignoring any trailing -N suffixes just before document filename extensions.
	- Match only the host part of URL file cards.

	We cannot resolve placeNames until after we are already in the space, so a typical use pattern is
	to activate a space and then do this, possibly through a script.
	
	By contrast, the location part of a service URL denotes a TFrame named by an island global.
	QPlaces are more general."
	| places wild |
	self stopScript: #goToPlaceName:.
	self readyForEvents ifFalse: [^self statusAlert: 'Unable to go to ', placeName].
	places := avatar currentLandmarks.
	(self goToPlaceNameOrFileCard: placeName landmarks: places) ifTrue: [^self].
	wild := QWebServices whitespaceMatcher: placeName.
	(self goToPlaceNameOrFileCard: '*', wild, '*' landmarks: places) ifTrue: [^self].
	self statusAlert: placeName, ' is not in this Forum.'.
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 10/2/2009 11:36'!
goToPlaceNameOrFileCard: aString landmarks: places
	"Find aString among places, posters, and the active space's file cards, and go there.
	Answer true if found, otherwise false."
	(places detect: [:lm | aString match: lm label] ifNone: [
				places detect: [:lm | 
					lm category = 'Panels' 
						and: [(self getPanelNameFor: lm) ifNil: [false] ifNotNilDo: [:panelName |
								aString match: panelName]
							]
					] ifNone: []
			])
		ifNotNilDo: [:place | avatar goToPlace: place. ^true].

	"#findFileCard: does not answer FileCards in posters, because unlike a poster on a panel or shelf,
	they are not proper descendents of the poster.  They also don't report true for isOpenInOwner.
	Finally, since they are not rooted in the space, the poster cards won't work for takeMeToFrame.
	So... we special-case posters here"
	(self activeSpace get: #findPoster: with: aString)
		ifNotNilDo: [:poster | avatar takeMeToFrame: poster. ^true].

	(self activeSpace get: #findFileCard: with: aString)
		ifNotNilDo: [:card | avatar takeMeToFrame: card. ^true].
	^false.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/8/2008 14:43'!
goToSpaceHome
	self readyForEvents ifFalse: [ ^ self ].
	avatar goToSpaceHome.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 17:47'!
gotoHelpMovieIndexPage

	"Bounce open a web browser on the page full of how-to movies."

	| url |

	"URLs served from qwaq web site."
	url := 'http://www.openqwaq.com/support/movies'.

	self shellOpen: url
		documentDescription: 'sound setup movie'
		applicationDescription: 'Web Browser'
	! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 17:48'!
gotoSoundHelpMoofie

	"Bounce open a web browser on the sound help."

	| url |

	"URLs served from qwaq web site."
	url := 'http://www.openqwaq.com/support/sound/soundwin.html'.
	CMacOSPlatform isActivePlatform ifTrue: [
		url := 'http://www.openqwaq.com/support/sound/soundmac.html'
	].

	self shellOpen: url
		documentDescription: 'sound setup movie'
		applicationDescription: 'Web Browser'
	! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 4/27/2008 01:59'!
handleError: anError
	"Handle an otherwise unhandled error. This method is called concurrently 
	and care must be taken to schedule it properly."
	| process preStack |
	process := Processor activeProcess.
	"Capture the stack trace *before* the actual error handling to avoid nasty effects due to the currently active process being terminated before we get around to handling the signal."
	preStack := anError signalerContext stackOfSize: 1000.
	self signal: #error withArguments: {anError. process. preStack}.

	"Make sure we don't kill Morphic below"
	Project spawnNewProcessIfThisIsUI: process.
	self isSealedApplication ifTrue: [
		process terminate	"So it cannot hold stray locks."
	] ifFalse: [
		process suspend.	"Leaves open the option of debugging."
	].! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'ar 4/16/2011 18:59'!
handleError: anError inProcess: aProcess stack: preStack
	"Handle an error in the given process"
	| dlg result editor button reportButton report sendSilently |
	<on: error>

	"If we're already in an error handling situation, just die on a second error."
	(session notNil and: [ session errorTripped ]) ifTrue: [
		SystemErrorHandler default: nil.
		[aProcess terminate] on: Error do: [].
		self exitConfirming: false.
	].
	sendSilently := (anError tag = #silently).
	sendSilently ifTrue:[
		self reportIncident: 'Silently-Tagged Unexpected Error' error: anError stack: preStack.
	].

	"Cut off interaction (most things would just lock up anyway.)"
	[
		self isDemonic ifTrue: [ self myDemon stopForError: anError ].
		"In an error state, nuke all controllers since otherwise someone might
		be trying to get a sync from us which results in the dreaded 'stuck join'
		problems on forums login or orgservices join."
		orgServices ifNotNil:[
			orgServices errorTripped: true.
			orgServices controller ifNotNil:[orgServices controller destroy].
		].
		session ifNotNil: [
			session errorTripped: true.
			session allControllers do:[:cc| cc destroy].
		].
		self hideTools.
	] on: Error do: [  : ex | QLogger logWarning: 'Error in error response: ', ex ].

	 "Quit out of OpenGL when error handling,
	 which avoids a lot of entanglements.    
	 If you need to debug render errors without killing render,
	 change the class variable KillRenderOnError "
	KillRenderOnError ifTrue: [
		session ifNotNil: [
			session killRender. 
		]].

	report :=  (self fullReportStringFor: 'Unexpected Error' error: anError stack: preStack).
	QTForums reportFailure: anError report: report.

	(CUnixPlatform isActivePlatform) ifTrue: [
		"Mostly for headless support - if on unix, dump the error here."
	 	[ (CrLfFileStream forceNewFileNamed: 'OpenQwaq-', (Time totalSeconds), '.err') 
		   ifNotNilDo: [ :err |
			err ascii.
			err nextPutAll: 'Runtime Error from OpenQwaq.'; crlf.
			err nextPutAll: report; crlf.
			err close.
		]] on: Error do: [].
	].

	dlg := CUserDialog new.
	dlg label: 'Unexpected Runtime Error'.
	dlg message: (
			'The following error has occurred:\	', 
			anError description, 
			'\\', (self appName), ' must quit.',
			(sendSilently ifTrue: [''] ifFalse: ['\\You can help improve this software by reporting the error.'])
		) withCRs.
	dlg messageText allowEditing: false.
	dlg messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.

	self isSealedApplication ifFalse: [
		dlg addButton: 'Debug' value: #debug.
	].

	"If we haven't already sent the report automatically, give the user the option:"
	sendSilently ifFalse: [
		reportButton := dlg addButton: 'Report'.
		self startScript: [
			dlg disable.
			reportButton disable.  "So we know when it's been done."
			self submitErrorReport: report dialog: dlg.
		] when: {reportButton. #fire}.
	].
	dlg addButton: 'Quit' value: #abort.

	editor := CTextEditor new.
	editor textFont: (QStyles smallTextFont).
	editor contents: report.
	editor vResizing: #rigid. editor height: 120.
	editor hResizing: #rigid. editor width: 450.
	editor hScrollable: false.
	dlg costume. "force #setupCostume"
	dlg group add: editor after: dlg messageText.
	editor visible: false.

	button := dlg addButton: 'Details...'.
	self startScript:[editor visible: true] when:{button. #fire}.

	mobileProxy ifNotNil: [^self submitErrorReport: report dialog: dlg].
	result := dlg run: nil centerAt: self bounds center.
	result == #abort ifTrue: [ 
		"Before calling orderly exit, kill the errored-process, as it may hold locks etc."
		[aProcess terminate] on: Error do: [].
		self exitConfirming: false
	].

	"See above - #debug is only available in the unsealed application."
	result == #debug ifTrue: [
		"Open the debugger directly so we can sugest a position for its window,
	        enough to hopefully make it visible and not entirely hidden by the rendered view."
		| debugger |
		debugger := Debugger new process: aProcess controller: nil context: aProcess suspendedContext.
		(debugger openFullNoSuspendLabel: (anError description)) 
			position: (self topRight).
		debugger errorWasInUIProcess: false.

		"And, since we're debugging, restore interactivity to the client.
	       (This can open the window to freezes and other grief, but that's  what debugging is all about.
	       If we didn't do this, simple fix&continue at the client would be impossible.)"
		session ifNotNil: [ 
			session errorTripped: false.
			"We treat debugging as if we were still healthy, so re-allow sync service.
	             (Otherwise, after fix&continue, disabledSync participants confuse the router.)"
			session allControllers do:[:cc| cc enableSyncs].
			self onReady.
		] ifNil: [
			serviceChooser visible
				ifTrue: [ self showServiceChooser ]
				ifFalse: [ self showLogin ]
		].
	].

! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'ar 6/15/2007 10:59'!
handleLowSpace
	"A low space situation has occurred. Not sure how to react to this situation..."! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 8/1/2008 19:24'!
handleSystemEvent: aSymbol
	"Handle a system-wide event. We just resignal those here since these events
	may occur in other processes."
	self signal: aSymbol.
! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'hrs 7/2/2008 13:17'!
hasAnimatedAvatars
	"Is this user (or their organization) permitted to 
	 select an Animated Avatar without already having one?

	 If they already have one, then they can keep using it regardless of this method."
	"not now:
	^ self hasExperimentalFeatures
			or: [ self hasFeatureEntitlement: #AnimatedAvatars ]."
	^true
! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 8/21/2008 15:23'!
hasCapability: capabilityName

	"Is the given capability current available to me ?"

	"These capabilities are the atoms from which account levels in tiers of service are built;
	 a given user in a given forum has a role,
	 and from the role we determine what capabilities obtain,
	 and that set of capabilities governs the affordances we offer the user."

	"Don't allow misformed cap-names to float around in the system.
	 This will help prevent mistyping, and encourage people to look at the capabilities list
 	 before inventing new ones."
	self isSealedApplication ifFalse: [
		(self class allCapabilities includes: capabilityName) ifFalse: [
			(self class allFeatureKeys includes: capabilityName) ifFalse: [
				QLogger logInfo: 'Unknown capability or feature: ', capabilityName.
				]]].

	^ capabilities includes: (capabilityName asSymbol)! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'das 10/12/2009 10:32'!
hasChat
	"A switchable feature 'ChatSuppressed - if true,
	 the chat dialogs are turned off. "

	^ (self hasFeatureEntitlement: #PublicChat)
	
! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 5/1/2008 18:05'!
hasExperimentalFeatures

	"Are x-features currently accessible to this user?

	 (NB. will return false if called prior to the serviceManager setup,
	 i.e. basic connection to forums host, 
	 as this currently comes from user-preferences.) "

	^ preferences notNil and: [ preferences enableExperimentalFeatures ].! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 9/6/2008 15:52'!
hasFeatureEntitlement: featurePropertyName

	"Is this user entitled to use the named feature?
	True if the prevailing forums.properties value for
	 	EnabledFeatures.<FeaturePropertyName> is 'true'. 

	 Any featurePropertyName used here should appear in my #class allFeatureKeys list.

	 If a value for the property is set in the user's forums.properties, it prevails.
	 Otherwise if a value for the property is set in the organization, it prevails.
	 Otherwise the global server scope is checked.
	 Failing both, the result is false."

	| dict result fullName |

	serviceMgr ifNil: [ ^ false ].

	"Capabilities can absorb feature-properties as they move out of beta, etc."
	(self hasCapability: featurePropertyName) ifTrue: [ ^true ].

	"Have we cached this result already?"
	featureProperties at: featurePropertyName ifPresent: [ : bool | ^ bool ].

	"The remainder is the path for doing a single-property check query;
	 probably meaning we are looking up a feature that is not
	 yet registered in (class allFeatureKeys),
	 or that the entitlement check is early in login/setup. "

	fullName := ('EnabledFeatures.', featurePropertyName).
	[
		dict := serviceMgr matchProperty: fullName scopes: #(user organization tier global).
	] on: Error do: [ :ex |
		QLogger logWarning: 'Reverting to scope-by-scope feature test (', ex description, ')'.
		^ self hasFeatureEntitlementByScope: featurePropertyName
	].

	dict ifNotEmpty: [
		"We got a nonempty dictionary for the property query,  so there is a setting for the flag."
		result :=  ((dict at: fullName ifAbsent: ['false']) = 'true').
		result ifTrue: [
			QLogger logInfo: (featurePropertyName,  ' available.').
		] ifFalse: [
			QLogger logInfo: (featurePropertyName,  ' not available.').
		].
		featureProperties at: featurePropertyName put: result.
		^ result
		].

	"No value for the property anywhere.
	 Assume false - no access - and cache that assumption."
	QLogger logInfo: (featurePropertyName,  ' not available').
	featureProperties at: featurePropertyName put: false.

	^ false! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 9/6/2008 15:55'!
hasFeatureEntitlementByScope: featurePropertyName

	"Fallback version of hasFeatureEntitlement:,
	 for service providers without multiscope property query."

	| dict result fullName |

	serviceMgr ifNil: [ ^ false ].

	fullName := ('EnabledFeatures.', featurePropertyName).

	"Possibly set for the user, or for the organization, or failing that, for the server globally."
	#(user organization tier global) do: [ :scope |
		dict := [serviceMgr matchProperty: fullName scope: scope]
				on: Error do: [ :ex |
					(scope = #tier)
						ifTrue: [ Dictionary new ]   "Tier scope is new, expected to fail on older SP."
				            ifFalse:[ 
							QLogger logWarning: 'Feature-check lookup failed: ', ex description.
							^ false ]].	 "Bail on fetch error, so we don't cache anything."
		dict ifNotEmpty: [
			"We got a nonempty dictionary for the property query, 
			 so there is a setting for the flag at this scope."
			result :=  ((dict at: fullName ifAbsent: ['false']) = 'true').
			result ifTrue: [
				QLogger logInfo: (featurePropertyName,  ' available (', scope, ')').
			] ifFalse: [
				QLogger logInfo: (featurePropertyName,  ' not available (', scope, ')').
			].
			featureProperties at: featurePropertyName put: result.

			"I.e. if organization included a value, that's what we use,
	    		 or failing that (second pass through), 
			 if the global scope has a value, we use that."
			^ result
		]].

	"No value for the property anywhere.
	 Assume false - no access - and cache that assumption."
	QLogger logInfo: (featurePropertyName,  ' not available').
	featureProperties at: featurePropertyName put: false.

	^ false! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'hrs 1/4/2010 14:24'!
hasForumPagesSupport

	"Is this user entitled to work with Forum Pages?
	(Everyone can follow a Landmark Link. This is about tools to generate and send them.)
	 True if experimental features is enabled,
	 or if the forums.properties feature entitlement for ForumPages is true."

	self hasExperimentalFeatures ifTrue: [ ^ true ].

	^self hasFeatureEntitlement: #ForumPages.! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'ar 5/6/2009 15:27'!
hasInvitePower

	"Is this user entitled to use video-chat?
	 True if experimental features is enabled,
	 or if the forums.properties feature entitlement for AnyUserCanInvite is true."

	^ (self hasCapability: #invite)
		or: [ 
			(self hasCapability: #modify)
			  and: [serviceMgr notNil 
			  and: [self hasFeatureEntitlement: #AnyUserCanInvite ]]].! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/23/2008 19:02'!
hasMiramarDesktopEdition

	"Is this user (or their organization) permitted to access Miramar?

	 This is the indicator as to whether Miramar is allowed at all;
	 i.e. it is the master flag from which we decide whether
	 or not to let the user into Miramar workspaces.

	 (NB. will return false if called prior to the serviceManager setup,
	 i.e. basic connection to forums host, 
	 as this currently comes from user-preferences.) "

	^ self hasExperimentalFeatures
		or: [ self isSealedApplication not ] 	" Per JDM "
		or: [ self hasFeatureEntitlement: #DesktopEdition ].

	"Temporary, I think... this will be converted to a tier-of-service or other rule when there are such."! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/23/2008 19:02'!
hasMiramarPersonalShare

	"Is this user (or their organization) permitted to 
	 use Miramar personal sharing?

	 (NB. will return false if called prior to the serviceManager setup,
	 i.e. basic connection to forums host, 
	 as this currently comes from user-preferences.) "

	^ self hasExperimentalFeatures
		or: [ self isSealedApplication not ] 	" Per JDM"
		or: [ self hasFeatureEntitlement: #PersonalShare ].

	"Temporary, I think... this will be converted to a tier-of-service or other rule when there are such."! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'hrs 4/30/2009 11:15'!
hasOfflineUploadSupport

	"Is this user entitled to work with offline uploads?"

	"Turning this on with x-features would be confusing, because it also has to be enabled
	on the server, where x-features does not apply."

	^  self hasFeatureEntitlement: #OfflineUploads.! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'ar 6/30/2009 09:12'!
hasPersonalShareColorKey

	"Allow color keying in personal share ?"

	^self hasExperimentalFeatures
		    or: [ self hasFeatureEntitlement: #PersonalShareColorKey]! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'ar 2/11/2009 12:21'!
hasPersonalShareInput

	"Allow in-island input into personal share ?"

	^self hasExperimentalFeatures
		    or: [ self hasFeatureEntitlement: #PersonalShareInput ]! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 3/18/2009 10:44'!
hasPersonalShareVNC

	"Connect-to-VNC available? "

	^ self hasExperimentalFeatures
		or: [ self hasFeatureEntitlement: #PersonalShareVNC ].

! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'jcg 11/21/2008 19:14'!
hasPhoneSupport

	"Is this user entitled to use video-chat?
	 True if experimental features is enabled,
	 or if the forums.properties feature entitlement for WebCam is true."

	"XXXX: Disabled because we 'need' to unload QAudioPlugin for OpenAL to recognize audio-device changes... if we don't disable it we get crashes due to a bad pointer in our QPhoneUserFeedbackChannel."
	true ifTrue: [^false].

	self hasExperimentalFeatures ifTrue: [ ^ true ].

	^  self hasFeatureEntitlement: #Phone.! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'das 10/12/2009 10:29'!
hasPrivateChat
	"A switchable feature 'PrivateChatSuppressed - if true,
	 the private chat dialogs are turned off. "

	^ (self hasFeatureEntitlement: #PrivateChat)
	
! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'crl 4/15/2010 14:41'!
hasSecureVideo

	"Is this user+org entitled to use the secure video features?
	 True if experimental features is enabled (for easier dev-testing),
	 or if the forums.properties feature entitlement for SecureVideo is true."

	self hasExperimentalFeatures ifTrue: [ ^ true ].

	^self hasFeatureEntitlement: #SecureVideo! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'ar 3/27/2009 17:49'!
hasSessionRecordingSupport

	"Is this user entitled to use the session recording features?
	 True if experimental features is enabled,
	 or if the forums.properties feature entitlement for SessionRecording is true."

	self hasExperimentalFeatures ifTrue: [ ^ true ].

	^  self hasFeatureEntitlement: #SessionRecording.! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'ar 5/3/2010 11:55'!
hasSharePoint

	"Is this user entitled to use SharePoint?
	SharePoint is always enabled and not dependent on feature flag"

	^true! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'jcg 6/15/2011 00:44'!
hasWebCamSupport

	"Is this user entitled to use video-chat?
	 Check the forums.properties feature entitlement for WebCam."

	"Everyone can use unwatermarked-webcam now that we're using Free codecs."
	^  true "self hasFeatureEntitlement: #WebCam."! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'jcg 12/10/2009 13:48'!
hasWebcast

	"Is this user+org entitled to use the session broadcast features?
	 True if experimental features is enabled (for easier dev-testing),
	 or if the forums.properties feature entitlement for SessionBroadcast is true."

	self hasExperimentalFeatures ifTrue: [ ^ true ].

	"XXXXX: in 2010, remove #SessionBroadcast."
	^  (self hasFeatureEntitlement: #SessionBroadcast) 
		or: [self hasFeatureEntitlement: #Webcast].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:32'!
helpMenu
	| menu |

	menu := CMenu new.
	menu add: 'User Guide...' translateMe target: self action: #openUserGuide.		"Stub"
	self runScript: #updateHelpMenu when: {menu . #open}.

	^menu! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/11/2010 13:57'!
hideFixedPanes 
	"Suppress the fixed-position panes and other chrome."
	controlBar visible: false.
	navPanel visible: false.
	self menu visible: false.
	keyboardIcon visible: false.
	usersPane visible: false.
	placesPane visible: false.
	peoplePlacesSash visible: false.
	viewAreaSash visible: false.
	videoAreaSash visible: false.
	webcamPane visible: false; disable.
	catalogPane visible: false.
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 9/12/2008 08:47'!
hidePanels
	"Suppress all but the scene view; this is for full-zoom onto panel applications."
	
	self hideFixedPanes.
	visiblePanes := self floatingPanes select: [:fp | fp visible].
	visiblePanes do:[:vp | vp visible: false].

	self signal: #extentChanged.
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/11/2010 13:56'!
hideTools

	"Hide or disable all the UI pieces that represent controls
	 we only want available during a healthy session.
	 This disables the control bar (stripping away its button-icons) 
	 but does not remove it from the UI,
	 so that it can serve as a plain status-line area."

	self floatingPanes do: [ : p | p visible: false ].

	self signal: #notReady.
	keyboardIcon visible: false.
	posReadout visible: false.

	usersPane visible: false; disable.
	placesPane visible: false; disable.
	peoplePlacesSash visible: false; disable.
	viewAreaSash visible: false; disable.
	webcamPane visible: false; disable.
	videoAreaSash visible: false; disable.
	navPanel visible: false; disable.
	viewPanel visible: false; disable.
	self menu visible: false; disable.

	controlBar disable.
	audioSession ifNotNil: [
		audioSession deviceManager future stopCheckingForDeviceChanges.
	].


! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/14/2010 12:13'!
hideWebcamPane
	"Make the web-cam pane invisible if it is not."
	
	webcamPane visible ifTrue: [
		webcamPane visible: false; disable.
		videoAreaSash visible: false; disable.
		self layoutEverything.
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/6/2007 21:45'!
historyPane
	"Answer the historyPane of the receiver"
	<bewareOf: #historyPaneChanged>
	^self propertyValueAt: #historyPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/6/2007 21:45'!
historyPane: aValue
	"Modify the receiver's historyPane"
	^self propertyValueAt: #historyPane put: aValue with: #historyPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/2/2010 12:03'!
homeOrgName
	"The organization name I should sign into, or return the service chooser to after place-logout."
	^ (login areOptionalFieldsShowing
	    or: [ preferences isNil ]
	    or: [ login forumGoal notNil ]		"A directed login; processCmdLine or generalTeleport"
	    or: [ preferences homeOrganization isEmptyOrNil ]
	) ifTrue: [
		login organization value asString
	] ifFalse: [
		preferences homeOrganization
	]! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 5/24/2007 14:52'!
hostCache
	"Answer the hostCache of the receiver"
	<bewareOf: #hostCacheChanged>
	^self propertyValueAt: #hostCache! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 5/24/2007 14:52'!
hostCache: aValue
	"Modify the receiver's hostCache"
	^self propertyValueAt: #hostCache put: aValue with: #hostCacheChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'ar 1/7/2009 14:08'!
icmpPingerScript
	| icmpStats info socket pinger |
	icmpStats := TWindowedStatistics forWindowDuration: 60000.
	icmpStats tagName: #ICMP.
	latencyPane stopWatchingTag: #ICMP.
	latencyPane startWatching: icmpStats.
	info := activeSpace controller hostInfo.
	info address ifNil:[^self inform: 'Error: No host address for ', info hostName].
	socket := ICMPSocket new.
	socket isValid ifFalse:[^self inform: 'Error: Could not create ICMP socket'].
	icmpSocket := socket.
	pinger := [[icmpSocket ping: info address at: 250 
					do:[:trip| icmpStats add: trip] 
					until:[icmpSocket ~~ socket]
				] on: Error do:[:ex|
					icmpSocket ifNotNil:[icmpSocket destroy].
					icmpSocket := nil.
				]] forkAt: Processor highIOPriority.
	pinger name: 'ICMP Pinger'.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 1/7/2009 12:08'!
icmpSocket
	"Answer the icmpSocket of the receiver"
	<bewareOf: #icmpSocketChanged>
	^self propertyValueAt: #icmpSocket! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 1/7/2009 12:08'!
icmpSocket: aValue
	"Modify the receiver's icmpSocket"
	^self propertyValueAt: #icmpSocket put: aValue with: #icmpSocketChanged! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 9/15/2009 20:26'!
idlingMessage
	"If in reduced-rate render,  say so."
	^ self isIdlingFrames 
		ifTrue: [
			'(Idling ', (self preferences idleFramesPerSecond), ' fps)'.
		] ifFalse: [
			 ''
		].
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/18/2009 10:38'!
ifAllowedToMove: aBlock 
	"Filter utility for various driving and teleporting actions...
	 if the user is free to move, evaluate the block;
	 otherwise explain why not in a status line message."
	 
	self allowedToMove ifTrue: aBlock ifFalse: [ 
		self statusInfo: 'The leader is controlling movement.'
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 11/28/2007 15:35'!
inBenchTest
	"Answer the inBenchTest of the receiver"
	<bewareOf: #inBenchTestChanged>
	^self propertyValueAt: #inBenchTest! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 11/28/2007 15:35'!
inBenchTest: aValue
	"Modify the receiver's inBenchTest"
	^self propertyValueAt: #inBenchTest put: aValue with: #inBenchTestChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/14/2009 13:33'!
infoCache
	"Answer the infoCache of the receiver"
	<bewareOf: #infoCacheChanged>
	^self propertyValueAt: #infoCache! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/14/2009 13:33'!
infoCache: aValue
	"Modify the receiver's infoCache"
	^self propertyValueAt: #infoCache put: aValue with: #infoCacheChanged! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/7/2006 22:58'!
infoFromAvatar: aString
	"Private handler; see updateStatusRelay."
	statusLine info: aString.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/7/2006 22:59'!
infoFromSpace: aString
	"Private handler; see updateStatusRelay."
	statusLine info: aString.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'ar 7/12/2011 12:53'!
inform: aString helpUrl: helpUrlPath
	"Inform the user with a note (just like #inform:), but with an additional button that, when 
	pressed, opens a URL in a Web browser."
	| result  dlg helpUrl |

	"We used to have an 'OK' and a 'More...' button, but people weren't reading the info
	from the latter. This dialog is a big-and-important event, so there's really no reason
	to allow people to duck out from a more-info browser experience."

	"Log the activity for database use"
	QActivityCapture action: #InvalidEnvironment details: helpUrlPath. 
	"And send their machine info to us"
	self reportIncident: aString.

	helpUrl := 'http://', self serviceHost, helpUrlPath.
	dlg := CUserDialog new.
	dlg dialogLabel textEmphasis: 1.
	dlg dialogLabel value: 'Note'.
	dlg messageText contents: aString asText.
	dlg messageText allowEditing: false.
	dlg messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	dlg buttonRow removeAll.
	"dlg addButton: 'OK' value: nil."
	(dlg addButton: 'More...' value: #more)
		balloonHelp: 'Get more info from ', helpUrl, '.'.
	result := dlg run: nil centerAt: nil.
	result = #more ifTrue: [
		self shellOpen: helpUrl
			documentDescription: 'OpenQwaq information'
			applicationDescription: 'Web browser'
		].
	^result.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/4/2009 21:05'!
informIfEnteringMeeting: serviceUrl
	"If the given serviceURL appears to have a meeting in progress,
	 put up a status-alert mentioning that.  Used on login and teleport."

	(servicesPane meetingLeaderLogin: (servicesPane entryForUrl: serviceUrl)) 
	 ifNotNilDo: [ : leader |
		| place |
		leader = self userLogin ifTrue: [ ^ false ].
		place := 100@50.	"Centering doesn't work, the thing resizes as we're doing this."
		statusLine
			floatingMessage: (' Meeting in progress, \ led by ', 
							(usersPane publicNameForLogin: leader), ' ')
			at: place
			extraSeconds: (viewPanel visible ifTrue: [1] ifFalse: [8]).
		^ true
	].
	^ false! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'ar 4/16/2011 17:31'!
initServerQPZMenu

	"Add a menu item for all of the server-based Python applications if it has not already been installed."
	| appName allApps toDir fromDir |

	"always recreate the Server Applications folder"
	toDir :=(FileDirectory cacheDirectory directoryNamed:'Server Applications').
	toDir exists ifTrue:[ toDir recursiveDelete. ].
	fromDir :=(FileDirectory on: (FileDirectory pathFromURI: 'scripts/Server Applications')).
	fromDir exists ifTrue:[fromDir recursiveCopy: toDir  pathName.].
	serviceMgr ifNil:[^self].
	allApps := [serviceMgr listFiles:'/system-resources/apps/OpenQwaqPythonApps/*.qpz']on: Error do:[:ex | ex return: #()].
	allApps := allApps, ([serviceMgr listFiles:'/system-resources/apps/OpenQwaqPythonApps/*.tpz']on: Error do:[:ex | ex return: #()]).
	allApps size = 0 ifTrue:[ "check the old location as well"
		allApps := allApps, ([serviceMgr listFiles:'/system-resources/apps/QwaqApps/*.qpz']on: Error do:[:ex | ex return: #()]).
	].
	allApps size = 0 ifTrue:[ QLogger logInfo: 'No Administrator installed applications'. ].
	allApps do:[:app |
		appName := app name.
		appName := appName copyReplaceAll: '.' with: '/'.
		((appName endsWith: 'qpz') or:[(appName endsWith:'tpz')]) ifTrue:[ "is this a qpz/tpz?"
			appName := appName copyUpTo:$/. "drop .xx.xx.xx.xx.qpz"
			QLogger logInfo: 'Adding Server Application: ', appName.
			self createC3XMenu: appName.
		].
	].
		
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 6/22/2009 17:56'!
initialNameSuggestion: templateSvc

	"If the screen name was all nonsense characters, substitute 'My' "
	^ (preferences publicName withBlanksTrimmed ifEmpty: [ 'My']),  ' - ', (templateSvc name).

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 7/12/2011 12:54'!
initialize
	super initialize.

	exitTripped := false.
	doOpenGL := true.

	Transcript cr.
	[self initializeDefaultDirectories] on: Error do:[:ignore].
	self initializeMemorySettings.
	self initializeMiscBits.
	"self initializePyScripts -- no longer called here but via #checkForValidEnvironment"
	self initializeOpenGL: false.

	"Resources - get'em loaded early"
	QForms importDirectory: 'icons/Qwaq'.
	QForms importDirectory: 'icons/Qwaq/Cards'.
	QForms importDirectory: 'icons/OpenQwaq'.
	QForms readBundledFormsTable.	"Loads the Content textures manifest."
	QTextures current.	"Ensure that TTextures bundler is installed early"
	QSounds current.

	self define: #undoRedoList as: QUndoRedoList new.
	self define: #isMeetingLeader as: false.
	self define: #meetingState as: QClientMeetingState new.
	self define: #login as: QwaqLogin new.
	self define: #proxyDlg as: QProxyDialog new.
	self define: #serviceChooser as: QServiceChooser new.
	self define: #prefsPanel as: QwaqPrefsPane new.
	self define: #aboutBox as: QwaqAbout new.
	self define: #viewPanel as: QMainViewPlayer new.
	self define: #usersPane as: QUsersPane new.
	self define: #placesPane as: QPlacesTool new.
	self define: #peoplePlacesSash as: QPaneSash new.
	self define: #viewAreaSash as: QPaneSash new.
	self define: #videoAreaSash as: QPaneSash new.
	self define: #awayOverlay as: QInertPlayer new.
	self define: #avatarPane as: QAvatarChooser new.
	self define: #chatPane as: QChatPane new.
	self define: #logView as: QLogView new.
	self define: #statisticsPane as: QStatisticsPane new.
	self define: #trafficPane as: QTrafficPane new.
	self define: #latencyPane as: QLatencyPane new.
	self define: #recorderPane as: QSessionRecorderPane new.
	self define: #webcastPane as: QSessionWebcastPane new.
	self define: #captureProxy as: QSessionCaptureProxy new.
	self define: #quickReference as: QQuickReference new.
	self define: #treeInspector as: QInspector new.
	self define: #materialCatalog as: QMaterialCatalog new.
	self define: #controlBar as: QControlBar new.
	self define: #keyboardIcon as: CImageButton new.
	self define: #navPanel as: QNavPanel new.
	self define: #statusLine as: QStatusLine new.
	self define: #posReadout as: QPositionReadout new.

	self define: #documentsPane as: QLiveDocumentTool new.
	self define: #servicesPane as: QServicesPane new.
	self define: #catalogPane as: QCatalogPane new.
	self define: #historyPane as: QHistoryPane new.
	self define: #webcamPane as: QVideoFeedPane new.
	self define: #crowdPane as: QCrowdPane new.
	self define: #phonePane as: QDemoPhonePane new.
	self define: #customsPane as: QCustomsPane new.
	self define: #backButton as: CImageButton new.
	self define: #forwardButton as: CImageButton new.
	self define: #recordingIndicator as: CImageButton new.
	self define: #webcastIndicator as: CImageButton new.
	self define: #ppToggle as: CImageButton new.
	self define: #audioSession as: QAudioSession new.
	audioSession ui: self.
	audioSession priority: Processor userScriptPriority + 2.

	self define: #windowSharePane as: QPersonalShareWindowPane new.
	self define: #repoBrowser as: QRepositoryBrowserTool new.

	Transcript addDependent: self. "get log messages here"

	firstLaunch := true.
	firstArrivalPending := true.
	crashRecovery := false.
	svcMgrCache := Dictionary new.
	urlCache := Dictionary new.
	flowControlFlag := false.
	navPanel visible: false.  "Hide until end of onOpen, as the preview looks like crap."
	navPanel collapse.
	capabilities := IdentitySet new.
	featureProperties := Dictionary new.
	hostCache := Dictionary new. 
	authCache := Dictionary new.
	recordingUsers := Set new.
	broadcastingUsers := Set new.

	self label: (self appName).
	self login owner: self.
	self viewPanel owner: self.
	self serviceChooser owner: self.
	self aboutBox appName:  self appName.
	self aboutBox licenseText contents: self licenseText.
	self define: #infoCache as: (QServiceInfoCache forOwner: self).
	self define: #sqScriptEditor as: (Smalltalk at: #CDemoScriptEditor ifPresent:[:sse| sse new]).

	"We install the menu last because menu setup links panes to menu items."
	self installMenu.

	"Prepare to accept logins..."
	login loadDefaultUser.

	"We don't want to be crazy."
	sanityChecker := QwaqSanityChecker new.
	sanityChecker ui: self.

	uiLatencyStats := TWindowedStatistics forWindowDuration: 60000.
	uiLatencyStats tagName: #UI.
	latencyPane startWatching: uiLatencyStats. "watch latency by default"

	fpsStats := TWindowedStatistics forWindowDuration: 60000.
	fpsStats tagName: #Render.
	latencyPane startWatching: fpsStats. "watch latency by default"

	"Quindi logging: If true, write log information for Quindi into the clipboard.
	May also create screenshots and thus slow down performance. Enable in
	developer menu."
	quindiLogging := false.
	snapshotBool := false.
	inBenchTest := false.
	logViewCharacterLimit := 50000.
	uiStyle := ''.
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 6/7/2010 10:50'!
initializeDefaultDirectories
	"Initialize the default directories."

	FileDirectory prefsDirectory assureExistence.
	FileDirectory cacheDirectory assureExistence.
	FileDirectory logDirectory assureExistence.
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 5/17/2007 13:36'!
initializeMemorySettings
	"Initialize the memory and GC settings to be more in line with QF requirements"

	"The following settings affect the rate incremental GCs and tenuring"

		"Limit incremental GC activity to run every 40k allocations"
		Smalltalk vmParameterAt: 5 put: 40000. "allocations between GCs (default: 4000)"
		"Limit tenuring threshold to only tenure w/ > 10k survivors"
		Smalltalk vmParameterAt: 6 put: 10000. "tenuring threshold (default: 2000)"

	"These settings affect overall memory usage"

		"Only give memory back to the OS when we have more than 16MB free"
		Smalltalk vmParameterAt: 24 put: 16*1024*1024. "default 8MB"
		"Try to keep 8MB headroom at all times"
		Smalltalk vmParameterAt: 25 put: 8*1024*1024. "default 4MB"

	"These settings describe what to do when we're close to running out of free space"

		"Tell the VM that we'd rather grow than spinning in tight GC loops"
		Smalltalk gcBiasToGrow: true. "default: false"
		"Tell the VM to do a fullGC for good measure if the above growth exceeded 16MB"
		Smalltalk gcBiasToGrowLimit: 16*1024*1024. "default: 0"
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 6/4/2010 16:09'!
initializeMiscBits
	"Initialize miscellaneous bits and pieces"

	QAudioDeviceManager soundIsEnabled: true.		"Until told otherwise."
	Collection randomForPicking: nil. "make sure it's using Processor activeIsland"
	IslandWriter signalFailure: true.
	Preferences enable:#canRecordWhilePlaying.
	QDocumentType initialize.		"make sure the psuedo-mime table is fresh."

	self isSealedApplication ifTrue:[
		"Handle low space and user interrupts ourselves"
		Sensor installInterruptWatcher:[self userInterruptWatcher].
		Smalltalk installLowSpaceWatcher:[self lowSpaceWatcher].
	].
	orgSetupMutex := Mutex new.
	QActivityCapture reset.
	QFileDialog reset. "make sure it's in a sane state"
	QLogger reset.
	[QLogger logDirectory: (FileDirectory prefsDirectory directoryNamed: 'logs')] on: Error do:[:ignore].
	TObjectID resetLogs.
	TObjectID logIDs: false.
	TMessageRelay resetMaxChannelBacklog. "use default"

	"Traffic logging."
	TTrafficCounter reset.
	TTrafficCounter current logger: 
		(QRevolvingLogger baseName: 'traffic' count: 6 size: 32000).
		"6 logs, 32k each, gives about 10-15 minutes worth per log, so an hour or so recorded."

	self startScript: #onFileDialogDone when: { QFileDialog . #fileDialogDone } .
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 9/17/2008 12:46'!
initializeOpenGL: bStereo

	"Initialize OpenGL if necessary. By default we set the OGL capabilities that we require
	and those that are optional. Subclasses can override this method to, for example, specify
	stereo rendering"
	| aa |
	stereoOn := bStereo.
	OpenGL minRendererFlags: 
		0 "1 - B3DSoftwareRenderer" + 2 "B3DHardwareRenderer" + 4 "B3DStencilBuffer".

	aa := preferences ifNil: [ true ] ifNotNil: [ preferences doAntialiasing ].
	OpenGL maxRendererFlags: 
		(stereoOn ifTrue:[
			0 "1 - B3DSoftwareRenderer" + 
			2 "B3DHardwareRenderer" + 
			4 "B3DStencilBuffer" + 
			16 "Stereo".
		] ifFalse:[
			0 "1 - B3DSoftwareRenderer" + 
			2 "B3DHardwareRenderer" + 
			4 "B3DStencilBuffer" + 
			(aa ifTrue: [8] ifFalse: [0])  "Antialiasing" .
		]).

	OpenGL disablePartialTextureUploads: false.
! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'osar 4/17/2011 20:01'!
initializePyScript
	| dllPath |
	"Compute the dllPath, e.g., the path to shared libraries 
	used by Python on the current platform."
	Smalltalk platformName caseOf:{
		['Mac OS'] -> [
			dllPath := Smalltalk getSystemAttribute:0.
			"Sanity check in case the VM name changes"
			(dllPath endsWith: '/Contents/MacOS/OpenQwaq') ifTrue:[
				dllPath := dllPath allButLast: '/MacOS/OpenQwaq' size.
				dllPath := dllPath, '/Frameworks/lib-dynload'.
			] ifFalse:[self error:'Unexpected VM Name: ', dllPath].
		].
		['unix'] -> [dllPath := '/usr/lib/python2.4/lib-dynload'.
					((FileDirectory default directoryExists: dllPath)
					and: [(FileDirectory on: dllPath) fileExists: 'binascii.so']) ifFalse:
						[self error: dllPath, ' does not exist or is empty']].
	} otherwise:[
		"Assume extra libraries are co-located with VM binaries"
		dllPath := Smalltalk vmPath.
	].

	[Smalltalk at: #Python ifPresent:[:pyBridge|
		"Ensure that the Python API Library is properly installed."
		self installDefaultPyLib.
		"Set the home directory to OpenQwaq install location"
		pyBridge pythonHome: FileDirectory default pathName.
		pyBridge initialize.
		"Set up the search paths:"
		pyBridge searchPaths: {
			dllPath.	"for addl. shared libraries"
			FileDirectory pathFromURI: 'scripts/lib'. "Python modules for OpenQwaq"
			FileDirectory cacheDirectory pathFromURI: 'OpenQwaqAPI/', QwaqVersion current version.   "Python API location"     
			FileDirectory cacheDirectory pathFromURI: 'OpenQwaqAPI/', QwaqVersion current version,'/Teleplace'.   "Python API location"     
			FileDirectory cacheDirectory pathName.
		}.
		"Trace information please"
		pyBridge reloadModule: 'Teleplace.Utils.Trace'.

		"A decent reload() please"
		"XXXX: This is currently broken it gets us a 'max recursion depth exceeded'
		pyBridge useReloadFromModule: 'Teleplace.Utils.Reload'."

		"Install the bridge's send() function"
		(pyBridge loadModule: 'Teleplace.Proxy.Proxy')
			set: 'send' to: ((pyBridge loadModule: 'Squeak') get: 'send').

		"Some platforms work this differently by default, so give Python
	       the same cwd as we assume for Squeak"
		(pyBridge loadModule: 'os') call: 'chdir' with: FileDirectory default pathName.

	]] ifError: [ : msg |
		QLogger logError: msg  from: self.
	]! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'ar 4/17/2011 20:14'!
installDefaultPyLib
	"If there is no OpenQwaq Python library directory in the cache, or it is a different version than this client, copy it from the main folder. This can be overridden by a server install. "
	| toPath fd versionFile doCopy oldVersion|

	doCopy := false. 
	toPath :=  FileDirectory cacheDirectory pathFromURI:'OpenQwaqAPI/', QwaqVersion current version,'/Teleplace'.
	fd := FileDirectory on: toPath.
	fd exists ifFalse:[
		QLogger logInfo: 'OpenQwaq Python API Library does not exist.'. 
		doCopy := true.] "there is no such directory"
	ifTrue:[ 
		versionFile := [FileStream readOnlyFileNamed: 
			(FileDirectory cacheDirectory pathFromURI:'OpenQwaqAPI/', QwaqVersion current version,'/OpenQwaqAPI-version').
		]on: Error do:[:ex |ex return: nil].
		versionFile ifNotNil:[
			[oldVersion := versionFile contents]ensure:[versionFile close].
			oldVersion = QwaqVersion current version ifFalse:[
				QLogger logInfo: 'OpenQwaq Python API Library version does not match client.'. 
				doCopy := true].
		]ifNil:[
			QLogger logInfo: 'OpenQwaq Python API Library version does not match client.'. 
			doCopy:= true
		].
		
	]. "directory exists - is it a different version?"
	doCopy ifTrue:[
		(FileDirectory on: (FileDirectory pathFromURI: 'scripts/Teleplace'))recursiveCopy: toPath.
		QLogger logInfo: 'Copying OpenQwaq Python API Library to cache.'. 
		versionFile := FileStream forceNewFileNamed:
						(FileDirectory cacheDirectory pathFromURI: 'OpenQwaqAPI/', QwaqVersion current version,'/OpenQwaqAPI-version').
		[versionFile nextPutAll: QwaqVersion current version.] ensure:[versionFile close].
	].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/22/2008 14:49'!
installMenu
	"Create and install the menu"
	self menu ifNotNil:[self menu destroy].
	self menu: CMenuBar new.
	self menu floating: false.
	self menu borderWidth: 1; borderStyle: #raised.
	self menu hResizing: #spaceFill.
	self addSubMenusTo: self menu.

	QStyles installMenuFontInMenu: self menu.! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'ar 4/16/2011 17:32'!
installServerPyLib
	"Install a server-based Python library.
	If the library is already installed, we compare the existing version with
	the one from the server and only install it if the server install has a higher number.
	If no library is installed, we need to copy the current library from.
	Note that this will return false if there is no server library, or if we are unable to load the data
	for some reason. If we return false, then we will use the normal installed library that is included
	with the client install. "
	| data versionedName versionFile oldVersion libPath libVersions unzipper fd |

	serviceMgr ifNil:[self installDefaultPyLib. ^false].
	libVersions := [serviceMgr listFiles:'/system-resources/apps/OpenQwaqAPI*.zip'] 
		on: Error do:[:ex| ex return: #()].
	libVersions size = 1 ifFalse:[self installDefaultPyLib. ^false]. "not installed"
	versionedName := libVersions first name.
	libPath := '/system-resources/apps/', versionedName.
	versionFile := [FileStream readOnlyFileNamed:
						(FileDirectory cacheDirectory pathFromURI: 'OpenQwaqAPI/', QwaqVersion current version,'/OpenQwaqAPI-version')
					] on: Error do:[:ex| ex return: nil].
	versionFile ifNotNil:[
		[oldVersion := versionFile contents] ensure:[versionFile close].
	].
	QLogger logInfo: 'Checking API Library'.
	QLogger logInfo: 'Current version ', oldVersion.
	QLogger logInfo: 'Server version ', versionedName.
	oldVersion = versionedName ifFalse:[
		QLogger logInfo: 'Downloading: ', versionedName.
		CProgressBar 
				displayProgress:  'Downloading API Library'
				at: self dialogCenter from: 0.0 to: 1.0 
				during:[:bar| 
					data := serviceMgr get: libPath informing: bar.
				].
		data ifNil:[self installDefaultPyLib.  ^false]. 
		QLogger logInfo: 'Installing ', versionedName.

		unzipper := ZipArchive new.
		unzipper readFrom: (ReadStream on:data).
		fd := FileDirectory on:(FileDirectory cacheDirectory pathFromURI: 'OpenQwaqAPI/', QwaqVersion current version).
		unzipper extractAllTo: fd informing: nil overwrite: true.

		versionFile := FileStream forceNewFileNamed:
			(FileDirectory cacheDirectory 
				pathFromURI: 'OpenQwaqAPI/', QwaqVersion current version,'/OpenQwaqAPI-version').
		[versionFile nextPutAll: versionedName] ensure:[versionFile close].
	].
	^true! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'ar 4/16/2011 17:32'!
installServerQPZ: appName
	"Install a server-based QPZ with the given appName (Module.AppName).
	If the application is already installed, we compare the existing version with
	the one from the server and only install it if they are different."
	| data versionedName versionFile versionFileName oldVersion appPath appVersions qpzPath |

	appPath := appName copyReplaceAll: '.' with: '/'.
	appPath := appPath copyReplaceAll: 'OpenQwaq/' with:'OpenQwaqPythonApps/'.
	appVersions := [serviceMgr listFiles:'/system-resources/apps/', appPath, '*.qpz'] 
		on: Error do:[:ex| ex return: #()].
	appVersions size = 1 ifFalse:[
		appVersions := [serviceMgr listFiles:'/system-resources/apps/', appPath, '*.tpz'] 
			on: Error do:[:ex| ex return: #()].
	].
	appVersions size = 1 ifFalse:[
		appPath := appPath copyReplaceAll: 'OpenQwaqPythonApps' with: 'QwaqApps'.
		appVersions := [serviceMgr listFiles:'/system-resources/apps/', appPath, '*.qpz'] 
			on: Error do:[:ex| ex return: #()].
		appVersions size = 1 ifFalse:[ ^false]. "not installed"
	].
	versionedName := appVersions first name.
	versionFileName := 'OpenQwaqPythonApps/', (appName copyAfterLast:$.), '-version'.
	qpzPath := '/system-resources/apps/', (appPath copyUpToLast: $/),'/', versionedName.
	versionFile := [FileStream readOnlyFileNamed:
						(FileDirectory cacheDirectory pathFromURI: versionFileName)
					] on: Error do:[:ex| ex return: nil].
	versionFile ifNotNil:[
		[oldVersion := versionFile contents] ensure:[versionFile close].
	].
	QLogger logInfo: 'Checking ', appName.
	QLogger logInfo: 'Current version ', oldVersion.
	QLogger logInfo: 'Server version ', versionedName.
	oldVersion = versionedName ifFalse:[
		QLogger logInfo: 'Downloading: ', versionedName.
		CProgressBar 
				displayProgress:  'Downloading ', appName
				at: self dialogCenter from: 0.0 to: 1.0 
				during:[:bar| 
					data := serviceMgr get: qpzPath informing: bar.
				].
		data ifNil:[^false].
		QLogger logInfo: 'Installing ', versionedName.
		QInstallPython installQPZ: data in: 'OpenQwaqPythonApps'.
		versionFile := FileStream forceNewFileNamed:
			(FileDirectory cacheDirectory pathFromURI: versionFileName).
		[versionFile nextPutAll: versionedName] ensure:[versionFile close].
	].
	^true! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 11/27/2009 20:49'!
installServiceManager: manager

	"Set the service manager (or clear it to nil) in all subpanes."
	
	serviceMgr := manager.
	servicesPane serviceManager: manager.
	serviceChooser serviceManager: manager.
	historyPane historyList contents: #().	"Reset just in case; it contains far refs."
	historyPane serviceManager: manager.
	usersPane serviceManager: manager.
	documentsPane serviceManager: manager.
	login serviceManager: manager.


! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'das 8/28/2009 15:28'!
installedAppsDirectory
	"The directory for installed applications."

	^FileDirectory on: (FileDirectory cacheDirectory pathFromURI: 'Installed Applications').! !

!QwaqParticipantUI methodsFor: 'Python' stamp: 'das 8/28/2009 15:28'!
installedServerDirectory
	"The directory for installed applications."

	^FileDirectory on: (FileDirectory cacheDirectory pathFromURI: 'Server Applications').! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'ar 4/16/2011 17:32'!
instantiateTemplate: templateSvc

	"Save the given service under a new name and activate it."
	| newName newService response |

	newName := CDialog request:'', templateSvc name, ' is a template.\\You can create a new Forum by giving it a name.\\New Forum Name:' withCRs
				initialAnswer: (self initialNameSuggestion: templateSvc).

	[(serviceMgr listServices detect:[:svc| svc name = newName] ifNone:[nil]) isNil]
	 whileFalse: [
		newName := CDialog 
			request: ('', newName, ' already exists.\\New Forum Name:' withCRs)
			initialAnswer: newName.
	].
	newName ifNil: [ 
		QActivityCapture action: #InstantiateTemplateAbandoned details: templateSvc name.
		CProgressCancel signal
	].

	QActivityCapture action: #InstantiateTemplate details: templateSvc name.

	"Clone newService"
	newService := QServiceRecord new copyFrom:templateSvc.
	newService name: newName.
	newService id: TObjectID new.
	newService baseUrl: templateSvc baseUrl.

	"Clone the island records."
	newService islands: (templateSvc islands collect:[:oldRecord|
		| newRecord | 
		newRecord := QIslandRecord forService: newService.
		newRecord name: oldRecord name.
		newRecord id: TObjectID new.
		newRecord makerName: oldRecord makerName.
		newRecord]).
	newService entryPostcard: templateSvc entryPostcard.
	newService isTemplate: false. "mark as non-template"

	"Make it official in the service manager."
	response := serviceMgr createService: newService.
	response name == #ok ifFalse:[
		self error: response contentString. 
	].

	"Copy the server-side resources from the template to the new forum."
	self copyResources: templateSvc to: newService informing: [:v|].

	activeSpace future signal: #infoMessage with: 'Created new Forum: ', newName.
	"Now we have a service that is -not- a template.  
	 Activate it, using the template as a fallback source of snapshot data."

	^ self activateService: newService url fromTemplate: templateSvc usingIslandFiles: nil	! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 2/19/2009 12:24'!
interceptBounds: aRect
	"Window has been resized... if the new size is not usable we override."
	| ext min |
	ext := aRect extent.
	min := self class minDisplaySize.
	((ext x < min x) or: [ ext y < min y ])
	ifTrue: [
		ext := (ext x max: (min x))@(ext y max: (min y)).
		self setWindowDisplayBounds: (aRect origin extent: ext).
		^ true
	] ifFalse: [
		^ false
	].! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/18/2008 22:09'!
isAdministrator

	"Does the user have adminstrative power in the current forum?
	 This is currently co-incident with hasAdminPowers (the org-level question)
	 but we may in future have situations where a user has admin powers
 	 only in some forums.   Such users would not be able to make new forums
	 but would be able to administer their own, for example."

	^ capabilities includes: #administer! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'bgf 5/18/2008 13:08'!
isAudioSuspended

	"Is the audio session temporarily suspended?
	 (Answers true if there is no audio session.)"
	^ self audioSession 
		ifNotNilDo: [ : as | as isSuspended ]
		ifNil: [ true ]! !

!QwaqParticipantUI methodsFor: 'testing' stamp: 'bgf 2/2/2008 01:34'!
isAway
	"Away is a user-controlled state, 
	 advertised to the world in an Org-Services notice,
	 that means we have an avatar present (and so can see into the world)
	 but that we are not participating in voice chat (cannot speak nor hear)."

	^ controlBar awayButton selected! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 11/15/2009 01:10'!
isDemon
	"Are we possessed by infernal creatures at this time?"
	"#isDemon asks whether the avatar is a demon, and #isDemonic asks whether it is currently acting demonically (i.e. not idle)."
	^ myDemon notNil ! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'jcg 10/17/2009 08:55'!
isDemonic
	"Are we possessed by infernal creatures at this time?"
	"#isDemon asks whether the avatar is a demon, and #isDemonic asks whether it is currently acting demonically (i.e. not idle)."
	^ myDemon notNil and: [myDemon isDemonic].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 11/12/2009 12:27'!
isEditingTemplate
	"Is my user currently engaged in editing the forum scene structure ?"
	^ treeInspector isEditingTemplate! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 9/15/2009 20:26'!
isIdlingFrames
	"True if I'm in deliberate slow render because the user is idle."
	^ (session notNil and: [ session idling] and: [ preferences idleFramesPerSecond < 1000]) 
	! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 11/24/2009 17:16'!
isMeetingFollower
	"Am I in a meeting, in which I am -not- the leader ?"
	^ self isMeetingInProgress and: [ self isMeetingLeader not ]! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/2/2009 11:30'!
isMeetingInProgress
	"Boolean - true if there is any leader running a meeting here. "
	
	^ meetingState leaderLogin notNil! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/25/2009 11:02'!
isMeetingLeader
	"True if my client is currently the designated meeting-leader of the room I'm in."
	<bewareOf: #isMeetingLeaderChanged>
	^self propertyValueAt: #isMeetingLeader! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/25/2009 11:02'!
isMeetingLeader: aValue
	"This is private.  
	 The org-services meeting-api's determine when to change this,
	 and the changes are sent here from the (people-pane) signal relays.
	 We mark the flag here so we always have a 'what my client last believed' flag
	 and can deal with transitions properly 
 	 when we update the client-state in sync with the agreed org-services state."

	^self propertyValueAt: #isMeetingLeader put: aValue with: #isMeetingLeaderChanged! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'bgf 1/15/2008 12:59'!
isMicrophoneOn
	"Is sound input on, so far as we know?"
	^ self audioSession notNil and: [self audioSession isRecording]! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/26/2007 19:24'!
isMute
	"Is sound suppressed?"
	self audioSession ifNotNilDo: [:audio | ^ audio isMute].
	^ false! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/26/2007 19:23'!
isMute: aBool
	"Suppress sound."
	self audioSession ifNotNilDo: [ :aud|aud isMute: aBool].! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 11/21/2009 10:54'!
isPoint: aPoint inPlayer: aPlayer ignoring: players
	"Is the point in the main viewing area, 
	 rather than in some obscuring floating pane?
	 (The ignore-Players are not considered as candidates to intercept.	
 	  This is useful when there you have a pane up that you know is the 'drag' icon.)"
	| aCostume iCostumes |

	aPlayer ifNil: [ ^ false ].
	(self containsPoint: aPoint) ifFalse:[^ false].
	
	iCostumes := players collect: [ : p | p costume ].
	costume contents size to: 1 by: -1 do:[:i|
		aCostume := costume contents at: i.
		(iCostumes includes: aCostume) ifFalse: [
			((aCostume fullContainsPoint: aPoint) and: [ aCostume visible ]) 
			 ifTrue: [
				^aCostume player = aPlayer
			]].
	].
	^false

! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 3/5/2009 03:43'!
isPointInMainView: aPoint
	"Is the point in the main viewing area, 
	 rather than in some obscuring floating pane?"
	^ self isPointInMainView: aPoint ignoring: #()
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 11/21/2009 10:54'!
isPointInMainView: aPoint ignoring: players
	"Is the point in the main viewing area, 
	 rather than in some obscuring floating pane?
	 (The ignore-Players are not considered as candidates to intercept.	
 	  This is useful when there you have a pane up that you know is the 'drag' icon.)"

	mainView ifNil: [ ^ false ].
	
	^ self isPoint: aPoint inPlayer: viewPanel ignoring: players! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/19/2008 12:36'!
isReadonlyUser

	"Indicates a user who cannot do -anything- to interact with stuff in the space.
	 This is (currently) the only implementation of action-access-control
	 for inspace objects and delivery of events into the space.
	 When the event-handling regime and all the in-space objects are enhanced
	 to honor user roles in there, then this will unfold into more specific capabilities."

	^ (self hasCapability: #modify) not
! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/19/2009 14:18'!
isReusableGuestUser

	"Hacky legacy definition.
	 these old-style guest users are users whose logins start with 'guest'.

	 Controls whether
	 - preferences are cleared on login
	 - all autosaves are ignored.

	But since these are for demos, other capabilites of isMember are allowed.
	This is -not- a role.  Roles can vary across forums and orgs.  A guest is only a guest.

	(In particular this is not 'isReadonlyUser' because these users traditionally
	 could participate.  Also, different because the software always
	 reset the user preferences of these users at login,
	 something we don't want to do for new-style 'come see my forum but don't touch' guests.)
	
	"

	^ (self userLogin asLowercase beginsWith: 'guest')
! !

!QwaqParticipantUI methodsFor: 'testing' stamp: 'jdm 12/10/2007 06:46'!
isSealedApplication
	"Answer whether we're currently running as a sealed application."
	^ QwaqVersion current isSealedApplication
! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'jcg 4/8/2009 23:37'!
isVideoSessionEnabled
	"Is video activity currently enabled?
       True if the preference is on, and the decoder hasn't been transiently disabled."
	^ preferences isNil or: [ 
		 preferences videoDisabled not 
		 and: [ 
			[session videoSession isSuspended not] on: Error do: [
				"still bootstrapping - it will come up enabled" true]
		]].! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 3/25/2010 12:26'!
islandAdded: anIsland postcard: postcard

	"An island has been added to our session."
	self startScript: #onIslandSaved:postcard:author: 
		withArguments: {anIsland . postcard }  
	      when: { anIsland . #saved }.

	self startScript: #onIslandAction:url:author:
		when: { anIsland . #action:url:author: }.

	"Assume clean on entry."
	(servicesPane entryForPostcard: postcard) ifNotNilDo: [ : entry |
		entry lastSaveTime: (Time utcSeconds).
	].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'bgf 6/9/2009 16:06'!
islandSideError: error fromController: aController
	"An error in t-island  message processing.
	 Disable island message processing globally,
	 so the client doesn't give the illusion of still being connected."

	self isSealedApplication ifTrue: [
		session ifNotNil: [ session allControllers do: [ : c | c disableMessages: true ] ] ].! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 3/26/2010 16:10'!
islandsInService: service

	"Return a list of currently active Islands in the given service."
	| result |
	result := OrderedCollection new.
	session ifNotNil: [
		service islands do: [:record |
			((session islandControllers ifNil: [ #() ])
					detect: [ :any | any sessionID = record id ] ifNone: [ nil ]) 
				ifNotNilDo:[ :cc | result add: cc island ].
		]].
	^ result.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 4/1/2010 16:51'!
keyDriveRedirect: player

	"Make the given player send me all its key activity so I can make keydriving work then."

	 viewPanel startScript: #onKeyDown: when: {player . #keyDown}.
	 viewPanel startScript: #onKeyStroke: when: {player . #keyStroke}.
	 viewPanel startScript: #onKeyUp: when: {player . #keyUp}.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/17/2006 12:42'!
keyboardIcon
	"Answer the keyboardIcon of the receiver"
	<bewareOf: #keyboardIconChanged>
	^self propertyValueAt: #keyboardIcon! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/17/2006 12:42'!
keyboardIcon: aValue
	"Modify the receiver's keyboardIcon"
	^self propertyValueAt: #keyboardIcon put: aValue with: #keyboardIconChanged! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'jcg 5/11/2010 15:50'!
keydownAcceleratorCases

	"CaseOf  case list for acceleration...
	 maps keyDown keyValue to an action block.
	 These are accelerators... intercepted at the top level of event input."

	Smalltalk platformName = 'Mac OS' ifTrue: [
		^ { 
			"F1 - toggle microphone"
			[122] -> [controlBar microphoneButtonFired: controlBar microphone selected not ].
		}].

	Smalltalk platformName = 'Win32' ifTrue: [
		^ { 
			"F1 - toggle microphone"
			[112] -> [controlBar microphoneButtonFired: controlBar microphone selected not ].
		}].

	^#()! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
lastBounds
	"Answer the lastBounds of the receiver"
	<bewareOf: #lastBoundsChanged>
	^self propertyValueAt: #lastBounds! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
lastBounds: aValue
	"Modify the receiver's lastBounds"
	^self propertyValueAt: #lastBounds put: aValue with: #lastBoundsChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/2/2008 17:16'!
latencyPane
	"Answer the latencyPane of the receiver"
	<bewareOf: #latencyPaneChanged>
	^self propertyValueAt: #latencyPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/2/2008 17:16'!
latencyPane: aValue
	"Modify the receiver's latencyPane"
	^self propertyValueAt: #latencyPane put: aValue with: #latencyPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 4/25/2007 18:56'!
latencyStatistics
	"Access the statistics gathered by the media controller."
	self session ifNil: [^nil].
	self session mediaController ifNil: [^nil].
	^ self session mediaController latencyStatistics! !

!QwaqParticipantUI methodsFor: 'pshare-utils' stamp: 'bgf 3/5/2009 02:51'!
launchPersonalVNC: embeddableApp
	"Launch an embeddable app for the receiver"
	| doc  vncDialog vncHost vncPort vncPass |
	<on: launchPersonalVNC>
	self ifTweak:[
		doc := embeddableApp app get: #document.
		vncDialog := QPersonalVNCDialog new.
		vncDialog hostInput value: (doc get: #vncHost).
		vncDialog portInput value: (doc get: #vncPort).
		vncDialog position: (self dialogCenter).
		self waitTick.
		self add: vncDialog.
		vncDialog waitUntil: #closed.
		vncDialog value == true ifFalse:[^embeddableApp cancelLaunch].
		[vncHost := vncDialog hostInput contents asString.
		vncPort := vncDialog portInput contents asString asNumber.
		vncPass := vncDialog passInput contents asString.
		embeddableApp connectTo: vncHost port: vncPort password: vncPass.
		doc isValidFarRef ifTrue: [
			doc future vncHost: vncHost port: vncPort.
		].
		] on: Error do:[:ex|
			embeddableApp cancelLaunch.
			CDialog inform: 'VNC connection failed:\' withCRs, ex description.
		].
	] ifNot:[
		self signal: #launchPersonalVNC with: embeddableApp.
	].! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/23/2010 12:24'!
layoutEverything
	| box viewHt viewTop viewWidth vidHeight |
	"Lay out all the parts in the current extent."
	box := self bounds.
	self menu topLeft: box topLeft-1.
	self menu width: box width+2.
	self menu fullBounds.
	controlBar width: box width+2.
	controlBar bottomLeft: (box bottomLeft + 
							(self isSealedApplication ifTrue: [-1@1] ifFalse: [-1@2])).
 				"The sealed app clips pixels off the bottom. 
				 I've no idea why (both platforms, so not just the vm-prims).	
				 Anyway, submerge the control bar an extra pixel when unsealed to mimic that."
	controlBar fullBounds.
	viewTop := (self menu font pointSize * 2) + 3.	"Echoing the menu bar layout."
	viewWidth := box width.
	viewHt :=  (controlBar top - viewTop).
	controlBar visible ifTrue: [
		"If the control bar is visible, we're not in Zoom-100.
	       If the menu bar is visible, we're in-session, otherwise login."
		| colWidth |
		colWidth := ppToggle selected 
						ifTrue: [ viewAreaSash memo ]
						ifFalse: [ 0 ].
		vidHeight := videoAreaSash memo.
		viewPanel topLeft: (0@viewTop);
					extent: (viewWidth - colWidth + 1) @ (viewHt - vidHeight);
				  	fullBounds.
		usersPane  visible: (self menu visible and: [colWidth > 0]);
					top: (viewTop).
		placesPane visible: (self menu visible and: [colWidth > 0]).
		peoplePlacesSash 
			     visible: ((ppToggle selected) and: [usersPane visible]);
			     right: (viewWidth+1).
		viewAreaSash 
			     height: viewHt;
			     visible: ((ppToggle selected) and: [usersPane visible]);
			     top: viewTop.
		videoAreaSash 
			     left: -1; width: viewWidth+3-colWidth.
		catalogPane 
			     visible: (self menu visible and: [ colWidth > 0] and: [ controlBar inventory selected ]);
			     height: (viewHt+1);
			     width: (colWidth);
			     topLeft:  (viewWidth+1) @ viewTop.
		self viewAreaRight: (width - colWidth).
		self viewAreaBottom: (controlBar top - vidHeight - viewTop ).
		self peoplePlacesSashTop: ((self viewHeight) * peoplePlacesSash memo) rounded.
	] ifFalse: [
		viewPanel topLeft: 0@0;
				   extent: (box extent).
	].
	viewPanel fullBounds.
	posReadout 
		bottom: (viewPanel bottom -2);
		left: (viewPanel left + 2).

	awayOverlay bounds: (viewPanel bounds).
	"The collapsed navigation panel touches the control-bar view-panel boundary."
	navPanel bottom: (viewPanel bottom - 2 + (navPanel collapsedHeight//3)).
	navPanel left: (viewPanel center x - (navPanel width/2)).
	sqScriptEditor ifNotNil:[
		sqScriptEditor extent: box width @ (box height // 4 min: 180).
		sqScriptEditor bottomLeft: controlBar topLeft.
	].
	serviceChooser visible ifTrue: [
		serviceChooser bounds: (box).
	].

	"Position the floating panes and other accessories."
	self floatingPanes do: [ :p | p alignToOwner].
	ppToggle topRight: (self width - 1) @ (viewTop - ppToggle height / 2 + 1) truncated.
	backButton topLeft: (self width - 82) @ (viewTop - backButton height /2 + 1) truncated.
	forwardButton topLeft: (backButton topRight +(1@0)).

	recordingIndicator topRight: (backButton topLeft -(6@0)).
	webcastIndicator topRight: (recordingIndicator topLeft -(7@0)).

	keyboardIcon topRight: (webcastIndicator left-7) @ 1.
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/3/2009 11:41'!
leaderBackFromAway
	"I'm the meeting leader, and I went 'away' for a moment.
	 Restore the state I left behind.  (We stash it in private properties.)"

	meetingState setAudienceIsMuted: (
					(self propertyValueAt: #saveAudienceIsMuted) ifNil: [ false ]).! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/3/2009 11:42'!
leaderGoingAway
	"I'm the meeting leader, and I'm going 'away' for a moment.
	 Save some state for when I come back, 
	 and release rules we don't want in force while I'm away (e.g. release chat)."

	self propertyValueAt: #saveAudienceIsMuted put: (meetingState audienceIsMuted).
	meetingState setAudienceIsMuted: false.
	

! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/12/2009 15:33'!
leaderPublicName

	^ meetingState leaderLogin ifNotNilDo: [ : leaderLogin |
		^ usersPane publicNameForLogin: leaderLogin
	] ifNil: [
		'No one'
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 4/20/2011 10:02'!
licenseText
	"Answer the license text associated with this app. Used by the aboutBox"
	^'Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.'! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 1/6/2010 16:56'!
linkIcon: aControlIcon toTool:aPanel

	"Make the given menu item checkmark track whether or not the given tool is visible."

	aControlIcon selected: (aPanel visible).
	self runScript: #toggleIcon:forTool: 
		withArguments: {aControlIcon . aPanel} 
		when: {aPanel . #visibleChanged}.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 2/24/2009 15:09'!
linkItem:aMenuItem toTool:aPanel

	"Make the given menu item checkmark track whether or not the given tool is visible."

	aMenuItem checked: (aPanel visible).
	self runScript: #toggleItem:forTool: withArguments: {aMenuItem . aPanel} when: {aPanel . #visibleChanged}.! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 8/16/2007 12:17'!
loadAuthInfo: realm
	"Load cached authentication info if available"
	| file data xmlDoc |
	file := [FileDirectory prefsDirectory readOnlyFileNamed: 'authInfo.xml'] 
			on: Error do:[:ex| ex return: nil].
	file ifNil:[^nil].
	[file converter: Latin1TextConverter new.
	data := file contents] ensure:[file close].
	[
		xmlDoc := (XMLDOMParser on: data readStream) nextEntity.
		(xmlDoc name == #authCache) ifFalse:[^nil].
		xmlDoc elementsDo:[:node|
			node name == #authInfo ifTrue:[
				(node attributeAt: 'realm') = realm 
					ifTrue:[^node contentString]]].
	] on: Error do:[:nothing|].
	^nil! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 2/7/2009 10:16'!
localPreferences
	"My preferences, or a temporary preference-object for convenience.
	 This result should not be held longer than is needed for a specific use.
	 Used by the login tools to get set up 
	 before the service-manager + preferences initialization is propagated,
	 to give them access to the locally-stored records."

	^ preferences ifNil: [
		QUserPreferences localForUser: (self userLogin) inClient: self
		].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 4/17/2007 11:03'!
logFile
	"Answer the logFile of the receiver"
	<bewareOf: #logFileChanged>
	^self propertyValueAt: #logFile! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 4/17/2007 11:03'!
logFile: aValue
	"Modify the receiver's logFile"
	^self propertyValueAt: #logFile put: aValue with: #logFileChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/7/2006 15:08'!
logView
	"Answer the logView of the receiver"
	<bewareOf: #logViewChanged>
	^self propertyValueAt: #logView! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/7/2006 15:08'!
logView: aValue
	"Modify the receiver's logView"
	^self propertyValueAt: #logView put: aValue with: #logViewChanged! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'das 3/23/2009 15:56'!
logViewCharacterLimit
	"Answer the logViewCharacterLimit of the receiver"
	<bewareOf: #logViewCharacterLimitChanged>
	^self propertyValueAt: #logViewCharacterLimit! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'das 3/23/2009 15:56'!
logViewCharacterLimit: aValue
	"Modify the receiver's logViewCharacterLimit"
	^self propertyValueAt: #logViewCharacterLimit put: aValue with: #logViewCharacterLimitChanged! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 12/6/2009 13:29'!
loggingOut: bExit
	"The user is leaving forums.
	 OpenGL rendering may have already been killed.
	 If bExit is true, the user is exiting to desktop, if false, back to the login screen.
	 After this method returns, the session will be destroyed.
	 Errors arising here will be ignored, and exit WILL proceed regardless."

	orgServices ifNotNil: [ orgServices beMeetingLeader: false ].
	bExit ifFalse: [ self isMeetingLeader: false ].

	(self propertyValueAt: #documentManager) ifNotNilDo: [ : dm | dm discardPrefetches ].

	self useQuickTimeouts.
	self statisticsPane closeSession.		"Grab the tail of the activity log."
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
login
	"Answer the login of the receiver"
	<bewareOf: #loginChanged>
	^self propertyValueAt: #login! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
login: aValue
	"Modify the receiver's login"
	^self propertyValueAt: #login put: aValue with: #loginChanged! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'hrs 11/11/2008 11:51'!
loginEntryComplete
	"Run after login, when first space entry is done."

	"No more preview-slow-animation"
	self stopScript: #loginUrl:forumPreviewChanged:byUser:.

	"Friendly reminder - if your microphone is off, you cannot speak."
	controlBar assuredSoundDialog announceMicrophoneStatus.

	"Chat and Users are expanded on first login... 
	 if you collapse them later, they'll stay collapsed
	 for subsequent transitions, but will become visible (see also #onReady)."

	(preferences restoreToolStates: (self toolTable))
	 ifFalse: [
		self restoreDefaultLayout.
	].
	self stopScript: #loginEntryComplete.! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'bgf 1/5/2009 22:24'!
loginFailure: user reason: reasonSymbol extra: extraInfo from: aController
	"Some other user fails entry due to version mismatch.
	 The router itself broadcasts a notification, and we pass it along."
	self signal: #loginFailure withArguments: { user . reasonSymbol . extraInfo . aController }.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 10/9/2008 16:59'!
loginUrl: targetUrl forumPreviewChanged: forumUrl byUser: aLoginString

	"This script is triggered by forum url update signals in org services,
 	 but only during my login process, when the main background is the forum chooser.

	 The first argument is preset when the #startScript: is issued; it is the forum we're entering.
	 The second and third arguments come from the org-services signal,
	 and identify the forum whose preview has changed and who changed it.
	 If the forum preview that has changed is our destination, draw the update."

	forumUrl = targetUrl ifTrue: [
		self fetchPreviewUpdate: targetUrl.
	].! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 11/10/2008 15:06'!
logoff
	"Logoff leaves the current session, but not the UI."
	^self logoffConfirming: true backToServiceChooser: true.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 2/15/2008 22:19'!
logoffConfirming: bConfirm

	"The default logoff method, for after we've been in a space and choose to logout."
	self logoffConfirming: bConfirm backToServiceChooser: false! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 5/13/2010 11:38'!
logoffConfirming: confirm backToServiceChooser: bSvcChooser

	"Logoff quits the entire current provider+session, but not the UI."

	self mouseCursor: Cursor wait.	
	self isDemonic ifTrue: [ self myDemon stop ].
	confirm ~=false "true or string" ifTrue:[
		"If voluntary, first check if there's an unsaved avatar."
		self readyForEvents ifTrue: [
			avatarPane onClose ifFalse: [ 
				self mouseCursor: nil.
				avatarPane expand.
				QLogger logInfo: 'Logoff cancelled due to unsaved avatar.'.
				^ nil ].
			"And kick the auto-save trigger..."
			preferences doAutoSave ifTrue: [ [ activeSpace future aboutToQuit] on: Error do: [] ].
		].
		activeSpace isValidFarRef ifTrue: [ 
			(self confirmQuit: confirm) ifFalse:[
				self mouseCursor: nil.
				self useNormalTimeouts.
				QLogger logInfo: 'Logoff cancelled by user.'.
				^nil].
			preferences doAutoSave ifTrue: [ (Delay forMilliseconds: 300) wait. ].
		].
		self useQuickTimeouts.
		self saveLayout.	"While everything's in place."
	].

	"At this point, we're confirmed and committed to logout."
	QLogger logInfo: 'Logoff confirmed.'.

	"Extension point for shutdown/cleanup."
	[ self loggingOut: false ] on: Error do: [ :ex |
		QLogger logWarning: 'Problem in shutdown: ', (ex description)
	].
	session ifNotNil: [
		session stopRender.
	].
	Tweak fullScreenMode ifTrue: [ Tweak fullScreenMode: false ].

	"Kick the display back to the goal state to clear all tools and panels."
	serviceSvc := nil.
	self useNormalTimeouts.
	bSvcChooser ifTrue: [
		self showServiceChooser.
	] ifFalse: [
		self showLogin
	].
	QLogger logInfo: 'Display cleared.'.

	"Close down the Croquet session if any."
	orgServices ifNotNil: [ 
		orgServices clearPresence.
		(Delay forMilliseconds: 100) wait.
	].
	self destroySession.
	self startScript: #restartAudioInterface.

	roundTripSound := nil.			"Real action is cleared by speaker manager."
	controlBar svcCount value: ''.
		"The above could go away once we have an observable List of active services."

	"At this point, if we're returning to service-chooser state, we're ready."
	bSvcChooser ifTrue: [
		historyPane discardAllServices.
		serviceChooser updateCurrentList.
		self updateTitle.
		^ self mouseCursor: nil.
	] ifFalse: [
		"To go back all the way to login, we have to clear the service manager state entirely."
		historyPane reset.
		login clearServices.
		self destroyServiceSession.

		"Poise for a new login."
		self setupLocalProvider.
		"We can get here in unattended situations, so we should respect a user's choice to protect
		their password. However, mobileProxy needs password to not be cleared when changing orgs."
		(login savePassword value or: [mobileProxy notNil]) ifFalse: [login password value: ''].
		Smalltalk windowLabel: (self appServiceLabelString).
		self mouseCursor: nil.
		self statusInfo: nil.

		"And signal a restart exception to bypass all pending actions"
		QwaqRestartException signal.
	].! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'ar 4/16/2011 17:32'!
longQuitDescription: logout
	"A description explaining the action that will be taken on quitting this forum."
	^logout caseOf: {
		[true] -> ['Log out of ', 
					((controlBar numForums = 1) ifTrue: [ 'this Forum'] ifFalse: [ 'these Forums'])].
		[false] -> ['Quit ', (self appName)]
	} otherwise: [ "going to another organization"
		'Leave ', self currentOrgName, ' and go to ', logout, 
			(logout last = $/ ifTrue: [' lobby'] ifFalse: ['']).
	]. 

	! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'ar 6/15/2007 10:59'!
lowSpaceWatcher
	"Wait until the low space semaphore is signalled, then take appropriate actions."
	| lowSpaceSemaphore |
	(Smalltalk garbageCollectMost <= Smalltalk lowSpaceThreshold
		and:[Smalltalk garbageCollect <= Smalltalk lowSpaceThreshold]) ifTrue: [
			"free space must be above threshold before starting low space watcher"
			^Smalltalk beep].
	lowSpaceSemaphore := Semaphore new.
	Smalltalk primLowSpaceSemaphore: lowSpaceSemaphore.
	"enable low space interrupts"
	Smalltalk primSignalAtBytesLeft: Smalltalk lowSpaceThreshold.
	"wait for a low space condition..."
	lowSpaceSemaphore wait.
	"disable low space interrupts"
	"User will be unprotected until the low space watcher is re-installed "
	self primSignalAtBytesLeft: 0.
	self primLowSpaceSemaphore: nil.
	self handleLowSpace.
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'ar 9/21/2007 16:41'!
magicDeveloperMenu

	"Not in sealed app; only in developer images."
	self isSealedApplication ifTrue: [^nil].
	^self developerMenu! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'ar 4/16/2011 17:32'!
mailDocument: docRef class: docClass
	"Populates the users mail client with the content text."
	| docName forumName |
	(docClass getContent: docRef forClient: self) ifNil: [
		self error: 'The document is not available.\It may not have been fetched yet.'  withCRs.
		^nil
	] ifNotNilDo: [:data |
		forumName := docRef get: #serviceName.
		docName := docRef get: #shortName.
		(self mailPeopleSubject: '[Forum ', forumName, '] ', docName body: data)
			ifFalse: [^nil].
		^docName.
	].! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'ar 4/16/2011 17:32'!
mailDocumentLink: docRef
	"We can't attach docs on Mac yet, so just send a QRL to the doc."
	| docName forum qrl data |
	self hasForumPagesSupport ifFalse: [self inform: 'Forum Pages access is not enabled for your account.'. ^nil].
	docRef isString ifTrue: [
		"e.g., a url"
		forum := self currentForumName.  "Only allowed from in-world. No lobby urls"
		data := docRef.
		docName := docRef asUrl authority.
	] ifFalse: [	
		docName := docRef get: #shortName.
		forum := docRef get: #serviceName.
		qrl := self serviceMgr qrlFromForum: forum.
		data := QWebServices qrl: qrl place: (docRef get: #identifier).
	].
	(self mailPeopleSubject: '[Forum ', forum, '] ', docName body: data)
		ifFalse: [^nil].
	^docName.! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'ar 4/16/2011 18:59'!
mailPeopleSubject: subject body: body
	"mails to current people listing"
	| valid n context users userPane |
	"What should we do in the lobby? Users can't control or even see the people list:
		Always show all users?
		Always none?
		Last setting?
		Don't allow mail out from lobby at all?
        Simplest to implement is to present no users in mail when in the lobby."
	userPane := self session ifNil: [self serviceChooser] ifNotNil: [self usersPane].
	users := userPane userList contents.
	n := users size.
	n > 10 ifTrue: [
		context := self usersPane filterChoices value caseOf: {
			['Everyone'] -> [' in your organization'].
			"Should we ask about many favorites? Should we stop asking after they say yes once? Would we then have
			to provide a way for someone to clear that beenAskedAboutSendingToFavorites machinePreference? Ugh."
			['Favorites'] -> [' in your favorites'].  
			['Online'] -> [' online'].
			['In This Forum'] -> [' in this Forum'].
		} otherwise: [''].
		(self confirm: ('People list is showing ',  n, ' people', context, '.\This may be too much for some mail services.\Are you sure you want to send to all those shown?\\(Choose Cancel if you want to change the People dropdown to something else.)') withCRs) ifFalse: [^false].
	].
	self mailTo: 
			(String streamContents:[:s|
				users
					do: [:entry | 
							(valid := entry userLogin includes: $@) ifTrue: [
								s nextPutAll: entry userLogin.
							].
						]
					separatedBy:[valid ifTrue: [s nextPutAll: ', ']]])
		subject: subject
		body: body.
	^true! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'hrs 8/10/2009 17:55'!
mailTo: to subject: subject body:bodyString
	"Open the local interactive mail client with fields as specified."
	| result body |
	body := bodyString isWideString ifTrue: [bodyString squeakToUtf8] ifFalse: [bodyString].
	result := nil.
	QActivityCapture action: #ComposeMail details: subject.
	[
	Smalltalk platformName caseOf: {
		['Win32'] -> [
		false ifTrue: [
			Win32Shell new shellOpen: 'mailto:', (to copyReplaceAll: ',' with: ';') encodeForHTTP,
				'?subject=', subject encodeForHTTP, '&body=', (body encodeForHTTP copyReplaceAll: '%0D' with: '%0A').
		] ifFalse: [
			"Until we find a reliable way to encode through MAPI, replace tabs with spaces."
			body := body copyReplaceAll: '	' with: '      '.
			"Something like this can be used to do attachments, but I don't need that yet."
			MAPIClient send: body from: nil to: to subject: subject.
		].
			result := 0.
			]. 
		['Mac OS'] -> [
			"In 10.4, only url-encoding is accepted.
			In 10.5, url-encoding is accepted if done consistently, otherwise it treats as quoted.
			We'll encode throughout to work in both versions."
			result := MacOSShell new system: '/usr/bin/open ''mailto:', to encodeForHTTP,
				'?subject=', subject encodeForHTTP, '&body=', body encodeForHTTP, ''''.
		false ifTrue: [  "old hack version"
			result := MacOSShell new system: '/usr/bin/open ''mailto:', to, 
						'?subject=', subject, '&body=', 
						"If we figure out the quoting, we could probably do attachments with MIME"
						((body replaceAll: $' with: $`) replaceAll: $& with: $%),  "how can we quote &?"
						 ''''.
		].
			]
	} otherwise: [self error: 'Interactive mail client not supported on this platform.']
	] on: Error do: [:e | result := e].
	^result = 0 ifTrue: [true] ifFalse: [
		self inform: 'Operating system mail operation failed.'.
		false
	]
		
		! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'bgf 1/31/2009 13:24'!
mailTo: to subject: subject body: body attachment: fullFilePath

	"Draft stuff, doesn't work.
	  Open the local interactive mail client with fields as specified, adding the attached file;
	 the attachment part is flaky, may only work on some Outlook/flavor mailers."
	| result |
	Smalltalk platformName caseOf: {
		['Win32'] -> [
			Win32Shell new shellOpen: 'mailto:', to,
				'?subject=', subject encodeForHTTP, 
				(fullFilePath ifNotNil: ['&attachment="', fullFilePath, '"'] ifNil: ['']),
				'&body=', (body encodeForHTTP copyReplaceAll: '%0D' with: '%0A').
			]. 
		['Mac OS'] -> [
			"This doesn't work.. there seems to be no attachment/support here."
			| mailto |
			mailto := 'mailto:', to, 
						'?subject=', subject, 
						(fullFilePath ifNotNil: ['&attach=', fullFilePath, ''] ifNil: ['']),
						'&body=', 
						((body replaceAll: $' with: $`) replaceAll: $& with: $%).  "how can we quote &?"
			result := MacOSShell new system: '/usr/bin/open ''', mailto, ''''.
			result isZero ifFalse: [self error: 'Shell mail operation failed.'].
			]
	} otherwise: [self error: 'Interactive mail client not supported on this platform.']
! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'ar 4/16/2011 17:33'!
mailUsers
	"Send mail to the the specified users, populating the message with info about the current 
	session."
	| body qrl  |
	qrl := serviceMgr qrlFromForum: (self currentForumName).
	body := String streamContents:[:s|
		s nextPutAll: 'Forum link: '.
		s nextPutAll: qrl.
		s cr. s cr.
		self chatPane chatPlayer chatLog writeTranscriptOn: s placeConverter: [:place | 
			(QWebServices qrl: qrl place: place)].
		s cr.
		s nextPutAll: 'http://openqwaq.com'.
		].
	^self mailPeopleSubject: '[OpenQwaq Chat Log] ', self serviceSvc name
		body: body! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'ar 4/16/2011 17:33'!
mailUsersAttaching: fullFilePath
	"Draft placeholder..
       Send mail to the the currently-listed people, attaching the given file;
	 at the moment, this does get the mailer open, but without the attachment.)"
	| valid body |
	body := String streamContents:[:s|
		s nextPutAll: 'Screenshot from Forum: ';
		  nextPutAll: (serviceMgr qrlFromForum: (self currentForumName)).
		s cr; cr.
		s nextPutAll: 'To learn more about OpenQwaq, visit: http://openqwaq.com'.
		].
	self mailTo: 
			(String streamContents:[:s|
				self usersPane userList contents
					do: [:entry | 
							(valid := entry userLogin includes: $@) ifTrue: [
								s print: entry userLogin.
							].
						]
					separatedBy:[valid ifTrue: [s nextPutAll: ', ']]])
		subject: '[Forum] ', self serviceSvc name
		body: body
		attachment: fullFilePath.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/9/2006 16:32'!
mainView
	"Answer the mainView of the receiver"
	<bewareOf: #mainViewChanged>
	^self propertyValueAt: #mainView! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/9/2006 16:32'!
mainView: aValue
	"Modify the receiver's mainView"
	^self propertyValueAt: #mainView put: aValue with: #mainViewChanged! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'ar 4/16/2011 17:48'!
makeMobileProxy: portOrPubID
	| dispatcher pubID |
	<on: makeMobileProxy>
	mobileProxy ifNotNil: [mobileProxy destroy].

	portOrPubID isNumber ifTrue: [ "Originally launched interactively in dev menu."

		"This allows foreign client testing of message relays and streams."
		"mobileProxy := QStreamingRouterTest new setUpMockProxy: port participant: self."
		dispatcher := HTTPSecureServiceDispatcher new.
		dispatcher accessLog: QLogger; errorLog: QLogger.
		dispatcher addFacetService: '/facet'.
		mobileProxy := QForeignClientProxy setUpProxy: self dispatcher: dispatcher.
		"UI streamID is hardcoded to QForeignClientProxy uiPubID / uiSubID."
		dispatcher listenOn: portOrPubID.
	] ifFalse: [  "We're given a pubID by OpenQwaq startup.  Clients connect to router, not us."
		mobileProxy ifNotNil: [mobileProxy destroy].
		TAnimationManager softwareAnimation. "In case QMachinePrefs is not already specifying it."
		pubID  := TObjectID readHexFrom: portOrPubID.
		QLogger logInfo: 'Creating proxy for ', portOrPubID.
		mobileProxy := QForeignClientProxy participantUI: self  uiPubID: pubID.
		mobileProxy runScript: #ensureHub  when: { self . #ready }.
		mobileProxy runScript: #ensureHub  when: { self . #alreadyPresent }.

		"for local proxy testing, listen for loginFacet on 9997"
		self isSealedApplication ifFalse: [
			dispatcher := HTTPSecureServiceDispatcher new.
			QForeignClientProxy localListener: mobileProxy dispatcher: dispatcher.
			dispatcher listenOn: 9997.
			QForeignClientProxy uiPubID: pubID. "so QExampleForeignClient can test in same image"
		].

		"Kill the proxy if nothing happens after a bit:
			- We don't want a load on server or for participants to be fooled.
			- We do want an explicit check with logging in Transcript."
		mobileProxy future: 180 action: #checkConnection.
	]! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 6/8/2008 14:39'!
makeNewCube

	| space island cube pos wrap |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].
	space := mainView activeSpace.
	island := space island.
	pos := avatar translation - (avatar lookAt * 10).

	self playSoundEffect: #tingle.
	cube := island future new: TCube.
	wrap := island future new: QMotionAdapter.
	wrap future isGroundBound: false.
	wrap future isDeletable: true.
	wrap future grip: cube.
	wrap future luggage: cube.
	cube future translation: pos.
	space future addChild: cube.

! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 4/5/2009 11:02'!
makePreview: blankForm viewAngle: vangle space: aTSpace

	"Make a snapshot from the home transform and answer the resulting form"
	| snap  timeout |

	"Relies on render, so can't complete if we aren't rendering."
	session ifNil: [ ^ nil ].
	session ogl ifNil: [ ^ nil ].

	"Windows gets it wrong if not foregrounded."
	Smalltalk isMainWindowObscured ifTrue: [ ^nil ].

	snap := TSnapshot new. 
	snap viewAngle: vangle.
	snap space: aTSpace transform: (aTSpace get: #homeTransform).
	snap form: blankForm.
	snap update: true.

	timeout := false.
	session snapshots add: snap.
	(snap waitUntil: #formChanged timeout: 500) ifNil: [
		"Returns an event-thing if the event fires, nil if not. (I think; undocumented)."
		timeout := true.
	].
	session ifNotNil: [
		session snapshots remove: snap ifAbsent: [].
	].
	^timeout ifFalse: [ snap form ] ifTrue: [ nil ]! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 4/10/2010 16:43'!
makeRoundTripSound
	"For testing purposes: Create a full round-trip sound setup which takes the participant's
	voice chat and subscribes to it via a subscription that is forced through the router.
	Intended as setup method for various follow-on tests.

	If the setup is successful, we leave this method with <roundTripSound> set
	 to the QClientSpeaker handling my avatar's QSpeaker,
	 and set to be in router-loopback so its subscription is on the wire.
 	 (Normally, a subscription to a locally-emitted stream would be a direct shortcut.)"

	| spkr client |

	"Find the client side-handler for my own avatar's speaker."
	spkr := avatar replica future speaker.

	"If we do setup dialogs on entry, there may not be a speaker yet."
	spkr ifNil: [^nil].
	self audioSession ifNil: [^nil].

	(spkr waitTimeoutSeconds: 10) ifTrue: [
		"Low 3-sec timeout
	       because too long a wait between request and result gets confusing."
		client := self audioSession speakerManager speakers at: (spkr get: #objectID) ifAbsent:[nil].
	].
	

	client ifNil: [ 
		"Not resource-synched yet, or disconnect, or what have you..."
		self inform: 'Cannot set up Hear Myself right now.\Try again in a moment.' withCRs.
		^ nil
	].

	"Force it into router-loopback mode."
	client future beLoopbackSound: true.
	client future play.

	"Keeping my avatar's client-speaker in roundTripSound is the marker that loopback is enabled."
	roundTripSound := client.! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 4/5/2009 11:00'!
makeServicePreview: aService
	"Create the default preview form for the given service, and return the Form.
	 May return nil if the service's entry url cannot be resolved at this time."

	| space ext |
	session ifNil: [ ^ nil ].
	session ogl ifNil: [ ^ nil].
	Smalltalk isMainWindowObscured ifTrue: [ ^ nil ].
	space := self findPreviewSpaceFor: aService.
	space ifNil: [ ^ nil ].

	"We can't make previews bigger than our OGL work area"
	ext := (520@360) scaledWithin: (extent).

	^self makePreview: (Form extent: ext depth: 32) 	"See also setupMainView"
		   viewAngle: 55.0 space: space! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/8/2008 19:12'!
makeSnapshot: blankForm
	"Make a snapshot from the avatar's point of view and answer the resulting form"
	| pc snap url |

	self readyForEvents ifFalse: [ 
		^ nil
	].

	pc := avatar replica future postcard.
	pc ifNil: [ ^ nil ].			"Nil replica - reconnect."
	pc wait.
	url := TPostcard makeUrl: pc value urlString relativeTo: (session findIslandUrl: self activeSpace island).
	snap := TSnapshot new postcard: (TPostcard urlString: url).
	snap form: blankForm.
	snap update: true.

	session snapshots add: snap.
	snap waitUntil: #formChanged.
	session snapshots remove: snap ifAbsent: [].
	^snap form! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/11/2010 16:16'!
makeThisTheHomePosition

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	((activeSpace get:#class) includesBehavior: QSpace) ifFalse: [
		^ self inform: 'This space does not allow changing the entry position.'.
	].

	(self confirm: 
		'This will make your avatar''s current position\the home position for new arrivals.' withCRs) 
	ifTrue: [
		activeSpace future setHomeTransformFromUser: (avatar replica)			   
	].! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 6/8/2008 12:39'!
mapObjectPos: vec3
	"Map the point into 2D coordinates relative to the main viewport "
	| pt2D xx yy |
	pt2D := mainView camera projectPoint: (vec3).
	xx := (1.0 + pt2D x) * 0.5 * mainView width.
	yy := (1.0 + pt2D y negated) * 0.5 * mainView height.
	^(xx@yy) truncated.
! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 4/8/2009 09:12'!
matchForum: aString
	"Answer the forum matching aString in the current service chooser, else the original arg.
	Looks for an exact match first, then treats underscores and dashes as wild cards with wildcard at end."
	| items found |
	items := self serviceChooser serviceList items collect: [:item | item name].
	found := QWebServices matchForum: aString among: items.
	(items includes: found) ifFalse: [
		"Maybe someone just created it."
		self serviceChooser onUpdateServicesList.
		items := serviceChooser serviceList items collect: [:item | item name].
		found := QWebServices matchForum: aString among: items.
	].
	^found! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 8/11/2010 10:09'!
materialCatalog
	"Answer the materialCatalog of the receiver"
	<bewareOf: #materialCatalogChanged>
	^self propertyValueAt: #materialCatalog! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 8/11/2010 10:09'!
materialCatalog: aValue
	"Modify the receiver's materialCatalog"
	^self propertyValueAt: #materialCatalog put: aValue with: #materialCatalogChanged! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 11/17/2009 18:14'!
meetingLeader
	"Return  UsersListEntry for the leader of this room's meeting,
	 or nil if there isn't one.
	 This is the basic test for whether a 'Meeting' is being led in the current space. "

	meetingState leaderLogin ifNotNilDo: [ : llogin |
		^ (usersPane entryForLogin: llogin) ifNil: [ 
			usersPane updateServiceUsers.
			usersPane entryForLogin: llogin
		]].
	^ nil! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'sn 5/13/2011 17:07'!
meetingMenu
	"Seed the 'Meeting' main menu.  See updateMeetingMenu, which rebuilds on the fly."
	| menu |
	menu := CMenu new.
	(menu add: 'Lead Meeting' translateMe target: self action: #toggleMeetingLeadership: arguments: { true })
		checked: false;	
		tag: #leadTheMeeting;
		separator: true.

	self runScript: #updateMeetingMenu when: {menu . #open}.
	^menu! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/25/2009 16:00'!
meetingState
	"Answer the meetingState of the receiver"
	<bewareOf: #meetingStateChanged>
	^self propertyValueAt: #meetingState! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/25/2009 16:00'!
meetingState: aValue
	"Modify the receiver's meetingState"
	^self propertyValueAt: #meetingState put: aValue with: #meetingStateChanged! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 7/30/2008 16:54'!
micActivity: level
	"The microphone has input"
	controlBar micActivity: level.

	"The user is not idle if speaking..."
	level > 0 ifTrue: [ self onUpdateCroquetView ].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'hrs 4/5/2010 14:30'!
microphoneStance

	"Quick summary of my microphone situation:
	 One of 
		#off		off altogether
		#on		full-time on
		#up		PTT engaged + off
		#down		PTT engaged + pressed
	"
	| ptt rec |
	controlBar microphone selected ifFalse: [ ^ #off ].
	self mobileProxy ifNotNil: [^#on].  "selected rules all."
	^ self audioSession ifNotNilDo: [ :sess |
			ptt := sess isPressToTalk.
			rec := ((sess isRecording) and: [ sess isRecorderPaused not ]).
			ptt ifTrue: [
				rec ifTrue: [ #down ] ifFalse: [ #up ]
			] ifFalse: [
				rec ifTrue: [ #on ] ifFalse: [ #off ]
			].
	] ifNil: [
		#off
	].
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 9/5/2009 16:54'!
minPPWidth
	"The lowest width for people&places."
	^ 186! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 6/9/2009 12:12'!
mobileProxy
	"Answer the mobileProxy of the receiver"
	<bewareOf: #mobileProxyChanged>
	^self propertyValueAt: #mobileProxy! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'hrs 6/9/2009 12:12'!
mobileProxy: aValue
	"Modify the receiver's mobileProxy"
	^self propertyValueAt: #mobileProxy put: aValue with: #mobileProxyChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 4/4/2007 23:25'!
myDemon
	"Answer the myDemon of the receiver"
	<bewareOf: #myDemonChanged>
	^self propertyValueAt: #myDemon! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 4/4/2007 23:25'!
myDemon: aValue
	"Modify the receiver's myDemon"
	^self propertyValueAt: #myDemon put: aValue with: #myDemonChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
navPanel
	"Answer the navPanel of the receiver"
	<bewareOf: #navPanelChanged>
	^self propertyValueAt: #navPanel! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
navPanel: aValue
	"Modify the receiver's navPanel"
	^self propertyValueAt: #navPanel put: aValue with: #navPanelChanged! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/11/2010 16:40'!
newPlaceMarker

	"For the New Place... entry in the Places menu; add a place marker."
	| marker |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].
	self mouseCursor: Cursor wait.
	marker := avatar replica future newPlaceMarker.
	marker whenResolved: [ self mouseCursor: nil ].! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 3/25/2010 13:32'!
noDemon

	"Turn off the demon and cancel my availability as a robot."
	myDemon ifNotNil: [
		myDemon stop.
		self stopScript: #demonAction:.
		myDemon stop.
		myDemon := nil.
		self setAsRobot: false mode: nil.
	].
	self updateTitle.! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 10/22/2008 21:31'!
objectsDirectory

	"The directory where the application finds persisted template files
	 for the object creation menu."

	^FileDirectory default directoryNamed: 'objects-menu/Default'! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
oglCache
	"Answer the oglCache of the receiver"
	<bewareOf: #oglCacheChanged>
	^self propertyValueAt: #oglCache! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
oglCache: aValue
	"Modify the receiver's oglCache"
	^self propertyValueAt: #oglCache put: aValue with: #oglCacheChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/21/2008 19:59'!
okToQuit
	"Answer the okToQuit of the receiver"
	<bewareOf: #okToQuitChanged>
	^self propertyValueAt: #okToQuit! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/21/2008 19:59'!
okToQuit: aValue
	"Modify the receiver's okToQuit"
	^self propertyValueAt: #okToQuit put: aValue with: #okToQuitChanged! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 4/16/2011 17:33'!
onActivateLink: pc forFrame: aFrame andGoThere: bGo

	"The user has activated a portal link through an in-space gesture. 
	 Resolve the associated link and (if needed) load the entire service.
	 If bGo is true, teleport into that service as the last step.

	 The frame that triggers this event is the holder of the postcard (usually a QGate.)
	 Any frame which triggers this event should respond to the callback
		 updatePostcard: aTPostcard
	We use this to setup and repair two way links, and to update the postcard
	if the original started life as the result of dropping a Forum-Template onto the gate.
	(Templates are not linked to; instead, we make a new forum from the template
	 and then go back and fix the postcard in the gate.)"

	| url svc svcUrl newLogin |

	<on: activateLinkForFrameAndGo in: avatar>

	pc ifNil:[^self].

	self playSoundEffect:#softClick.
	self statusInfo: 'Checking link: ', pc urlString, '...'; waitTick.

	"Compute the full url for the postcard"
	url := pc makeUrlRelativeTo: (session findIslandUrl: aFrame island).

	"Looking up a Url can time out on the network if the address is bad..."
	[ 
		self statusInfo: ('Checking link: ', pc shortDisplayString, '...'); waitTick.
		svcUrl := self serviceUrlForUrl:url.
		self statusInfo: 'Link is available.'; waitTick.

	] on: Error do: [ svcUrl := nil ].
	svcUrl ifNil: [ 
		self playSoundEffect:#click.
		self inform: 'The destination is not available.', String cr, String cr, (pc shortDisplayString).
		^ self.
	].

	"Check to see if we have that island already"
	svc := urlCache at: svcUrl ifAbsent: [nil].
	newLogin := svc isNil.
	svc ifNil: [
		"Not logged in there, yet..."
		[ 
			self checkLinkHost: svcUrl.
			svc := self activateService: svcUrl  " -- Logs in, creates instance if svc was a template."
		] on: Error, CProgressCancel do: [ :ex |
			(ex isKindOf: CProgressCancel) ifFalse: [
				self inform: 'Cannot open ', pc shortDisplayString, String cr, String cr, ex messageText.			
			].
		].	
		svc ifNil: [ ^ self ].
		servicesPane signal: #updateServiceList.		"A new forum is open"

		"Activating the service may have changed its name (if it was a template); 
		update the inbound link so the gate new points to the result, not the template."
		(self updatePostcard: pc inFrame: aFrame toService: svc) 
		 ifNotNilDo: [ :replacement |
			url := replacement
		].
	].

	"-- At this point, the destination service has been activated. -- 
	 If the origin is a QGate, and the far end isn't, try to set up return links;
	 this is how a generic new link (dropping a forum onto a gate)
	 gets transformed into a 2-way interconnection."

	((aFrame get:#class) includesBehavior:QGate) ifTrue: [
		(self checkReturnLinkFor: aFrame linkedTo: pc url: url atService: svc) ifFalse: [
			newLogin ifTrue: [
				self inform: ('The other side of the gate is not a gate back here.\',
							'(No unused gates are available.)\\',
							'Once in the other Forum,\',
							'you can use the arrow buttons to come back.') withCRs
		]]].
	
	bGo ifTrue: [ 
		"Only do the teleport if we (by now) have the service loaded;
	  	 this prevents double-login attempts if there are accidents or timing problems above.
 	       (teleportToUrl: will happily attempt a login, but we don't want that here.)"
		(self urlCache at: (QServiceLink normalizedServiceUrl: svcUrl) ifAbsent: [nil])
	        ifNotNil: [
			self statusInfo: '', pc shortDisplayString.
			self waitTick.	"Just a bit of extra chance for triggers to catch up."

			QActivityCapture action: #TeleportPortalClick details: url.

			QLogger logInfo: 'Click teleporting: ', url.
			self startScript: #teleportToUrl: withArguments: {url} 
		].
	].
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'jcg 3/3/2010 14:45'!
onAddAllowedToSpeak: login
	"A user has been added to the allowed-to-speak list.
	 If it's me, then update my UI accordingly."
	<on: addAllowedToSpeak in: meetingState>
	login = self userLogin ifTrue: [
		self controlBar leaderAllowingToSpeak: true.
		self playSoundEffect: #online.
		self updateCanSpeakInMeeting.
		meetingState audienceIsMuted ifTrue: [
			self statusAlert: 'You may speak now.'.
		]].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 4/7/2007 18:44'!
onAlertMessage: message
	<on: alertMessage>

	"From other processes, say 'participant signal #alertMessage with: msg 
	 to get text into the status line."
	statusLine alert: message! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'hrs 11/27/2007 12:03'!
onAntialiasingPrefChanged
	<on: doAntialiasingChanged in: preferences>

	self resetOpenGL.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 2/28/2009 14:04'!
onAppSave
	"We're being saved as sealed app - get a few things in place"
	<on: appSave>

	"Clear the lists of services and users."
	serviceChooser serviceList removeAll.
	servicesPane allServices:#().
	servicesPane serviceList removeAll.
	documentsPane clear.

	serviceChooser userList removeAll.
	usersPane allUsers: #().
	usersPane userList removeAll.
	placesPane placesList removeAll.
	placesPane allPlaces: #().

	"cancel local services"
	self cancelLocalProvider.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 10/19/2007 13:06'!
onAppStart
	"We're starting as sealed app - get a few things in place"
	<on: appStart>
	login loadDefaultUser.
	login staticServerList: nil.
	login updateServerList.
	login clearServices.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 2/25/2008 15:42'!
onAppendEntry: aString with: anEvent
	"This is dog stupid and cheap... but it works."
	<on: appendEntry>

	logView panel selectAt: logView panel contents size+1.
	logView panel paste: anEvent arguments first select: false.
	logFile ifNotNil:[logFile nextPutAll: anEvent arguments first. logFile flush].
	logView panel contents size > self logViewCharacterLimit ifTrue:[
		logView panel 
			contents: (logView panel contents last: self logViewCharacterLimit // 2);
			textFont: (QStyles smallTextFont).
	].	
	! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'jcg 3/3/2010 14:44'!
onAudienceIsMutedChanged: bMuteAudience
	<on: audienceIsMutedChanged in: meetingState>
	bMuteAudience ifTrue: [
		QLogger logInfo: 'Audience is muted (meeting in progress.)'.
	] ifFalse: [
		QLogger logInfo: 'Audience free to chat.'.
	].

	controlBar leaderControllingSpeech: bMuteAudience.
	self updateCanSpeakInMeeting.

	usersPane invalidate.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/9/2009 19:39'!
onAudienceStayPutChanged: bStayPutAudience
	<on: audienceStayPutChanged in: meetingState>
	bStayPutAudience ifTrue: [
		QLogger logInfo: 'Audience movement is not allowed.'.
		"PLACEHOLDER"
		self isMeetingLeader ifFalse: [ avatar stopDriving ].
	] ifFalse: [
		QLogger logInfo: 'Audience movement is allowed.'.
	].

	self updateMobility.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 11/24/2007 12:58'!
onAutoChat
	<on: autoChat in: avatar>

	"When the avatar raises #autoChat, go directly to the chat pane focus."

	chatPane visible ifTrue: [
		self playSoundEffect: #softClick.
		chatPane takeFocus.
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'hrs 4/30/2009 10:44'!
onAutoImport: aSpace
	| |
	<on: processCustoms>
	"We've been nominated to check for items waiting to be released from customs."
	DropAllButLastEvent signal.
	aSpace isValidFarRef ifFalse: [ ^ nil ].
	self hasOfflineUploadSupport ifFalse: [^nil].
	self isReusableGuestUser ifTrue: [ ^ nil ].		"Guests don't autoimport."	
	self canModifyForum ifFalse: [ ^nil ].
	
	customsPane processOfflineUploadsWithNotice: false.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 3/31/2010 17:32'!
onAutoSave: aSpace
	| nowTime lastTime totalTime island service |
	<on: autoSave>
	DropAllButLastEvent signal.
	aSpace isValidFarRef ifFalse: [ ^ nil ].
	self isReusableGuestUser ifTrue: [ ^ nil ].		"Guests don't autosave."	
	self canModifyForum ifFalse: [ ^nil ].

	lastTime := Time millisecondClockValue.
	statusLine status: 'Autosaving ... '.	
	island := aSpace island.
	service := self urlCacheAt: (session serviceUrlForIsland: island)  ifAbsent: [serviceSvc].
	service ifNil: [ 
		QLogger logInfo: 'Autosave: ', (aSpace), ' not found in url cache.'.
		^ nil ].

	(servicesPane entryForUrl: (service url)) ifNotNil: [ : entry | 	
		(self serviceEntryIsDirty: entry ) ifFalse: [
			QLogger logInfo: 'Autosave ', (entry name), ' not needed (no change)'.
			^ nil
		]].

	totalTime := [
		[ 
			self autoSave: island in: service notifying: [:amt|
				nowTime := Time millisecondClockValue.
				nowTime - lastTime > 500 ifTrue:[
					lastTime := nowTime.
					statusLine status: 'Autosaving ... ', (amt * 100 truncateTo: 0.1), '% done'.
				]].
		] on: Error do: [ : ex |
			QLogger logWarning: 'Autosave abandoned: ', (ex description).
			^ nil.
		].
	] timeToRun.
	statusLine status: 'Autosave finished.'; waitTick.
	QLogger logInfo: 'Autosave (', service name,  '):', (totalTime // 1000 max: 1), ' seconds'.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/22/2009 17:55'!
onAvatarArrived: atar at: placeLabel
	"For Quindi logging. Note the fact that an avatar has arrived at a named place."
	| atarName |
	<on: avatarArrived in: activeSpace>
	quindiLogging ifTrue:[
		atarName := atar get: #publicName.
		self quindiLog: atarName, ' arrived at ', placeLabel
			hiddenText: 'This event happened at ', Time now print24.
	].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 1/6/2010 14:34'!
onAvatarChanged
	"Hook up triggers for a new avatar-user"
	<on: avatarChanged>

	self stopScript: #showPanels:.
	self stopScript: #onOverWarpPlace:.
	self updateStatusRelay.

	avatar ifNil: [ ^ self ].

	"Whenever we drop out of a sticky viewpoint, we restore the normal UI...
	 this is the reset from 'Zoom100' closeup focus on an application."
	self startScript: #showPanels: when: {avatar. #stickyChanged}.
	"Bullet-indicator when the mouse hovers over a warpable destination."
	self runScript: #onOverWarpPlace: when: { avatar . #overWarpPlace }.

! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'jcg 12/28/2006 21:50'!
onAvatarColorPrefChanged
	<on: avatarColorChanged in: preferences>

	"Pass this on to the Avatar, so that painting draws in my preferred color."
	avatar ifNotNil: [
		avatar laserColor: (preferences avatarColorValue ifNil:[^self]).
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 1/13/2010 11:59'!
onAvatarEmbarked: placeName
	<on: avatarEmbarked in: avatar>

	DropAllButLastEvent signal.

	"The avatar has started a goto sequence towards a named destination;
	 this happens when we tab, or hit a goto button, or otherwise use a navigation shortcut.
	 So we're not in keyboard input anymore.
	 For sound effect,  rule out 'With ' place names, as those are used for Go With ... 
	 when you follow an avatar around, and the embarked chimes are irksome then.)"

	self readyForEvents ifTrue: [
		"Forget keyboarding  if pulled away from an app by 'Follow Me' or other leader stuff."
		avatar stopActivity: #Typing.
		session keyboardDriving: true.
	].

	(placeName notNil and: [placeName beginsWith: 'With ']) ifFalse: [
		self playSoundEffect: #warpPosition.
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/9/2009 14:26'!
onAvatarEnter: avtr
	<on: avatarEnter in: activeSpace>

	"Don't play sounds to me for me."
	((avtr get:#sessionID) = self sessionID)
 	 ifFalse: [
		self playSoundEffect: #avatarEnter.
		self statusInfo: '', ((avtr get: #publicName) ifNil: ['A user']), ' has arrived.'.
	] ifTrue: [
		self resumeAudio.	"In case it was suspended due to heartbeat lag while teleporting."
	].
	orgServices ifNil: [
		"We rely on org-services user:location:localTime: messages when available;
	 	 without them (legacy case, soon to be retired) update the lists manually."
		servicesPane signal: #updateServiceList.
	].
	usersPane signal: #updateUserList.
	infoCache faces cacheGraphics: avtr.
	self updateRecordingUsers.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/9/2009 14:35'!
onAvatarHide: avtr

	"Signal from replica after an avatar has been hideen due to possible timeout."
	<on: avatarHide in: activeSpace>

	mainView ifNil: [ ^ self ].

	"Make sure it's not about me.  Own-avatar warnings are done in checkHeartbeatLag,
	 since often a network problem will mean the island is not heartebeating and won't signal."
	((avtr get:#sessionID) = self sessionID)  ifFalse: [
		self isMeetingFollower ifFalse: [
			"Don't bother the audience in a meeting."
			statusLine alert: '', ((avtr get: #publicName) ifNil: ['A user']), ' has a connection problem.'.
		].
		usersPane signal: #updateUserList.
	].

	self updateRecordingUsers.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/9/2009 14:25'!
onAvatarLeave: atarID sessionID: sessionID

	<on: avatarLeave in: activeSpace>

	"If logging out, just let it be..."
	mainView ifNil: [ ^ self ].
	mainView avatar ifNil: [ ^ self ].

	"Departure updates for users other than me:"
	(sessionID = self sessionID)  ifFalse: [
		self playSoundEffect: #avatarLeave.
		usersPane signal: #updateUserList.	"In case it is a self-destruct (no presence signal)."	
		self updateRecordingUsers	
	].
	! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/9/2009 14:25'!
onAvatarRestore: avtr

	"Signal from replica after an avatar is in danger of timing out."
	<on: avatarRestore in: activeSpace>
	mainView ifNil: [ ^ self ].

	"Is it about someone else, or about me?"
	((avtr get:#sessionID) = self sessionID)  ifFalse: [
		self statusInfo: '', ((avtr get: #publicName) ifNil: ['A user']), ' is recovering from delays.'.
	] ifTrue: [
		"If we get a recovery signal about ourself from the space,
	       then our Croquet is back in action.
	 	 Undo the effects of #onHeartbeatLag: ..."
		self statusAlert: nil.
		self statusInfo: 'Network has recovered.  Activity resumed.'.
		self resumeAudio.
	].
	usersPane signal: #updateUserList.
	self updateRecordingUsers.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/22/2009 17:55'!
onAvatarWarning: avtr msRemaining: ms

	"Signal from replica after an avatar has been suppressed due to possible timeout."
	<on: avatarTimeoutWarning in: activeSpace>

	mainView ifNil: [ ^ self ].

	"Make sure it's not about me.  Own-avatar warnings are done in checkHeartbeatLag,
	 since often a network problem will mean the island is not heartebeating and won't signal."

	((avtr get:#sessionID) = self sessionID)  ifFalse: [
		usersPane transientNotice: 'Delay' 
				    forSession: (avtr get: #sessionID)
				    forMillis: (ms+3000).
		statusLine info: '', ((avtr get: #publicName) ifNil: ['A user']), ' may have a delay.'.
		QActivityCapture countAction: #DelayOtherUser.
	].

	! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 2/2/2008 01:22'!
onAwayOver
	<on: mouseMove in: awayOverlay>
	
	controlBar onAwayButtonEnter.	"So the messages are consistent."! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'jcg 12/23/2009 10:46'!
onBroadcastingSessionChanged: replica
	"Someone started or stopped recording"
	| userName broadcasting |
	<on: isBroadcastingSessionChanged in: activeSpace>
	userName := replica get: #publicName.
	broadcasting := replica get: #isBroadcastingSession.
	broadcasting ifTrue:[
		broadcastingUsers add: userName.
	] ifFalse:[
		broadcastingUsers remove: userName ifAbsent:[].
	].
	self statusInfo: userName, 
		(broadcasting ifTrue:[' started '] ifFalse:[' stopped ']), 
		'webcasting this session'.

	recordingIndicator visible: (recordingUsers notEmpty).
	webcastIndicator visible: (broadcastingUsers notEmpty).! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 8/30/2009 00:00'!
onCallForVote: pollRef withSeedVotes: seeds fromUser: publicName
	"Signal relay for QPollBarChart ... used to prompt for votes."
	<on: callForVote in: avatar>
	QPollBarChart 
		askForVote: pollRef 
		suggesting: (seeds sort) 
		fromUser: publicName
		forClient: self.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/24/2007 20:06'!
onCancelLogin

	<on: cancel in: login>

	exitTripped := true.
	self exitConfirming: false.! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 10/26/2006 10:38'!
onCancelPreferences
	<on: cancel in: prefsPanel>

	self remove: prefsPanel.
	self menu enable.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 11/27/2009 12:43'!
onCancelServiceChooser

	"Abandon the service-manager,
	 and go back to the initial phase 1 login panel."

	<on: cancel in: serviceChooser>

	self loggingOut: false.
	self mouseCursor: Cursor wait.
	[
		
		login clearServices.
		self destroyServiceSession.
		self showLogin.

	] ensure: [
		self mouseCursor: nil.
	].! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/16/2011 17:33'!
onCheckinDocument: checkinInfo

	"Perform a document checkin. If upload is true, upload the document to the server.
	If locked is true, unload the document on the server. Tries a server-sice upload first,
	if that fails attempts to check in through the client."

	| resp repository mtime shortName comment pb |
	<on: checkinDocument>
	checkinInfo success: false.

	"Construct the check-in comment"
	comment := String streamContents:[:s|
		s nextPutAll: 'Checked in from Forum "'.
		s nextPutAll: (self currentForumName ifNil:['Lobby']).
		s nextPutAll: '" in organization "'.
		s nextPutAll:  (self currentOrgName ifNil:['Unknown']).
		s nextPutAll: '"'.
	].


	repository := (QDocumentRepository cache)
		at: checkinInfo repoUrl 
		ifAbsentPut:[QDocumentRepository type: checkinInfo repoType url: checkinInfo repoUrl owner: self].
	repository ifNil:[^self inform: 'Could not check in document.\\Unknown repository: ', checkinInfo repoType].

	shortName := checkinInfo serverPath copyAfterLast: $/.
	QLogger logInfo: 'Checking in ', checkinInfo repoPath.
	QLogger logInfo: 'Repository ', (checkinInfo repoType, '+', checkinInfo repoUrl).

	[QDocumentRepository mutex critical:[
		pb := CProgressBar new.
		pb displayProgress: 'Checking in...' at: self dialogCenter from: 0 to: 4 during:[:bar|
			mtime := (serviceMgr listFiles: checkinInfo serverPath) 
				ifNotEmpty:[:entries| entries first modificationTime].
			checkinInfo mtime = mtime ifTrue:[
				QLogger logInfo: 'Document unmodified: ', shortName.
			].
			pb value: 1. (self wait: 0.2).
			(checkinInfo upload and:[checkinInfo mtime ~= mtime]) ifTrue:[
				"Try a server-side upload first"
				resp := self serverCheckin: checkinInfo serverPath
							repository: checkinInfo repoUrl 
							path: checkinInfo repoPath 
							type: checkinInfo repoType.
				resp name == #ok ifFalse:[
					"Log the failure just in case"
					QLogger logWarning: 'Server checkin failed: ', resp contentString.
					"If the server doesn't support repositories, try it from the client"
					resp := self clientCheckin: checkinInfo serverPath 
								repository: checkinInfo repoUrl 
								path: checkinInfo repoPath 
								type: checkinInfo repoType.
					resp ifNotNil:[^self inform: resp].
				].
				pb value: 2. self wait: 0.2.
				"Commit the document on the server"
				resp := repository commit: checkinInfo repoPath 
						comment: comment.
			].
			pb value: 3. self wait: 0.2.
			checkinInfo locked ifTrue:[
				"Unlock the document"
				QLogger logInfo: 'Unlocking ', shortName.
				resp := repository removeLock: checkinInfo repoPath.
				resp isSuccess ifFalse:[^self inform: resp].
			].

			pb value: 3. self wait: 0.2.
			"Check in succeeded; delete document on server"
			checkinInfo deleteFile ifTrue:[
				QLogger logInfo: 'Deleting ', shortName.
				serviceMgr deleteFiles: checkinInfo serverPath forReal: true.
			].

			checkinInfo success: true.
			pb value: 4. self wait: 0.2.
		].
	]] on: Error do:[:ex|
		^self inform: 'Could not check in document.\\' withCRs, ex description.
	].
! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 5/10/2010 10:42'!
onCheckoutDocument: docRef andLaunch: embeddableApp

	"Check out a document from a repository. Try a server-side
	checkout first, if that fails attempt a checkout through the client.

	Finish the launch of the embeddable app when successful."

	| repoType repoUrl resp serverPath  repoPath repository needLock shortName |
	<on: checkoutDocumentAndLaunch>

	serverPath := docRef get: #asServiceManagerPath.
	repoType := docRef get: #repoType.
	repoUrl := docRef get: #repoUrl.
	repoPath := docRef get: #repoPath.
	needLock := docRef get: #requiresLocking.

	repository := (QDocumentRepository cache)
		at: repoUrl 
		ifAbsentPut:[QDocumentRepository type: repoType url: repoUrl owner: self].
	repository ifNil:[^self inform: 'Could not check out document.\\Unknown repository: ', repoType].

	shortName := serverPath copyAfterLast: $/.
	QLogger logInfo: 'Checking out ', repoPath.
	QLogger logInfo: 'Repository ', repoType, ' ',  repoUrl.

	[QDocumentRepository mutex critical:[

		"Lock the document"
		needLock ifTrue:[ | lockMsg |
			QLogger logInfo: 'Locking ', shortName.
			[lockMsg := nil.
			[resp := repository lock: repoPath] 
				on: QDocumentIsLocked 
				do:[:ex| lockMsg := ex messageText].
			lockMsg == nil] whileFalse:[
				(self confirm: lockMsg, '\\Continue check-out and override this lock?' withCRs) 
					ifFalse:[^embeddableApp app future stopRunning: session userID].
				repository removeLock: repoPath.
			].
			embeddableApp locked: true.
		].

		"Check for a left-over document on the server first"
		resp := self verifyCheckoutOverwrite: docRef.

		(resp == #useExisting) ifTrue:[
			QLogger logInfo: 'Using existing document'.
			embeddableApp checkedOut: true.
			^embeddableApp launchApp
		].

		"If in doubt, err on the side of no harm and cancel"
		(resp == #useRepository)  
			ifFalse:[^embeddableApp app future stopRunning: session userID].

		"Try a server-side checkout first"
		resp := self serverCheckout: serverPath repository: repoUrl path: repoPath type: repoType.
		resp name == #ok ifFalse:[
			"Log the failure just in case"
			QLogger logWarning: 'Server checkout failed: ', resp contentString.
			"If the server doesn't support repository ops, try it from the client"
			resp := self clientCheckout: serverPath repository: repoUrl path: repoPath type: repoType.
			resp ifNotNil:[
				"Unlock the document and inform user about failure"
				needLock ifTrue:[repository removeLock: repoPath].
				embeddableApp app future stopRunning: session userID.
				^self inform: resp
			].
		].

		(serviceMgr listFiles: serverPath) ifNotEmpty:[:entries|
			embeddableApp mtime: entries first modificationTime.
		].

		"Everything is fine, keep launching the document"
		embeddableApp checkedOut: true.
		^embeddableApp launchApp
	]] on: Error do:[:ex|
		embeddableApp app future stopRunning: session userID.
		^self inform: 'Could not check out document.\\' withCRs, ex description.
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/11/2009 18:47'!
onClickRecordingIndicator
	<on: mouseDown in: recordingIndicator>

	"Convenienience: click in the recording indicator, bring up the tool if I'm recording."

	recorderPane isRecording ifTrue: [
		recorderPane expand; comeToFront
	].
	! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'jcg 12/23/2009 10:20'!
onClickWebcastIndicator
	<on: mouseDown in: webcastIndicator>

	"Convenienience: click in the indicator, bring up the tool if I'm casting."

	webcastPane isRecording ifTrue: [
		webcastPane expand; comeToFront
	].
	! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'hrs 10/9/2009 11:37'!
onClientSpeakersChanged
	| stats |
	<on: clientSpeakersChanged>
	self audioSession ifNil: [^self].
	latencyPane stopWatchingTag: #Audio.
	self audioSession speakerManager speakers do:[:spkr|
		spkr jitterStats ifNil:[
			stats := TWindowedStatistics forWindowDuration: 60000.
			stats tagName: #Audio.
			spkr jitterStats: stats.
		].
		latencyPane startWatching: spkr jitterStats.
	].
	roundTripSound ifNotNilDo:[:spkr|
		spkr jitterStats ifNil:[
			stats := TWindowedStatistics forWindowDuration: 60000.
			stats tagName: #Audio.
			spkr jitterStats: stats.
		].
		latencyPane startWatching: spkr jitterStats.
	].
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 11/24/2009 14:00'!
onCommitPreferences
	<on: commitPreferences>
	DropAllButLastEvent signal.
	preferences ifNotNil: [
		preferences storeXml		"Logs any problems."
	].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 8/31/2010 15:52'!
onControlStreamLost: aController
	"Notify the participant that we lost the control stream in a streaming controller.
	There are two possible causes:
		a) A connection failure (in which case the hub will have reported a connection error)
		b) We got kicked off the router (most likely due to a login by the same user)
	"
	| kickedIsland |
	<on: controlStreamLost>

	self exitTripped ifTrue: [ ^ nil ].		"Expected cases, depending on timing."
	session ifNil: [ ^ nil ].

	(session notNil and:[session reconnecting]) ifTrue:[^nil]. "we are in reconnect"

	"If there's no reason for the logoff, reconnect"
	aController logoffReason ifNil:[^self signal: #reconnect].

	kickedIsland := aController island.
	kickedIsland == mainView activeIsland ifTrue:[
		"If we get kicked off the current space, log off and inform the user"
		self logoffConfirming: false backToServiceChooser: true.
		self inform: 'You have been logged off.'
	] ifFalse:[
		"If we get kicked off an inactive space, just drop that space"
		self deactivateService: (session serviceUrlForIsland: kickedIsland).
		self inform: 'You have been logged out of ', aController serviceName.
	].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 2/11/2007 11:33'!
onCursorModeChange
	<on: cursorModeChanged in: controlBar>

	"These should coordinate with the cursors in the TTasks.
	  (The task cursors are only installed at the time the task is pushed, however,
	  which is usually upon pointerDown.  
	 These cursors take effect when the mode control is changed.)"

	controlBar cursorMode caseOf: {
		[#thumbtack] -> [viewPanel mouseCursor: CCursorLibrary downLeftCursor].
		[#paintbrush] -> [viewPanel mouseCursor: Cursor marker].
		[#editBox] -> [viewPanel mouseCursor: Cursor crossHair].
		[#materialEditor] -> [viewPanel mouseCursor: Cursor square].
	} otherwise: [
		viewPanel mouseCursor: nil
	].! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 6/8/2008 19:12'!
onDoAutoSaveChanged
	| |
	<on: doAutoSaveChanged in: preferences>

	"There is no surgical way to unhook one event-trigger startScript:when:,
	 so to reset this, we simply redo the job handler listener map.
	 Note that QAutoSaveJob has a class 'isEnabledForClient: method 
	 that checks and enforces the auto-save mechanism,
	 this trigger is just a relay to let the mechanism know a change has been made."

	(session notNil and: [session jobHandler notNil]) ifTrue: [
		session jobHandler resetAllListeners.
	].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 7/19/2010 10:45'!
onDropFiles: anEvent
	"Handle a number of dropped files from the OS."
	| numFiles streams ndirs dirName names up |
	<on: dropFiles>

	numFiles := anEvent contents.
	names := Array streamContents: [ : all |
			1 to: numFiles do: [ :i|  
				all nextPut: (FileStream requestDropName: i) ]].
	streams := #().
	[[
		"We'll accept a directory, or a bunch of files, but not a mix, in the original payload."
		ndirs := names count: [ :str | FileDirectory isDirectory: str].
		((ndirs > 1) or: [ ndirs > 0 and: [names size > ndirs]])  ifTrue: [
			^ self inform: 'Directories must be dropped one at a time.'.
		].

		"If the drop is a single directory,  take all the files (not subdirs) within it... No recursion".
		((ndirs = 1) and: [ names size = 1 ]) ifTrue: [
			| dir |
			dir := FileDirectory on: (names first).
			dirName := dir localName.
			streams := Array streamContents: [ : all |
				dir fileNames do: [ :n |
					n first isAlphaNumeric ifTrue: [
						all nextPut: (FileStream readOnlyFileNamed: (dir fullNameFor: n)) 
					]]]
		] ifFalse:[
			"otherwise request the drop streams and skip any embedded directories"
			streams := (1 to: names size) 
				collect:[:i| FileStream requestDropStream: i]
				thenSelect:[:file| file notNil].
		].
		streams ifEmpty: [
			^ self inform: 'That directory does not contain any files.'
		].
		
		"special case: neither a file card doc nor an object import, and we want
		it to work at any time -- even when we don't have a session yet and are
		not ready for events."
		(streams size = 1 and: [
			up := streams first name asUppercase.
			(up endsWith: '.TLL') or: [up endsWith: '.QWQ']
		]) ifTrue: [
			^self processQwqStream: streams first.
		].

		"If we're not in-forum, but in the lobby, let the SC handle the drop.
	 	 it does very different things than the importer... sends the data for offline drop."
		serviceChooser visible ifTrue: [
			^ serviceChooser dropFiles: streams at: anEvent position directoryName: dirName.
		].

		"If the drop happened into a tool that handles dropFiles, so be it."
		self floatingPanes do: [ : pane |
			(pane visible 
			 and: [ pane respondsTo: #dropFiles:at:directoryName: ]
			 and: [ self isPoint: (anEvent position) inPlayer: pane ignoring: #() ] )
			 ifTrue: [
				^ pane dropFiles: streams at: anEvent position directoryName: dirName. 
				] ].
			
		self readyForEvents ifFalse: [ ^ self ].
		(meetingState isEditBlocked and: [ self isMeetingLeader not]) ifTrue: [ 
			self statusAlert: 'Imports are not permitted by the meeting leader.'.
			^ self ].
		self canModifyForum ifFalse: [ 
			self statusInfo: 'As a visitor, you cannot add documents.'.
			^ self ].

		"If the drop includes many images, we offer to make a slide show."
		(session checkDropSlides: streams event: anEvent initialName: dirName) 
			ifTrue: [ ^ self ].
		"Otherwise, just grind through the items, dropping each according to its nature."
		streams do: [ : stream |
			self readyForEvents ifFalse: [ ^ self ].	"Quit, or connection-loss..."
			session dropFile: stream event: anEvent editBox: true.
		].
	] ifError: [ : msg|
		self inform: ('An unexpected problem interrupted the drop.\\', msg) withCRs.
	]] ensure: [
		streams do: [ : str | [  str close ] ifError: [ :ignore ] ]
	].! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 11/24/2009 19:20'!
onEditBlockedChanged: bBlocked
	<on: isEditBlockedChanged in: meetingState>
	"The meeting control flag to block imports and scene editing has been changed."
	bBlocked ifTrue: [
		QLogger logInfo: 'Scene editing is disallowed (meeting in progress).'.
	].

	"The meat is in my #canModifyForum, just re-apply it."
	self applyUserRole.

	"Since we're taking away the Tests submenu, 
	 stop the tests that are running.  Gits their attention.
	 Especially, this stops those annoying voice-transport tests."
	bBlocked ifTrue: [ self stopTests ].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 8/23/2010 17:12'!
onEditMaterial: frameRef
	<on: editFrameMaterial in: avatar>
	"From in-space, a request to edit the material of the given TFrame"
	QFrameMaterialsPane editMaterials: frameRef inOwner: self! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 7/12/2009 13:39'!
onEnableExperimentalFeaturesChanged
	| enabled |
	<on: enableExperimentalFeaturesChanged in: preferences>

	DropAllButLastEvent signal.
	enabled := preferences enableExperimentalFeatures.
	(self menu itemTagged: #experimentalFeatures)
		ifNotNilDo: [:item | item checked: enabled].

	false ifTrue: [
		"This puts the meeting menu under x-features; (code just in case.) "
		enabled ifTrue: [
			(self menu itemNamed: 'Meeting') ifNil: [
				(self menu addItem: 'Meeting' after: (self menu itemNamed: 'View'))
					subMenu: (self meetingMenu).
			].
		] ifFalse: [
			self menu removeMenuItems.
			self addSubMenusTo: self menu.
			orgServices ifNotNil: [ orgServices beMeetingLeader: false ].
		]].

	self applyUserRole.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 6/8/2008 23:44'!
onExit
	"The menu/window-close #exit operation signals rather than directly invokes,
	 so that in the unsealed window close case, the  work is done in the CProject context."
	<on: exit>
	self exitConfirming: true.! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 3/6/2009 03:36'!
onExtentChanged
	"Window has been resized"
	<on: extentChanged>
	DropAllButLastEvent signal.

	"The current controlling view can enforce limits by forcibly re-resizing my window."
	login visible ifTrue: [
		(login interceptBounds: (self windowDisplayBounds)) ifTrue: [ ^ self ].
	] ifFalse: [	
		serviceChooser visible ifTrue: [
			(serviceChooser interceptBounds: (self windowDisplayBounds)) ifTrue: [ ^ self ].
		] ifFalse: [
			(self interceptBounds: (self windowDisplayBounds)) ifTrue: [ ^ self ].
		]].

	self layoutEverything.
	statusLine setStyleForWindowWidth: (self width).
	self saveLayout.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/11/2009 12:27'!
onFileDialogDone
	"When a file dialog completes, we have to recreate OpenGL 	
	 because the goddamity coverflow-garbage can debilitate the renderer."
	Smalltalk platformName = 'Mac OS' ifFalse: [ ^ self ].
	session ifNotNil: [ 
		session ogl ifNotNil: [
			session killRender.
			session startRender.
		]].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 1/11/2009 16:52'!
onFloatForEmbeddedApp: anEmbeddedAppRef
	"Given a far ref to a TEmbeddedApp, 
	 open (or raise to front) a QFloatingEmbeddedApp for it.
	 An in-space object can  say  
		'avatar signal: #floatForEmbeddedApp with: aTEmbeddedApp'
	to show the floating pane.  
	The avatar is used since this a per-client action,
	we don't want to raise floating panes on other user's clients."

	<on: floatForEmbeddedApp in: avatar>

	self showFloatingAppPane: anEmbeddedAppRef withAlignment: #rightCenter.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 1/16/2008 14:32'!
onFloatMessage: message
	<on: floatMessage>
	"From other processes, say 'participant floatMessage:msg'
	 to get text into the floating message line"
	DropAllButLastEvent signal.
	statusLine floatingMessage: message.
	self waitTick.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 6/19/2009 14:25'!
onGeneralTeleportHost: host organization: org forum: forum location: location
	"Go to the specified place.
	If this cannot be done without logging out of where we are (e.g., if on a different host/org),
	then ask they user if they really want to log out."
	| isNewHost isNewOrg url isInitial adjustedOrg adjustedForum qrl |
	<on: generalTeleport>
	DropAllButLastEvent signal.
	isNewHost := host ~= self login serverList value.
	isNewOrg := org ~= self currentOrgName.
	isInitial := self readyForEvents not.
	login forumGoal ifNotNil: [^self]. "Don't stack operations."
	forum isEmptyOrNil ifFalse: [login forumGoal: forum].  "stake our claim now. make nice below"
	"If we came here (through a QRL) while waiting on phase2 of an autoLogin 
	(from an earlier QRL that stopped to give the guy a chance to complete login info),
	cancel completion of the first autologin."
	self stopScript: #autoLogin2:forum:location:robotic:.
	
	(isNewHost | isInitial) not & isNewOrg ifTrue: [ "If we have a valid login, avoid false positives on changed org."
		adjustedOrg := QWebServices matchOrg: org among: serviceChooser orgChoices items.
		isNewOrg := adjustedOrg ~= self login organization value.
	] ifFalse: [adjustedOrg := org].

	adjustedForum := forum ifEmpty: [] ifNotEmpty: [self matchForum: forum].
	login forumGoal: adjustedForum.  "for use by client update machinery"
	login placeGoal: (location ifEmpty: [] ifNotEmpty: [location]).

	"This is adjusted only as well as we can without changing the service manager."
	qrl := '//', host, '/', adjustedOrg, '/', (adjustedForum ifNil: ['']).
	QActivityCapture action: #GeneralTeleport details: qrl.
	QLogger logInfo: 'Teleporting ', qrl.

	(isInitial | isNewHost | isNewOrg | forum isEmpty) ifTrue: [
		isInitial ifFalse: [(self logoffConfirming: qrl backToServiceChooser: false) ifNil: [^nil]].
		^self autoLogin: host organization: adjustedOrg forum: adjustedForum location: location robotic: false.
		].

	url := serviceSvc baseUrl, '/', adjustedForum.  "Not the qrl (orgName), but a serviceUrl (realm guid)."

	"I'd rather do the next two statements, but we don't get a #ready if we're already in the space.""
	location isEmptyOrNil ifFalse: [
		self startScript: #goToPlaceName: withArguments: {location} when: {self. #ready}. 
	].
	self  signal: #signalTeleportToUrl with: url."
	self  signal: #teleportToUrlPlace with: url with: location.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/2/2008 19:34'!
onGlobalKeyDown: evt

	"For accelerator and custom hotkeys - observes all events into the project."
	| accelerators |
	mainView ifNotNil: [
		self isAway ifTrue: [
			"Typing clears the Away state,
	  		 except for  meta-keys (Cmd, Alt) ... so alt-tab focus-shift doesn't cancel Away."
			(evt buttons anyMask: (64 + 32)) ifFalse: [
				controlBar awayButton selected: false.
			].
		].
		accelerators := self keydownAcceleratorCases.
		evt keyValue caseOf: accelerators otherwise: [].
		self checkMicrophoneKeyEvent: evt onOff: true.

		avatar ifNotNil: [ 
			mainView keyboardDriving ifFalse: [ avatar typingActivity ]
		]].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 3/14/2009 17:31'!
onGlobalKeyStroke: evt
	"For accelerator and custom hotkeys - observes all events into the project."
	
	statusLine startCounting.
	(evt commandKeyPressed and: [ evt keyCharacter = $T ]) ifTrue: [	
		self reportIncident: 'Command-Shift-T Transcript at user request'. 
		self statusInfo: 'Cmd-Shift-T:  Situation Report Sent'.
	].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 11/21/2008 15:17'!
onGlobalKeyUp: evt
	"For accelerator and custom hotkeys - observes all events into the project."
	mainView ifNotNil: [
		self checkMicrophoneKeyEvent: evt onOff: true.
		avatar ifNotNil: [ 
			mainView keyboardDriving ifFalse: [ avatar typingActivity ]
		]].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 2/2/2008 12:03'!
onGlobalMouseDown: evt

	"For accelerator and custom hotkeys - observes all events into the project."

	"Clicking anywhere clears the away state."
	self isAway ifTrue: [

		"We don't flip the state if the event is going to the awayButton
	 	 since it will toggle the state on fire... we don't want to toggle it twice."
		(controlBar awayButton globalBounds containsPoint: (evt position)) ifFalse: [
			controlBar awayButton selected: false.
		]
	].

! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 3/14/2009 17:31'!
onGlobalMouseMove: evt

	"For accelerator and custom hotkeys - observes all events into the project."

	statusLine startCounting.	"The floating (balloon-like) text line persists until a mouse move;
							 at that point it starts to fade unless re-asserted."
	mainView ifNotNil:[
		self checkMicrophoneMouseEvent: evt.	
			"The mac uses mouse events as a vehicle for Control-key state changes."
	].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/10/2009 14:05'!
onGlobalMouseUp: evt
	"Observes all events into the project.
	 Workaround to prevent run-on mouse driving when yellow-down starts in a menu-state.

	 * The underlying problem is that if Tweak sees a right-down to a player other than a live menu,
	   that cancels the menu, and the player starts getting mouse-moves (so we drive)	 
 	   but the player does not get the up, because the up goes to whomever got the down... (the menu).
	   So we never see the up normally, and keep driving..."

	avatar ifNotNil: [ 
		avatar isDriving ifTrue: [
			avatar stopMouseDriving.
			mainView ifNotNil: [ 
				viewPanel mouseCursor: (mainView task cursor).
			]]].! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 6/8/2008 19:12'!
onIdleFramesPerSecondChanged
	<on: idleFramesPerSecondChanged in: preferences>

	session ifNotNil: [
		session idleFramesPerSecond: preferences idleFramesPerSecond. "1-1000"
	].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 1/16/2008 14:32'!
onInfoMessage: message
	<on: infoMessage>
	"From other processes, say 'participant signal #infoMessage with: msg '
	 to get text into the status line."

	DropAllButLastEvent signal.
	statusLine info: message.
	self waitTick.! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 11/22/2007 18:03'!
onInitializePreferences
	| |
	<on: initializePreferences>
	
	"Raise each value-changed signal in the preferences, 
	 to fire all the value-change listeners.
	 This is a trampoline, done indirectly from onPreferencesChanged."
	preferences signalChanges.
	! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 8/4/2010 16:46'!
onIsMeetingLeaderChanged
	<on: isMeetingLeaderChanged>
	"This is triggered (from people-list) when org-services changes my leader state.
	 Propagate the change.

	 A become-leader request is first sent to org-services (#beMeetingLeader:),
	 the resulting change in the presence manager is handled by the users list,
	 which uses this method to update the client as a whole.  The leader client
	 is responsible to pass the change on to the forum island (registerLeader:, below.)"

	(self readyForEvents) ifTrue: [
		isMeetingLeader ifTrue: [
			meetingState ref future registerLeader: (self userLogin).
			avatar replica future isLeader: true.
		] ifFalse: [
			meetingState ref future releaseLeader: (self userLogin).
			avatar replica future isLeader: false.
		]
	] ifFalse: [
		"Get this right... keep the forum-island and org-services consistent."
		QLogger logWarning: 'Meeting leadership transfer retry.'.
		(Delay forMilliseconds: 500) wait.
		session ifNotNil: [ 
			self signal: #isMeetingLeaderChanged
		].
		^ self
	].

	isMeetingLeader 
		ifTrue: [self playSoundEffect: #online] 
		ifFalse: [self playSoundEffect: #offline].
	self updateCanSpeakInMeeting.
	self updateMobility.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 8/8/2009 03:08'!
onIsPrivateTextBlockedChanged: bBlocked
	<on: isPrivateTextBlockedChanged in: meetingState>
	"The meeting control flag to block imports and scene editing has been changed."
	bBlocked ifTrue: [
		QLogger logInfo: 'Private Text is disallowed (meeting in progress).'.
	].
	
	(self isMeetingLeader or: [ bBlocked not ]) ifTrue: [
		orgServices ifNotNil: [	
			 orgServices setMeetingPrivateTextBlocked: bBlocked
		]].
	
	"TBD ... the implementation will disable send on any open private-dialogs,
	 and disable access to it from people menus, etc.
	 Most of that is on-the-fly in those tools, but some parts need broadcast."
	self applyUserRole.! !

!QwaqParticipantUI methodsFor: 'actions' stamp: 'ar 4/16/2011 17:48'!
onIslandAction: action url: svcURL author: actorSessionID
	"A signal notifying us that someone - maybe me - has made a formal change to the given forum."
	(servicesPane entryForUrl: svcURL) ifNotNilDo: [ :e |
		e lastChangeTime: (Time utcSeconds).
	] ifNil: [ 
		QLogger logInfo: 'Action ', action, ' for unknown url ', svcURL
	].

	! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 8/10/2008 18:44'!
onIslandSaved: island postcard: postcard author: author
	"The given island has been saved."

	(servicesPane entryForPostcard: postcard) ifNotNilDo: [ : entry |
		entry lastSaveTime: (Time utcSeconds).
		entry lastAuthor: author.
		self statusStatus: '', entry name, ' has been saved.'.
	] ifNil: [
		self statusStatus: '', postcard shortDisplayString, ' has been saved.'.
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 8/16/2008 17:07'!
onKeyIconEnter
	<on: mouseEnter in: keyboardIcon>
	self statusInfo: 'Click to exit typing mode. <Escape> to toggle.'.
	keyboardIcon dropShadowEnabled: true.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 1/27/2008 02:16'!
onKeyIconLeave
	<on: mouseLeave in: keyboardIcon>
	keyboardIcon dropShadowEnabled: false.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/8/2008 16:44'!
onKeyStrokeBar: evt
	<on: keyStroke in: controlBar>

	mainView ifNil: [ ^ self ].
	(evt keyCharacter = Character escape) ifTrue: [
		mainView keyboardDriving: mainView keyboardDriving not
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 1/27/2008 01:44'!
onKeyboardFire
	<on: fire in: keyboardIcon>
	
	mainView ifNotNil: [ 
		mainView keyboardDriving: true.
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 11/21/2008 14:42'!
onKeyboardModeChanged

	<on: keyboardDrivingChanged in: session>

	mainView keyboardDriving ifFalse: [
		self add: keyboardIcon
	] ifTrue: [ 
		self remove: keyboardIcon.
	].

	navPanel visible: (preferences doNavPanel and: [mainView keyboardDriving]).
	mainView keyboardDriving ifTrue: [
		avatar stopActivity: #Typing
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 12/4/2008 12:05'!
onLatencyTest: startTime
	<on: latencyTest>
	uiLatencyStats add: (Time millisecondsSince: startTime).
	(Delay forMilliseconds: 250) wait.
	self startScript:[self signal: #latencyTest with: Time millisecondClockValue].! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'das 9/10/2009 14:24'!
onLaunchPythonApp: pyPanelWorld
	"Installs a server-based QPZ with the given appName (Module.AppName).
	If the application is already installed, we compare the existing version with
	the one from the server and only install it if they are different."
	| appName |
	<on: launchPythonApp>
	appName := pyPanelWorld pyAppName.
	"Don't even attempt to download modules into the Qwaq namespace since
	Python will not be able to find it anyway if it's not in the local scripts/Qwaq 
	directory."
	self installServerQPZ: appName.
	pyPanelWorld launchApp.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 11/24/2009 19:18'!
onLeaderLoginChanged: login
	<on: leaderLoginChanged in: meetingState>
	"Make this basic change more easily observable by other tools."
	self signal: #leaderLoginChanged with: login.

	login ifNotNil: [
		self stopTests.		"So that when meeting starts, voice-transport testing stops."
	].
	self onAudienceIsMutedChanged: (meetingState audienceIsMuted).! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 11/23/2009 10:37'!
onLogin: userName password: password
	"Phase 1 login - transition from user/password login, to the org/service chooser.
	 The login panel has authenticated with the forums host.
	 It has a service manager, but the service manager may not yet know a realm.

	 We hand this off directly to the service chooser."

	self checkForValidEnvironment. "double-check just in case"

	"Note that we don't set the participant's own ServiceManager here...
	 we just pass to the serviceChooser.  
	 It will choose a realm, and give us the <ready> serviceManager."

	login serviceManager ifNotNilDo: [ : mgr |
		serviceMgr := mgr.
		preferences := (QUserPreferences forUser: (login user value) inClient: self).
		"Fetched here, but will be activated in setupPreferences:"
		serviceMgr := nil.
		].

	serviceChooser setServiceManager: login serviceManager.
	self showServiceChooser.
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 4/16/2011 17:33'!
onLoginFailure: user reason: reasonSymbol extra: extraInfo from: aController
	| yours msg |
	<on: loginFailure>
	"This is the report when some other user fails entry due to version mismatch.
	 The router itself broadcasts a notification, and we pass it along."

	QLogger logInfo: '', user, ' failed to enter ', (aController serviceName), ' due to ', reasonSymbol.

	self readyForEvents ifTrue: [
		| nm |
		"Since leader-moderated meetings are about keeping the audience experience predictable,
	 	 if we know who to tell about this, just tell the leader. "
		(self isMeetingInProgress and:  [ self isMeetingLeader not ]) ifTrue:[ ^ self ].

		nm := (aController serviceName = (self currentForumName)) 
					ifTrue: [ 'this Forum' ]
					ifFalse: [aController serviceName].

		reasonSymbol = #versionMismatch ifTrue: [
			yours := QwaqVersion current infoString.
			msg := 	('Someone has just tried to join using a different client version.\\', 
					  user, ' has tried to enter ', nm,'\', 
					'using ', ( (yours < extraInfo) ifTrue: [ 'a newer' ] ifFalse: [ 'an older'] ),
						' version of the software.\\',
					'They used:\	', (extraInfo), '\',
					'You have:\	', (QwaqVersion current infoString),
					'\\All participants in a meeting must use the same version.') withCRs.
			self inform: msg.
		]].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 8/1/2008 19:26'!
onMainWindowCloseRequest
	"User pressed the big red X in the main window"
	<on: mainWindowCloseRequest>
	self exitConfirming: true.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/4/2009 15:08'!
onMeetingPlaceChanged: qPlace
	"The leader-assigned place is a field in the meeting state
	 that contains the place where the leader wants everyone to be placed."
	<on: meetingPlaceChanged in: meetingState>
	qPlace ifNotNil: [
		QLogger logInfo: 'Meeting place is now ', (qPlace label ifNil: [ ' a place' ]).
	].

	"Observable from tools."
	self signal: #meetingPlaceChanged with: qPlace.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 4/1/2010 18:26'!
onMouseLeave
	<on: mouseLeave>

	"Attempt to keep keyboard driving workable when the mouse drifts out of the window.
 	 This may adversely affect focus in input windows, but the net should be positive.
	 since stray laptop-user loss of driving/control ought to be reduced.
	 The main culprit is mouse-leave induced when the mouse trails across the title bar,
	 which renders the ui inert to key input even tho it remains the top window on the app screen."

	mainView ifNotNil: [
		hand keyboardFocus: viewPanel
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/11/2009 18:41'!
onMouseOverRecordingIndicator
	<on: mouseEnter in: recordingIndicator>

	"If recording is on, the indicator is visible, and can explain itself..."
	recordingIndicator visible ifTrue: [
		| recorder msg |
		msg := nil.
		recordingUsers ifNotEmpty: [
			msg := 'Session being recorded by ', recordingUsers anyOne, '.'
		].
		statusLine floatingMessage: msg
				   at: nil 
				   extraSeconds: 1.
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/11/2009 19:05'!
onMouseOverWebcastIndicator
	<on: mouseEnter in: webcastIndicator>

	"If recording is on, the indicator is visible, and can explain itself..."
	webcastIndicator visible ifTrue: [
		| broadcaster msg |
		msg := nil.
		broadcastingUsers ifNotEmpty: [
			msg := 'Session being webcast by ', (broadcastingUsers anyOne), '.'.
		].
		statusLine floatingMessage: msg
				   at: nil 
				   extraSeconds: 1.
	].! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 7/10/2009 08:52'!
onNavPanelPrefChanged
	<on: doNavPanelChanged in: preferences>

	self layoutEverything.
	self updateMobility.
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 5/15/2007 13:38'!
onOKPreferences
	<on: OK in: prefsPanel>

	self remove: prefsPanel.
	self menu enable.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'ar 7/12/2011 12:54'!
onOpen
	| crashDumpPath |
	<on: open>
	Smalltalk windowLabel: (self appName).

	"FIXME: Suspend the standard EToysProject blueButton script which puts up an in-world halo that we can't get rid of otherwise. This should really be done by not having an EtoysProject (and no associated blueButton script) but for now this workaround do the trick just as well."
	self project pauseScript: #onBlueButtonDown.
	"Similar problem here - the mouseDown script can screw up z-order (not sure exactly why...) and dragging wouldn't work anyway since it ignores OpenGL rendering."
	self project pauseScript: #onMouseDown.

	"If we're in the sealed app handle system events and errors here."
	QwaqVersion current isSealedApplication 
		ifTrue:[
			SystemErrorHandler default: self.
			SystemEventHandler default: self
		].

	projectMorph ifNotNil:[
		WorldState addDeferredUIMessage:[
			projectMorph noRedraw: false.
			projectMorph changed.
			self signal: #completeAfterShowingScreen].
		self waitUntil: #completeAfterShowingScreen.
		Display forceToScreen.].

	"Check if we are recovering from a crash"
	self isSealedApplication ifTrue:[
		crashRecovery := (crashDumpPath := self crashDumpPathOrNil) notNil].
	crashRecovery ifTrue:[
		"We are recovering from a VM crash, take appropriate action"
		[ 	FileDirectory cacheDirectory directories do: [ :dir | dir deleteLocalFiles].	"Per-login caches"
		 	FileDirectory cacheDirectory deleteLocalFiles
		 	]  on: Error do:[:ex| ex return].
		"Send a crash dump if necessary"
		self isSealedApplication 
			ifTrue:[self startScript: #submitCrashDump: withArguments: {crashDumpPath}].
		"And tell the user know about the problem"
		self displayInitialScreen.
		login disable.
		self inform: self appName, ' has recovered from a crash.\\These actions have been taken:\\	* Antialiasing disabled in Preferences.\	* Local resource cache cleared.\' withCRs.
		login enable.
	].

	proxyDlg loadProxySettings.

	"Process command line arguments"
	self processCommandLine.

	"Check to see that we are running in a sane environment;
	 we do this after checking command-line args in case OpenGL is suppressed."
	self checkForValidEnvironment.

	Smalltalk platformName = 'Mac OS' ifTrue:[
		"XXXX: Work around an issue with Mac switchable graphics.
		For some reason, doing a 'full' graphics initialization here causes
		a proper switch to the discrete graphics whereas doing it later
		(in the 'real' call to checkForValidFullEnvironment) the OS hands us
		a context for the integrated (Intel) graphics but then switches to the
		discrete (NV) graphics which then causes errors when attempting to
		create OpenGL contexts when we actually need them."
		self checkForValidFullEnvironment.
	].

	"Start the listener to update the CroquetView with events"
	self startScript: #onUpdateCroquetView when:{
		self hand. #mouseDown.
		self hand. #mouseMove.
		self hand. #mouseUp.
		self hand. #keystroke.		"Morphic spelling, but the hand emits it (at least unsealed).."
		self hand. #keyStroke.		".. the hand never emits these currently (March/08)"
		self mainView. #activeSpaceChanged.
	}.

	self startScript: #onGlobalMouseMove: when: { hand . #mouseMove }.
	self startScript: #onGlobalMouseDown: when: { hand . #mouseDown }.
	self startScript: #onGlobalMouseUp: when: { hand . #mouseUp }.
	self startScript: #onGlobalKeyDown: when: { hand. #keyDown }.
	self startScript: #onGlobalKeyUp: when: { hand . #keyUp }.
	self startScript: #onGlobalKeyStroke: when: { hand . #keystroke }.

	self signal: #latencyTest with: Time millisecondClockValue.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 11/12/2009 12:45'!
onOverWarpPlace: placeOrNil
	"A run-script to indicate when the avatar has intercepted an over-event
	 that constitutes a clickable shortcut warp-to.
	 This is invoked through runScript: from island events and must not block. 
	 We suppress warps in full-editing mode to allow freer use of the mouse to click-select, etc."

	( self allowedToMove and: [ self isEditingTemplate not ] ) ifTrue: [
		self mouseCursor: (placeOrNil ifNil: [nil] ifNotNil: [ QStyles warpPlaceCursor ]).
		placeOrNil ifNotNil: [
			self signal: #floatMessage with: (placeOrNil label).
		]].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/27/2008 15:25'!
onPPToggleFire	
	<on: fire in: ppToggle>
	"We're using an image button to serve as a toggle..."
	ppToggle selected: (ppToggle selected not).! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 9/8/2009 21:46'!
onPPToggleSelectedChanged
	<on: selectedChanged in: ppToggle>
	"The people-places view has been hidden or revealed..."
	self layoutEverything.
	ppToggle balloonHelp: ( 
		ppToggle selected
			ifTrue: [ 'Hide People & Landmarks' ]
			ifFalse: [ 'Show People & Landmarks' ] ).! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bobw 8/2/2010 10:36'!
onPluginEvent
	"The browser plugin (currently ActiveX Control) has signaled the input 
	 process with a event.  Currently a stub method which does nothing."
	<on: pluginEvent>
	! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'sn 5/13/2011 14:49'!
onPointOfViewPrefChanged
	<on: pointOfViewChanged in: preferences>

	"The avatar could do this, since it has a handle on the preferences,
	  but doing it here keeps the prefs handlers together."

	self changeViewpointValue: (preferences povValue).

	"Make the dynamic menu match."
	(self menu itemNamed: 'View' translateMe) subMenu menuItems asArray
		from: 1 to: (QUserPreferences viewpointTable size) do: [ :item|
		item checked: (item label = preferences pointOfView).
	].
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 1/9/2010 13:20'!
onPreferencesChanged
	<on: preferencesChanged>
	"Fire the changes markers in the preferences, 
	 to fire all the value-change listeners.
	 This is done after loading preferences to activate the initial choices
	 through the change-listener code.
	 Done with a deferral (an indirect signal)
	 to let setup of the individual value-change listeners take place
	 before we signal the value-change events."
	preferences ifNotNil: [ self signal: #initializePreferences ]
	! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 2/17/2009 09:55'!
onProfile	
	"Send me #profile to get a 30-second system profiler."
	<on: profile>
	self systemProfiler: 30.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 10/22/2009 12:48'!
onPyDropCard: fileName island: island whenComplete: aBlock
	"Trampoline for Python"
	| card file |
	<on: pyDropCard>
	file := (FileStream readOnlyFileNamed: fileName).
	card := session importDocument: file island: island.
	card ifNotNil:[aBlock value: card].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 9/29/2008 14:11'!
onPyLoad3DFile: fileName from: panelDir whenComplete: aBlock
	"Trampoline for Python"
	| fname model |
	<on: pyLoad3DFile>
	fname := panelDir pathFromURI: fileName.
	"For compatibility: Also look relative to the install location"
	(FileDirectory default fileExists: fname)
		ifFalse:[fname := FileDirectory default pathFromURI: fileName].
	model := session load3DFile: fname.
	model future objectName: fileName.
	aBlock value: model.

! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'das 2/13/2009 15:40'!
onPySaveForum
	"Trampoline for Python"
	<on: pySaveForum>

	self saveForum.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'das 1/2/2010 15:34'!
onPyUploadDocumentData: data shortName: fileName

	<on: pyUploadData>

		session documentManager uploadDocumentData: data 
			shortName: fileName 
			collisions: #clobber owner: false.
	
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 10/22/2009 14:27'!
onPyUploadTexture: fileName extension: extension island: island whenComplete: aBlock
	"Trampoline for Python"
	| fname aFile doc |
	<on: pyUploadTexture>
	fname := fileName.
	"Upload file and assign shared name.
	 Note that these are textures, but we put them into the sharedDocuments system,
	 since they are user-supplied images, rather than model-support resources."
	aFile := FileStream readOnlyFileNamed: fname.
	doc := session importImageAsTexture: aFile shrinkFit: true extension: extension island: island.
	"doc := self documentManager uploadDocumentFile: aFile collisions: #leaveOriginal ."
	aBlock value: doc.
	
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/11/2010 12:03'!
onReady
	| arrival |
	<on: ready>

	"Signalled when the active space is ready,
	 after any teleport including the first login entry.
	 We use a signal for this, so that all UI-setup handlers get their chance
	 before the controls go live."

	"Is this our first entry into this forum in this session?"
	arrival := firstArrivalPending.
	firstArrivalPending := false.

	"Make sure that no logout-shortened timeouts leak into normal running."
	self useNormalTimeouts.

	self menu visible ifFalse: [
		"If we're arriving in the first space, make the UI ready."
		self showFixedPanes.
		viewPanel visible: true; enable.
		self graphic: nil.
		self layoutEverything.
		peoplePlacesSash borderColor: (QColors paneBorderColor). 
		videoAreaSash  borderColor: (QColors paneBorderColor). 
			" Above line is needed (without it, the sash border
	 		  comes up wrong on first login) but I really don't know why;
	 		  something changes the ppSash player border color without doing borderColor:"
	].

	self readyForEvents ifTrue: [
		controlBar comeToFront.
		self menu comeToFront.
		self resetActivities.
		self audioSession ifNotNilDo: [:sess | sess audioInterface ifNotNil: [
			"That check mostly prevents launching a sure-to-fail device check on linux,
	 		 where primitive failures always prevent creating the audio interface."
			sess deviceManager future initializeDevicesFromPreferences.
			sess deviceManager future startCheckingForDeviceChanges.
		] ].

		"We've probably changed the CB layout."
		statusLine setStyleForWindowWidth: (self width).

		"Rudimentary - open the welcome-booklet app if one is configured."
		arrival ifTrue: [ self showForumWelcome ].

		"Check for a pending session recording that wasn't processed"
		recorderPane checkForCurrentRecording.
	].
	self isMeetingLeader: (orgServices notNil 
						and: [ orgServices meetingLeaderSessionID = self sessionID ]).

	self setWantsToSpeak: false.
	self informIfEnteringMeeting: (serviceSvc url).

	login forumGoal: nil. "allow  general teleports."
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 3/5/2008 17:32'!
onRecordLocation
	<on: recordLocation>
	"Writing the locator record is a bit blocky, so we give its own trigger/timeslice."
	DropAllButLastEvent signal.
	avatar recordLocation.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 12/11/2009 18:04'!
onRecordingSessionChanged: replica
	"Someone started or stopped recording"
	| userName recording |
	<on: isRecordingSessionChanged in: activeSpace>
	userName := replica get: #publicName.
	recording := replica get: #isRecordingSession.
	recording ifTrue:[
		recordingUsers add: userName.
	] ifFalse:[
		recordingUsers remove: userName ifAbsent:[].
	].
	self statusInfo: userName, 
		(recording ifTrue:[' started '] ifFalse:[' stopped ']), 
		'recording this session'.

	recordingIndicator visible: (recordingUsers notEmpty).
	webcastIndicator visible: (broadcastingUsers notEmpty).
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'jcg 3/3/2010 14:45'!
onRemoveAllowedToSpeak: login
	<on: removeAllowedToSpeak in: meetingState>
	login = self userLogin ifTrue: [
		self controlBar askToSpeak selected: false.
		self controlBar leaderAllowingToSpeak: false.
		self updateCanSpeakInMeeting.
	].! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/25/2009 21:50'!
onRemoveWantsToSpeak: login
	<on: removeWantsToSpeak in: meetingState>
	login = self userLogin ifTrue: [
		self controlBar askToSpeak selected: (meetingState hasLoginPermissionToSpeak: login).
		].
	! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 5/4/2010 18:57'!
onRender
	"Do one render/step cycle"
	| box loadmsg |
	<on: render>
	DropAllButLastEvent signal.
	"FIXME: Set the current hand explicitly to the primary hand of this project since render may be signaled without an associated hand (via CProjectMorph>>step) which will cause problems later on when we try to refer to that (non-existing) hand"
	Processor activeProcess hand: Processor activeProject primaryHand.

	self checkHeartbeatLag.

	session ifNil:[^nil].
	session step. "this can take time"
	session ifNil:[^nil].
	self documentManager ifNil: [ ^ nil ].
	usersPane lastActivity: (session lastEventTime).

	loadmsg := self documentManager resourceLoadingStatus.
	loadmsg ifNotNil: [ self signal: #infoMessage with: loadmsg ].
	(flowControlFlag or: [ loadmsg notNil ]) ifTrue: [
		| msecs|
		session ifNil: [ ^ self].
		msecs := 25 + ((session percentRenderTime max: 0.0) * 80) rounded.	 "Longer if render is hogging"
		exitTripped ifFalse: [(Delay forMilliseconds: msecs) wait ]	 "Free cycles for others."
				     ifTrue: [ ^ nil ].
	].

	session ifNil: [ ^ nil ].	"In case we overlapped a logout."
	box := self globalBounds.
	box := box translateBy:  (window boundsInWorld origin + window borderWidth).
	lastBounds ~= box ifTrue:[
		self waitTick.		"For tweak extent changes to propagate."
		session ifNil: [ ^ nil ].
		session renderLock critical: [ session globalBounds: box ].
		lastBounds := box.
	].

	"Try to decouple OpenGL render from Tweak render by first updating Tweak 
	and then invoking OpenGL render"
	mainView ifNil: [ ^ nil ].
	session ogl ifNotNil: [
		session renderLock critical: [
			self updateTweak.
			session ogl ifNil: [ ^ nil ].
			mainView ifNil: [ ^ nil ].
			mainView camera viewAngle: self getCurrentViewAngle.
			session ogl maxPortalDepth: 2.  "Enough to see reflections but not tunnels."
			session render.
		].

	] ifNil: [
		"Pretend - block for a bit; let Croquet things advance."
		exitTripped ifFalse: [ (Delay forMilliseconds: 10) wait ].
	].! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'jjl 8/27/2008 09:17'!
onReturnPrepDone: target atTransform: aTransformOrNil

	<on: returnPrepDone>

	self teleportTo: target atTransform: aTransformOrNil.! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'bgf 1/8/2009 15:40'!
onSendTimout: secs from: aController
	"Notify the participant about a send timeout in the given controller."
	<on: sendTimeout>
	self statusInfo: 'Network timeout: ', secs, ' second(s) lag'.

	"Latency stats don't generally get updated during connection timeout.  
	 We want the svcCount readout color to mean 'connection goodness'
	 in the user's mind... so on connection timeout, make it red."

	controlBar updateNetHealthLabel: 1000.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'hrs 4/7/2009 14:34'!
onServiceLogin: svcListEntry 
	"Phase 2 of initial-login... a service has been chosen;  go there.
	 May fail, or be canceled by the user, in which case informs are done,
	 and the result s false.   If the login seems to complete, returns true."
	| t |
	self checkForValidFullEnvironment ifFalse: [^serviceChooser thinClient: true].
	#none "ProfilerPreference" "#QSystemProfiler" "#QVMProfiler"
		caseOf: {[#QSystemProfiler] ->
					[QSystemProfiler spyOn:
						[ t := [ self onServiceLogin: svcListEntry usingIslandFiles: nil ] timeToRun]].
				  [#QVMProfiler] ->
					[(Smalltalk classNamed: #QVMProfiler) spyOn:
						[ t := [ self onServiceLogin: svcListEntry usingIslandFiles: nil ] timeToRun]]}
		otherwise:
			[t := [ self onServiceLogin: svcListEntry usingIslandFiles: nil ] timeToRun].
	QLogger logInfo: 'Start service ', svcListEntry name, '  ', t, 'ms'.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'ar 4/16/2011 18:59'!
onServiceLogin: svcListEntry usingIslandFiles: islandFiles

	"Phase 2 of initial-login... a service has been chosen;  go there.
	 May fail, or be canceled by the user, in which case informs are done,
	 and the result s false.   If the login seems to complete, returns true.

	 IslandFiles is usually nil; if not-nil, it is a dictionary mapping islands to persistence dirEntries
	 on the service manager, and will be used to de-persist a chosen saved-version of the forum."

	svcListEntry ifNil: [
		"Safety check... callers should not do this."
		QLogger logError: 'Login failed - service is absent' from: self.
		self signal: #serviceLoginFailed.	"For robot bailout."
		self inform: 'You must select a forum\to log into.' withCRs.
		^ false
	].

	self statusInfo: 'Entering  ', (svcListEntry name).
	self serviceManager: login serviceManager.
	serviceMgr ifNil: [ ^ false 	"Cancelled or reconnected. " ].
	svcMgrCache at: serviceMgr orgUID put: serviceMgr.

	firstLaunch := false. "now considered compromised in case something below goes wrong"

	Smalltalk windowLabel: (self appName), '  |  ', (serviceMgr currentOrg name), '  |  ',
							(svcListEntry name).

	"One last-ditch attempt to prevent overlapping login confusion...
	 try to prevent the user from getting in using non-default island files
	 if the forum has been activated by someone else in the meantime.)"
	(islandFiles notNil and: [ svcListEntry isActive ]) 
	  ifTrue: [
		self inform: 'The forum ', svcListEntry name, '\now has active users, and cannot be reverted.' withCRs.
		^ false.
		].

	"The activation code should signal CProgressCancel for issues 
	 that have already been shown to the user (such as when the user cancels)
	 and Error for problems that do need an appropriate error dialog raised."
	[
		login isLocal ifFalse: [ self cancelLocalProvider ].
		self setupService: svcListEntry usingIslandFiles: islandFiles.
		self invalidate.
		self startScript: #loginEntryComplete when: {self . #ready }.

	] on: Error, CProgressCancel, QwaqRestartException
	  do: [:ex |
			self stopScript: #loginUrl:forumPreviewChanged:byUser:.
			(ex isKindOf: Error) ifTrue: [

				self signal: #serviceLoginFailed.	"For robot bailout."

				(ex description includesSubString: 'Access to the Forum:') ifTrue: [
					self inform: ex messageText
				] ifFalse: [
					QLogger logError: ex from: self.
					self inform: 'A problem has prevented login.', String cr, ex messageText.
				].
			] ifFalse: [
				self statusInfo: 'Login canceled...'.
			].
		self logoffConfirming: false backToServiceChooser: true.
		^ false
	].

	^ true
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 5/11/2010 12:08'!
onSoundEffect: aSymbol atFrame: aFrame

	<on: soundEffect in: activeSpace>

	(preferences doSoundEffects and: [ self isAway not ])
	 ifTrue: [
		self audioSession ifNotNilDo: [ :audio |
			(aFrame isValidFarRef and: [aFrame isResolved])
				ifTrue: [
				 	audio playSoundEffect: aSymbol 
						   at: (aFrame get: #globalPositionReal).
				]]]! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 10/26/2006 20:00'!
onSoundPrefChanged
	<on: doSoundEffectsChanged in: preferences>

	self audioSession ifNotNil: [
		self audioSession useSoundEffects: (preferences doSoundEffects).
	].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'jcg 5/11/2010 18:27'!
onSpaceChange
	"Fires when the avatar has entered a new space, after replica+costume injection."
	<on: activeSpaceChanged in: mainView>

	exitTripped ifTrue: [ ^ self ].
	activeSpace ifNotNil:[
		"Reduce the heartbeat of the outgoing island to reduce network traffic"
		activeSpace controller ifNotNilDo:[:cc|
			(cc respondsTo: #latencyStats) ifTrue: [
				latencyPane stopWatching: cc jitterStats.
				latencyPane stopWatching: cc latencyStats.
			].
			cc userIsInSpace: false.
			cc heartbeat: cc class reducedHeartbeat].
		"Notify embedded apps about leaving the space they are in"
		session onSpaceLeave: activeSpace.
	].
	activeSpace := mainView activeSpace.
	activeSpace ifNil: [
		self menu disable.
		self controlBar disable.
		^ self
	].
	"Set the heartbeat of the incoming island to the full rate"
	activeSpace controller ifNotNilDo:[:cc| 
		cc userIsInSpace: true.
		cc heartbeat: cc class defaultHeartbeat.
		(cc respondsTo: #latencyStats) ifTrue: [
			latencyPane startWatching: cc jitterStats.
			"Larger window duration by default on the client so we have some context"
			cc latencyStats windowDuration: 60000.
			latencyPane startWatching: cc latencyStats.
		]].

	"Update the current service for the space we're in."
	serviceSvc := (self urlCacheAt: (session serviceUrlForIsland: (activeSpace island)) 
						ifAbsent: [serviceSvc]).
	QActivityCapture action: #EnterForum details: (self currentForumName ifNil: [ 'Unknown' ]).

	"Window labelling."
	exitTripped ifTrue: [ ^ self ].
	self playSoundEffect: #changeSpace.
	"Connect listeners to avatar replica in new space."
	activeSpace wait.
	exitTripped ifTrue: [ ^ self ].
	self updateVoiceRelay.
	self updateStatusRelay.
	self updateRecordingUsers.

	"We make new ClientSpeakers when we teleport...
	 so we can just forget the loopback state after a space jump;
	 the QClientSpeaker for the old space's instance of my QSpeaker unsubscribes as it dies."
	roundTripSound := nil.	

	self updateUserRole.
	exitTripped ifTrue: [ ^ self ].

	self signal: #ready.
	self signal: #recordLocation.

	"Notify embedded apps about entering the space they are in"
	session onSpaceEnter: activeSpace.! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'jcg 3/3/2010 14:47'!
onSpeakerListsCleared
	"The meeting state has reset wantsToSpeak and allowedToSpeak."
	<on: speakerListsCleared in: meetingState>
	
	controlBar askToSpeak selected: false.
	controlBar leaderAllowingToSpeak: false.
	self updateCanSpeakInMeeting.
! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 1/16/2008 14:32'!
onStatusMessage: message
	<on: statusMessage>

	"From other processes, say 'participant signal #statusMessage with: msg '
	 to get text into the status line."
	DropAllButLastEvent signal.
	statusLine status: message.
	self waitTick.! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 3/13/2010 09:07'!
onUnready
	<on: unready>
	"Followup after going or returning to login or lobby.
	 Clear up any dangling activities that might hold indirect island references."

	self stopScript: #forkEntryZoom.

	undoRedoList reset.
	(self propertyValueAt: #documentManager) ifNotNilDo: [  : dm | dm discardPrefetches ].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/8/2008 19:13'!
onUpdateCroquetView
	"Explicitly started to notify the CroquetView about events that are never passed to it so it doesn't slow down rendering when we have a preference dialog open or so."

	session ifNotNil:[session updateLastEventTime].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 11/3/2006 16:27'!
onUploadEnd
	"Inform the user that we are done uploading a snapshot."
	<on: uploadEnd>
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'bgf 3/19/2010 21:42'!
onUploadProgress: amount for: clientName
	"Notify the user about our upload progress."
	<on: uploadProgress>
	clientName ifNil:["we're phasing this out but for now..."
		self statusInfo: 'Uploading for new user: ',(amount * 100) truncated, '% done'.
	] ifNotNil:[
		self statusInfo: 'Uploading for ', (QStyles simplerLoginName: clientName),': ',(amount * 100) truncated, '% done'.
	].
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'bgf 3/19/2010 21:42'!
onUploadStart: clientName
	"Inform the user that we are currently uploading a snapshot which could take a while."
	<on: uploadStart>
	clientName ifNil:["we're phasing this out but for now..."
		self statusInfo: 'Uploading for new user: 0.0% done'.
	] ifNotNil:[
		self statusInfo: 'Uploading for ', (QStyles simplerLoginName: clientName),': 0.0% done'.
	].
! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 6/28/2009 16:26'!
onUserInterrupt
	"Handle a user interrupt"
	| pass hash devMenu |
	<on: userInterrupt>
	pass := UIManager default requestPassword: 'Enter the interrupt password:'.
	pass ifNil:[^nil].
	pass = 'x-features' ifTrue:[
		self toggleExperimentalFeatures.
		^self inform: 'Experimental features have been ',
				(preferences enableExperimentalFeatures ifTrue:['enabled'] ifFalse:['disabled']).
	].
	pass = 'dev-menu' ifTrue:[
		(self menu itemNamed: 'Developer') ifNil: [
			devMenu := self menu addItem: 'Developer' after: (self menu itemNamed: 'Extras').
			devMenu subMenu: self developerMenu.
		].
		^self].

	hash := TCryptoRandom md5HashMessage: pass.
	hash = InterruptPassword ifFalse:[^self inform:'Incorrect password'].
	doOpenGL ifTrue:[self toggleOpenGL].
	"fork the debugger off to allow repeated interrupt keys to work properly"
	[Processor activeProcess debug: thisContext title: 'User Interrupt'] fork.
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 8/21/2008 15:21'!
onUserIsIdle
	"The main view has fallen into render-idle;
	 take the opportunity to refresh resources.
	 (Also triggered when the user steps Away, as a diagnostic trick.)"

	<on: userIsIdleChanged in: controlBar>

	DropAllButLastEvent signal.

	activeSpace ifNotNil:[activeSpace controller ifNotNilDo:[:cc| cc userIsIdle: controlBar userIsIdle]].

	flowControlFlag ifTrue: [ ^ self ].		"Already busy loading summat"
	
	(controlBar userIsIdle or: [self isAway]) ifTrue: [
		self statisticsPane putActivityRecords: true.		"Get latest data up"
		self readyForEvents ifTrue: [
			QLogger logInfo: 'Idle refresh.'.
			session updateResourceType: (TFrameThumb resourceType) in: (activeSpace island).
			session updateResourceType: (TFileTexture resourceType) in: (activeSpace island).
			session updateResourceType: (QSharedDocument resourceType) in: (activeSpace island).
		]].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'jjl 8/26/2008 11:54'!
onUserNeedsInformation: info

	<on: userNeedsInformation>

	self inform: info.! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 5/29/2008 10:28'!
onVideoDisabledChanged
	<on: videoDisabledChanged in: preferences>
	
	preferences videoDisabled ifTrue: [
		self disableVideoSession
	] ifFalse: [
		self enableVideoSession
	].! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/14/2010 17:32'!
onVideoItemsChanged
	<on: itemsChanged in: webcamPane>

	controlBar visible ifTrue: [
		webcamPane numItems > 0 ifTrue: [
			self showWebcamPane.
		] ifFalse: [
			self hideWebcamPane.
		]].

! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 11/11/2006 16:09'!
onViewAnglePrefChanged
	<on: viewAngleChanged in: preferences>

	"Trigger a rest of the ogl view and camera."
	lastBounds := nil.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 8/11/2009 13:55'!
onWelcomeAppResolved: anEmbeddedAppRef
	"Signal-handler to get the resolved far-ref for the welcome application
	 back into a tweak processing context."
	<on: welcomeAppResolved>

	"Allow time for appExtent to propagate to the embeddable display.
	 But this is a tradeoff with menu-responsiveness, so not too much."
	(Delay forMilliseconds: 100) wait.

	"Only open it if we're still in that forum.!!"
	(self readyForEvents 
	 and: [ anEmbeddedAppRef isValidFarRef ]	
	 and: [ anEmbeddedAppRef island = activeSpace island ]) ifTrue: [
		self showFloatingAppPane: anEmbeddedAppRef withAlignment: #topLeft.

		"If a sound-change alert was raised during entry, keep it the priority."
		controlBar soundDialog ifNotNilDo: [ 
			:sd | sd visible ifTrue: [ sd comeToFront ]].
	].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'jcg 10/15/2007 18:26'!
openInformDialog: informationalString label: labelString 
	<on: openInformDialog>
	CDialog
		inform: informationalString
		label: labelString
		centerAt: self globalBounds center.
! !

!QwaqParticipantUI methodsFor: 'external interfaces' stamp: 'ar 4/16/2011 17:48'!
openMP4InfoPage
	"Originally from WindowedSession popopMP4Info  ...  called out for reuse.
	 Open the web-page where we tell people the details about MP4/M4V file support."

	self 
		shellOpen: 'http://www.openqwaq.com/support/streaming-info.php' 
		documentDescription: 'user guide'
		applicationDescription: 'Web browser'
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 4/14/2010 17:53'!
openQrlForCurrentForum
	"Open a QRL info page for the current forum in the operating system default Web browser."
	self openQrlForForum: (self currentForumName) place: nil! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 4/14/2010 17:53'!
openQrlForCurrentPlace
	"Open a QRL info page for the current form/place in the operating system default Web browser."
	self openQrlForForum: (self currentForumName) place: placesPane currentStablePlaceName! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 17:33'!
openQrlForDocument: docRef
	"Open a QRL info page for docRef in the operating system default Web browser."
	| place forum |
	self hasForumPagesSupport ifFalse: [self inform: 'Forum Pages access is not enabled for your account.'. ^nil].
	place := docRef get: #identifier.
	forum := docRef get: #serviceName.
	self openQrlForForum: forum place: place.
	^place! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'hrs 5/26/2009 11:46'!
openQrlForForum: forumName place: placeNameOrNil
	"Open a QRL info page for this place in the operating system default Web browser."
	| url |
	url := self serviceMgr qrlFromForum: forumName place: placeNameOrNil.
	QActivityCapture action: #OpenForumPage details: url.
	self shellOpen: url
		documentDescription: (placeNameOrNil ifNil: [forumName])
		applicationDescription: 'Web browser'.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 10/23/2008 20:25'!
openUserGuide

	QActivityCapture countAction: #HelpMenuUserGuide.
	self openUserGuide: 'index.html'.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 7/12/2011 12:55'!
openUserGuide: fragment
	"Open to the desired place within the user guide.
	 Fragment is the tail-end of a url to the page we want (a html file uri),
	 such as 'index.html' or 'navigation.html'"

	| base |

	"If the service provider gives a URL, we use that and go to those online;
	 otherwise if we have docs in the package, use the local ones;
	 otherwise default to a guessed hosted path for online docs."

	base := (serviceMgr ifNotNil: [ serviceMgr helpUrl ]).
	base ifNil: [
		(FileDirectory default fileExists: (FileDirectory pathFromURI: 'docs/index.html')) ifTrue: [
			base := ((FileDirectory default directoryNamed: 'docs') asUrl) asString.
		] ifFalse: [
			base := 'http://', 
				"If there's no serviceMgr (e.g., not logged in yet), we have a standard place on host.
				Otherwise (serviceMgr but no hlepUrl) we've got an old server: go to qwaq.com"
				(serviceMgr ifNil: [self serviceHost, '/docs'] ifNotNil: ['www.qwaq.com']),
				'/userguide'.
		].
	].

	self shellOpen: (base, ((base endsWith: '/') ifFalse: ['/'] ifTrue: ['']), fragment)
		documentDescription: 'user guide'
		applicationDescription: 'Web browser'.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/20/2007 00:46'!
orgServices
	"Answer the orgServices of the receiver"
	<bewareOf: #orgServicesChanged>
	^self propertyValueAt: #orgServices! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/20/2007 00:46'!
orgServices: aValue
	"Modify the receiver's orgServices"
	^self propertyValueAt: #orgServices put: aValue with: #orgServicesChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/23/2008 20:07'!
orgSetupMutex
	"Answer the orgSetupMutex of the receiver"
	<bewareOf: #orgSetupMutexChanged>
	^self propertyValueAt: #orgSetupMutex! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/23/2008 20:07'!
orgSetupMutex: aValue
	"Modify the receiver's orgSetupMutex"
	^self propertyValueAt: #orgSetupMutex put: aValue with: #orgSetupMutexChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 1/3/2009 19:05'!
peoplePlacesSash
	"Answer the peoplePlacesSash of the receiver"
	<bewareOf: #peoplePlacesSashChanged>
	^self propertyValueAt: #peoplePlacesSash! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 1/3/2009 19:05'!
peoplePlacesSash: aValue
	"Modify the receiver's peoplePlacesSash"
	^self propertyValueAt: #peoplePlacesSash put: aValue with: #peoplePlacesSashChanged! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/16/2010 11:33'!
peoplePlacesSashTop: newY
	| yTop |
	"The argument is the new y coordinate, within me,
	 for the top edge of the sash between People and Places.
	 The adjustment numbers are partly gibberish to make up for odd menuBar height result."

	yTop := newY 
				min: (self controlBar top - (peoplePlacesSash height) + 1)
				max: (self menu bottom-1 ).
	
	peoplePlacesSash
		top: yTop;
		memo: (yTop / (self viewHeight)).

	placesPane 
		height: ((controlBar top - yTop + 2 - (peoplePlacesSash height)) 
				   max: (0));
		top: (controlBar top - (placesPane height) + 1).

	usersPane 
		height: (placesPane top - self menu bottom) + 4 - (peoplePlacesSash height).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/8/2008 14:33'!
performActionInSpace: anActionSelectorSymbol

	"Invoke the given action selector on the current TSpace.
      If the selector takes an argument, it is to passed the avatar replica."

	|  result |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	QActivityCapture countAction: #ActionsMenu details: (anActionSelectorSymbol asString).

	self mouseCursor: Cursor wait.
	anActionSelectorSymbol numArgs = 1 ifTrue: [
		result := activeSpace future 
			perform: anActionSelectorSymbol with: avatar replica
	] ifFalse: [
		result := activeSpace future perform: anActionSelectorSymbol
	].

	result ifNotNil: [
		result whenResolved: [
			self mouseCursor: nil.
		].
	] ifNil: [
		self cursor: nil.
	].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/8/2008 14:33'!
performActionInSpace: anActionSelectorSymbol with: anArgument

	"Invoke the given action selector on the current TSpace, supplying the given first argument.
      If the selector takes a second argument, it is to passed the avatar replica."

	|  result |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	"This is used for in-space create-object commands."
	QActivityCapture countAction: #ObjectsMenu 
					 details: (anActionSelectorSymbol asString).

	self mouseCursor: Cursor wait.
	anActionSelectorSymbol numArgs = 2 ifTrue: [
		result := (activeSpace future) 
			perform: anActionSelectorSymbol with: anArgument with: avatar replica 
	] ifFalse: [
		result := (activeSpace future) 
			perform: anActionSelectorSymbol with: anArgument
	].

	result ifNotNil: [
		result whenResolved: [
			self mouseCursor: nil.
		].
	] ifNil: [
		self cursor: nil.
	].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 2/15/2009 01:09'!
persistenceTest
	"Testing only: Make a round-trip conversion of the current space, 
	e.g., persist to a string and read back from the string. This test isn't
	quite perfect as it only replaces the children of the current space instead
	of dumping the space entirely but it's a start."
	| archive island controller pc |
MessageTally spyOn:[
	island := self activeSpace island.
	controller := island controller.

	"Save existing island"
	archive := ZipArchive new.
	session saveIsland: island asZip: archive notifying: [:v|].

	"Create new island and load it"
	island := TIsland new.
	controller island: island.
	session loadIsland: island fromZip: archive.

	"Create a fake registration for the new island"
	pc := TPostcard urlString: '../', (island get: #name).
	session addIsland: island postcard: pc.

	"And update the resources in the island so that embeded apps get mapped properly"
	session updateResourcesIn: island newSession: true.
].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 2/13/2009 14:44'!
personalShareVNC
	"Share a window with the given ID"
	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].
	session ifNotNil:[session personalShareVNC].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 7/18/2008 15:02'!
phonePane
	"Answer the phonePane of the receiver"
	<bewareOf: #phonePaneChanged>
	^self propertyValueAt: #phonePane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 7/18/2008 15:02'!
phonePane: aValue
	"Modify the receiver's phonePane"
	^self propertyValueAt: #phonePane put: aValue with: #phonePaneChanged! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 12/15/2008 11:13'!
pingConnectionInfo: info
	"Determine whether we can connect using the given connection info.
	To figure out whether we can connect, we send a ping and if we hear an ok
	response within five seconds everything is fine and we answer true."
	| mgr ok response timedOut |
	mgr := QServiceManagerV2 for: self.
	mgr info: info.
	ok := timedOut := false.
	[
		[
			response := mgr ping.
			ok := response name == #ok.
			ok ifFalse:[self error: response contentString].
		] valueWithin: 5 seconds onTimeout:[
			timedOut := true.
			self error: 'Timeout']
	] on: Error do:[:ex|
		"If we're using a proxy note the problem in proxy.log" 
		mgr proxy notNil ifTrue:[
			QProxyHandler log: 'Service manager ping error: ', info printString, ' -- ', ex description.
			timedOut ifFalse:[
				QProxyHandler logRequest: (ex signalerContext stackOfSize: 100).
			].
		].
		QLogger logWarning: 'Service manager ping error: ', info printString, ' -- ', ex description. 
		ex return].
	^ok! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'hrs 6/11/2009 13:30'!
pipeTranscriptToFile
	logFile := [FileDirectory logDirectory
					forceNewFileNamed: ('Transcript-', self userLogin, '.log')] 
				on: Error do:[:ex| ex return: nil].

	logFile ifNotNil:[
		logFile wantsLineEndConversion: true.
		logFile nextPutAll: logView panel contents asString].! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'das 7/20/2009 09:05'!
placeStreamingVideoFileCard
	| filePath doc card |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	QActivityCapture countAction: #ActionsMenu details: 'PlaceStreamingVideoCard'.

	"Find the file that we want to stream."
	filePath := QFileDialog 
					openFile: #(('MPEG-4 Videos (*.mp4)' '*.mp4')) 
					label: 'Please choose an MPEG-4 video file (H.264/AAC only)'.

	"Hand the filePath to a helper function."
	doc := self streamingVideoDocument: filePath.
	card := QFileCard forDocument: doc.
	card isTransient: true.
	card := TIslandCopier copy: card to: session activeIsland.

	"And treat the injection just like a dropped-document from the desktop."
	session dropCard: card event: nil.! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'das 7/20/2009 09:00'!
placeStreamingVideoFileCard: filePath
	"Helper function... called from #placeStreamingVideoFileCard in the Tools menu."
	| doc sharedArea |

	QNetVidServer processForClientStreaming: filePath.

	sharedArea := 'sharedDocuments/', (self simplerLogin).

	doc := (QSharedDocument new)
					ctime: (self documentManager newTimeStamp: 'Insert');
					sha: (TContentManager generateSHA: 'foobar');
					sharedName: sharedArea, '/', (filePath copyAfterLast: FileDirectory pathNameDelimiter);
					suffix: (filePath copyAfterLast: $.);
					yourself.
	^ doc.
! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 7/13/2008 15:55'!
placeWebFileCard
	| url space island card |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	QActivityCapture countAction: #ActionsMenu details: 'PlaceWebCard'.

	"Those spaces after URL force the default dialog to be a bit wider."
	url := (CDialog request: 'Enter Web Page URL                                        '
			initialAnswer: 'http://google.com').
	(url isNil or: [url isEmpty]) ifTrue: [^ self].

	space := mainView activeSpace.
	island := space island.

	"Not especially useful for users, but a backdoor to provide QService links by hand.
	 Note that 'qurl:' is swallowed here, it isn't a protocol name, just a trick."
	(url beginsWith: 'qurl:') ifTrue: [
		card := QFileCard forServiceURL: (url allButFirst:5).
	] ifFalse: [
		card := QFileCard forURL: url.
	].

	card := TIslandCopier copy: card to: island.

	"And treat the injection just like a dropped-document from the desktop."
	session dropCard: card event: nil.! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'ar 2/18/2009 18:33'!
placeXSetRootCard
	"Create and place a card for the simplest possible test of a QMS round-trip request... 
	 just set the background color."
	| space island card app |

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].

	app := QRemoteServerApp 
			name: #QRemoteXSetRoot
			extent: QRemoteXSetRoot defaultExtent
			data: { '' . '' . 'lightblue' }.  	"App-server, media stream, color "

	space := mainView activeSpace.
	island := space island.
	card := QFileCard forApplication: app.
	card := TIslandCopier copy: card to: island.
	"And treat the injection just like a dropped-document from the desktop."
	session dropCard: card event: nil.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/21/2008 10:53'!
placesPane
	"Answer the placesPane of the receiver"
	<bewareOf: #placesPaneChanged>
	^self propertyValueAt: #placesPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/21/2008 10:53'!
placesPane: aValue
	"Modify the receiver's placesPane"
	^self propertyValueAt: #placesPane put: aValue with: #placesPaneChanged! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'bgf 8/20/2009 14:13'!
playSoundEffect: aSymbol

	"Play a spatialized sound effect at the current avatar position; 
	 QSoundEffects has a table of named sounds."

	| where |

	self audioSession ifNil: [ ^ self ].

	"In case the avatar is unrealized or being removed..."
	where := avatar ifNotNil: [ avatar localPosition + (avatar lookAt * 3) ] ifNil: [ 0@2@0.1 ].
	where ifNil: [ where :=  0@2@0.1 ].

	self audioSession playSoundEffect: aSymbol at: (where).! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/17/2006 14:54'!
posReadout
	"Answer the posReadout of the receiver"
	<bewareOf: #posReadoutChanged>
	^self propertyValueAt: #posReadout! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/17/2006 14:54'!
posReadout: aValue
	"Modify the receiver's posReadout"
	^self propertyValueAt: #posReadout put: aValue with: #posReadoutChanged! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 7/29/2009 10:07'!
postPatchDoItHolder
	"This is reserved for updates that have to run code in the tweak context of the UI.
	 This method is run once, in the tweak project island, after all updates are in place.
	 (By contrast, naked do-its in update.cs files are run in global Morphic context.)
	 Your patches can add their needs into this method body.
	 If there is a chain of patches, each needs to accumulate the stuff here,
	 one patch must not clobber a previous patch's updates.  
	 (We also know in updates.xml and in the patch 000nName.st fileins,
	 what chain of updates is being applied, since the patches are explicit per version.)"



	"Leave this as the final patches-complete  log message."
	QLogger logInfo: 'Version is now ', (QwaqVersion current infoString).
	
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/27/2008 13:34'!
ppToggle
	"Answer the ppToggle of the receiver"
	<bewareOf: #ppToggleChanged>
	^self propertyValueAt: #ppToggle! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/27/2008 13:34'!
ppToggle: aValue
	"Modify the receiver's ppToggle"
	^self propertyValueAt: #ppToggle put: aValue with: #ppToggleChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/25/2006 18:04'!
preferences
	"Answer the preferences of the receiver"
	<bewareOf: #preferencesChanged>
	^self propertyValueAt: #preferences! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/25/2006 18:04'!
preferences: aValue
	"Modify the receiver's preferences"
	^self propertyValueAt: #preferences put: aValue with: #preferencesChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'eem 11/18/2008 16:32'!
preferredProfilerClass
	ProfilerPreference isNil ifTrue:
		[self profilerPreference: #QSystemProfiler].
	^Smalltalk classNamed: ProfilerPreference! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/26/2006 10:38'!
prefsPanel
	"Answer the prefsPanel of the receiver"
	<bewareOf: #prefsPanelChanged>
	^self propertyValueAt: #prefsPanel! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/26/2006 10:38'!
prefsPanel: aValue
	"Modify the receiver's prefsPanel"
	^self propertyValueAt: #prefsPanel put: aValue with: #prefsPanelChanged! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 3/21/2009 17:38'!
preloadFeatureEntitlements

	"Just forces early efficient checks of forums.properties for all the EnabledFeatures keys
	 we care about, so that later checks against them
	 don't introduce random delays in (e.g.) menu appearance."

	|  fullKey flag dict |
	serviceMgr ifNil: [ ^ self ].
	[
		dict := serviceMgr matchProperty: 'EnabledFeatures.*' scopes: #(user organization tier global).
	] on: Error do: [ :ex |
		QLogger logWarning: 'Reverting to scope-by-scope feature setup (', ex description, ')'.
		^ self preloadFeatureEntitlementsByScope
	].

	"Got a merged dictionary - pour the results into the cache as booleans."
	self class allFeatureKeys do: [ : key |
		fullKey := 'EnabledFeatures.', key.	
		flag := nil.
		(dict includesKey: fullKey) ifTrue: [	"Have a specified value."
			flag := ( (dict at: fullKey) = 'true' )
		].
		featureProperties at: key put: (flag ifNil: [false]).
	].! !

!QwaqParticipantUI methodsFor: 'roles-private' stamp: 'bgf 9/6/2008 15:55'!
preloadFeatureEntitlementsByScope

	"Fallback for older service providers, for preloadFeatureEntitlements.
	 If the SP doesn't do multiscope queries, we have to query all the scopes.."

	| userDict orgDict tierDict  sysDict fullKey flag |
	[ 
		"First fetch the forums.properties EnabledFeatures.*  tables from each scope, once;
	       so we're doing at most 3 fetches rather than (3 * num-features)"

		serviceMgr ifNil: [ ^self ].
		userDict := serviceMgr matchProperty: 'EnabledFeatures.*' scope: 'user'.
		serviceMgr ifNil: [ ^self ].
		orgDict := serviceMgr matchProperty: 'EnabledFeatures.*' scope: 'organization'.
		serviceMgr ifNil: [ ^self ].

		"The tier layer is new, and may not be recognized by the SP"
		[ tierDict := serviceMgr matchProperty: 'EnabledFeatures.*' scope: 'tier' ]
		  on: Error do: [
			QLogger logWarning: 'SP does not support tier property scope.'.
			tierDict := Dictionary new.
			serviceMgr currentOrg ifNotNil: [
				self defaultFeaturesForTier: (serviceMgr currentOrg tierName) intoDictionary: tierDict.
			]].

		serviceMgr ifNil: [ ^self ].
		sysDict := serviceMgr matchProperty: 'EnabledFeatures.*' scope: 'global'.
		serviceMgr ifNil: [ ^ self ].

	] on: Error do: [
		^ self 	"So we don't cache false negatives."
	].

	"Now pour the results into the cache, according to the precedence rules."

	self class allFeatureKeys do: [ : key |
		fullKey := 'EnabledFeatures.', key.	
		flag := nil.
		{userDict . orgDict . tierDict . sysDict } do: [ :scope |	"Order of precedence."
			flag ifNil: [									"squeak has no loop-break."
				(scope includesKey: fullKey) ifTrue: [		"this scope has a value"
					flag := ( (scope at: fullKey) = 'true' )		"so it rules"
				]].
		].
		featureProperties at: key put: (flag ifNil: [false]).
		].! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 8/9/2008 15:02'!
printStackReport: preStack on: strm
	"Create the stack report as part of an error report stream"
	| cnt |

	strm nextPutAll: '------- Stack Details -------'; cr.
	cnt := 0.
	(preStack copyFrom: 1 to: (5 min: preStack size)) do:[:ctx|
		ctx printDetails: strm.	"variable values"
		strm cr.
	].
	strm cr; nextPutAll: '------- Full Stack -------'; cr.
	cnt := 0.
	preStack do:[:ctx|
		cnt := cnt + 1.
		cnt = 5 ifTrue: [
			strm nextPutAll: ' - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -'; cr].
		strm print: ctx; cr.  "just class>>selector"
	].
	
! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 6/1/2010 14:42'!
privateResetSound: reasonString
	| speaker mic reset |

	session ifNil: [ ^ self ].
	audioSession ifNil: [^self].

	QLogger logWarning: 'Resetting sound system (reason: ', reasonString, ')'.
	QAudioSession pluginLogMsg: 'RESET-SOUND: Resetting sound system.'.

	"Step 0: turn off basic user controls, to stop activity at origin."

	sanityChecker suspended: true.
	speaker := controlBar speaker selected.
	mic := controlBar microphone selected.
	controlBar microphone selected: false.
	controlBar speaker selected: false.
	self waitTick.	"Let the handlers for button state changes fire."

	"Main reset - with all TIslands locked to prevent runScript deadlocks from audio-phase render."
	reset := (audioSession future resetSoundAfterLockingSession: session) wait.
	session ifNil: [ ^ self ].

	"Restore the basic controls."
	controlBar microphoneButtonFired: mic.
	controlBar speaker selected: speaker.

	sanityChecker suspended: false.
	QLogger logInfo: (reset ifTrue: [ 'Sound reset complete.' ] ifFalse: [ 'Sound-reset canceled, no lock.']).
	QAudioSession pluginLogMsg: (reset ifTrue: [ 'RESET-SOUND: Sound reset complete.' ] ifFalse: [ 'RESET-SOUND: canceled, no lock.']).

	self waitTick.
	session ifNil: [ ^ self ].

	"Update the UI to reflect any changes."
	self signal: #audioDeviceSettingsChanged.

	self playSoundEffect: #blip.
	reset ifTrue: [
		self statusInfo: 'Sound has been reset.'.
	] ifFalse: [
		self statusAlert: 'System busy. Sound was not reset.'.
	].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 6/22/2010 16:46'!
privateRunGroupClientExperienceTest: reasonString
	| test notifying testDetails |

	QLogger logWarning: 'Running group client experience test (reason: ', reasonString, ')'.
	mobileProxy ifNotNil: [ "Now, before we bog the proxy down."
		mobileProxy startScript: #requestClientExperience.  "Reports to text chat asyncronously on its own."	
		[
			(Delay forSeconds: 5) wait.  "for a response"
			self profileToChat: 15.
		] forkAt: Processor userBackgroundPriority. 
		^self   "Sound test is too unreliable on proxy. Why?"
	].

	test := QSoundTest forSeconds: 5.
	test controller: self session mediaController.

	"As long as the test is still going, keep informing the user."
	notifying := true.
	[ [notifying] whileTrue: [
			self signal: #infoMessage with: 'Running group client experience test...'.
			(Delay forSeconds: 1) wait]
	] fork.

	[	[test testDecoding.
		 testDetails := test detailsFor: self] 
			on: TestFailure, Error 
			do: [:ex |
				ex isResumable
					ifTrue: [ex resume "Ignore assertion failures."]
					ifFalse:
						[QLogger logError: ex.
						testDetails := 'Error during client experience test: ', ex messageText]].
	] ensure:
		[notifying := false.
		 test destroy].

	self startScript: [
		self activeSpace future chatLog future
			addMessage: testDetails
			fromUser: (self activeSpace future chatLog future userForAvatar: self avatar replica)
			atTime: QChatLog clientUtcTime].

	self statusInfo: 'Ran group client experience test.'! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'jcg 12/1/2008 18:43'!
privateTestVoiceTransportWithDefaultSound: allAtOnce
	"If 'allAtOnce' is true, we send audio data downstream as fast as we can; given sufficient bandwidth, this should allow us to simulate someone who has a network lag and then receives a torrent of audio data all at once."
	| path  |

	path := FileDirectory pathFromURI: 
						'Objects/Sounds/MicrophoneCheck/Qwaq-MicCheck-16000Hz.wav'.
	self privateTestVoiceTransportWithFileNamed: path allAtOnce: allAtOnce.
! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 2/21/2009 13:50'!
privateTestVoiceTransportWithFileNamed: fullPath allAtOnce: allAtOnce
	"If 'allAtOnce' is true, we send audio data downstream as fast as we can; given sufficient bandwidth, this should allow us to simulate someone who has a network lag and then receives a torrent of audio data all at once."
	| sound |
	[sound := SampledSound fromWaveFileNamed: fullPath]
		on: Error 
		do: [:err | ^self statusAlert: 'Cannot open WAV file to test voice transport'].

	[ self privateTestVoiceTransportWithSound: sound allAtOnce: allAtOnce ] 
		on: Error 
		do: [:err | QLogger logWarning: 'Private voice test failed: ', err description.
			self statusAlert: 'Unexpected error in voice transport test.' ].! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'jcg 3/4/2010 18:04'!
privateTestVoiceTransportWithSound: sound "a SampledSound" allAtOnce: allAtOnce
	"If 'allAtOnce' is true, we send audio data downstream as fast as we can; given sufficient bandwidth, this should allow us to simulate someone who has a network lag and then receives a torrent of audio data all at once."
	| codec samplingRate startTime bufferSize remaining buffer delayTime elapsed targetElapsed |
	
	startTime := Time millisecondClockValue.

	self statusInfo: 'Transmitting prerecorded voice sample.'.

	codec := nil.
	samplingRate := sound originalSamplingRate.
	samplingRate == 8000 ifTrue: [codec := GSMCodec new].
	samplingRate == 16000 ifTrue: [codec := QSpeexCodec new].
	codec ifNil: [
		self statusAlert: 'WAV file must be 8kHz or 16kHz'.
		(Delay forMilliseconds: 1000) wait.
		^self
	].
	
	"Break sound down into buffers the same size as sent by the voice recorder."
	"XXXX: is this also right for GSM/8000Hz?  Don't care right now."
	bufferSize := 1280.
	remaining := sound samples size.
	
	delayTime := (bufferSize / samplingRate * 1000) rounded.
	startTime := Time millisecondClockValue.
	targetElapsed := 0.

	["Fork the actual sound delivery in a high-priority process just as the mic would"

	"Break the sound up into chunks."
	1 to: sound samples size by: bufferSize do: [:start |
		buffer := SoundBuffer new: bufferSize.
		buffer 
			replaceFrom: 1
			to: (bufferSize min: remaining)
			with: sound samples
			startingAt: start.

		remaining := remaining - bufferSize.
		audioSession voiceRelay ifNotNil: [:relay |
			relay disableMicrophone.
			relay
				reallyPlaySound: (codec encodeSoundBufferNoReset: buffer)
				at: (Time millisecondsSince: startTime)
				samplingRate: samplingRate
				codec: codec class
				level: 20.
		].

		elapsed := Time millisecondsSince: startTime.
		targetElapsed := targetElapsed + delayTime.

		allAtOnce ifFalse: [
			elapsed < targetElapsed ifTrue: [
				(Delay forMilliseconds: (targetElapsed - elapsed)) wait.
			].
		].
	].

	self signal: #voiceTestComplete] forkAt: Processor highIOPriority.
	self waitUntil: #voiceTestComplete.

	self statusInfo: 'Finished transmitting prerecorded voice sample.'.
	audioSession voiceRelay sendEndOfAudiblePeriod.
	(Delay forMilliseconds: 200) wait.! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 5/19/2010 18:40'!
privateUpdateVoiceRelay
	"Reconnect the voice relay to the avatar replica's speaker.
	 This is the body implementation, called for both normal updates and for retries."
	| spkr check |

	"This is retried on a timer; if the user has given up on a reconnect cycle
	 or just plain logged out / stopped, then we've nothing to do."
	(session isNil or: [ self exitTripped ] or: [ session errorTripped ]) ifTrue: [
		^ self 
	].

	"We do this on reconnects and teleports... make sure we're hip to current rules."
	self updateCanSpeakInMeeting.

	"Clean up the connections to the previous space, if any."
	audioSession 
		ifNil: [QLogger logWarning: 'no audio-session in #privateUpdateVoiceRelay']
		ifNotNil: [:sess | sess voiceRelay 
										stop; 
										speaker: nil].
	self readyForEvents ifFalse: [ 
		"Can happen if we get here before rendering setup during login. Just retry. "
		(session notNil and: [session errorTripped not]) ifTrue: [
			"Not an error/shutdown state yet, so on the normal scripting queue."
			^ self signal: #retryUpdateVoiceRelay
		]
	].

	check := self avatarReplica.		"For comparison below, after the future-round-trip."
	spkr := check get:#speaker.
	spkr ifNil: [
		"Must be recent reconnect, since we just validated the replica-ref,
	 	and the nil means we didn't event get a far-ref back from the get.
	      DO NOT retry here, as reconnect will force a reset later,
	 	and we don't want a busy loop in the meantime."
		QLogger logError: 'Failed get from replica in voice relay setup.'.
		 ^ self ]. 

	spkr whenResolved: [
		self readyForEvents ifFalse: [ ^ self ].
		spkr value ifNotNil:[ 
			"Cross-check that after resolve, the harness hasn't reconnected or teleported."
			((check = self avatarReplica) and: [check isValidFarRef] and: [spkr isValidFarRef]) ifTrue: [
				audioSession voiceRelay
									speaker: spkr value;
									start.
				QLogger logInfo: 'just updated voice-relay'.
			] ifFalse: [
				self readyForEvents ifTrue: [
					"If we're still healthy, then we've overlapped teleport (or disconnect);
					 Just retry."
					self signal: #retryUpdateVoiceRelay.
				].
			].
		] ifNil: [
			"The replica speaker is set up at replica creation time, if supported.
	 		 If we didn't get one after all the validity checks, it really doesn't have one."
			QLogger logWarning: 'Avatar has no speaker for voice.'.
		].
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 5/29/2010 11:25'!
processCommandLine
	"Check whether we've been started -demonic"
	| index attr arg forumToLoad robot |
	self isSealedApplication ifFalse:[^nil]. "only process command line for sealed apps"
	index := 0.
	robot := false.

	"The browser plugin inserts blank attributes to delimit sets of arguments.
  	 they end with a nil after the last set (the PARAMs from the html block."

	[attr := Smalltalk getSystemAttribute: (index := index + 1).
	attr isNil] whileFalse:[
		attr caseOf:{
			['-demonic:']	-> [
				"Uses of this might be replaceable by some combination of 
				-robot:, -forSeconds:, and -staf:."
				arg := (Smalltalk getSystemAttribute: (index := index + 1)) asNumber.
				self startScript: #runDemonic: withArguments: {arg}.
			].
			['-upgrade:']	-> [
				arg := Smalltalk getSystemAttribute: (index := index + 1).
				WorldState addDeferredUIMessage:[self class upgrade: arg restart: true].
			].
			['-runAllTests']	-> [
				WorldState addDeferredUIMessage:[self class runAllTests: true].
			].
			['-noOGL' ] -> [
				self doOpenGL: false
			].
			['-noSound'] -> [
				"Tells the audio manager to pretend there are no audio devices here."
				QAudioDeviceManager soundIsEnabled: false.
			].
			['-logIDs' ] -> [
				TObjectID logIDs: true
			].
			['-robot:'] -> [
				arg := (Smalltalk getSystemAttribute: (index := index + 1)) asSymbol.
				self startScript: #runRobotic: withArguments: {arg}.
				robot := true.
			].
			['-staf:'] -> [
				StafProxy fromFileNamed: (Smalltalk getSystemAttribute:(index := index + 1)).
			].
			['-forSeconds:'] -> [
				arg := (Smalltalk getSystemAttribute: (index := index + 1)) asNumber.
				self startScript: #runTime: withArguments: {arg}.
			].
			['-qwqFile:' ] -> [
				arg := (Smalltalk getSystemAttribute: (index := index + 1)) asSymbol.
				self processQwqFile: arg.
			].
			['-proxy:']	-> [
				arg := Smalltalk getSystemAttribute: (index := index + 1).
				self startScript: #runAsProxyClient: withArguments: {arg}.
			].


			"The following are not delimited with - :, so that they can
 			be placed as attributes within the EMBED tag for a browser-based launch;
	 		mozilla requires this."
			['server']	-> [
				arg := Smalltalk getSystemAttribute: (index := index + 1).
				login serverList value: arg.
			].
			['organization']	-> [
				arg := Smalltalk getSystemAttribute: (index := index + 1).
				login organization value: arg.
			].
			['forum' ] -> [
				forumToLoad := Smalltalk getSystemAttribute: (index := index + 1).
				login forumGoal: forumToLoad.
			].
			['user'] ->[
				arg := Smalltalk getSystemAttribute: (index := index + 1).
				login user value: arg.
			].
			['password'] ->[
				"For testing use only, since cmd line args are so insecure!!"
				arg := Smalltalk getSystemAttribute: (index := index + 1).
				login password value: arg.
			].
				
		} otherwise:[
			"Check if this is just a file that was associated with forums, like a qwq"
			(attr asLowercase endsWith: '.qwq') 
				ifTrue:[self processQwqFile: attr].
		].
	].

	"If a forum name was provided, or we're starting robotic, log in all the way."
	((forumToLoad notNil) | robot) ifTrue: [
		self autoLogin: login serverList value
			organization: login organization value
			forum: forumToLoad
			location: ''
			robotic: robot.
		self clearReturnQwqFile..  
	] ifFalse: [ "Check for an auto-return.qwq file only if we're not otherwise auto starting above"
		(FileDirectory default fileExists: self returnQwqPath) ifTrue: [
			WorldState addDeferredUIMessage:[  "give us a moment to receive an odoc on Mac"
				self processQwqFile: self returnQwqPath.  
			].
		]
	].
	! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'hrs 4/9/2010 14:48'!
processQwqFile: fileName
	"Process the .qwq file at fileName, and delete it if contains credentials."
	| file contents |
	[
		file := FileStream readOnlyFileNamed: fileName.
		[contents := file contents] ensure: [file close].
	]
		on: Error
		do: [:ex | 
			QLogger logError: ex.
			^nil
			].
	(self processQwqStream: contents) ifTrue: [FileDirectory default deleteFileNamed: fileName].! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 5/13/2010 11:39'!
processQwqStream: aFile
	"We have received some sort of 'event' (loosely speaking) to open a document of type .qwq.
	For example, on Mac, this may have originated as an odoc Apple Event that the VM translated
	into a drop event.  Answers true IFF the file contains credentials (so that senders can clean up)."
	| payload wrapper node qwqOrg qwqForum qwqLocation qwqHost isCredentialed |
 
	wrapper := QMimeWrapper importFrom: aFile.
	payload := wrapper payload.
	node := (XMLDOMParser on: payload readStream) nextEntity.
	qwqHost := (node contentStringAt: 'host') ifEmpty: [self login serverList value].
	qwqOrg := (node contentStringAt: 'organization') ifEmpty: [self currentOrgName].
	qwqForum := node contentStringAt: 'forum'.
	qwqLocation := node contentStringAt: 'location'.
	QLogger logInfo: 'Processing ', qwqHost, '/', qwqOrg, '/', qwqForum, '/', qwqLocation.
	isCredentialed := false.
	(node contentStringAt: 'user') ifNotEmpty: [:user | 
		"safety check against misuse"
		(self serviceMgr notNil and: [user ~= self login user]) ifTrue: [
			"We're already logged in as someone else. That is not a supported use of <user> in .qwq files."
			^self inform: 'Cannot change user. Please log out.' "Alt., we could confirm w/ sufficient expl."
		].
		isCredentialed := true.
		self login userInput value: user.
		(node contentStringAt: 'password') ifNotEmpty: [:pass | 
			self login savePassword value: false.
			self login password value: pass.
		].
		(node contentStringAt: 'proxyPublicationId') ifNotEmpty: [:hex | 
			self startScript: #runAsProxyClient: withArguments: {hex}
		].
	].

	self clearReturnQwqFile.  "Doing what we're doing supercedes any auto-return processing"
	self signal: #generalTeleport with: qwqHost with: qwqOrg with: qwqForum with: qwqLocation.
	^isCredentialed
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'hrs 5/16/2010 13:51'!
profileToChat: seconds
	"Profile performance and report to text chat."
	| user profiler results |
	user := self avatar publicName.

	self statusInfo: 'Profiling for ', seconds, ' seconds.'.

	profiler := self preferredProfilerClass default.
	[profiler spyOn: [(Delay forSeconds: seconds) wait]]  ensure: [ 
		results := String streamContents: [:s |  profiler report: s.
		].
	].

	self startScript: [
		self activeSpace future chatLog future
			addMessage: results
			fromUser: (self activeSpace future chatLog future userForAvatar: self avatar replica)
			atTime: QChatLog clientUtcTime].

	self statusInfo: 'Profile complete for ', user, '.'.! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'eem 11/18/2008 16:29'!
profilerPreference: aSymbol
	ProfilerPreference := aSymbol! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 6/6/2007 23:37'!
projectMorph
	"Answer the projectMorph of the receiver"
	<bewareOf: #projectMorphChanged>
	^self propertyValueAt: #projectMorph! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 6/6/2007 23:37'!
projectMorph: aValue
	"Modify the receiver's projectMorph"
	^self propertyValueAt: #projectMorph put: aValue with: #projectMorphChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 5/24/2007 13:45'!
proxy
	"Answer the proxy of the receiver"
	<bewareOf: #proxyChanged>
	^self propertyValueAt: #proxy! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 5/24/2007 13:45'!
proxy: aValue
	"Modify the receiver's proxy"
	^self propertyValueAt: #proxy put: aValue with: #proxyChanged! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/13/2010 13:58'!
proxyAuth: webClient realm: authRealm fail: failBlock
	"Ask the user for authentication credentials.

	authRealm is expected to be the user visible description for the authentication request.
	For Sharepoint (and other repositories) this should be the repository URL, NOT the realm
	used by WebClient (which is not available for NTLM and friends)."


	| dialog pw |
	dialog := QUserLoginDialog new.
	dialog title value: 'Proxy Authentication Required '.
	dialog explanation value:
		'Enter username and password to access\' withCRs, 
		'Enter username and password to access', String cr,  authRealm.
	dialog userName value: (webClient username ifNil:[login user value]).
	dialog password value: (webClient  password value ifNil:['']).
	dialog savePassword visible: false.
	dialog acceptAction:[pw := (QPasswordHolder with: dialog password value).  dialog close].
	dialog cancelAction:[pw := nil. dialog close].
	dialog openCenteredAt: self dialogCenter.
	dialog waitUntil: #closed.
	pw ifNil:[^failBlock value].
	webClient username: dialog userName value.
	webClient password: pw.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 7/12/2011 12:51'!
proxyDlg
	"Answer the proxyDlg of the receiver"
	<bewareOf: #proxyDlgChanged>
	^self propertyValueAt: #proxyDlg! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 7/12/2011 12:51'!
proxyDlg: aValue
	"Modify the receiver's proxyDlg"
	^self propertyValueAt: #proxyDlg put: aValue with: #proxyDlgChanged! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'bgf 10/22/2008 21:31'!
qmsExtensionDirectory
	"The directory for objects common to all applications."
	^FileDirectory default directoryNamed: (FileDirectory default pathFromURI: 'objects-menu/QMS').! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 4/15/2007 21:18'!
quickReference
	"Answer the quickReference of the receiver"
	<bewareOf: #quickReferenceChanged>
	^self propertyValueAt: #quickReference! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 4/15/2007 21:18'!
quickReference: aValue
	"Modify the receiver's quickReference"
	^self propertyValueAt: #quickReference put: aValue with: #quickReferenceChanged! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 10/1/2007 17:27'!
quickServiceUrlForUrl: url

	"See realServiceUrlForUrl: for the proper way to this.
	 This is a simplified method that assumes names are always
	 host-stuff/realm-name/service-name/island-etc.
	 I.e. we just decompose the string without consulting the realm server."

	| host path realm svcName svcUrl nextSlash |
	(url beginsWith: '//') ifFalse:[^self error: 'Service url must be absolute'].
	"Decompose url to find host/port"
	host := url copyFrom: 3 to: (url indexOf: $/ startingAt: 3)-1.
	path := url copyFrom: host size + 3 to: url size.
	"Carefully treat both, urls with and without ports here"
	host := (host upTo: $:) asLowercase. "normalize"

	"This is where the assumptions start: 
	 just take the next word as realm,
	 and the one after that as service name..."

	realm := path copyFrom: 1 to: (path indexOf: $/ startingAt:2)-1.
	nextSlash := (path indexOf: $/ startingAt: realm size+2).
	nextSlash = 0 ifTrue: [
		svcName := path copyFrom: (realm size+2) to: (path size).
	] ifFalse: [
		svcName := path copyFrom: (realm size+2) to: nextSlash-1.
	].
	svcUrl := '//',host, realm,'/', svcName.
	^svcUrl! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 6/8/2008 19:14'!
quindiLog: noteText hiddenText: hiddenText
	"Log a message for Quindi Meeting Companion. The noteText appears is visible 
	by default, the hiddenText only in the web-ui (but searchable by default).
	This method currently always takes a screen shot of the user's view."
	| fileName screen clipText |
	fileName := (FileDirectory on: (QFileDialog getFileLocation: 'pictures'))
		fullNameFor: (Date today printFormat: #(3 2 1 $- 2 1 2)), '-',
						(Time now print24 copyReplaceAll: ':' with: '-'),'.bmp'.
	(session notNil and:[session ogl notNil])
		ifTrue:[screen := session ogl screenShot].
	screen ifNil:[screen := Display].
	BMPReadWriter putForm: screen onFileNamed: fileName.
	"Strip the tabs off the input text since it is used as separators for Quindi"
	clipText := 'QwaqData:', (noteText copyReplaceAll: String tab with: '    '), 
		String tab, (hiddenText copyReplaceAll: String tab with: '    '), 
		String tab, fileName.
	Clipboard clipboardText: clipText.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 9/26/2007 11:53'!
quindiLogging
	"Answer the quindiLogging of the receiver"
	<bewareOf: #quindiLoggingChanged>
	^self propertyValueAt: #quindiLogging! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 9/26/2007 11:53'!
quindiLogging: aValue
	"Modify the receiver's quindiLogging"
	^self propertyValueAt: #quindiLogging put: aValue with: #quindiLoggingChanged! !

!QwaqParticipantUI methodsFor: 'testing' stamp: 'bgf 8/24/2009 02:41'!
readyForEvents
	"Use this preflight check before operating on the avatar replica far ref 
         or any far-ref, to protect against disconnects 
	 and to suppress action in the face of errors."

	^ (exitTripped not) 
		and: [mainView notNil and: [ mainView readyForEvents ] ]
		and: [ session notNil and: [ session readyForEvents] ]
	    	and: [ activeSpace isValidFarRef ] 
! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 8/26/2009 02:16'!
realServiceUrlForUrl: url

	"The CORRECT way to do serviceUrlForUrl:...
	 For doing this we first need to know what realm / service the url belongs to 
	 and we find out by asking the service manager at the target host."

	| mgr host path response realm svcName svcUrl info |
	(url beginsWith: '//') ifFalse:[^self error: 'Service url must be absolute'].
	"Decompose url to find host/port"
	host := url copyFrom: 3 to: (url indexOf: $/ startingAt: 3)-1.
	path := url copyFrom: host size + 3 to: url size.
	"Carefully treat both, urls with and without ports here"
	host := (host upTo: $:) asLowercase. "normalize"

	"Retrieve connection info"
	[info := self connectionInfoFor: host] 
		on: Error do:[:ex| QLogger logWarning: ('Problem in url-check: ', ex).
					^nil].

	mgr := QServiceManagerV2 for: self.
	mgr info: info.
	response := mgr resolveRealm: path.
	response name == #ok ifFalse:[
		QLogger logWarning: 'Service ', url, ' not available: ', response contentString.
		^ nil
	].
	mgr destroy.

	realm := response contentStringAt: 'realm'.
	"Now that we got the realm, we can construct both service name and service url"
	svcName := (path copyFrom: realm size+2 to: path size) copyUpTo: $/. 
	svcUrl := '//',host, realm,'/', svcName.
	^svcUrl! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 8/31/2010 10:39'!
reconnect
	"Reconnect the client"

	| dlg giveUp success |
	<on: #reconnect>

	"In case we ever accumulate #reconnect signals"
	(Delay forSeconds: 1) wait.
	DropAllButLastEvent signal.

	(self exitTripped) ifTrue: [ ^ self ].

	session ifNil: [ 
		"If not in full session, check for org services and repair that."
		orgServices ifNil: [ ^ self ].
		orgServices errorTripped ifTrue:[^self].
	] ifNotNil:[
		"Remember that we were reconnecting"
		session clearEvent.
		session errorTripped ifTrue: [ ^ self ].
		session reconnecting: true.
	].

	giveUp := false.
	sanityChecker suspended: true.
	
	QLogger log: 'Start reconnect'.
	QActivityCapture action: #Reconnect details: ''.

	"Inform the user about what we're going to do"
	dlg := self showReconnectDialog: [
		giveUp := true.
		session ifNotNil:[self logoffConfirming: false].
	].

	self flowControlFlag: true.

	[success := self reconnectEverything:[giveUp].
	success or:[giveUp or:[exitTripped or:[self errorTripped]]]] 
		whileFalse:[self wait: 2].

	"Clean up some outdated state"
	session ifNotNil:[
		"Remove the old (invalid) controllers"
		(session islandControllers reject:[:cc| cc isValid]) do:[:cc| 
			cc island ifNotNil:[
				self audioSession ifNotNil:[:as | as flushForIslandReconnect: cc island].
			].
			session removeController: cc.
			cc destroy.
		].
	].
	dlg ifNotNil:[
		dlg destroy. 
		dlg := nil.
	].
	self mouseCursor: nil.
	self statusAlert: nil.

	self exitTripped ifTrue: [ ^ self ].
	self errorTripped ifTrue: [ ^ self ].

	giveUp ifFalse: [
		self finishMediaReconnection.
		self errorTripped ifTrue: [ ^ self ].
		self finishOrgReconnection.
		self errorTripped ifTrue: [ ^ self ].
		sanityChecker suspended: false.
	].
	self flowControlFlag: false.

	"Print stats to transcript."
	self errorTripped ifTrue:[^self].

	QLogger log: 'Reconnect completed' from: self.
	self statusAlert: 'Connection restored'.
	"Give us some time before we consider reconnect finished"
	(Delay forSeconds: 2) wait.
	session ifNotNil: [ session reconnecting: false ].

	self readyForEvents ifTrue: [
		avatar setActivities.		"Flush out transient inconsistencies."
		self updateRecordingUsers.
		recorderPane checkForCurrentRecording.
	].! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 8/31/2010 10:44'!
reconnectEverything: giveUpBlock
	"The basic reconnect logic. If anything fails in any phase of the reconnect
	we return false to the caller and let it retry. Otherwise we simply return true
	and let the caller proceed."

	| link |
	"First, reconnect the service manager"
	(giveUpBlock value or:[exitTripped or:[self errorTripped]]) ifTrue:[^false].
	(serviceMgr reconnect) ifFalse:[^false].
	QLogger logInfo: 'Service manager reconnect successful'.
		
	"Next, reconnect org services"
	(giveUpBlock value or:[exitTripped or:[self errorTripped]]) ifTrue:[^false].
	(orgServices reconnect) ifFalse:[^false].
	QLogger logInfo: 'Org services reconnect successful'.

	"Next, reconnect all the active services"
	self activeServices do:[:svc|
		(giveUpBlock value or:[exitTripped or:[self errorTripped]]) ifTrue:[^false].
		link := QServiceLink forClient: self.
		link serviceManager: serviceMgr.
		link session: session.
		link service: svc.
		(link reconnect) ifFalse:[^false].
	].

	^true! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 8/31/2010 10:32'!
reconnectTest
	"Testing only: Reconnect all existing (and working) remote connections to see what happens."
	self signal: #reconnect.
! !

!QwaqParticipantUI methodsFor: 'actions' stamp: 'bgf 3/25/2010 12:51'!
recordAction: urAction
	"Add this user action to my undo-redo list,
	 and note that this represents a formal change to the active space."
	undoRedoList recordAction: urAction.
	session ifNotNil: [
		session activeIsland future
			signal: #action:url:author: 
			withArguments: { (urAction description) . (serviceSvc url) . self sessionID }.
	].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 3/17/2009 11:20'!
recordRawMicrophoneData
	"Tell the microphone to start recording it's data to disk for debugging.  It will do so for a fixed time period.  See QMicrophone."
	| mic |
	[mic := self audioSession microphone]
		on: Error 
		do: [:err |
			^QLogger logWarning: 
				'could not obtain microphone because: ', err description.
		].
	mic startRawRecording! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/11/2009 10:52'!
recorderPane
	"Answer the recorderPane of the receiver"
	<bewareOf: #recorderPaneChanged>
	^self propertyValueAt: #recorderPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/11/2009 10:52'!
recorderPane: aValue
	"Modify the receiver's recorderPane"
	^self propertyValueAt: #recorderPane put: aValue with: #recorderPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/23/2009 23:14'!
recordingIndicator
	"Answer the recordingIndicator of the receiver"
	<bewareOf: #recordingIndicatorChanged>
	^self propertyValueAt: #recordingIndicator! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/23/2009 23:14'!
recordingIndicator: aValue
	"Modify the receiver's recordingIndicator"
	^self propertyValueAt: #recordingIndicator put: aValue with: #recordingIndicatorChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/24/2009 13:11'!
recordingUsers
	"Answer the recordingUsers of the receiver"
	<bewareOf: #recordingUsersChanged>
	^self propertyValueAt: #recordingUsers! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/24/2009 13:11'!
recordingUsers: aValue
	"Modify the receiver's recordingUsers"
	^self propertyValueAt: #recordingUsers put: aValue with: #recordingUsersChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 4/12/2010 12:07'!
refreshRoundTripSound
	"If round-trip sound is on, re-create it."
	<on: refreshRoundTripSound>
	roundTripSound ifNotNil: [self makeRoundTripSound].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 10/9/2006 16:25'!
registerPortal: portalName in: island
	| entry pc |
	entry := island future at: portalName.
	pc := entry future postcard.
	pc whenResolved:[session addIsland: island postcard: pc value.].
	^entry! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/8/2008 19:14'!
render360

	self statusInfo: 'Rendering for a measured turn; please wait...'; waitTick.
	session ifNotNil: [session startScript: #render360: withArguments:#(10)]! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 3/7/2009 19:15'!
renderOverlay: ogl
	"Set up gl to provide overlay drawing - this is done in pixel coordinates"
	ogl glDisable: GLLighting.
	self beginOverlay: ogl.
	self glRenderOverlayOn: ogl.
	self endOverlay: ogl.
	ogl glEnable: GLLighting.
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 6/8/2008 19:15'!
renderProfile: seconds

	"Developer menu operation - do a 360 render test for n seconds,  and profile it."
	self startScript: #systemProfiler: withArguments: {seconds+1}.
	session startScript: #render360: withArguments: {seconds}.
! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 1/3/2009 10:17'!
renderTest

	| dlog |
	session ifNotNil: [ 
		inBenchTest ifFalse:[
			false ifTrue:[mainView startScript: #renderTest. ]
			ifFalse:[
				dlog := (QOKCancelDialog new).
				dlog title value: 'This test will take some',
				Character cr,     'time to complete. Click',
				Character cr, 	'mouse to terminate test',
				Character cr, 	'at any time. Proceed? '.
				dlog acceptBlock:[ 
					session startScript: #renderTest. 
				].
				dlog openCenteredAt: (self dialogCenter).
			]]].
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/30/2010 17:03'!
repoBrowser
	"Answer the repoBrowser of the receiver"
	<bewareOf: #repoBrowserChanged>
	^self propertyValueAt: #repoBrowser! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 3/30/2010 17:03'!
repoBrowser: aValue
	"Modify the receiver's repoBrowser"
	^self propertyValueAt: #repoBrowser put: aValue with: #repoBrowserChanged! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/16/2013 11:03'!
repopulateToolsMenu: menu
	| item |
	menu removeMenuItems.
	"--- Principal tools first ---"

	self hasExperimentalFeatures ifTrue: [
		item := menu add: 'Crowd...' translateMe target: crowdPane action: #toggle.
		self linkItem: item toTool: crowdPane
	].

	item := menu add: 'Documents...' translateMe target: documentsPane action: #toggle.
	self linkItem: item toTool: documentsPane.

	self hasPhoneSupport ifTrue: [
		item := menu add: 'Phone...' translateMe target: phonePane action: #toggle.
		self linkItem: item toTool: phonePane.
	].

	self hasSharePoint ifTrue:[
		item := menu add: 'SharePoint ...' translateMe target: repoBrowser action: #toggle.
		self linkItem: item toTool: repoBrowser.
	].

	self hasChat ifTrue:[
		item := menu add: 'Text Chat...' translateMe target: chatPane action: #toggle.
		self linkItem: item toTool: chatPane.
	].

	"Next group - construction-biased tools..."

	menu addSeparator.
	item := menu add: 'Object Catalog...' translateMe target: self action: #toggleCatalog.
	self linkItem: item toTool: catalogPane.

	item := menu add: 'Scene Inspector...' translateMe target: treeInspector action: #toggle.
	self linkItem: item toTool: treeInspector.
	item enabled: (self canModifyForum).

	item := menu add: 'Theme Materials...' translateMe target: materialCatalog action: #toggle.
	self linkItem: item toTool: materialCatalog.
	item enabled: (self canModifyForum).
	
	"-- Administration - user permissions ---"
	menu addSeparator.

	(menu add: 'Manage Users...' translateMe target: usersPane action: #editForumPermissions)
		tag: #addRemoveUsers;
		enabled: (self isAdministrator).

	"--- Tail group for personal settings ---"

	menu addSeparator.
	item := menu add: 'My Avatar...' translateMe target: avatarPane action: #toggle.
	self linkItem: item toTool: avatarPane.

	menu add: 'Preferences...' translateMe target: self action: #showPreferences.
	"Preferences is a dialog rather than a float, so it is not linked."

	menu add: 'Sound...' translateMe target: controlBar action: #adjustSound.

	QStyles installMenuFontInMenu: menu.
	
	self hasExperimentalFeatures ifTrue: [
		menu addSeparator.
		item := (
			menu
				add: 'Disable Video/PShare Decoding/Rendering' translateMe
				target: self
				action: #toggleReceiveButIgnoreVideo).
		item tag: #toggleReceiveButIgnoreVideo.
		item checked: (QJoshHacks receiveButIgnoreVideo).
	].

	^menu

! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'ar 2/2/2007 17:37'!
reportError: msgString details: details label: dialogLabel button: buttonLabel
	"Report an error message with extra details which can be shown upon request. Use the dialog and button labels specified."
	| dlg button editor |
	dlg := CUserDialog new.
	dlg label: dialogLabel.
	dlg message: msgString asText.
	dlg messageText allowEditing: false.
	dlg messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	dlg costume. "force #setupCostume"

	editor := CTextEditor new.
	editor textFont: (QStyles textEditFont).
	editor contents: details.
	editor vResizing: #rigid. editor height: 100.
	editor hScrollable: false.
	editor selectAt: editor contents size.
	dlg group add: editor after: dlg messageText.
	editor visible: false.

	dlg addButton: buttonLabel value: nil.
	button := dlg addButton: 'Details...'.
	self startScript:[editor visible: true] when:{button. #fire}.
	dlg run: nil centerAt: self bounds center.
! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 5/19/2010 11:14'!
reportIncident: aMessage

	"Asynchronously report the given message as part of a <silent> bug report mail,
	 adding all our conventional bug-report details, transcript, etc."

	QLogger logWarning: 'Incident: ', aMessage.		"Get something onto the transcript asap."
	self signal: #incidentReport with: aMessage with: nil with: nil! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 8/7/2008 14:07'!
reportIncident: aString error: anError stack: preStack

	"Asynchronously report the error incident to Qwaq, without any notice to the user,
	 adding all our conventional reporting details."
	self signal: #incidentReport with: aString with: anError with: preStack.

! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/16/2011 17:34'!
repositoryDrop: incoming at: aPoint fromService: svcRecord

	"Dispatcher for drag-drop of QSharedDocuments from the repository tool.
	Note: incomingis *always* an instance of QSharedDocument since the repo
	browser doesn't know which documents are acceptable and which ones aren't.
	Consequently the code here deals with the 'proper' import for anything that
	shouldn't be a QSharedDocument upon importing."

	| docType docClass document |
	docType := QDocumentType forSuffix: incoming suffix.
	docType ifNil:[
		self startScript: #inform: withArguments: {'Cannot import ', incoming shortName, String cr, String cr, 'The document type ".', incoming suffix asLowercase, '" is not supported.'}.
		^false
	].

	"Thanks Bill!!  We really appreciate your fine work."
	docType isMicrosoftXML ifTrue: [
		(self confirm: '', incoming shortName, (String cr), (String cr),
				'is in Microsoft Office Open XML Format.', (String cr), (String cr),
				'You can open this document with OpenQwaq MultiShare,', (String cr),
				'but any changes you make will not be saved.') ifFalse:[^false]].

	"Convert the incoming document to the proper document class"
	docClass := docType docClass.
	document := docClass fromRepository: incoming repoInfo 
		docType: docType session: mainView session.
	document ifNotNil:[
		document 
			suffix: incoming suffix;
			ctime: incoming ctime.
	].

	"Double-check that the document supports repository operations."
	(document == nil or:[document fullySupportsRepositories not]) ifTrue:[
		| fileToImport file |
		"Fallback for non-supported document types"
		(self confirm: 'Cannot check out ', incoming shortName,'
Import the document instead?') ifFalse:[^false].
		fileToImport := QOpenDocExternalJob localPrivateCopy: incoming forClient: self.
		QActivityCapture 
			action: #ImportRepositoryDrop 
			details: (QActivityCapture fileDetails: fileToImport).
		file := FileDirectory default readOnlyFileNamed: fileToImport.
		file ifNotNil:[session dropFile: file event: nil editBox: true].
		^true
	].

"	((self floatingPanes asArray), {catalogPane})  do: [ : pane | 
		(pane visible and: [ pane respondsTo: #repositoryDrop:at:fromService: ]) ifTrue: [
			(self isPoint: aPoint inPlayer: pane ignoring: #() ) ifTrue: [
				^ pane repositoryDrop: incoming at: aPoint fromService: svcRecord
			]]].
"

"	serviceChooser visible ifTrue: [
		^ serviceChooser docToolDrop: document at: aPoint fromService: svcRecord
	].
"
	
	"OK, its for the logged-in case, and probably for dropping into the forum."
	self readyForEvents ifFalse: [
		self statusAlert: 'Not ready to receive document transfer.'.
		^ false
	].

	(meetingState isEditBlocked and: [ self isMeetingLeader not]) ifTrue: [ 
		self statusAlert: 'Imports are not permitted by the meeting leader.'.
		^ false ].

	self canModifyForum ifFalse: [
		self statusAlert: 'As a visitor, you cannot add documents.'.
		^ false
	].

	(self isPointInMainView: aPoint)  ifFalse: [
		self statusAlert:  'Cannot be dropped here.\Drop only in the space.'.
		^ false
	].

	"Do the main work as followup script so the return is immeidate; it starts the drop-in icon animation."
	documentsPane startScript: #dropDocument:at: withArguments: {document . aPoint}.

	^ true! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 8/18/2010 15:04'!
requestDisplay: cardName
	self stopScript: #requestDisplay:.
	self readyForEvents ifFalse: [^self statusAlert: 'Unable to display ', cardName].
	avatar replica future requestDisplay: cardName.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 11/12/2009 12:27'!
resetActivities
	"On login/reconnect, space entry, or back-from-away."
	| activities |
	activities := OrderedCollection new.
	self isEditingTemplate ifTrue: [ activities add: #EditingTemplate ].
	avatar resetActivities: activities.

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'hrs 11/27/2007 12:03'!
resetOpenGL
	self resetOpenGL: stereoOn.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 6/8/2008 19:15'!
resetOpenGL: bStereo
	"Initialize OpenGL if necessar, and restart rendering if appropriate."
	self initializeOpenGL: bStereo.
	(session notNil and: [session ogl notNil]) ifTrue: [
		session killRender.
		session startRender.
		]! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 2/17/2010 18:56'!
resetSound
	"Backstop in case somebody still calls this."
	self resetSound: 'unknown reason'.
! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 2/17/2010 18:36'!
resetSound: reasonString
	"This is the public API (also available though signal to participant
	 for a thorough reset of the sound system."
	| resetLock |
	<on: resetSound>
	session ifNil: [ ^ self ].
	resetLock := self propertyValueAt: #privateSoundLock ifAbsentPut: [ Mutex new ].
	resetLock critical: [
		self privateResetSound: reasonString
	] ifLocked: [ 
		QLogger logWarning: 'Sound reset requested but already in progress.'.
	].! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 3/16/2009 22:34'!
resetTweak
	| lock |
	<on: resetTweak>
	"Clear OpenGL-caches from the tweak portion of the UI,
	 so we can return to normal drawing."

	lock := session ifNotNil: [ session renderLock] ifNil: [ Mutex new ].
	lock critical: [	
		self asPrimCostume do:[:aCostume|
			aCostume localCache: nil ].
		"Also update the hands if any of them are software rendering"
		Processor activeProject ifNotNilDo:[:proj|
			proj hands do:[:h|
				h asPrimCostume localCache: nil ]].
	] ifLocked: [
		"Probably an error condition in render is holding the lock during update."
	].! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 11/23/2007 19:19'!
resolvePostcard: pc within: mseconds info: aString

	"Utility - ask the session to resolve the given postcard to a frame;
	 wait and keep trying for the given number of milliseconds,
	 displaying the given string as an info-status explanation."

	| frame waited delay |

	waited := 0.
	delay := ((mseconds//20) max: 100).
	frame := session findViewpointByPostcard: pc.

	[frame = nil and: [waited < mseconds]] whileTrue: [
		(Delay forMilliseconds: delay) wait.	  "Allow harness time to resolve the viewpoint."
		frame := session findViewpointByPostcard: pc.
		frame ifNil: [self statusInfo: aString; waitTick].
		waited := waited + delay.
	].
	self statusInfo: nil.	"Clear the info"
	^ frame! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/9/2010 17:45'!
restartAudioInterface

	"Specialized support for logoff, 
	  to release openAL resources after a forum session
	  and prepare the lobby for next time.
	 Meant to run concurrently with logoff/return-to-chooser, in a startscript."

	(Delay forMilliseconds: 200) wait.	"Let the Croquet Session (speakers, etc) clean up."
	OpenAL allSubInstancesDo:[:oal| oal destroy].
	ProcessorScheduler relinquishProcessorForMicroseconds: 5000.	"Let open al clean itself up."
	audioSession ifNotNilDo: [:as | as future resetAudioInterface ].! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/14/2010 10:59'!
restoreDefaultLayout
	
	"Return all the floating panes to their initial posture."

	ppToggle selected: true.
	self floatingPanes do: [ : p |
		p restoreToFactory.
	].
	chatPane collapse.
	peoplePlacesSash memo: (self defaultPPSashFraction).
	viewAreaSash  memo: (self defaultPPWidth).
	videoAreaSash memo: (self defaultVidPaneHeight).

	self layoutEverything.
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 2/8/2009 11:59'!
restoreShape
	"If there's a preferred size recorded for the main window,  apply it;
	 otherwise, just take on my default size."
	self isSealedApplication ifFalse: [
		"Inside the dev image, stay small."
		 self restoreShape: (self topLeft extent: (self class defaultSmallWindowSize)).
		 ^true
		 ].
	preferences ifNotNil: [
		  preferences restoreWindow: 'window' 
			    shape: self 
			    default: ((0@0) extent: (self class defaultFullWindowSize)).
	] ifNil: [
		self restoreShape: ((0@0) extent: (self class defaultFullWindowSize)).
	].

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 2/13/2009 20:31'!
restoreShape: aRectangle
	"Restore the preferred full window size... we're going into a live forum."
	aRectangle width > 0 ifTrue: [
		self setWindowBounds: aRectangle.
	].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/8/2010 08:38'!
resumeAudio
	"Restore audibility if suspended.  Cheap, and harmless if already resumed."

	self isAway ifFalse: [
		"We test away first because the normal avatar-lag observer
	 	resumes audio often, to recover from any previous lag-induced suspend...	 
 		 but we don't want audio to come back on for that,  if we're officially Away."
		audioSession future resumeAudio.
	].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/8/2009 14:56'!
resumeVideo
	"Restore video if suspended.  Cheap, and harmless if already resumed."

	self isAway ifFalse: [
		"We test away first because the normal avatar-lag observer
	 	resumes video often, to recover from any previous lag-induced suspend...	 
 		 but we don't want video to come back on for that,  if we're officially Away."
		self videoSession ifNotNilDo: [:vs | vs resumeVideo].
	].! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 8/30/2010 10:32'!
resynced: newController from: oldController
	"We have resync'ed newController. Update the necessary state in the client"

	| url pc oldIsland |
	oldIsland := oldController island.
	self audioSession ifNotNilDo: [ :as | as flushForIslandReconnect: oldIsland ].

	session ifNotNil: [ 
		url := session findIslandUrl: oldIsland.
		mainView activeSpace island == oldIsland ifTrue:[
			"This is the island where rendering starts; refresh the root space"
			pc := TPostcard urlString: 
				(TPostcard makeUrl: serviceSvc entryUrl relativeTo: serviceSvc url).
		].
	].

	url ifNotNil: [
		"If the island has a url, treat it as a service(forum) island... lots of state to reset."
		session flushLinksMatching: url.
		pc ifNotNil:[mainView entryPostcard: pc].
		self urlCacheAt: url put: newController island.
	].

	"Discard stale state that might still have references or triggers in the old island."
	self discardIsland: oldIsland.

	"Retire the old island.  HAS to be syncSend, there is no message queue."
	[ 
		oldIsland syncSend: #discard withArguments:#().
	] on: Error do: [ :ex |
		QLogger logWarning:  'After reconnect: Exception retiring island: ', ex 
	].

	^true
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 11/19/2008 10:46'!
retractWindow
	"Make the Smalltalk window, or the system window around me if unsealed,
	 go away (or hide it, or collapse it, or something)... but do not stop Squeak.
	 This is done once exit-confirm has done the last user interaction."
	self isSealedApplication ifTrue: [
		"This does nothing but blank the content of the Squeak window...
	  	 I'm not sure how to suppress the Squeak Window without stopping Squeak,
	  	 (the point is to take the window away while the exit-time housework goes on.)"
		"ar 11/19/2008: Also lock the root morph or else clicks can pop up the 
		Morphic desktop menu."
		self window world lock.
		self window hide.
	] ifFalse: [
		self window window visible: false
	].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'bgf 3/16/2008 16:11'!
retryUpdateVoiceRelay
	"Reconnect the voice relay to the avatar replica's speaker;
 	 this is a followup retry if the first attempt fails."

	DropAllButLastEvent signal.
	(Delay forMilliseconds: 100) wait.
	self privateUpdateVoiceRelay.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 3/4/2009 10:49'!
returnQwqPath
	"Answer the pathname of a file in which we can leave a .qwq file before exiting, and
	the client will return there (and delete the file) the next time it is started."
	^ FileDirectory prefsDirectory pathFromURI: 'auto-return.qwq'.! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 3/20/2008 00:53'!
robot: aSessionID mode: aSymbol

	"Signal from org services;
	 If this is my user login, become that sort of demon."
	aSessionID = (self sessionID) ifTrue: [
		"Normal case... be that sort of demon."
		aSymbol ifNotNil: [
			self beADemon: aSymbol.
		].
	].! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 11/29/2007 21:09'!
robotMode
	"If I am robot-enabled, return my current robot mode symbol; else nil."
	^ myDemon ifNotNil: [ myDemon mode ]! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/20/2008 01:26'!
roleInCurrentForum
	"Get authoritative information from the service provider.
	 Note that for now, this does not vary by forum."

	^ serviceMgr ifNil: [ 'user' ] ifNotNil: [ serviceMgr role ]! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 10/17/2007 09:11'!
roundTripSound
	"Answer the roundTripSound of the receiver"
	<bewareOf: #roundTripSoundChanged>
	^self propertyValueAt: #roundTripSound! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 10/17/2007 09:11'!
roundTripSound: aValue
	"Modify the receiver's roundTripSound"
	^self propertyValueAt: #roundTripSound put: aValue with: #roundTripSoundChanged! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 2/5/2010 19:10'!
runAsProxyClient: pubID
	"Set up to run as proxy for a foreign client. Very much like runRobotic:."

	self startScript: #makeMobileProxy:  withArguments: {pubID}.

	"If the robot cannot do the desired initial login, just die, since we cannot control it until it is present."
	self startScript: #exitConfirming: withArguments: {false} when: { self . #serviceLoginFailed }.

	self pipeTranscriptToFile.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 11/29/2007 18:31'!
runDemonic: seconds
	"Run demonically - this supports the (Spring 2007) QTest arrangements;
	 for instance, when the demon logs itself in this way, it uses the last
 	 digit of its IP address as the digit and logs in to test user 'test<digit>'"

	logFile := [FileStream forceNewFileNamed: 'Transcript.log'] on: Error do:[:ex| ex return: nil].
	logFile ifNotNil:[logFile nextPutAll: logView panel contents asString].
	Transcript cr; show: 'Starting demon...'.
	Smalltalk windowLabel: self appName, ' - Possessed By Demons'.
	myDemon := QDemon new participant: self.
	myDemon doLogin.
	self waitUntil: #ready.
	self startScript: #demonAction: when: {myDemon . #action}.
	myDemon start.
	self wait: seconds.
	self exitConfirming:false.! !

!QwaqParticipantUI methodsFor: 'testing' stamp: 'crl 4/14/2010 16:31'!
runGroupClientExperienceTest: reasonString
	"This is the public API."

	| experienceTestLock |

	session ifNil: [ ^ self ].
	experienceTestLock := self propertyValueAt: #privateExperienceTestLock ifAbsentPut: [ Mutex new ].
	experienceTestLock critical: [
		self privateRunGroupClientExperienceTest: reasonString
	] ifLocked: [ 
		QLogger logWarning: 'Group client experience test requested but already in progress.'.
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'hrs 6/11/2009 13:30'!
runRobotic: aMode 

	"Followup to configure robotic operation (for command-line -robot option).
	 Log the transcript to a file.
	 Later, once login gets us into a forum,
	 then become a demon in the desired mode."

	self startScript: #startAsDemon: withArguments: {aMode} when: { self . #ready }.

	"If the robot cannot do the desired initial login, just die, since we cannot control it until it is present."
	self startScript: #exitConfirming: withArguments: {false} when: { self . #serviceLoginFailed }.

	self pipeTranscriptToFile.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'hrs 5/23/2008 12:55'!
runTime: seconds
	"Run only for the specified amount of time after SignIn, and exit. Useful for automation."
	self waitUntil: #ready.
	self wait: seconds.
	QTForums reportSuccess.
	self exitConfirming:false.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 4/9/2007 14:39'!
sanityChecker
	"Answer the sanityChecker of the receiver"
	<bewareOf: #sanityCheckerChanged>
	^self propertyValueAt: #sanityChecker! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 4/9/2007 14:39'!
sanityChecker: aValue
	"Modify the receiver's sanityChecker"
	^self propertyValueAt: #sanityChecker put: aValue with: #sanityCheckerChanged! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'ar 4/16/2011 17:34'!
saveAllForums
	"Save all active services."

	| services barBlock |
	self statusInfo: 'Saving all open Forums...'.
	self flowControlFlag: true.
	QActivityCapture action: #SaveAllForumsFileMenu.

	[
		services := servicesPane allServices 
					select: [ : svc | (self urlCacheAt: svc url ifAbsent: [ nil ]) notNil ].
		CProgressBar 
			displayProgress: 'Saving' at: self bounds center 
			from: 0.0 to: 1.0 during:[:bar|
				services doWithIndex: [ : svc : i |
					"Confirm still active to protect from ticking logouts, or quits, or whatever."
					(self urlCacheAt: svc url ifAbsent:[nil]) ifNotNil: [

						barBlock := [ : v | bar value: (i-1 + v) / (services size)] fixTemps.
						self saveForum: svc svcRecord notifying: barBlock.
					].
					self statusInfo: 'Saving all open Forums...'.	 "The cycles can take a while"
				]].
	] ensure: [
		self flowControlFlag: false.
	].
	
	self statusInfo: 'The open Forums have been saved.'.
	self saveLayout.	"Why not?"
! !

!QwaqParticipantUI methodsFor: 'connecting' stamp: 'ar 8/16/2007 12:17'!
saveAuthInfo: realm value: authInfo
	"Store cached authentication info for the given realm"
	| file data xmlDoc xmlNode changed |

	file := [FileDirectory prefsDirectory readOnlyFileNamed: 'authInfo.xml'] 
			on: Error do:[:ex| ex return: nil].
	file ifNil:[
		data := '<authCache/>'.
	] ifNotNil:[
		[file converter: Latin1TextConverter new.
		data := file contents] ensure:[file close].
	].
	[
		xmlDoc := (XMLDOMParser on: data readStream) nextEntity.
		(xmlDoc name == #authCache) ifFalse:[^nil].
	] on: Error do:[:nothing| ^nil].

	changed := false.

	"Take out all previous authInfo's that match the previous realm"
	(xmlDoc elements select:[:node|
			node name == #authInfo and:[(node attributeAt: 'realm') = realm]])
			do:[:node| changed := true. xmlDoc removeElement: node].

	"Add the new authInfo (unless it got removed)"
	authInfo ifNotNil:[
		changed := true.
		xmlNode := XMLElement new.
		xmlNode name: #authInfo.
		xmlNode addContent: (XMLStringNode string: authInfo).
		xmlNode setAttributes: (Dictionary newFromPairs: {'realm'. realm}).
		xmlDoc elements add: xmlNode.
	].

	changed ifTrue:[
		data := String streamContents:[:s|
			xmlDoc printXMLOn: ((XMLWriterIndenting on: s)
				lineEnd: String crlf; yourself)].
		file := [FileDirectory prefsDirectory forceNewFileNamed: 'authInfo.xml'] 
			on: Error do:[:ex| ex return: nil].
		[file converter: Latin1TextConverter new.
		file nextPutAll: data] ensure:[file close].
	].! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 4/14/2010 17:53'!
saveForum
	"Save the current service island"
	
	self flowControlFlag: true.
	[ 
		self saveForum: serviceSvc.
		QActivityCapture action: #SaveForumFileMenu details: (self currentForumName).

	] ensure: [
		self flowControlFlag: false.
	].
	self saveLayout.	"Why not?"! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 8/2/2008 13:44'!
saveForum: aService

	"Save the given service (all its islands, and its preview.)"

	CProgressBar displayProgress: 'Saving' at: self bounds center 
		from: 0.0 to: 1.0 during:[:bar|
			self saveForum: aService notifying: bar
		].
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 10/8/2008 11:30'!
saveForum: aService notifying: bar

	"Save the given service island, reporting progress. "
	|  islands response barBlock |

	"XXX  ... right now it only saves the preview of the active service,
	 using the island we happen to be in right now,
	 but it should find the entry island and make its preview."

	islands := self islandsInService:aService.
	islands do: [ :island|
		ProgressNotification signal: ''
						      extra: 'Saving ', (aService name) , ' (', ((island get:#name) ifNil:['Main']), ')'.
		barBlock := [ : v | bar value: (0.75 * v) ] fixTemps.
		response := self saveIsland: island in: aService notifying: barBlock.
		response name == #ok ifFalse:[
			okToQuit := false.
			^self inform: 'The save was not completed:', String cr, response contentString
		]].

	ProgressNotification signal: '' extra: 'Saving ', (aService name), ' (Preview)...'.
	bar value: 0.8.
	barBlock := [ : v | bar value: 0.8 + (0.2 * v) ] fixTemps.
	self saveServicePreview: aService informing: barBlock.

! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'ar 4/16/2011 17:34'!
saveForumAs

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The Forum is not ready.'.
	].

	[
		self flowControlFlag: true.
		QActivityCapture action: #SaveForumAsFileMenu details: (self currentForumName).
		^ self saveForumAsTemplate:false
	] ensure: [
		self flowControlFlag: false.
	].
	self saveLayout.	"Why not?"! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'ar 4/16/2011 17:34'!
saveForumAsTemplate

	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The Forum is not ready.'.
	].

	[
		QActivityCapture action: #SaveForumAsTemplateFileMenu details: (self currentForumName).
		self flowControlFlag: true.
		^ self saveForumAsTemplate: true
	] ensure: [
		self flowControlFlag: false.
	].
	self saveLayout.	"Why not?"! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'sn 3/10/2013 14:42'!
saveForumAsTemplate: bTemplate
	"Save (Copy) the current service under a new name.  
	 If bTemplate is true, we are generating a template from this forum.
	 (See #saveIsland:in:notifying, which checks 'service isTemplate')"

	| newName island metadata |

	newName := 
		CDialog request: (bTemplate 
							ifTrue: ['Enter a name for the new template:'translateMe]
					 		ifFalse: ['Enter a name for the new Forum:'translateMe]) 
				initialAnswer: (self currentForumName).
	newName ifNil: [ ^ self ].

	"To change the space/island properties, we have to change them on-island.
	 So we stash thems and ensure-restore them."	
	self readyForEvents ifFalse: [ ^ self ].
	island := self activeSpace island.
	metadata := island get: #forumDocData.
	[
		bTemplate ifFalse: [
			island future appendToSummaryNote: ('Copy of: ', (self currentForumName)).
		] ifTrue: [
			island future originalTemplate: nil.
		].
		island future ctime: (QTimeStamp new 
										who: (self userLogin)
										what: (bTemplate ifTrue: [ 'Create' ] ifFalse: [ 'Copy' ])
										utc: (Time utcSeconds)).
		CProgressBar 
			displayProgress: 'Copying ', (serviceSvc name), ' to ', newName,'...' at: self bounds center 
			from: 0.0 to: 1.0 
			during: [ :bar | self saveForumAsTemplate: bTemplate name: newName informing: bar ].
	] ensure: [
		island isValidFarRef ifTrue: [ 
			island future forumDocData: metadata.
		].
	].
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'ar 4/16/2011 17:34'!
saveForumAsTemplate: bTemplate name: newName informing: bar

	"Save the current service under a new name.  
	 If bTemplate is true, save it as a template.
	 The copy includes all the islands, and all the resources, used in the original.
	 For templates, documents and resources are copied conservatively after reference-checking,
	 but normal copies do a whole-sweep copy (the whole sweep is a great deal faster). "

	|  oldService newService response cc oldIsland snapData snapLoc snap islands i barBlock names where |

	oldService := serviceMgr listServices detect:[:svc| svc name = newName] ifNone:[nil].
	oldService ifNotNil:[
		oldService isActive ifTrue: [
			^ self inform: 'Cannot replace ', newName printString, String cr,
					         'The Forum exists and is in use.'.
		].
		(self confirm: '', newName printString,' already exists.',
			'\Replace the existing forum?' withCRs) ifFalse:[^nil].
		response := serviceMgr deleteService: oldService.
		response name == #ok ifFalse:[
			^self inform: 'Could not delete Forum:\'withCRs, response contentString.
		].
	].

	bar value: 0.01.
	snap := self makeServicePreview: serviceSvc.
	snap ifNotNil: [
		snapData := ByteArray streamContents: [ : stream |	
				| writer | 		
				writer := JPEGReadWriter2 on: stream.
				writer nextPutImage: snap quality: 60 progressiveJPEG: false.
					"Use decent quality for template previews"
				writer close.
			]].
	bar value: 0.1.

	"Let participants know"
	activeSpace future signal: #infoMessage 
					with: '', (self currentForumName), ' is being copied to ', newName.

	"Clone newService from oldService"
	newService := QServiceRecord new copyFrom: (oldService ifNil:[serviceSvc]).
	newService name: newName.
	newService id: (oldService ifNil:[TObjectID new] ifNotNil:[oldService id]).
	newService baseUrl: serviceSvc baseUrl.
	newService users: #().
	newService isTemplate: bTemplate.

	"Switch oldService to the current service and clone the island structure"
	oldService := serviceSvc.
	newService islands: (oldService islands collect:[:oldRecord|
		| newRecord | 
		newRecord := QIslandRecord forService: newService.
		newRecord name: oldRecord name.
		newRecord id: TObjectID new.
		newRecord makerName: oldRecord makerName.

		cc := session islandControllers detect:[:any| any sessionID = oldRecord id] ifNone:[nil].
		cc ifNil:[oldIsland := nil] ifNotNil:[oldIsland := cc island].
		oldIsland ifNil:[
			(self confirm: 'A part of the Forum is missing: ', oldRecord name,'.\Proceed anyway?' withCRs) ifFalse:[^nil].
		].
		newRecord ]).

	bar value: 0.1.
	newService entryPostcard: oldService entryPostcard.
	response := serviceMgr createService: newService.
	response name == #ok
		ifFalse:[ ^self inform: 'The copy could not be created:\\' withCRs, (response contentString) ].

	"Copy the resources. "
	bTemplate ifFalse: [
		"Normal case - a fast sloppy copy (includes old unused documents, etc.)"
		self copyResources: oldService
			to: newService 
			informing: ( [ : v | bar value: (0.3*v + 0.1) ] fixTemps ).
	] ifTrue: [
		"Slow but precise form for templates...
	       Copy only things that remain in use in the service,
		 so that deleted documents are not part of the template payload."
		islands := self islandsInService: oldService.
		islands doWithIndex: [ : island : idx |
			names := session resourceFilesForCopy: island.
			names doWithIndex: [ : fname  : fIndex |
				response := serviceMgr copyFiles: ('services/', oldService name, '/', fname)
										to: ('services/', newService name, '/', fname).
				response name == #ok 
					ifFalse: [ QLogger logWarning: (
								'Cannot copy resource ', fname, ' to ', 
								    newService name, ': ', response contentString) ].
				where := ((idx-1) / islands size asFloat) + (fIndex / names size asFloat).
				bar value: (0.1 + (0.3 * where)).
			]]].

	"Upload islands only *after* the service has been created successfully"
	i := 0.
	bar value: 0.4.
	islands := newService islands.
	islands with: oldService islands do:[:newRec :oldRec|
		cc := session islandControllers detect:[:any| any sessionID = oldRec id] ifNone:[nil].
		i := i + 1.
		(cc notNil and: [cc island notNil]) ifTrue:[ 
			barBlock := [ : v | bar value: ( 0.4 +   (0.5 * ((i-1) / (islands size)) + (v / islands size))) ] fixTemps.
			response := self saveIsland: cc island in: newService notifying: barBlock.
			response name == #ok ifFalse:[ 
				[serviceMgr deleteService: newService] on: Error do: [ ].	"Clean up orphan."
				^self inform: 'The copy could not be created:\\' withCRs, (response contentString)
			].
		].
	].

	snapData ifNotNil: [
		snapLoc := 'services/',newService name, '/preview.jpg'.
		barBlock := [ : v | bar value: ( 0.9  + (0.1 * v)) ] fixTemps.
		response := serviceMgr put: snapData name: snapLoc backup: false informing: barBlock.
	].

	response name == #ok
		ifFalse:[^self inform: response contentString]
		ifTrue: [ 
			self statusAlert:  'Forum copied as ', newName, (bTemplate ifTrue: [' [Template]'] ifFalse: ['']).
			bTemplate ifFalse: [
				self concludeForumCopy: newName serviceRecord: newService.
			].
			servicesPane updateServiceList: false.
		]! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 4/8/2009 13:17'!
saveIsland: anIsland in: qService notifying: bar
	"Regular save of the main forum island."
	^ self saveIsland: anIsland in: qService notifying: bar forAutoSave: false.! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 6/11/2009 23:35'!
saveIsland: anIsland in: qService notifying: bar forAutoSave: bAuto
	"Save the given island."
	| archive data location response barBlock |

	bAuto ifFalse: [
		"Don't incur any of this foolishness for autosave."
		self convertToLazyTextures: anIsland.
		bar value: 0.05.
		self waitTick.
	].

	anIsland future mtime: (
		QTimeStamp new
			who: (bAuto ifTrue: [ 'auto' ] ifFalse: [ self userLogin] ) 
			what: 'Save'
			utc: (Time utcSeconds)).
		"Save-island uses a future-send and result, so it synchronizes."
 
	archive := ZipArchive new.
	barBlock := [ : v | bar value: 0.05 + (v * 0.35). ].
	session saveIsland: anIsland asZip: archive notifying: barBlock.
	bar value: 0.4.
	self waitTick.

	data := ByteArray streamContents:[:s| archive writeTo: s].
	bar value: 0.45.
	self waitTick.

	"Modern servers: Save forum as forum.user.c3x"
	bAuto ifTrue: [
		location := 'services/',qService name, '/', (anIsland get: #name) , '.auto.c3z'.
	] ifFalse: [
		location := 'services/',qService name, '/', (anIsland get: #name) , '.', serviceMgr user, '.c3z'.
	].

	barBlock := [ : v | bar value: 0.45 + (v * 0.45) ].
	[
		self flowControlFlag: true.
		response := serviceMgr put: data name: location informing: barBlock.
		[anIsland future signal: #saved with: (self userLogin) ] on: Error do: [].
	] ensure: [
		self flowControlFlag: false.
	].
	
	response name == #ok ifTrue: [
		ProgressNotification signal: '' extra: 'Saving ', (qService name), ' (Manifest)...'.
		barBlock := [ : v | bar value: 0.9 + (v * 0.1) ].
		[ self saveManifest: anIsland inService: qService notifying: barBlock ] 
		  on: Error 
		  do: [ : ex | QLogger logWarning: 'Manifest not saved: ', (ex description) ].

		"Allow the server to garbage collect any new-imports-since-save."
		serviceMgr ifNotNil: [
			self flag: #AssumesOneSpacePerIsland. "See QFileCard>>space"
			(anIsland syncSend: #resourcesAt: withArguments: {TSpace resourceType})
				ifNotNilDo: [:spaces |
					(spaces first get: #lastCustomsEntryServerTimestamp) ifNotNilDo: [:stamp |
						serviceMgr trimNewImports: qService name since: stamp.
						].
					].
			].
	].
	^response
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 10/20/2009 14:02'!
saveLayout
	"If we're in a meaningful session, save the window and tool layout."

	Tweak fullScreenMode ifTrue: [ ^ self ].	"Not if fullscreen since that's not re-adopted."

	(self readyForEvents and: [ controlBar visible ]) ifTrue: [
		self isSealedApplication ifTrue: [ 
			preferences saveWindow: 'window' shape: (self shapeToRestore).
		].
		preferences saveToolStates: (self toolTable).
	].! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 5/30/2009 19:42'!
saveManifest: anIsland inService: qService notifying: bar
	"Write the QForumManifest (summary of external resources used by this island),
	 as a zipped xml file on the service provider."
	| manifest |
	serviceMgr ifNil: [	
		^ QLogger logWarning: 'Manifest for ', (qService name), ' not written (no SM).'
	].
	manifest := QForumManifest saveManifest: anIsland inService: qService to: serviceMgr notifying: bar.
	manifest ifNotNil: [
		orgServices ifNotNil: [ 
			orgServices sendSignal: #forumManifestUpdated forForum: (qService url) ].
		"Pop the update straight into the local cache.
		 (The cache ignores the org-services signal from self.)
		 Backgrounded to defer to more urgent business during logout-saves."
		infoCache startScript: #updateManifest:forService: 
				   withArguments: { manifest . (qService url) }.
	].	
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 8/2/2009 10:54'!
saveServicePreview: aService informing: bar

	"Save the given service jpg preview. "
	|   snapData snapLoc preview barBlock |

	"Launched as a startscript from the preview ticker, so watch for logout."
	aService ifNil: [ ^ self ].
	self readyForEvents ifFalse: [ ^ self ].

	preview := self makeServicePreview: aService.
	preview ifNotNil: [
		snapData := ByteArray streamContents: [ : stream |	
				| writer | 		
				writer := JPEGReadWriter2 on: stream.
				writer nextPutImage: preview quality: 36 progressiveJPEG: false.
				writer close.
			].
		snapLoc := 'services/', aService name, '/preview.jpg'.
	].

	bar value: 0.1.
	snapData ifNotNil: [
		barBlock := [ : v | bar value: 0.1 + (0.9 * v) ] fixTemps.
		"Save a little server space .. there's no need to back up previews."
		serviceMgr ifNotNil:[
			serviceMgr put: snapData name: snapLoc backup: false informing: barBlock.
		].

		"Announce the update to the organization."
		orgServices ifNotNil: [ 
			orgServices sendSignal: #forumPreviewUpdated forForum: (aService url) ].
		"Finally, update my local caches."
		infoCache storePreview: preview forService: (aService url).
	].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 12/7/2006 13:11'!
sendTimeout: secs from: aController
	"Notify the participant about a send timeout in the given controller.
	We only signal an event here as the method is called from a different thread."
	self signal: #sendTimeout with: secs with: aController.! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/6/2010 16:48'!
serverAuth: aBlock forRepository: repository

	"Manage server authentication for an action on the given repository"

	| token webClient resp request response |
	webClient := repository webClient.
	webClient flushAuthState.

	token := nil.
	[[resp := aBlock value: token.
	resp name == #authenticate] whileTrue:[
		request := WebRequest new readFrom: (resp contentStringAt: 'request') readStream.
		response := request newResponse readFrom: (resp contentStringAt: 'response') readStream.
		webClient authenticate: request from: response.
		token := request headerAt: (response code = 401 
					ifTrue:['Authorization'] 
					ifFalse:['Proxy-Authorization']).
		token isEmptyOrNil  ifTrue:[^resp].
	]] ensure:[webClient flushAuthState].
	^resp
! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/5/2010 15:48'!
serverCheckin: serverPath repository: repoUrl path: repoPath type: repoType

	"Attempt a server-side checkin of the given document.
	May require user credentials to be supplied for authentication.
	Returns the final server-response to indicate success/failure."

	|  repository |
	repository := QDocumentRepository cache at: repoUrl ifAbsent:[nil].
	repository ifNil:[^self error: 'No repository found for ', repoUrl].
	^self serverAuth: [:token| 
		serviceMgr checkin: serverPath repository: repoUrl path: repoPath type: repoType auth: token
	] forRepository: repository.
! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 4/5/2010 15:48'!
serverCheckout: serverPath repository: repoUrl path: repoPath type: repoType

	"Attempt a server-side checkout of the given document.
	May require user credentials to be supplied for authentication.
	Returns the final server-response to indicate success/failure."

	| repository |
	repository := QDocumentRepository cache at: repoUrl ifAbsent:[nil].
	repository ifNil:[^self error: 'No repository found for ', repoUrl].
	^self serverAuth: [:token| 
		serviceMgr checkout: serverPath repository: repoUrl path: repoPath type: repoType auth: token
	] forRepository: repository.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/14/2008 14:25'!
serviceChooser
	"Answer the serviceChooser of the receiver"
	<bewareOf: #serviceChooserChanged>
	^self propertyValueAt: #serviceChooser! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/14/2008 14:25'!
serviceChooser: aValue
	"Modify the receiver's serviceChooser"
	^self propertyValueAt: #serviceChooser put: aValue with: #serviceChooserChanged! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'ar 4/16/2011 17:48'!
serviceEntryIsDirty: svcEntry
	"Should we indicate that the given service(forum) needs to be saved?
	 True if there is a newer change timestamp in it, or if any app inside is marked dirty."
	^ svcEntry needsSave 
	  or: [ svcEntry svcRecord notNil 
		   and: [ (self islandsInService: svcEntry svcRecord) 
			       anySatisfy: [ : i | (i get: #hasAnyDirtyApps) = true ]]].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 7/12/2011 12:55'!
serviceHost
	"Answer the serviceHost of the receiver"
	<bewareOf: #serviceHostChanged>
	^self propertyValueAt: #serviceHost ifAbsent: [self login serviceHost]
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 7/12/2011 12:51'!
serviceHost: aValue
	"Modify the receiver's serviceHost"
	^self propertyValueAt: #serviceHost put: aValue with: #serviceHostChanged! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 6/9/2008 14:56'!
serviceManager: manager
	"Public access to set the service manager.
	 This is done ONLY AFTER the service-manager realm has been specified
	 in the service chooser."
	self serviceManager: manager forceChanges: false.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 11/27/2009 12:45'!
serviceManager: manager forceChanges: bForce

	"A new (or re-configured) service manager to propagate.
       Stop propagation of the previous one, and then broadcast this one instead."

	| script |

	(self exitTripped | bForce) ifTrue: [ 
		orgServices ifNotNil: [
			orgServices destroy.
			orgServices := nil.
			exitTripped ifTrue: [ ^ self]  ]].

	(bForce 
	 or: [ manager ~= serviceMgr]
	 or: [ orgServices isNil]
	 or: [ orgServices isConnected not ])
	  ifTrue: [
		self stopScript: #serviceManagerProcess:forceChanges:.
		(manager ~= serviceMgr and:[serviceMgr notNil]) ifTrue:[
			serviceMgr destroy. 
			svcMgrCache keysAndValuesRemove: [ : k : v | v = serviceMgr ].
			serviceMgr := nil].
		script := self startScript: #serviceManagerProcess:forceChanges: 
					  withArguments: {manager . bForce}.
		[ script waitUntil: #done ] 
			on: Error do: [ :ex |
				manager ifNotNil: [
					"Manager is nil for logout - we expect this process to be curtailed."
					QLogger logWarning: 'Service Manager abandoned: ', (ex description)
				]].
	].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 6/17/2010 10:09'!
serviceManagerProcess: manager forceChanges: bForce

	"PRIVATE Stopabble worker script to initialize the service manager connection;
	 see serviceManager: forceChanges: "
	[
		self assureOrgServices: manager.
		bForce ifFalse: [
			serviceMgr = manager ifTrue: [ ^ self ].
		].
		serviceMgr := manager.
		manager ifNotNil: [manager resetQrl].
		self clearDocumentManager.
		manager ifNotNil: [
			"We have the preferences cached from the provider, but - specifically for federated st.eu,
			 we need to refresh the preferences each time we switch orgs."
			self setupPreferences: ( QUserPreferences forUser: (login user value) inClient: self ) .
			self initServerQPZMenu. "generate menus for all server-side Python apps"
			self installServerPyLib.	   "install updated Python API library from server if required"
		].

		featureProperties := Dictionary new.
		self updateUserRole.		"Do this now so all the tools and panes get the right info below."

		"Set up the tools after org-services, because the tools use org-services."
		self installServiceManager: manager.

	] ifCurtailed: [
		"(Not unexpected, but helpful when deciphering headless transcripts.)"
		QLogger logInfo: 'Service Manager setup canceled.'.
		(self propertyValueAt: #documentManager)
			ifNotNilDo: [ : dm | dm destroy.
						self removePropertyNamed: #documentManager ].
	].
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 1/22/2007 17:32'!
serviceMgr
	"Answer the serviceMgr of the receiver"
	<bewareOf: #serviceMgrChanged>
	^self propertyValueAt: #serviceMgr! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 1/22/2007 17:32'!
serviceMgr: aValue
	"Modify the receiver's serviceMgr"
	^self propertyValueAt: #serviceMgr put: aValue with: #serviceMgrChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/1/2007 09:44'!
serviceSvc
	"Answer the serviceSvc of the receiver"
	<bewareOf: #serviceSvcChanged>
	^self propertyValueAt: #serviceSvc! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/1/2007 09:44'!
serviceSvc: aValue
	"Modify the receiver's serviceSvc"
	^self propertyValueAt: #serviceSvc put: aValue with: #serviceSvcChanged! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 10/1/2007 17:27'!
serviceUrlForUrl: url

	^ self realServiceUrlForUrl: url! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/8/2007 16:59'!
servicesPane
	"Answer the servicesPane of the receiver"
	<bewareOf: #servicesPaneChanged>
	^self propertyValueAt: #servicesPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/8/2007 16:59'!
servicesPane: aValue
	"Modify the receiver's servicesPane"
	^self propertyValueAt: #servicesPane put: aValue with: #servicesPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/9/2006 16:05'!
session
	"Answer the session of the receiver"
	<bewareOf: #sessionChanged>
	^self propertyValueAt: #session! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/9/2006 16:05'!
session: aValue
	"Modify the receiver's session"
	^self propertyValueAt: #session put: aValue with: #sessionChanged! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 6/9/2008 08:55'!
sessionClass
	"The TSessionWindow subclassed used to set up sessions;
	 it in turn defines the class for its viewpanes."

	"This is the main Miramar integration point; see also the users of MAvatarUser"
	^ MWindowedSession! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/6/2009 16:59'!
sessionID
	"The session ID of this user's current session."
	^ session ifNotNil: [ session sessionID ]
			   ifNil: [  orgServices ifNotNil: [ orgServices lobbySessionID ]]! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 3/20/2008 01:04'!
setAsRobot: aBool mode: aSymbol

	"If aBool is true, declare myself a robot, currently operating the given mode symbol;
	 if false, retract my availability as a robot."

	orgServices ifNotNil: [
		aBool ifTrue: [
			orgServices setUser: (self sessionID) robotMode: aSymbol.
		] ifFalse: [
			orgServices stopBeingRobot: (self sessionID).
		].
	].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 5/17/2010 10:46'!
setJitterBufferActivityThreshold: zeroTo100
	"The Speex jitter-buffer has a way to check the 'activity' (roughly, 'volume') of the previously decoded frame.  It also has a way to hint that 'now would be a good time to extrapolate or drop a frame'.  We use a heuristic that says 'when the activity is below a threshold, hint to the jitter-buffer that now is a good time to make time adjustments'.  This method sets this threshold."
	audioSession speakerManager future jitterBufferActivityThreshold: zeroTo100
! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 2/4/2007 23:08'!
setNewUserPreferences: aUserPreferences

	"Subclass opportunity - specify default user preferences for this application.
	See QUserPreferences #initialize for the base defaults."

	aUserPreferences pointOfView: 'First Person'! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:25'!
setWantsToSpeak: bWants
	"Notice from a UI control that my user wants to speak (or not)."
	self readyForEvents ifFalse: [
		^ controlBar askToSpeak selected: false.
	].
	(bWants not and: [ self isMeetingLeader not ])  ifTrue: [
		"Cede the speaker."
		meetingState setLogin: self userLogin allowedToSpeak: false.
	].
	meetingState setLogin: self userLogin wantsToSpeak: bWants.
	controlBar askToSpeak selected: bWants.
	QActivityCapture action: #MeetingWantsToSpeak details: bWants asString.
	! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 3/27/2009 14:19'!
setWindowBounds: aRectangle
	"Set the size of the UI window
	 to accomodate the given inner (Tweak- panel) rectangle "
	self isSealedApplication ifFalse: [ 
		self setWindowDisplaySize: (aRectangle extent).
	] ifTrue: [
		| chrome |
		"To handle legacy saved shapes with zero origin... 
		 detect that, and divert to just size-change."
		(aRectangle top ~= 0 or: [aRectangle left ~= 0]) ifTrue: [
			HostWindows squeakWindowSetBounds: aRectangle.
		] ifFalse: [
			self setWindowSize: (aRectangle extent).
		].
	].

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 3/27/2009 14:20'!
setWindowDisplayBounds: aRectangle
	"Set the size of the UI window
	 to accomodate the given inner (Tweak- panel) rectangle "
	self isSealedApplication ifFalse: [ 
		self setWindowDisplaySize: aRectangle extent
	] ifTrue: [
		"To handle legacy saved shapes with zero origin... 
		 detect that, and divert to just size-change."
		(aRectangle top ~= 0 or: [aRectangle left ~= 0]) ifTrue: [
			HostWindows squeakDisplaySetBounds: aRectangle.
			self window invalidate.
		] ifFalse: [
			self setWindowDisplaySize: aRectangle extent
		].
	].

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 6/21/2010 17:27'!
setWindowDisplaySize: anExtent
	"Set the size of the UI (complete, chrome-decorated) window
	 to accomodate the given inner (Tweak- panel) extent."
	self isSealedApplication ifFalse: [ 
		self setWindowSize: (anExtent + (10@30)).	"SystemWindow wrappings"
	] ifTrue: [
		(HostWindows squeakDisplayRectangle extent ~= anExtent) ifTrue: [
			HostWindows squeakDisplaySetSize: anExtent.
		].
		self window invalidate.
	].

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 2/7/2009 21:01'!
setWindowSize: aPoint
	"Set the outer window size to the given extent. 
	 The given size has to include allocations for the platform window chrome,
	 so that this operation is idempotent with (Smalltalk windowSize)."
	self isSealedApplication ifTrue: [ 
		HostWindows squeakWindowSetSize: aPoint
	] ifFalse: [
		self window window extent: aPoint.
	].

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 4/2/2009 10:46'!
setupAvatar
	"--- set up the avatar model ---"
	|  avtr |

	avtr := self avatarUserClass new.
	avtr tweakHand: (self hand island asFarRef: self hand). "Eeeeeek!!"
	avtr replicaClass: self avatarReplicaClass.
	avtr userPreferences: (self preferences).

	"And now let all the other avatar change listeners fire."
	self avatar: avtr.

	"Install the avatar model (costume), which may be application-defined
	 or extracted from user preferences."
	self setupAvatarCostume.

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 8/21/2008 15:06'!
setupAvatarCostume

	"Setup up the avatar-user with the user's preferred avatar costume,
	 supplying my default if the user has no saved preference."

	|  cstm |

	[
		cstm := self preferences avatarXMLZip.
	] on: Error do: [:exc|
		QLogger logWarning:  'Ignoring the stored avatar XML: ', exc description.
		cstm := nil
	].

	(cstm isNil) ifTrue:[
		cstm := self defaultAvatarCostume.
		self preferences avatarCostumeData: cstm.
	].

	avatar costumeXMLZip: cstm. 
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'sn 3/9/2013 23:21'!
setupCostume
	| w |
	super setupCostume.
	self fillColor: QColors paneDarkerColor.
	self clipping: true.
	self add: viewPanel.
	self add: usersPane.
	usersPane owner: self;  visible: false.
	self add: placesPane.
	placesPane owner: self; visible: false.
	peoplePlacesSash
			target: self; 
			selector: #peoplePlacesSashTop:;
			memo: (self defaultPPSashFraction);
			visible: false.
	self add: peoplePlacesSash.
	self keyDriveRedirect: peoplePlacesSash.
	viewAreaSash
			target: self; 
			motion: #horizontal;
			selector: #viewAreaRight:;
			width: 4;
			borderWidth: 0;
			visible: false;
			fillAlpha: 0;
			memo: (self defaultPPWidth).
	self add: viewAreaSash.
	self keyDriveRedirect: viewAreaSash.

	videoAreaSash
			target: self; 
			selector: #viewAreaBottom:;
			visible: false;
			memo: self defaultVidPaneHeight.
	self add: videoAreaSash.
	self keyDriveRedirect: videoAreaSash.
	self add: webcamPane.
	webcamPane owner: self; visible: false; enabled: false.

	self add: self menu.
	self add: controlBar.
	controlBar owner: self.

	self add: navPanel.
	navPanel owner: self.
	self keyDriveRedirect: navPanel.

	self add: (statusLine floater).
	posReadout owner: self.

	self extent: (self class defaultExtent).

	w := self window window.
	w ifNotNil:[w extent: (self class defaultExtent)].
	self color: (QColors paneDarkerColor).		"See also class #runAsApp"

	self menu borderColor: (QColors mainBorderColor).
	self addFloatingPane: treeInspector.
	self addFloatingPane: materialCatalog.
	self addFloatingPane: chatPane.
	self addFloatingPane: avatarPane.

	self addFloatingPane: crowdPane.
	self addFloatingPane: phonePane.
	self addFloatingPane: servicesPane.
	self addFloatingPane: documentsPane.
	self addFloatingPane: customsPane.

	self addFloatingPane: logView.
	self addFloatingPane: statisticsPane.
	self addFloatingPane: trafficPane.
	self addFloatingPane: latencyPane.
	self addFloatingPane: recorderPane.
	self addFloatingPane: webcastPane.
	self addFloatingPane: captureProxy.
	self addFloatingPane: quickReference.
	self addFloatingPane: windowSharePane.
	self addFloatingPane: repoBrowser.

	self add: catalogPane.
	catalogPane owner: self.
	historyPane owner: self.		"No longer a user-visible panel, just a list-manager"

	keyboardIcon borderWidth: 0;
		 color: Color transparent;
		 normalImage:  (QForms find: 'icons/Qwaq/Keyboard.png');
		 balloonHelp: 'You are in typing mode. Input goes to applications. Keyboard driving is off.\Hit <Escape> to return to driving mode.'translateMe withCRs;
		 extent: 68@25;	
		 layoutFrame: CLayoutFrame new;
		 fillAlpha: 0.9;
		 visible: false;
		 sticky: true.
	self startScript: #onKeyStrokeBar: when: { keyboardIcon . #keyStroke }.

	awayOverlay borderWidth: 0;
		borderStyle: #simple;
		graphicForm: (QForms find: 'AwayOverlay.png');
		graphicFit: #rigid;	"Would prefer scale, but it whacks the transparency."
		fillAlpha: 0.0;
		visible: true;
		sticky: true.

	ppToggle borderWidth: 0;
		extent: 28@21;
		fillAlpha: 0;
		balloonHelp: 'Hide People & Landmarks'translateMe;
		normalImage: (QForms find: 'pp-expand.png');
		overImage: (QForms find: 'pp-expand-over.png');
		selectedImage: (QForms find: 'pp-collapse.png');
		selectedOverImage: (QForms find: 'pp-collapse-over.png');
		select;
		sticky: true.
	self menu add: ppToggle.

	backButton borderWidth: 0;
		defaultImage: (QForms find: 'BackButtonOff.gif');
		balloonHelp: 'Go back to the previous space.'translateMe;
		extent: 24@24;
		fillAlpha:0;
		sticky: true.
	self menu add: backButton.
	forwardButton borderWidth: 0;
		defaultImage: (QForms find: 'FwdButtonOff.gif');
		balloonHelp: 'Go back to the previous space.'translateMe;
		extent: 24@24;
		fillAlpha:0;
		sticky: true.
	self menu add: forwardButton.
	historyPane backButton: backButton; forwardButton: forwardButton.

	recordingIndicator borderWidth: 0;
		defaultImage: (QForms find: 'Recording.png');
		selectedImage: (QForms find: 'Recording2.png');
		extent: (recordingIndicator normalImage extent x @ 24);	"Pin height to protect menubar"
		fillAlpha: 0;
		sticky: true.
	self menu add: recordingIndicator.
	self startTicking: #flickerRecordingIndicator at: 1.1.
	recordingIndicator visible: false.
	
	webcastIndicator borderWidth: 0;
		defaultImage: (QForms find: 'Webcasting.png');
		selectedImage: (QForms find: 'Webcasting2.png');
		extent: (webcastIndicator normalImage extent x @ 24); 	"Pin height to protect menubar"
		fillAlpha: 0;
		sticky: true.
	self menu add: webcastIndicator.
	self startTicking: #flickerWebcastIndicator at: 1.7.
	webcastIndicator visible: false.

	posReadout visible: false.
	self add: posReadout.

	sqScriptEditor ifNotNil: [
		sqScriptEditor visible: false.
		sqScriptEditor color: (Color white alpha: 0.8).
		sqScriptEditor borderWidth: 2.
		sqScriptEditor borderColor: CWidgetLibrary default windowColor.
		sqScriptEditor borderStyle: #raised.
		self add: sqScriptEditor.
	].

	self linkIcon: (controlBar inventory)  toTool: (catalogPane).

	statusLine label: (login messageArea); owner: self.
	self showLogin.
	self setupLocalProvider.	"Has to follow login-setup, where <local> is detected."
	meetingState owner: self.

	self keyDriveRedirect: self menu.! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 2/3/2007 00:45'!
setupError: aString
	"Handle an error during connection setup. Inform the user about the problem, give him the option to show the log (if requested) and start over when done."
	self reportError: aString details: logView panel contents 
			label: 'Connection Error' button: 'Start over'.

	^ self logoffConfirming:  false.
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'ar 3/12/2009 17:44'!
setupLocalProvider

	"If the login screen includes the 'My Local Services' option, start a local provider."
	exitTripped ifFalse: [
		login hasLocalServices ifTrue: [
			QLocalServiceProvider reset.
			(QLocalServiceProvider default) startForClient: self.	
		]].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'jcg 5/11/2010 18:27'!
setupMainView

	"Set up our main croquet window and viewpane."
	| view |

	session owner: self.

	view := session addMainView.
	view avatar: avatar.
	view bounds: ((0@0) extent: (520@360)).	"See makeServicePreview: "

	session audioSession: audioSession.
	audioSession ifNotNil: [latencyPane startWatching: audioSession micStats].

	session addOverlay: self.
	session idleFramesPerSecond: preferences idleFramesPerSecond.

	"Initialize menu states..."
	(self menu itemTagged:#ignoreMotionRules) 
		ifNotNilDo:[:item| item checked: false].

	"This relies on the view's avatar already being set."
	self setupVideoSessionInView: session.

	"xxx: Temporary workaround for an apparent timing-related bug which would eventually cause openAL to be nil in the main view and thusly disable the microphone even though the initialization went fine. Not sure where the timing comes in though, possibly in QwaqALSession creation?"
	viewPanel view: view.
	mainView := view.
	navPanel view: view.

	session updateLastEventTime.
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 11/23/2009 12:05'!
setupPreferences: prefs
	"Fetch the preferences from the service; this is at service-manager login. "

	"Nasty convenience for GUEST users: fresh preferences each login."
	self isReusableGuestUser ifTrue: [
		self statusInfo: 'Guest access granted.'.
		QLogger logInfo: 'Clearing preferences for new guest session.'.
		[ 
			QUserPreferences resetUser: (self userLogin) forOrganization: serviceMgr 
		] on: Error do: [ : ex| 
			QLogger logWarning: 'Cannot clear preferences: ', ex description 
		]].

	(prefs notNil and: [preferences = prefs]) ifTrue: [	
		self onPreferencesChanged.		"Installing what we prefetched at login."
	] ifFalse: [
		self preferences: (prefs ifNil: [ QUserPreferences forUser: (login user value) inClient: self ]).
	].

	"Disable risky/expensive preferences for VM crash recovery"
	crashRecovery ifTrue: [
		"preferences useOpenAL: false."    "It is now probably safer to leave openAL on"
		preferences doAntialiasing: false.
	].

	[
		audioSession initializeFromPreferences: preferences
	] on: Error do: [ : ex |
		QLogger logWarning: 'Cannot initialize audio: ', (ex description).
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'jcg 5/11/2010 15:50'!
setupService: qService usingIslandFiles: islandFiles
	"Initialize the client session, logging in to the given service (forum.)
	 This starts our Croquet Session, and arranges for us to teleport into the service.
	 IslandFiles is usually nil - meaning use the most recent available persistence;
	 if specified, it is a list of persistence files (service-provider paths)
	 used to open a forum to some previous saved state."
	| preview |
	self mouseCursor: Cursor wait.

	self showNormalView.
	preview := (infoCache previewForUrl: qService url).
	preview ifNotNil: [
		self graphic: (qService isActive
						ifTrue: [ self class asMutedColor: preview] 
						ifFalse: [ self class asTonedGray: preview ] ).
		self graphicFit: #stretch.
	] ifNil: [
		self graphic: self class defaultSplashForm.
		self graphicFit: #stretch.
	].
	self startScript: #loginUrl:forumPreviewChanged:byUser: withArguments: {qService url}
		when: { orgServices . #forumPreviewUpdated }.

	urlCache := Dictionary new.
	session := self sessionClass new.
	session sessionID: (orgServices lobbySessionID).
	session frameStats: fpsStats.
	session owner: self.
	session userLogin: (self userLogin).
	session versionInfo: QwaqVersion current infoString.
	session documentManager: (self documentManager).
	QActivityCapture default sessionID: (session sessionID hex).
	orgServices ifNotNil: [
		"Reassert my presence under my new Croquet sessionID,
	       claiming middling quality because I'm synciing."
		orgServices noteLobbyPresenceWithQuality: 3
	].

	self setupAvatar.
	self setupMainView.
	preview ifNotNil: [ viewPanel preview: preview ].
	self setupSounds.
	self mouseCursor: nil.
	"Animation: The configuration category methods on TAnimatedMesh class needs access in 
	order to reset openGL. TAnimatedMesh also tracks rendering speed, so the following is done
	after mainView is set."
	TAnimationManager ui: self.

	[
		" -- This is where we load the forum ---"
		self flowControlFlag: true.
		serviceSvc := self activateService: (qService url) fromTemplate: nil usingIslandFiles: islandFiles
	] ensure: [
		self flowControlFlag: false.
	].

	session updateLastEventTime.
	serviceSvc ifNotNil: [
		" --- This tells the WindowedSession to take our avatar into that forum ---"
		mainView entryPostcard: (TPostcard urlString:
					(TPostcard makeUrl: serviceSvc entryUrl relativeTo: serviceSvc url)).
	] ifNil: [
		(CProgressCancel new)	"I.e. assume a nil result has already been 'inform:'ed "
			messageText: 'Cannot join ', qService url;
			signal.
	].

	Processor activeProject displayHook: self.
	self startEntryZoom.
	sanityChecker startSpeakerChecker.

	"If the user favors press-to-talk, get it set up. Otherwise, default will be mic off.
	We don't want to override any other setting that might have been set by early sound
	or a setup dialog."
	preferences pressToTalk ifTrue: [
		controlBar microphone micOnPressToTalk.
		controlBar microphoneButtonFired: true 
	].

	QLogger logInfo: 'Participant ready for session.'! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 8/9/2007 11:04'!
setupSounds

	"Initialize user preferences."
	self audioSession ifNotNilDo: [ :as|
		as useSoundEffects: (preferences doSoundEffects).
	].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 10/15/2009 19:46'!
setupVideoSessionInView: view "a QMainView"
	| videoSession camMgr camBtn |
	videoSession := QwaqVideoSession new.
	view videoSession: videoSession.
	
	camMgr := videoSession cameraManager.
	camBtn := controlBar camera.
	camBtn runScript: #selected: withArguments: #(true) when: {camMgr. #startedVideoChat}.
	camBtn runScript: #selected: withArguments: #(false) when: {camMgr. #endedVideoChat}.
	
	self startTicking: #videoSessionTick at: 0.333.	"Every 3 seconds."! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 2/7/2009 19:44'!
shapeToRestore
	"For preferences tool states."
	^ self isSealedApplication ifTrue: [
		HostWindows squeakWindowRectangle
	] ifFalse: [
		(0@0) extent: ((self window window extent) ifNil: [ self class defaultExtent ]).
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 1/26/2009 18:19'!
shareFile
	"Share a file"
	| fileName fileStream |
	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].
	fileName := QFileDialog openFile: {{'All Files' . '*.*'}} label: 'Choose a File to Import'.
	fileName ifNil: [ ^ nil .].
	fileStream := FileStream readOnlyFileNamed: fileName.
	[session dropPersonalShare: fileStream event: nil] ensure:[fileStream close].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 1/28/2009 18:28'!
shareWindow: wid
	"Share a window with the given ID"
	self readyForEvents ifFalse: [ 
		^ self statusAlert: 'The space is not ready.'.
	].
	session ifNotNil:[session personalShareWindow: wid].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'hrs 5/5/2009 08:41'!
shellOpen: argument documentDescription: docDescription applicationDescription: appDescription
	| platformClass |
	
	[
		platformClass := CPlatform current class.
		platformClass == CMacOSXPlatform ifTrue: [
			| result | 		
			result := (MacOSShell new 
						system: ('/usr/bin/open ', (MacOSShell escapeFileName: argument))).
			(result = 0 ) ifFalse: [
				QLogger logInfo: ' ', (MacOSShell escapeFileName: argument).
				self error: 'Open failed (', result, ')' 
			].
			self statusInfo: 'Opening ', docDescription, ' in ', appDescription.
			^ self 
		].
		platformClass == CWin32Platform ifTrue: [
			Win32Shell new shellOpen: argument.
			self statusInfo: 'Opening ', docDescription, ' in ', appDescription.
			^ self
		].
	] on: Error do: [:ex | 	
		| msg |
		(ex description includes: 'system error, code:') ifTrue: [
			QLogger logWarning: ex description, ' for ', appDescription.
			"Useless generic windows errror"
			msg := 'You may not have the necessary application.'
		] ifFalse: [
			msg := ('There was a problem starting the ', appDescription, '.'),
					(String cr), (String tab), (ex description)
		].
		^ self inform: 
				('The ', docDescription, ' could not be opened.', (String cr), 
				msg,
				String cr, String cr,
				'(There may be other information in system dialogs.)' )
	].

	"Fallthrough"
	self inform: 'External ', docDescription, ' launch is not yet supported on this platform'.! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 3/31/2009 18:02'!
shootOrgServicesController
	"Kill the OrgServices controller's socket."
	| socket |
	[socket := orgServices controller connection socket]
		on: Error 
		do: [^self 
				signalInformDialog: 'Could not find a socket to kill' 
				label: 'Did not pull the trigger...'].
	socket destroy.! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 2/21/2008 17:36'!
shootTheMediaController
	"Kill the media controller's socket."
	| socket |
	[socket := session mediaController connection socket]
		on: Error 
		do: [^self 
				signalInformDialog: 'Could not find a socket to kill' 
				label: 'Did not pull the trigger...'].
	socket destroy.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'sn 5/13/2011 13:56'!
shortQuitLabel: logout
	"A short label suitable for a button, indicating the action that will be taken on quitting this forum."
	^logout caseOf: {
		[true] -> ['Go to Lobby' translateMe].
		[false] -> ['Quit' translateMe]
	} otherwise: ['Go' translateMe]. "going to another service"
	! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 7/12/2011 12:55'!
showAboutBox

	QActivityCapture countAction: #HelpMenuAbout.
	serviceMgr ifNotNil:[
		aboutBox serviceName value: ('Organization:	', (serviceMgr orgName)).
		aboutBox serviceHost value: ('Hosted At:	'), (self serviceHost).
		aboutBox tosLabel value: (self class tosLabelForOrg: (serviceMgr currentOrg)).

		self userLogin ifNotNilDo: [ : ln |
				aboutBox userLogin value: ('Logged In As:	', ln)
			] ifNil:[
				aboutBox userLogin value:('Not Logged In')
			]].

	aboutBox versionLabel label: (aboutBox infoString).
	aboutBox position: self dialogCenter;
			  layoutEverything.
	self add: aboutBox.
	aboutBox visible:true.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:21'!
showFeedbackDialog
	QExitQuizDialog 
		runFor: self 
		quitLabel: 'Submit' translateMe
		questions: (QExitQuizDialog 
						surveyQuestions: self
						defaultStatement: 'My OpenQwaq experience is excellent.'
						verb: 'is'
						serviceManager: self serviceMgr)
		name: 'feedback'.! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/14/2010 10:58'!
showFixedPanes
	"Reveal all the in-session chrome."
	controlBar visible: true; enable.
	navPanel visible: (preferences doNavPanel and: [ self allowedToMove]); enable; collapse.
	self menu visible: true; enable.
	keyboardIcon visible: true.
	usersPane visible: true; enable.
	placesPane visible: true; enable.
	peoplePlacesSash visible: true; enable.
	viewAreaSash visible: true; enable.
	webcamPane numItems > 0 ifTrue: [
		videoAreaSash visible: true; enable.
		webcamPane visible: true; enable.
	] ifFalse: [
		videoAreaSash visible: false; disable.
		webcamPane visible: false; disable.
	].
	self menu comeToFront.
	controlBar comeToFront.
	navPanel comeToFront.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 12/20/2008 12:59'!
showFloatingAppPane: anEmbeddedAppRef withAlignment: anAlignment
	"Given a far ref to a TEmbeddedApp, 
	 open (or raise to front) a QFloatingEmbeddedApp for it."

	self floatingPanes do:[ : pane |
		pane isFloatingEmbeddedApp ifTrue:[
			pane targetApp = anEmbeddedAppRef ifTrue:[ 
				pane expand.
				pane alignToOwner.
				pane comeToFront. 
				^ pane
			]]].

	QFloatingEmbeddedApp openApp: anEmbeddedAppRef in: self alignment: anAlignment.

! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 1/17/2009 10:58'!
showForumWelcome
	"Menu operation - open the welcome-booklet app if one is configured,
	 in a floating-app window distinctively positioned."
	self readyForEvents ifTrue: [ 
		self showForumWelcomeSoonFor: activeSpace
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 3/27/2009 13:05'!
showForumWelcomeSoonFor: aSpace
	"Open the entry-time welcome-booklet app if one is configured,
	 in a floating-app window distinctively positioned.
	 We build in some background delay 
	 so the app + harness signaling madness can get the display size set up."

	(self readyForEvents and: [ activeSpace island = avatar replica island ])
		ifFalse: [ ^ nil ].
	(activeSpace future welcomeDocumentAsApplicationForUser: (avatar replica))
	 whenResolved: [ :appRef |
		(Delay forMilliseconds: 250) wait.
		((aSpace = activeSpace) and: [appRef isValidFarRef])	"We're still here?"
		 ifTrue: [
			self signal: #welcomeAppResolved with: appRef 
		]].

! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 3/25/2010 13:32'!
showLogin
	"Posture the UI for initial login, by hiding all else and showing the login pane."
	self signal: #unready.
	self useNormalTimeouts.
	self graphic: nil.
	self hideTools.		"These are restored by onReady"
	controlBar visible: false.
	serviceChooser visible: false.
	login fullBounds.
	login position: (login extent // 2).
	self setWindowDisplaySize: (login extent).
	login position: (login extent // 2).
	statusLine label: (login messageArea).
	self updateTitle.
	self remove: login; add: login.
	login visible: true.

	self startScript: #invalidate.! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 7/16/2009 15:40'!
showNormalView
	"Hide the login dialogs, and switch to the normal session posture,
       preparatory to entering an actual forum space."

	login visible: false. "So login can tell if it's up or not."
	serviceChooser visible: false.
	self remove: login.
	self remove: serviceChooser.
	(Delay forMilliseconds: 25) wait.	"Get some paint."
	self restoreShape.
	controlBar visible: true.
	statusLine label: (controlBar messageArea).
	statusLine setStyleForWindowWidth: (self width).

	(Delay forMilliseconds: 25) wait.	"Get some paint."
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/21/2008 01:07'!
showPane: aToolPane
	"Make the given tool pane fully visible if it is not; draw attention to it."

	aToolPane comeToFront.	
	aToolPane isOpen ifFalse: [
		aToolPane expand.
	] ifTrue: [
		aToolPane fillAlpha: 1.0.		"Flash opaque so you see which one it is."
		self waitTick.
		self waitTick.
		aToolPane startScript: #tookFocus.
	].! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 9/13/2008 01:05'!
showPanels
	"Restore from a previous hidePanels"
	controlBar visible ifFalse:[
		self showFixedPanes.
		visiblePanes ifNotNil:[
			visiblePanes do:[:vp | vp visible: true].
		].

		"Heuristic step away - dropping out of zoom-to-window, 
		 move back so context comes back into view."
		self readyForEvents ifTrue: [ avatar moveBack: 2.5 ].

		self startScript: #layoutEverything.
	].
	! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 7/18/2009 15:37'!
showPanels: stickyViewpoint
	
	"If we've dropped out of the sticky viewpoint
	 (presumably set up by a hidePanels: for zoom-100)
	 then go back to normal posture."

	(stickyViewpoint isNil or: [ stickyViewpoint > 0]) ifTrue: [
		self showPanels
	].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 7/18/2008 15:05'!
showPhonePane
	"Make the phone pane fully visible if it is not."
	
	self showPane: phonePane.! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 11/20/2009 20:12'!
showPreferences
	"Activate the user preferences dialog."

	QActivityCapture action: #PreferencesDialog.
	self menu disable.

	"Make sure it will fit."
	self assureDisplaySize: (QwaqPrefsPane fixedExtent + 6).
	self waitTick.
	prefsPanel ifNil: [
		prefsPanel := QwaqPrefsPane new.
	].
	prefsPanel position: (self dialogCenter).

	prefsPanel organizations: (serviceChooser orgChoices contents).
	prefsPanel use: preferences.
	self add: prefsPanel.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/15/2008 20:19'!
showQuickReference: fileName url: aURL

	"Show one of the help flash cards.  The second argument is a docs URI for more info."

	[
		QActivityCapture countAction: #HelpMenuQuickReference details: fileName.

		quickReference card:  fileName url: aURL.
		quickReference restoreToFactory; 
						alignToOwner;
						expand.
	] on: Error do: [ 
		self inform: 'The quick reference could not be found.\Your software installation may be damaged.' withCRs.
	].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 4/16/2011 17:34'!
showReconnectDialog: giveUpBlock
	| dlg button |
	"Reconnection in progress... show the user"

	self mouseCursor: Cursor wait.

	dlg := CUserDialog new.
	dlg label: 'Connection Problem'.
	dlg message: 'Lost connection to OpenQwaq host.\Recovery in progress.\\Please wait...\' withCRs.
	dlg messageText allowEditing: false.
	dlg messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	button := dlg addButton: 'Quit OpenQwaq'.
	button extent:100@24.

	"Confirm that the user insists on quitting."
	dlg startScript:[
		(self confirm: 'Log Out of OpenQwaq?') ifTrue: [
			dlg close.
			giveUpBlock value.
		].
	] when: {button. #fire}.

	"Don't show the dialog right away, maybe we can recover faster than that."
	dlg startScript: [
		|dots| dots := '.............'.
		1 to: 12 do: [ :i |		"Counting the seconds..."
			dlg ifNil: [ ^ self ].
			self statusAlert: 'Lost a connection.  Recovery in progress', (dots copyFrom:1 to:i).
			(Delay forSeconds: 1) wait.
		].
		"After that number of seconds, relent and offer the dialog, with its drop-dead option."
		dlg ifNotNil: [ dlg openCenteredAt: self dialogCenter. dlg mouseCursor: Cursor normal. ].
	].

	^ dlg! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 3/25/2010 13:32'!
showServiceChooser
	"Pose the UI for the second phase of login (the org/service chooser).  
	 Assumes the service manager has been authenticated."
	self signal: #unready.
	self useNormalTimeouts.
	self graphic: nil.
	self hideTools.		"These are restored by onReady"
	controlBar visible: false.
	login visible: false.
	statusLine label: (serviceChooser messageArea).
	serviceChooser mouseCursor: Cursor wait.
	serviceChooser disable.
	serviceChooser restoreShape.
	[
		self remove: serviceChooser; add: serviceChooser.
		serviceChooser visible: true.
		self updateTitle.
	]  ensure: [
		serviceChooser mouseCursor: nil.
		serviceChooser enable.		
	].
	self window invalidate.
	self waitTick.
	self window invalidate.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/21/2008 00:23'!
showServicesPane
	"Make the services pane fully visible if it is not."

	self showPane: servicesPane.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 12/16/2006 18:09'!
showSqScriptEditor
	|  |

	sqScriptEditor ifNotNil: [
		sqScriptEditor space: mainView activeSpace.
		sqScriptEditor visible: sqScriptEditor visible not.
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 12/16/2006 17:41'!
showTranscript
	|  |
	logView visible: logView visible not.
	! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 6/16/2010 17:46'!
showWebcamPane
	"Make the web-cam pane visible if it is not."

	self menu visible ifTrue: [	"... if we're in full-session poise."
		webcamPane visible: true; enable.
		videoAreaSash visible: true; enable.
		self signal: #extentChanged with: extent.
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 1/29/2009 22:42'!
showWindowSharePane
	"Make the window share pane fully visible if it is not."
	self showPane: windowSharePane.
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'jcg 10/15/2007 18:27'!
signalInformDialog: informationalString label: labelString 
	self signal: #openInformDialog with: informationalString with: labelString! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 3/13/2009 13:39'!
simplerLogin
	"This is a simplified version of the email login, with the dots and @host stuff removed.
	 It is used as a fallback when using the login to make non-unique but user-friendly tags,
	 since we cannot count on having a screen name (and the screen name varies.)"
	^ QStyles simplerLoginName: (self userLogin).
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 4/7/2008 11:45'!
snapshotBool
	"Answer the snapshotBool of the receiver"
	<bewareOf: #snapshotBoolChanged>
	^self propertyValueAt: #snapshotBool! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 4/7/2008 11:45'!
snapshotBool: aValue
	"Modify the receiver's snapshotBool"
	^self propertyValueAt: #snapshotBool put: aValue with: #snapshotBoolChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 6/8/2008 16:38'!
snapshotIslands: everySeconds

	"This is to force many snapshots over a short period of time. This is an attempt to find and 
	force errors."

	| delay controller |

	self viewPanel mouseCursor: (Cursor wait).
	delay := Delay forMilliseconds: everySeconds*1000. " seconds"
	[snapshotBool] whileTrue: [
		delay wait.
		self readyForEvents ifTrue: [
			self statusStatus: 'Taking snapshot.'.
			controller := activeSpace controller.
			controller snapshot.
		].
	].
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 9/6/2006 11:19'!
sqScriptEditor
	"Answer the sqScriptEditor of the receiver"
	<bewareOf: #sqScriptEditorChanged>
	^self propertyValueAt: #sqScriptEditor! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 9/6/2006 11:19'!
sqScriptEditor: aValue
	"Modify the receiver's sqScriptEditor"
	^self propertyValueAt: #sqScriptEditor put: aValue with: #sqScriptEditorChanged! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 6/11/2009 23:33'!
stampIsland: island asCreatedFrom: aTemplateService
	"Utility used from service-link when instantiating from a template.
	 Mark the creation info and default summaryNote suitably.
	 (Needs to be done after the island is alive but before the initial save is generated,
	 so this is called from the service-link code.) "

	island isValidFarRef ifTrue: [
		aTemplateService ifNotNil: [
			island future originalTemplate: (aTemplateService name).
		].
		island future ctime: (QTimeStamp new
						who: (self userLogin)
						what: 'Create'
						utc: (Time utcSeconds))
	].

	"We know a new service entry has been made... this is just a convenience."
	servicesPane updateServiceList: false.


! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 12/3/2007 22:31'!
startAsDemon: aMode 
	"See runRobotic: -
	 this is a one-shot startup time trigger to become a demon once logged inn."

	self beADemon: aMode.
	self stopScript: #startAsDemon:.	 "Don't want every space change to restart the demon."

! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 8/22/2009 18:49'!
startEntryZoom
	"When first entering the space,
	 we zoom gently in from far third-person to the current view preference,
	 to orient the user and remind them of their avatar appearance."

	"The zoom-in begins once we're in the space."
	self stopScript: #forkEntryZoom.
	self runScript: #forkEntryZoom when: {mainView . #activeSpaceChanged}.
! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'bgf 6/8/2008 19:16'!
startVideoChat
	self statusInfo: 'Starting camera'.
	session videoSession ifNotNilDo: [:vs | vs cameraManager startVideoChat]! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/30/2010 11:21'!
startVoiceChat
	"Tell the audio-session to start voice-chat, and register for mic-activity changes to provide UI feedback."
	audioSession ifNotNilDo: [ :as|
		as future startVoiceChat.
		self statusInfo: 
				'Microphone On. ', 
				(as isPressToTalk ifTrue: ['  Hold down control key to talk.'] ifFalse: ['']).
		self stopScript: #micActivity:.  "Belt&braces, endVC also does this."
		self startScript: #micActivity: when: {as voiceRelay . #activity}.
	].


! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 5/12/2007 19:02'!
statisticsPane
	"Answer the statisticsPane of the receiver"
	<bewareOf: #statisticsPaneChanged>
	^self propertyValueAt: #statisticsPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 5/12/2007 19:02'!
statisticsPane: aValue
	"Modify the receiver's statisticsPane"
	^self propertyValueAt: #statisticsPane put: aValue with: #statisticsPaneChanged! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/26/2007 19:52'!
statusAlert: message
	"Deliver the message to the status line at high importance.
	 This is done through a signal so it is safe from any Process."
	self signal: #alertMessage with: message.	
	(Delay forMilliseconds: 1) wait.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/26/2007 19:48'!
statusFromAvatar: aString
	"Private handler; see updateStatusRelay."
	statusLine status: aString.
	self waitTick.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/26/2007 19:47'!
statusFromSpace: aString
	"Private handler; see updateStatusRelay."
	statusLine status: aString.
	self waitTick.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/26/2007 19:51'!
statusInfo: message
	"Deliver the message to the status line at medium priority.
	 This is done through a signal so it is safe from any Process."
	self signal: #infoMessage with: message.
	(Delay forMilliseconds: 1) wait.! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/20/2006 10:47'!
statusLine
	"Answer the statusLine of the receiver"
	<bewareOf: #statusLineChanged>
	^self propertyValueAt: #statusLine! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/20/2006 10:47'!
statusLine: aValue
	"Modify the receiver's statusLine"
	^self propertyValueAt: #statusLine put: aValue with: #statusLineChanged! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 8/26/2007 19:51'!
statusStatus: message

	"Deliver the message to the status line, as low-importance status.
	 This is done through a signal so it is safe from any Process."
	self signal: #statusMessage with: message.
	(Delay forMilliseconds: 1) wait.! !

!QwaqParticipantUI methodsFor: 'stereo' stamp: 'sn 5/17/2011 16:42'!
stereoAnaglyph

	QActivityCapture action: #ToolsStereo details: 'Anaglyph'.
	self resetOpenGL: false.
	session cameraClass: QCameraAnaglyph.
	(self menu subItemNamed: 'Stereoscopic' translateMe) subMenu menuItems do: [ :item|
		item checked: (item label = 'Anaglyph' translateMe).
	].
! !

!QwaqParticipantUI methodsFor: 'stereo' stamp: 'sn 5/17/2011 16:43'!
stereoHardware

	QActivityCapture action: #ToolsStereo details: 'Hardware' translateMe.

	self resetOpenGL: true.
	session cameraClass: QCameraStereo.
	(self menu subItemNamed: 'Stereoscopic' translateMe) subMenu menuItems do: [ :item|
		item checked: (item label = 'Hardware' translateMe).
	].
! !

!QwaqParticipantUI methodsFor: 'stereo' stamp: 'sn 5/17/2011 16:43'!
stereoOff

	QActivityCapture action: #ToolsStereo details: 'Off' translateMe.
	self resetOpenGL: false.
	session cameraClass: TCamera.
	(self menu subItemNamed: 'Stereoscopic' translateMe) subMenu menuItems do: [ :item|
		item checked: (item label = 'Off' translateMe).
	].
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 4/17/2007 18:12'!
stereoOn
	"Answer the stereoOn of the receiver"
	<bewareOf: #stereoOnChanged>
	^self propertyValueAt: #stereoOn! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 4/17/2007 18:12'!
stereoOn: aValue
	"Modify the receiver's stereoOn"
	^self propertyValueAt: #stereoOn put: aValue with: #stereoOnChanged! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 11/24/2009 19:05'!
stopTests
	<on: #stopAllTests>

	self signal: #stopOpenALTest.
	self signal: #stopBuiltinSoundTest.
	self signal: #stopVoiceTransportTest.
! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'das 7/20/2009 09:04'!
streamingVideoDocument: filePath
	"Helper function... called from #placeStreamingVideoFileCard in the Tools menu."
	| doc sharedArea |

	QNetVidServer processForClientStreaming: filePath.

	sharedArea := 'sharedDocuments/', (self simplerLogin).

	doc := (QSharedDocument new)
					ctime: (self documentManager newTimeStamp: 'Insert');
					sha: (TContentManager generateSHA: 'foobar');
					sharedName: sharedArea, '/', (filePath copyAfterLast: FileDirectory pathNameDelimiter);
					suffix: (filePath copyAfterLast: $.);
					yourself.
	^ doc.
! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 4/14/2010 17:52'!
submitCrashDump: fullPath
	"Report a crash.dmp from a previous execution. Note: We do NOT ask the user
	here since a) the crash dump never contains any sensitive information and b)
	crash dumps are too hard to hear about otherwise"
	| report location response directory file |
	file := [FileStream readOnlyFileNamed: fullPath] on: Error do:[:ex| ex return: nil].
	file ifNil:[^nil]. "no pending crash dump"
	report := [file binary contents asString] ensure:[file close].
	"We need a valid service, so wait until user logged on"
	[serviceSvc == nil] whileTrue:[(Delay forSeconds: 10) wait].
	"And send the error"
	location := 'services/',(self currentForumName), '/errors/', TObjectID new hex, '.log'.
	response := serviceMgr errorReport: report location: location.
	response name == #ok ifTrue:[
		directory := FileDirectory forFileName: fullPath.
		directory
			rename: (FileDirectory localNameFor: fullPath)
			toBe: (directory nextNameFor: 'sent-crash' extension: 'txt').
	] ifFalse: [
		QLogger logWarning: 'Crash report submission failed: ', response contentString
	]! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 4/14/2010 17:52'!
submitErrorReport: anErrorReport dialog: errDialog

	"Report the error to Qwaq,
	 and quit the application if we're sealed,
	 or re-enable the original error dialog if unsealed."

	| location response |

	self mouseCursor: Cursor write.
	self waitTick.

	location := 'services/',(self currentForumName), '/errors/', TObjectID new hex, '.log'.
	response := serviceMgr errorReport: anErrorReport location: location.
	response name == #ok 
		ifTrue:[self inform: 'The error has been successfully reported.\Thank you for your help.' withCRs]
		ifFalse:[self inform: response contentString].

	"For regular use, once the error is submitted... just go away.
	 Saves any misunderstandings, and ensures there's no lingering connection."
	self isSealedApplication ifTrue: [
		self mouseCursor: Cursor wait.
		self exitConfirming: false.
	] ifFalse: [
		errDialog enable.
		self mouseCursor: nil.
	].! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'bgf 4/14/2010 17:52'!
submitIncidentSilently: incidentSummaryString with: anError with: aStack
	| reportString location response |
	"Handler for ui signal: #incidentReport with: msgString with: (error or nil) with: (stack or nil).
	 You would usually call 
			reportIncident: msgString
			reportIncident: msgString error: anError stack: aStack
 	 which raise the signal for me to handle here.
	 Send the given message as part of a bug report mail
	 with all the usual details, transcript, etc."

	<on: incidentReport>

	[
		"Ensure that all relevant information has time to appear on the Transcript."
		(Delay forSeconds: 1) wait.
		self waitTick; waitTick; waitTick.
		(Delay forSeconds: 1) wait.

		reportString := String streamContents: [:s | 
			s nextPutAll: '[SENT SILENTLY]'; crlf.
			self fullReportFor: incidentSummaryString error: anError stack: aStack on: s.
		].
		location := 'services/',(self currentForumName), '/errors/', TObjectID new hex, '.log'.
		response := serviceMgr errorReport: reportString location: location.
		(response name == #ok) ifFalse: [
			QLogger logError: 'Incident report transmission failed: ', response contentString from: self.
		]
	] on: Error do: [ :ex|
		QLogger logError: 'Incident report failed: ', (ex description) from: self.
	]
		

! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/9/2010 17:33'!
suspendAudio
	"Temporarily suspend sound audibility, without perturbing streaming or player relationships.
	 Cheap and idempotent."
	self  audioSession ifNotNilDo: [ :as | as future suspendAudio ].! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 4/8/2009 14:54'!
suspendVideo
	"Temporarily suspend capturing/publish video, as well as playing incoming video."
	self videoSession ifNotNilDo: [:vs | vs suspendVideo].! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'jcg 12/28/2009 12:02'!
suspendVideoBecauseOfLag
	"Suppress video in this session (regardless of the persistent preference)
	 without changing the persistence setting... 
 	 but only if video is active (so there is something to be gained by stopping it).

	 Returns true if an active video session was actually disabled."
	| active vidSession |

	"We disable if either I am running my camera, or I am receiving video."
	active := false.
	[ active := controlBar camera selected.
	  controlBar camera selected: false ] on: Error do: [ ].

	"Just a way to detect if we're fully underway.  I believe that testing
	 'session videoSession isNil' would also do, but I'm taking no risks
	 so close to the 2.0 release."
	[session videoSession webcastManager] 
		on: Error 
		do: [ "still bootstrapping" ^ false ].
	vidSession := session videoSession.

	active := active or: [vidSession hasVideoPlayers].
	active ifTrue: [
		vidSession isSuspended ifFalse: [
			QLogger logInfo: 'Suspending video session.'.
			vidSession suspendVideo.
		].
	].
	^ active

! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/16/2007 10:45'!
svcMgrCache
	"Answer the svcMgrCache of the receiver"
	<bewareOf: #svcMgrCacheChanged>
	^self propertyValueAt: #svcMgrCache! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/16/2007 10:45'!
svcMgrCache: aValue
	"Modify the receiver's svcMgrCache"
	^self propertyValueAt: #svcMgrCache put: aValue with: #svcMgrCacheChanged! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'eem 11/18/2008 16:32'!
systemProfiler: seconds

	| profiler |
	profiler := self preferredProfilerClass default.
	[ ^ profiler spyOn: [(Delay forSeconds: seconds) wait]]  ensure: [ 
		| pane report |
		report := String streamContents: [:s | profiler report: s].
		pane := QFloatingTextPane new.
		pane text: report.
		self addFloatingPane: pane. 
		pane title: seconds printString, '-Second Profile'.
		pane toggle.
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 2/12/2009 12:28'!
takeScreenShot
	| form |
	self waitTick; waitTick.	"Let the menu itself go away."
	session ifNil: [ ^ self ].

	self playSoundEffect: #iris.
	form := session screenShot.
	form ifNil: [ 		
		^ self inform: 'Could not capture the screen;\please try again.'.
	].
	[
		self disposeScreenShot: form.
	] ifError: [ :msg |
		self inform: 'There was a problem saving the capture.\\Details: ' withCRs,  
						msg, 
					('\\The screen shot was not saved.' withCRs).
	].! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 8/22/2009 18:48'!
teleportTo: target atTransform: aTransformOrNil

	| newSpace |

	"Nb. Far-ref root is one of the few things we can say directly."
	newSpace := target root.
	"Don't set up for entry zoom-in on same-space transitions, 
	 as the space-change trigger that unrolls the zoom doesn't fire."
	(newSpace ~= activeSpace) ifTrue: [
		self startEntryZoom.	
	].
	avatar newSpace: newSpace.

	"If we know where to go, go there; else let the space place us."
	aTransformOrNil ifNotNil: [
		avatar newTransform: aTransformOrNil.
	] ifNil: [
		"If the following get returns nil, 
	 	 we'll just follow the default space-entry pattern,
		 so no harm is done..."
		avatar newTransform: (target get:#teleportTransform).
	].


! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 8/26/2008 17:50'!
teleportToUrl: locationUrl
	"Teleport to the service described by the given service url"

	<on: signalTeleportToUrl>
	
	self teleportToUrl: locationUrl atTransform: nil! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'jjl 8/27/2008 09:28'!
teleportToUrl: locationUrl atTransform: aTransformOrNil

	"Teleport to the location described by the given service url"
	| service targetUrl target |

	"First check if there's an unsaved avatar."
	avatarPane onClose ifFalse: [ 
		self mouseCursor: nil.
		avatarPane visible: true; isExpanded: true.
		historyPane noteSpaceActivate: self activeSpace.
		^self].

	[
		[
			service := self activateServiceFromUrl: locationUrl.
		] on: Error do: [ :error|
			"Kick the history pane to reset its back/forward state."
			historyPane noteSpaceActivate: self activeSpace.
			QActivityCapture action: #TeleportFail details: locationUrl.
			^ self inform: (error messageText)
		]
	] on: CProgressCancel do: [ :ex|
		QActivityCapture action: #TeleportCancel details: locationUrl.
		^ self
	].

	service ifNil:[ ^self ].

	targetUrl := locationUrl.
	target := session resolvePostcardUrl: targetUrl.
	(target notNil and:[(target get: #class) includesBehavior: TFrame]) ifFalse:[
		"The target link is not to a frame - use default entry point"
		targetUrl := (TPostcard makeUrl: service entryUrl relativeTo: service url).
		target := session resolvePostcardUrl: targetUrl.
	].

	(target isFarRef and: [target isValidFarRef]) ifFalse: [
		"No teleport. Kick the history pane to reset its back/forward state;
		  otherwise the back-forward buttons would stay gray,
		  as there will be no #spaceChange to reset them."
		historyPane noteSpaceActivate: self activeSpace.
		QActivityCapture action: #TeleportIncomplete details: locationUrl.
		^self
	].

	(mainView performReturnPrep: target atTransform: aTransformOrNil) ifFalse:[
		"#performReturnPrep will finish teleport by signalling #returnPrepDone,
		 which will call #teleportTo:atTransform:"
		^self.
	].

	self teleportTo: target atTransform: aTransformOrNil.! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'hrs 5/7/2010 13:09'!
teleportToUrl: locationUrl place: placeName
	"Teleport to the service described by the given service url (if not already there) and 
	then go to place."
	<on: teleportToUrlPlace>
	self currentForumURL = locationUrl ifFalse: [
		self teleportToUrl: locationUrl atTransform: nil.
		placeName isEmptyOrNil ifFalse: [
			(Delay forMilliseconds: 500) wait.		"Long waitTick... let it go there."
		].
	] ifTrue: [
		"Above branch signals #ready on space change. This branch gives 
		generalTeleport  clients an alternative trigger."
		self signal: #alreadyPresent.
	].
	placeName isEmptyOrNil ifFalse: [
		"Don't send direct, because the various scripts must be ordered properly."
		self startScript: #goToPlaceName: withArguments: {placeName}
		].
	login forumGoal: nil. "allow  general teleports."! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'crl 5/6/2010 15:03'!
testAudioPipeline
	"Send the test signal, speex-encoded, to the media router, receive it, and ensure that the end-to-end content, maximum latency, and maximum jitter are appropriate."

	| test infoMessageDelay notifying |

	test := (
		(QSoundTest forSeconds: 5)
			controller: self session mediaController;
			yourself).

	QActivityCapture action: #TestAudioPipelineExtrasMenu.
	infoMessageDelay := Delay forSeconds: 1.

	notifying := true.
	[
		[notifying] whileTrue: [
			self signal: #infoMessage with: 'Testing end-to-end voice audio pipeline...'.
			infoMessageDelay wait]
	]
		fork.

	[
		[test testDecoding]
			on: TestFailure
			do: [:exception |
				"Ignore assertion failures."
				exception resume]
	]
		ensure: [
			notifying := false.
			test destroy]! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 11/24/2009 18:54'!
testBuiltinSound
	| sampledSound stopper dialog |
	QActivityCapture action: #TestBuiltinSoundExtrasMenu.
	sampledSound := self testSoundToPlay.
	sampledSound ifNil:[^self inform: 'Oops, test sound not found'].


	dialog := (CUserDialog new).
	stopper := self startScript: [  dialog close. sampledSound := nil ] when: {self. #stopBuiltinSoundTest }.
	self startScript:[
		dialog inform: 'Now playing test sound.\You should hear a bird chirp.\\Click to stop the test.' withCRs
			    label: 'Testing'
			    centerAt: nil.
		self signal: #stopBuiltinSoundTest.
	].

	[sampledSound == nil] whileFalse:[
		sampledSound isPlaying ifFalse:[sampledSound play].
		self waitTick.
	].

	self stopScript: stopper.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 4/16/2011 19:00'!
testClientExperience
	"Send the test signal, speex-encoded, to the media router, receive it, and ensure that the end-to-end content, maximum latency, and maximum jitter are appropriate."

	| test testSummary testDetails dialog sendDetails playOriginalTestSignal playReceivedTestSignal pasteDetails infoMessageDelay notifying |

	test := (
		(QSoundTest forSeconds: 5)
			controller: self session mediaController;
			yourself).

	QActivityCapture action: #TestAudioPipelineExtrasMenu.
	infoMessageDelay := Delay forSeconds: 1.

	notifying := true.
	[	[notifying] whileTrue: [
			self signal: #infoMessage with: 'Testing end-to-end voice audio pipeline...'.
			infoMessageDelay wait]
	] fork.

	[	[test testDecoding.
		 testSummary := test summary.
		 testDetails := test detailsFor: self]
			on: TestFailure, Error
			do: [:exception |
				"Ignore assertion failures."
				exception isResumable
					ifTrue: [exception resume]
					ifFalse: [
						QLogger logError: exception.
						^CDialog inform: 'Error occurred during client-experience test']]
	] ensure: 
		[notifying := false.
		 test destroy].

	dialog := (
		(CUserDialog new)
			label: 'Client Experience Test Results';
			message: testSummary;
			yourself).

	(dialog messageText)
		allowEditing: false;
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.

	sendDetails := (
		(dialog addButton: 'Report to OpenQwaq')
			width: 150;
			yourself).
	self
		startScript: [
			self
				mailTo: 'support@openqwaq.com'
				subject: 'client experience report'
				body: testDetails]
		when: {sendDetails. #fire}.

	pasteDetails := (
		(dialog addButton: 'Paste to Text Chat')
			width: 150;
			yourself).
	self
		startScript: [
			self activeSpace future chatLog future
				addMessage: testDetails
				fromUser: (self activeSpace future chatLog future userForAvatar: self avatar replica)
				atTime: QChatLog clientUtcTime]
		when: {pasteDetails. #fire}.

	playOriginalTestSignal := (
		(dialog addButton: 'Play Original')
			width: 150;
			yourself).
	self
		startScript: [test playOriginalSound]
		when: {playOriginalTestSignal. #fire}.

	playReceivedTestSignal := (
		(dialog addButton: 'Play Received')
			width: 150;
			yourself).
	self
		startScript: [test playDecodedSound]
		when: {playReceivedTestSignal. #fire}.

	dialog addButton: 'Dismiss' value: #cancel.

	dialog run: nil centerAt: self bounds center! !

!QwaqParticipantUI methodsFor: 'test-menu' stamp: 'rmp 12/29/2008 15:07'!
testMenu
	"Tests that are not going to be exposed all the time."
	| menu |

	menu := CMenu new.

      menu add: 'Voice Extrapolation' target: self action: #voiceTest.

	^menu
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 11/24/2009 19:07'!
testOpenALSound
	| sampledSound oal instantiatedOpenAL msg details buffer source dialog stopper |

	QActivityCapture action: #TestOpenALSoundExtrasMenu.

	sampledSound := self testSoundToPlay.
	sampledSound ifNil:[^self inform: 'Oops, test sound not found'].

	"Grep OpenAL from the mainView if present"
	instantiatedOpenAL := false.
	self audioSession ifNotNilDo: [ :as|
		oal := as openAL
	].

	"Try to create a new openAL if necessary"
	oal ifNil:[
		[
			instantiatedOpenAL := true.
			oal := OpenAL new.
			oal makeCurrent.
		] on: Error do:[:ex|
			msg := ex description.
			details := String streamContents:[:s|
				ex signalerContext errorReportOn: s.
			].
			oal := nil.
			ex return.
		].
	].

	"If still no OpenAL then report error from previous step and get outta here"
	oal ifNil:[
		^self reportError: msg details: details
			label: 'Error initializing OpenAL' button: 'OK'.
	].

	buffer := QOpenALBuffer 
		data: sampledSound samples 
		samplingRate: sampledSound originalSamplingRate 
		format: (sampledSound isStereo 
			ifTrue:[OpenALConstants classPool at: #ALFormatStereo16] 
			ifFalse:[OpenALConstants classPool at: #ALFormatMono16]).

	source := QSoundSource new.
	source referenceDistance: 5.
	source maxDistance: 500.0.
	source gain: 1.0.
	source position: 0@0@0.
	source direction: 0@0@-1.
	source looping: true.
	source buffer: buffer.

	dialog := (CUserDialog new).
	stopper := self startScript: [ dialog close . source stop . source delete . source := nil. instantiatedOpenAL ifTrue: [ oal destroy ]]
		 			when: { self . #stopOpenALTest }.
	self startScript:[
		dialog inform:'Now playing test sound.\You should hear a bird chirp.\\Click to stop the test.' withCRs
			    label: 'Testing'
			    centerAt: nil.
		self signal: #stopOpenALTest.
	].
	source play.
	source render: oal.
	[ source ] whileNotNil: [ (Delay forMilliseconds: 50) wait ].
	self stopScript: stopper.


! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'ar 2/2/2007 17:37'!
testSoundToPlay
	"Answer the sampled sound used for testing sound feedback"
	^QSounds find: 'Bird_Idle07.wav'! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 3/3/2010 15:06'!
testVoiceTransport
	self testVoiceTransportWithBlock: [
		self privateTestVoiceTransportWithDefaultSound: false.
	]
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 3/3/2010 15:06'!
testVoiceTransportAllAtOnce
	self testVoiceTransportWithBlock: [
		self privateTestVoiceTransportWithDefaultSound: true.
	]
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 12/1/2008 18:43'!
testVoiceTransportRepeatedly
	| path  |
	path := FileDirectory pathFromURI: 
						'Objects/Sounds/MicrophoneCheck/Qwaq-MicCheck-16000Hz.wav'.
	self testVoiceTransportRepeatedly: path.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 3/3/2010 15:11'!
testVoiceTransportRepeatedly: fullPathToWAV
	| stop stopper dialog |

	QActivityCapture action: #TestVoiceTransportExtrasMenu.

	stop := false.
	dialog := (CUserDialog new).
	self startScript:[
		dialog inform: ('Now sending test sound to other participants.\They should hear a voice sample.\\Click to stop the test.' withCRs)
			    label: 'Testing'
			    centerAt: nil.

		self signal: #stopVoiceTransportTest.
	].
	stopper := self startScript: [ dialog close. stop := true ] when: { self . #stopVoiceTransportTest }.

	self testVoiceTransportWithBlock: [
		[stop] whileFalse:[
			self 
				privateTestVoiceTransportWithFileNamed: fullPathToWAV 
				allAtOnce: false
		].
		self stopScript: stopper.
	].

	! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'eem 3/19/2010 17:28'!
testVoiceTransportRepeatedlyWithUserFile
	| fileName |
	
	[
		fileName := QFileDialog
					openFile: #(('WAV files (*.wav)' '*.wav')) 
					path: (QFileDialog getFileLocation: #home)
					label: 'Select 8kHz or 16kHz WAV file'.
	] on: Error do: [
		self inform: 'Could not open file chooser'.
	].

	fileName ifNotNil: [self testVoiceTransportRepeatedly: fileName]
! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'jcg 3/4/2010 18:26'!
testVoiceTransportWithBlock: aBlock
	"Common code for running several voice-transport tests.  All run in a separate script, disable the microphone before starting, do the test, and finally reset the microphone state to reflect the current can-speak-in-meeting setting. "
	audioSession ifNil: [^QLogger logWarning: 'Test Voice Transport: no audio-session'].
	audioSession voiceRelay ifNil: [^QLogger logWarning: 'Test Voice Transport: no voice-relay'].
	self startScript: [
		[	audioSession voiceRelay disableMicrophone.
			aBlock value.
		] ensure: [self updateCanSpeakInMeeting]
	].
! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'bgf 8/29/2008 16:30'!
textMessage: aString fromUser: uname atUtcSeconds: seconds
	"For Quindi logging. Note the fact that an avatar has sent text chat"

	quindiLogging ifTrue:[
		self quindiLog: Time now print24, ' Text chat from ', uname, ':\', String crlf, aString
			hiddenText: ''.
	].
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 3/20/2009 01:46'!
throttleInboundTraffic
	"Open a dialog that allows the user to specify a maximum inbound data rate."
	| rate |
	rate := CDialog 
			request: 'Enter maximum inbound traffic (in KB/s)\(zero to turn off)' withCRs
			initialAnswer: 0.
	[rate := rate asInteger * 1000]
		on: Error 
		do: [^CDialog inform: 'invalid value (must be a number)'].
	[session mediaController connection throttleInboundTraffic: rate]
		on: Error
		do: [:err | QLogger logWarning: 'cannot throttle inbound traffic: ', err description]! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 1/11/2010 12:13'!
toggleCatalog
	"From the tools menu bar..."

	self toggleCatalog: (catalogPane visible not).
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 1/11/2010 12:13'!
toggleCatalog: bflag
	"From the control bar... make the Object Catalog visible."
	self readyForEvents ifTrue: [
		catalogPane visible: bflag.
		bflag ifTrue: [
			ppToggle selected ifFalse: [
				ppToggle select.
			]]].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'hrs 6/10/2008 15:57'!
toggleDropModelThumbs

	session dropModelThumb: session dropModelThumb not.
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:29'!
toggleEditBlocked

	"Meeting control - disallow importing and editing the forum."
	self readyForEvents ifTrue: [
		meetingState setEditBlocked: ((meetingState  isEditBlocked) not).
		QActivityCapture action: #MeetingEditBlocked details: meetingState isEditBlocked asString.
	].
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'das 8/27/2008 20:11'!
toggleEmbeddedApps

	embeddedOverlay ifNil:[
		embeddedOverlay := QEmbeddedOverlay new.
		session addOverlay: embeddedOverlay.
	] ifNotNil:[
		session removeOverlay: embeddedOverlay.
		embeddedOverlay := nil.
	].! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'jcg 8/25/2008 09:03'!
toggleExperimentalFeatures
	"Switch the current menu font-size choice;
	 we support two alternatives, so this need only swap that option."
	preferences enableExperimentalFeatures: (preferences enableExperimentalFeatures not).
	preferences storeXml.

	controlBar toggleExperimentalFeatures. "let it make any adjustments"! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'hrs 2/4/2010 08:18'!
toggleForeignClientProxy
	"Switch whether or not we are acting as a proxy for a foreign (e.g., mobile) proxy."
	| port |
	mobileProxy ifNil: [
		port := CDialog request: 'Port for proxy to listen on'
				initialAnswer: '9997'.
		port isEmptyOrNil ifTrue: [^self].
		self makeMobileProxy: port asNumber.
	] ifNotNilDo: [:mp | 
		mobileProxy := nil. 
		mp destroy.
	]! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 12/5/2009 08:23'!
toggleFullScreen

	Tweak fullScreenMode: Tweak fullScreenMode not.
	self wait:0.5. "gives the system time to settle before another render"
	projectMorph window ifNotNil:[window fullScreenMode: Tweak fullScreenMode.].
	(self menu itemTagged: #fullscreen) 
		checked: (Tweak fullScreenMode).
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'ar 1/7/2009 12:13'!
toggleICMP
	"Toggle the ICMP pinger"
	icmpSocket ifNil:[
		self startScript: #icmpPingerScript.
	] ifNotNil:[
		self stopScript: #icmpPingerScript.
		icmpSocket destroy.
		icmpSocket := nil.
	].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 7/31/2007 19:04'!
toggleIDLog
	TObjectID logIDs: (TObjectID logIDs not).
	(self menu itemTagged: #logIDs) checked: (TObjectID logIDs).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 1/6/2010 17:18'!
toggleIcon: anIcon forTool:aPanel
	"Callback for visibility/checkmark tracking for control toggle icons; 
	 see my method #linkIcon:toTool: "
	anIcon selected: (aPanel visible).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 12/16/2006 20:01'!
toggleItem:aMenuItem forTool:aPanel

	"Callback for visibility/checkmark tracking for menu items; 
	 see my method #linkItem:toTool: "

	aMenuItem checked: (aPanel visible).! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'das 3/23/2009 15:57'!
toggleLongTranscript

	logViewCharacterLimit := (logViewCharacterLimit = 50000) ifTrue:[ 250000] ifFalse:[ 50000].
	(self menu itemTagged: #longTranscript) checked: (logViewCharacterLimit = 250000).! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 6/10/2008 20:27'!
toggleMaterials

	session materialsEnabled: session materialsEnabled not.
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'ar 4/16/2011 17:35'!
toggleMeetingLeadership: bWasLeader
	"Meeting menu action: become, or retire as, meeting leader."
	self readyForEvents ifTrue: [
		orgServices beMeetingLeader: (bWasLeader not).
		QActivityCapture action: #MeetingBeLeader details: bWasLeader not  asString.
	] ifFalse: [
		self statusAlert: 'You must be in an OpenQwaq session\to take up meeting leadership.'.
	].! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:23'!
toggleMeetingWhereIAm
	"Make my current place the assigned meeting place (or clear it)."

	self readyForEvents ifTrue: [
		| wasHere |
		wasHere := ((meetingState meetingPlace) = (avatar replica get: #nominalPlace)).
		wasHere ifTrue: [
			avatar replica future setMeetingPlace: nil
		] ifFalse: [
			avatar replica future setMeetingWhereIAm.
		].
		QActivityCapture action: #MeetingMeetHere details: wasHere not asString.
	].
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:28'!
toggleMuteAudience
	"Meeting control - send everyone the command to mute their audio."

	self readyForEvents ifTrue: [
		meetingState setAudienceIsMuted: ((meetingState audienceIsMuted) not).
		QActivityCapture action: #MeetingMuteAudience details: (meetingState audienceIsMuted asString).
	].
! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 1/21/2009 11:44'!
toggleOpenGL
	"Developer-menu item... turn on and off the 3D rendering."
	doOpenGL := (self doOpenGL not).

	(mainView notNil and: [ mainView ogl notNil & (doOpenGL not)]) ifTrue: [
		session killRender.
	] ifFalse: [
		(mainView notNil and: [ mainView ogl isNil & (doOpenGL) ]) ifTrue: [
		session startRender.
		]].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'das 1/5/2009 11:18'!
togglePanelLog
	QLogger doPanelLog: (QLogger doPanelLog not).
	(self menu itemTagged: #panelLog) checked: (QLogger doPanelLog).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 10/2/2007 01:37'!
toggleQMSAutoPause
	preferences doQMSAutoPause: preferences doQMSAutoPause not.
	preferences storeXml.
	! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'ar 9/26/2007 11:56'!
toggleQuindiLogging
	"Toggle quindi logging"
	quindiLogging := quindiLogging not.
	(self menu itemTagged: #quindiLogging) checked: quindiLogging.! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'bgf 8/15/2008 01:05'!
toggleReadonlyBehavior

	"Pretend to be 'guest' user, or go back to my real role."
	(self hasCapability: #modify) ifFalse: [
		capabilities := self capabilitiesForRole: (self roleInCurrentForum).
	] ifTrue: [
		capabilities := self capabilitiesForRole: 'visitor'.
	].

	self applyUserRole.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 7/28/2010 16:42'!
toggleReceiveButIgnoreVideo
	"In cases where turning on video/pshare causes performance to fall off a cliff, we want to figure out whether it's because of network-effects or not.  We do this by continuing to receive the data, but not decoding/rendering it."
	QJoshHacks receiveButIgnoreVideo: QJoshHacks receiveButIgnoreVideo not.
	(self menu itemTagged: #toggleReceiveButIgnoreVideo) ifNotNil: [:item |
		item checked: QJoshHacks receiveButIgnoreVideo.
	].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'das 4/11/2008 21:01'!
toggleReplicatedRandom

	| isle |
	isle := mainView activeIsland.
	(isle get: #isRandomTesting) ifTrue:[
		isle future stopRandomTesting.
	] ifFalse:[
		isle future startRandomTesting.
	].
	! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'bgf 5/18/2008 12:33'!
toggleRespectForTheLaw

	"Toggle respect for gravity and motion boundaries...
	 turning these off is essential for space-definers.
	 We'll need a way to restrict this only for special people in future."
	| item |

	avatar driver followRules: (avatar driver followRules not).

	item := (self menu subMenuNamed: 'Extras') itemTagged:#ignoreMotionRules.
	item checked: (avatar driver followRules not).

	QActivityCapture action: #ToggleFlyingExtrasMenu details: (item checked asString).! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:22'!
toggleRestrictPrivateText

	"Toggle the 'Restrict Private Text' menu item."
	self readyForEvents ifTrue: [
		meetingState setPrivateTextBlocked: ((meetingState isPrivateTextBlocked) not).
		QActivityCapture action: #MeetingRestrictPrivateText details: (meetingState isPrivateTextBlocked asString).
		].
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 5/11/2010 18:27'!
toggleRoundTripSound
	"For testing purposes: Create a full round-trip sound setup which takes the participant's
	voice chat and subscribes to it via a different media controller, thereby forcing a round
	trip to the router."
	roundTripSound ifNil:[
		QLogger  logInfo: 'Enabling Hear-Myself (round-trip sound)'.
		self makeRoundTripSound.
	] ifNotNil:[
		QLogger logInfo: 'Disabling Hear-Myself (round-trip sound)'.
		roundTripSound ifNotNil: [:clientSpkr |
			roundTripSound := nil.
			clientSpkr future stop.
			clientSpkr future beLoopbackSound: false.
		].
	].
	self signal: #clientSpeakersChanged.! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 6/17/2010 10:31'!
toggleSecureBroadcast
	| replica |
	replica := avatar ifNotNil: [:av | av replica].
	replica ifNil: [^QLogger logWarning: 'no avatar in #toggleSecureBroadcast'].

	"If Secure Broadcast is already enabled, we don't need to do this test."
	(replica get: #cameraEnabled) ifFalse: [
		audioSession audioInterface supportsRecording ifFalse: [
			| yesNo |
			yesNo := CDialog 
						confirm: 
							'Secure Broadcast requires OpenAL to be enabled.',
							String cr, String cr,
							'Open the Sound Settings pane to make this change?'
						yes: 'Yes'
						no: 'No'
						label: 'OpenAL is disabled'.
			yesNo ifTrue: [controlBar adjustSound].
			^self.
		].
	].

	"Toggle it!!"
	replica future toggleCamera.
	self showWebcamPane! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'das 4/10/2008 19:30'!
toggleSnapshotLoad

	snapshotBool := snapshotBool not. 
	snapshotBool ifTrue:[
		self startScript: #snapshotIslands: withArguments: {5}.
	].! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 9/13/2009 19:21'!
toggleStayPut
	| wasSP |
	"Toggle the 'Permit Movement' menu item."
	self readyForEvents ifTrue: [
		wasSP := meetingState audienceStayPut.

		QActivityCapture action: #MeetingAudienceStayPut details: (wasSP not asString).
		meetingState setAudienceStayPut: (wasSP not).
		wasSP ifFalse: [
			avatar replica future refreshLeaderPlace.
		]].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 8/11/2010 10:11'!
toolTable
	"The dictionary of tools whose shapes are preserved across logins."

	^ Dictionary newFromPairs: {
		  'chat' . chatPane .
		  'forums' . servicesPane .
		  'documents' . documentsPane .
		  'repositories' . repoBrowser.
		  'people' . usersPane .
		  'history' . historyPane .
		  'phones' . phonePane .
		  'inspector' . treeInspector .
		  'materials' . materialCatalog .
		  'network' . trafficPane .
		  'latency' . latencyPane .
		  'recorder' . recorderPane .
		  'broadcaster' . webcastPane .
		  'performance' . statisticsPane .
		  'transcript' . logView.
		  'customs' . customsPane
		}! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jcg 1/24/2008 09:48'!
toolsMenu
	| menu |
	menu := CMenu new.
	self repopulateToolsMenu: menu.
	^menu! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/30/2007 14:01'!
trafficPane
	"Answer the trafficPane of the receiver"
	<bewareOf: #trafficPaneChanged>
	^self propertyValueAt: #trafficPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 10/30/2007 14:01'!
trafficPane: aValue
	"Modify the receiver's trafficPane"
	^self propertyValueAt: #trafficPane put: aValue with: #trafficPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 9/1/2006 15:38'!
treeInspector
	"Answer the treeInspector of the receiver"
	<bewareOf: #treeInspectorChanged>
	^self propertyValueAt: #treeInspector! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 9/1/2006 15:38'!
treeInspector: aValue
	"Modify the receiver's treeInspector"
	^self propertyValueAt: #treeInspector put: aValue with: #treeInspectorChanged! !

!QwaqParticipantUI methodsFor: 'proxy' stamp: 'ar 7/12/2011 12:56'!
tryProxyServer: proxyString forHost: hostName
	"Take a proxy list (array of proxy server strings) and answer one that is reachable"
	| proxy info |
	QProxyHandler log: 'Trying proxy ', proxyString from: self class.
	self statusInfo: 'Trying proxy server ', proxyString.
	proxyString = '(no proxy)' ifFalse:[
		proxy := QProxyHandler new.
		proxy userAgent: self appName.
		proxy proxyHost: (proxyString copyUpToLast: $:).
		[proxy proxyPort: (proxyString copyAfterLast: $:) asInteger] on: Error do:[:ex|].
		proxy isValid ifFalse:[
			QProxyHandler log: proxyString, ' - proxy invalid ' from: self class.
			^nil].
		proxy isReachable ifFalse:[
			QProxyHandler log: proxyString, ' - proxy unreachable ' from: self class.
			^nil].
	].
	self proxy: proxy.
	QProxyHandler log: proxyString, ' - Connecting to ', hostName from: self class.
	info := [self connectionInfoFor: hostName] on: Error do:[:ex| 
		QProxyHandler log: proxyString, ' - ', ex from: self class.
		QProxyHandler logRequest: (ex signalerContext stackOfSize: 100).
		QLogger logWarning: ' Proxy error from ', proxyString, ' - ', (ex description).
		QLogger logError: ex  from: self.
		ex return: nil].
	info ifNotNil:[QProxyHandler log: proxyString, ' - Success for ', hostName from: self class].
	^info
! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 2/2/2007 17:24'!
uiClass
	"This is for subclasses to override
	 if they want to pretend to be a different uiClass
	 for purposes of matching loadable services."

	^ self class name! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/4/2008 11:47'!
uiLatencyStats
	"Answer the uiLatencyStats of the receiver"
	<bewareOf: #uiLatencyStatsChanged>
	^self propertyValueAt: #uiLatencyStats! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 12/4/2008 11:47'!
uiLatencyStats: aValue
	"Modify the receiver's uiLatencyStats"
	^self propertyValueAt: #uiLatencyStats put: aValue with: #uiLatencyStatsChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jjl 3/4/2008 10:14'!
uiStyle
	"Answer the uiStyle of the receiver"
	<bewareOf: #uiStyleChanged>
	^self propertyValueAt: #uiStyle! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jjl 3/4/2008 10:14'!
uiStyle: aValue
	"Modify the receiver's uiStyle"
	^self propertyValueAt: #uiStyle put: aValue with: #uiStyleChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/18/2010 10:51'!
undoRedoList
	"Answer the undoRedoList of the receiver"
	<bewareOf: #undoRedoListChanged>
	^self propertyValueAt: #undoRedoList! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 3/18/2010 10:51'!
undoRedoList: aValue
	"Modify the receiver's undoRedoList"
	^self propertyValueAt: #undoRedoList put: aValue with: #undoRedoListChanged! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 5/23/2007 09:33'!
update: aSymbol
	"Update from Transcript"
	[
		aSymbol == #appendEntry ifTrue:[self signal: #appendEntry with: Transcript contents].
	] on: Error do: [
		"Get it later."
	].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:18'!
updateActionsMenu
	"Query the space for the currently available user Actions
	 and update the menu to present them."
	|  actions menu item |
	menu := self menu subMenuNamed: 'Actions' translateMe.
	menu ifNil: [ ^ self ].
	menu menuItems removeAll.

	self isReadonlyUser ifTrue: [
		"For readonly visitors, omit all the space-defined actions."
		self addFixedItemsToActionsMenu: menu.
		QStyles installMenuFontInMenu: menu.
		^ self.
	].
	(menu add: 'Add Landmark Here...' translateMe target: self action: #newPlaceMarker)
		enabled: (self canModifyForum);
		separator: true.

	self canModifyForum ifTrue: [
		actions := (activeSpace get:#actionVerbs) ifNil: [#()].
		actions do: [:a|
			(a first == #separator) ifTrue: [
			 	(menu menuItems size > 0) ifTrue: [
					menu menuItems last separator: true.
				]
			] ifFalse: [
				item :=menu add: (a second) 
						   target: self 
						   action: #performActionInSpace: 
						   argument: (a first).
				item enabled:(a first ~= #noSuchAction).
			]].
		self hasOfflineUploadSupport ifTrue: [
			menu add: 'Check for Offline Uploads' translateMe target: customsPane 
				action: #processOfflineUploadsWithNotice: arguments: {true}.
		].
	].

	"See up top also..."
	menu menuItems last separator: true.
	self addFixedItemsToActionsMenu: menu.

	QStyles installMenuFontInMenu: menu.
! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'hrs 5/26/2010 12:38'!
updateCanSpeakInMeeting
	"Update whether we're currently allowed to speak in this meeting."
	| canSpeak |
	audioSession ifNotNilDo: [:sess |
		canSpeak := self canSpeakInMeeting.
		sess voiceRelay isMicrophoneEnabled: canSpeak.
		mobileProxy ifNotNil: [mobileProxy canSpeakInMeeting: canSpeak].
	].! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 12/20/2009 14:11'!
updateCostume: aCostume
	"Update the damage rectangles for the given costume"
	| box formCache material rectList bb |
	aCostume visible ifFalse:[^self].
	box := aCostume drawingBounds.
	box := box origin asIntegerPoint corner: box corner asIntegerPoint.
	formCache := aCostume localCache.
	(formCache isKindOf: QTextureFormCache) 
		ifFalse:[aCostume localCache: (formCache := QTextureFormCache new)].

	"figure out changes in size/depth"
	(formCache form == nil 
		or:[formCache form extent isPowerOfTwo not
			or:[formCache form extent asLargerPowerOfTwo ~= box extent asLargerPowerOfTwo]]) ifTrue:[
		formCache form: (Form extent: box extent asLargerPowerOfTwo depth: 32).
	].
	"repair damage"
	rectList := formCache 
		repairDamage: aCostume drawingBounds 
		using:[:cc| aCostume localFullDrawOn: cc].

	"grab the texture cache"
	material := formCache material.
	"make sure it hasn't changed"
	(material == nil or:[material textureForm ~~ formCache form]) ifTrue:[
		"Try to prevent async changes (?) from leading to MNU UndefinedObject extent:   
	 	 apparently from loss/destruction of costume localCache.
		 The error appears to arise during device-change-induced sound-resets,
		 but the transcripts I've seen give little else to go on."
		formCache = (aCostume localCache) ifFalse: [ 
			QLogger logWarning: 'Costume cache lost during update.' every: 60000.
			^ self ].

		formCache form ifNotNilDo: [ : f | 
			material := TMaterial new.
			material useForm: f mipmap: false  shrinkFit: true.
			material texture invalidate: f boundingBox.
			formCache material: material.
		] ifNil: [ 
			QLogger logWarning: 'Costume cache form lost during update.' every: 60000.
			^ self 
		]].

	"fix up alpha and remember upload rectangle"
	bb := BitBlt toForm: material textureForm.
	bb combinationRule: 40 "fixAlpha:with:".
	bb halftoneForm: (Bitmap with: 16rFF000000).
	rectList do:[:r|
		| aRect |
		aRect := r translateBy: box origin negated.
		bb destRect: aRect.
		bb copyBits.
		material texture invalidate: aRect.
	].! !

!QwaqParticipantUI methodsFor: 'menu-developer' stamp: 'jcg 2/27/2010 15:12'!
updateDeveloperMenu

	"Pre-show for the magic dev menu."

	(self menu itemTagged: #logIDs) checked: (TObjectID logIDs).
	(self menu itemTagged: #panelLog) checked: (QLogger doPanelLog).

	(self menu itemTagged: #noDemon) checked: (myDemon isNil).
	QDemon modes do: [ :sym |
		(self menu itemTagged: sym) checked: false.
	].
	myDemon ifNotNil: [
		myDemon mode ifNil: [
			(self menu itemTagged: #idle) checked: true.
		] ifNotNil: [
			(self menu itemTagged: (myDemon mode)) checked: true.
		].
	].

	(self menu itemTagged: #devGuestToggle) checked: (self isReadonlyUser).
	(self menu itemTagged: #zoomApps) checked: embeddedOverlay notNil.
	(self menu itemTagged: #thumbToggle) checked: session dropModelThumb.
	(self menu itemTagged: #foreignClientProxy) checked: mobileProxy notNil.
	(self menu itemTagged: #materialsEnabled) checked: session materialsEnabled not.
	(self menu itemTagged: #qmsAutoPause) checked: preferences doQMSAutoPause.
	(self menu itemTagged: #snapshotToggle) checked: snapshotBool.
	(self menu itemTagged: #replicatedRandom) 
		checked: ((session activeIsland get:#isRandomTesting) ifNil: [ false ]).
	(self menu itemTagged: #jitterBufferLatencyProfile) checked: 
		((Smalltalk at: #QAudioSinkEventProfiler) isGlobalProfilingEnabled).
	(self menu itemTagged: #icmpPing) checked: icmpSocket notNil.

	"Only offer choice of which profiler to use if more than one is loaded ;-)
	 (also see #developerMenu)."
	(Smalltalk includesKey: #QVMProfiler) ifTrue: [
		#(QSystemProfiler QVMProfiler) do: [:profiler |
			(self menu itemTagged: profiler) checked: ProfilerPreference == profiler
		]
	].
	
	(self menu itemTagged: #toggleOpenGL) checked: (self doOpenGL).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:14'!
updateEditMenu
	| menu hasSelection selSummary suffix hasPastes canModify |
	menu := self menu subMenuNamed: 'Edit' translateMe.
	menu ifNil: [ ^ self ].
	menu menuItems removeAll.

	selSummary := (avatar ifNotNil: [ avatar selectionSummary ]).
	hasSelection := (selSummary notNil).
	canModify := self canModifyForum.
	hasPastes := (avatar ifNotNil: [ avatar hasPastedFrames ] ifNil: [ false ]).
	suffix := selSummary ifNil: [''] ifNotNil:  [ ' ', selSummary ].

	(menu add: (undoRedoList undoLabel)  target: self action: #editUndo:)
		tag: #editUndo;
		enabled: (canModify and: [ undoRedoList canUndo ] ).

	(menu add:  (undoRedoList redoLabel) target: self action: #editRedo: )
		tag: #editRedo;
		enabled: (canModify and: [ undoRedoList canRedo ] ).

	menu addSeparator.

	(menu add: 'Cut' translateMe target: self action: #editCut:) 
		tag: #editCut;
		enabled: (hasSelection & canModify ) .

	(menu add: 'Copy' translateMe target: self action: #editCopy:) 
		tag: #editCopy;
		enabled: (hasSelection) .

	(menu add: 'Paste' translateMe target: self action: #editPaste:)
		tag: #editPaste;
		enabled: (canModify & hasPastes).

	menu addSeparator.

	(menu add: 'Delete' translateMe, suffix  target: self action: #editDelete:)
		tag: #editDelete;
		enabled: (canModify & hasSelection).

	hasSelection ifTrue: [
		menu addSeparator.
		(menu add: 'Locate' translateMe, suffix target: self action: #editLocate:)
			tag: #editLocate;
			enabled: (hasSelection).
		].! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:14'!
updateExtrasMenu
	"Preflight firing of the xtras menu."
	| menu deItem lmItem |
	menu := self menu subMenuNamed: 'Extras' translateMe.
	menu ifNil: [ ^ self ].
	
	"Put on-the-fly updates to the extras menu here:"
	(menu itemTagged: #extrasForumDetails) enabled: (self canModifyForum).
	(menu itemTagged: #extrasEntryPosition) 
		enabled: (self canModifyForum and: [ (self activeSpace get: #isPlacementLocked) == false ]).

	(menu itemTagged: #extrasChangePassword) enabled: (self isReusableGuestUser not).

	(menu itemTagged: #extrasTestMenu) 
		enabled: (meetingState isEditBlocked not or: [ self isMeetingLeader] ).
		"Not canmodifyforum above, so that visitors can test."

	"Miramar theme command thingy only shown to DE-enabled."
	deItem := menu itemTagged: #extrasWriteTheme.
	lmItem := menu itemTagged: #extrasEntryPosition.
	(deItem notNil and: [ self hasMiramarDesktopEdition not]) ifTrue: [
		menu menuItems remove: deItem.
		lmItem separator: true.
	] ifFalse: [
		(self hasMiramarDesktopEdition and: [ deItem isNil ]) ifTrue: [
			deItem := (menu add: 'Write Theme...' translateMe target: self action: #writeThemeFile).
			deItem tag: #extrasWriteTheme.
			menu menuItems remove: deItem.
			deItem := menu menuItems add: deItem after: lmItem.
			deItem separator: true.
			lmItem separator: false.
		]].
	
	menu signal: #menuItemsChanged.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 13:59'!
updateFileMenu
	| menu sub |
	menu := self menu subMenuNamed: 'File' translateMe.
	menu ifNil: [ ^ self ].
	menu menuItems removeAll.

	(menu add: 'Import...' translateMe target: self action: #fileImport) 
		tag: #fileImport;
		enabled: (self canModifyForum).
	(menu add: 'Web Link...' translateMe target: self action: #placeWebFileCard) 
		tag: #webCard;
		enabled: (self canModifyForum).
	menu addSeparator.

	"Personal Share. (Assumed not available unless the user gets permission in applyUserRole.)"
	sub := CMenu new.
	self runScript: #updateShareMenu when: {sub. #open}.
	(menu add: 'Share a Window' translateMe subMenu: sub) tag: #shareMenu.

	self hasPersonalShareVNC ifTrue: [
		(menu add: 'Connect to VNC' translateMe target: self action: #personalShareVNC) 
			tag: #vncShare;
			enabled: self canModifyForum.
	].
	menu addSeparator.
	(menu add: 'Save Forum' translateMe target: self action: #saveForum) 
		tag: #fileSave;
		enabled: (self canModifyForum).

	(menu add: 'Save Copy As...' translateMe target: self action: #saveForumAs) 
		tag: #fileSaveAs;
		enabled: (self isAdministrator).

	(menu add: 'Save As Template...' translateMe target: self action: #saveForumAsTemplate) 
		tag: #fileSaveAsTemplate;
		enabled: (self isAdministrator).

	menu addSeparator.
	(menu add: 'Save All Forums' translateMe target: self action: #saveAllForums) 
		tag: #fileSaveAll;
		enabled: (self canModifyForum).

	menu addSeparator.
	menu add: 'Go to Lobby' translateMe target: self action: #logoff.
	menu add: 'Exit' translateMe target: self action: #onExit.
	^menu! !

!QwaqParticipantUI methodsFor: 'objects' stamp: 'ar 4/16/2011 17:35'!
updateForumsMenu

	"Load the forums menu with top part of the forums list..."

	| menu |

	menu := self menu subMenuNamed: 'Forums'.
	menu ifNil: [ ^ self ].
	servicesPane forumMenuInto: menu.
! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:09'!
updateHelpMenu
	| menu |

	menu := self menu subMenuNamed: 'Help' translateMe.
	menu ifNil: [ ^ self ].
	menu menuItems removeAll.

	menu add: 'User Guide...' translateMe target: self action: #openUserGuide.
	menu addSeparator.
	(menu add: 'Forum Welcome...' translateMe target: self action: #showForumWelcome) 
		enabled: (activeSpace isValidFarRef and: [ (activeSpace get: #welcomeDocument) notNil ]).

	menu addSeparator.
	menu add: 'Moving With the Keyboard...' translateMe target: self action: #showQuickReference:url:
		 	arguments: #('qhelp_navkeys.jpg' 'navigation.html').
	menu add: 'Moving With the Mouse...' translateMe target: self action: #showQuickReference:url:
			arguments: #('qhelp_navmouse.jpg' 'navigation.html').
	menu add: 'Importing Content...' translateMe target: self action: #showQuickReference:url:
			arguments: #('qhelp_import_content.jpg' 'importing_content.html').
	menu add: 'Traveling By Portal...' translateMe target: self action: #showQuickReference:url: 
			arguments: #('qhelp_portals.jpg' 'portals_and_spaces.html').
	menu add: 'Using Sound...' translateMe target: self action: #showQuickReference:url:
			arguments: #('qhelp_sound.jpg' 'communication.html').

	"no valid content any more
	menu addSeparator.
	menu add: 'Introductory Movies...' target: self action: #gotoHelpMovieIndexPage.
	menu add: 'Getting Started With Voice...' target: self action: #gotoSoundHelpMoofie.
	"

	(serviceMgr notNil 
		and: [ serviceMgr currentOrg notNil] 
		and: [ serviceMgr currentOrg isFreeTrial]
		and: [ serviceMgr currentOrg contact = self userLogin]
	) ifTrue: [
		menu addSeparator.
		menu add: 'Buy Now...' translateMe target: login action: #goToBuyNow: argument: serviceMgr currentOrg.
	].
	menu addSeparator.
	menu add: 'Feedback...' translateMe target: self action: #showFeedbackDialog.
	menu add: 'About...' translateMe target: self action: #showAboutBox.
	^menu! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'sn 12/27/2012 19:46'!
updateMeetingMenu
	| menu bringing topLabel canChangeLeader leader here |
	menu := self menu subMenuNamed: 'Meeting' translateMe.
	menu ifNil: [ ^ self ].
	menu menuItems removeAll.
	self readyForEvents ifFalse: [
		menu add: 'System Not Ready' translateMe  target: self action: #noSuchThing.
		^ menu
	].

	here := (avatar replica get: #nominalPlace).	
	bringing := (meetingState meetingPlace ifNotNilDo: [ : place | place isFollowMePlace ] ifNil:[false]).
	leader := self meetingLeader.

	"Leader is changeable by any normal member unless 'Permit Changes' is off.
	 Also changeable if the leader is away, or idle (no action or sound for > 1 minute)"
	canChangeLeader := 
		leader ifNil: [ self isReadonlyUser not ]
			    ifNotNil: [ 
		      		isMeetingLeader
					or: [ self isReadonlyUser not and: [ leader isAway or: [ leader isIdle ] ] ]
					or: [ self canModifyForum ]].
	topLabel := canChangeLeader
					ifTrue: [ (leader isNil or: [ leader isMe ])
								ifTrue: [ 'Lead Meeting' translateMe  ] 
								ifFalse: [ 'Take Over From ' translateMe , (leader publicName)] ]
					ifFalse: [ leader 
								ifNil: [ 'Lead Meeting' translateMe  ]
								ifNotNil: [ 'Led by ' translateMe , (leader publicName) ] ].
	(menu add: topLabel target: self action: #toggleMeetingLeadership: arguments: {isMeetingLeader} )
			checked: isMeetingLeader;	
			tag: #leadTheMeeting;
			separator: true;
			enabled: canChangeLeader.

	(menu add: 'All Follow Me'translateMe  
			target: self
			action: #bringEveryoneWithMe:
			argument: (bringing not))
			enabled: isMeetingLeader;
			checked: bringing.

	(menu add: 'Meet Here' translateMe 
			target: self 
			action: #toggleMeetingWhereIAm)
			enabled: (here notNil 
						and: [ here isStatic not ] and: [ here posture ~= #Sitting ] 
						and: [ here isFollowMePlace not or: [ isMeetingLeader ] ]
						and: [ canChangeLeader ]);
			checked: (here notNil and:[here = (meetingState meetingPlace)]);
			separator: true.

	(menu add: 'Tally a Question...' translateMe  
			target: self
			action: #askEveryoneAQuestion:)
			enabled: (isMeetingLeader or: [ leader isNil and: [ self canModifyForum ] ]);
			separator: true.

	(menu add: 'Restrict Changes' translateMe 
			target: self
			action: #toggleEditBlocked)
			enabled: isMeetingLeader;
			checked: (meetingState isEditBlocked) .

	(menu add: 'Restrict Movement' translateMe 
			target: self
			action: #toggleStayPut)
			enabled: (isMeetingLeader and: [ bringing not ]);
			checked: (meetingState audienceStayPut) | bringing.

	(menu add: 'Restrict Private Text' translateMe 
		target: self
		action: #toggleRestrictPrivateText)
		enabled: isMeetingLeader;
		checked: (meetingState isPrivateTextBlocked).
	(menu add: 'Restrict Voice Chat' translateMe 
			target: self 
			action: #toggleMuteAudience)
			enabled: isMeetingLeader;
			checked: (meetingState audienceIsMuted).
	menu menuItems last separator: true.
	(menu add: 'All Reset Sound' translateMe 
		    target: self
		    action: #broadcastResetSound)
		    enabled: (isMeetingLeader or: [ leader isNil and: [self canModifyForum]]).
	(menu add: 'All Test Experience' translateMe  
		    target: self
		    action: #broadcastTestClientExperience)
		    enabled: (isMeetingLeader or: [ leader isNil and: [self canModifyForum]])		! !

!QwaqParticipantUI methodsFor: 'meetings' stamp: 'bgf 7/10/2009 08:53'!
updateMobility
	"Update UI to reflect changes on whether or not I can move."
	| canMove |
	canMove := self allowedToMove.
	controlBar sitStand enabled: canMove.
	navPanel visible: (self menu visible and: [canMove] and: [ preferences doNavPanel ]).

	! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'bgf 5/6/2010 16:13'!
updateNametagMenu: menu
	"Match the name tag choices menu to the Preferences choices and value"
	menu menuItems removeAll.

	"In case we haven't set up the lazy prefs-panel bindings yet..."
	prefsPanel costume.
	serviceChooser ifNotNil: [ 
		prefsPanel organizations: (serviceChooser orgChoices contents).
		prefsPanel use: preferences.
	].
	prefsPanel showNamesMenu contents do: [ : item |
		(menu add: item target: prefsPanel action: #setShowNamesChoice: argument: item)
			checked: (item = prefsPanel showNamesMenu value)
		].
! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 6/19/2009 14:18'!
updatePassword: newPass forUser: aUserName

	"Followup to a password change.
	 If the password was one I'm using, 
	 update it so my session can continue."

	(serviceMgr user = aUserName) ifTrue: [
		serviceMgr password: newPass.
	].
	(self userLogin = aUserName) ifTrue: [
		login password value: newPass value.
		login saveDefaultUser.
	].
	self inform: 'Your password has been changed.'.! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'bgf 11/23/2007 15:14'!
updatePostcard: pc inFrame: aFrame toService: svc

	"Ensure the postcard's URL refers to the given service.
	 Called to update template URL references after instantiation,
	 so the in-space links refer to the new service instance."

	| replacement original oldPrefix |

	original := pc urlString.
	original ifNil: [ ^ nil ].
	oldPrefix := self serviceUrlForUrl: original.
	oldPrefix ifNil: [ ^ nil ].

	replacement := svc url, (pc urlString copyFrom:(oldPrefix size +1) to: pc urlString size).
	(replacement ~= pc urlString) ifTrue: [
		pc urlString: replacement.
		aFrame future updatePostcard:pc originalUrl: original.
	].
	^ replacement! !

!QwaqParticipantUI methodsFor: 'events' stamp: 'jcg 5/21/2010 16:21'!
updateRecordingUsers
	"Update my list of recording+webcasting users after a space transition"
	| allUsers |
	<on: updateRecordingUsers>

	"Ensure consistency of our own avatar's state."
	self avatarReplica ifNotNil: [:rep |
		| isCapturing |
		isCapturing := recorderPane isRecording.
		(rep get: #isRecordingSession) = isCapturing 
			ifFalse: [rep future isRecordingSession: isCapturing].
		isCapturing := webcastPane isRecording.
		(rep get: #isBroadcastingSession) = isCapturing 
			ifFalse: [rep future isBroadcastingSession: isCapturing].
	].

	allUsers := activeSpace get: #allUsers.
	recordingUsers := (allUsers 
		select:[:atar| atar get: #isRecordingSession]
		thenCollect:[:atar| atar get: #publicName]) asSet.
	broadcastingUsers := (allUsers 
		select:[:atar| atar get: #isBroadcastingSession]
		thenCollect:[:atar| atar get: #publicName]) asSet.

	recordingIndicator visible: (recordingUsers notEmpty).
	webcastIndicator visible: (broadcastingUsers notEmpty).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:06'!
updateShareMenu
	"Query the space for the currently available user actions
	 and update the menu to present them."
	|  menu enabled list |
	menu := (self menu itemTagged: #shareMenu) subMenu.
	menu ifNil: [ ^ self ].
	menu menuItems removeAll.
	enabled := self canModifyForum and:[self hasMiramarPersonalShare].
	(menu add: 'Open a File...' translateMe target: self action: #shareFile)
		tag: #shareFile;
		enabled: enabled.
	menu addSeparator.
	"Sort list alphabetically"
	list := (QPersonalShare sharableWindows collect:[:wid|
		(QPersonalShare windowLabelFor: wid) -> wid.
	]) sort.
	list size > 20 ifTrue:[list := list copyFrom: 1 to: 20].
	list do:[:assoc| | lbl |
		lbl := assoc key.
		(menu add: (lbl truncateWithElipsisTo: 50)
			target: self 
			action: #shareWindow:
			argument: assoc value)
				icon: (QPersonalShare windowIconFor: assoc value);
				tag: #shareWindow;
				enabled: enabled].
	menu addSeparator.
	(menu add: 'All Sharable Windows...' translateMe target: self action: #showWindowSharePane) enabled: enabled.! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 11/14/2009 21:53'!
updateStatusRelay

	"Connect the status line event handlers to the active space and my avatar replica.

	These handlers detect certain events raised on the avatar replica,
	and on the active space, and show their content as status line messages.

	We support two message paths from the replicated space.
	   Message to all participants via  space signal: #statusMessage with: aString
	   Message to one participant via  avatar signal: #statusMessage with: aString.
	(Other harnesses can detect per-avatar messages.
	 the point isn't privacy, but rather to reduce message noise).

	There are three grades of message corresponding to event names:
		#statusMessage 	- for background status, like texture-to-load count.
		#infoMessage 	- for context information, or confirmation of actions.
		#alertMessage 	- attention-getting, to explain unexpected outcomes. 

	There is separately 
		#floatMessage	- for info floated above the current pointer location
		#cornerMessage	- for to-the-side floating transients.
	"

	self stopScript: #avatarMenuPosted:itemData:.
	self stopScript: #floatMessage:.
	self stopScript: #statusFromSpace:.
	self stopScript: #infoFromSpace:.
	self stopScript: #alertFromSpace:.

	self stopScript: #statusFromAvatar:.
	self stopScript: #infoFromAvatar:.
	self stopScript: #alertFromAvatar:.
	self stopScript: #floatFromAvatar:.
	self stopScript: #cornerMessageFromAvatar:.
	
	avatar ifNotNil: [
		self startScript: #avatarMenuPosted:itemData:  when: {avatar . #menuPosted:itemData: }.
		self startScript: #statusFromAvatar: when: {avatar . #statusMessage}.
		self startScript: #infoFromAvatar: 	when: {avatar . #infoMessage}.
		self startScript: #alertFromAvatar:  	when: {avatar . #alertMessage }.
		self startScript: #floatFromAvatar:	when: {avatar . #floatMessage }.
		self startScript: #cornerMessageFromAvatar: when: { avatar. #cornerMessage }.
	].

	activeSpace ifNotNil: [
		self startScript: #statusFromSpace: when: {activeSpace . #statusMessage}.
		self startScript: #infoFromSpace: when: {activeSpace . #infoMessage}.
		self startScript: #alertFromSpace: when: {activeSpace . #alertMessage }.
	].! !

!QwaqParticipantUI methodsFor: 'demon' stamp: 'bgf 3/25/2010 13:32'!
updateTitle
	"Refresh my window title with the most specific info about where I am."
	Smalltalk windowLabel: (self appServiceLabelString).! !

!QwaqParticipantUI methodsFor: 'render' stamp: 'bgf 3/16/2009 22:05'!
updateTweak
	"Render (update) the tweak portion of the UI for OpenGL-based display."
	[
		self asPrimCostume do:[:aCostume|
			self updateCostume: aCostume.
			"This used to (before Feb-2009) have a special-case for the viewPanel-costume,
 			 but that panel is now a proper player and the extra step seems unneeded."
		].
		"Also update the hands if any of them are software rendering"
		Processor activeProject ifNotNilDo:[:proj|
			proj hands do:[:h|
				h needsSoftwareCursor
					ifTrue:[self updateCostume: h asPrimCostume]
					ifFalse:[h asPrimCostume localCache: nil]]].
	] on: Error do: [ :ex|
		session ifNotNil: [ session killRender ].	"If the problem is in 2d, just kill render, even if in dev mode."
		ex pass
	].
! !

!QwaqParticipantUI methodsFor: 'roles&capabilities' stamp: 'bgf 6/23/2008 20:45'!
updateUserRole

	"Get the current role from the service infrastructure,
	 update my capabilities set, and refresh the UI availabilities."

	self orgSetupMutex critical: [
		"The mutex keeps us from doing this in parallel during login / space-entry script-chaining."
		self capabilities: (self capabilitiesForRole: self roleInCurrentForum).
		self preloadFeatureEntitlements.
		self applyUserRole.
	].! !

!QwaqParticipantUI methodsFor: 'preferences' stamp: 'sn 5/13/2011 14:55'!
updateViewMenu
	"Match the checkmark in the viewpoint menu to the current state of the avatar camera."
	| menu minDelta minItem |

	(mainView notNil and: [avatar notNil]) ifFalse: [ ^ self ].

	menu := (self menu itemNamed: 'View' translateMe) subMenu.

	"These choices are the first n entries in the menu."
	minDelta := Float infinity.
	minItem := nil.
	menu menuItems asArray
	from: 1 to: (QUserPreferences viewpointTable size) do: [ :item|
		| di |
		item checked: false.
		di := (avatar currentThirdPerson - 
				(QUserPreferences viewpointTable at: (item label) ifAbsent:[0])) abs.
		di < minDelta ifTrue: [
			minDelta := di.
			minItem := item.
		].
	].
	minDelta < 0.5 ifTrue: [
		minItem ifNotNil: [ minItem checked: true ].
	].

	self updateNametagMenu: (menu itemNamed: 'Name Tags' translateMe) subMenu.
	(self menu itemTagged: #fullscreen) 
		checked: (Tweak fullScreenMode).
! !

!QwaqParticipantUI methodsFor: 'sound' stamp: 'jcg 5/19/2010 18:40'!
updateVoiceRelay
	"Reconnect the voice relay to the avatar replica's speaker."
	QLogger logInfo: 'about to update voice-relay'.

	"Abandon any retries-in-progress from earlier updates."
	self stopScript: #retryUpdateVoiceRelay.
	self startScript: #retryUpdateVoiceRelay when: {self . #retryUpdateVoiceRelay}.
	self privateUpdateVoiceRelay.
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 2/14/2008 09:57'!
uploadEnd: aController
	"Notify the receiver that the given controller just finished to upload a snapshot.
	We only signal an event here as the method is called from a different thread."
	^self uploadEnd: aController for: nil! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 2/14/2008 09:57'!
uploadEnd: aController for: clientName
	"Notify the receiver that the given controller just finished to upload a snapshot.
	We only signal an event here as the method is called from a different thread."

	flowControlFlag := false.
	self signal: #uploadEnd.
! !

!QwaqParticipantUI methodsFor: 'pshare-utils' stamp: 'bgf 3/14/2009 01:20'!
uploadPShareDocument: document ifNotModifiedSince: modificationTime deleteWhenDone: aBool
	"Utility method. Upload a personal share document if it has not been modified since the given modification time. If deleteWhenDone: is true, delete the document after the upload is complete."
	| localName aFile fTime |
	<on: uploadPShareDocument>
	self ifTweak:[
		localName := FileDirectory pathFromURI: (document get: #asLocalFileUri).
		(FileDirectory default fileExists: localName) ifFalse:[^nil].
		fTime := (FileDirectory default directoryEntryFor: localName) modificationTime.
		fTime = modificationTime ifFalse:[
			aFile := FileStream readOnlyFileNamed: localName.
			[ 
				self documentManager uploadDocumentFile: aFile collisions: #clobber. 
			] ensure: [ aFile close ].
		].
		aBool ifTrue:[FileDirectory default deleteFileNamed: localName ifAbsent:[]].
	] ifNot:[
		self signal: #uploadPShareDocument with: document with: modificationTime with: aBool.
	].! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 2/14/2008 09:57'!
uploadProgress: amount
	"Notify the receiver that the given controller made progress in uploading a snapshot.
	We only signal an event here as the method is called from a different thread."
	^self uploadProgress: amount for: nil
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 2/14/2008 09:56'!
uploadProgress: amount for: clientName
	"Notify the receiver that the given controller made progress in uploading a snapshot.
	We only signal an event here as the method is called from a different thread."

	amount < 1 ifTrue: [ flowControlFlag := true].
	self signal: #uploadProgress with: amount with: clientName.
! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'ar 2/14/2008 09:56'!
uploadStart: aController
	"Notify the receiver that the given controller just started to upload a snapshot.
	We only signal an event here as the method is called from a different thread."
	^self uploadStart: aController for: nil.! !

!QwaqParticipantUI methodsFor: 'notifier' stamp: 'bgf 2/10/2010 15:56'!
uploadStart: aController for: clientName
	"Notify the receiver that the given controller just started to upload a snapshot.
	We only signal an event here as the method is called from a different thread."
	self signal: #uploadStart with: clientName.

	"Slow down the drive/render cycle to reduce message flow,
	 which competes for bandwidth with the upload."
	flowControlFlag := true.

	"Give  a little chance for feedback to appear before freezing for the sync.."
	(Delay forMilliseconds: 100) wait.


! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/16/2007 14:22'!
urlCache
	"Answer the urlCache of the receiver"
	<bewareOf: #urlCacheChanged>
	^self propertyValueAt: #urlCache! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/16/2007 14:22'!
urlCache: aValue
	"Modify the receiver's urlCache"
	^self propertyValueAt: #urlCache put: aValue with: #urlCacheChanged! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 2/27/2007 21:35'!
urlCacheAt: urlString ifAbsent: aBlock
	"Look up the entity with the given unique urlString and answer if there.
	Otherwise evaluate aBlock and store it under the given urlString which is assumed
	to be along the lines of //host:port/realm/yourid."
	^urlCache at: urlString ifAbsent: aBlock! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 2/16/2007 14:23'!
urlCacheAt: urlString ifAbsentPut: aBlock
	"Look up the entity with the given unique urlString and answer if there.
	Otherwise evaluate aBlock and store it under the given urlString which is assumed
	to be along the lines of //host:port/realm/yourid."
	^urlCache at: urlString ifAbsentPut: aBlock! !

!QwaqParticipantUI methodsFor: 'service linking' stamp: 'ar 2/19/2007 11:43'!
urlCacheAt: urlString put: anObject
	^urlCache at: urlString put: anObject! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 1/22/2009 12:21'!
useNormalTimeouts
	"During logout, for non-essential record-putbacks etc,
	 we use a much shorter default timeout to prevent log hangs on connection-losses."
	serviceMgr ifNotNil: [ serviceMgr resetTimeout ].! !

!QwaqParticipantUI methodsFor: 'logon-off' stamp: 'bgf 1/22/2009 12:20'!
useQuickTimeouts
	"During logout, for non-essential record-putbacks etc,
	 we use a much shorter default timeout to prevent log hangs on connection-losses."
	serviceMgr ifNotNil: [ serviceMgr standardTimeout: 5 ].! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 11/14/2009 20:57'!
userAway

	"When Away, we inform the world through a notice (handled by the Users tool)
	 and we also disable voice chat and video... so we're not eavesdropping.
	 We also place the awayOverlay over the main view panel, to intercept all events."

	"Don't add me twice, no matter what."
	awayOverlay container = self ifFalse: [ self add: awayOverlay ].

	self suspendAudio.
	self suspendVideo.
	self statusInfo: nil.
	self isMeetingLeader ifTrue: [ self leaderGoingAway ].
	mainView ifNotNil: [
		self statusInfo: 'Away. Voice chat suspended. Click to return.'.
		QActivityCapture action: #Away.
	].
	avatar setActivities: #(Away).
	self signal: #userAway.

	"Since idle resource-refresh happens only after 60-second idle,
	 we make the Away button into a trick for forcing a refresh,
	 to help diagnosis any further resource-loading problems."
	self onUserIsIdle.

! !

!QwaqParticipantUI methodsFor: 'status' stamp: 'bgf 7/3/2009 10:37'!
userBackFromAway

	"When Away, we inform the world through a notice (handled by the Users tool)
	 and we also disable voice chat and video... so we're not eavesdropping.
	 We also place the awayOverlay over the main view panel, to intercept all events."
	self remove: (awayOverlay).
	avatar ifNotNil: [ 
		self resetActivities.
		QActivityCapture action: #BackFromAway.
	].
	self resumeAudio.
	self resumeVideo.
	self playSoundEffect: #blip.
	self isMeetingLeader ifTrue: [ self leaderBackFromAway ].
	self statusInfo: nil.
	mainView ifNotNil: [
		self statusInfo: 'Activity resumed. (Back from away.)'.
	].


! !

!QwaqParticipantUI methodsFor: 'error handling' stamp: 'ar 6/15/2007 11:01'!
userInterruptWatcher
	"Wait for user interrupts and open a notifier on the active process when one occurs."
	| interruptSemaphore |
	interruptSemaphore := InputSensor interruptSemaphore.
	[true] whileTrue: [
		interruptSemaphore wait.
		self signal: #userInterrupt.
	].! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 2/14/2008 14:29'!
userLogin
	"The login name currently in use."
	^ self login user value asString
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/8/2007 16:59'!
usersPane
	"Answer the usersPane of the receiver"
	<bewareOf: #usersPaneChanged>
	^self propertyValueAt: #usersPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'ar 2/8/2007 16:59'!
usersPane: aValue
	"Modify the receiver's usersPane"
	^self propertyValueAt: #usersPane put: aValue with: #usersPaneChanged! !

!QwaqParticipantUI methodsFor: 'sharepoint' stamp: 'ar 5/10/2010 10:53'!
verifyCheckoutOverwrite: docRef

	"Check if there is a left-over document from a previous launch on the server.
	If there is, ask the user to confirm overwriting it with the latest from the repository.
	If not confirmed, lock the repository version and continue editing the document
	currently on the server.

	Returns a symbol indicating the action to take:
		#cancel - cancel the app launch entirely
		#useExisting - use the document on the server
		#useRepository - use the document from the repository
	"

	| path entries shortName dlg |
	path := docRef get: #asServiceManagerPath.
	entries := serviceMgr listFiles: path.
	entries ifEmpty:[^#useRepository].

	"XXXX: Need to be more clever here. Should give additional info about file size, modification date etc."
	shortName := docRef get: #shortName.
	dlg := CUserDialog new.
	dlg label: 'Confirmation'.
	dlg message: 'This document \\	' withCRs, shortName, '\\has a working copy on this server which may contain changes\you have not checked in.\' withCRs.
	dlg messageText allowEditing: false.
	dlg messageText hResizing: #shrinkWrap; vResizing: #shrinkWrap.
	(dlg addButton: 'Use Working Copy' value: #useExisting) width: 125.
	(dlg addButton: 'Get Fresh Copy' value: #useRepository) width: 125.
	(dlg addButton: 'Cancel Launch' value: #cancel) width: 125.
	^dlg run: #cancel centerAt: self dialogCenter.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/11/2010 11:41'!
videoAreaSash
	"Answer the videoAreaSash of the receiver"
	<bewareOf: #videoAreaSashChanged>
	^self propertyValueAt: #videoAreaSash! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 6/11/2010 11:41'!
videoAreaSash: aValue
	"Modify the receiver's videoAreaSash"
	^self propertyValueAt: #videoAreaSash put: aValue with: #videoAreaSashChanged! !

!QwaqParticipantUI methodsFor: 'video chat' stamp: 'bgf 6/8/2008 19:17'!
videoSession
	^ session ifNotNil: [session videoSession]! !

!QwaqParticipantUI methodsFor: 'private' stamp: 'bgf 10/15/2009 17:08'!
videoSessionTick
	"Tweak-timed periodic tick for the video session, which is not itself tweak.
	 This provokes the polling of available camera devices."

	self videoSession ifNotNilDo: [:vs | vs tick ].! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/23/2010 12:23'!
viewAreaBottom: newViewBottom
	| viewHt viewTop |
	"The argument is the new x coordinate, within me,
	 for the top of the sash between the mainView and the video-feeds panel."

	viewTop := (self menu font pointSize * 2) + 3.	"Echoing the menu bar layout."
	webcamPane visible ifTrue: [
		| ht |
		"Some of the offsets below are a bit empirical, since the sash uses an internal offset
	 	 when tracking its motion, making all the factors come out correctly is tedious."

		viewHt := (newViewBottom max:  (self height // 3)).
		viewHt := (viewHt min: (controlBar top - (viewTop) + 17)).
		ht := (controlBar top -  viewTop )  - viewHt.
		videoAreaSash 
			top: viewHt + 1 ;
			memo: ht;
			fullBounds.
		
		webcamPane 
			topLeft: (-1@(videoAreaSash bottom-1));
			extent: (viewPanel width+2 @ (ht+18)).					
		webcamPane isVideoAreaVisible ifTrue: [webcamPane keepItemsAliveForAWhile].

		viewHt := viewHt - 25.		"Allowance for title strip of lower area."
	] ifFalse: [
		viewHt := controlBar top - viewTop.
	].

	viewPanel height: viewHt.
	"The collapsed navigation panel touches the control-bar view-panel boundary."
	navPanel bottom: (viewHt + 25 + (navPanel collapsedHeight//3)).
! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 6/11/2010 14:16'!
viewAreaRight: newX
	| xLeft |
	"The argument is the new x coordinate, within me,
	 for the left edge of the sash between the mainView and the right-column widgets."

	xLeft := (newX max:  (self width // 2)).
	
	"Attract the default width, for tidiness."
	((xLeft - (width - self defaultPPWidth)) abs < 2) ifTrue: [
		xLeft := (width - self defaultPPWidth).
	].
	
	viewAreaSash visible ifTrue: [
		xLeft := xLeft min: (self width - (self minPPWidth)).
								"See also control bar layoutButtons !!!!."
		viewAreaSash 
			left: xLeft;
			memo: (width - xLeft);
			comeToFront.
		placesPane 
			width: (width - xLeft);
			left: (xLeft+1).
		usersPane 
			width: (width - xLeft);
			left: (xLeft+1).
		catalogPane
			width: (width - xLeft);
			left: (xLeft+1).
		peoplePlacesSash width: (width - xLeft);
			left: (xLeft+1).
	].

	viewPanel width: xLeft+1.
	videoAreaSash width: xLeft + 3.
	webcamPane width: xLeft + 3.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 1/4/2009 19:12'!
viewAreaSash
	"Answer the viewAreaSash of the receiver"
	<bewareOf: #viewAreaSashChanged>
	^self propertyValueAt: #viewAreaSash! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 1/4/2009 19:12'!
viewAreaSash: aValue
	"Modify the receiver's viewAreaSash"
	^self propertyValueAt: #viewAreaSash put: aValue with: #viewAreaSashChanged! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 1/4/2009 21:28'!
viewHeight
	"Nominal height of the main view panel, 
	 for people-places layout.  Does not return zero."
	^ (viewPanel height max: 1).! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'sn 5/13/2011 14:48'!
viewMenu
	| menu vps item subMenu |
	menu := CMenu new.

	"Have to get the order right.
	 Note that several update routines assume that the first n elements
	 of this menu are the viewpoint positions from the viewpoint table."
	vps := {'First Person'. 'Near'. 'Normal'. 'Far'}.
	(QUserPreferences viewpointTable keys includesAllOf: vps) ifFalse: [
		self error: 'Viewpoint menu doesn''t match preference options.'.
	].
	vps do:[:lbl |
		menu add: lbl translateMe target: self action: #changeViewpoint: argument: lbl.
	].

	menu addSeparator.
	menu add: 'Name Tags' translateMe subMenu: (CMenu new).	 "See updateNametagMenu:"

	"--- Middle-group for some readouts (i.e. view this information...)  ---"
	menu addSeparator.
	item := menu add: 'Network Activity...' translateMe target: trafficPane action: #toggle.
	self linkItem: item toTool: trafficPane.
	item := menu add: 'Performance...' translateMe target: statisticsPane action: #toggle.
	self linkItem: item toTool: statisticsPane.
	item := menu add: 'Transcript...' translateMe target: logView action: #toggle.
	self linkItem: item toTool: logView.

	"Useless stereo-rendering toy."
	menu addSeparator.
	subMenu := CMenu new.
	item := subMenu add: 'Off' translateMe target: self action: #stereoOff.
	item checked:true.
	item := subMenu add: 'Hardware' translateMe target: self action: #stereoHardware.
	item := subMenu add: 'Anaglyph' translateMe target: self action: #stereoAnaglyph.
	menu add: 'Stereoscopic' translateMe subMenu: subMenu.

	menu addSeparator.
	(menu add: 'Full Screen' translateMe target: self action: #toggleFullScreen)
		checked: (Tweak fullScreenMode);
		"Full Screen does not work if we are running in a browser 
		 as an ActiveX Control, so ensure that the menu pick is disabled."
		disabled: (QActiveXControl current isActive);
		tag: #fullscreen.

	self startScript: #updateViewMenu when: {menu . #open }.
	^menu! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
viewPanel
	"Answer the viewPanel of the receiver"
	<bewareOf: #viewPanelChanged>
	^self propertyValueAt: #viewPanel! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 7/17/2006 14:11'!
viewPanel: aValue
	"Modify the receiver's viewPanel"
	^self propertyValueAt: #viewPanel put: aValue with: #viewPanelChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 2/26/2008 21:35'!
visiblePanes
	"Answer the visiblePanes of the receiver"
	<bewareOf: #visiblePanesChanged>
	^self propertyValueAt: #visiblePanes! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'das 2/26/2008 21:35'!
visiblePanes: aValue
	"Modify the receiver's visiblePanes"
	^self propertyValueAt: #visiblePanes put: aValue with: #visiblePanesChanged! !

!QwaqParticipantUI methodsFor: 'test-menu' stamp: 'jcg 3/3/2010 15:13'!
voiceTest
	"Test voice transport and report percent extrapolated."
	| hearMyselfOn fileName startingExtrapolations stats endingExtrapolations speakers speaker startingGets endingGets spaceRef panelRef  totalVoiceExtrapolations totalVoiceGets percentVoiceExtrapolations |
	self statusInfo: 'Starting voice test...'; waitTick.
            
	audioSession ifNil: [^QLogger logWarning: 'Test Voice: no audio-session'].
      
      "turn on hear myself"
	"is it on?"
	hearMyselfOn := roundTripSound notNil.
     	hearMyselfOn ifFalse: [self toggleRoundTripSound].
      (Delay forSeconds: 1) wait. 

	"get the speaker"
	speakers := self audioSession speakerManager speakers.
	speakers := speakers select: [:spkr | spkr isConnectedAndStreaming and: [spkr isMe]].
      speakers size = 1 ifFalse: [^QLogger logWarning: 'Unable to locate loopback speaker.'].
	speaker := speakers anyOne.

      "Launch QMS app"
       spaceRef := mainView activeSpace.
       
	"Add display panel here to load the URL"
       panelRef := spaceRef future openQMSForExtrapolationTest: 'http://www.qwaq.com/' andAvatar: self avatar replica.
	 panelRef wait.
       panelRef isValidFarRef ifFalse: [ ^ self ].

         (panelRef island = mainView avatar replica island) ifFalse: [
       ^ self
      ].
      
      mainView avatar replica future takeMeToFrame: panelRef.

	(Delay forSeconds: 3) wait.
	                
	"start voice transport"
	fileName := 'Qwaq-MicCheck-16000Hz.wav'.
	fileName := (FileDirectory pathFromURI: 
							'Objects/Sounds/MicrophoneCheck/', fileName).
      stats := speaker speakerOutput speexSink stats.
	startingExtrapolations := stats totalExtrapolations.
	startingGets := stats totalGets.
	self testVoiceTransportWithBlock: [
		5 timesRepeat: [
			self privateTestVoiceTransportWithFileNamed: fileName allAtOnce: false ].
	].

	stats := speaker speakerOutput speexSink stats.
	endingExtrapolations := stats totalExtrapolations.
	endingGets := stats totalGets.

	"turn off hear myself"
	hearMyselfOn ifFalse: [ self toggleRoundTripSound].
            
	totalVoiceExtrapolations := endingExtrapolations - startingExtrapolations.
      totalVoiceGets := endingGets - startingGets.
     
      totalVoiceGets == 0 ifFalse: [percentVoiceExtrapolations := (100 * totalVoiceExtrapolations asFloat / totalVoiceGets) roundTo: 0.1].
	"display results"
       totalVoiceGets == 0 ifTrue:[self inform: ('Voice Extrapolation Test Result:\\Total Extrapolation: '
                   ,  endingExtrapolations, ' - ', startingExtrapolations, 
		' = ', totalVoiceExtrapolations, ' \Total Gets (Voice Transport launch): ', startingGets,  '\Total Gets (End of Voice Transport): ' ,endingGets) withCRs]
	ifFalse: [self inform: ('Voice Extrapolation Test Result:\\Total Extrapolation: '
                   ,  endingExtrapolations, ' - ', startingExtrapolations, 
		' = ', totalVoiceExtrapolations, ' \Total Gets (Voice Transport launch): ', startingGets,  '\Total Gets (End of Voice Transport): ' ,endingGets, '\Percent Extrapolation = ', percentVoiceExtrapolations , '%')
            withCRs].
 
      "Destroy panel"
       panelRef future destroy. 

	self statusInfo: 'End of test.'.
! !

!QwaqParticipantUI methodsFor: 'load/save' stamp: 'bgf 2/13/2009 22:37'!
volunteerSlowAnimationPreview
	"Periodic call.  If the current forum wants a fresh preview saved, do it.
       This is part of the mechanism for slow-rate changing previews of live forums,
	 so that (for instance) the splash-background during login can show current activity."

	"Am I in condition to make previews?"
	self readyForEvents ifFalse: [ ^ self ].
	session ogl ifNil: [ ^ self ].
	orgServices ifNil: [ ^ self ].
	(orgServices numUsersLoggedIn < 2) ifTrue: [ ^ self  "No audience" ].

	(Smalltalk platformName = 'Win32' and: [ Smalltalk isMainWindowObscured = true ])
	 ifTrue: [ 
		"On windows, OpenGL clips even offscreen rendering if the window is obscured."
		^ self ].

	"The check/confirm/generate business: 
	 this is as close to transactional as we easily can,
 	 without using promise-result booleans that we cannot easily clean up."

	((activeSpace get: #needsPreviewUpdate) = true) ifFalse: [  ^ self ].
	activeSpace future previewUpdated.	

	self startScript: #saveServicePreview:informing: 
		withArguments: { serviceSvc . [:v|]}.
! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 1/23/2008 22:20'!
webcamPane
	"Answer the webcamPane of the receiver"
	<bewareOf: #webcamPaneChanged>
	^self propertyValueAt: #webcamPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 1/23/2008 22:20'!
webcamPane: aValue
	"Modify the receiver's webcamPane"
	^self propertyValueAt: #webcamPane put: aValue with: #webcamPaneChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/11/2009 18:00'!
webcastIndicator
	"Answer the webcastIndicator of the receiver"
	<bewareOf: #webcastIndicatorChanged>
	^self propertyValueAt: #webcastIndicator! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'bgf 12/11/2009 18:00'!
webcastIndicator: aValue
	"Modify the receiver's webcastIndicator"
	^self propertyValueAt: #webcastIndicator put: aValue with: #webcastIndicatorChanged! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 12/23/2009 10:19'!
webcastPane
	"Answer the webcastPane of the receiver"
	<bewareOf: #webcastPaneChanged>
	^self propertyValueAt: #webcastPane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'jcg 12/23/2009 10:19'!
webcastPane: aValue
	"Modify the receiver's webcastPane"
	^self propertyValueAt: #webcastPane put: aValue with: #webcastPaneChanged! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 2/7/2009 20:38'!
windowDisplayBounds
	"The screen-relative version of my inner bounds"
	^ self isSealedApplication ifTrue: [
		HostWindows squeakDisplayRectangle
	] ifFalse: [
		self bounds  "Don't really care... unsealed, we don't save/restore shapes."
	].! !

!QwaqParticipantUI methodsFor: 'initialize' stamp: 'bgf 2/7/2009 19:21'!
windowDisplaySize
	"My size within my window decorations (convenient alias to extent)"
	^ self extent! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'clw 8/20/2008 12:14'!
windowSharePane
	"Answer the windowSharePane of the receiver"
	<bewareOf: #windowSharePaneChanged>
	^self propertyValueAt: #windowSharePane! !

!QwaqParticipantUI methodsFor: 'accessing' stamp: 'clw 8/20/2008 12:14'!
windowSharePane: aValue
	"Modify the receiver's windowSharePane"
	^self propertyValueAt: #windowSharePane put: aValue with: #windowSharePaneChanged! !

!QwaqParticipantUI methodsFor: 'menu' stamp: 'jjl 1/9/2009 17:08'!
writeThemeFile

	self mainView writeTheme.! !

!QwaqParticipantUI methodsFor: 'layout' stamp: 'bgf 7/18/2009 15:48'!
zoomToWindow100: place translation: trans extent: ext
	"Bring the given destination into full view, 
	 retracting all my screen chrome temporarily."

	| ac dx distance front trans2 targetAspect |
	<on: zoomToWindow100 in: avatar>
	ac := mainView camera.
	dx := ext y/2.0.

	"That distance fits the ext height into the view height.
	 If the target has a wider aspect than the view, 
	 move back to fit the horizontal dimension."
	targetAspect := (bounds width / bounds height).
	(ext x / ext y) > (targetAspect) ifTrue: [ 
		dx :=  dx * ((ext x / ext y) / targetAspect).
	].
	distance := dx/((ac viewAngle degreesToRadians/2.0) tan).

	"distance := distance + windowFrame radius."
	front := Vector3 x: 0 y: 0 z: distance.
	front := trans orientation transformPoint: front.
	front := trans translation + front.
	trans2 := trans copy.
	trans2 translation: front.

	self ifAllowedToMove: [ 
		avatar goToPlace: place.
	].
	"NB. if not allowed to move, we're just gonna look from where-ever."

	avatar driver stickyViewpoint: 0.
	mainView keyboardDriving: false.
	self waitTick.
	avatar replica future focalCameraTransform: trans2.
	self hidePanels.
	self waitTick.
	statusLine floatingMessage: '\Zoomed to window.\\Type <Escape> to exit.\' at: self center extraSeconds: 2.! !

Object subclass: #HostWindows
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Platform'!

!HostWindows commentStamp: '<historical>' prior: 0!
A test class for accessing native window handles and their sizes and positions via the extended HostWindowPlugin

	{HostWindows squeakWindowRectangle. HostWindows squeakDisplayRectangle}

	HostWindows primitiveSetCursorPositionX: 1760 Y: -310

	| p |
	p := HostWindows squeakDisplayRectangle center.
	HostWindows primitiveSetCursorPositionX: p x Y: p y!

!HostWindows class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!HostWindows class methodsFor: 'primitives' stamp: 'eem 11/19/2008 19:08'!
nativeDisplayPositionFor: aNativeWindowHandle
	"Answer the origin position of the user area of the given native window in pixels from the topleft corner of the screen.
	 Fail if the windowHandle is invalid."
	<primitive: 'primitiveNativeDisplayPosition' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'eem 11/19/2008 19:09'!
nativeDisplaySizeFor: aNativeWindowHandle
	"Answer the size of the user area of the given native window in pixels.
	 Fail if the windowHandle is invalid."
	<primitive: 'primitiveNativeDisplaySize' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'eem 11/19/2008 19:09'!
nativeWindowPositionFor: aNativeWindowHandle
	"Answer the origin position of the given native window in pixels from the topleft corner of the screen.
	 Fail if the windowHandle is invalid."
	<primitive: 'primitiveNativeWindowPosition' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'eem 11/19/2008 19:09'!
nativeWindowSizeFor: aNativeWindowHandle
	"Answer the size of the given native window in pixels.
	 Fail if the windowHandle is invalid."
	<primitive: 'primitiveNativeWindowSize' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 2/8/2009 09:03'!
primScreenWorkArea
	"The available working area of the screen occupied by the main window."
	^ (self primitiveHostWindowScreenWorkAreaPositionFor: (self squeakWinIndex))
		extent: (self primitiveHostWindowScreenWorkAreaSizeFor: (self squeakWinIndex)).
! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 3/28/2009 10:36'!
primSqueakDisplaySetBounds: aRectangle
	"Set the bounds of the main window (index 1) to the given rectangle,
	 or its best-fit equivalent in the current screen's work area.
	 If we have to adjust, adjust position first, 
	 and reduce size only if the work area is smaller than the requested window.
	 But never put anything offscreen.
	 This is not a true primitive, but uses a bunch... and might primitive-Fail."
	| top left width height wh winP usrP winS usrS |

	"The position-setter we have is for the inner window.
	 So we get the current state of both to figure what offset to use."
	wh := self squeakWindowHandle.
	winP := (self nativeWindowPositionFor: wh).
	usrP := (self nativeDisplayPositionFor: wh).
	winS := (self nativeWindowSizeFor: wh).
	usrS := (self nativeDisplaySizeFor: wh).
	
	width := (aRectangle width).
	height := (aRectangle height).
	top := (aRectangle top).
	left := (aRectangle left).

	"Set the width and height back to 'full-window' basis."
	width := (width + (winS x) - (usrS x)).
	height := (height + (winS y) - (usrS y)).

	"On Windows, set the position back to 'full-window' basis.  (The HostWindows impls disagree.)"
	Smalltalk platformName = 'Win32' ifTrue: [
		top := top - (usrP y) + (winP y).
		left := left - (usrP x) + (winP x).
	].

	"If the new geometry would push the transition offscreen, move first. otherwise resize first."
	self primitiveHostWindowPositionSet: (self squeakWinIndex) x: left y: top.
	Smalltalk windowSize: (width @ height).
	! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 3/28/2009 10:39'!
primSqueakWindowSetBounds: aRectangle
	"Set the bounds of the main window (index 1) to the given rectangle,
	 or its best-fit equivalent in the current screen's work area.
	 The rectangle is the full window, not just the inner display.
	 This is not a true primitive, but uses a bunch... and might primitive-Fail."
	|  top left width height wh win usr |

	"The only position-setter we have is for the inner window (on Mac),
	 but most legacy stuff wants to use the outer window+chrome.
	 So we get the current state of both to figure what offset to use."
	wh := self squeakWindowHandle.
	win := (self nativeWindowPositionFor: wh).
	usr := (self nativeDisplayPositionFor: wh).

	width := (aRectangle width).
	height := (aRectangle height).
	top := (aRectangle top).
	left := (aRectangle left).

	"On Mac, the prim sets the content-window origin; on windows, the full chromed window."
	(Smalltalk platformName = 'Mac OS') ifTrue: [
		self primitiveHostWindowPositionSet: (self squeakWinIndex) x: (left + usr x - win x) y: (top + usr y - win y).
	] ifFalse: [
		self primitiveHostWindowPositionSet: (self squeakWinIndex) x: left y: top.
	].
	Smalltalk windowSize: width@height.	"This sets the full-window, as we want."
	! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 2/7/2009 15:25'!
primitiveHostWindowPositionSet: windowIndex x: x y: y
	"Set the size of the user area of the window in pixels - return what is actually set by the OS/GUI/window manager. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	<primitive: 'primitiveHostWindowPositionSet' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 2/7/2009 16:08'!
primitiveHostWindowScreenWorkAreaPositionFor: windex
	"Answer the origin  of host-defined screen work area for the screen holding the given window.
	 Fail if the windowHandle is invalid."
	<primitive: 'primitiveHostWindowScreenWorkAreaPosition' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 2/7/2009 16:08'!
primitiveHostWindowScreenWorkAreaSizeFor: windex
	"Answer the origin  of host-defined screen work area for the screen holding the given window.
	 Fail if the windowHandle is invalid."
	<primitive: 'primitiveHostWindowScreenWorkAreaSize' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'bgf 2/7/2009 15:25'!
primitiveHostWindowSizeSet: windowIndex x: x y: y
	"Set the size of the user area of the window in pixels - return what is actually set by the OS/GUI/window manager. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	<primitive: 'primitiveHostWindowSizeSet' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'primitives' stamp: 'eem 11/20/2008 16:14'!
primitiveSetCursorPositionX: x Y: y
	"Move the cursor to the specified point in desktop coordinates"
	<primitive: 'primitiveSetCursorPosition' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'utilities' stamp: 'bgf 2/20/2009 14:45'!
squeakDisplayRectangle
	| swh |
	[
		swh := self squeakWindowHandle.
		^(self nativeDisplayPositionFor: swh) extent: (self nativeDisplaySizeFor: swh)
	] ifError: [ :msg |
		"Rather than letting primitive fails ripple up, just fake it if unsupported."
		^ World bounds
	]
! !

!HostWindows class methodsFor: 'utilities' stamp: 'ar 3/30/2010 10:25'!
squeakDisplaySetBounds: aRectangle
	"Set the bounds of the main display (contents of main window 1) to the given rectangle,
	 or its best-fit equivalent in the current screen's work area.
	 If the primitive support fails, fall back to just setting the size."

	[ self primSqueakDisplaySetBounds: aRectangle ] on: PrimitiveFailed do:[
		| chrome |
		"Fallback to the older size-only primitives.
		Assumes we're not doing this back-to-back with other size changes,
		since we get the chrome sizes by asking for the current difference..."
		chrome := Smalltalk windowSize - (World extent).
		Smalltalk windowSize: (aRectangle extent + chrome).
	].! !

!HostWindows class methodsFor: 'utilities' stamp: 'ar 3/30/2010 10:25'!
squeakDisplaySetSize: anExtent
	"Set the inner extent of the main display (contents of main window 1) to the given rectangle,
	 or its best-fit equivalent in the current screen's work area.
	 If the primitive support fails, fall back to just setting the size."

	[ self primSqueakDisplaySetBounds: (
		self squeakDisplayRectangle origin extent: anExtent) ] on: PrimitiveFailed do:[
			| chrome |
			"Fallback to the older size-only primitives.
			Assumes we're not doing this back-to-back with other size changes,
			since we get the chrome sizes by asking for the current difference..."
			chrome := Smalltalk windowSize - (World extent).
			Smalltalk windowSize: (anExtent + chrome).
	].! !

!HostWindows class methodsFor: 'private' stamp: 'eem 3/23/2010 17:08'!
squeakWinIndex

	"What winIndex refers to Squeak's main window?"
	^ Smalltalk platformName caseOf: {
		['Win32'] -> [ (self squeakWindowHandle) ].	"Window Index is just the handle here."
		['unix'] -> [ self squeakWindowHandle ]. 	"Window Index is just the handle here."
		['Mac OS'] -> [ 1 ] 	"The Host Window plugin shares the window-ref lookup table."
	} otherwise: [ -1 ].! !

!HostWindows class methodsFor: 'primitives' stamp: 'eem 11/19/2008 17:00'!
squeakWindowHandle
	"Answer the native Squeak window handle, presumably a pointer, as an unsigned integer.
	 This value should be usable in native calls to the window manager as well as
	 the argument to primitiveNativeWindowPosition:"
	<primitive: 'primitiveSqueakWindowHandle' module: 'HostWindowPlugin'>
	self primitiveFailed! !

!HostWindows class methodsFor: 'utilities' stamp: 'bgf 2/20/2009 14:45'!
squeakWindowRectangle
	| swh |
	[
		swh := self squeakWindowHandle.
		^( self nativeWindowPositionFor: swh) extent: (self nativeWindowSizeFor: swh)
	] ifError: [ :msg|
		"Rather than letting primitive fails ripple up, just fake it if unsupported."
		^ World bounds
	]! !

!HostWindows class methodsFor: 'utilities' stamp: 'ar 3/30/2010 10:25'!
squeakWindowSetBounds: aRectangle
	"Set the bounds of the main window (index 1) to the given rectangle,
	 or its best-fit equivalent in the current screen's work area.
	 If the primitive support fails, fall back to just setting the size."

	[ self primSqueakWindowSetBounds: aRectangle ] on: PrimitiveFailed do:[
		Smalltalk windowSize: (aRectangle extent)
	].! !

!HostWindows class methodsFor: 'utilities' stamp: 'ar 3/30/2010 10:26'!
squeakWindowSetSize: anExtent
	"Set the inner extent of the main display (contents of main window 1) to the given rectangle,
	 or its best-fit equivalent in the current screen's work area.
	 If the primitive support fails, fall back to just setting the size."

	[ self primSqueakWindowSetBounds: (
		(self squeakWindowRectangle origin) extent: anExtent) ] on: PrimitiveFailed do:[
			"Fallback to the older size-only primitives."
			Smalltalk windowSize: (anExtent).
	].! !

Object subclass: #QActiveXControl
	instanceVariableNames: ''
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'Qwaq-Client-Plugin'!

!QActiveXControl commentStamp: 'ar 4/16/2011 18:50' prior: 0!
I'm a simple, singleton, interface to the ActiveX Control that is running in Internet Explorer.    If the client was launched from the browser, then the testing method, #isActive will return true.  Always access the control through the class side #current method.  For example, to check if the control is active:

     QActiveXControl current isActive

The ActiveX Control implements one method and its prototype is:

   HRESULT OpenQwaqCommand([in] BSTR command)

To access the parameter, command, sent by OpenQwaqCommand, use #currentCommand.   When the ActiveX Control sends this command, it will be sent up through the image's input process.  The OpenQwaq client is expected to register itself with SystemEventHandler so that it can receive plugin event (#pluginEvent is sent see HandMoprh>>#generatePluginEvent:).

Here is an example of invoking OpenQwaqCommand() in javascript (assuming that OpenQwaqClient is the control):

   function goToPlace(place)
  {
    OpenQwaqClient.OpenQwaqCommand('goToPlace:'+place);
  }

Here is how one would access the command that was sent:

  QActiveXControl current currentCommand

If there is no command, #currentCommand will return an empty string.

There are currently two implementation details that clients should be aware of:

  1) A side-effect of calling #currentCommand is that it calling it again will not return the same command
  2) Commands sent from the ActiveX Control are *not* queued

For (1), a side-efect of calling #currentCommand is that the underlying resource associated with the command is free()'d.  So, it is clients responsiblities to current hold onto the string returned by #currentCommand.  Otherwise, if you call #currentCommand expecting to get the same result, you'll simplly get an empty string.

For (2) if the ActiveX Control sends multiple commands to the OpenQwaq Client, #currentCommand will only return the last command received.

Both 1 & 2 are likely to be improved at some point, but for now that is the current state.

The ActiveX Control can raise one event and its prototype is:

  void OpenQwaqEvent(BSTR message)

This is the mechanism to use to communicate from the client back into the ActiveX Control.  For example:

   QActiveXControl current sendEvent: 'hello world'

In the browser, one can write javascript to register itself with the OpenQwaqEvent event.  For example (assuming that OpenQwaqClient is the control)

     OpenQwaqClient.attachEvent('OpenQwaqEvent', handleOpenQwaqEvent);

Then you can implement a javascript function to do something with the event:
  
  function handleOpenQwaqEvent(message)
  {
    /* do something with message */
  }

The best place to register the event handling in javascript once the document has loaded.!

!QActiveXControl class methodsFor: 'LICENSE' stamp: 'osar 4/20/2011 10:25'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QActiveXControl class methodsFor: 'accessing' stamp: 'bobw 8/2/2010 10:42'!
current

	^Current ifNil: [Current := self new]! !

!QActiveXControl class methodsFor: 'class initialization' stamp: 'bobw 8/2/2010 10:42'!
initialize
	"self initialize"
	
	"Add the receiver to the list of classes to be notified when the system has
	 returned from snapshot."
	Smalltalk addToStartUpList: self! !

!QActiveXControl class methodsFor: 'class initialization' stamp: 'bobw 8/2/2010 10:42'!
obsolete

	Smalltalk removeFromStartUpList: self.
	super obsolete! !

!QActiveXControl class methodsFor: 'class initialization' stamp: 'bobw 8/2/2010 10:42'!
startUp: resuming

	Current := nil! !

!QActiveXControl methodsFor: 'accessing' stamp: 'bobw 8/2/2010 10:44'!
currentCommand
	
	^((self primitivePluginBrowserCommand) ifNil: ['']) asString! !

!QActiveXControl methodsFor: 'testing' stamp: 'bobw 8/2/2010 10:43'!
isActive
	"Am I associated with a plugin in a browser?"

	^StandardFileStream isRunningAsBrowserPlugin! !

!QActiveXControl methodsFor: 'private' stamp: 'bobw 8/2/2010 10:43'!
primitivePluginBrowserCommand
	"self current primitivePluginBrowserCommand"

	<primitive:'primitivePluginCommandFromBrowser'>
	^nil! !

!QActiveXControl methodsFor: 'private' stamp: 'bobw 8/2/2010 10:43'!
primitivePluginSendEvent: aString
	"self current primitivePluginSendEvent: 'hello world'"

	<primitive:'primitivePluginCommandToBrowser'>
	^nil! !

!QActiveXControl methodsFor: 'accessing' stamp: 'bobw 8/2/2010 10:45'!
sendEvent: aString
	
	self primitivePluginSendEvent: aString! !

Object subclass: #QDriveKeyMap
	instanceVariableNames: ''
	classVariableNames: 'KeyMap'
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QDriveKeyMap commentStamp: '<historical>' prior: 0!
Keyboard mappings for keyboard "ASDW" driving.

Provides platform-specific mappings of keyPress/keyRelease keyValues
into standard driving operation symbols:

	#moveBack .
	#moveForward.
	#moveLeft.
	#moveRight.
	#moveUp .
	#moveDown .
	#resetPosition.
	#rotateLeft .
	#rotateRight.

	#lookDown.
	#lookUp.
	#resetLook.

	#return.
	#escape.

"QKeyMap map: aKeyValue"  returns the current default mapping
for the given key value (from a keyPress/keyRelease event)
or nil for unmapped values. 

The map also contains some entries whose keys are symbols
rather than (integer) keyValues.  These are (with usual values):
	#scrollUp -> #moveUp			
	#scrollDown -> #moveDown

These help detect mouse scroll wheel motion:
	(self map: #scrollUp) = (self map:(aKeyStrokeEvent keyValue)) 
is true when aKeyStrokeEvent 
looks like a synthetic event representing mouse scroll wheel motion.!

!QDriveKeyMap class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QDriveKeyMap class methodsFor: 'class initialization' stamp: 'bgf 11/2/2006 16:51'!
initialize

	"QDriveKeyMap initialize"
	Smalltalk removeFromStartUpList: self.
	Smalltalk addToStartUpList: self after: CPlatform.
! !

!QDriveKeyMap class methodsFor: 'class initialization' stamp: 'bgf 8/7/2009 00:49'!
keymapMacOS

	"Reloaded on startUp:, so saving the image updates the maps."

	^ Dictionary newFromPairs: {

			"asdwqe"
			0 . #rotateLeft .
			1 . #moveBack .
			2 . #rotateRight.
			13 . #moveForward.
			12 . #moveLeft.
			14 . #moveRight.
			
			"Arrow keys."
			126 . #moveForward.
			123 . #rotateLeft.
			124 . #rotateRight.
			125 . #moveBack.

			"Leaning forward and back (r, f, c)"
			15. #lookDown.
			3.	#resetLook.
			8.	#lookUp.

			"HOME"
			115 . #resetPosition.

			"Page up, Page down."
			116 . #moveUp .
			121 . #moveDown .

			"Special control characters."
			53 . #escape.
			-1 . #shift.  "The Mac doesn't generate key events for shift."

			"Aliasing - the synthetic keystroke values for scroll wheel motion."

			"The Mac VM generates keyDown/Up events with key value 126,125  for the scroll wheel,
			 but also generates keyStroke events with key value 30 and 31."

			30 . #scrollUp.
			31 . #scrollDown.

			"Symbolic reverse mappings for keyStroke: processing"
			#scrollUp . #scrollUp .		
			#scrollDown . #scrollDown.
	}.

	! !

!QDriveKeyMap class methodsFor: 'class initialization' stamp: 'eem 5/26/2010 16:47'!
keymapUnix

	"Reloaded on startUp:, so saving the image updates the maps.
	 Almost exact clone of the Win32 map (i.e. translated keys not raw scan codes)"

	^ Dictionary newFromPairs: {

			"asdwqe"
			97 . #rotateLeft .
			115 . #moveBack .
			100 . #rotateRight.
			119. #moveForward.
			113. #moveLeft.
			101. #moveRight.
			
			"Arrow keys."
			30 . #moveForward.
			28 . #rotateLeft.
			29 . #rotateRight.
			31 . #moveBack.

			"Leaning forward and back (r, c, f)"
			114. #lookDown.
			99.	#lookUp.
			102.	#resetLook.

			"HOME"
			1 . #resetPosition.

			"Page up, Page down."
			11 . #moveUp .
			12 . #moveDown .

			"Special control characters."
			27 . #escape.
			16 . #shift.

			"Symbolic reverse mappings for keyStroke: processing"

			"The synthetic keystroke values for scroll wheel motion;
			 the windows & unix VMs VM translate these into key 30 and 31 keyStrokes, with no corresponding up/down."
			#scrollUp . #moveForward .
			#scrollDown . #moveBack .
	}.

	! !

!QDriveKeyMap class methodsFor: 'class initialization' stamp: 'eem 5/26/2010 16:47'!
keymapWin32

	"Reloaded on startUp:, so saving the image updates the maps."

	^ Dictionary newFromPairs: {

			"ASDWQE"
			65 . #rotateLeft .
			83 . #moveBack .
			68 . #rotateRight.
			87 . #moveForward.
			81 . #moveLeft.
			69 . #moveRight.
			
			"Arrow keys."
			30 . #moveForward.
			28 . #rotateLeft.
			29 . #rotateRight.
			31 . #moveBack.

			"Leaning forward and back (R, C, F)"
			82. #lookDown.
			67.	#lookUp.
			70.	#resetLook.

			"HOME"
			1 . #resetPosition.

			"Page up, Page down."
			11 . #moveUp .
			12 . #moveDown .

			"Special control characters."
			27 . #escape.
			16 . #shift.

			"Symbolic reverse mappings for keyStroke: processing"

			"The synthetic keystroke values for scroll wheel motion;
			 the windows VM translates these into key 30 and 31 keyStrokes, with no corresponding up/down."
			#scrollUp . #moveForward .
			#scrollDown . #moveBack .
	}.

	! !

!QDriveKeyMap class methodsFor: 'system startup' stamp: 'eem 5/26/2010 16:36'!
startUp: resuming

	super startUp:resuming.

	"Load the keymap for the current platform."

	KeyMap := Smalltalk platformName caseOf: {
			['Mac OS'] -> [ self keymapMacOS ].
			['Win32'] -> [ self keymapWin32 ].
			['unix'] -> [ self keymapUnix ].
		} otherwise: [Dictionary new].

! !

!QDriveKeyMap methodsFor: 'accessing' stamp: 'bgf 11/3/2006 01:17'!
map: aKeyValue
	
	"Stub for now... use the class default directly."

	KeyMap ifNotNil: [
		^ KeyMap at: aKeyValue ifAbsent: [nil].
	].
	^ nil! !

Object subclass: #QMimeWrapper
	instanceVariableNames: 'description payload'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Platform'!

!QMimeWrapper commentStamp: 'jcg 3/24/2008 11:22' prior: 0!
A simple XML wrapper file incorporating state-of-the-art security-by-obscurity.  An XML attribute specifies whether obfuscation is enabled (the XML calls it 'encryption' to heighten the false sense of security).  If 'encryption' is enabled, then the payload consists of:
	- a MIME-encoded string of:
		- the concatenation of:
			- 13 random bytes (ignored when decrypting)
			- 16-byte RC4 key 
			- message encoded with the above RC4 key

The number 13 is picked for its connotation of unluckiness... don't rely on the obfuscation provided by this class for any real security.!

!QMimeWrapper class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMimeWrapper class methodsFor: 'utility' stamp: 'hrs 4/30/2010 16:04'!
asStringFromHost: host organization: org forum: forum location: location
	"Generate a content string for a .qwq file."
	^self asStringFromHost: host organization: org forum: forum location: location 
		user: nil password: nil proxyPublicationId: nil.! !

!QMimeWrapper class methodsFor: 'utility' stamp: 'hrs 5/3/2010 11:33'!
asStringFromHost: host organization: org forum: forum location: location user: user password: password
	"Generate a content string for a .qwq file."
	^self asStringFromHost: host organization: org forum: forum location: location 
		user: user password: password proxyPublicationId: nil.! !

!QMimeWrapper class methodsFor: 'utility' stamp: 'hrs 4/30/2010 16:04'!
asStringFromHost: host organization: org forum: forum location: location user: userOrNil password: passwordOrNil proxyPublicationId: pubIdOrNil
	"Generate a content string for a .qwq file."
	| payload |
	^String streamContents:[:s| 
		(XMLWriterIndenting on: s)
			startTag: 'qwq'; attribute: 'version' value: '1.0'; endTag;	
			  "Older versions of QMimeWrapper require a description."
                    tag: 'description' pcData: host, '/', org, '/', (forum ifNil: ['']), 
					(location ifNil: [''] ifNotNil: ['?go=', location]);	  
			  tag: 'payload' pcData: 
					(String streamContents: [:inner |
							payload := XMLWriter on: inner.
							payload
								startTag: 'qrl'; endTag;
									tag: 'host' pcData: host;
									tag: 'organization' pcData: org.
							forum ifNotNil: [payload tag: 'forum' pcData: forum].
							location ifNotNil: [payload tag: 'location' pcData: location].
							userOrNil ifNotNil: [payload tag: 'user' pcData: userOrNil].
							passwordOrNil ifNotNil: [payload tag: 'password' pcData: passwordOrNil].
							pubIdOrNil ifNotNil: [payload tag: 'proxyPublicationId' pcData: pubIdOrNil hex].
							payload		
								endTag: 'qrl']);
			endTag: 'qwq'.
		]! !

!QMimeWrapper class methodsFor: 'example' stamp: 'hrs 1/8/2009 13:39'!
exampleQwqFile
	^String streamContents:[:s| 
	(XMLWriterIndenting on: s)
		startTag: 'qwq'; attribute: 'version' value: '1.0'; endTag;
		  startTag: 'description'; endTag;
 		     pcData: 'An example .qwq file.
Comment elements are NOT allowed. (Bug.)
Note that the payload is pcdata that happens to be XML.';
		  endTag: 'description';
		  startTag: 'payload'; endTag;
		    pcData: (String streamContents: [:inner |
						(XMLWriterIndenting on: inner)
							startTag: 'qrl'; endTag;
							   startTag: 'organization'; endTag;
								pcData: 'Qwaq-internal';
							   endTag: 'organization';
							   startTag: 'forum'; endTag;
							      pcData: 'Qwaq-PA-MainOffice';
							   endTag: 'forum';
							endTag: 'qrl']);
		  endTag: 'payload';
		endTag: 'qwq'.
	]! !

!QMimeWrapper class methodsFor: 'instance creation' stamp: 'jcg 3/24/2008 11:30'!
importFrom: aStringOrStream
	^ [self new importFrom: aStringOrStream; yourself] on: Error do: [nil]! !

!QMimeWrapper methodsFor: 'private' stamp: 'jcg 3/24/2008 11:16'!
decryptPayloadFrom: aStringOrStream
	"Decrypt payload from stream, but don't store it in 'payload'; just answer it."
	| stream demimed cipher |
	aStringOrStream isStream 
						ifTrue: [stream := aStringOrStream]
						ifFalse: [stream := aStringOrStream readStream].
	demimed := (Base64MimeConverter mimeDecodeToBytes: stream) contents.
	cipher := TARC4Cypher key: (demimed copyFrom: 14 to: 29).
	cipher process: demimed from: 30 to: demimed size.
	^ (demimed copyFrom: 30 to: demimed size) asString.! !

!QMimeWrapper methodsFor: 'accessing' stamp: 'jcg 3/24/2008 13:49'!
description
	^description! !

!QMimeWrapper methodsFor: 'accessing' stamp: 'jcg 3/24/2008 13:50'!
description: descriptionString
	description := descriptionString! !

!QMimeWrapper methodsFor: 'export' stamp: 'jcg 3/24/2008 11:17'!
encryptedPayload
	| key cypher encoded |
	key := TObjectID new.
	cypher := TARC4Cypher key: key.
	encoded := (TObjectID new: 13), key, (cypher process: payload asByteArray).
	^ (Base64MimeConverter mimeEncode: (ReadStream on: encoded)) contents.
! !

!QMimeWrapper methodsFor: 'export' stamp: 'jcg 3/21/2008 15:32'!
export
	| stream | 
	stream := WriteStream on: ''.
	self exportOn: stream.
	^stream contents! !

!QMimeWrapper methodsFor: 'export' stamp: 'jcg 3/24/2008 10:51'!
exportOn: aStream
	self exportOn: aStream encrypted: true! !

!QMimeWrapper methodsFor: 'export' stamp: 'jcg 3/24/2008 11:36'!
exportOn: aStream encrypted: isEncrypted
	| writer |
	writer := XMLWriterIndenting on: aStream.
	writer 
		startTag: 'dot-qwq';
		attribute: 'version' value: '1.0';
		endTag;
			startTag: 'description'; endTag;
			pcData: description;
			endTag: 'description';
			startTag: 'payload';
			attribute: 'type' value: (isEncrypted ifTrue: ['encrypted'] ifFalse: ['plaintext']);
			endTag;
			pcData: (isEncrypted ifTrue: [self encryptedPayload] ifFalse: [payload]);
			endTag: 'payload';
		endTag: 'dot-qwq'.
! !

!QMimeWrapper methodsFor: 'export' stamp: 'jcg 3/24/2008 10:52'!
exportPlaintext
	| stream | 
	stream := WriteStream on: ''.
	self exportOn: stream encrypted: false.
	^stream contents! !

!QMimeWrapper methodsFor: 'import' stamp: 'hrs 1/7/2009 14:54'!
importFrom: aStreamOrString
	| stream node |
	stream := aStreamOrString isStream 
								ifTrue: [aStreamOrString] 
								ifFalse: [aStreamOrString readStream].
	stream ascii.
	node := (XMLDOMParser on: stream) nextEntity.
	(node attributeAt: 'version') = '1.0' 
		ifFalse: [self error: 'expecting version 1.0'].
	description := (node elementAt: 'description') ifNotNilDo: [:desc | desc contentString].
	node := node elementAt: 'payload'.
	(node attributeAt: 'type') = 'encrypted' 
		ifTrue: [payload := self decryptPayloadFrom: node contentString readStream]
		ifFalse: [payload := node contentString].
	! !

!QMimeWrapper methodsFor: 'initialize' stamp: 'jcg 3/5/2008 20:43'!
initialize
	super initialize.
	description :=  'no description'.
	payload := ''.! !

!QMimeWrapper methodsFor: 'accessing' stamp: 'jcg 3/24/2008 13:49'!
payload
	^payload! !

!QMimeWrapper methodsFor: 'accessing' stamp: 'jcg 3/24/2008 13:50'!
payload: payloadString
	payload := payloadString! !

Object subclass: #QTextLine
	instanceVariableNames: 'message location alignment form style foreground background inset dropShadowEnabled material textAlignment textEmphasis viewBounds standoff'
	classVariableNames: ''
	poolDictionaries: 'OpenGLConstants'
	category: 'Qwaq-Client'!

!QTextLine commentStamp: '<historical>' prior: 0!
Text lines that can appear on an OpenGL view;
used for status and debug message display in participant UI,
and to render 2D text for the QLabel2D widget.

Rendered by filling a rectangular surround with background
(usually a translucent contrast color) before rendering the text in its foreground color.

For efficiency, the rendered appearance is only updated when the #message: is set.
Set style and color attributes first, then set the message.

The default coloring is white (alpha .75) on black (alpha 0.25).

Text rendering is done by compositing to a Form, so there
are no restrictions on font faces or formats.

This is neither a CPanel nor a Croquet TFrame; 
it is a specialized handler for a transient display string 
rendered on an OpenGL display.!

!QTextLine class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:17'!
alignment

	"Usually #bottomLeft or #topLeft; alignment of the text relative to my location."

	^alignment! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 1/27/2008 09:43'!
alignment: anAlignment

	"Usually #bottomLeft or #topLeft; 
	 alignment of the text rectangle as a whole relative to my location."
	alignment := anAlignment! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:22'!
background 
	^ background! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:22'!
background: aColor
	background := aColor! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/3/2007 03:00'!
dropShadowEnabled
	^ dropShadowEnabled! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/3/2007 02:50'!
dropShadowEnabled: aBool
	dropShadowEnabled := aBool.! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:22'!
foreground
	^ foreground! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:21'!
foreground: aColor
	foreground := aColor! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/2/2007 23:43'!
form
	"The Form containing the current message.
	 In OpenGL, this wants blend function GLSrcAlpha, GLOneMinusSrcAlpha"
	^ form! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:34'!
height
	" The height of the current message."
	^ form ifNotNil: [
		form height.
	] ifNil: [
		0
	].! !

!QTextLine methodsFor: 'initialize-release' stamp: 'bgf 7/19/2008 12:21'!
initialize

	super initialize.
	message := nil.
	location := 0@0.
	standoff := 0@0.
	inset := 8@3.
	dropShadowEnabled := false.
	alignment := #topLeft.
	foreground := Color black alpha:1.0.
	background := Color white alpha: 0.7.
	material := TMaterial new.
	textAlignment := #centered.! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/2/2007 22:55'!
inset
	^ inset! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/2/2007 22:55'!
inset: aPoint
	"The x@y inset margins around the text."
	inset := aPoint isNumber ifTrue: [aPoint@aPoint] ifFalse: [aPoint]! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:15'!
location
	"The xy location of the text in window coordinates."
	^ location! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:15'!
location: aPoint
	location := aPoint! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:20'!
message
	"The message string currently being displayed; nil when the status line is empty."
	^ message! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 1/27/2008 08:55'!
message: aMessage
	"The message string currently being displayed; nil when the text line is empty."

	(message ~= aMessage) ifTrue: [
		message := aMessage.
		self reset
	].! !

!QTextLine methodsFor: 'render' stamp: 'bgf 2/26/2009 01:20'!
prepareMessage: aString

	"Composite the message into my display form, to prepare for rendering."
	| para dim at text |
	message := aString.
	message ifNil: [
		form := nil.
		^ self
	].

	text := aString asString asText.
	(textEmphasis notNil and: [ textEmphasis > 0 ]) ifTrue: [
		| code attr |
		1 to: 3 do:[:i|
			code := 1 bitShift: i-1.
			(textEmphasis bitAnd: code) = 0 ifFalse:[
				attr := TextEmphasis new emphasisCode: code.
				text addAttribute: attr]].
	].

	"Render the text into a Form, for display later by OGL"
	para := Paragraph withText: text style: (self style).
	para foregroundColor: foreground  
		  backgroundColor: background;
		  perform: textAlignment.
		 "Most all my uses are single-line.  The exceptions, at time of writing,
	 	  is the avatar label when there is an away notice
	  	  (for which multiline text should always be centered),
		  and QThumbtacks, for which centering seems to work perfectly well."

	dim := para boundingBox extent + (2*inset) -(0@1).
	at := inset x @ (inset y - 2).
	form := Form extent: dim depth: 32.
	form fill: form boundingBox fillColor: background.
	para displayOn: form at: at 
						clippingBox: (form boundingBox) 
						rule: (Form over) 
						fillColor: foreground.

	form borderWidth: 1 color: (background muchDarker alpha: (background alpha)*0.8).
	

	"Needed for ogl drawPixels"
	form := form flipBy: #vertical centerAt: (form center).	
! !

!QTextLine methodsFor: 'render' stamp: 'bgf 4/25/2007 17:14'!
redraw
	self prepareMessage: message! !

!QTextLine methodsFor: 'render' stamp: 'bgf 2/26/2009 01:19'!
render: ogl

	| offset shadow box |

	form ifNil: [ self redraw ].
	form ifNil: [ ^ nil ].

	material enable: ogl.
	ogl pushBlend: GLSrcAlpha with: GLOneMinusSrcAlpha.

	box := (0@0 extent: (form extent)).
	offset := box perform: alignment.
	box := box translateBy: (location - offset - standoff).
	viewBounds ifNotNil: [
		box := box translatedToBeWithin: viewBounds.
	].

	true ifTrue: [
		shadow := Form extent: form extent -(2@2) depth: 32.
		shadow fillColor: ((Color gray: 0.1) alpha: (background alpha*0.2)).
		ogl glRasterPos3f: (box left + 5) 
			with: (box bottom + 4) 
			with: 0.
		 ogl glDrawPixels: shadow width with: shadow height 
			with: ogl imagePixelFormat32 with: ogl imagePixelType32 with: shadow bits.
	].

	ogl glRasterPos3f: (box left) 
		with: (box bottom ) 
		with: 0.
	ogl glDrawPixels: form width with: form height 
		with: ogl imagePixelFormat32 with: ogl imagePixelType32 with: form bits.

	ogl popBlend.
	material disable: ogl.

! !

!QTextLine methodsFor: 'render' stamp: 'bgf 1/27/2008 08:55'!
reset
	form := nil! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 7/19/2008 12:18'!
standoff
	"The offset from 'location' to the actual alignment point."
	^ standoff! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 7/19/2008 12:17'!
standoff: aPoint
	"The desired standoff from the location to where the floating message alignment point falls;
	 used to keep the message away from the actual location."

	standoff := aPoint! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 7/21/2007 16:55'!
style 
	"My current text style.. lazily initialized to default."
	style ifNil: [
		style := QStyles find:#ControlStyle
	].
	^ style! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:20'!
style: aTextStyle 
	"My text style."
	style := aTextStyle.
! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 1/27/2008 09:44'!
textAlignment
	^ textAlignment

	"Paragraph alignment of the text when it is multiline"
! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 1/27/2008 09:44'!
textAlignment: anAlignmentSelector

	"Paragraph-alignment for this message... centered, justified, rightFlush, leftFlush.
	Default is #centered"

	textAlignment == anAlignmentSelector ifFalse: [
		textAlignment := anAlignmentSelector
							ifNotNil: [anAlignmentSelector asSymbol].
		self reset.
	].! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 1/27/2008 08:53'!
textEmphasis
	^ textEmphasis

	"Tweak-style (0,1,2...) emphasis code"! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 1/27/2008 08:55'!
textEmphasis: anEmphasisCode

	textEmphasis ~= anEmphasisCode ifTrue: [
		textEmphasis := anEmphasisCode.
		self reset.
	].! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 7/19/2008 11:55'!
viewBounds
	"The bounds of the area in which the line should appear."
	^ viewBounds ! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 7/19/2008 11:46'!
viewBounds: aRectangle
	"The bounds of the area in which the line should appear."
	viewBounds := aRectangle! !

!QTextLine methodsFor: 'accessing' stamp: 'bgf 11/1/2006 07:34'!
width
	" The width of the current message."
	^ form ifNotNil: [
		form width.
	] ifNil: [
		0
	].! !

Object subclass: #QUndoableAction
	instanceVariableNames: 'description requiredFrames'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Undo'!

!QUndoableAction commentStamp: '<historical>' prior: 0!
Abstract class of entries in the undo-redo list;
each instance represents a user's action that can be undone or redone.

Each action also has a description - very short word or phrase summarizing what was done (e.g. 'Rotate')

Each also has a list of 'requiredFrames'  - an array of TFrame far-refs.
All those frames must be verifiable as rooted tframes in the space, 
for the action to be available for undo or redo.!

QUndoableAction subclass: #QBlockAction
	instanceVariableNames: 'undoBlock redoBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Undo'!

!QBlockAction commentStamp: '<historical>' prior: 0!
An undoable action using block closures for undo and redo.
This is a 'generic' undoable action when the undo and redo need no special dependency tracking.

The blocks are performed (in client harness context) to carry out the undo or redo.
The blocks are involed with 2 optional arguments 
		avatarReplicaFarRef
		participantUI !

!QBlockAction class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QBlockAction class methodsFor: 'instance creation' stamp: 'bgf 3/14/2010 18:09'!
undoBlock: aBlock1 redoBlock: aBlock2 description: aString 
	"Factory method for a block-closure based undoable action record.
	 The blocks are performed (in client harness context) to carry out the undo or redo.
	 The blocks are involed with 2 optional arguments 
		avatarReplicaFarRef  participantUI  "

	^ (self new)
		undoBlock: aBlock1 redoBlock: aBlock2 description: aString requiredFrames: {}

! !

!QBlockAction class methodsFor: 'instance creation' stamp: 'bgf 3/14/2010 18:09'!
undoBlock: aBlock1 redoBlock: aBlock2 description: aString requiredFrames: refs
	"Factory method for a block-closure based undoable action record.
	 The blocks are performed (in client harness context) to carry out the undo or redo.
	 The blocks are involed with 2 optional arguments 
		avatarReplicaFarRef  participantUI  "

	^ (self new)
		undoBlock: aBlock1 redoBlock: aBlock2 description: aString requiredFrames: refs

! !

!QBlockAction methodsFor: 'actions' stamp: 'bgf 3/14/2010 17:55'!
redo: participantUI
		
	redoBlock valueWithPossibleArgs: { participantUI avatar replica .  participantUI }.
! !

!QBlockAction methodsFor: 'actions' stamp: 'bgf 3/14/2010 17:55'!
undo: participantUI
		
	undoBlock valueWithPossibleArgs: { participantUI avatar replica .  participantUI }.
! !

!QBlockAction methodsFor: 'initialize-release' stamp: 'bgf 2/24/2010 14:43'!
undoBlock: aBlock1 redoBlock: aBlock2 description: aString

	description := aString.
	undoBlock := aBlock1.
	redoBlock := aBlock2.

! !

!QBlockAction methodsFor: 'initialize-release' stamp: 'bgf 3/5/2010 09:02'!
undoBlock: aBlock1 redoBlock: aBlock2 description: aString requiredFrames: refs

	description := aString.
	undoBlock := aBlock1.
	redoBlock := aBlock2.
	self addRequiredFrames: refs.
! !

QUndoableAction subclass: #QStateVectorAction
	instanceVariableNames: 'primaryFrame oldState newState'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client-Undo'!

!QStateVectorAction commentStamp: '<historical>' prior: 0!
An Undo record for a primary frame that supplies the method
	applyStateVector: stateVec
where the state vector is a copy-as-argument plain array of data
from which the user-editable undoable state can be reconstituted.!

!QStateVectorAction class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QStateVectorAction class methodsFor: 'instance creation' stamp: 'bgf 3/12/2010 07:31'!
frame: frameRef oldState: oldVec newState: newVec description: descrip  

	^ (self new)
		frame: frameRef oldState: oldVec newState: newVec description: descrip  
! !

!QStateVectorAction class methodsFor: 'instance creation' stamp: 'bgf 3/12/2010 07:31'!
frame: frameRef oldState: oldVec newState: newVec description: descrip  requiredFrames: reqs

	^ (self new)
		frame: frameRef oldState: oldVec newState: newVec description: descrip  requiredFrames: reqs
! !

!QStateVectorAction methodsFor: 'initialize-release' stamp: 'bgf 3/12/2010 07:30'!
frame: frameRef oldState: oldVec newState: newVec description: descrip  

	primaryFrame := frameRef.
	oldState := oldVec.
	newState := newVec.
	description := descrip.
	self addRequiredFrame: frameRef.
! !

!QStateVectorAction methodsFor: 'initialize-release' stamp: 'bgf 3/12/2010 07:25'!
frame: frameRef oldState: oldVec newState: newVec description: descrip  requiredFrames: reqs

	primaryFrame := frameRef.
	oldState := oldVec.
	newState := newVec.
	description := descrip.
	self addRequiredFrame: frameRef.
	self addRequiredFrames: reqs.	! !

!QStateVectorAction methodsFor: 'actions' stamp: 'bgf 3/12/2010 07:26'!
redo: client

	primaryFrame future applyStateVector: newState.! !

!QStateVectorAction methodsFor: 'actions' stamp: 'bgf 3/12/2010 07:26'!
undo: client

	primaryFrame future applyStateVector: oldState.! !

!QUndoableAction class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUndoableAction methodsFor: 'accessing' stamp: 'bgf 3/5/2010 09:57'!
addRequiredFrame: aFarRef
	"The given far-ref refers to a TFrame that must be rooted in-space for this action to be applicable."
	requiredFrames := (self requiredFrames) copyWith: aFarRef! !

!QUndoableAction methodsFor: 'accessing' stamp: 'bgf 3/5/2010 18:00'!
addRequiredFrames: arrayOfFarRef
	"The given far-refs refer to a TFrames 
	 that must already be rooted in-space for this action to be available for undo/redo.   
	 These frames are not ones destroyed or created by the action...
	 since they must be present before the action can fire, 
	 as well as before the undo of the action can fire."
	requiredFrames := (self requiredFrames), arrayOfFarRef! !

!QUndoableAction methodsFor: 'testing' stamp: 'bgf 3/10/2010 13:26'!
canBePerformed
	requiredFrames isEmptyOrNil ifTrue: [ ^ true ].

	"We allow far-ref aliasing to keep far-refs alive between delete  and delete-undo;
	 so we cannot do 'Get' on these preserved far refs to ensure that they are rooted.
	 Instead we send the operation into the TIsland and have it look up the oids and verify them."

	^  (self requiredFrames allSatisfy: [ : ref | ref isValidFarRef   ] )
	    and: [
		requiredFrames first island 
			get: #allAreRootedFrames: 
			with: (requiredFrames collect: [ : ref | ref value refName ] ) ].

! !

!QUndoableAction methodsFor: 'accessing' stamp: 'bgf 3/12/2010 07:21'!
description
	"Word or very short phrase describing what this action changes ... e.g. 'Delete Frames' "
	^ description! !

!QUndoableAction methodsFor: 'actions' stamp: 'bgf 2/24/2010 14:31'!
redo: aClient
	self subclassResponsibility! !

!QUndoableAction methodsFor: 'accessing' stamp: 'bgf 3/5/2010 18:00'!
requiredFrames
	"TFrames far-refs
	 that must already be rooted in-space for this action to be available for undo/redo.   
	 These frames are not ones destroyed or created by the action...
	 since they must be present before the action can fire, 
	 as well as before the undo of the action can fire."

	^ requiredFrames ifNil: [ #() ]! !

!QUndoableAction methodsFor: 'actions' stamp: 'bgf 2/24/2010 14:30'!
undo: aClient
	self subclassResponsibility! !

Object subclass: #QUserFaceCache
	instanceVariableNames: 'tinyFaces normalFaces colors owner'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QUserFaceCache commentStamp: '<historical>' prior: 0!
A cache of user faces drawn from the _small_picture resources of users we encounter.
(And the user's avatar-color preference, as gleaned from avatar contact.)

Keeps two flavors of each cached entry
	- normal size (the same as QUserPreferences creates when the user gives a picture)
	- tiny size (32x32 opaque icon, for use as a list tag)

The cache can be checked/loaded from a userLogin string, or from an avatar far-ref;
the avatar far-ref path is better since it piggybacks on normal texture sharing.

The keys in the  dictionaries are login-strings.  
An empty (not-found) marker entry is created by an unsatisfied lookup,
so asking twice won't query the service twice.
use discard:login or discardUnsatisfied to 'flush' those not-found entries.
!

!QUserFaceCache class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QUserFaceCache class methodsFor: 'forms' stamp: 'bgf 8/28/2008 18:32'!
asTinyFace: aForm into: aTinyFormOrNil

	"Generate a square-cropped, scaled-down copy of the given form,
	 composited atop an opaque base-color form 
	 (transparency in the picture looks bad on translucent UI panels).
	 Note that we always cache the result of this operation."

	| size picture copy |

	aForm ifNil: [ ^ nil ].
	size := aForm extent.
	picture := aForm.

	"Extract a square."
	size x > size y ifTrue: [
		picture := picture contentsOfArea: 
						(((size x - size y / 2) rounded @0) extent: (size y@size y)).
	].
	size x < size y ifTrue: [
		picture := picture contentsOfArea: 
						(0@((size y - size x / 2)rounded) extent: (size x@size x)).
	].

	"Scale and composite onto the final result."
	copy := aTinyFormOrNil ifNil: [ Form extent: (self tinyFaceSize) depth: 32 ].
	copy fill: copy boundingBox fillColor: Color paleTan.
	(WarpBlt current toForm: copy)	"display scaled, using blend mode..."
			 sourceForm: picture destRect: copy boundingBox;
			combinationRule: Form blend;
			cellSize: 2;
			warpBits.

	^ copy
! !

!QUserFaceCache class methodsFor: 'forms' stamp: 'bgf 9/10/2008 22:02'!
defaultFaceShas
	"This is the sha of the default smiley-face,
	 which we don't want to keep as a cached face
	 so we know the difference."

	^{ 'PDEDROMTRWUHCA78XORMS9KTDY8U586' }! !

!QUserFaceCache class methodsFor: 'instance creation' stamp: 'bgf 8/28/2008 13:45'!
forOwner: aParticipant
	"There is one user-face-cache for the participant UI."
	^ (self new)
		owner: aParticipant;
		yourself! !

!QUserFaceCache class methodsFor: 'forms' stamp: 'bgf 8/28/2008 14:15'!
tinyFaceSize
	"Pixel dimensions of the tiny-face images."
	^ 32@32! !

!QUserFaceCache methodsFor: 'private' stamp: 'bgf 8/28/2008 22:54'!
cacheColor: aColor forLogin: aString
	"Private path for seeding my own color at prefs setup time."
	colors at: aString put: aColor! !

!QUserFaceCache methodsFor: 'private' stamp: 'bgf 8/28/2008 18:34'!
cacheFace: normalFaceForm forLogin: aUserLogin
	"Stow the given face as the normal face for that login,
	 and make a tiny variant.
	 We recycle the tiny-face form if possible,
	 since that makes live ui that much easier."
	normalFaces at: aUserLogin put: normalFaceForm.
	tinyFaces at: aUserLogin
			  put: (self class asTinyFace: normalFaceForm
							into: (tinyFaces at: aUserLogin ifAbsent: [ nil ])).
! !

!QUserFaceCache methodsFor: 'populating' stamp: 'bgf 2/14/2009 13:58'!
cacheFacesFromActiveSpace
	"Sent when we enter a new space...
	 examine the users there to get their face-pictures into the cache."
	| space |
	owner ifNil: [ ^ self ].
	space := owner activeSpace.

	space isValidFarRef ifTrue: [
		"Snarf up all the avatar graphics to see if anyone's new."
		(space get:#allUsers) ifNotNilDo: [ :array | array do: [ : avFar |
			self cacheGraphics: avFar
		]]].! !

!QUserFaceCache methodsFor: 'populating' stamp: 'bgf 8/28/2008 18:28'!
cacheGraphics: anAvatar
	"If we don't already have it,
	 store the face picture for this avatar, rescaled to a small icon,
	 and store the avatar color."
	| login |
	(anAvatar notNil and: [anAvatar isResolved]) ifFalse: [ ^ self ].

	login :=  (anAvatar get:#userLogin).
	login ifNil: [ ^ nil ].

	normalFaces at: login ifPresent: [ :face | ^ face ].

	^ self cacheGraphics: anAvatar userLogin: login.
! !

!QUserFaceCache methodsFor: 'populating' stamp: 'bgf 9/10/2008 22:11'!
cacheGraphics: anAvatar userLogin: uLogin
	"Store the face picture for this avatar, rescaled to a small icon."
	| picture login sha texture |
	
	(anAvatar notNil and: [anAvatar isResolved]) ifFalse: [ ^ nil ].
	login :=  uLogin ifNil: [ anAvatar get:#userLogin ].
	login ifNil: [ ^ nil ].

	"Try to get it from the avatar first, since we may already have the texture cached."
	normalFaces at: login ifAbsent: [
		[
			"First get it from the avatar"
			sha := anAvatar get:#smallPictureSHA.
			(self class defaultFaceShas includes: (sha)) ifFalse: [
				texture := TContentManager default lookup: sha.
				texture ifNotNil: [ picture := texture form ].
			].
		] ifError: [:msg|
			picture := nil.
		].
		picture ifNil: [
			picture := self fetchPictureForLogin: login.
		].
		picture ifNotNil: [
			self cacheFace: picture forLogin: login
		].
	].

	"While we're here, grab the color"
	(anAvatar get:#avatarColor) ifNotNilDo: [ :col |
		colors at: login put: col asColor
	].

	^ normalFaces at: login ifAbsent: [ nil ].! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 22:19'!
colorForAvatar: anAvatar
	"The given user's avatar color, which we cache under the login name as a side effect."
	| login clr |
	login := self loginFromAvatarRef: anAvatar.
	login ifNil: [ ^ anAvatar get: #avatarColor ].

	"Got the avatar, may as well  grab the current color from it."
	clr := (anAvatar get: #avatarColor).
	clr ifNotNil: [
		^ colors at: login put: clr.
	].
	"Maybe we had an old one..."
	^ colors at: login ifAbsent: [ nil ]! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 21:41'!
colorForLogin: userLogin
	^ colors at: userLogin ifAbsent: [ nil ].! !

!QUserFaceCache methodsFor: 'populating' stamp: 'bgf 8/28/2008 18:29'!
discard: userLogin
	"Uncache the data for the given user."
	normalFaces removeKey: userLogin ifAbsent: [].
	tinyFaces removeKey: userLogin ifAbsent: [].
! !

!QUserFaceCache methodsFor: 'populating' stamp: 'bgf 8/28/2008 18:39'!
discardUnsatisfied

	"Flush the cache of the 'don't know' entries we've accumulated,
	 so we'll try again when next someone asks."

	| unsatisfied |
	unsatisfied := OrderedCollection new.
	normalFaces keysAndValuesDo: [ : login : form |
		form ifNil: [ unsatisfied add: login ].
	].
	unsatisfied do: [ : login | self discard: login ].
! !

!QUserFaceCache methodsFor: 'private' stamp: 'bgf 1/22/2009 12:36'!
fetchPictureForLogin: aLogin
	"Get the picture for the given user-login from the service provider."
	| data |
	owner serviceMgr ifNil: [ ^ nil ].
	[ 
		data :=	owner serviceMgr getFile: (QUserPreferences faceFilePath: aLogin).
		data ifNotNil: [
			^ (ImageReadWriter formFromStream: (ReadStream on: data)) 
		] ifNil: [ 
			^ nil
		]
	] on: Error do: [ : ex |
		(ex description includesSubString: 'FileDoesNotExistException') ifFalse: [
			QLogger logInfo: 'No face file for user ', aLogin, ': ', (ex description).
		].
		^ nil
	].
! !

!QUserFaceCache methodsFor: 'initialize-release' stamp: 'bgf 8/28/2008 21:50'!
initialize
	super initialize.
	normalFaces := Dictionary new.
	tinyFaces := Dictionary new.
	colors := Dictionary new.! !

!QUserFaceCache methodsFor: 'private' stamp: 'jcg 8/30/2008 00:14'!
loginFromAvatarRef: anAvatarRef
	"It's not paranoia - far refs are out to get you."
	anAvatarRef isValidFarRef ifFalse: [ ^ nil ].
	^ anAvatarRef get: #userLogin.
	
! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 20:51'!
normalFaceForAvatar: anAvatarRef
	"Find the face for the given avatar; return the normal sized form (as saved by user-preferences).
	 Note this triggers lookup / provider fetching if the face hasn't been checked yet."
	| login |
	login := self loginFromAvatarRef: anAvatarRef.
	login ifNil: [ ^ nil ].


	^ self normalFaceForLogin: login ifAbsent: [ 
			self cacheGraphics: anAvatarRef userLogin: login.
			^ normalFaces at: login ifAbsent: nil 
		].
! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 20:51'!
normalFaceForAvatar: anAvatarRef ifAbsent: aBlock
	"Find the face for the given user; return the larger-sized form if we have it.
	 Return the block result if we have no cached value."
	| login |
	login := self loginFromAvatarRef: anAvatarRef.
	login ifNil: [ ^ nil ].

	^ self normalFaceForLogin: login ifAbsent: aBlock! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 20:51'!
normalFaceForLogin: aLoginString
	"Find the face for the given user; return the normal sized form (as saved by user-preferences).
	 Note this triggers lookup / provider fetching if the face hasn't been checked yet."
	^ normalFaces at: aLoginString ifAbsent: [
			self cacheFace: (self fetchPictureForLogin: aLoginString) 
			      forLogin: aLoginString.
			normalFaces at: aLoginString
	   ].! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 20:51'!
normalFaceForLogin: aLoginString ifAbsent: aBlock
	"Find the face for the given user; return the larger-sized form if we have it.
	 Return the block result if we have no cached value."
	^ normalFaces at: aLoginString ifAbsent: aBlock! !

!QUserFaceCache methodsFor: 'initialize-release' stamp: 'bgf 8/28/2008 13:44'!
owner: aParticipant
	owner := aParticipant! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 20:50'!
tinyFaceForAvatar: anAvatarRef
	"Find the face for the given avatar; return the tiny form (32x32 icon).
	 Note this triggers lookup / provider fetching if the face hasn't been checked yet."
	| login |
	login := self loginFromAvatarRef: anAvatarRef.
	login ifNil: [ ^ nil ].

	^ self tinyFaceForLogin: login ifAbsent: [ 
			self cacheGraphics: anAvatarRef userLogin: login.
			^ tinyFaces at: login ifAbsent: nil 
		].! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 20:51'!
tinyFaceForAvatar: anAvatarRef ifAbsent: aBlock
	"Find the face for the given user; return the icon-sized form if we have it.
	 Return the block result if we have no cached value."
	| login |
	login := self loginFromAvatarRef: anAvatarRef.
	login ifNil: [ ^ nil ].

	^ self tinyFaceForLogin: login ifAbsent: aBlock! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 18:26'!
tinyFaceForLogin: aLoginString

	"Find the face for the given user; return the normal sized form (as saved by user-preferences).
	 Note this triggers lookup / provider fetching if the face hasn't been checked yet."

	^ tinyFaces at: aLoginString ifAbsent: [
			self cacheFace: (self fetchPictureForLogin: aLoginString) 
			      forLogin: aLoginString.
			tinyFaces at: aLoginString
	   ].! !

!QUserFaceCache methodsFor: 'cache-lookup' stamp: 'bgf 8/28/2008 18:27'!
tinyFaceForLogin: aLoginString ifAbsent: aBlock

	"Find the face for the given user; return the icon-sized form if we have it.
	 Return the block result if we have no cached value."

	^ tinyFaces at: aLoginString ifAbsent: aBlock! !

Object subclass: #QwaqSanityChecker
	instanceVariableNames: 'ui speakerChecker suspended'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QwaqSanityChecker commentStamp: '<historical>' prior: 0!
Checks for bad conditions that we occasionally get into, and either fix or log them.!

!QwaqSanityChecker class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QwaqSanityChecker methodsFor: 'tests' stamp: 'jcg 3/11/2010 21:31'!
areAllSpeakersInOtherRoomsInaudible
	"Answer true if all QSpeakers in other rooms are inaudible (as they should be), and false otherwise."

	"This runs in a background thread so we have to be conservative about 'activeSpace' and whether it is up to date."
	| myActiveSpace clientSpeakers activeIsland | 
	ui audioSession ifNil: [ ^ true ].
	ui readyForEvents ifFalse: [ ^ true ].
	(((myActiveSpace := ui activeSpace) isValidFarRef) and: [myActiveSpace isResolved])
		ifTrue: [activeIsland := myActiveSpace island]
		ifFalse: [^true "cannot check right now"].

	"That get: crossed a mutex, so make sure we're still kosher..."
	ui audioSession ifNil: [ ^ true ].
	ui readyForEvents ifFalse: [ ^ true ].
	[clientSpeakers := ui audioSession speakerManager speakers values] 
		on: Error 
		do: [
			QLogger logWarning: 'failed to get list of client-speaker refs (shutting down?)'.
			^true "we're probably shutting down"
		].
	
	"Only care about client-speakers referring to speakers in the current island."
	clientSpeakers := clientSpeakers select: [:cs | cs ref island ~~ activeIsland].
	(clientSpeakers allSatisfy: [:cs | cs isAudible not]) ifTrue: [ ^ true ].

	"We 'failed' the test, but it might be because we teleported."
	ui activeSpace = myActiveSpace ifFalse: [ ^ true  "Ignore transitional result" ].
	QLogger logWarning: 'Failed speaker audibility test'.
	clientSpeakers do: [:cs |
		cs isAudible ifFalse: [
			| streamID parentName msg |
			streamID := cs ref get: #streamID.
			parentName := cs ref get: #parentObjectName.
			msg := '        audible speaker in other room: ', parentName.
			streamID ifNotNil: [msg := msg, '  stream-id: ', streamID hex].
			QLogger logWarning: msg.
		].
	].
	^ false
! !

!QwaqSanityChecker methodsFor: 'initialize' stamp: 'jcg 4/9/2007 11:20'!
destroy
	self stopSpeakerChecker! !

!QwaqSanityChecker methodsFor: 'tests' stamp: 'jcg 3/11/2010 15:13'!
doAllSpeakersHaveACorrespondingClientSpeaker
	"Answer true if all QSpeakers in the current space have a corresponding QClientSpeaker outside of the island, and false otherwise."

	"This runs in a background thread so we have to be conservative about 'activeSpace' and whether it is up to date."
	| myActiveSpace islandSpeakers clientSpeakers unmatchedSpeakers activeIsland | 

	ui audioSession ifNil: [ ^ true ].
	ui readyForEvents ifFalse: [ ^ true ].
	(((myActiveSpace := ui activeSpace) isValidFarRef) and: [myActiveSpace isResolved])
		ifTrue: [activeIsland := myActiveSpace island]
		ifFalse: [^true "cannot check right now"].
	islandSpeakers := activeIsland get: #resourcesAt: with: QSpeaker resourceType.

	"That get: crossed a mutex, so make sure we're still kosher..."
	ui audioSession ifNil: [ ^ true ].
	ui readyForEvents ifFalse: [ ^ true ].
	[clientSpeakers := 
		ui audioSession speakerManager speakers values collect: [:cs | cs ref]
	] on: Error do: [
		QLogger logWarning: 'failed to get list of client-speaker refs (shutting down?)'.
		^true "we're probably shutting down"
	].
	"Only care about client-speakers referring to speakers in the current island."
	clientSpeakers := clientSpeakers select: [:ref | ref island = activeIsland].

	"Don't want false-positives due to shadow far-refs 
	 (I don't think this is an issue here, but it's better to be sure)."
	islandSpeakers := islandSpeakers collect: [:ref | ref value]. 
	clientSpeakers := clientSpeakers collect: [:ref | ref value]. 

	unmatchedSpeakers := islandSpeakers difference: clientSpeakers.
	unmatchedSpeakers ifNotEmpty: [
		ui activeSpace = myActiveSpace ifFalse: [ ^ true  "Ignore transitional result" ].
		QLogger logWarning: 'FAILED TO MATCH ', unmatchedSpeakers size, '/', islandSpeakers size, ' SPEAKERS'.
		unmatchedSpeakers do: [:ref |
			| streamID parentName msg  |
			streamID := ref get: #streamID.
			parentName := ref get: #parentObjectName.
			msg := '        unmatched speaker: ', parentName.
			streamID ifNotNil: [msg := msg, '  stream-id: ', streamID hex].
			QLogger logWarning: msg
		].
		^false "uh-oh... couldn't match all the speakers"
	].
	^true "jolly good!!"
! !

!QwaqSanityChecker methodsFor: 'initialize' stamp: 'bgf 7/26/2007 19:00'!
initialize
	suspended := true.	"Until started."! !

!QwaqSanityChecker methodsFor: 'tests' stamp: 'jcg 3/11/2010 15:28'!
runAllTests
	"Answer true if all tests succeeded, and false otherwise."
	self doAllSpeakersHaveACorrespondingClientSpeaker ifFalse: [^false].
	self areAllSpeakersInOtherRoomsInaudible ifFalse: [^false].
	^true
! !

!QwaqSanityChecker methodsFor: 'watchers' stamp: 'jcg 3/11/2010 15:35'!
runSpeakerChecker
	"Ensure that each voice-chat speaker has a corresponding off-island QClientSpeaker.  If the state is insane, fix it by resettting the sound."
	| count |
	count := 0.
	[ui exitTripped] whileFalse: [
		(Delay forSeconds: 3) wait.
		ui exitTripped ifTrue: [ ^ self].
		ui mainView ifNil: [ ^ self ].
		(suspended or: [self runAllTests])
			ifTrue: [count := 0 "All is well"]
			ifFalse: [
				count := count + 1.
				count > 1 ifTrue: [
					"Try to fix the situation."
					ui signal: #resetSound with: 'failed speaker sanity check'.
					count := 0.
				]
			]
	].
	speakerChecker := nil. 	"Since I'm going away anyway, nothing to terminate."
	! !

!QwaqSanityChecker methodsFor: 'watchers' stamp: 'jcg 1/20/2009 14:40'!
startSpeakerChecker
	self stopSpeakerChecker.
	speakerChecker := self 
							forkSuspended: #runSpeakerChecker 
							at: Processor userSchedulingPriority.
	speakerChecker resume.
	self suspended: false.

! !

!QwaqSanityChecker methodsFor: 'watchers' stamp: 'jcg 4/9/2007 11:23'!
stopSpeakerChecker
	speakerChecker ifNotNil: [
		speakerChecker terminate.
		speakerChecker := nil
	]! !

!QwaqSanityChecker methodsFor: 'initialize' stamp: 'bgf 7/26/2007 19:05'!
suspended: bFlag

	"Suspending the checker (setting suspended:true) 
	 makes all the checks do nothing  until suspended:false."

	suspended := bFlag! !

!QwaqSanityChecker methodsFor: 'initialize' stamp: 'jcg 4/9/2007 11:19'!
ui: aQwaqParticipantUI
	ui := aQwaqParticipantUI.
! !

Object subclass: #QwaqVersion
	instanceVariableNames: 'isSealedApplication version relevantPackageNames highestUpdate sealedBuildString'
	classVariableNames: 'Current'
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QwaqVersion commentStamp: '<historical>' prior: 0!
A version identifier for Qwaq. Use as in:
	QwaqVersion current infoString.!

!QwaqVersion class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QwaqVersion class methodsFor: 'accessing' stamp: 'ar 11/6/2006 14:20'!
current
	^Current ifNil:[Current := self new]! !

!QwaqVersion class methodsFor: 'accessing' stamp: 'ar 11/6/2006 14:20'!
current: aQwaqVersion
	Current := aQwaqVersion! !

!QwaqVersion class methodsFor: 'class initialization' stamp: 'ar 3/20/2007 15:10'!
initialize	"QwaqVersion initialize"
	"Initialize the default version"
	self newVersion: 'Qwaq 1.0'.! !

!QwaqVersion class methodsFor: 'accessing' stamp: 'ar 4/16/2011 17:48'!
newVersion: versionString
	"
	 QwaqVersion newVersion: 'OpenQwaq 1.0.01'

	 Note that parts of the system expect the version string
	 to be a legal file name - don't get too creative."

	self current: (self new version: versionString)
! !

!QwaqVersion class methodsFor: 'utilities' stamp: 'ar 4/16/2011 17:48'!
parseInfoString: infoString 
	"Parse the 'infoString' (as created by 'QwaqVersion current versionString') into:
		- the version string (eg: 'OpenQwaq 1.0.01')
		- the update number
		- the build-string"
	| versionString updateNumber buildString ind1 ind2 |

	ind1 := infoString indexOfSubCollection: 'Update:'.
	ind1 = 0 ifTrue: [self error: 'QwaqVersion info-string does not contain "Update:"'].
	ind2 := infoString indexOfSubCollection: 'Build: '.
	ind2 = 0 ifTrue: [self error: 'QwaqVersion info-string does not contain "Build:"'].

	versionString := (infoString copyFrom: 1 to: ind1-1) withBlanksTrimmed.
	updateNumber := infoString copyFrom: ind1 + 'Update:' size to: ind2-1.
	updateNumber := updateNumber asInteger 
		ifNil: [self error: 'Could not read update-number from info-string'].
	buildString := infoString copyFrom: ind2 + 'Build:' size to: infoString size.
	buildString := buildString withBlanksTrimmed.

	^{versionString. updateNumber. buildString}! !

!QwaqVersion class methodsFor: 'accessing' stamp: 'eem 6/18/2009 09:13'!
sysProfile

	"A string summarizing the system...  image, vm, OS, whatever else we can scrape."
	^ String streamContents: [ :strm |

		strm nextPutAll: 'Software: ';
			 nextPutAll: QwaqVersion current infoString;
			 cr.
		strm nextPutAll: 'VM: ';
			nextPutAll:  Smalltalk platformName asString;
			nextPutAll:  ' (', Smalltalk osVersion asString, ')';
			nextPutAll: ' - ';
			nextPutAll: (Smalltalk getSystemAttribute: 1006) asString;
			cr.
		(Smalltalk getSystemAttribute: 1007) ifNotNil:
			[:interpreterVersion|
			strm nextPutAll: '    Interpreter: '; nextPutAll: interpreterVersion asString; cr].
		(Smalltalk getSystemAttribute: 1008) ifNotNil:
			[:cogitVersion|
			strm nextPutAll: '    Cogit: '; nextPutAll: cogitVersion asString; cr].
		
		"Additional info available from the windows VM"
		Smalltalk platformName = 'Win32' ifTrue:[
			10001 to: 10003 do:[:i| | info |
				info := (Smalltalk getSystemAttribute: i) ifNil:[''].
				info := info copyReplaceAll: String lf with: String cr.
				info isEmpty ifFalse:[strm cr; nextPutAll: info]]].

		strm cr.
	]! !

!QwaqVersion methodsFor: 'accessing' stamp: 'bgf 10/16/2008 15:12'!
buildString	"QwaqVersion current buildString"
	"Derive the info (version) string."
	| newest relevant modified da db |

	"If we are sealed, use the sealed build string"
	self isSealedApplication ifTrue:[^sealedBuildString].
	relevant := (MCWorkingCopy allManagers  select: [ :a | 
					self relevantPackageNames anySatisfy: [:rpn |
						a packageName includesSubString: rpn 
					]]).
	modified := (relevant anySatisfy: [ : wc | wc modified]).
	newest := (relevant asSortedCollection: [ :a :b |
				[
					da := a ancestors first date.	
					db := b ancestors first date.
					((da < db) or: [ da = db and: [  a ancestors first time <= b ancestors first time ]])
				] ifError: [true].
				"Occasionally a package will show up with no archived ancestor;
				 such a package is probably not loaded, and so goes early in the list so it is never last."
			]) last.

	"Make a numerical version string (that is actually the date)"
	^newest ancestors first date yyyymmdd, '-', 
			newest ancestors first time hhmm24, 
			(modified ifTrue: [ ' *'] ifFalse: [ '  ']).! !

!QwaqVersion methodsFor: 'updates' stamp: 'ar 6/26/2007 17:09'!
highestUpdate
	"Answer the highest update number installed in the system so far"
	^highestUpdate ifNil:[0]! !

!QwaqVersion methodsFor: 'updates' stamp: 'ar 6/26/2007 17:09'!
highestUpdate: aNumber
	"Set the highest update number installed in the system so far"
	highestUpdate := aNumber! !

!QwaqVersion methodsFor: 'accessing' stamp: 'ar 6/27/2007 10:16'!
infoString	"QwaqVersion current infoString"
	"Derive the info (version) string."
	^self version, ' Update: ', self highestUpdate, ' Build: ', self buildString! !

!QwaqVersion methodsFor: 'initialize' stamp: 'ar 6/26/2007 17:08'!
initialize
	super initialize.
	highestUpdate := 0.
	self relevantPackageNames: {'Croquet' . 'Qwaq'}.! !

!QwaqVersion methodsFor: 'testing' stamp: 'jdm 12/10/2007 06:48'!
isSealedApplication
	isSealedApplication ifNil:[^false].
	^isSealedApplication! !

!QwaqVersion methodsFor: 'accessing' stamp: 'bgf 9/8/2009 20:18'!
numericVersion
	"The tail portion of the version consisting of dots and digits.
	 Only for display purposes; not robust.
	 Returns  <self version> if the tail isn't dots&digits."
	| v i |

	v := self version.
	i := v size.
	[ i > 0 and: [ (v at: i) isDigit or: [ (v at: i) = $. ]]] whileTrue: [ i := i - 1 ].

	i < v size ifTrue: [
		^ v copyFrom: i+1 to: v size 
	].
	^ v! !

!QwaqVersion methodsFor: 'printing' stamp: 'ar 3/20/2007 15:11'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(; print: self infoString; nextPut: $).
! !

!QwaqVersion methodsFor: 'updates' stamp: 'ar 6/26/2007 17:09'!
registerUpdate: updateNum
	"Register an update with the given number, bumping the highest version if needed"
	updateNum ifNotNil:[
		highestUpdate := self highestUpdate max: updateNum.
	].! !

!QwaqVersion methodsFor: 'accessing' stamp: 'ar 11/6/2006 14:19'!
relevantPackageNames
	^relevantPackageNames! !

!QwaqVersion methodsFor: 'accessing' stamp: 'ar 11/6/2006 14:19'!
relevantPackageNames: aCollection
	relevantPackageNames := aCollection! !

!QwaqVersion methodsFor: 'initialize' stamp: 'bgf 12/10/2007 10:17'!
seal
	sealedBuildString := self buildString.
	isSealedApplication := true.
! !

!QwaqVersion methodsFor: 'initialize' stamp: 'bgf 12/10/2007 10:17'!
unseal
	isSealedApplication := false.! !

!QwaqVersion methodsFor: 'accessing' stamp: 'ar 3/20/2007 15:06'!
version
	^version ifNil:['Unknown']! !

!QwaqVersion methodsFor: 'accessing' stamp: 'ar 3/20/2007 15:06'!
version: aString
	version := aString.! !

CroquetEvent subclass: #QCroquetEvent
	instanceVariableNames: 'action'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QCroquetEvent class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCroquetEvent methodsFor: 'accessing' stamp: 'jjl 5/12/2008 11:41'!
action

	^action! !

!QCroquetEvent methodsFor: 'accessing' stamp: 'jjl 5/12/2008 11:42'!
action: anAction

	"The list of possible actions is given in QCroquetEvent class."
	action := anAction.! !

CPlayer subclass: #QControlBar
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QControlBar commentStamp: '<historical>' prior: 0!
The control strip at the bottom of the participant UI.
Comprises the pointer tool set, the media controls,
the messageArea (a label, handled by the QStatusLine message manager),
counters for applications and open forums,
the inventory button,
and a qwaq logo doubling as the position-readout access toggle.!

QControlBar defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="owner" toSet="owner:" changeEvent="ownerChanged">owner</field>
	<field toGet="view" toSet="view:" changeEvent="viewChanged">view</field>
	<field toGet="cursorMode" toSet="cursorMode:" changeEvent="cursorModeChanged">cursorMode</field>
	<field toGet="thumbtack" toSet="thumbtack:" changeEvent="thumbtackChanged">thumbtack</field>
	<field toGet="editBox" toSet="editBox:" changeEvent="editBoxChanged">editBox</field>
	<field toGet="askToSpeak" toSet="askToSpeak:" changeEvent="askToSpeakChanged">askToSpeak</field>
	<field toGet="microphone" toSet="microphone:" changeEvent="microphoneChanged">microphone</field>
	<field toGet="speaker" toSet="speaker:" changeEvent="speakerChanged">speaker</field>
	<field toGet="camera" toSet="camera:" changeEvent="cameraChanged">camera</field>
	<field toGet="sitStand" toSet="sitStand:" changeEvent="sitStandChanged">sitStand</field>
	<field toGet="readoutButton" toSet="readoutButton:" changeEvent="readoutButtonChanged">readoutButton</field>
	<field toGet="inventory" toSet="inventory:" changeEvent="inventoryChanged">inventory</field>
	<field toGet="svcCount" toSet="svcCount:" changeEvent="svcCountChanged">svcCount</field>
	<field toGet="svcCountLabel" toSet="svcCountLabel:" changeEvent="svcCountLabelChanged">svcCountLabel</field>
	<field toGet="appCount" toSet="appCount:" changeEvent="appCountChanged">appCount</field>
	<field toGet="appCountLabel" toSet="appCountLabel:" changeEvent="appCountLabelChanged">appCountLabel</field>
	<field toGet="netHealth" toSet="netHealth:" changeEvent="netHealthChanged">netHealth</field>
	<field toGet="renderHealth" toSet="renderHealth:" changeEvent="renderHealthChanged">renderHealth</field>
	<field toGet="messageArea" toSet="messageArea:" changeEvent="messageAreaChanged">messageArea</field>
	<field toGet="editIndicator" toSet="editIndicator:" changeEvent="editIndicatorChanged">editIndicator</field>
	<field toGet="ioIcon" toSet="ioIcon:" changeEvent="ioIconChanged">ioIcon</field>
	<field toGet="buttonDownTime" toSet="buttonDownTime:" changeEvent="buttonDownTimeChanged">buttonDownTime</field>
	<field toGet="awayButton" toSet="awayButton:" changeEvent="awayButtonChanged">awayButton</field>
	<field toGet="soundDialog" toSet="soundDialog:" changeEvent="soundDialogChanged">soundDialog</field>
	<field toGet="userIsIdle" toSet="userIsIdle:" changeEvent="userIsIdleChanged">userIsIdle</field>
	<field toGet="dividers" toSet="dividers:" changeEvent="dividersChanged">dividers</field>
</fields>'!

!QControlBar class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/3/2009 10:16'!
adjustSound
	"Open the sound settings dialog."
	self assuredSoundDialog openCenteredAt: (owner dialogCenter).	
	QActivityCapture action: #ShowFromToolsMenu details: 'Sound Dialog'.! !

!QControlBar methodsFor: 'private' stamp: 'jcg 5/4/2010 14:08'!
allMicrophoneButtonsDo: aBlock
	"Evaluate aBlock for all of the microphone buttons throughout the system.  There are currently two: one in the control-bar and one in the pop-up sound-dialog."
	aBlock value: microphone.
	soundDialog ifNotNil: [:sd | aBlock value: sd microphone].
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 10/12/2007 14:07'!
appCount
	"Answer the appCount of the receiver"
	<bewareOf: #appCountChanged>
	^self propertyValueAt: #appCount! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 10/12/2007 14:07'!
appCount: aValue
	"Modify the receiver's appCount"
	^self propertyValueAt: #appCount put: aValue with: #appCountChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 9/8/2009 16:12'!
appCountLabel
	"Answer the appCountLabel of the receiver"
	<bewareOf: #appCountLabelChanged>
	^self propertyValueAt: #appCountLabel! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 9/8/2009 16:12'!
appCountLabel: aValue
	"Modify the receiver's appCountLabel"
	^self propertyValueAt: #appCountLabel put: aValue with: #appCountLabelChanged! !

!QControlBar methodsFor: 'initialize' stamp: 'bgf 8/28/2009 09:19'!
applyUserRole
	"Adjust control visibility and layout for the current user's capabilities."
	| markup |

	owner ifNil: [  ^ self disable ].		"No controls but the disabled ones"
	locked ifTrue: [ ^ self ].				"Already disabled... stay that way."

	markup := (owner canModifyForum).
	thumbtack enabled: markup.
	inventory enabled: markup.
	markup ifFalse: [
		"Make sure they're not open, or in active-state."
		thumbtack selected: false.
		inventory selected: false.		"Closes inventories when meeting-leader restricts change."
	].

	"The edit-mode indicators do not appear at all unless you're editing."
	markup := (owner canModifyForum and: [editIndicator visible]).
	editBox enabled: markup; visible: markup.

	"Everyone has webcam support, but it is watermarked unless you pay."
	camera visible: true.
	self setupAskToSpeakButton.	"Also lays out the buttons."
	! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 7/5/2009 19:38'!
askToSpeak
	"Answer the askToSpeak of the receiver"
	<bewareOf: #askToSpeakChanged>
	^self propertyValueAt: #askToSpeak! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 7/5/2009 19:38'!
askToSpeak: aValue
	"Modify the receiver's askToSpeak"
	^self propertyValueAt: #askToSpeak put: aValue with: #askToSpeakChanged! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 7/5/2009 19:40'!
askToSpeakTipString
	| audiencemute youcan |
	owner isMeetingInProgress ifFalse: [ ^ '' ].

	audiencemute := owner meetingState audienceIsMuted.
	youcan := (owner meetingState isLoginAllowedToSpeak: (owner userLogin)).

	^  (owner isMeetingLeader ifTrue: [
		audiencemute 
			ifTrue: ['Click to permit voice chat from the audience.' ] 
			ifFalse: ['Click to stop voice chat from the audience.'. ].
	  ] ifFalse: [
		(audiencemute & youcan) ifTrue: [
			'Click to stop speaking.'
		] ifFalse: [
			askToSpeak selected
				ifTrue: [	'Click if you no longer want to speak.' ] 		
				ifFalse: [
					audiencemute
						ifTrue: [ 'Click to ask to speak.' ] 
						ifFalse: [ 'Click to ask for attention.' ]
				]]]).! !

!QControlBar methodsFor: 'accessing' stamp: 'hrs 11/11/2008 11:56'!
assuredSoundDialog
	^soundDialog ifNil: [
		soundDialog := QSoundDialog new.
		soundDialog owner: owner.
		soundDialog
	].
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/2/2008 01:23'!
awayButton
	"Answer the awayButton of the receiver"
	<bewareOf: #awayButtonChanged>
	^self propertyValueAt: #awayButton! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/2/2008 01:23'!
awayButton: aValue
	"Modify the receiver's awayButton"
	^self propertyValueAt: #awayButton put: aValue with: #awayButtonChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/20/2008 03:04'!
buttonDownTime
	"Answer the buttonDownTime of the receiver"
	<bewareOf: #buttonDownTimeChanged>
	^self propertyValueAt: #buttonDownTime! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/20/2008 03:04'!
buttonDownTime: aValue
	"Modify the receiver's buttonDownTime"
	^self propertyValueAt: #buttonDownTime put: aValue with: #buttonDownTimeChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'jcg 8/7/2007 18:11'!
camera
	"Answer the camera of the receiver"
	<bewareOf: #cameraChanged>
	^self propertyValueAt: #camera! !

!QControlBar methodsFor: 'accessing' stamp: 'jcg 8/7/2007 18:11'!
camera: aValue
	"Modify the receiver's camera"
	^self propertyValueAt: #camera put: aValue with: #cameraChanged! !

!QControlBar methodsFor: 'private' stamp: 'bgf 5/25/2010 16:43'!
cameraMenu
	"Post the press-hold menu for the web-camera button."
	| menu |
	owner videoSession ifNotNilDo: [ :sess |
		menu := sess qualityMenu.
		"Redirect the menu back to me,
		 since I want it to mean set-quality and then start-camera."
		menu menuItems do: [ : item |
			item action: (MessageSend receiver: self 
									    selector: #startCameraAtQuality: 
									    arguments: (item action arguments)).
			item emphasis: ((view owner preferences videoQuality = item label) ifTrue: [1] ifFalse: [0]).
			item label: item label, ' Quality'.
			item checked: (item checked and: [camera selected]).
		].
		menu menuItems addFirst: ((CMenuItem new)
						label: 'Off';
						action: (MessageSend receiver: self selector: #stopVideo);
						checked: (camera selected not);
						yourself).

		(sess deviceMenu) ifNotNilDo: [ : sub |
			sub menuItems size caseOf: {
			[0] -> [
				" No cameras - turn off the resolution-choice menus altogether" 
				menu menuItems do: [ : item | item disable ]].
			[1]-> [
				" Use the camera name as a disabled menu item. "
				menu menuItems addFirst: ((CMenuItem new)
											label: (sub menuItems first label);
											action: (MessageSend receiver: self selector: #yourself);
											checked:  true;
											enabled: false;
											yourself). 
				menu menuItems first separator: true. ]
			} otherwise: [
				" A choice of cameras - cascade to a submenu. "
				menu menuItems addFirst: ( (CMenuItem new) label: 'Camera'; subMenu: sub; yourself ).
				menu menuItems first separator: true.
				sub menuItems do: [ : item |
						item action: (MessageSend receiver: self 
									    selector: #setVideoDeviceName:deviceUID: 
									    arguments: (item action arguments)).
				]].
		] ifNil: [
			"No indicator at top - we're limited to using the platform default."
		].
	
		"A way to restart seeing web cams (i.e. if  disabled due to delays or in preferences.)"
		sess webcastManager videoDisabled ifTrue: [
			menu addSeparator.
			menu add: 'Resume Video Streaming'
				   target: self
				   action: #checkVideoSession.
		].

		QStyles installMenuFontInMenu: menu.
		menu fullBounds.
		menu openAt: (camera center + self topLeft - (0@(menu extent y*2))).
	].! !

!QControlBar methodsFor: 'events' stamp: 'jcg 4/8/2009 23:49'!
checkVideoSession
	"Get the video session running if permitted.
	 Return true if the owner's video session is already on or we turn it back on in here.
	 If the session is disabled, ask for permission to restart it.
	 (The preference may be off, 
	 or there may have been a session-stop due to network lag);
	 owner handles both cases."

	| msg |
	owner isVideoSessionEnabled ifFalse: [

		"May be disabled because we said so in prefs, or because checkAvatarLag 
		 in the participant UI shut it off to save traffic."

		owner preferences videoDisabled ifTrue: [
			msg := 'You have disabled video streaming in your preferences.\Enable it now?' withCRs.
		] ifFalse: [
			msg := 'Video streaming has been disabled to conserve bandwidth.\Re-enable it now?' withCRs.
		].

		(CDialog confirm: msg centerAt: owner center)
			ifFalse: [ ^ false ]
			ifTrue: [
				owner enableVideoSession.
				owner statusInfo: 'Video streaming enabled.'.
			]].
	^ true! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 11:37'!
cursorMode
	"Answer the cursorMode of the receiver"
	<bewareOf: #cursorModeChanged>
	^self propertyValueAt: #cursorMode! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 11:37'!
cursorMode: aValue
	"Modify the receiver's cursorMode"
	^self propertyValueAt: #cursorMode put: aValue with: #cursorModeChanged! !

!QControlBar methodsFor: 'selecting' stamp: 'bgf 1/4/2009 21:32'!
disable
	| highlight |

	"Override the default behavior for a gentler gray,
	 and hide all the icons except the status message and logo.
	 This is the posture during login, before the session is going."

	highlight := CFillingHighlight new.
	highlight name: #disableHighlight.
	highlight fill: (Color black alpha: 0.025).
	self costume addHighlight: highlight.
	self locked: true.
	self elements do: [:e| ({messageArea . readoutButton . ioIcon } includes: e)
							ifFalse: [e visible: false ]].

	self remove: ioIcon.	"It's misplaced, and we don't want its triggers to show it."

	soundDialog ifNotNil: [ soundDialog close ].
	self layoutButtons.! !

!QControlBar methodsFor: 'private' stamp: 'bgf 11/3/2007 13:02'!
disableSound

	speaker selected: false.
	speaker disable.
	microphone selected: false.
	microphone disable.

	false ifTrue: [
		speaker balloonHelp: 'Sound is disabled because OpenAL could not be  loaded.'.
		microphone balloonHelp: 'Sound is disabled because OpenAL could not be loaded.'.
	].! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/3/2008 12:25'!
dividers
	"Answer the dividers of the receiver"
	<bewareOf: #dividersChanged>
	^self propertyValueAt: #dividers! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/3/2008 12:25'!
dividers: aValue
	"Modify the receiver's dividers"
	^self propertyValueAt: #dividers put: aValue with: #dividersChanged! !

!QControlBar methodsFor: 'events' stamp: 'bgf 8/17/2009 19:08'!
dragOutAStickyNote
	"Dragging directly from the sticky-note icon into the space;
	 an easier creation path than click-clicking."
	| dragLabel offset nextPos delta |
	owner readyForEvents ifFalse: [ ^ false ].

	dragLabel := QDocumentDragIcon new.
	dragLabel plainIcon: (QForms find: 'StickyNoteIcon.png').
	dragLabel dropShadowEnabled: false.
	dragLabel addToWorld: world.

	offset := dragLabel cursorPoint.
	hand forEach: #mouseMove do:[
		nextPos := dragLabel cursorPoint.
		delta := nextPos - offset.
		dragLabel transform: ((COffsetTransform withOffset: delta) 
			transformedBy: dragLabel transform).
	] until: #mouseUp.
	
	"Got the final position... still on track for doing a note-insert?"
	(owner readyForEvents and: [owner isPointInMainView: (hand position)]) 
	 ifFalse: [
		dragLabel dragReject.
		^ false
	].

	"Send the actual annotate request  through the owner."
	owner viewPanel pickTarget.
	(owner addAnnotation: QStickyNote asDrop: true) ifTrue: [
		dragLabel dragAccept.
	] ifFalse: [
		dragLabel dragReject.
	].
! !

!QControlBar methodsFor: 'accessing' stamp: 'das 12/22/2006 23:29'!
editBox
	"Answer the editBox of the receiver"
	<bewareOf: #editBoxChanged>
	^self propertyValueAt: #editBox! !

!QControlBar methodsFor: 'accessing' stamp: 'das 12/22/2006 23:29'!
editBox: aValue
	"Modify the receiver's editBox"
	^self propertyValueAt: #editBox put: aValue with: #editBoxChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 6/3/2009 00:30'!
editIndicator
	"Answer the editIndicator of the receiver"
	<bewareOf: #editIndicatorChanged>
	^self propertyValueAt: #editIndicator! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 6/3/2009 00:30'!
editIndicator: aValue
	"Modify the receiver's editIndicator"
	^self propertyValueAt: #editIndicator put: aValue with: #editIndicatorChanged! !

!QControlBar methodsFor: 'selecting' stamp: 'bgf 6/3/2009 00:30'!
enable

	"See disable; this is its complement."
	self costume removeHighlight: #disableHighlight.
	self elements do: [:e| ( {ioIcon . editIndicator } includes: e) ifFalse: [e visible: true]].

	self locked: false.
	self add: ioIcon.

	self applyUserRole.! !

!QControlBar methodsFor: 'private' stamp: 'bgf 11/3/2007 13:00'!
enableSound

	"Restore the microphone and speaker buttons to their usual default state."

	speaker selected: true.
	speaker enable.
	microphone selected: false.
	microphone enable.

	false ifTrue: [
		speaker balloonHelp: 'Turn sound on or off..'.
		microphone balloonHelp: 'Activate your microphone so you can speak to others in the space.'.
	].! !

!QControlBar methodsFor: 'initialize' stamp: 'ar 4/16/2011 17:10'!
initialize
	super initialize.
	self define: #askToSpeak as: CImageButton new.
	self define: #thumbtack as: CImageButton new.
	self define: #editBox as: CImageButton new.

	self define: #awayButton as: CImageButton new.

	self define: #microphone as: QMicrophoneMenuButton new.
	self define: #speaker as: QSpeakerMenuButton new.
	self define: #camera as: CImageButton new.
	self define: #sitStand as: CImageButton new.

	self define: #readoutButton as: CImageButton new.
	self define: #inventory as: CImageButton new.
	self define: #messageArea as: CLabel new.
	self define: #editIndicator as: CLabel new.

	self define: #svcCountLabel as: (CLabel on:('Forums:')).
	self define: #appCountLabel as: (CLabel on:('TMS Apps:')).
	self define: #svcCount as: CLabel new.
	self define: #appCount as: CLabel new.
	self define: #netHealth as: CLabel new.
	self define: #renderHealth as: CLabel new.

	self define: #ioIcon as: CPlayer new.
	self define: #dividers as: OrderedCollection new.
	self define: #cursorMode as: #arrow.

	userIsIdle := false.! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 4/28/2007 14:14'!
inventory
	"Answer the inventory of the receiver"
	<bewareOf: #inventoryChanged>
	^self propertyValueAt: #inventory! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 4/28/2007 14:14'!
inventory: aValue
	"Modify the receiver's inventory"
	^self propertyValueAt: #inventory put: aValue with: #inventoryChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 7/23/2007 14:04'!
ioIcon
	"Answer the ioIcon of the receiver"
	<bewareOf: #ioIconChanged>
	^self propertyValueAt: #ioIcon! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 7/23/2007 14:04'!
ioIcon: aValue
	"Modify the receiver's ioIcon"
	^self propertyValueAt: #ioIcon put: aValue with: #ioIconChanged! !

!QControlBar methodsFor: 'private' stamp: 'hrs 12/8/2008 16:31'!
isPTT
	"Are we using push to talk"

	^ owner notNil 
	   and: [ owner audioSession notNil ] 
	   and: [ owner audioSession isPressToTalk ]! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/9/2009 23:34'!
isSeatNearby: bool
	"An update from the avatar about whether there is a free seat nearby."
	sitStand normalImage: (bool ifTrue: [ QForms find: 'SitDownIcon.png' ]
							     ifFalse: [ QForms find: 'SitDownNoneIcon.png' ]).
	sitStand overImage: (bool ifTrue: [ QForms find: 'SitDownOver.png' ]
	 						  ifFalse: [ QForms find: 'SitDownNoneIcon.png' ]).! !

!QControlBar methodsFor: 'private' stamp: 'bgf 4/1/2010 18:07'!
keyDrivePlayers
	"Would like to just say 'elements' but most aren't added as children until we go to full-session."

	^ { askToSpeak . thumbtack . editBox . 
	      awayButton .
	      microphone . speaker . camera . sitStand .
	      readoutButton .
	      inventory . messageArea . editIndicator .
	      svcCountLabel . appCountLabel .
		svcCount . appCount .
		netHealth . renderHealth .
		ioIcon .
		}, 
	dividers
! !

!QControlBar methodsFor: 'initialize' stamp: 'ar 5/25/2011 19:56'!
layoutButtons
	"Layout the control bar.  This done various times during a session,
	 to adjust the messageArea width depending on whether we're logged in or not,
	 and also to update the controls after login to make sure restricted features are checked."
	| mid curX |
	mid := 17.

"Left-hand groups"
	readoutButton topLeft: (2@1).	"OpenQwaq logo"

	curX := readoutButton right - 4.
	awayButton topLeft: ((curX + 13)@(mid-13)).
	dividers first topLeft: (curX+52 @1).
	curX := curX + 57.

	microphone topLeft: curX@(mid-15).
	camera visible ifTrue: [
 		"Camera is restricted, either x-features or forums.properties"
		camera topLeft: curX + 29 @(mid-17).
		speaker topLeft: curX + 65 @(mid-15). 
	] ifFalse: [
		speaker topLeft: curX + 30 @(mid-15)
	].
	dividers second topLeft: ( speaker right + 4) @ 1.
	sitStand topLeft: dividers second right + 3 @ 1.

	askToSpeak visible ifTrue: [
		askToSpeak topLeft: (dividers second right + 37 @2).
		dividers third
				 topLeft: (askToSpeak right + 5)@1.
	] ifFalse: [
		dividers third topLeft: (dividers second topLeft + (39@0)).
	].

"Right-hand group, working inwards."
	curX := right-183.	"See also Participant viewAreaRight: and defaultPPWidth."

	"Counters and health labels."
	netHealth topLeft: (curX+98)@(mid-15).
	renderHealth topLeft: (curX+98)@(mid-2).
	svcCount fullBounds;
			  topRight:  (curX+85)@(mid-15).
	appCount fullBounds;
			   topRight: (curX+85)@(mid-2).
	svcCountLabel topLeft: (curX+8)@(mid-15).
	appCountLabel topLeft: (curX+8)@(mid-2).

	dividers sixth topLeft: ( curX-2)@1.		
	inventory visible ifTrue: [
		inventory topRight: (curX-3)@1.
		dividers fifth topLeft: (curX-37@1); visible: true.
		curX := curX - 37.
	] ifFalse: [
		dividers fifth visible: false.
	].

	editBox visible ifTrue: [
		editBox topLeft: (curX-31)@(mid-12).
		curX := curX-30
	].
	thumbtack visible ifTrue: [
		thumbtack topLeft: (curX-29)@(mid-12).
		curX := curX - 30.
	].
	dividers fourth topLeft: ((curX-6)@1);
			visible: ( editBox visible or: [ thumbtack visible ]).

"Message area group"
	self locked ifTrue: [
		messageArea  left:140; 
					   width: (width - 160).
	] ifFalse: [
		messageArea left: (dividers third right + 6);  
					   width: (dividers fourth left - messageArea left - 5).
	].
	messageArea top: 1; height: (self height -3); textAnchorPoint: #leftCenter; textOffset: (2@-1).
	ioIcon topLeft: (messageArea right - 16)@(mid-13).
	messageArea label: ''; fullBounds.
	editIndicator height: (self height-1); right: (messageArea right+1).
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/5/2009 19:44'!
leaderAllowingToSpeak: bOn
	"When voice-chat restrictions are set by the leader,
	 and then they allow us to speak, we turn the want-to-speak icon to a green dot."
	(bOn and: [ owner isMeetingLeader not ]) ifTrue: [
		askToSpeak selectedImage: (QForms find: 'QuestionOn.png').
		askToSpeak normalImage: (QForms find: 'QuestionOn.png').
	] ifFalse: [
		self setupAskToSpeakButton.
	].
	askToSpeak onSelectedChanged.! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/5/2009 20:08'!
leaderControllingSpeech: bOn
	"The meeting-state for audienceIsMuted has been changed."
	self setupAskToSpeakButton.
	owner isMeetingLeader ifTrue: [
		askToSpeak selected: bOn
	].
	! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 8/1/2009 00:15'!
meetingStatusString
	"The participant uses this as the default status-line message
	 when a meeting is being led in the forum."

	| audiencemute youcan lead |

	lead := owner isMeetingLeader.
	audiencemute := owner meetingState audienceIsMuted.
	youcan := (owner meetingState isLoginAllowedToSpeak: (owner userLogin)).

	^ String streamContents: [ : str |
		audiencemute
			ifTrue: [ lead 
						ifTrue: [str nextPutAll: 'Audience chat off. ']
						ifFalse: [youcan ifFalse: [str nextPutAll: 'Chat is off. ']  ] ].
		lead ifFalse: [
			(audiencemute & youcan) 
				ifTrue: [ str nextPutAll: 'You may speak. ' ]
				ifFalse: [			
					askToSpeak selected 
						ifTrue: [ str nextPutAll: 'Asking to speak. ' ] ].
			str nextPutAll: (owner leaderPublicName);
					nextPutAll: ' is leading. '
		] ifTrue: [
			audiencemute ifFalse: [ str nextPutAll: 'Audience chat on. ' ].
			str nextPutAll: 'You are leading. '
		].
	 ].
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/1/2008 21:30'!
messageArea
	"Answer the messageArea of the receiver"
	<bewareOf: #messageAreaChanged>
	^self propertyValueAt: #messageArea! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/1/2008 21:30'!
messageArea: aValue
	"Modify the receiver's messageArea"
	^self propertyValueAt: #messageArea put: aValue with: #messageAreaChanged! !

!QControlBar methodsFor: 'events' stamp: 'bgf 11/26/2009 19:40'!
micActivity
	"Report the level from the recorder meterLevel as blips in the mic icon."
	| level |

	view ifNil: [ ^ self ].
	view window audioSession ifNotNilDo: [ :as|
		level := as microphone isRecording 
					ifTrue: [as microphone meterLevel] 
					ifFalse: [0].
		self micActivity: level.
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 6/25/2008 21:14'!
micActivity: micLevel

	"Report the given (raw) microphone meterLevel as blips in the mic icon."

	(self locked or: [micLevel < 1] or: [self pttAndControlIsUp] ) ifTrue: [	
		self microphoneLevel: 0.
	] ifFalse: [
		"Empirical - for a one-to-six steps display, from a 0-100 level"
		self microphoneLevel: ((((micLevel * micLevel * 10) log * 1.2) rounded min:6 max:1)).
	]
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:40'!
microphone
	"Answer the microphone of the receiver"
	<bewareOf: #microphoneChanged>
	^self propertyValueAt: #microphone! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:40'!
microphone: aValue
	"Modify the receiver's microphone"
	^self propertyValueAt: #microphone put: aValue with: #microphoneChanged! !

!QControlBar methodsFor: 'events' stamp: 'bgf 5/29/2010 11:13'!
microphoneButtonFired: isSelected
	"A microphone was clicked, and we have its new (provisional) selection value.  We don't just send this on actual clicks... sometimes we want to turn recording on/off programatically."

	self owner mainView ifNil: [ ^self].

	(isSelected and: [QAudioDeviceManager recorderDevices isEmpty]) ifTrue: [
		"Do not set proxies back to false, because it drives #microphoneStance -> org services notice."
		self owner mobileProxy ifNotNil: [^self].  

		self microphoneMessage: 'There is no microphone device.'.
		^self allMicrophoneButtonsDo: [:mic | mic selected: false].
	].
	self allMicrophoneButtonsDo: [:mic | 
		mic pressToTalkPressed: false.
		mic selected: isSelected
	].
	self owner controlBar setPressToTalk: (self owner preferences pressToTalk).

	isSelected ifTrue: [
		self owner playSoundEffect: #blip.
		self owner startVoiceChat.
		QActivityCapture action: 
			((self owner preferences pressToTalk)
				ifTrue: [ #MicrophoneOnPTT ]
				ifFalse: [ #MicrophoneOn ]).

	] ifFalse: [
		self owner endVoiceChat.
		self owner playSoundEffect: #blip.
		QActivityCapture action: #MicrophoneOff.
	].
! !

!QControlBar methodsFor: 'private' stamp: 'hrs 12/8/2008 11:41'!
microphoneLevel: numBlips
	"A microphone volume from 0 to 6"

	microphone level: numBlips
! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 10/15/2009 20:22'!
microphoneMessage: msg
	"Put up an obvious message near the microphone/media control area"
	owner statusLine floatingMessage: msg
			at: (self bottomLeft + (190@-5)) extraSeconds: 2.! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/8/2009 11:11'!
netHealth
	"Answer the netHealth of the receiver"
	<bewareOf: #netHealthChanged>
	^self propertyValueAt: #netHealth! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/8/2009 11:11'!
netHealth: aValue
	"Modify the receiver's netHealth"
	^self propertyValueAt: #netHealth put: aValue with: #netHealthChanged! !

!QControlBar methodsFor: 'private' stamp: 'bgf 9/8/2009 17:34'!
numApps
	"We stow the app count in the label value as a string."
	appCount value ifNotNilDo: [ :v |
		^ (v = '') ifTrue: [ 0 ] ifFalse: [ v asInteger ].
	].
	^ 0! !

!QControlBar methodsFor: 'private' stamp: 'bgf 9/8/2009 17:34'!
numForums
	"We stow the service count in the label "
	svcCount value ifNotNilDo: [ :v |
		^ (v = '') ifTrue: [ 0 ] ifFalse: [ v asInteger ].
	].
	^ 0! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 7/5/2009 19:42'!
onAskToSpeakEnter

	<on: mouseEnter in: askToSpeak>

	owner statusInfo: (self askToSpeakTipString)
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/5/2009 19:40'!
onAskToSpeakFire
	<on: fire in: askToSpeak>

	askToSpeak selected: (askToSpeak selected not).
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 8/16/2009 17:09'!
onAskToSpeakSelectedChanged
	<on: selectedChanged in: askToSpeak>

	owner isMeetingLeader ifTrue: [
		owner meetingState setAudienceIsMuted: askToSpeak selected
	] ifFalse: [
		owner setWantsToSpeak: (askToSpeak selected).
		(owner readyForEvents and: [ askToSpeak selected ])
		ifTrue: [
			owner avatar replica future attentionGesture
		].
	].! !

!QControlBar methodsFor: 'tips' stamp: 'sn 5/13/2011 22:48'!
onAwayButtonEnter
	<on: mouseEnter in: awayButton>

	awayButton selected ifTrue: [
		owner statusInfo: 'Away.  Voice chat suspended.  Click to return.'  translateMe.
	] ifFalse: [
		userIsIdle ifTrue: [
			owner statusInfo: 'You are present but Idle. Click for Away.'  translateMe.
		] ifFalse: [
			owner statusInfo: 'You are present and active.  Click for Away.' translateMe.
		].
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 2/2/2008 01:28'!
onAwayButtonFire
	<on: fire in: awayButton>
	
	"The away button is 'selected' when we are Away."
	awayButton selected: awayButton selected not.
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 4/29/2008 14:40'!
onAwaySelectedChanged
	<on: selectedChanged in: awayButton>

	"We want to respect the selected state in the button... 
	 if we get busy and these signal-handlings stack up,
	 always take us to the current button value."

	DropAllButLastEvent signal.

	"Let the world know ..."
	awayButton selected ifFalse: [
		owner userBackFromAway.
	] ifTrue: [
		owner userAway.
	].
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 1/20/2008 14:38'!
onCameraDown

	<on: mouseDown in: camera>

	"The camera button has two behaviors...
	 if clicked quickly, it starts video in the last quality this user chose;
	 if held, it offers a menu of qualities at which to start video.
	 When the camera is alread on, the button simply turns it off."
	buttonDownTime := Time millisecondClockValue.
! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 10/15/2009 20:12'!
onCameraEnter
	| sess setting noun |
	<on: mouseEnter in: camera>
	
 	sess := owner videoSession ifNil: [ ^ self ].
	setting := sess qualitySettingString.
	noun := sess cameraName ifNil: [ 'Camera'].

	camera enabled ifFalse: [
		owner statusInfo: 'No webcam device detected on this computer.'
	] ifTrue: [
		owner statusInfo: (
			camera selected 
				ifTrue: [ noun, ' is on at ', setting, ' quality.']
				ifFalse: [ noun, ' is off (', setting, ' quality).  Hold to choose quality.']).
	].
		! !

!QControlBar methodsFor: 'events' stamp: 'bgf 8/22/2009 23:47'!
onCameraFire

	<on: fire in: camera>

	"The camera button has two behaviors...
	 if clicked quickly, it starts video in the last quality this user chose;
	 if held, it offers a menu of qualities at which to start video."

	buttonDownTime = 0 ifTrue: [ ^ self ].	"the menu took over"

	"Normal click - means toggle the camera."
	camera selected ifFalse: [
		"Turning it on, so make sure video is enabled first."
		self checkVideoSession ifFalse: [ ^ self ].
	].
	self owner playSoundEffect: #blip.
	camera selected: (camera selected not)! !

!QControlBar methodsFor: 'events' stamp: 'bgf 1/20/2008 03:05'!
onCameraLeave

	<on: mouseLeave in: camera>

	"Forget the press-time...
	 makes sure we don't fire the menu on a down-leave-up sequence."
	buttonDownTime := 0! !

!QControlBar methodsFor: 'events' stamp: 'bgf 2/16/2008 13:48'!
onCameraRightDown
	<on: yellowButtonDown in: camera>
	self cameraMenu.
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 9/19/2007 09:18'!
onCameraSelectedChanged

	<on: selectedChanged in: camera>

	view ifNil: [ ^self].

	(owner respondsTo: #startVideoChat) ifFalse: [
		"No live video support."
		camera selected: false.
		QLogger log: 'Owner (', owner, ') does not support video chat' from: self.
		^ self.
	].

	camera selected ifTrue: [
		owner startVideoChat.
	] ifFalse: [
		owner endVideoChat.
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 1/20/2008 14:37'!
onCameraStillDown
	"Press-hold check in the camera button."
	<on: mouseStillDown in: camera>

	"If we're counting, and if the button has been down long enough,
	 fire the quality-choice menu."
	(buttonDownTime > 0) ifTrue: [
		(Time millisecondClockValue > (buttonDownTime + (QStyles delayedActionMenuDelay))) 
		ifTrue: [
			buttonDownTime := 0.
			self cameraMenu.
		].
	].! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 11/2/2007 13:26'!
onControlBarLeave

	"By clearing info when the mouse leaves the control bar,
	 we prevent lingering tip statusInfo messages,
	 so they feel more like in-place tooltips."

	<on: mouseLeave>
	owner statusInfo: nil! !

!QControlBar methodsFor: 'events' stamp: 'bgf 9/8/2009 17:27'!
onCountChanged
	<on: valueChanged in: svcCount>
	(self numForums = 0) ifTrue:[
		svcCount textColor: Color transparent. 
		appCount textColor: Color transparent.
		svcCountLabel textColor: Color transparent.
		appCountLabel textColor: Color transparent.
	] ifFalse: [
		svcCount textColor: (Color gray: 0.12).	
		appCount textColor: (Color gray: 0.12).
		svcCountLabel textColor: (Color gray: 0.12).	
		appCountLabel textColor: (Color gray: 0.12).
	].
! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 6/30/2009 18:30'!
onEditBoxEnter
	<on: mouseEnter in: editBox>
	owner statusInfo: 'Tool for selecting objects to resize.'.! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/31/2009 00:43'!
onEditBoxFire
	<on: fire in: editBox>
	editBox selected ifFalse: [
		self cursorMode: #editBox.
		"paintbrush deselect. removed for now"
		thumbtack deselect.
		editBox select.
	] ifTrue: [
		self resetArrow.
	].

		! !

!QControlBar methodsFor: 'events' stamp: 'bgf 6/3/2009 11:13'!
onEditIndicatorClick
	<on: mouseDown in: editIndicator>
	"The edit indicator is visible when the user has chosen to edit the full scene;
	 this is done in the inspector, so we put a click-shortcut from the indicator
	 to bring back the inspector."
	owner treeInspector expand! !

!QControlBar methodsFor: 'events' stamp: 'bgf 2/1/2008 20:49'!
onExtentChanged
	<on: extentChanged>
	self layoutButtons.! !

!QControlBar methodsFor: 'tips' stamp: 'sn 5/13/2011 22:43'!
onInventoryEnter
	"Status-line info for the Inventory button."

	<on: mouseEnter in: inventory>
	owner avatar ifNil:[ ^self].
	owner avatar replica ifNil: [ ^ self ].
	inventory selected ifFalse: [
		owner statusInfo: 'Click to show the Object Catalog.' translateMe.
	] ifTrue: [
		owner statusInfo: 'Click to hide the Object Catalog.' translateMe
	].	! !

!QControlBar methodsFor: 'events' stamp: 'bgf 4/28/2007 14:22'!
onInventoryFire
	<on: fire in: inventory>
	inventory selected: inventory selected not.
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 1/11/2010 12:13'!
onInventorySelectedChanged
	<on: selectedChanged in: inventory>

	owner ifNotNil: [ owner toggleCatalog: (inventory selected) ].
	owner readyForEvents ifTrue: [
		QActivityCapture countAction: #ToggleInventory
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/25/2009 21:52'!
onIsLeaderChanged: bLeader
	"When the owner changes to or from being the meeting leader himself."
	<on: isMeetingLeaderChanged in: owner>

	self setupAskToSpeakButton.

	"If you are the leader, you don't use the q-button to ask to speak,
	   so turn it off.  For the leader, it governs blocking everyone else's voice."
	askToSpeak selected: false.
! !

!QControlBar methodsFor: 'events' stamp: 'jcg 4/30/2010 16:12'!
onIsMuteChanged: isMute
	"The global mute state has changed.  Update the speaker icon."
	<on: isMuteChanged in: owner>
	speaker selected: isMute not.
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/5/2009 19:44'!
onLeaderChanged: leaderLogin 
	"Whenever the leader changes (including to or from being led at all.)"
	<on: leaderLoginChanged in: owner>

	self setupAskToSpeakButton.
! !

!QControlBar methodsFor: 'events' stamp: 'jcg 5/10/2010 12:40'!
onMainViewChanged

	"The owner sets the mainView after harness setup;
	 we catch the changes here and keep my 'view' up to date."

	<on: mainViewChanged in: owner>

	self view: owner mainView.

	self enableSound.
	self applyUserRole.
			"After login, we can determine which server-controlled features are available."

	"Ensure that all speaker- and microphone-buttons start off in the correct state."
	owner signal: #isMuteChanged with: owner isMute.
	self allMicrophoneButtonsDo: [:mic | mic selected: false].

	"When we first log in, camera is not active, so de-select the camera button (which might have been left on when we last logged off)."
	camera selected: false.
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/11/2009 01:39'!
onMessageChanged
	<on: valueChanged in: messageArea>

	"There's no signaling of the actual message font size changes,
	 so we check this when the message itself changes."
	messageArea top: ((self height // 2) - 3 - ((messageArea font pointSize)//2 )).! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 1/8/2009 12:52'!
onOverAppCount
	"When the mouse is over the app-count label,
	 recap what it means in the status line."

	<on: mouseMove in: appCount>

	self onOverCounterLines.! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 9/8/2009 17:27'!
onOverAppCountLabel
	"When the mouse is over the app-count label,
	 recap what it means in the status line."

	<on: mouseMove in: appCountLabel>

	self onOverCounterLines.! !

!QControlBar methodsFor: 'tips' stamp: 'ar 4/16/2011 17:10'!
onOverCounterLines
	"When the mouse is over the app- or service-count label,
	 explain what they mean in the status line."
	| count msg |

	owner avatar ifNil:[ ^self].
	owner avatar replica ifNil: [ ^ self ].

	count := self numForums.
	msg := count caseOf: {
			[nil] -> ['Not logged in.'].
			[0] -> ['No Forum open.'].
			[1] -> ['One Forum open.'].
		} otherwise: [ '', count, ' Forums open.'].

	count := self numApps.
	msg := msg, ' ', 
			(count caseOf: {
				[nil] -> ['No remote apps in use.'].
				[0] -> ['No remote apps in use.'].
				[1] -> ['One remote app in use.'].
			} otherwise: [ '', count, ' remote apps in use.']).

	owner statusInfo: msg.
! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 1/8/2009 16:06'!
onOverNetHealth
	"When the mouse is over the net-health label,
	 recap what it means in the status line."
	<on: mouseMove in: netHealth>
	
	self onOverStatusLines.! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 1/8/2009 16:06'!
onOverRenderHealth
	"When the mouse is over the render-health label,
	 recap what it means in the status line."

	<on: mouseMove in: renderHealth>
	
	self onOverStatusLines.! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 1/8/2009 12:52'!
onOverServiceCount
	"When the mouse is over the service-count label,
	 recap what it means in the status line."

	<on: mouseMove in: svcCount>

	self onOverCounterLines.! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 9/8/2009 17:27'!
onOverServiceCountLabel
	"When the mouse is over the service-count label,
	 recap what it means in the status line."

	<on: mouseMove in: svcCountLabel>

	self onOverCounterLines.! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 1/8/2009 16:07'!
onOverStatusLines
	"When the mouse is over a health label,
	 explain them in the status line."
	| msg latency netLevel frameTime renderLevel |
	
	owner avatar ifNil:[ ^self].
	owner avatar replica ifNil: [ ^ self ].
	latency := self recentLatency.
	netLevel := self tagForLatency: latency.
	frameTime := self recentRenderTime.
	renderLevel := self tagForRenderTime: frameTime.

	msg := ''.
	latency ifNotNil: [
		msg := msg, 'Network delay is ', netLevel, '.  '.
		self updateNetHealthLabel: latency.
	].
	frameTime ifNotNil: [
		msg := msg, 'Display speed is ', renderLevel, '. '.
		self updateGraphicsHealthLabel: frameTime.
	].

	owner statusInfo: msg.
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 4/1/2010 18:03'!
onOwnerChanged
	<on: ownerChanged>

	owner ifNotNil: [ 
		self setupKeyDriveRedirect
	]! !

!QControlBar methodsFor: 'tips' stamp: 'bgf 11/2/2007 13:23'!
onReadoutEnter
	<on: mouseEnter in: readoutButton>

	owner statusInfo: (readoutButton selected
		ifTrue: ['Hides the position readout.']
		ifFalse: ['Shows the position readout.']).! !

!QControlBar methodsFor: 'events' stamp: 'bgf 12/18/2006 15:16'!
onReadoutFire
	<on: fire in: readoutButton>

	readoutButton selected: (readoutButton selected not).	

! !

!QControlBar methodsFor: 'events' stamp: 'bgf 6/21/2008 01:09'!
onReadoutSelectedChanged
	<on: selectedChanged in: readoutButton>

	owner posReadout visible: (readoutButton selected).
	owner posReadout visible ifTrue: [
		owner posReadout comeToFront.
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 12/30/2009 14:42'!
onSessionChanged
	"The owner sets the mainView after harness setup;
	 we catch the changes here and keep my 'view' up to date."
	<on: sessionChanged in: owner>
	self stopScript: #updateAppCount.
	owner session ifNotNil: [
		self startScript: #updateAppCount when: {owner session . #embeddedAppsUpdated }.	
	].

	self isSeatNearby: false.
	self stopScript: #postureChangedFrom:to:.
	self stopScript: #isSeatNearby:.
	owner avatar ifNotNil: [
		self startScript: #postureChangedFrom:to: 
			when: { owner avatar . #postureChangedFrom:to: }.
		self startScript: #isSeatNearby: 
			when: { owner avatar . #isNearASeat }.
	].! !

!QControlBar methodsFor: 'tips' stamp: 'sn 5/13/2011 22:41'!
onSitStandEnter
	<on: mouseEnter in: sitStand>

	owner avatar ifNil:[ ^self].
	owner avatar replica ifNil: [ ^ self ].
	owner allowedToMove ifFalse: [
		^ owner statusInfo: 'The leader is controlling movement.' translateMe
	].

	sitStand selected ifTrue: [
		owner statusInfo: ('Click to stand up.' translateMe).
	] ifFalse: [
		sitStand overImage = (QForms find: 'SitDownOver.png') ifTrue: [ 
			owner statusInfo: ('Click to take a seat.' translateMe)
		] ifFalse: [ 
			owner statusInfo: ('No seats nearby.' translateMe)
		].
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/9/2009 20:00'!
onSitStandFire
	<on: fire in: sitStand>
	owner readyForEvents ifTrue: [
		"Clear the status line mouse-enter help which looks odd once you've hit the button."
		owner statusInfo: nil.	
		owner avatar replica future togglePosture.
	].! !

!QControlBar methodsFor: 'events' stamp: 'bgf 5/29/2010 11:12'!
onSpeakerSelectedChanged

	<on: selectedChanged in: speaker>

	self owner mainView ifNil: [ ^self].

	(speaker selected and: [QAudioDeviceManager playerDevices isEmpty]) ifTrue: [
		self microphoneMessage: 'There is no speaker device.'.
			^speaker selected: false.
		].
! !

!QControlBar methodsFor: 'events' stamp: 'bgf 8/28/2009 09:02'!
onStickyNoteDrag
	<on: mouseDrag in: thumbtack>
	"Dragging directly from the sticky-note icon into the space;
	 an easier creation path than click-clicking."

	owner ifNil: [ ^ self ].
	owner readyForEvents ifFalse: [ ^ self ].
	owner canModifyForum ifFalse: [ ^ self ].

	thumbtack selected: false; disable.
	self resetArrow.
	[self dragOutAStickyNote ] ensure: [ 
		thumbtack selected: false; enable.
		self resetArrow
	].
! !

!QControlBar methodsFor: 'tips' stamp: 'sn 5/13/2011 22:38'!
onThumbtackEnter
	<on: mouseEnter in: thumbtack>
	owner statusInfo: 'Tool for adding a sticky note.' translateMe.! !

!QControlBar methodsFor: 'events' stamp: 'bgf 6/30/2009 18:26'!
onThumbtackFire
	<on: fire in: thumbtack>

	thumbtack selected ifFalse: [
		self cursorMode: #thumbtack.
		editBox deselect.
		thumbtack select.
	] ifTrue: [ 
		self resetArrow
	].

		! !

!QControlBar methodsFor: 'events' stamp: 'bgf 3/25/2010 13:31'!
onTick
	<ticking: 3>

	"Keep the health indicators up-to-date."
	self visible ifFalse: [ ^ nil ].

	self updateNetHealthLabel: (self recentLatency).
	self updateGraphicsHealthLabel: (self recentRenderTime).
	self updateServiceCount.
	self updateAppCount.

	self micActivity.		"Clear leftover zero"

	"Check the avatar activity faders..."
	owner avatar ifNotNil: [ 
		owner avatar checkTypingActivity.
		sitStand selected: (owner avatar driver posture = #Sitting) ].


	owner updateTitle.! !

!QControlBar methodsFor: 'events' stamp: 'bgf 5/17/2008 21:06'!
onUserIsIdleChanged
	<on: userIsIdleChanged>

	owner avatar ifNotNil: [
		QActivityCapture action: (userIsIdle ifTrue: [ #Idle ] ifFalse: [ #NotIdle ]).
		userIsIdle ifTrue: [
			owner avatar startActivity: #Idle
		] ifFalse: [
			owner avatar stopActivity: #Idle
		]].

	"As a bit of feedback, when we enter 'Idle' in the presence system,
	 we fade the user silhouette in the Away button."

	userIsIdle ifTrue: [
		awayButton normalImage: (QForms find: 'UserIdle.png').
		awayButton overImage: (QForms find: 'UserIdleOver.png').
	] ifFalse: [
		awayButton normalImage: (QForms find: 'UserPresent.png').
		awayButton overImage: (QForms find: 'UserPresentOver.png').
	].

! !

!QControlBar methodsFor: 'events' stamp: 'bgf 6/13/2009 18:37'!
onYellowDown: event
	"This is here simply to capture tweak hand mouse focus to prevent drive-start leakage.
	 The Tweak prim costume handleMouseMove: sends events regardless of focus,
	 but handleMouseUp: requires focus.  Focus is taken by the original down-target.
	 So if you y-down on the control bar, and move into the main view
	 (which sees yellow moves and makes sure you're driving for them)
	 and release, you don't get an up event to stop driving through."

	<on: yellowButtonDown>! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:46'!
owner
	"Answer the owner of the receiver"
	<bewareOf: #ownerChanged>
	^self propertyValueAt: #owner! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:46'!
owner: aValue
	"Modify the receiver's owner"
	^self propertyValueAt: #owner put: aValue with: #ownerChanged! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/9/2009 18:42'!
postureChangedFrom: oldPosture to: newPosture

	"On posture change in the avatar "
	sitStand selected: (newPosture = #Sitting)! !

!QControlBar methodsFor: 'private' stamp: 'jcg 9/5/2007 11:05'!
preferences
	^self owner ifNotNilDo: [:o | o preferences]! !

!QControlBar methodsFor: 'private' stamp: 'bgf 5/28/2009 18:59'!
pttAndControlIsDown
	"True if the current state is a press to talk microhone with the press-to-talk control key down."

	^ self isPTT
	   and: [ microphone micPTT graphicForm isNil		"Control is down" ]. ! !

!QControlBar methodsFor: 'private' stamp: 'hrs 12/8/2008 16:31'!
pttAndControlIsUp
	"Used to suppress green-feedback when press to talk is off;
	  it is just too confusing as information.   
	  We use of pause/resume in the Recorder for PTT.
	  A paused SoundRecorder still sends activity messages, just not actual sound data.
	  We don't want to show that activity if the PTT key is up (not sending)."

	^ self isPTT
	   and: [ microphone micPTT graphicForm notNil		"Control is up" ]. ! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 14:52'!
readoutButton
	"Answer the readoutButton of the receiver"
	<bewareOf: #readoutButtonChanged>
	^self propertyValueAt: #readoutButton! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 14:52'!
readoutButton: aValue
	"Modify the receiver's readoutButton"
	^self propertyValueAt: #readoutButton put: aValue with: #readoutButtonChanged! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/8/2009 14:30'!
recentLatency
	"Recent 5-second network latency, or nil if not known."
	| stats five |
	owner avatar ifNil:[ ^nil].
	owner avatar replica ifNil: [ ^ nil ].
	stats := owner latencyStatistics.
	stats ifNotNil: [
		five := stats nSecondStats: 5.
		five ifNotNil: [ ^ five first].
	].

	^ nil
! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/8/2009 16:08'!
recentRenderTime
	"Recent average time to render a frame."
	| stats three |
	owner avatar ifNil:[ ^nil].
	owner avatar replica ifNil: [ ^ nil ].
	stats := owner fpsStats.
	stats ifNotNil: [
		"We use a fairly short time here since it depends on POV,
	 	 and a given second usually has enough data to make it useful."
		three := stats nSecondStats: 3.
		three ifNotNil: [ ^ three first].
	].
	^ nil
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/8/2009 11:11'!
renderHealth
	"Answer the renderHealth of the receiver"
	<bewareOf: #renderHealthChanged>
	^self propertyValueAt: #renderHealth! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/8/2009 11:11'!
renderHealth: aValue
	"Modify the receiver's renderHealth"
	^self propertyValueAt: #renderHealth put: aValue with: #renderHealthChanged! !

!QControlBar methodsFor: 'events' stamp: 'bgf 6/30/2009 17:35'!
resetArrow
	"Reset the cursor mode to the normal arrow."
	self thumbtack selected:false.
	self editBox selected: false.
	self cursorMode: #arrow.! !

!QControlBar methodsFor: 'private' stamp: 'jcg 2/17/2010 18:37'!
resetSound
	"Reset sound - called from the mic/speaker menus"
	QActivityCapture action: #ResetSoundControlBar.
	owner resetSound: 'user action in QControlBar'.! !

!QControlBar methodsFor: 'initialize' stamp: 'bgf 6/30/2009 18:30'!
setEditing: bEditing
	"Whether to include scene-editing indications."
	bEditing = (editIndicator visible) ifFalse: [
		editIndicator visible: bEditing.
		editBox visible: bEditing; enabled: bEditing.
		bEditing ifFalse: [ self resetArrow ].
		self layoutButtons.
	].! !

!QControlBar methodsFor: 'accessing' stamp: 'jcg 4/30/2010 11:19'!
setPressToTalk: yesNo

	"This is the entry API for changing the press-to-talk preference, and the client posture.
	 Press to talk is a remembered preference that is governed by the most-recent 
	 choice the user makes in the microphone menu."

	(owner preferences pressToTalk == yesNo) ifFalse: [
		owner preferences pressToTalk: yesNo.
		owner commitPreferencesInBackground.
	].

	microphone pressToTalkOn: yesNo & microphone selected.

	self isPTT == yesNo ifTrue: [^self].
	owner audioSession ifNotNilDo: [:a | a future isPressToTalk: yesNo].! !

!QControlBar methodsFor: 'private' stamp: 'bgf 10/16/2009 14:29'!
setVideoDeviceName: cameraName deviceUID: deviceUID

	"Menu callback from the select-camera submenu."
	(self checkVideoSession) ifFalse: [ ^ self ].
	owner videoSession cameraManager 
			desiredDeviceName: cameraName uid: deviceUID.
	camera selected ifFalse: [ camera selected: true ].
! !

!QControlBar methodsFor: 'private' stamp: 'bgf 7/5/2009 19:44'!
setupAskToSpeakButton

	"Update the question button visibility and labeling.
	 It is visible when a meeting (a leader) is in effect.
	 The decorations change for the leader himself."

	askToSpeak visible: ( owner isMeetingInProgress ).
	owner isMeetingLeader ifTrue: [
		self setupAskToSpeakButtonForLeader
	] ifFalse: [
		self setupAskToSpeakButtonForAudience
	].
	self layoutButtons.! !

!QControlBar methodsFor: 'private' stamp: 'bgf 7/5/2009 20:09'!
setupAskToSpeakButtonForAudience
	"The audience (everyone but leader) sees a ? button that they use to ask to speak."
	askToSpeak 
		pressedImage: (QForms find: 'QuestionDown.png');
		selectedImage: (QForms find: 'QuestionDown.png');
		selectedPressedImage: (QForms find: 'QuestionOver.png');
		overImage: (QForms find: 'QuestionOver.png').
	
	
	(owner notNil and: [ owner meetingState audienceIsMuted ]) 
	 ifTrue: [
		askToSpeak normalImage: (QForms find: 'QuestionUpMute.png')
	] ifFalse: [
		askToSpeak normalImage: (QForms find: 'QuestionUp.png')
	].! !

!QControlBar methodsFor: 'private' stamp: 'bgf 7/5/2009 19:44'!
setupAskToSpeakButtonForLeader
	"The leader sees a stylized L in the button; they use it to turn on controlled-speaked."
	askToSpeak 
		normalImage: (QForms find:'QuestionLeaderUp.png');
		pressedImage: (QForms find: 'QuestionLeaderDown.png');
		selectedImage: (QForms find: 'QuestionLeaderDown.png');
		selectedPressedImage: (QForms find: 'QuestionLeaderOver.png');
		overImage: (QForms find: 'QuestionLeaderOver.png').
	! !

!QControlBar methodsFor: 'initialize' stamp: 'sn 5/8/2011 18:54'!
setupCostume

	"@ Note - balloon help is coded, but suppressed in favor
	 of status-line mouseover messages which are less disruptive.
	 Some controls (e.g. microphone) are dynamically placed
	 and so their balloon help is commented... the others 
	 are suppressed by a children-do: at the bottom of this method."

	| invisible lblColor |
	super setupCostume.
	self sticky: true. "prevent me from being picked up willy nilly"
	invisible := Color r: 0.0 g: 0.0 b: 0.0 alpha: 0.0.

	self color: (Color white).
	self fillAlpha: 1.0.
	self fillColors: (CWidgetLibrary default menuColors).
	self fillStyle: #verticalGradient.

	"The border is set up to give a top line to match the bottom of the menu bar;
	 the participant ui submerges the other three border edges to save space."
	self height: 34.
	self borderWidth: 1.
	self borderColor: (QColors mainBorderColor).
	self borderStyle: #simple.

	lblColor := (Color r: 0.01 g: 0.15 b: 0.2). 

	messageArea color: invisible;
		font: QStyles controlFont;
		extent: 300@20;
		hResizing: #rigid;
		textColor: lblColor;
		balloonHelp: 'Status and information messages.';
		top: 9;
		label: ''.
	self add: messageArea.

	editIndicator 	label: ' [Editing]' translateMe;
		font: QStyles smallControlFont;
		textColor: lblColor;
		textOffset: (0@-2); top: (1); 
		vResizing: #rigid;
		hResizing: #shrinkWrap;
		fullBounds;
		 fillAlpha: 1.0;
		 fillColors: (CWidgetLibrary default menuColors);
		 fillStyle: #verticalGradient;		
		 visible: false.
	self add: editIndicator.

	svcCountLabel color: invisible;
		textFont: (QStyles smallControlFont);
		hResizing: #shrinkWrap;
		textColor: (Color r:0.1 g:0.1 b:0.7).
	self add: svcCountLabel.
	svcCount color: invisible;
		textFont: (QStyles smallControlFont);
		hResizing:#shrinkWrap;
		textColor: (Color r:0.0 g:0.4 b:0.5);
		" balloonHelp: 'The number of forums you are connected to.';	 
			Deliberately absent,
			because the svcCount has a status-line mouseover and the b-h distracts from it."
		value: '';
		textAnchorPoint: #rightCenter.
	self  add: svcCount.

	appCountLabel color: invisible;
		textFont: (QStyles smallControlFont);
		hResizing: #shrinkWrap;
		textColor: (Color r:0.1 g:0.1 b:0.7).
	self add: appCountLabel.

	appCount color: invisible;
		textFont: (QStyles smallControlFont);
		hResizing:#shrinkWrap;
		textColor: (Color r:0.1 g:0.1 b:0.7);
		value: '';
		textAnchorPoint: #rightCenter.
	self add: appCount.
	netHealth color: invisible;
		textFont: (QStyles smallControlFont);
		hResizing:#shrinkWrap;
		textColor: (Color r:0.1 g:0.1 b:0.7);
		value: 'Network:'.
	self add: netHealth.
	renderHealth color: invisible;
		textFont: (QStyles smallControlFont);
		hResizing:#shrinkWrap;
		textColor: (Color r:0.1 g:0.1 b:0.7);
		value: 'Graphics:'.
	self add: renderHealth.

	askToSpeak color: invisible;
		extent: 32@32.
	self setupAskToSpeakButtonForAudience.
	self add: askToSpeak.

	thumbtack color: invisible;
		extent: 24@24;
		balloonHelp: 'Set the pointer mode to create a sticky note.' translateMe;
		normalImage: (QForms find:'StickyNoteNormal.png');
		pressedImage: (QForms find: 'StickyNoteDown.png');
		selectedImage: (QForms find: 'StickyNoteDown.png');
		selectedPressedImage: (QForms find: 'StickyNoteDown.png');
		disabledImage: (QForms find: 'StickyNoteDisabled.png');
		overImage: (QForms find: 'StickyNoteOver.png').
	self add: thumbtack.

	editBox color: invisible;
		extent: 24@24;
		balloonHelp: 'Set the pointer mode for editing objects position and scale.';
		normalImage: (QForms find: 'edit.png');
		pressedImage: (QForms find:'editDown.png');
		selectedImage: (QForms find: 'editDown.png');
		selectedPressedImage: (QForms find: 'editDown.png');
		overImage: (QForms find: 'editOver.png').
	self add: editBox.

	microphone 
		selected: false.
	self add: microphone.

	speaker 
		selected: QAudioDeviceManager playerDevices isEmpty not.
	self add: speaker.

	"only for developers, for now"
	camera color: invisible;
		extent: 32@32;
	"@	balloonHelp: 'Turn video camera on or off.  Press and hold to select video-chat quality.'; "
		normalImage: (QForms find: 'cameraOff.png');
		pressedImage: (QForms find: 'cameraDown.png');
		overImage: (QForms find: 'cameraOffOver.png');
		selectedImage: (QForms find: 'cameraOn.png');
		selectedOverImage: (QForms find: 'cameraOnOver.png');
		selectedPressedImage: (QForms find: 'cameraDown.png');
		selected: false.
		self add: camera.

	awayButton color: invisible;
		extent: 30@30;
		normalImage: (QForms find: 'UserPresent.png');
		overImage: (QForms find: 'UserPresentOver.png');
		selectedImage: (QForms find: 'UserAbsent.png');
		selectedOverImage: (QForms find: 'UserAbsentOver.png');
		selected: false.
	self add: awayButton.

	sitStand color: invisible;
		extent: 32@32;
		normalImage: (QForms find: 'SitDownIcon.png');
		overImage: (QForms find: 'SitDownOver.png');
		selectedImage: (QForms find: 'StandUpIcon.png');
		selectedOverImage: (QForms find: 'StandUpOver.png');
		selected: false.
	self add: sitStand.


	readoutButton color: invisible;
		extent: 115@30;
		balloonHelp: '(Bring up the position readout.)';
		normalImage: (QForms openqwaqLogoClearCBForm);
		selectedImage: (QForms openqwaqLogoClearCBForm);
		overImage: (QForms openqwaqLogoClearCBForm);
		selectedOverImage: (QForms openqwaqLogoClearCBForm);
		graphicFit: #rigid;
		graphicAnchorPoint: #center.
	self add: readoutButton. 

	inventory color: invisible;
		extent: 32@32;
		balloonHelp: 'Show or hide your inventory.';
		normalImage: (QForms find:'inventoryClosed.png');
		selectedImage: (QForms find: 'inventoryOpen.png');
		overImage: (QForms find:'inventoryClosedOver.png');
		selectedOverImage: (QForms find:'inventoryOpenOver.png');
		disabledImage: (QForms find: 'inventoryDisabled.png');
		fillAlpha:0.9.
	self add: inventory. 

	ioIcon color: invisible;
		extent: 16@24;
		balloonHelp: 'Indicates resource transfer activity.';
		graphic: (QForms find: 'io.png');
		visible: false.
	self add: ioIcon.

	"Suppress all those balloon helps (keeping them in case someone changes my mind.
	 See 'tips' category... we put terse tips into the status line itself for each control.
	 This works well for the control bar since the user's attention is already here."
	self collection do: [ :c | c balloonHelp: nil ].
		
	6 timesRepeat: [
		| divider |
		divider := CPlayer new.
		divider fillAlpha: 1.
		divider graphicForm: (QForms find:  'ControlBarDivider.png').
		divider extent: 2@32.
		divider borderWidth: 0.
		self add: divider.
		dividers add: divider.
	].

	"Layout comes later, and is the responsibility of the owner."! !

!QControlBar methodsFor: 'private' stamp: 'bgf 4/1/2010 18:02'!
setupKeyDriveRedirect
	"Somewhat specialized - used to redirect keyStrokes 
	 from selected components to the owner's main view panel,
	 so that keyboard driving continues to work."

	self keyDrivePlayers do: [ :pl |
		owner keyDriveRedirect: pl
	]
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 7/9/2009 18:31'!
sitStand
	"Answer the sitStand of the receiver"
	<bewareOf: #sitStandChanged>
	^self propertyValueAt: #sitStand! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 7/9/2009 18:31'!
sitStand: aValue
	"Modify the receiver's sitStand"
	^self propertyValueAt: #sitStand put: aValue with: #sitStandChanged! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/8/2009 16:09'!
slowGraphicsThreshold
	"The one-frame-render time (ms) that makes the indicator say 'Graphics: Slow' "
	^ 105

	"Which gives a frame rate of maybe 7 or so, when we're in normal operation."! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/8/2009 16:09'!
slowLatencyThreshold
	"The degree of controller latency (ms) that makes the indicator say 'Network: Slow'. "
	^ 225! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/20/2008 18:45'!
soundDialog
	"Answer the soundDialog of the receiver"
	<bewareOf: #soundDialogChanged>
	^self propertyValueAt: #soundDialog! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 1/20/2008 18:45'!
soundDialog: aValue
	"Modify the receiver's soundDialog"
	^self propertyValueAt: #soundDialog put: aValue with: #soundDialogChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:40'!
speaker
	"Answer the speaker of the receiver"
	<bewareOf: #speakerChanged>
	^self propertyValueAt: #speaker! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:40'!
speaker: aValue
	"Modify the receiver's speaker"
	^self propertyValueAt: #speaker put: aValue with: #speakerChanged! !

!QControlBar methodsFor: 'private' stamp: 'bgf 5/29/2008 13:18'!
startCameraAtQuality: aQuality

	"Make sure the quality matches, and launch the camera."

	(self checkVideoSession) ifFalse: [ ^ self ].

	owner videoSession qualitySetting: aQuality.
	camera selected ifFalse: [ camera selected: true].
			"Button selected-event starts the camera, giving the setting change a chance to take."! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/20/2008 15:40'!
stopVideo
	"Turn off the camera.  Currently, selected-change triggers do the business..."
	camera selected: false.! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 5/30/2007 21:02'!
svcCount
	"Answer the svcCount of the receiver"
	<bewareOf: #svcCountChanged>
	^self propertyValueAt: #svcCount! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 5/30/2007 21:02'!
svcCount: aValue
	"Modify the receiver's svcCount"
	^self propertyValueAt: #svcCount put: aValue with: #svcCountChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 9/8/2009 16:12'!
svcCountLabel
	"Answer the svcCountLabel of the receiver"
	<bewareOf: #svcCountLabelChanged>
	^self propertyValueAt: #svcCountLabel! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 9/8/2009 16:12'!
svcCountLabel: aValue
	"Modify the receiver's svcCountLabel"
	^self propertyValueAt: #svcCountLabel put: aValue with: #svcCountLabelChanged! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/8/2009 15:24'!
tagForLatency: latency

	"Arbitrary classification name for the current 10-second latency we're observing"

	^ (true caseOf: {
		[latency isNumber not "sometimes it might be nil"] -> ['unknown'].
		[(0 to: 40) includes: latency] -> ['very low'] .
		[(40 to: 90) includes: latency] -> ['low' ].
		[(90 to: 140) includes: latency] -> ['normal' ].
		[(140 to: 220) includes: latency] -> ['moderate'].
		[(220 to: 360) includes: latency] -> ['high' ].
		[(360 to: 9999) includes: latency] -> ['very high']
	} otherwise: ['unknown']).
! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/8/2009 15:49'!
tagForRenderTime: frameMillis
	"Arbitrary  name for the recent-average per-frame drawing time we're observing"
	^ (true caseOf: {
		[frameMillis isNumber not "sometimes it might be nil"] -> ['variable'].
		[frameMillis <= 25] -> ['excellent'] .
		[frameMillis between: 25 and: 45] -> ['very good'] .
		[frameMillis between: 45 and: 70] -> ['good' ].
		[frameMillis between: 70 and: 95] -> ['OK' ].
		[frameMillis between: 95 and: 110] -> ['fair'].
		[frameMillis between: 110 and: 160] -> ['poor' ].
		[frameMillis >= 160]  -> ['very poor']
	} otherwise: ['variable']).
! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:40'!
thumbtack
	"Answer the thumbtack of the receiver"
	<bewareOf: #thumbtackChanged>
	^self propertyValueAt: #thumbtack! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:40'!
thumbtack: aValue
	"Modify the receiver's thumbtack"
	^self propertyValueAt: #thumbtack put: aValue with: #thumbtackChanged! !

!QControlBar methodsFor: 'events' stamp: 'bgf 7/5/2009 19:50'!
toggleAskToSpeak
	"For the avatar '?' accelerator.  Toggle the ask-to-speak button."

	askToSpeak visible ifTrue: [
		askToSpeak selected: askToSpeak selected not.
		].! !

!QControlBar methodsFor: 'initialize' stamp: 'jcg 10/8/2008 00:11'!
toggleExperimentalFeatures

	"Sound dialog needs to be recreated, if it exists."
	soundDialog ifNotNilDo: [:sd |
		soundDialog := nil.
		sd destroy
	].
! !

!QControlBar methodsFor: 'private' stamp: 'bgf 11/24/2009 14:59'!
updateAppCount
	"Update the application count indicator."
	| v |
	owner session ifNotNilDo: [ : sn | 
		v := '', sn numRemoteApps.
	] ifNil: [
		v := ''.
	].
	"Check first using equality (rather than identity) to avoid spurius UI-invalidation."
	v ~= appCount value ifTrue: [ appCount value: v ].! !

!QControlBar methodsFor: 'private' stamp: 'bgf 7/11/2009 00:39'!
updateGraphicsHealthLabel: frameMillis
	"The graphics health label.  Color it reddish as rendering slows;
	 beyond the threshold, change the label to indicate Slow."
	| latency extra clr |

	self numForums > 0 ifTrue: [
		| lbl |
		(frameMillis isNil or:[ frameMillis < 1]) ifTrue: [
			lbl :=  'Display:	?'.
		] ifFalse: [
			lbl := 'Display:	',
				((frameMillis > (self slowGraphicsThreshold)) ifTrue: [ 'Slow'] ifFalse: [ 'OK' ]).
		].
		lbl = renderHealth label ifFalse: [  renderHealth label: lbl ].
		latency := (frameMillis ifNotNil: [ (frameMillis / 120.0) min: 1.0 max: 0.0 ]) ifNil: [ 0.0 ].
		extra :=  ((latency*0.3) min:0.3 max:0.00).
		clr := Color r: 0.07 + extra g: 0.12 b: 0.12.
	] ifFalse: [
		clr := Color transparent
	].

	(QColors sameColor: renderHealth textColor as: clr) ifFalse: [
		renderHealth textColor: clr.
	].! !

!QControlBar methodsFor: 'private' stamp: 'bgf 7/11/2009 00:38'!
updateNetHealthLabel: latMillis
	"The network health label.  Color it reddish as latency grows;
	 beyond the threshold, change the label to indicate Slow."
	| latency extra clr |

	self numForums > 0 ifTrue: [
		| lbl |
		(latMillis isNil or:[ latMillis < 1]) ifTrue: [
			lbl := 'Network:	?'.
		] ifFalse: [
			lbl :=  'Network:	',
				( latMillis < (self slowLatencyThreshold) ifTrue: [ 'OK' ] ifFalse: [ 'Slow' ] ).
		].
		lbl = netHealth label ifFalse: [ netHealth label: lbl ].
		latency := (latMillis ifNotNil: [ (latMillis / 500.0) min: 1.0 max: 0.0 ]) ifNil: [ 0.0 ].
		extra :=  ((latency*0.3) min:0.3 max:0.00).
		clr := Color r: 0.07 + extra g: 0.12 b: 0.12.
	] ifFalse: [
		clr := Color transparent
	].

 	(QColors sameColor: netHealth textColor as: clr) ifFalse: [
		netHealth textColor: clr.
	].! !

!QControlBar methodsFor: 'private' stamp: 'bgf 1/6/2010 16:20'!
updateServiceCount
	"Update the service count indicator value."
	"At the moment we have to go a bit low-level to do this."
	| v |
	owner session ifNotNilDo: [ : sn |
		v := '', (owner historyPane activeList size).
	] ifNil: [
		v := ''.
	].
	"Check first using equality (rather than identity) to avoid spurius UI-invalidation."
	v ~= svcCount value ifTrue: [ svcCount value: v ].

! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/1/2008 22:49'!
userIsIdle
	"Answer the userIsIdle of the receiver"
	<bewareOf: #userIsIdleChanged>
	^self propertyValueAt: #userIsIdle! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 2/1/2008 22:49'!
userIsIdle: aValue
	"Modify the receiver's userIsIdle"
	^self propertyValueAt: #userIsIdle put: aValue with: #userIsIdleChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:46'!
view
	"Answer the view of the receiver"
	<bewareOf: #viewChanged>
	^self propertyValueAt: #view! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 12/17/2006 10:46'!
view: aValue
	"Modify the receiver's view"
	^self propertyValueAt: #view put: aValue with: #viewChanged! !

!QControlBar methodsFor: 'accessing' stamp: 'bgf 8/17/2008 11:39'!
windowColor: aColor
	"Do nothing - we don't want the inherited border-color clobber."! !

CPlayer subclass: #QNavPanel
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QNavPanel commentStamp: '<historical>' prior: 0!
A Tweak panel for avatar control in a Croquet harness interface.
The panel provides pushbutton control for motion and orientation of the avatar.

The navPanel appears in the interface as a floating panel of height 64.
It is best to make it invisible when
initially created, and make it visible only after Croquet is ready (because
it is rendered as an openGL overlay, and the preview looks quite rough.)

Beyond the usual Tweak panel creation and management, 
the other client responsibilities are:

  - Call owner: participantUI during the main participant costume setup.
    This allows the navigation panel
    to establish event-handlers for various avatar and space change events.


!

QNavPanel defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="owner" toSet="owner:" changeEvent="ownerChanged">owner</field>
	<field toGet="controller" toSet="controller:" changeEvent="controllerChanged">controller</field>
	<field toGet="compassRose" toSet="compassRose:" changeEvent="compassRoseChanged">compassRose</field>
	<field toGet="compassArrow" toSet="compassArrow:" changeEvent="compassArrowChanged">compassArrow</field>
	<field toGet="yawControl" toSet="yawControl:" changeEvent="yawControlChanged">yawControl</field>
	<field toGet="resetPosition" toSet="resetPosition:" changeEvent="resetPositionChanged">resetPosition</field>
	<field toGet="lookDown" toSet="lookDown:" changeEvent="lookDownChanged">lookDown</field>
	<field toGet="lookUp" toSet="lookUp:" changeEvent="lookUpChanged">lookUp</field>
	<field toGet="moveBack" toSet="moveBack:" changeEvent="moveBackChanged">moveBack</field>
	<field toGet="moveForward" toSet="moveForward:" changeEvent="moveForwardChanged">moveForward</field>
	<field toGet="moveLeft" toSet="moveLeft:" changeEvent="moveLeftChanged">moveLeft</field>
	<field toGet="moveRight" toSet="moveRight:" changeEvent="moveRightChanged">moveRight</field>
	<field toGet="moveUp" toSet="moveUp:" changeEvent="moveUpChanged">moveUp</field>
	<field toGet="moveDown" toSet="moveDown:" changeEvent="moveDownChanged">moveDown</field>
	<field toGet="resetLook" toSet="resetLook:" changeEvent="resetLookChanged">resetLook</field>
	<field toGet="resetNorth" toSet="resetNorth:" changeEvent="resetNorthChanged">resetNorth</field>
	<field toGet="rotateLeft" toSet="rotateLeft:" changeEvent="rotateLeftChanged">rotateLeft</field>
	<field toGet="rotateRight" toSet="rotateRight:" changeEvent="rotateRightChanged">rotateRight</field>
	<field toGet="lastEvent" toSet="lastEvent:" changeEvent="lastEventChanged">lastEvent</field>
	<field toGet="newAngle" toSet="newAngle:" changeEvent="newAngleChanged">newAngle</field>
	<field toGet="mouseOut" toSet="mouseOut:" changeEvent="mouseOutChanged">mouseOut</field>
</fields>'!

!QNavPanel class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/23/2010 12:22'!
collapse
	
	"Shrink to collapsed size"
	| bot |

	mouseOut ifFalse: [ ^ self ].
	bot := self bottom.
	self height: self collapsedHeight.
	self fillColor: (QColors paneDarkerColor).
	self fillAlpha: 0.18.
	self bottom: bot.
	self allElementsDo: [:e | e visible:false].	
	self dropShadowEnabled: false.
	self borderWidth: 1.
	self borderColor: (Color lightGray alpha: 0.15).
! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/23/2010 12:20'!
collapsedHeight

	"Just enough to not distort the corners,
	 and be reasonable accessible as a mouseover target."
	^ 12! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 12:53'!
compassArrow
	"Answer the compassArrow of the receiver"
	<bewareOf: #compassArrowChanged>
	^self propertyValueAt: #compassArrow! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 12:53'!
compassArrow: aValue
	"Modify the receiver's compassArrow"
	^self propertyValueAt: #compassArrow put: aValue with: #compassArrowChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 12:53'!
compassRose
	"Answer the compassRose of the receiver"
	<bewareOf: #compassRoseChanged>
	^self propertyValueAt: #compassRose! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 12:53'!
compassRose: aValue
	"Modify the receiver's compassRose"
	^self propertyValueAt: #compassRose put: aValue with: #compassRoseChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/14/2006 10:47'!
controller
	"Answer the controller of the receiver"
	<bewareOf: #controllerChanged>
	^self propertyValueAt: #controller! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/14/2006 10:47'!
controller: aValue
	"Modify the receiver's controller"
	^self propertyValueAt: #controller put: aValue with: #controllerChanged! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 2/1/2009 09:26'!
expand
	| bot |
	<on: expand>
	"grow to normal size"

	bot := self bottom.
	self height: 67 + (self collapsedHeight//2).
	self bottom: bot.
	self borderWidth: 1.
	self allElementsDo: [:e | e visible:true].
	self dropShadowEnabled: true.
	self fillColor: (QColors innerPaneColor).
	self fillAlpha: 0.7.
	self borderColor: (Color lightGray alpha: 0.3).
	self roundCornerSpec: #(1 2 3 4).
	self comeToFront.! !

!QNavPanel methodsFor: 'initialize' stamp: 'bgf 3/23/2009 15:52'!
initialize
	super initialize.
	mouseOut := true.
	controller := QNavController new.

	compassRose := CImageButton new.
	compassArrow := CImageButton new.
	moveForward := CImageButton new.
	resetPosition  := CImageButton new.
	lookDown  := CImageButton new.
	lookUp  := CImageButton new.

	moveBack  := CImageButton new.
	moveForward  := CImageButton new.
	moveLeft  := CImageButton new.
	moveRight  := CImageButton new.
	moveUp  := CImageButton new.
	moveDown := CImageButton new.

	resetLook  := CImageButton new.
	resetNorth  := CImageButton new.
	rotateLeft  := CImageButton new.
	rotateRight  := CImageButton new.

	controller controls: self.! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 3/1/2008 11:49'!
isExpanded
	"Am I onscreen at full height?"
	^self visible and: [ self height > self collapsedHeight ]! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/15/2006 16:01'!
lastEvent
	"Answer the lastEvent of the receiver"
	<bewareOf: #lastEventChanged>
	^self propertyValueAt: #lastEvent! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/15/2006 16:01'!
lastEvent: aValue
	"Modify the receiver's lastEvent"
	^self propertyValueAt: #lastEvent put: aValue with: #lastEventChanged! !

!QNavPanel methodsFor: 'components' stamp: 'bgf 1/16/2008 23:34'!
layoutButtons

	"Layout the navigation control pane."

	| mid |
	mid := (320@68) // 2.

	lookDown topLeft: mid + (82@1).
	lookUp topLeft:  mid + (82@-25).

	resetPosition topLeft: mid - (12@12).
	moveBack topLeft: mid + (-13@8).
	moveForward topLeft: mid + (-13@-34).
	moveLeft topLeft: mid + (-34@-12).
	moveRight topLeft: mid + (9@-12).

	resetLook topLeft: mid + (48@2).
	resetNorth topLeft: mid + (-76@2).

	rotateLeft topLeft: mid + (-60@-29).
	rotateRight topLeft: mid+ (34@-29).
	moveUp  topLeft: mid+ (116@-24).
	moveDown topLeft: mid+ (116@1).

	compassRose position: 38@(mid y).
	compassArrow position: 38@(mid y).

! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
lookDown
	"Answer the lookDown of the receiver"
	<bewareOf: #lookDownChanged>
	^self propertyValueAt: #lookDown! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
lookDown: aValue
	"Modify the receiver's lookDown"
	^self propertyValueAt: #lookDown put: aValue with: #lookDownChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
lookUp
	"Answer the lookUp of the receiver"
	<bewareOf: #lookUpChanged>
	^self propertyValueAt: #lookUp! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
lookUp: aValue
	"Modify the receiver's lookUp"
	^self propertyValueAt: #lookUp put: aValue with: #lookUpChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 12/22/2006 14:09'!
mouseOut
	"Answer the mouseOut of the receiver"
	<bewareOf: #mouseOutChanged>
	^self propertyValueAt: #mouseOut! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 12/22/2006 14:09'!
mouseOut: aValue
	"Modify the receiver's mouseOut"
	^self propertyValueAt: #mouseOut put: aValue with: #mouseOutChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveBack
	"Answer the moveBack of the receiver"
	<bewareOf: #moveBackChanged>
	^self propertyValueAt: #moveBack! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveBack: aValue
	"Modify the receiver's moveBack"
	^self propertyValueAt: #moveBack put: aValue with: #moveBackChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/14/2006 12:29'!
moveDown
	"Answer the moveDown of the receiver"
	<bewareOf: #moveDownChanged>
	^self propertyValueAt: #moveDown! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/14/2006 12:29'!
moveDown: aValue
	"Modify the receiver's moveDown"
	^self propertyValueAt: #moveDown put: aValue with: #moveDownChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveForward
	"Answer the moveForward of the receiver"
	<bewareOf: #moveForwardChanged>
	^self propertyValueAt: #moveForward! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveForward: aValue
	"Modify the receiver's moveForward"
	^self propertyValueAt: #moveForward put: aValue with: #moveForwardChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveLeft
	"Answer the moveLeft of the receiver"
	<bewareOf: #moveLeftChanged>
	^self propertyValueAt: #moveLeft! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveLeft: aValue
	"Modify the receiver's moveLeft"
	^self propertyValueAt: #moveLeft put: aValue with: #moveLeftChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveRight
	"Answer the moveRight of the receiver"
	<bewareOf: #moveRightChanged>
	^self propertyValueAt: #moveRight! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
moveRight: aValue
	"Modify the receiver's moveRight"
	^self propertyValueAt: #moveRight put: aValue with: #moveRightChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/14/2006 12:28'!
moveUp
	"Answer the moveUp of the receiver"
	<bewareOf: #moveUpChanged>
	^self propertyValueAt: #moveUp! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/14/2006 12:28'!
moveUp: aValue
	"Modify the receiver's moveUp"
	^self propertyValueAt: #moveUp put: aValue with: #moveUpChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/19/2006 12:29'!
newAngle
	"Answer the newAngle of the receiver"
	<bewareOf: #newAngleChanged>
	^self propertyValueAt: #newAngle! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/19/2006 12:29'!
newAngle: aValue
	"Modify the receiver's newAngle"
	^self propertyValueAt: #newAngle put: aValue with: #newAngleChanged! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/19/2006 15:37'!
onCompassDown:evt

	"Mouse down in the compass rose.  Turn toward the heading indicated by the mouse position."
	| ehand |

	<on: mouseDown in: compassRose>

	ehand := evt hand.
	self onCompassDownProportional:evt.
	[ehand redButtonPressed] whileTrue: [
		lastEvent := hand lastEvent transformedBy: compassRose globalToLocal.
		self onCompassDownProportional:lastEvent.
		self waitTick.
	].
! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/19/2006 15:42'!
onCompassDownDirect:evt

	"Mouse down in the compass rose.  Direct immediate variant.
	 Turn to the heading indicated by the mouse position."

	| |

	newAngle := 0 - ((32@32) bearingToPoint: (evt position)).
	self signal: #rotateToNewAngle.
	^ self.
	! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/19/2006 15:42'!
onCompassDownIncremental:evt

	"Mouse down in the compass rose.  Turn (stepwise) toward the heading indicated by the mouse position."

	| hitAngle oldAngle diff |

	hitAngle := (90-(32@32 bearingToPoint:(evt position))).
	hitAngle < 0 ifTrue: [hitAngle := hitAngle + 360].
	oldAngle := compassArrow angle .

	diff := (oldAngle - hitAngle).
	(diff abs < 4) ifTrue: [
		^self.
	].

	lastEvent := evt.
	"Take the shorter path around the compass..."
	((diff > 0 and: [diff < 180]) or: [ diff < -180]) ifTrue: [
		self signal: #rotateRight.
	] ifFalse: [
		self signal: #rotateLeft.
	].! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/19/2006 15:43'!
onCompassDownProportional:evt

	"Mouse down in the compass rose.  Turn toward the heading indicated by the mouse position.
	The further the heading is from our current one, the more we turn."

	| hitAngle oldAngle diff factor |

	factor := 1.5.  "Divisor of turn to take at each tick.  1 gives immediate full rotation."

	hitAngle := 360 - ((32@32) bearingToPoint:(evt position)).
	hitAngle < 0 ifTrue: [hitAngle := hitAngle + 360].
	oldAngle := compassArrow angle - 90.
	oldAngle < 0 ifTrue: [oldAngle := oldAngle + 360.].

	diff := (hitAngle - oldAngle).
	lastEvent := evt.

	(diff abs) < 1 ifTrue: [
		^ self
	].

	"Take the shorter path around the compass..."
	diff > 180 ifTrue: [
		diff := diff - 360. 
	] ifFalse: [
		diff < -180 ifTrue: [
			diff := 360 + diff.
		].
	].

	"unless direct, put a ceiling on the amount of spin in a single increment."
	factor > 1 ifTrue: [
		diff positive ifTrue: [
			diff := diff min: 120.
		] ifFalse: [
			diff := diff max: -120.
		].
	].

	newAngle := oldAngle + (diff/factor).

	self signal: #rotateToNewAngle.
! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/7/2008 14:16'!
onOwnerChanged

	<on: ownerChanged>

	controller owner: (self owner).! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 1/17/2008 00:19'!
onPointerEnter
	"pop up the HUD nav panel"
	<on: mouseEnter>

	mouseOut := false.
	(Delay forMilliseconds: 250) wait.		"Don't spring if the mouse is just passin' through"
	mouseOut ifFalse: [ 
		self expand
		].
! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 3/10/2008 10:28'!
onPointerLeave
	| |
	<on: mouseLeave>

	mouseOut := true.
	(Delay forMilliseconds: 500) wait.
	mouseOut ifTrue: [
		self collapse.
		].! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 6/7/2006 22:48'!
onResetNorth
	<on: mouseDown in: resetNorth>
	self signal: #resetNorth.
		! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 5/20/2007 19:08'!
onResetPosition
	<on: mouseUp in: resetPosition>
	self signal: #resetPosition! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 12/17/2006 13:55'!
owner
	"Answer the owner of the receiver"
	<bewareOf: #ownerChanged>
	^self propertyValueAt: #owner! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 12/17/2006 13:55'!
owner: aValue
	"Modify the receiver's owner"
	^self propertyValueAt: #owner put: aValue with: #ownerChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 14:14'!
pitchAngle: pa

! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
resetLook
	"Answer the resetLook of the receiver"
	<bewareOf: #resetLookChanged>
	^self propertyValueAt: #resetLook! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
resetLook: aValue
	"Modify the receiver's resetLook"
	^self propertyValueAt: #resetLook put: aValue with: #resetLookChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
resetNorth
	"Answer the resetNorth of the receiver"
	<bewareOf: #resetNorthChanged>
	^self propertyValueAt: #resetNorth! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
resetNorth: aValue
	"Modify the receiver's resetNorth"
	^self propertyValueAt: #resetNorth put: aValue with: #resetNorthChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 5/20/2007 19:04'!
resetPosition
	"Answer the resetPosition of the receiver"
	<bewareOf: #resetPositionChanged>
	^self propertyValueAt: #resetPosition! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 5/20/2007 19:04'!
resetPosition: aValue
	"Modify the receiver's resetPosition"
	^self propertyValueAt: #resetPosition put: aValue with: #resetPositionChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
rotateLeft
	"Answer the rotateLeft of the receiver"
	<bewareOf: #rotateLeftChanged>
	^self propertyValueAt: #rotateLeft! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
rotateLeft: aValue
	"Modify the receiver's rotateLeft"
	^self propertyValueAt: #rotateLeft put: aValue with: #rotateLeftChanged! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
rotateRight
	"Answer the rotateRight of the receiver"
	<bewareOf: #rotateRightChanged>
	^self propertyValueAt: #rotateRight! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
rotateRight: aValue
	"Modify the receiver's rotateRight"
	^self propertyValueAt: #rotateRight put: aValue with: #rotateRightChanged! !

!QNavPanel methodsFor: 'initialize' stamp: 'sn 5/8/2011 19:04'!
setupCostume

	| invisible |
	super setupCostume.
	"self sticky := true."  "prevent me from being picked up willy nilly"
	invisible := Color r: 0.0 g: 0.0 b: 0.0 alpha: 0.0.

	self fillColor: QColors paneColor.
	self fillAlpha: 0.6.
	self borderStyle: #simple.
	self roundCorners: true.
	self borderWidth: 1.	"Until we're up and in-use."
	self borderColor: (Color lightGray alpha: 0.1).

	self extent: 320@64.

	compassArrow color: invisible;
		extent: 39@39;
		pivot: 20@20;
		normalImage: (QForms find: 'compassArrow.png');
		pressedImage: (QForms find: 'compassArrow.png');
		overImage: (QForms find:  'compassArrow.png');
		color: (Color white).

	compassArrow layoutFrame: CLayoutFrame new.
	self add: compassArrow.
	compassArrow angle: 1.

	compassRose color: invisible;
		extent: 64@64;
		balloonHelp: ('Indicates the direction you are facing. Click in the rim to turn.' translateMe );
		normalImage: (QForms find: 'compassRose.png');
		pressedImage: (QForms find: 'compassRoseDown.png');
		overImage: (QForms find: 'compassRoseOver.png').
	compassRose layoutFrame: CLayoutFrame new.
	self add: compassRose.

	resetPosition color: invisible;
		balloonHelp: 'Return to the center of the space.' translateMe;
		extent: 24@24;
		normalImage: (QForms find: 'center.png');
		pressedImage: (QForms find: 'centerDown.png');
		overImage: (QForms find: 'centerOver.png').
	resetPosition layoutFrame: CLayoutFrame new.
	self add: resetPosition.

	lookDown color: invisible;
		balloonHelp: 'Look downward.' translateMe;
	  	extent:24@24;
		normalImage: (QForms find: 'lookDown.png');
		pressedImage: (QForms find: 'lookDownDown.png');
		overImage: (QForms find: 'lookDownOver.png').
	lookDown  layoutFrame: CLayoutFrame new.
	self add: lookDown.

	lookUp color: invisible;
	   balloonHelp: 'Look upward.' translateMe;
	   extent:24@24;
	   normalImage:(QForms find:  'lookUp.png');
		pressedImage: (QForms find:'lookUpDown.png');
		overImage: (QForms find: 'lookUpOver.png').
	lookUp  layoutFrame: CLayoutFrame new.
	self add: lookUp.

	moveBack color: invisible;
		balloonHelp: 'Move backward.' translateMe;
		extent: 24@24;
		normalImage:(QForms find: 'moveBack.png');
		pressedImage: (QForms find: 'moveBackDown.png');
		overImage: (QForms find: 'moveBackOver.png').
	moveBack  layoutFrame: CLayoutFrame new.
	self add: moveBack.

	moveForward color: invisible;
		extent: 24@24;
		balloonHelp: 'Move forward.' translateMe;
		normalImage: (QForms find: 'moveForward.png');
		pressedImage: (QForms find: 'moveForwardDown.png');
		overImage: (QForms find: 'moveForwardOver.png').
	moveForward  layoutFrame: CLayoutFrame new.
	self add:moveForward.

	moveLeft color: invisible;
		extent: 24@24;
		balloonHelp: 'Move left.' translateMe;
		normalImage: (QForms find: 'moveLeft.png');
		pressedImage: (QForms find: 'moveLeftDown.png');
		overImage: (QForms find: 'moveLeftOver.png').
	moveLeft layoutFrame: CLayoutFrame new.
	self add: moveLeft.

	moveRight color: invisible;
		extent: 24@24;
		balloonHelp: 'Move right.' translateMe;
		normalImage: (QForms find: 'moveRight.png');
		pressedImage: (QForms find: 'moveRightDown.png');
		overImage: (QForms find: 'moveRightOver.png').
	moveRight  layoutFrame: CLayoutFrame new.
	self add: moveRight.

	resetLook color: invisible;
		extent: 24@24;
		balloonHelp: 'Reset the viewing angle to horizontal.' translateMe;
		normalImage: (QForms find: 'resetLook.png');
		pressedImage: (QForms find: 'resetLookDown.png');
		overImage: (QForms find: 'resetLookOver.png').
	resetLook  layoutFrame: CLayoutFrame new.
	self add: resetLook.

	resetNorth color: invisible;
		extent: 24@24;
		balloonHelp: 'Turn to face north.' translateMe;
		normalImage: (QForms find: 'resetNorth.png');
		pressedImage: (QForms find: 'resetNorthDown.png');
		overImage: (QForms find: 'resetNorthOver.png').
	resetNorth layoutFrame: CLayoutFrame new.
	self add: resetNorth.

	rotateLeft color: invisible;
		balloonHelp: 'Turn to the left.' translateMe;
		extent: 24@24;
		normalImage: (QForms find: 'rotateLeft.png');
		pressedImage: (QForms find: 'rotateLeftDown.png');
		overImage: (QForms find: 'rotateLeftOver.png').
	rotateLeft  layoutFrame: CLayoutFrame new.
	self add: rotateLeft.

	rotateRight color: invisible;
		balloonHelp: 'Turn to the right.' translateMe;
		extent: 24@24;
		normalImage: (QForms find: 'rotateRight.png');
		pressedImage: (QForms find: 'rotateRightDown.png');
		overImage: (QForms find: 'rotateRightOver.png').
	rotateRight  layoutFrame: CLayoutFrame new.
	self add: rotateRight.

	moveUp color: invisible;
		extent: 24@24;
		balloonHelp: 'Move upwards.' translateMe;
		normalImage: (QForms find: 'zoomIn.png');
		pressedImage: (QForms find: 'zoomInDown.png');
		overImage: (QForms find: 'zoomInOver.png').
	moveUp  layoutFrame: CLayoutFrame new.
	self add: moveUp.

	moveDown color: invisible;
		extent: 24@24;
		balloonHelp: 'Move downwards.' translateMe;
		normalImage: (QForms find: 'zoomOut.png');
		pressedImage: (QForms find: 'zoomOutDown.png');
		overImage: (QForms find: 'zoomOutOver.png').
	moveDown  layoutFrame: CLayoutFrame new.
	self add: moveDown.

	self layoutButtons.
! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 16:02'!
stopSpinning
	"legacy protocol."! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2008 14:18'!
view: aCroView
	"Temporary during mainview - mainwindow transition"
	controller view: aCroView.! !

!QNavPanel methodsFor: 'events' stamp: 'bgf 11/2/2006 21:52'!
whileStillDown: evt in:widget signal: event

	| ehand |
	ehand := evt hand.
	lastEvent := evt.
	self signal: event.
	self waitTick; waitTick; waitTick; waitTick.	"Lazy man's tiny measure of hysteresis."

	[ehand redButtonPressed] whileTrue: [
		lastEvent := hand lastEvent transformedBy: widget globalToLocal.
		self signal: event.
		self waitTick.
	].
! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/9/2006 23:02'!
yawAngle: ya

	"Display the given yaw angle as the compass arrow direction."

	compassArrow angle:(90+(ya asInteger)).

		" asInteger so that 0/90 etc are exact 
	       -- improves drawing the angled image of the arrow at the axes."! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
yawControl
	"Answer the yawControl of the receiver"
	<bewareOf: #yawControlChanged>
	^self propertyValueAt: #yawControl! !

!QNavPanel methodsFor: 'accessing' stamp: 'bgf 6/7/2006 16:53'!
yawControl: aValue
	"Modify the receiver's yawControl"
	^self propertyValueAt: #yawControl put: aValue with: #yawControlChanged! !

CPlayer subclass: #QViewPlayer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QViewPlayer commentStamp: '<historical>' prior: 0!
A CPlayer tied to a TViewPane of some sort.

Ensures that the 'view' 
(a TViewPanel, as specified by the app session's viewPaneClass)
is created when non-nil session is supplied,
and removed when the session is set to nil.

Maintains the TViewPanel's bounds to match the player's.

This is an abstract base class; subclasses will need to coordinate
the assignment of space and camera to the viewPane.!

QViewPlayer subclass: #QMainViewPlayer
	instanceVariableNames: ''
	classVariableNames: 'AvatarLabelBorderColor DelayedColor HiliteLabelFill NormalLabelFill'
	poolDictionaries: 'OpenGLConstants'
	category: 'Qwaq-Client'!

!QMainViewPlayer commentStamp: '<historical>' prior: 0!
A Tweak panel holding a croquet viewpane,
and allowing avatar/event management in it.

(TViewPane is not a Tweak player itself; this wraps it as a Player,
currently limited to event dispatch and some housekeeping.)

Basic state:
	owner - the participant UI holding this panel
	view   - the QViewPane I manage (aka the owner's mainView.)!

QMainViewPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="owner" toSet="owner:" changeEvent="ownerChanged">owner</field>
	<field toGet="crosshairs" toSet="crosshairs:" changeEvent="crosshairsChanged">crosshairs</field>
	<field toGet="lastAnimationCheckTime" toSet="lastAnimationCheckTime:" changeEvent="lastAnimationCheckTimeChanged">lastAnimationCheckTime</field>
</fields>'!

!QMainViewPlayer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMainViewPlayer class methodsFor: 'class variables' stamp: 'bgf 5/3/2009 14:00'!
avatarLabelBorderColor
	^ AvatarLabelBorderColor! !

!QMainViewPlayer class methodsFor: 'class variables' stamp: 'bgf 5/3/2009 12:32'!
hiliteAvatarLabelFill
	 ^HiliteLabelFill! !

!QMainViewPlayer class methodsFor: 'class initialization' stamp: 'bgf 5/3/2009 14:30'!
initialize
	" QMainViewPlayer initialize.
	 Avatar decoration constants.  We use instances to avoid genning,
	 and so identity-based checks work, to avoid sending Tweak labels more updates than needed."

	DelayedColor := Color r: 0.04 g:0.42 b:0.02.
	NormalLabelFill := Color gray: 0.95.
	HiliteLabelFill := Color r: 0.94 g: 0.96 b: 0.99.
	AvatarLabelBorderColor := (Color gray alpha: 0.14).! !

!QMainViewPlayer class methodsFor: 'class variables' stamp: 'bgf 7/23/2009 19:45'!
maxAvatarLabels
	"Maximum number of avatar labels to show before we start dropping those further away."
	^ 12! !

!QMainViewPlayer class methodsFor: 'class variables' stamp: 'bgf 5/3/2009 12:27'!
normalAvatarLabelFill
	 ^ NormalLabelFill! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/28/2008 12:03'!
avatar
	^ view ifNotNil: [ view avatar ]! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/28/2008 12:24'!
crosshairs
	"Answer the crosshairs of the receiver"
	<bewareOf: #crosshairsChanged>
	^self propertyValueAt: #crosshairs! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/28/2008 12:24'!
crosshairs: aValue
	"Modify the receiver's crosshairs"
	^self propertyValueAt: #crosshairs put: aValue with: #crosshairsChanged! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'bgf 6/12/2008 04:10'!
destroyView

	"Do nothing - the main view is created explicitly and passed in."
	^ view! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'bgf 12/4/2008 18:02'!
ensureView
	"The main view is created explicitly and passed in."
	view ifNotNil: [ view player: self ].
	^ view! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'bgf 7/25/2009 15:12'!
initialize
	super initialize.
	crosshairs := QInertLabel new.

! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 5/8/2009 02:53'!
isLookMode
	"Look mode is implemented as a task installed in the viewpane,
	 with some support here to integrate it as a right-button behavior."
	^ view notNil and: [ view isLookMode ]! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/14/2009 01:04'!
lastAnimationCheckTime
	"Answer the lastAnimationCheckTime of the receiver"
	<bewareOf: #lastAnimationCheckTimeChanged>
	^self propertyValueAt: #lastAnimationCheckTime! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/14/2009 01:04'!
lastAnimationCheckTime: aValue
	"Modify the receiver's lastAnimationCheckTime"
	^self propertyValueAt: #lastAnimationCheckTime put: aValue with: #lastAnimationCheckTimeChanged! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/10/2009 09:08'!
onBlueButtonDown: evt
	<on: blueButtonDown>

	view ifNotNil:[
		view mouseDown: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/10/2009 09:10'!
onBlueButtonMove: evt
	<on: blueButtonMove>

	view ifNotNil:[
		view mouseMove: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/10/2009 09:08'!
onBlueButtonUp: evt
	<on: blueButtonUp>

	view ifNotNil:[
		view mouseUp: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:01'!
onKeyDown: evt

	<on: keyDown>
	"Indirection for debug-convenience.. events queue properly during halt."
	self signal: #keyboardEvent with: evt.
! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:01'!
onKeyStroke: evt

	<on: keyStroke>
	"Indirection for debug-convenience.. events queue properly during halt."
	self signal: #keyboardEvent with: evt.
! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:01'!
onKeyUp: evt

	<on: keyUp>
	"Indirection for debug-convenience.. events queue properly during halt."
	self signal: #keyboardEvent with: evt.
! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/14/2009 00:57'!
onKeyboardEvent: evt
	"process messages on event queue"
	<on: keyboardEvent>
	view ifNil:[^self].

	"Look mode is a task but also handles keyboard events for seated keyboard-driven look."
	( (view checkLookMode: evt) and: [ view task keyboardLookEvent: evt ] ) ifTrue: [
		^ self
	].

	evt type caseOf: {
		[#keyDown]		->	[
			self setDriveSpeed: evt.
			view keyDown:  (evt copy setPosition: self pointerXY)	.
		].
		[#keyUp]		->	[
			self setDriveSpeed: evt.
			view keyUp:  (evt copy setPosition: self pointerXY)	.
		].
		"XXXX: Yes it's keystroke with lower case s below."
		[#keystroke]		->	[
			view keyStroke:  (evt copy setPosition: self pointerXY)	.
		].
	} otherwise:[
		QLogger logWarning: ('Unknown keyboard event: ', evt) from: self.
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/14/2010 12:38'!
onMouseDown: event2D

	| tsk mode |
	<on: mouseDown>

	hand keyboardFocus: nil. "shake off keyboard focus"
	owner inBenchTest ifTrue:[
			^ owner session ifNotNil: [ owner session endBenchTest]].

	mode := owner controlBar cursorMode.
	view ifNotNil:[
		self setDriveSpeed: event2D.
		mode = #thumbtack ifTrue: [
			| aClass |
			[
				"What kind of annotation?
				 Shift-click to place a thumbtack; default is sticky note."
				aClass := event2D shiftPressed
							ifFalse: [QStickyNote] 
							ifTrue: [QThumbtack].  
				(owner addAnnotation: aClass asDrop: false) ifNotNil: [
					^ self.
				].
			] ensure: [
				owner controlBar resetArrow.
			].
		].

		mode = #editBox ifTrue:[
			tsk := QEditTask new. 
			view pushTask: tsk.
			view fastPick. "force the pick of the object here"
			view mouseDown: (event2D copy setPosition: self pointerXY).
			^ self.
		].

		mode = #paintbrush ifTrue:[
			tsk := QDrawTask new. 
			tsk color: self avatar laserColor.
			tsk stroke: 2.0.
			view pushTask: tsk.
			view fastPick. "force the pick of the object here"
			view mouseDown: (event2D copy setPosition: self pointerXY).
			^ self.
		].

		"Sometimes the mouseUp event that clears the arrow here 
	       doesn't get delivered through to the draw task.  This pops the TTask 
		 that handled a previous drawing stroke; it doesn't change the mode itself. "
		(view task isKindOf: QDrawTask) ifTrue: [
			view popTask.
		].

		view mouseDown: (event2D copy setPosition: self pointerXY)
	]. ! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:35'!
onMouseEnter: evt
	<on: mouseEnter>

	self setDriveSpeed: evt.
	owner audioSession ifNotNilDo: [ :as|as makeCurrent].
	view ifNotNil: [
		view mouseEnter: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:35'!
onMouseLeave: evt
	<on: mouseLeave>
	view ifNotNil: [
		view mouseLeave: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 5/7/2009 17:01'!
onMouseMove: evt
	<on: mouseMove>

	view ifNotNil:[
		self setDriveSpeed: evt.
		view checkLookMode: evt.
		view mouseMove: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:35'!
onMouseUp: evt
	| persistentModes |
	<on: mouseUp>

	persistentModes := {#paintbrush}.

	view ifNotNil:[

		self setDriveSpeed: evt.
		view mouseUp: (evt copy setPosition: self pointerXY).

		"Some modes (painting) remain in force until cleared by the users,
		 others revert after a single mouse stroke."
		(persistentModes includes: owner controlBar cursorMode) ifFalse: [
			owner controlBar resetArrow.
			self mouseCursor: nil.
		].
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 3/7/2009 19:38'!
onOwnerChanged
	<on: ownerChanged>
	owner ifNotNil: [
		"Given the wierd background-cutout way we're rendered,
		 my children aren't visible... so put the crosshairs in my parent."
		owner add: crosshairs after: self
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/25/2009 15:08'!
onReady
	<on: ready in: owner>
	"We can start volunteering, but only after things settle."
	lastAnimationCheckTime := Time totalSeconds + 5.
! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'bgf 7/25/2009 15:10'!
onRenderEnd
	"Protocol - a rendered frame is just complete."! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'bgf 7/25/2009 15:11'!
onRenderStart
	"Protocol - about to render the frame."
	crosshairs visible: (self avatar notNil and: [ self avatar driver isDriving ]).	
	crosshairs position ~= (self center) ifTrue: [
		crosshairs position: (self center + (0@4)); comeToFront.
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/14/2009 01:12'!
onSlowAnimationTick

	"This ticker controls how often we volunteer to upload a slow-animation forum preview.
	 The rate of actual preview upload is determined by the space's #previewInterval"

	<ticking: 0.5>	"Two/second, but the real action is guarded by the second clock,
					 so we can defer by setting the check ahead in time."

	lastAnimationCheckTime ifNil: [ ^ self ].
	Time totalSeconds - lastAnimationCheckTime > 0 ifTrue: [
		"We stagger these to help round-robin them."
		(Delay forMilliseconds: (500 atRandom + 1)) wait.
		owner volunteerSlowAnimationPreview.
		lastAnimationCheckTime := Time totalSeconds.
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/25/2009 15:09'!
onUnready
	<on: unready in: owner>
	"No volunteering when not in space."
	lastAnimationCheckTime := nil.
! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:35'!
onYellowButtonDown: evt
	<on: yellowButtonDown>

	view ifNotNil:[
		view mouseDown: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 7/10/2009 09:15'!
onYellowButtonMove: evt
	<on: yellowButtonMove>

	view ifNotNil:[
		view mouseMove: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:35'!
onYellowButtonUp: evt
	<on: yellowButtonUp>

	view ifNotNil:[
		view mouseUp: (evt copy setPosition: self pointerXY)
	].! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/28/2008 11:40'!
owner
	"Answer the owner of the receiver"
	<bewareOf: #ownerChanged>
	^self propertyValueAt: #owner! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/28/2008 11:40'!
owner: aValue
	"Modify the receiver's owner"
	^self propertyValueAt: #owner put: aValue with: #ownerChanged! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 12:36'!
pickTarget
	"Do a pick at the given mouse point and return the frame-far ref picked."

	^ self pickTargetAt: (self pointerXY).! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 4/26/2008 15:07'!
pickTargetAt: pXY
	"Do a pick at the given mouse point and return the frame-far ref picked."

	| sel |
	(view notNil and: [view ogl notNil]) ifFalse: [ ^ nil ].

	view pointerXY: pXY.
	view pushTask: TPickAll new.
	sel := view fastPickSelection.
	view popTask.
	^ sel ifNotNil: [ sel target ].
! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 6/8/2008 02:15'!
pointerXY

	"The XY mouse position (from the current live hand) in view raster coordinates."

	^ self cursorPoint! !

!QMainViewPlayer methodsFor: 'events' stamp: 'bgf 2/28/2008 11:55'!
setDriveSpeed: event

	"Update the speed settting in the avatar according to the current
	 command-key states.   My avatar's driving speed is normal unless
	 	shift pressed -> go fast
		option or alt pressed -> go slow.

	We don't use Control for slow because it interacts with Control+Up-Down arrow events
	from the mouse wheel emulation... which control POV zooming.
	
	This method just sets the 'shiftSpeed' parameter in the avatar; it doesn't cause motion.
 	Driving may or may not be happening at any given time.

	We catch these modifier changes here,  via the UI 'on<Event>' methods,
      rather than deeper inside the driving, to simplify cases in the driving system.
	(The Mac and Windows VM's differ on how changes to modifier keys are reported;
	 the Mac sends mouseMove (!!) events when shift/cmd/ctl/alt change,
 	 while the Windows VM sends key events for them.
	Meanwhile, the speed settings apply to keyboard, nav panel,
	and mouse-based driving, which have different control paths.)	"

	| isWindows |

	self avatar ifNil: [ ^ self ].
	isWindows := Smalltalk platformName = 'Win32'.

	self avatar shiftSpeed: (
		(event buttons anyMask: (isWindows ifTrue: [64] ifFalse:[32]))	"Mac option, Win Alt"
			ifTrue: [#low]
			ifFalse: [ (event buttons anyMask:8) 	"Shift"
				ifTrue: [#high] 
				ifFalse: [#normal]])			"Not modified."! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'sn 5/8/2011 20:16'!
setupCostume
	super setupCostume.
	self borderWidth: 0.

	crosshairs 
		label: '';
		graphicForm: ( QForms find: 'DrivingCrosshairs.png');
		graphicFit: #rigid;
		fillAlpha: 0;
		borderWidth: 0;
		hResizing: #rigid; vResizing: #rigid;
		extent: (crosshairs graphicForm extent);
		visible: false.

	self add: crosshairs.! !

!QMainViewPlayer methodsFor: 'initialize' stamp: 'bgf 9/29/2009 13:47'!
view: aView
	super view: aView.
	self stopScript: #onRenderStart.
	self stopScript: #onRenderEnd.
	aView ifNotNil: [
		self startScript: #onRenderStart when: { aView . #aboutToRender }.
		self startScript: #onRenderEnd when: { aView . #endOfRender }.
	].! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/12/2008 01:48'!
visible: aBool
	"Override autocreation inherited from QViewPlayer."
	costume visible: aBool! !

!QMainViewPlayer methodsFor: 'accessing' stamp: 'bgf 8/17/2008 11:39'!
windowColor: aColor

	"Do nothing - we don't want the inherited border-color clobber."! !

QViewPlayer defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="session" toSet="session:" changeEvent="sessionChanged">session</field>
	<field toGet="view" toSet="view:" changeEvent="viewChanged">view</field>
	<field toGet="useCutout" toSet="useCutout:" changeEvent="useCutoutChanged">useCutout</field>
	<field toGet="cutoutForm" toSet="cutoutForm:" changeEvent="cutoutFormChanged">cutoutForm</field>
	<field toGet="renderInterval" toSet="renderInterval:" changeEvent="renderIntervalChanged">renderInterval</field>
	<field toGet="preview" toSet="preview:" changeEvent="previewChanged">preview</field>
</fields>'!

!QViewPlayer class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/11/2008 20:37'!
cutoutForm
	"Answer the cutoutForm of the receiver"
	<bewareOf: #cutoutFormChanged>
	^self propertyValueAt: #cutoutForm! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/11/2008 20:37'!
cutoutForm: aValue
	"Modify the receiver's cutoutForm"
	^self propertyValueAt: #cutoutForm put: aValue with: #cutoutFormChanged! !

!QViewPlayer methodsFor: 'initialize' stamp: 'bgf 2/12/2009 14:32'!
destroyView
	view ifNotNil: [
		view destroy.
		view := nil.
	].

	"Scrub any left-over miniview capture."
	self renderInterval ifNotNil: [
		self graphicForm: (self preview)
	].! !

!QViewPlayer methodsFor: 'drawing' stamp: 'bgf 2/12/2009 14:18'!
drawOn: canvas in: drawingBounds

	self costume drawOn: canvas in: drawingBounds.	"Border, etc."
	useCutout ifTrue: [ 
		"Cut a hole through my ancestral background,
		 because the tweak layer is rendered above the open gl layer "
		cutoutForm ifNil: [
			"The cutout form is an all-0 form; drawing it in full paint mode
		 	 causes anything drawn into my canvas before me to be erased to transparent."
			cutoutForm := Form extent: (self costume innerBounds extent) depth: 32
		].
		(view notNil 	
		 and: [ view camera notNil ] 
		 and: [ view space notNil ]
		 and: [ view ogl notNil]) ifTrue: [
			canvas drawImage: cutoutForm 	
				     at: (drawingBounds origin + self borderWidth)
		]].

! !

!QViewPlayer methodsFor: 'initialize' stamp: 'bgf 12/4/2008 17:57'!
ensureView

	"Make sure I have a view if I should, and don't if I shouldn't."

	"How to ask Tweak if I am actually mapped into the screen,
	 and how to get updates when that changes?"
	(session notNil and: [visible] and: [container notNil] and: [container visible])
	 ifTrue: [
		view ifNil: [
			view := session viewPaneClass new.
			view player: self.
		].
		renderInterval ifNotNil: [ view renderInterval: renderInterval ].
		view bounds: (self globalBounds insetBy: (self borderWidth)).
	] ifFalse: [
		self destroyView.
	].
	^ view.! !

!QViewPlayer methodsFor: 'initialize' stamp: 'bgf 3/7/2009 19:23'!
initialize
	super initialize.
	useCutout := false.
! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 1/3/2009 18:38'!
onExtentChanged
	<on: extentChanged>
	
	useCutout ifTrue: [
		cutoutForm := nil.
	].
	view ifNotNil: [
		self fullBounds.
		view bounds: (self globalBounds insetBy: (self borderWidth)).
	].
	self invalidate.		"Redraw my tweak stuff to get the hole-puncher into play."! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 1/3/2009 18:37'!
onFrameChanged
	<on: frameChanged>
	"I have moved; the view has to follow."

	(self visible and: [ view notNil ]) ifTrue: [
		self fullBounds.
		view bounds: (self globalBounds insetBy: (self borderWidth)).
	].
		! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 2/13/2009 21:40'!
onOglDestroyed
	<on: oglDestroyed in: view>

	"If we've been using capture-interval,
	 flush it to avoid confusion, since it will be stale."
	self renderInterval ifNotNil: [
		self graphicForm: (self preview)
	] ifNil: [
		"If I'm not a mini-view, I go blank when OGL goes off."
		self graphicForm: nil.
	].! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 6/12/2008 04:37'!
onSessionChanged
	<on: sessionChanged>

	"I have a view when I have a session and am visible...."
	self ensureView
! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 8/16/2008 21:26'!
onUseCutoutChanged
	<on: useCutoutChanged>

	"The cutout form is an all-0 form; drawing it in full paint mode
 	 causes anything drawn into my canvas before me to be erased to transparent.
	 It is created on demand in drawOn:in:"

	cutoutForm := nil.! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 1/3/2009 18:38'!
onViewChanged
	<on: viewChanged>

	view ifNotNil: [
		self fullBounds.
		view bounds: (self globalBounds insetBy: (self borderWidth)).
	].! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 2/12/2009 14:24'!
preview
	"Answer the preview of the receiver"
	<bewareOf: #previewChanged>
	^self propertyValueAt: #preview! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 3/1/2009 07:08'!
preview: aForm
	"Modify the receiver's preview"
	(view notNil and: [ aForm notNil ] ) ifTrue: [
		view fallbackPreview: aForm 
	].

	"Set my background to the preview, but only if we can do so
	 without flickering the non-preview normal rendered scene. 
	 (Otherwise graphicForm: can induce an invalidate that substitues the preview for something fresher.)"
	(view isNil 
        or: [ view savedSceneForm isNil ] 
	  or: [ view useOGLForSavedScene ]
	  or: [ view ogl isNil] ) 
	ifTrue: [
		self graphicForm: aForm.
		aForm ifNotNil: [ self graphicFit: #stretch ].
	].

	^self propertyValueAt: #preview put: aForm with: #previewChanged! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/14/2008 23:32'!
renderInterval
	"Answer the renderInterval of the receiver"
	<bewareOf: #renderIntervalChanged>
	^self propertyValueAt: #renderInterval! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 1/17/2009 14:52'!
renderInterval: aValue
	"Modify the receiver's renderInterval"

	"Propagate to the live viewpane."
	(aValue notNil and: [ view notNil ]) ifTrue: [
		view renderInterval: aValue.
	].

	^self propertyValueAt: #renderInterval put: aValue with: #renderIntervalChanged! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/11/2008 20:37'!
session
	"Answer the session of the receiver"
	<bewareOf: #sessionChanged>
	^self propertyValueAt: #session! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/11/2008 20:37'!
session: aValue
	"Modify the receiver's session"
	^self propertyValueAt: #session put: aValue with: #sessionChanged! !

!QViewPlayer methodsFor: 'initialize' stamp: 'bgf 3/7/2009 19:23'!
setupCostume
	super setupCostume.
	self userDraw: true.
	self color: QColors paneDarkerColor.
	self hResizing: #spaceFill.
	self vResizing: #spaceFill.
! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/12/2008 01:23'!
useCutout
	"Answer the useCutout of the receiver"
	<bewareOf: #useCutoutChanged>
	^self propertyValueAt: #useCutout! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/12/2008 01:23'!
useCutout: aValue
	"Modify the receiver's useCutout"
	^self propertyValueAt: #useCutout put: aValue with: #useCutoutChanged! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 6/12/2008 01:16'!
view
	"Answer the view of the receiver"
	<bewareOf: #viewChanged>
	^self propertyValueAt: #view! !

!QViewPlayer methodsFor: 'accessing' stamp: 'bgf 12/4/2008 17:59'!
view: aViewPane
	"Modify the receiver's view"
	aViewPane ifNotNil: [
		aViewPane player: self
	].
	^self propertyValueAt: #view put: aViewPane with: #viewChanged! !

!QViewPlayer methodsFor: 'events' stamp: 'bgf 6/21/2008 01:11'!
visible: aBool
	"I have a view when I have a session and am visible,
	 but not otherwise.  (This keeps me cheap when unseen.)"

	super visible: aBool.
	visible ifFalse: [
		session ifNotNil: [
			self destroyView.
		].
	] ifTrue: [
		self ensureView ifNotNilDo: [ : mv |
			session addPane: view.
			self onFrameChanged.
		].
		self comeToFront.	"So the cut-out form trick can work."
	].
! !

!DateAndTime methodsFor: '*Qwaq-Client' stamp: 'bgf 1/30/2009 14:15'!
asFileNameString

	"Handy for dated temp files.... 
	 DateAndTime now asFileNameString '2009-01-30-1307-40' "

	^ String streamContents: [ : aStream |
	
		self printYMDOn: aStream withLeadingSpace: false.
		aStream nextPut: $-.
		aStream
			nextPutAll: (self hour asString padded: #left to: 2 with: $0);
			nextPutAll: (self minute asString padded: #left to: 2 with: $0);
			nextPut: $-;
			nextPutAll: (self second asString padded: #left to: 2 with: $0).
	].

! !

TObject subclass: #QFrameRate
	instanceVariableNames: 'timeData frameData index lastPeriod lastCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QFrameRate commentStamp: '<historical>' prior: 0!
Encapsulates and independent counter for #frameRate.
See #windowSize: and class #newForView:.

See #render360 and QwaqStatsTest for examples.
!

QFrameRate subclass: #QCounterRate
	instanceVariableNames: 'rateCounter lastTick'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QCounterRate commentStamp: 'hrs 11/20/2007 09:53' prior: 0!
A QFrameRate interface for TRateCounter:
* TRateCounter shares data between all uses, so you don't get to use the cool adjustments without messing up other uses. QCounterRate wraps your own individual QFrameRate object (e.g, with your own window sizes and such) around the raw shared data of a TRateCounter.
* TRateCounter will wrap it's raw data after a long while of counting (not just a long while between asking for results). QCounterRate takes care of this.

Usage:
stats := QCounterRate newFor: (TTrafficCounter current outboundCountsFor: VideoTag).
stats windowSize: 5.
stats frameRate.  "=> 35000"
stats frameRate.  "=> 20000, etc"!

!QCounterRate class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QCounterRate class methodsFor: 'instance creation' stamp: 'hrs 11/20/2007 09:55'!
newFor: aTRateCounter
	^self new rateCounter: aTRateCounter! !

!QCounterRate methodsFor: 'accessing' stamp: 'hrs 11/20/2007 10:29'!
frameRate
	"We don't care when the counts happen, only that they were between the previous #frameRate
	and this one.  So we just let the rateCounter do it's thing, and then settle-up when we we're
	asked for a #frameRate."
	| tick increment |
	tick := rateCounter total.
	increment := tick - lastTick.
	"Guard against rollover in case TRateCounter is someday changed to not go into bignum arith.
	(Defensive programming.) Consider counting video traffic, which we will control to max out at
	20Kb/s.   SmallInteger maxVal / 20000.0 / 60 / 60 =>14.9 hours"
	increment < 0 ifTrue: [increment := SmallInteger maxVal + increment].
	lastTick := tick.
	self countFrames: increment.
	^super frameRate.! !

!QCounterRate methodsFor: 'private' stamp: 'hrs 11/20/2007 09:54'!
rateCounter
	^rateCounter! !

!QCounterRate methodsFor: 'private' stamp: 'hrs 11/20/2007 09:57'!
rateCounter: aTRateCounter
	rateCounter:= aTRateCounter.
	lastTick := rateCounter total.! !

!QFrameRate class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QFrameRate class methodsFor: 'instance creation' stamp: 'hrs 10/23/2007 16:36'!
newForView: aCroquetView
	"Answer an instance that automatically counts each render done by aCroquetView.

	Except for the behind-the-scenes script machinery, the instance retains no knowledge
	of aCroquetView.

	Clients must send #stopAllScripts when done, in order to clean up."
	| stats |
	stats := self new.
	stats runScript: #countFrame when: {aCroquetView . #aboutToRender}.
	^stats.! !

!QFrameRate methodsFor: 'infrastructure' stamp: 'hrs 11/20/2007 09:41'!
countFrame
	self countFrames: 1.! !

!QFrameRate methodsFor: 'infrastructure' stamp: 'hrs 11/20/2007 09:40'!
countFrames: n
	"See class newForView:.
	Note that this will go to bignums when the count exceeds SmallInteger maxVal.
	SmallInteger maxVal / 100.0 / 60 / 60 / 24 > 100 days between #frameRate sends @ 100fps."
	self frameCount: self frameCount + n.! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:08'!
frameCount
	^frameData at: index! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:09'!
frameCount: anInteger
	frameData at: index put: anInteger! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:15'!
frameCountWindowed
	^frameData sum
! !

!QFrameRate methodsFor: 'accessing' stamp: 'hrs 10/19/2007 11:07'!
frameRate
	"Answer the number of frames per second counted since:
           	- the previous send of #frameRate, in the simple case, or
  	 	- the time this object was instantiated, if no previous send of #frameRate, or
		- an average of frames per second counted over the last #windowSize of the above periods.
	#windowSize defaults to 1, reducing the last case to the previous two.
	If no time has elapsed in the entire period, answers zero."
	| time |
	time := Time millisecondClockValue.
	lastPeriod := time - self timeCountWindowed.

	lastCount := self frameCountWindowed.

	self nextIndex.
	self frameCount: 0.
	self timeCount: time.

	"Not an error, but not necessarilly meaningful. Maybe Float infinity would be better?
	QwaqStatsTest>>testZeroTime expects zero."
	lastPeriod isZero ifTrue: [^0.0].	
	^lastCount * 1000.0 / lastPeriod! !

!QFrameRate methodsFor: 'initialize' stamp: 'hrs 10/19/2007 10:20'!
initialize
	super initialize.
	self reset.
! !

!QFrameRate methodsFor: 'accessing' stamp: 'hrs 10/19/2007 10:29'!
lastFrameCount
	"Break out the counts part of the last #frameRate result.
	That's total counts in the window."
	^lastCount! !

!QFrameRate methodsFor: 'accessing' stamp: 'hrs 10/19/2007 10:30'!
lastPeriodInMilliseconds
	"Break out the time part of the last #frameRate result.
	That's total period in the window."
	^lastPeriod! !

!QFrameRate methodsFor: 'accessing' stamp: 'hrs 10/19/2007 10:31'!
lastPeriodInSeconds
	^self lastPeriodInMilliseconds / 1000.0! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:30'!
load: anInteger from: oldData
	| next |
	next := Array new: anInteger.
	1 to: anInteger do: [:x |
		next at: x put: (oldData atWrap: index + x).
		].
	^next! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:14'!
nextIndex
	index := index \\ timeData size + 1.
! !

!QFrameRate methodsFor: 'initialize' stamp: 'hrs 10/19/2007 10:20'!
reset
	lastCount := lastPeriod := 0.
	frameData := Array with: 0.
	timeData := Array with: Time millisecondClockValue.
	index := 1.! !

!QFrameRate methodsFor: 'private' stamp: 'hrs 10/23/2007 16:44'!
stopAllScripts
	self stopScript: #countFrame
! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:09'!
timeCount: aNumber
	timeData at: index put: aNumber! !

!QFrameRate methodsFor: 'private-windowing' stamp: 'hrs 10/18/2007 20:11'!
timeCountWindowed
	^timeData atWrap: index + 1.! !

!QFrameRate methodsFor: 'accessing' stamp: 'hrs 10/19/2007 10:53'!
windowSize
	"See #windowSize:."
	^timeData size.
! !

!QFrameRate methodsFor: 'accessing' stamp: 'hrs 10/19/2007 10:48'!
windowSize: n
	"Have #frameRate use a rolling avarage of the last n #frameRate sends.
	(Instantiation counts as a #frameRate send.)
	The actual window can be less than n if #frameRate has been sent less than N+1 times.

	The default window size is 1, so in normal use we just work on the counts between each 
	successive send to #frameRate.

	Suppose everything is working properly on a heavilly loaded machine:
 	  1. We'll be getting about 10 fps.
        2. We'll be checking #frameRate about 10 times / desired response period = 10x/sec.
	  3. The counter will get an average 1 count on each invocation of #frameRate.
        4. Performance will fluctuate, so some time intervals will get zero counts.
	Oops.

	Specifying #windowSize: 5, for example, will make #frameRate report a rolling average
	over the last five reporting periods, which will give us a little smoother (non-zero) results.

	There are other ways of getting better results, but this class does not (yet) do them.
	For example, we could record the time of the last frame counted from the previous period
	and then only use the time interval from that to the time the last frame was counted in
	the new period."

	| anInteger |
	anInteger := n asInteger. 
	frameData := self load: anInteger from: frameData.
	timeData := self load: anInteger from: timeData.
	index := 1.! !

TObject subclass: #QNavController
	instanceVariableNames: 'owner view controls spaceCompassRotation lastUpdate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QNavController commentStamp: '<historical>' prior: 0!
Implements the event callbacks for a QNavPanel; acts as a scripted relay between the QNavPane events and an underlying CroquetHarness.  

This class need not be used directly; a QNavPanel creates and manages its QNavController.
The controller is shared with the main QwaqCroquetView to coordinate
keyboard and button-driving driving actions; the QwaqCroquetView keyboard driving 
actions are sent through the nav controller so that it can hilite the appropriate nav panel buttons.!

!QNavController class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QNavController methodsFor: 'accessing' stamp: 'bgf 6/15/2006 16:07'!
controlFactor
	"Use keyboard state to return a multiplier against a default motion increment.
	 Return 1 if the event had no keyboard modifiers,
	 10 if the shift key was pressed, 0.1 if the control key was pressed."

	controls lastEvent ifNotNil: [
		^ self controlFactor: controls lastEvent.
	].
	^ 1.0! !

!QNavController methodsFor: 'accessing' stamp: 'bgf 6/15/2006 16:18'!
controlFactor: aCroquetEvent
	"Use keyboard state to return a multiplier against a default motion increment.
	 Return 1 if the event had no keyboard modifiers,
	 10 if the shift key was pressed, 0.1 if the control key was pressed."

	| step |
	step := 1.
	aCroquetEvent shiftPressed ifTrue: [
		step := 5
	].
	aCroquetEvent controlKeyPressed ifTrue: [
		step := 0.2
	].
	
	^ step
! !

!QNavController methodsFor: 'initialize' stamp: 'bgf 3/21/2009 01:02'!
controls: aNavPanel
	"The nav panel I provide logic for."
	controls := aNavPanel.

	self startScript: #onPitch: when: {controls. #pitchAngle}.
	self startScript: #onYaw: when: {controls. #yawAngle}.
	self startScript: #resetNorth when: {controls. #resetNorth}.
	self startScript: #rotateToNewAngle when: {controls. #rotateToNewAngle}.

	"Avatar symbolic speed control" 
	{#moveBack . #moveForward . #moveLeft . #moveRight . 
	  #moveUp . #moveDown . #rotateLeft . #rotateRight . #resetPosition . 
	  #lookDown . #lookUp . #resetLook
	} do: [ :sym |
		"For each, when the corresponding control button is pressed, start the motion of that name;
	       when released, stop the motion of that name.   The nav panel button instance variables
	 	 use the same name as the avatar driving symbols,  so the 'perform:sym' is a lookup
	 	 for the nav panel button that drives the motion indicated by 'sym'.
		 These symbols appear as avatar replica methods (and as entries in QDrivingKeyMap)."

		self startScript: #startMotion: withArguments: {sym} when: {(controls perform:sym) . #mouseDown}.
		self startScript: #stopMotion: withArguments: {sym} when: {(controls perform:sym) . #mouseUp}.
	].! !

!QNavController methodsFor: 'initialize' stamp: 'bgf 12/17/2006 14:09'!
initialize
	super initialize.

	lastUpdate := 0.! !

!QNavController methodsFor: 'events' stamp: 'bgf 3/21/2009 00:50'!
onAvatarChanged

	" on avatarChanged in participant ui "
	self stopScript: #pitchYawRoll:.
	self startScript: #pitchYawRoll: when: {owner avatar . #pitchYawRoll}.
	! !

!QNavController methodsFor: 'events' stamp: 'bgf 10/10/2006 03:23'!
onPitch: p

	view avatar pitch: p.! !

!QNavController methodsFor: 'events' stamp: 'das 11/8/2006 22:36'!
onYaw: y

	view avatar yaw: y - spaceCompassRotation.! !

!QNavController methodsFor: 'initialize' stamp: 'bgf 6/7/2008 14:00'!
owner: aParticipantUI

	owner := aParticipantUI.
! !

!QNavController methodsFor: 'events' stamp: 'bgf 3/1/2008 11:36'!
pitchYawRoll: pyr

	controls isExpanded ifTrue: [
		controls pitchAngle:(pyr at:1).
		controls yawAngle: (pyr at:2) + (spaceCompassRotation ifNil: [0]).
	].! !

!QNavController methodsFor: 'events' stamp: 'bgf 4/5/2008 04:53'!
resetNorth

	view resetToNormalDriving.
	view avatar yaw: 0.
! !

!QNavController methodsFor: 'events' stamp: 'bgf 4/5/2008 04:53'!
rotateToNewAngle

	view resetToNormalDriving.
	view avatar lockYaw: (controls newAngle - spaceCompassRotation)
! !

!QNavController methodsFor: 'events' stamp: 'bgf 12/17/2006 15:00'!
spaceChanged

	| |

	"A space can declare that its internal compass is rotated from 'true';
	 the nav controller will rotate its displays and steering behavior accordingly."

	spaceCompassRotation := view activeSpace get:#compassRotation.
	spaceCompassRotation ifNil: [
		spaceCompassRotation := 0.
	] ifNotNil: [
		spaceCompassRotation := spaceCompassRotation asInteger.
	].
! !

!QNavController methodsFor: 'private' stamp: 'bgf 8/13/2007 11:44'!
startMotion: aSymbol
	"Callback from a keyboard driving key or nav panel drive button;
	 add this to the list of active driving pedals..."
	view addDriveSymbol: aSymbol! !

!QNavController methodsFor: 'private' stamp: 'bgf 8/13/2007 11:44'!
stopMotion: aSymbol
	"Callback from a keyboard driving key or nav panel drive button;
	remove this from the list of active driving pedals..."
	view removeDriveSymbol: aSymbol.! !

!QNavController methodsFor: 'accessing' stamp: 'bgf 11/6/2006 13:30'!
syntheticDown: aControlName forEvent: anEvent

	"Keyboard driving sends it commands by synthetically driving
	 the navigation buttons (so we also get button highlighting."
	(controls perform: aControlName) signal: #mouseDown with: anEvent; state:#pressed.! !

!QNavController methodsFor: 'accessing' stamp: 'bgf 11/6/2006 13:31'!
syntheticUp: aControlName forEvent: anEvent

	"Keyboard driving sends it commands by synthetically driving
	 the navigation buttons (so we also get button highlighting."
	(controls perform: aControlName) state:#normal; signal: #mouseUp with: anEvent.! !

!QNavController methodsFor: 'initialize' stamp: 'bgf 3/21/2009 00:54'!
view: aViewPortal
	
	view := aViewPortal.

	#(onAvatarChanged pitchYawRoll: spaceChanged) do: [ :sel | self stopScript: sel ].

	view ifNil: [ ^ self ].

	self startScript: #onAvatarChanged when: {view . #avatarChanged}.
	self startScript: #pitchYawRoll: when: {view avatar . #pitchYawRoll}.
	self startScript: #spaceChanged when: {view . #activeSpaceChanged}.! !

CMenu subclass: #QMenuCheckables
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QMenuCheckables commentStamp: 'hrs 11/27/2007 15:18' prior: 0!
I am a simple declarative way to specify a group of optional choices in a UI:
   CMenuCheckable new
	 subLabel: 'Foo' action: #fireFoo;
	 subLabel: 'Bar' action: #fireBar;
       label 'Foo Bar Choices' target: anObject container: menu
anObject will be sent #fireFoo when the user checks 'Foo', and similarly for 'Bar.

Other objects can use #checkItem: or #checkItem:value: to keep me up to date. E.g., anObject 
might define:
    fireFoo
        "do something here"
         myCMenuCheckable checkItem: #fireFoo.

#selectItem: also checks the indicated item, but also fires the indicated message to the target. This is useful for clients that have the UI object, but do not have any knowledge of the target.

Frankly, I mostly exist to support QMenuRadios.

The current implementation is a modest extension of CMenu. This means that it can intersperse other items and submenus that are not options to be checked or not checked.!

QMenuCheckables defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="target" toSet="target:" changeEvent="targetChanged">target</field>
</fields>'!

!QMenuCheckables class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMenuCheckables methodsFor: 'accessing' stamp: 'hrs 5/12/2008 17:37'!
checkItem: anActionSymbol
	"Mark an item as checked if the item was previously defined by sending me #sublLabel:action:,
	with anActionSymbol as the second argument."
	self checkItem: anActionSymbol value: nil.! !

!QMenuCheckables methodsFor: 'accessing' stamp: 'hrs 5/12/2008 17:37'!
checkItem: anActionSymbol value: aBooleanOrNil
	"Mark an item as checked or unchecked (based on aBoolean) if the item was previously defined
	by sending me #sublLabel:action:, with anActionSymbol as the second argument.
	A value of nil means to toggle the item."
	self menuItems do: [:item |
		item tag = anActionSymbol ifTrue: [^item checked: (aBooleanOrNil ifNil: [item checked not])].
		].
! !

!QMenuCheckables methodsFor: 'private' stamp: 'hrs 5/12/2008 17:21'!
invokeAction: anActionSymbol
	"anActionSymbol is sent to my target."
	self menuItems do: [:item |
		item tag = anActionSymbol ifTrue: [^target perform: anActionSymbol with: item checked].
		].
	! !

!QMenuCheckables methodsFor: 'construction' stamp: 'hrs 11/27/2007 15:31'!
label: aCapitalizedString target: anObject container: aMenu
	"The group of items I define will be headed with the label aCapitalizedString, and I will be added to aMenu. When objects defined with #subLabel:action: are selected, their associated symbol will be performed on anObject target."
	target := anObject.
	aMenu add: aCapitalizedString subMenu: self.! !

!QMenuCheckables methodsFor: 'accessing' stamp: 'hrs 5/12/2008 17:17'!
selectItem: anActionSymbol
	"Take the action appropriate for the user checking a sub item that was previously added to me using #subLabel:action: with anActionSymbol as the second argument. In particular, the item is marked checked and anActionSymbol is sent to my target."
	self checkItem: anActionSymbol.
	target ifNotNil: [self invokeAction: anActionSymbol].! !

!QMenuCheckables methodsFor: 'construction' stamp: 'hrs 11/27/2007 15:30'!
subLabel: aCapitalizedString action: aSymbol
	"Define an item labeled aCapitalizedString, which, when checked by the user or mentioned in #selectItem:, will have aSymbol sent to my target. 
	In the current implementation -- but not my necessarilly my subclasses -- uses of this method can be interspersed with other CMenu messages such as #addSeparator and #add:subMenu:."
	| item |
	item := self add: aCapitalizedString target: self action: #selectItem: arguments: {aSymbol}.
	item tag: aSymbol.
	^item! !

!QMenuCheckables methodsFor: 'accessing' stamp: 'hrs 11/27/2007 09:12'!
target
	"Answer the target of the receiver"
	<bewareOf: #targetChanged>
	^self propertyValueAt: #target! !

!QMenuCheckables methodsFor: 'accessing' stamp: 'hrs 11/27/2007 09:12'!
target: aValue
	"Modify the receiver's target"
	^self propertyValueAt: #target put: aValue with: #targetChanged! !

QMenuCheckables subclass: #QMenuRadios
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QMenuRadios commentStamp: 'hrs 11/27/2007 15:22' prior: 0!
I am a simple declarative way to specify a group of radio-button like choices.  I differ from CMenuCheckable only in that just one of my sub items can be checked at a time -- selecting one deselects the others.

Created for (and so far only used by) animation choices, other candidates for conversion are:
	Tools -> Point of View
	Tools -> Stereo
	Developers -> Graphics -> Choose GPU
These currently all implement the same visual effect with different coding styles.

!

QMenuRadios defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!QMenuRadios class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMenuRadios methodsFor: 'accessing' stamp: 'hrs 11/27/2007 15:34'!
checkItem: anActionSymbol
	"Check the item previously defined using #subLabel:action: (with anActionSymbol as second argument), and clear the checkmark from all other sub items. The behavior is unspecified if #subLabel:action: had been interspersed with other sub-element-defining messages."
	self menuItems do: [:item |
		item checked: item tag = anActionSymbol.
		].
! !

!QMenuRadios methodsFor: 'private' stamp: 'hrs 5/12/2008 17:20'!
invokeAction: anActionSymbol
	"anActionSymbol is sent to my target."
	target perform: anActionSymbol.! !

CImageButton subclass: #QMenuImageButton
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

QMenuImageButton defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="buttonDownTime" toSet="buttonDownTime:" changeEvent="buttonDownTimeChanged">buttonDownTime</field>
</fields>'!

!QMenuImageButton class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:30'!
baseImageName
	self subclassResponsibility.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 08:56'!
buttonDownTime
	"Answer the buttonDownTime of the receiver"
	<bewareOf: #buttonDownTimeChanged>
	^self propertyValueAt: #buttonDownTime! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 08:56'!
buttonDownTime: aValue
	"Modify the receiver's buttonDownTime"
	^self propertyValueAt: #buttonDownTime put: aValue with: #buttonDownTimeChanged! !

!QMenuImageButton methodsFor: 'utility' stamp: 'hrs 12/2/2008 09:22'!
formFrom: suffixString
	| base ext |
	base := self baseImageName copyUpTo: $..
	ext := self baseImageName copyAfter: $..
	name := base, suffixString, '.', ext.
	^QForms find: name.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 08:52'!
home
	^self componentBinding home! !

!QMenuImageButton methodsFor: 'events' stamp: 'hrs 12/8/2008 08:59'!
menu
	self subclassResponsibility.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:37'!
normalForm
	^self formFrom: 'Off'.! !

!QMenuImageButton methodsFor: 'events' stamp: 'hrs 12/8/2008 08:57'!
onDown
	<on: mouseDown in: self>
	buttonDownTime := Time millisecondClockValue.

	"May fire, may raise menu, depending on press duration."! !

!QMenuImageButton methodsFor: 'events' stamp: 'hrs 12/8/2008 08:57'!
onFire
	<on: fire in: self>
	buttonDownTime = 0 ifTrue: [ ^ self ].	"the menu took over"
	buttonDownTime := 0.

	self selected: (self selected not).	

! !

!QMenuImageButton methodsFor: 'events' stamp: 'hrs 12/8/2008 08:59'!
onLeave
	<on: mouseLeave in: self>
	buttonDownTime := 0.! !

!QMenuImageButton methodsFor: 'events' stamp: 'hrs 12/8/2008 09:05'!
onRightDown
	<on: yellowButtonDown in: self>
	self menu.
! !

!QMenuImageButton methodsFor: 'events' stamp: 'hrs 12/8/2008 08:58'!
onStillDown
	<on: mouseStillDown in: self>

	((buttonDownTime > 0)
	 and: [Time millisecondClockValue > (buttonDownTime + QStyles delayedActionMenuDelay) ])
		ifFalse: [ ^ self ].
	
	"Press has been held long enough to get the delayed menu."
	buttonDownTime := 0.
	self menu.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:39'!
overForm
	^self formFrom: 'Over'.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 08:53'!
owner
	^self home owner! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:37'!
pressedForm
	^self formFrom: 'On'.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:38'!
selectedForm
	^self formFrom: 'On'.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:39'!
selectedOverForm
	^self formFrom: 'OffOver'.! !

!QMenuImageButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:40'!
selectedPressedForm
	^self formFrom: 'Off'.! !

!QMenuImageButton methodsFor: 'initialize' stamp: 'hrs 12/1/2008 16:45'!
setupCostume
	super setupCostume.
	self 
		color: Color transparent;
		extent: 30@30;
		normalImage: self normalForm;
		pressedImage: self pressedForm;
		overImage: self overForm;
		selectedImage: self selectedForm;
		selectedOverImage: self selectedOverForm;
		selectedPressedImage: self selectedPressedForm.! !

QMenuImageButton subclass: #QMicrophoneMenuButton
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

!QMicrophoneMenuButton commentStamp: '<historical>' prior: 0!
An icon for the forum microphone,
with click control, visual feedback about state and activity,
and a popup menu with further control options.

A microphone button can be slaved to a master,
and will echo state changes and activity from the master.!

QMicrophoneMenuButton defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="micLevelDots" toSet="micLevelDots:" changeEvent="micLevelDotsChanged">micLevelDots</field>
	<field toGet="micPTT" toSet="micPTT:" changeEvent="micPTTChanged">micPTT</field>
	<field toGet="master" toSet="master:" changeEvent="masterChanged">master</field>
</fields>'!

!QMicrophoneMenuButton class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QMicrophoneMenuButton methodsFor: 'initialize' stamp: 'hrs 12/8/2008 11:38'!
addDots
	micLevelDots do: [:dot | self add: dot]! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 09:28'!
baseImageName
	^'microphone.png'! !

!QMicrophoneMenuButton methodsFor: 'initialize' stamp: 'hrs 12/8/2008 12:49'!
initialize
	super initialize.
	self define: #micLevelDots as: OrderedCollection new.
	self define: #micPTT as: CPlayer new.
! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 16:49'!
level: numBlips
	"A microphone volume from 0 to 6"

	micLevelDots doWithIndex: [ :dot :index |
		dot visible: (index-1 < numBlips).
 
	].
	self signal: #level with: numBlips! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'bgf 2/8/2009 22:36'!
master
	"Answer the master of the receiver"
	<bewareOf: #masterChanged>
	^self propertyValueAt: #master! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'bgf 2/8/2009 22:36'!
master: aValue
	"Modify the receiver's master"
	^self propertyValueAt: #master put: aValue with: #masterChanged! !

!QMicrophoneMenuButton methodsFor: 'events' stamp: 'jcg 2/17/2010 18:48'!
menu
	
	"Post the press-hold menu for the microphone button."
	| menu |
	self owner audioSession ifNotNilDo: [ :sess |

		menu := CMenu new.
		(menu add: 'Off' target: self owner controlBar microphone action: #micOff)
			checked: self selected not.

		(menu add: 'On Full-Time' target: self owner controlBar microphone action: #micOnFulltime)
			checked: (self selected and: [self owner preferences pressToTalk not]);
			emphasis: (self owner preferences pressToTalk ifFalse: [1] ifTrue: [0]).
	
		(menu add: 'On Push-to-Talk' target: self owner controlBar microphone action: #micOnPressToTalk)
			checked: (self selected and: [self owner preferences pressToTalk]);
			emphasis: (self owner preferences pressToTalk ifTrue: [1] ifFalse: [0]).

		menu addSeparator.
		(menu add: 'Reset Sound' target: self owner action: #resetSound: argument: 'microphone button menu clicked').

		menu addSeparator.
		"Intentionally disabled when self home doesn't have #adjustSound, as in QSoundDialog."
		(menu add: 'Settings...' target: self home action: #adjustSound).

		QStyles installMenuFontInMenu: menu.
		menu openAt: (self center + self home topLeft - (0@90))	
						"90 rather than menu height because the menu doesn't have that figured yet."
	].! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 11:24'!
micLevelDots
	"Answer the micLevelDots of the receiver"
	<bewareOf: #micLevelDotsChanged>
	^self propertyValueAt: #micLevelDots! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 11:24'!
micLevelDots: aValue
	"Modify the receiver's micLevelDots"
	^self propertyValueAt: #micLevelDots put: aValue with: #micLevelDotsChanged! !

!QMicrophoneMenuButton methodsFor: 'mode' stamp: 'hrs 12/8/2008 16:36'!
micOff
	"Go to the off state, and make sure the PTT bar is absent."
	self selected ifTrue: [
		self selected: false	
	].
	self showPttIndicator: false.
	self signal: #micOff.! !

!QMicrophoneMenuButton methodsFor: 'mode' stamp: 'hrs 12/8/2008 16:37'!
micOnFulltime

	self owner controlBar setPressToTalk: false.
	self selected ifFalse: [
		self selected: true.
	].
	self showPttIndicator: false.
	self signal: #micOnFulltime.! !

!QMicrophoneMenuButton methodsFor: 'mode' stamp: 'hrs 12/8/2008 16:37'!
micOnPressToTalk

	self owner controlBar setPressToTalk: true.
	self selected ifFalse: [
		self selected: true.
	].
	self showPttIndicator: true.
	self signal: #micOnPressToTalk.! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 12:49'!
micPTT
	"Answer the micPTT of the receiver"
	<bewareOf: #micPTTChanged>
	^self propertyValueAt: #micPTT! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 12:49'!
micPTT: aValue
	"Modify the receiver's micPTT"
	^self propertyValueAt: #micPTT put: aValue with: #micPTTChanged! !

!QMicrophoneMenuButton methodsFor: 'events' stamp: 'jcg 5/11/2010 15:50'!
onFire
	<on: fire in: self>
	super onFire.
	self owner controlBar microphoneButtonFired: self selected.
! !

!QMicrophoneMenuButton methodsFor: 'events' stamp: 'bgf 2/8/2009 22:43'!
onLevel: level
	<on: level in: master>
	self level: level
! !

!QMicrophoneMenuButton methodsFor: 'slave events' stamp: 'bgf 2/8/2009 22:57'!
onMasterMicOff
	<on: micOff in: master>
	self micOff
! !

!QMicrophoneMenuButton methodsFor: 'slave events' stamp: 'bgf 2/8/2009 22:57'!
onMasterMicOnFulltime
	<on: micOnFulltime in: master>
	self micOnFulltime
! !

!QMicrophoneMenuButton methodsFor: 'slave events' stamp: 'bgf 2/8/2009 22:57'!
onMasterMicOnPressToTalk
	<on: micOnPressToTalk in: master>
	self micOnPressToTalk
! !

!QMicrophoneMenuButton methodsFor: 'slave events' stamp: 'bgf 2/8/2009 22:57'!
onMasterPressToTalkOn: bOn
	<on: pressToTalkOn in: master>
	self pressToTalkOn: bOn
! !

!QMicrophoneMenuButton methodsFor: 'slave events' stamp: 'bgf 2/8/2009 22:57'!
onMasterPressToTalkPressed: bPressed
	<on: pressToTalkPressed in: master>
	self pressToTalkPressed: bPressed
! !

!QMicrophoneMenuButton methodsFor: 'tips' stamp: 'sn 5/13/2011 22:52'!
onMicrophoneEnter
	| |
	<on: mouseEnter in: self>
	
 	self owner audioSession 
		ifNil: [ ^ self ]
		ifNotNilDo: [ :sess|
			self owner statusInfo: (
				self selected 
				ifTrue: [ 
					'Microphone is on.' translateMe,
					(sess isPressToTalk ifTrue: [ 'Hold down control key to talk.' translateMe ] ifFalse: [''])]
				ifFalse: [ 
					'Microphone is off. (Click to start. Press for settings.)' translateMe] ) ].! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 09:26'!
overForm
	^self formFrom: 'OffOver'.! !

!QMicrophoneMenuButton methodsFor: 'events' stamp: 'bgf 8/22/2009 00:03'!
positionDots

	<on: geometryChanged in: self>

	micLevelDots doWithIndex: [:dot :i| 
		dot position: ((self width // 2 -1)@(21 - (3*i))).
	].

	"The dot at 6 shares space with the dot at 5, but is red."
	(micLevelDots at: 6) position: (micLevelDots at: 5) position.! !

!QMicrophoneMenuButton methodsFor: 'mode' stamp: 'hrs 12/9/2008 10:01'!
pressToTalkOn: yesNo
	self showPttIndicator: yesNo.
	self addDots.	"In cause we caught a push-transition."
	self level: 0.		"Clear transient blips if we catch the mic timing badly"
	self signal: #pressToTalkOn with: yesNo! !

!QMicrophoneMenuButton methodsFor: 'mode' stamp: 'bgf 8/22/2009 01:02'!
pressToTalkPressed: yesNo

	"To be called while in press-to-talk mode, to enable actual speech.
	 I.e this is called by the key-down for ptt, and again for the key-up.
	 (or whatever the equivalent gesture is during master keyboard mode, which is as yet unsolved.)"

	yesNo ifTrue: [
		"We only want in-mic metering when sending."
		self addDots.
		self micPTT graphicForm: nil
	] ifFalse: [
		self removeDots.
		self micPTT graphicForm ifNil: [
			self selected ifTrue:[
				self micPTT graphicForm: (QForms find: 'microphonePTTBar.png').
			].
			self level: 0.		"Clear transient blips if we catch the mic timing badly"
		].
	].
	self signal: #pressToTalkPressed with: yesNo.! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 09:26'!
pressedForm
	^self formFrom: 'Over'.! !

!QMicrophoneMenuButton methodsFor: 'initialize' stamp: 'hrs 12/8/2008 11:38'!
removeDots
	micLevelDots do: [:dot | self remove: dot]! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 09:27'!
selectedOverForm
	^self formFrom: 'Over'.! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 09:27'!
selectedPressedForm
	^self formFrom: 'Over'.! !

!QMicrophoneMenuButton methodsFor: 'initialize' stamp: 'bgf 8/22/2009 01:03'!
setupCostume
	| form dot |
	super setupCostume.
		
	form := (QForms find: 'microphoneDot.png').
	1 to: 6 do: [:i|
		dot  := CPlayer new.
		dot borderWidth: 0; borderStyle: #simple;
			graphicForm: form.
		dot extent: 5@3.
		self add: dot.
		micLevelDots add: dot.
		dot visible: true.
 	].
	 (QForms find: 'microphoneDotRed.png') ifNotNilDo: [ :red |
		(micLevelDots at: 6) graphicForm: red; graphicOffset: (1@0).
	].

	micPTT color: self color;
		extent: self extent;
		graphicForm: (QForms find: 'microphonePTTBar.png');
		graphicFit: #rigid.

! !

!QMicrophoneMenuButton methodsFor: 'accessing' stamp: 'hrs 12/8/2008 14:41'!
showPttIndicator: yesNo

	"Enable or disable the push-to-talk crossbar.
	 This is the semi-static state for mode changes (off, always-on, ptt).
	 (Doing the control-press feedback is by setting/removing the graphic form.) "
	yesNo ifTrue: [
	 	micPTT container = self ifFalse: [ self add: micPTT ].
	] ifFalse: [
		self remove:  micPTT.
	].! !

QMenuImageButton subclass: #QSpeakerMenuButton
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-Client'!

QSpeakerMenuButton defineFields: '<?xml  version="1.0" ?>
<fields>
</fields>'!

!QSpeakerMenuButton class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QSpeakerMenuButton methodsFor: 'accessing' stamp: 'hrs 12/1/2008 16:48'!
baseImageName
	^'speaker.png'! !

!QSpeakerMenuButton methodsFor: 'events' stamp: 'jcg 2/17/2010 18:48'!
menu
	
	"Post the press-hold menu for the speaker button."
	| menu |
	self owner audioSession ifNotNilDo: [ :sess |

		menu := CMenu new.
		menu add: 'Reset Sound' target: self owner action: #resetSound: argument: 'speaker button menu clicked'.

		menu addSeparator.
		"Intentionally disabled when self home doesn't have #adjustSound, as in QSoundDialog."
		(menu add: 'Settings...' target: self home action: #adjustSound).

		QStyles installMenuFontInMenu: menu.
		menu openAt: (self center + self home topLeft - (0@40))	
						"40 rather than menu height because the menu doesn't have that figured yet."
	].! !

!QSpeakerMenuButton methodsFor: 'events' stamp: 'jcg 4/30/2010 17:19'!
onFire
	"Someone clicked on us, so mute or unmute the application-wide audio."
	<on: fire>
	super onFire.
	(self owner respondsTo: #isMute:) ifTrue: [
		self owner isMute: (self selected not).   
		self owner playSoundEffect: #blip.
	].
! !

!QSpeakerMenuButton methodsFor: 'tips' stamp: 'sn 5/13/2011 22:59'!
onSpeakerEnter

	<on: mouseEnter in: self>
	
 	self owner audioSession ifNil: [ ^ self ].
	self owner statusInfo: (
		self selected 
			ifTrue: [ 'Audio output is on.' translateMe]
			ifFalse: [ 'Audio output is off.' translateMe]).! !
QwaqParticipantUI initialize!
QActiveXControl initialize!
QDriveKeyMap initialize!
QwaqVersion initialize!
QMainViewPlayer initialize!
